/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
 * \file Decoder.H
 *
 * \brief Definition of Decoder class
 *
 *
 */

#ifndef __Decoder_H__
#define __Decoder_H__

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MotionField.H"
#include "EntropyDecoder.H"
#include "ContextMap.H"
#include "ZeroTree.H"
#include "LoopFilter.H"
#include "RefFrameManager.H"
#include "MLC_Stat.H"
#include "FastDCT.H"
#include "Quantizer.H"
#include "WFrame.H"
#include "IntraPrediction.H"
#include "WienerFilter.H"
#include "Deblocking_Filter.H"

#ifdef SUBSTREAM_MT
#include <pthread.h>
//-----------------------------------------------------------------------------
/*!
 * \brief Arguments given to a superblock-line decoding thread.
 */
//-----------------------------------------------------------------------------
class Decoder;
struct SubstreamThreadArgs
{
  class Decoder *decoder;
  int thr;
  Frame *cur;
};
#endif // SUBSTREAM_MT

#ifdef FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
 * \brief Shared work area, one element per substream.
 */
//-----------------------------------------------------------------------------
class SubstreamWorkInfo
{
public:
  SubstreamWorkInfo() {beingWorked = sb_i = sb_j = sb_avail = 0;}
  ~SubstreamWorkInfo() {};
  
  int beingWorked; // true implies a thread's working on this ss.
  // superblock coords of sb being worked on, or last worked on.
  int sb_i; // Superblock line in image.
  int sb_j; // Superblock column in image.
  int sb_avail; // # subsequent sbs that can be decoded.
};
#endif // FRAME_SUBSTREAMS

//-----------------------------------------------------------------------------
/*!
 * \brief %Decoder class
 *
 * Contains main decoding loop excluding I/O
 */
//-----------------------------------------------------------------------------
class Decoder 
{
  MotionField *mf; //!< Pointer to motion field
#ifdef FRAME_SUBSTREAMS
  EntropyDecoder **edecs; //!< Table of entropy decoders.
  Buffer **substream_bufs; //!< Table of substream buffers.
  
  int frame_width_in_sbs;
  int frame_height_in_sbs;
  SubstreamWorkInfo *ss_work_info; //!< element per substream.
  int ss_work_n_waiting_threads;
  
  int get_ss_work(int threadno, int &ss_no);
  void post_ss_work(int threadno, int ss_no, int nwork);
  void advance_ss_work(int threadno, int ss_no);
  void find_substream_bufs(int first_ss, int n_ss, int bits_offset, int payload_bits, Buffer *frame_buf);
  void init_substream_bufs(Buffer *frame_buf);
#else
  EntropyDecoder *edec; //!< Pointer to entropy decoder
#endif // FRAME_SUBSTREAMS
#ifdef SUBSTREAM_MT
  struct SubstreamThreadArgs *substream_args;
  pthread_mutex_t ss_work_info_mutex; // protects ss_work_info and ss_work_n_waiting_threads.
  sem_t ss_work_waiting_threads;
#endif // SUBSTREAM_MT
  ContextMap *zmap[TWO]; //!< Zero block maps
  LoopFilter *lf; //!< Pointer to loop filter
  WienerFilter *wlf; //!< Pointer to Wiener filter  
  Deblocking_Filter *dbf; //!< Pointer to Deblocking filter  
  AIF_adaptive_filter *saif_half;
  
  Transform *trans[3]; //!< Pointers to transforms (DCT, Alternate intra, Alternate Inter)
  ContextMap *zmap_chroma[TWO][TWO]; ///!< Zero block maps for chroma
  
#ifdef FRAME_SUBSTREAMS
  IntraPrediction **intra_preds; //!< Pointer to intra predictor
#else
  IntraPrediction *intra_pred; //!< Pointer to intra predictor
#endif // FRAME_SUBSTREAMS
  
  WFrame *wcur;
  WFrame *wcur_pred; //!< Pointer to predicted frame
  WFrame *wcur_qpe;
  RefFrameManager *ref_mgr; //!< Pointer to reference frame manager
  
  int decoded_frames; //!< Number of frames decoded so far
  
  int frame_type_stat[4]; //!< Counts of frame type usage
  int total_frames; //!< Number of frames decoded
  int ref_frame_stat[MAX_REF_FRAMES]; //!< Counts of reference frame index usage
  
  char *intra_map; //!< Intra block map
  Quantizer *quant[4]; //!< Array of quantizers (intra/inter and luma/chroma)
  
  void decode_one_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_filtered);
  
  void decodemvs_blk(int frame_type, int i, int j, bool largeIntra);
  
  void decode_intra_luma_blk(int i, int j, int sip_type);
  
  void decode_luma_inter_block(int i, int j, int emodel);
  
  void decode_luma_inter_block_8x8(int i, int j,  int emodel);
  
  void decode_chroma_inter_block(int i, int j, int c, int emodel);
  
  void decode_superblock(int q8, int p8, Frame *cur);
  
  void decode_luma_blocks_8x8(int i, int j, int sip_type, Frame *cur);
  void decode_intra_luma_blk16x16(int i, int j);
  void decode_intra_chroma_blk8x8(int i, int j, Frame *cur);
  void decode_intra_chroma_blk(int i, int j, Frame *cur, int sip_type);
  
public:
  void decode_ss(int ss, Frame *cur); //!< called in order to decode a substream.
#ifdef FRAME_SUBSTREAMS
  EntropyDecoder::ZT_STATS_64 *getzt64(int idx) {return &edecs[0]->zt64[idx];}
  MLC_Stat *mlcs; //!< Motion/Luma/Chroma stats for current coded frame
  void collect_mlc_stats(); // Called at end of decode.
#else
  MLC_Stat mlc; //!< Motion/Luma/Chroma stats for current coded frame
#endif
  
  Decoder(int h, int w);
  ~Decoder();
  
#ifdef DEC_TRANSCODER
  // buf_out added.
  void decode_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_filtered, Buffer *buf, Buffer*buf_out = NULL);
#else
  void decode_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_filtered, Buffer *buf);
#endif
  void reset_frame_type_stat();
  int get_frame_type_stat(int type);
  int get_total_frames();
  
#ifdef VERBOSE_STATS
  void print_stats_summary(FILE *f, int verbose);
#else
  void print_stats_summary(FILE *f);
#endif
  void print_stats(FILE *f);
  
};

#endif

