/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file EntropyModel.H
* \brief Definition of EntropyModel class
*/
//-----------------------------------------------------------------------------

#ifndef __EntropyModel_H__
#define __EntropyModel_H__

#include "Global.H"
#include "SimpleProbEst.H"
#include "FixedProbEst.H"
#include "UnaryProbEst.H"
#include "Buffer.H"

#define MAX_ID_CTX 1000

#ifdef CABAC_SYNC_POINTS
//-----------------------------------------------------------------------------
/*!
* \brief SynchroState class.
*
* Contains saved information related to # symbols and probas during treatment
* of a single line of superblocks.
*/
//-----------------------------------------------------------------------------
class SynchroState
{
public:
  unsigned int mnb[MAX_ID_CTX][MAX_AC_SYNCHRO]; //!< mean number of symbols
  unsigned int p2d[MAX_ID_CTX][MAX_AC_SYNCHRO]; //!< stored probas
  unsigned int s2d[MAX_ID_CTX][MAX_AC_SYNCHRO]; //!< stored probas
};
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Entropy model class
*
* Contains set of estimators used for arithmetic coding.
*/
//-----------------------------------------------------------------------------
class EntropyModel 
{
  double bstat[4]; //!< Bit usage by category
  double bstat_intra_block; //!< Bit usage for mode
  double bstat_refindex_lsb; //!< Bit usage for LSB of reference index
  double bstat_refindex_msb; //!< Bit usage for MSB of reference index
  double bstat_intcomp; //!< Bit usage for intensity compensation parameters
  double bstat_mvpdir; //!< Bit usage for motion vector prediction direction
  double bstat_mvdsign; //!< Bit usage for motion data sign information
  double bstat_mvdzero; //!< Bit usage for signalling zero MVD
  double bstat_mvd; //!< Bit usage for motion vector differences
  
  public:
  enum 
  {
    I_luma = 0, //!< Intra-predicted luma block
    I_chroma = 1, //!< Intra-predicted chroma block
    P_luma = 2,   //!< Inter-predicted luma block
    P_chroma = 3, //!< Inter-predicted chroma block
  };
  int max_idFCtx; //!< maximum number of fixed context 
  int max_idACtx; //!< maximum number of adaptive context (SimpleProbEst or UnaryProbEst)
  SimpleProbEst *p_aCtx[MAX_ID_CTX]; //!< pointer to initialising adaptive context.
  ProbEst *p_fCtx[MAX_ID_CTX]; //!< pointer to initialising fixed context.
#ifdef CABAC_SYNC_POINTS
  SynchroState *synchro_state; //!< collection of probability and # symbol info for a single line of superblocks.
  int synchro_state_allocated; //!< have we allocated our own synchro state?
#endif
  void copyInitialProbasFrom(EntropyModel *cs); //!< copy initial proba settings.
  void printProbas(char const *msg); //!< print proba information.
  // Add name and classes parameters.
  void storeFProbCtx(ProbEst &p, int idCtx, char const *name, unsigned int classes); //!< save pointer to proba object.
  void pushFProbCtx(ProbEst &p, char const *name, unsigned int classes)  { storeFProbCtx(p, ++max_idFCtx, name, classes); }
  void storeAProbCtx(SimpleProbEst &p, int idCtx, char const *name, unsigned int classes); //!< save pointer to proba object.
  void pushAProbCtx(SimpleProbEst &p, char const *name, unsigned int classes)  {   storeAProbCtx(p, ++max_idACtx, name, classes); }
  void storeAProbCtx(UnaryProbEst &p, int idCtx, char const *name, unsigned int classes); //!< save pointer to proba object.
  void pushAProbCtx(UnaryProbEst &p, char const *name, unsigned int classes)  { storeAProbCtx(p, ++max_idACtx, name, classes); }
#ifdef CABAC_SYNC_POINTS
  void resetSynchro(); //!< reset synchro info at start of frame.
  void initSynchro(int iSB, int jSB); //!< initialise synchro info at start of superblock.
  void updateSynchro(int iSB); //!< update synchro info at end of superblock.
#endif

#ifndef MVCOMP_ZERORES
  SimpleProbEst p_mvp[2]; //!< Estimators for motion vector predictor
#else
  SimpleProbEst p_mvp[3]; //!< Estimators for motion vector predictor
#endif
  SimpleProbEst p_mv_zero[NCTX_ZMV]; //!< Estimator for zero MVD indicator
  SimpleProbEst p_refidx_lsb0[LISTS][3]; //!< Estimators for LSB of reference list index
  SimpleProbEst p_refidx_lsb1[LISTS][3]; //!< Estimators for LSB of reference list index
  SimpleProbEst p_refidx_msb[LISTS][3]; //!< Estimators for MSB of reference list index
  SimpleProbEst p_zblk[4][NCTX_ZBLK]; //!< Estimators for zero block indicator
  SimpleProbEst p_DUMMY; //!< Explicit unused estimator.  Line present in prob.dat file, not in stream header.
  SimpleProbEst p_intra_blk; //!< Unused estimator
#ifdef FRAME_SUBSTREAMS
  // There are a number of SimpleProbEst contexts reserved for ztrees, the ztrees
  // found param.ztree[0..NZTREES-1] and param.ztree_64.
#define MAX_ZTREE_CTXS (1000) // (140+126)
  SimpleProbEst p_ztree_pe[MAX_ZTREE_CTXS];
#endif
  SimpleProbEst p_sip_type_b1[NCTX_SIP_TYPE]; //!< Estimator for sip type bit 1
  SimpleProbEst p_sip_type_b2[NCTX_SIP_TYPE]; //!< Estimator for sip type bit 2
  SimpleProbEst p_sip_type_b3[NCTX_SIP_TYPE]; //!< Estimator for sip type bit 3
  SimpleProbEst p_sip_type_b4[NCTX_SIP_TYPE]; //!< Estimator for sip type bit 2
  FixedProbEst p_inter_type; //!< Estimators for mv precision
  SimpleProbEst p_mode[3];    //!< Estimators for prediction mode indicator
  SimpleProbEst p_mode_P1[3];
  SimpleProbEst p_mode_P2[3];

  SimpleProbEst p_mvshare_type_zero[3];
  FixedProbEst p_mvshare_four;
  SimpleProbEst p_mvshare_pred;
  
  SimpleProbEst p_mode_TP1[3];
  SimpleProbEst p_mode_TP2[3];
  SimpleProbEst p_mode_TB1[3];
  SimpleProbEst p_mode_TB2[3];
  
  SimpleProbEst p_prev_intra_pred_mode_flag;  //!< Estimators for intra prediction mode indicator
  SimpleProbEst p_rem_intra_pred_mode_flag[5]; //!< Estimators for remainder intra prediction mode indicator
  SimpleProbEst p_intra_chroma_pred_mode[3]; //!< Estimators for intra chroma prediction mode indicator
  SimpleProbEst p_intra_chroma_pred_mode_1; //!< Estimators for intra chroma prediction mode indicator
  SimpleProbEst p_intra_chroma_pred_mode_2; //!< Estimators for intra chroma prediction mode indicator
  
  // prob for SAIF information
  SimpleProbEst p_AIF_use_all_subpel_positions;
  SimpleProbEst p_AIF_subpel_positions_pattern_zero;
  UnaryProbEst  p_AIF_subpel_positions_pattern;
  FixedProbEst  p_AIF_subpel_positions_pattern_suffix;
  UnaryProbEst  p_AIF_filter_pattern;
  FixedProbEst  p_AIF_filter_pattern_suffix;
  SimpleProbEst p_AIF_diff_qfilter_coef_zero;
  UnaryProbEst  p_AIF_diff_qfilter_coef;
  FixedProbEst  p_AIF_diff_qfilter_coef_sign;
  SimpleProbEst p_AIF_mode_flag1;
  SimpleProbEst p_AIF_mode_flag2;
  
  UnaryProbEst  p_WLF;
  FixedProbEst  p_WLF_suffix;
  SimpleProbEst p_WLF_coef_zero;
  UnaryProbEst  p_WLF_coef;
  FixedProbEst  p_WLF_coef_sign;

  UnaryProbEst  p_Dummy1;
  SimpleProbEst p_Dummy2;
  UnaryProbEst  p_Dummy3;

  FixedProbEst p_dummy;

  FixedProbEst p_tsfm_flag;

  UnaryProbEst p_coeff_level_64[ZTsize64]; //!< Estimators for 8x8 luma coefficient levels 
  UnaryProbEst p_coeff_level_64_CT8[ZTsize64]; //!< Estimators for 8x8 chroma coefficient levels (8x8 transform)
  UnaryProbEst p_coeff_level[4][16]; //!< Estimators for coefficient levels 4x4
  UnaryProbEst p_mv_level[2]; //!< Estimators for MVD levels
  UnaryProbEst p_scale[NCTX_SCALE]; //!< Estimators for scale difference levels
  UnaryProbEst p_offset[2]; //!< Estimators for offset difference levels
  
  FixedProbEst p_coeff_sign[4]; //!< Estimator for coefficient signs
  FixedProbEst p_coeff_suffix[4]; //!< Estimator for coefficient suffix
  FixedProbEst p_mv_suffix; //!< Estimator for MVD suffix
  FixedProbEst p_coeff_suffix_64; //!< Estimator for coefficient suffix 64
  FixedProbEst p_coeff_sign_64; //!< Estimator for coefficient signs 64
  FixedProbEst p_coeff_suffix_64_CT8; //!< Estimator for coefficient suffix 64 (chroma 8x8 transform)
  FixedProbEst p_coeff_sign_64_CT8; //!< Estimator for coefficient signs 64 (8x8 transform)
  FixedProbEst p_offset_suffix; //!< Estimator for offset suffix
  FixedProbEst p_scale_suffix; //!< Estimator for scale suffix
  FixedProbEst p_offset_sign; //!< Estimator for offset signs
  FixedProbEst p_scale_sign; //!< Estimator for scale signs

  FixedProbEst p_mv_sign; //!< Estimator for MVD signs
  FixedProbEst p_frame_type; //!< Estimator for frame type
  FixedProbEst p_reference_idc; //!< Estimator for reference frame flag
  FixedProbEst p_frm_zero; //!< Estimator for zero frame count difference
  FixedProbEst p_framecount_sign; //!< Estimator for sign of frame count difference
  FixedProbEst p_framecount; //!< Estimator for level of frame count difference
  FixedProbEst p_num_of_refs; //!< Estimator for indicating number of references used
  FixedProbEst p_dqp_zero; //!< Estimator for zero delta QP indicator
  FixedProbEst p_dqp_sign; //!< Estimator for sign of delta QP
  FixedProbEst p_deltaqp; //!< Estimator for level of delta QP
#ifdef MVCOMP_ZERORES
  FixedProbEst p_mvcomp_config; //!< Estimator for indicating our mvcomp config choice.
#endif

#ifdef Q_OFFSET
  FixedProbEst p_q_offset_zero[2];
  FixedProbEst p_q_offset8x8_zero[2];
  FixedProbEst p_q_offset_sign[2];
  FixedProbEst p_q_offset8x8_sign[2];
  FixedProbEst p_q_offset[2];
  FixedProbEst p_q_offset8x8[2];
#endif

  FixedProbEst quantization_set_id_value_zero; // cq_start
  FixedProbEst quantization_set_id_value;
  FixedProbEst qmatrix_overwrite_flag[2];
  FixedProbEst qmatrix_value_zero[2];
  FixedProbEst qmatrix_value_sign[2];
  FixedProbEst qmatrix_value[2];
  FixedProbEst qmatrix_list_present_flag[2]; // cq_end

  FixedProbEst p_rpps_type; //!< Estimator for RPPS type
  FixedProbEst p_rpps_id; //!< Estimator for RPPS id
  FixedProbEst p_filter_type; //!< Estimator for interpolation filter type
  FixedProbEst p_ic_disable; //!< Estimator for disabling intensity compensation for reference frames
  FixedProbEst p_round_type; //!< Estimator for rounding type for reference frames
  FixedProbEst p_edge_prediction_flag;  //!< Estimator for edge prediction flag
  FixedProbEst p_threshold_edge;
#ifdef CABAC_SYNC_POINTS
  EntropyModel(SynchroState *share = NULL);
#else
  EntropyModel();
#endif
  ~EntropyModel();

  void read_prob_file(FILE *f); //!< read proba settings from .dat.
  void read_stream_probs(Buffer *buf); //!< read initial probas from coded stream.
  int  write_stream_probs(Buffer *buf); //!< write initial probas to coded stream.

#ifndef MVCOMP_ZERORES
  void reset();
#else
  void reset(int mv_keep_movement = 0);
  void recalc_movement();
#endif
  void reset_counts();
  
#ifdef FRAME_SUBSTREAMS
  void sum_stats_from(EntropyModel *src);
#endif
#ifdef VERBOSE_STATS
  void print_bar(FILE *f, const char *caption, double sum_val, double sum_all);
  void print_by_category_full(FILE *f);
#endif
  void print_by_category();
  void print_stat(FILE *f);
  void print_summary();
  
  double entropy();
  void copy_coding_state(EntropyModel  *cs);
};

#endif // __EntropyModel_H__
