/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
 * \file IntraPrediction.H
 * \brief Definition of IntraPrediction class
 */
//-----------------------------------------------------------------------------

#ifndef __IntraPrediction_H__
#define __IntraPrediction_H__

#include "WFrame.H"
#include "Global.H"

#define DC_PRED_VALUE_LUMA 0
#define DC_PRED_VALUE_CHROMA 0

#define EDGE_ADD_PREC 6

//-----------------------------------------------------------------------------
/*!
 * \brief Main Intra Prediction class
 *
 */
//-----------------------------------------------------------------------------
class IntraPrediction 
{
public:
  enum LumaPredModeIdx
  {
    VERT_PRED_MxN = 0,
    HOR_PRED_MxN = 1,
    TMA_L9_PRED_MxN = 2,
    DC_PRED_MxN = 3,
    DIAG_DOWN_RIGHT_PRED_MxN = 4,
    VERT_RIGHT_PRED_MxN = 5,
    HOR_DOWN_PRED_MxN = 6,
    VERT_LEFT_PRED_MxN = 7,
    HOR_UP_PRED_MxN = 8,
    DIAG_DOWN_LEFT_PRED_MxN = 9,
    TMA_L17_PRED_MxN = 10,
    
    NO_INTRA_PMODE_LUMA_MxN = 11 //!< Number of intra prediction modes for luma block
  };
  
  enum ChromaPredModeIdx
  {
    DC_PRED_C = 0,
    HOR_PRED_C = 1,
    VERT_PRED_C = 2,
    PLANE_PRED_C = 3,
    TMA_L5_PRED_C = 4,
    TMA_L9_PRED_C = 5,
    
    NO_INTRA_PMODE_CHROMA = 6 //!< Number of intra prediction modes for chroma block
  };
  
  int getNumPredModes(int sip_type) { return NO_INTRA_PMODE_LUMA_MxN - (sip_type == SIP8x2 || sip_type == SIP2x8); } //!< Get number of prediction modes
  
#ifdef FRAME_SUBSTREAMS
  int *get_NxN_type() {return NxN_type;}
  int *get_mode_y()   {return mode_y;}
  int *get_mode_u()   {return mode_u;}
#endif
  
  Block mprr_blk[NSIP][NO_INTRA_PMODE_LUMA_MxN]; //!< Array of blocks holding intra prediction samples
  Block mprr_c_blk[2][2][NO_INTRA_PMODE_CHROMA];   //!< Array of blocks holding intra prediction samples (indexed by component:size:mode)
  int use_mode_luma[NO_INTRA_PMODE_LUMA_MxN];    //!< Array of flag indicating if the luma prediction mode is in use
  int use_mode_chroma[2][NO_INTRA_PMODE_CHROMA]; //!< Array of flag indicating if the chroma prediction mode is in use
  
  int avail_up_right_4x4;
  int avail_up_right_8x8;
  int avail_up_right_16x16;
  
  int L_mode_pred_luma[4];                       //!< the predicted mode is stored here, 4 entries to cater for 4 4x4 in 8x8 blocks.
  int tmp_sip_type[2][2];
  int Large_mode_y[16];                              //!< Array holding the luma SIP mode for the 16x16 intra block
  int Large_L_mode_pred_luma[4][4];             //!< the prediciton mode is stored here, 4 entries to cater for 4 4x4 in 8x8 blocks, 4 entries for 4 8x8 blocks in a 16x16 block.
  
#ifdef FRAME_SUBSTREAMS
  // Pass shared tables.
  IntraPrediction(int h, int w, int *NxN_type_given = NULL, int *mode_y_given = NULL, int *mode_u_given = NULL);
#else
  IntraPrediction(int h, int w);
#endif
  ~IntraPrediction();
  
  void reset_mode();
  void set_sip_type(int i, int j, int sip_type, int h8 = 1, int w8 = 1);
  int get_sip_type(int i, int j);
  void reset_sip_type();
  int get_sip_ctx(int i, int j);
  void copy_sip_type_from_NxN_type(int fi, int fj, int ti, int tj, int h8, int w8);
  void copy_sip_type_to_NxN_type(int fi, int fj, int ti, int tj, int h8, int w8);
  void copy_from_mode_y(int* pDes, int i8, int j8, int h8, int w8);
  void copy_to_mode_y(int* pSrc, int i8, int j8, int h8, int w8);
  void copy_from_L_mode_pred_luma(int i8, int j8, int h8, int w8);
  void copy_to_L_mode_pred_luma(int i8, int j8);
  
  void put_mode_y(const int i, const int j, int subblock, int mode);
  void put_mode_y_MxN(int i, int j, int mode, int h, int w);
  int get_mode_y(const int i, const int j, int subblock);
  int get_mode_y(const int i_p, const int j_p);
  void put_mode_u(const int i, const int j, int mode);
  int get_mode_u(const int i, const int j);
  
  void set_L_mode_pred_luma(int count, int sip_type);          //!< compares the L SAD and decides the predicted mode.
  
  void generate_prediction(int i, int j, int sip_type, int pred_mode, Block *blk_pre, WFrame *ref);
  void generate_prediction_chroma(int i, int j, int c, int sip_type, int pred_mode, Block *blk_pre, int qpC, WFrame *ref);

private:
  int m_height4;
  int m_width4;
  int m_height8;
  int m_width8;
  
  int candidate_search_region;     //!< defines the candidate search region for Intra TMA
  int candidate_search_region_intra16x16; //!< defines the candidate search region for Intra TMA
  int num_candidate_averaged;      //!< number of candidates used in the MultiHypothesis
  int candidate_sad_threshold;     //!< defines the threshold for candidate selection
  int *search_area; //!< Array holding samples used for TMA search
  
  int *NxN_type;      //!< Array holding the partition type for the 8x8 intra block
  int *mode_y;        //!< Array holding the luma SIP mode for the 4x4 intra block
  int *mode_u;        //!< Array holding the chroma SIP mode for the 4x4 intra block
#ifdef FRAME_SUBSTREAMS
  int  fields_allocated; // The above NxN_* and mode_* fields allocated by us, or given to us?
#endif
    
  int intrapred_luma_edge(const int i, const int j, int sip_type, Block *blk);
  int check_direction_edge_vector(const int ref_x, const int ref_y, int *dir_x, int *dir_y, int sip_type);
  
//-----------------------------------------------------------------------------
/*!
 * \brief Structure to hold TMA parameters
 *
 */
//-----------------------------------------------------------------------------
  struct tma_param
  {
    int tmp_width, tmp_height;              // template size
    int tmp_blk_width, tmp_blk_height;      // sub block size (2x2, 4x4, 8x8)
    int blk_width, blk_height;              // block size  (16x16, 8x8, 4x4, 2x8, 8x2)
    int search_range;                       // tma search range
    int stride;                             // stride in prediction buffer
    int x, y;                               // search point location
    int sx, sy;                             // scalar for x and y directions
    int is_chroma;                          // is_chroma flag
    int chroma_channel;                     // chroma channel number
    int default_value;                      // default value
    Block* psPred;                          // prediction buffer
    int is_up_right_available;              // up right available flag for the block
    int Lnum;                               // number of pixels in L shape template
    int sip_type;                           // sip type
    int is_print;                           // for debug
  };
  
  tma_param stma_param;
  tma_param atma_base_param[5][2];
  
  void init_tma_param_table();
  void set_tma_param(int sip_type, int mode, int chroma_channel,int is_up_right_available, Block* blk);
  int template_matching(int i, int j, tma_param* ptma_param);
  int get_right_edge(int i, int j, int ii, int jj, tma_param* ptma_param);
  void reset_use_mode_luma();
  
  int intrapred_luma_MxN(const int i, const int j, int sip_type, int predmode, Block *blk);   
  int intrapred_chroma_MxN(const int i, const int j, const int c, int sip_type, const int iqpC, int pred_mode, Block *blk_pre);
  WFrame *cur2;

  void LowPassForLumaMxNPred(short *PredPel, int bheight, int bwidth, int available_top, int available_left, int available_topleft);
#ifdef CHROMA_INTRA_ADAPTIVE_LPF
  void LowPassForChromaPred(short *PredPel, int block_up, int block_left, int bsize, const int iqpC);
#else
  void LowPassForChromaPred(short *PredPel, int block_up, int block_left, int bsize);
#endif // CHROMA_INTRA_ADAPTIVE_LPF
};

#endif // __IntraPrediction_H__
