/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file LoopFilter.H
*
* \brief Definition of LoopFilter class
*
*/
//-----------------------------------------------------------------------------

#ifndef __LoopFilter_H__
#define __LoopFilter_H__

#include "MotionField.H"
#include "ContextMap.H"
#include "Frame.H"
#include "WFrame.H"

//-----------------------------------------------------------------------------
/*!
* \brief Base loop filter class
*
* Provides methods for in-loop filtering
*/
//-----------------------------------------------------------------------------
class LoopFilter 
{
public:
  int height; //!< Height of frame in 4x4 luma block units
  int width; //!< Width of frame in 4x4 luma block units
  int height2; //!< Height of frame in 4x4 chroma block units
  int width2; //!< Width of frame in 4x4 chroma block units
  int *lumaH; //!< Array of filter modes for filtering horizontal luma block boundaries
  int *lumaV; //!< Array of filter modes for filtering vertical luma block boundaries
  int *chromaH[2]; //!< Array of filter modes for filtering horizontal chroma block boundaries
  int *chromaV[2]; //!< Array of filter modes for filtering vertical chroma block boundaries
  
  LoopFilter(int h, int w);
  virtual ~LoopFilter();
  
  void reset(int v);

  //-----------------------------------------------------------------------------
  /*!
  * \brief Determine luma boundary modes
  *
  * \param[in] mf Motion field of current frame
  * \param[in] zmap Zero block map of current frame
  */
  //-----------------------------------------------------------------------------
  virtual void set_luma(MotionField *mf, ContextMap *zmap) = 0;
  
  //-----------------------------------------------------------------------------
  /*!
  * \brief Determine chroma boundary modes
  *
  * \param[in] mf Motion field of current frame
  * \param[in] zmap Zero block map of current frame
  * \param[in] c Chroma component index
  */
  //-----------------------------------------------------------------------------
  virtual void set_chroma(MotionField *mf, ContextMap *zmap, int c) = 0;
  
  //-----------------------------------------------------------------------------
  /*!
  * \brief Filter frame
  *
  * This method assumes that boundary modes have been set for luma and chroma
  * by invoking set_luma and set_chroma.
  *
  * \param[in,out] cur Current frame to be filtered
  */
  //-----------------------------------------------------------------------------
  virtual void filter(Frame *cur) = 0;
  
  static LoopFilter *create(int h, int w, int type);
};

#endif // __LoopFilter_H__
