/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file LoopFilterOvercomplete3.H
*
* \brief Definition of LoopFilterOvercomplete3 class
*
*/
//-----------------------------------------------------------------------------

#ifndef __LoopFilterOvercomplete3_H__
#define __LoopFilterOvercomplete3_H__

#include "Global.H"


#include "LoopFilter.H"
#include "DenoiseTransform.H"
#include "DenoiseThreshold.H"

#define NUM_LOOP_MODES3 9
#define POW_TABLE_LIMIT 128 //!< Size of pow_table
#define NDF_SAMPLE_OFFSET 127 // Ugly! Should be changed to 128

#define INVALID_THRES 255 //!< Invalid threshold identifier
#define INVALID_EXTENT 255 //!< Invalid extent identifier
#define DENOMINATOR 100

enum Loop_Filter_Mode 
{
  DCM_SKIP           =  0,
  DCM_INTRA_QUANT    =  1,
  DCM_PRED_SIGNIF    =  2,
  DCM_PRED_QUANT     =  3,
  DCM_PRED_MOT       =  4,
  DCM_MOT_DISC       =  5,
  DCM_IC_STRONG      =  6,
  DCM_IC_INTERMED    =  7,
  DCM_IC_WEAK        =  8,
  // Note: NUM_LOOP_MODES3 macro defines number of normal modes

  // START OF SPECIAL HANDLING MODES
  DCM_BS_QUARTER = 32,
  DCM_BS_HALF = 64,
  DCM_DENOISE_THIS_BLOCK = 128
};

//-----------------------------------------------------------------------------
/*!
* \brief window class
*
* Carries parameters to identify a window of 2D data in a 1D data structure
*/
//-----------------------------------------------------------------------------
class FilterWindow
{
public:
  int row_begin;
  int row_end;
  int col_begin;
  int col_end;
  int row_offset;
  int col_offset;
};

//-----------------------------------------------------------------------------
/*!
* \brief Non-linear Denoising Filter (NDF) class
*
* This filter uses integer operations to filter a frame in an
* overcomplete transform domain.
*/
//-----------------------------------------------------------------------------
class LoopFilterOvercomplete3 : public LoopFilter 
{
  bool m_use_weighting; //!< Flag to enable weighting of blocks
  bool m_do_two_rounds; //!< Flag to enable 2 rounds of denoising
  bool m_aggressive;
  int m_signif_cnt;     //!< Threshold for number of significant coefficients
  int m_mvd_th; //!< Motion difference threshold
  
  int m_override_strength[NUM_LOOP_MODES3];
  int m_override_extent[NUM_LOOP_MODES3];
  int strength_chroma;
  
  int mode_decision_4x4(PredBlk *mvA, PredBlk *mvB, int ncoeffA, int ncoeffB, int mth, int i, int mvshare_type, int chroma);
  
  static int pow_table[POW_TABLE_LIMIT];

  static int lum_thres_adpround[NUM_LOOP_MODES3];
  static int chr_thres_adpround[NUM_LOOP_MODES3];
  static int lum_thres_adpround_aggr[NUM_LOOP_MODES3];
  static int chr_thres_adpround_aggr[NUM_LOOP_MODES3];
  static int lum_extent[NUM_LOOP_MODES3];
  static int lum_extent_aggr[NUM_LOOP_MODES3];
  static int chr_extent[NUM_LOOP_MODES3];
  static int chr_extent_aggr[NUM_LOOP_MODES3];

  static int lum_threshold_table[NUM_LOOP_MODES3][2]; //!< Luma threshold table for the current QP
  static int lum_extent_table[NUM_LOOP_MODES3]; //!< Luma extent table for the current QP
  static int chr_threshold_table[NUM_LOOP_MODES3][2]; //!< Chroma threshold table for the current QP
  static int chr_extent_table[NUM_LOOP_MODES3]; //!< Chroma extent table for the current QP
  
  int lum_threshold_values[NUM_LOOP_MODES3];
  int lum_extent_values[NUM_LOOP_MODES3];  
  int chr_threshold_values[NUM_LOOP_MODES3];
  int chr_extent_values[NUM_LOOP_MODES3];
  
  static int s_qptable[32];
  
#ifdef DENOISE_MT
  // Element per thread.
  DenoiseTransform **trans_luma_thrd;
  DenoiseTransform **trans_chroma_thrd;
  DenoiseThreshold **thres_luma_thrd;
  DenoiseThreshold **thres_chroma_thrd;
  FilterWindow *m_window_luma_thrd;
  FilterWindow *m_window_chroma_thrd;
#else
  DenoiseTransform *trans_luma;
  DenoiseTransform *trans_chroma;
  DenoiseThreshold *thres_luma;
  DenoiseThreshold *thres_chroma;
  FilterWindow m_window_luma;
  FilterWindow m_window_chroma;
#endif
  
  int **thres_mask;  //!< 2D array of thresholds corresponding to each location in a frame.
  int **effected;    //!< 2D data structure showing the locations of pixels which denoising is allowed to change
#ifdef DENOISE_MT
  struct DataLinesThread
  {
    int *lines[8];
  };
  struct DataLinesThread *m_data_accum_thrd;
  struct DataLinesThread *m_norm_accum_thrd;
#define m_data_accum m_data_accum_thrd[denoise_thrd_no].lines
#define m_norm_accum m_norm_accum_thrd[denoise_thrd_no].lines
  int **m_norm_accum2_thrd;
#define m_norm_accum2 m_norm_accum2_thrd[denoise_thrd_no]
#else
  int *m_data_accum[8];
  int *m_norm_accum[8];
  int *m_norm_accum2;
#endif
#ifdef DENOISE_MT
  int *m_indata;
  int *m_outdata;
#else
  int *m_data;
#endif
  int *m_est;

  int m_frame_width;
  int m_frame_height;
  
#ifdef DENOISE_MT
public:
  void init_filter_plane(unsigned char *img, int /*height*/, int width, int threshold, int *H, int *V, int /*qp*/, bool is_luma);
#endif
  void filter_plane(unsigned char *img, int height, int width, int threshold, int *H, int *V, int qp, bool is_luma);
#ifdef DENOISE_MT
private:
#endif
  void calc_thres_values(int threshold, bool is_luma);
  void calc_extent_values(bool is_luma);
  void set_weights(int bits, DenoiseTransform *t);
  void set_thres_mask(const int *lcH, const int *lcV, int bsize, bool is_luma);
  void skip_transforms(int skip_pattern);
  bool set_thres_extent(int mode, int *thres, int *extent, int *bs_inside, bool is_luma);
  void set_threshold(int k_begin, int k_end, int l_begin, int l_end, int thres);
  void set_effected(int k_begin, int k_end, int l_begin, int l_end);
  void adjust_thres(int scalar);
  void set_thresholds();
  void set_extents();

  void rotate(int **handle, int size);
  void denoise_row(const int *src, const int *est, int src_row, int src_pitch, int col_begin, int col_end, DenoiseTransform *t, DenoiseThreshold *h, int *th);
  void normalize_row(int row, int col_begin, int col_end, const int *orig, int pitch, int *out);

  void set_param_from_config_file();

public:
  LoopFilterOvercomplete3(int h, int w);
  ~LoopFilterOvercomplete3();
  
  virtual void set_luma(MotionField *mf, ContextMap *zmap);
  virtual void set_chroma(MotionField *mf, ContextMap *zmap, int c);
  
  virtual void filter(Frame *cur);
};

#endif // __LoopFilterOvercomplete3_H__
