/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file PredBlk.H
* \brief Definition of PredBlk and PredBlkI classes
*/
//-----------------------------------------------------------------------------

#ifndef __PredBlk_H__
#define __PredBlk_H__

#include "RefFrameManager.H"
#include "Param.H"


//-----------------------------------------------------------------------------
/*!
* \brief Intra prediction data class
*
* Contains intra prediction data of an 8x8 block.
*/
//-----------------------------------------------------------------------------
class PredBlkI 
{
public:
  int mode_luma[2][2]; //!< 4 4x4 intra prediction mode for luma
  int mode_chroma; //!< 4x4 intra prediction mode for u and v 
  int lmode_luma[2][2]; //!< 4 4x4 L-mode prediction mode for luma
  int sip_type; //!< Block type

  int sad; //!< Sum of absolute differences, may be weighted

  int cost[8]; //!< Rate cost for coding block in intra mode. Depends on coding mode of neighbouring blocks.

  void reset_cost(); // sets all parameters to zero
};

//-----------------------------------------------------------------------------
/*!
* \brief Motion data class
*
* Contains motion data of an 8x8 block.
*/
//-----------------------------------------------------------------------------
class PredBlk 
{
public:
  PredBlkI *intra; //!< Pointer to intra prediction data. NULL if not using intra for this block.
  int mv[2][2]; //!< Motion vectors for both lists
#ifdef MV_CODING_ENHANCE
  int mv_norm[2][2]; //!< Normalized motion vectors
#endif
  int ref_index[2]; //!< Reference list indices for both lists
  int gsad; //!< SAD
  int list; //!< Prediction mode (LIST_0, LIST_1, or BI_PRED)
  int DC[2]; //!< DC offset for intensity compensation (for LIST_0 and LIST_1 modes)
  int scale[2]; //!< Scale factor for intensity compensation (for LIST_0 and LIST_1 modes)
  int BP_DC; //!< DC offset for intensity compensation (for BI_PRED mode)
  int BP_scale[2]; //!< Scale factors for intensity compensation (for BI_PRED mode)

  int list8;
  int ref_index8[2];
  int mvshare_size[2];  //!< partition size for motion sharing (0; ver, 1; hor)

#ifdef MVCOMP_ZERORES
  int mv_zero_res;  //!<  1 indicates we should force residue to zero.
#endif
  PredBlk();
  void clear();

  void copy(PredBlk *blk);
  int equal(PredBlk *blk);
  int equal(PredBlk *blk, int cur_list);
  int soft_equal(PredBlk *blk, int th);
#ifdef MVCOMP_ZERORES
  int mv_vector_diff(PredBlk *blk, int list_idx);
  int mv_complete_diff(PredBlk *blk, int list_idx); // include illumination compensation.
#endif
  PredBlk *rescale(RefFrameManager *ref_mgr, PredBlk *blk, int cur_list);

#ifdef MV_CODING_ENHANCE
  void sub(PredBlk *blk, RefFrameManager *ref_mgr);
  void add(PredBlk *blk, RefFrameManager *ref_mgr);
#else
  void sub(PredBlk *blk);
  void add(PredBlk *blk);
#endif

  int equal_pred(PredBlk *blk, int cur_list);

  int equal_mot(PredBlk *blk) const;
  int equal_ic_DC(PredBlk *blk) const;
  int equal_ic_scale(PredBlk *blk) const;

#ifndef MVCOMP_ZERORES
  int get_pred_context(PredBlk *left, PredBlk *top, int target_list);
#else
  int get_pred_context(PredBlk *left, PredBlk *top, int target_list, int extra_pred);
#endif

#ifdef MV_CODING_ENHANCE
  void compute_mv_norm(RefFrameManager *ref_mgr);

#ifdef PMV0_AT_OUT_OF_BOUNDARY
  PredBlk *get_prediction(int i, int j,PredBlk *blk, PredBlk *target, RefFrameManager *ref_mgr);
#else
  PredBlk *get_prediction(PredBlk *blk, PredBlk *target, RefFrameManager *ref_mgr);
#endif

#else
  PredBlk *get_prediction(PredBlk *blk, PredBlk *target);
#endif

#ifdef MVCOMP_ZERORES
  void copy_mvinfo_from(PredBlk *neighbour, int list_idx);
  int get_best_median_refs(PredBlk *left, PredBlk *top, PredBlk *topleft);
  void ref_scale_mvs(RefFrameManager *ref_mgr, PredBlk *curblk, int list_idx);
  int get_temporal_prediction(RefFrameManager *ref_mgr, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft, PredBlk *curblk, int list_idx);
  PredBlk *get_temporal_prediction(RefFrameManager *ref_mgr, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft, PredBlk *curblk);
  int get_best_temporal_refs(RefFrameManager *ref_mgr, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft);
#endif

  int mode();

  int mode_P();
  int mode_B();

  int get_B_scale_sum() { return power2(SAMPLE_ADD_PREC) + (BP_scale[0]); }
  int get_B_scale_diff() { return BP_scale[1] << param.weight_diff_scale; }

  int inbounds(RefFrameManager *ref_mgr, int curr_list, int ref);
  int offset_inbounds(RefFrameManager *ref_mgr, int curr_list, int ref, int mv0, int mv1);
  int inbounds_wp();

  void print();
};

#ifdef MVCOMP_ZERORES
//-----------------------------------------------------------------------------
/*!
* \brief MV predictor set class
*
* Contains a set of MV predictors
*/
//-----------------------------------------------------------------------------
class PredBlkSet
{
public:
    PredBlkSet() {}
    ~PredBlkSet() {}
private:
#ifdef FASTER_MVC
#ifdef SUBSTREAM_MT
  /*static*/ PredBlk predictor[MV_MAX_PRED_INDEX+1]; // working area. // SUBSTREAM_MT remove statics!
#else
  static PredBlk predictor[MV_MAX_PRED_INDEX+1]; // working area.
#endif
#else
  PredBlk predictor[MV_MAX_PRED_INDEX+1]; // working area.
#endif
  PredBlk *predictor_set[MV_MAX_PRED_INDEX+1]; // available predictors.
  PredBlk *collapsed_predictor_set[MV_MAX_PRED_INDEX+1]; // unique predictors.

public:
  PredBlk **get_predictor_set_complete(RefFrameManager *ref_mgr, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft, PredBlk *curblk);
  PredBlk **collapse_predictor_set(int list, int &n_diff_preds);
  void signal_treatment(int &n_diff_preds, int &extra_pred, int list,
                        int t_proba, int t_cnt0, int t_cnt1, int thresh);
  void print(int n_diff_preds, int i, int j);
};
#endif

extern PredBlk zero_mv; //!< Default motion data

#endif
