/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file Quantizer.H
* \brief Definition of Quantizer class
*/
//-----------------------------------------------------------------------------

#ifndef __Quantizer_H__
#define __Quantizer_H__

#include "Global.H"
#include "Transform.H"
#include "EntropyEncoder.H"
#define NQPS 128 //!< Number of valid QP values (0-127)

//-----------------------------------------------------------------------------
/*!
* \brief %Quantizer class
*
* Provides methods for quantizing and inverse quantizing blocks. Also provides
* means to adaptively round coefficients during quantization based on past
* history.
*/
//-----------------------------------------------------------------------------
class Quantizer 
{
  void quant4x4(Block *blk, int sip_type, int parity = -1); 
  void quant8x8(Block *blk, int firstcoeff, int parity = -1); 
  void quant_loop4x4(Block *blk, int sip_type, int parity = -1);
  void quant_rdoq4x4(Block *blk, EntropyEncoder *eenc, int type, double labmda_mode, int context, ZeroTree *ztree, int sip_type, int parity = -1); 
  void quant_rdoq8x8(Block *blk, EntropyEncoder *eenc, int type, double lambda_mode, int context, ZeroTree *ztree, int parity = -1);

  void iquant4x4(Block *blk, int sip_type);
  void iquant8x8(Block *blk);

public:

  int *fwd_scale[NSIP];         //!< Array of scale factors for quantization (for all QPs)
  int *inv_scale[NSIP];    //!< Array of scale factors for inverse quantization (for all QPs)
  int *quant_scale[NSIP];       //!< Array of scale factors for quantization of 4x4 coefficients
  int *iquant_scale[NSIP]; //!< Array of scale factors for inverse quantization of 4x4 coefficients

  int *fwd_scale8x8;            //!< Array of scale factors for quantization (for all QPs) for 8x8
  int *inv_scale8x8;       //!< Array of scale factors for inverse quantization (for all QPs) for 8x8
  int *quant_scale8x8;          //!< Array of scale factors for quantization of 8x8 coefficients
  int *iquant_scale8x8;    //!< Array of scale factors for inverse quantization of 8x8 coefficients
  int offset8x8[4][TX_NCOEFF8]; //!< Array of offsets for quantization rounding of 8x8 coefficients
  int offset16x16[4][TX_NCOEFF16]; //!< Array of offsets for quantization rounding of 8x8 coefficients
  int *qp16x16;                 //!< Array of scale factors for quantization of 8x8 coefficients

  int offset[4][TX_NCOEFF];     //!< Array of offsets for quantization rounding of 4x4 coefficients
  int I_offset[2][4][TX_NCOEFF];//!< Array of offsets for quantization rounding of 4x4 intra coefficients
  int intra_frame; //!< I-Pic indicator

  double *coeff_norm_rdoq[NSIP];     //!< Norm for RDOQ
  double *coeff_norm_rdoq_qp[NSIP];  //!< Norm for RDOQ through all QPs
  double *coeff_norm_rdoq8x8;        //!< Norm for RDOQ8x8
  double *coeff_norm_rdoq_qp8x8;     //!< Norm for RDOQ8x8 through all QPs

#ifdef Q_OFFSET
  int iTheta[2];
  int iTheta8x8[2];
#endif

  int *active_qmatrix_4x4[2]; // cq_start       //!< active 4x4 matrix to be used in quantization
  int *active_qmatrix_8x8[2];                   //!< active 8x8 matrix to be used in quantization
  int *active_qmatrix_16x16[2]; // cq_end       //!< active 16x16 matrix to be used in quantization

  //S.Kato quant offset parameters
  int AIF_offset[4][TX_NCOEFF];     //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_1[4][TX_NCOEFF]; //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_2[4][TX_NCOEFF]; //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_3[4][TX_NCOEFF]; //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_4[4][TX_NCOEFF]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_1[4][TX_NCOEFF]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_2[4][TX_NCOEFF]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_3[4][TX_NCOEFF]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_4[4][TX_NCOEFF]; //!< Array for storage of offsets for quantization rounding for SAIF

  int AIF_offset_8x8[4][TX_NCOEFF8];     //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_1_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_2_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_3_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_4_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_1_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_2_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_3_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_4_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF

  int P_AIF_offset_5[4][TX_NCOEFF];      //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_5[4][TX_NCOEFF];      //!< Array for storage of offsets for quantization rounding for SAIF
  int P_AIF_offset_5_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF
  int B_AIF_offset_5_8x8[4][TX_NCOEFF8]; //!< Array for storage of offsets for quantization rounding for SAIF

  int intra; //!< Intra quantizer flag
  int low_zeromv;  //!< Flag indicating whether the number of zero MV in a frame is smaller than TH
  int reference_idc; //!< Flag indicating the reference IDC of current frame 
  int frame_type;  //!< frame type for quantizer

  void quant(Block *blk, int sip_type, int parity = -1); 
  void quant_loop(Block *blk, int sip_type, int parity = -1);
  void quant_rdoq(Block *blk, EntropyEncoder *eenc, int type, double labmda_mode, int context, ZeroTree *ztree, int sip_type, int parity = -1);

  void iquant(Block *blk, int sip_type);
  
  void rescan_d(double *val, int type1, int type2);

  int update_offset(int offset, int v, int vq, int bound);
  int get_offset_limit();
  int get_offset_index(int intra, int intra_frame, int v, int low_zeromv);
  bool get_update_offset(int v, int vq, int low_zeromv);
  void quant16x16 (Block *blk, int type);
  void iquant16x16(Block *blk); 
  void set_qp16x16(int *qp);
  void set_qp8x8(int *qp);

  void set_qp(int *qp);


  // quant parameter load and save
  void I_load(int bank);
  void I_save(int bank);

#ifdef Q_OFFSET
  void setOffset(int iTheta[2], int iTheta8x8[2]); 
#endif

  void loadmatrix(int emodel, int CbCrFlag, int sip_type, int matrix_type); // cq_start, cq_end

  void load();
  void save();
  void P_save_1();
  void B_save_1();
  void P_save_2();
  void B_save_2();  
  void P_save_all(int i);
  void B_save_all(int i);

  Quantizer(int mode);
  ~Quantizer();

  void read_qtables(FILE *f, int sip_type);
  void read_qtables_8x8(FILE *f);
  
#ifndef READ_QTABLES_FROM_FILE
  void copy_qtables(int *f, int sip_type);
  void copy_qtables_8x8(int *f);  
#endif
};

#endif
