/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*
 * pthreads stuff for windows.
 */
#ifdef WIN32

#include <pthread.h>
#include <process.h>

//-----------------------------------------------------------------------------
/*!
* \brief local_args structure.
*
* Indicates the target function and arguments to call when a new thread is created.
*/
//-----------------------------------------------------------------------------
struct local_args
{
  void *(*startfunc)(void *); //!< target function to call.
  void *real_args; //!< args for target function.
};

//-----------------------------------------------------------------------------
/*!
* \brief Function to translate between __stdcall and __cdecl when launching a thread.
*
* \param[in] local_args function pointer and args of target __cdecl function.
*
* \return 0
*/
//-----------------------------------------------------------------------------
unsigned int __stdcall local_function(void *local_args)
{
  struct local_args *args = (struct local_args *)local_args;
  args->startfunc(args->real_args);
  delete args;
  return 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Create a new thread.
*
* \param[out] tid Thread identifier (handle) stored here.
* \param[in]  attr Thread attributes (not supported).
* \param[in]  startfunc The function which the new thread will call.
* \param[in]  arg Arguments to give to the called function.
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int pthread_create(pthread_t *tid,
                   const pthread_attr_t *attr,
                   void *(*startfunc)(void *),
                   void *arg)
{
  if (tid == NULL || attr != NULL || startfunc == NULL)
    return -1;
  // We must apply a wrapper to convert our __cdecl parameter to a __stdcall required
  // by _beginthreadex.
  struct local_args *args = new local_args;
  args->startfunc = startfunc;
  args->real_args = arg;

  *tid = (HANDLE)_beginthreadex(NULL, 0, local_function, args, 0, NULL);
  // check for _beginthread and _beginthreadex bad returns in case we change stuff.
  return (long)*tid == -1L || *tid == NULL ? -1 : 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Wait for a thread to terminate.
*
* The thread identifier is no longer valid after this function returns.
*
* \param[in]  tid Identifies the thread to wait for.
* \param[in]  resultptr Thread function result stored here (not supported).
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int pthread_join(pthread_t tid, void **resultptr)
{
  if (resultptr != NULL)
    return -1; // not supported.
  int result = WaitForSingleObject(tid, INFINITE) == WAIT_OBJECT_0 ? 0 : -1;
  if (!CloseHandle(tid))
    result = -1;
  return result;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Create a new thread-local-storage key.
*
* \param[out] key_result The resulting key for pthread_{get,set}specific.
* \param[in]  destructor Function to call when key is destroyed (not supported).
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int pthread_key_create(pthread_key_t *key_result, void (*destructor)(void *))
{
  if (key_result == NULL)
    return -1;
  if (destructor != NULL)
    return -1; // not supported.
  *key_result = TlsAlloc();
  return *key_result == TLS_OUT_OF_INDEXES ? -1 : 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Return current value assigned to a thread-local-storage key.
*
* \param[in] key The thread-local-storage key for which to return a value.
*
* \return The void * generic value of the key.
*/
//-----------------------------------------------------------------------------
void *pthread_getspecific(pthread_key_t key)
{
  return TlsGetValue(key);
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Assign a value to a thread-local-storage key.
*
* \param[in] key The thread-local-storage key for which to set a value.
* \param[in] value The value to assign to the key.
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int pthread_setspecific(pthread_key_t key, void *value)
{
  return TlsSetValue(key, value) ? 0 : -1;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Create a new mutex.
*
* \param[out] mutex Identifies the new mutex.
* \param[in]  attr Mutex attributes (not supported).
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *attr)
{
  if (mutex == NULL)
    return -1;
  if (attr != NULL)
    return -1; // not supported.
  *mutex = CreateMutex(NULL, FALSE, NULL);
  return *mutex == NULL ? -1 : 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Delete an existing mutex.
*
* \param[in] mutex Identifies the mutex.
*
* The mutex identifier is no longer valid when this function returns.
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
  if (mutex == NULL)
    return -1;
  return CloseHandle(*mutex) ? 0 : -1;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Exclusively lock a mutex, waiting if necessary.
*
* \param[in] mutex Identifies the mutex.
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int pthread_mutex_lock(pthread_mutex_t *mutex)
{
  if (mutex == NULL)
    return -1;
  return WaitForSingleObject(*mutex, INFINITE) == WAIT_OBJECT_0 ? 0 : -1;
}

//-----------------------------------------------------------------------------
/*!
* \brief POSIX threads API support.  Release a locked mutex.
*
* \param[in] mutex Identifies the mutex.
*
* \return 0 on success, -1 on failure.  No errno is set here on failure.
*/
//-----------------------------------------------------------------------------
int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  if (mutex == NULL)
    return -1;
  return ReleaseMutex(*mutex) ? 0 : -1;
}


#endif
