/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ArithDecoder.cpp
*
* \brief Implementation of ArithDecoder class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "ArithDecoder.H"
#include "Global.H"
#include "Param.H"
#ifdef TRACE_DEBUG
#include "Trace.H"
#endif
//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Allocates memory for and initializes ArithDecoder::log_t array.
*/
//-----------------------------------------------------------------------------
ArithDecoder::ArithDecoder() 
{
  int i;
  
  log_t = new double[power2(AC_BITS)];
  
  for (i = ZERO; i < power2(AC_BITS); i++)
  log_t[i] = AC_BITS - log((double)i) / log(2.0);
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory allocated by constructor.
*/
//-----------------------------------------------------------------------------
ArithDecoder::~ArithDecoder() 
{
  delete[] log_t;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Initialize decoder
*
* Initializes internal registers and reads initial data from byte buffer. This
* method is invoked before decoding the first symbol of each segment of coded
* data (e.g., each frame or slice).
*
* \param[in] b Pointer to byte buffer containing arithmetic coded data.
* \param[in] b_out Pointer to byte buffer to contain re-encoded output if transcoding.
*/
//-----------------------------------------------------------------------------
#ifdef DEC_TRANSCODER
void ArithDecoder::init(Buffer *b, Buffer *b_out) 
#else
void ArithDecoder::init(Buffer *b) 
#endif
{
  buf = b;
#ifdef DEC_TRANSCODER
  buf_out = b_out;
#endif
  
  R = power2(AC_BITS) - TWO;
  
  V = ZERO;
  V |= buf->read_word() << 16;
  V |= buf->read_word();
  
  bits = TWO * BITS_IN_BYTE - ONE;
}

#ifdef DEC_TRANSCODER
//-----------------------------------------------------------------------------
/*!
* \brief Indicate a specific context in which to re-encode symbols.
*
* When multiple re-encoding contexts are possible (ie, in the transcoder)
* a specific context in which to encode symbols can be indicated.  This is
* normally used to indicate '0' when decoding the trailing frame syntax
* after the macroblocks and superblocks have been decoded and encoded.
*
* \param[in] thrd The context number to use.
*/
//-----------------------------------------------------------------------------
void ArithDecoder::setEncodeThread(int thrd)
{
    (void)thrd; // unused parameter
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Decode a binary symbol
*
* Decodes a binary symbol using a probability estimate given by context ctx.
* Updates the probability estimate according to the decoded value.
*
* \note For decoding speed improvement the update of the probability estimate
* is incorporated into this method. Hence it does not invoke
* SimpleProbEst::update().
*
* \param[in,out] ctx Pointer to probability estimator
* \return Decoded binary symbol
*/
//-----------------------------------------------------------------------------
int ArithDecoder::decode_symbol(SimpleProbEst *ctx)
{
  unsigned int R0, R0s;
  int val;
  int n;
  
  #ifdef TRACE_STATS
  ctx->fbitcount -= log_t[R];
  #endif
#ifdef TRACE_DEBUG
  SVPRINT_DTRACE("R ", R);
  SVPRINT_DTRACE("p ", ctx->p);
#endif
  // compute interval for value 0
  R0 = shift_right_round_us(R * ctx->p, AC_BITS); 
  R0s = R0 << (TWO * BITS_IN_BYTE - ONE);

  if (V < R0s) 
  {
#ifdef CABAC_SYNC_POINTS
    ctx->update(0);
#else
    ctx->p -= shift_right_round(ctx->p - power2(PROB_BITS) + param.ac_exp_limit, param.loss_rate);
#endif
    val = ZERO;
    R = R0;
#ifdef TRACE_STATS
#ifndef CABAC_SYNC_POINTS
    ctx->symcount[ZERO]++;
#endif
    ctx->fbitcount += log_t[R];
#endif
  }
  else 
  {
#ifdef CABAC_SYNC_POINTS
    ctx->update(1);
#else
    ctx->p -= shift_right_round(ctx->p - param.ac_exp_limit, param.loss_rate);
#endif
    val = ONE;
    R = R - R0;
    V = V - R0s;

#ifdef TRACE_STATS
#ifndef CABAC_SYNC_POINTS
    ctx->symcount[ONE]++;
#endif
    ctx->fbitcount += log_t[R];
#endif
  }

  // renormalize  
  if (R < (unsigned int)power2(AC_BITS - ONE)) 
  {
    n = ZERO;
    do 
    {
      R = R + R;
      n = n + ONE;
    } while (R < (unsigned int)power2(AC_BITS - ONE));


    V = V << n;
    bits -= n;
    if (bits < ZERO) 
    {
      V += buf->read_word() << (-bits - ONE);
      bits += TWO * BITS_IN_BYTE;
    }

#ifdef TRACE_STATS
    ctx->bitcount[val] += n;
#endif
  }
#ifdef TRACE_DEBUG
  SVPRINT_DTRACE("val ", val);
  RPRINT_DTRACE;
#endif
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decodes a binary symbol with 50/50 probability
*
* \note Any probability estimate stored in ctx is ignored and assumed to be
* 0.5.
*
* \param[in,out] ctx Pointer to probability estimator
* \return Decoded binary symbol
*/
//-----------------------------------------------------------------------------
int ArithDecoder::decode_symbol_half(FixedProbEst *ctx)
{
  unsigned int R0, R0s;
  int val;
  
  R0 = R + (R & ONE);
  R0s = R0 << (TWO * BITS_IN_BYTE - ONE);
  V = V + V;
  
  if (V < R0s) 
  {
    R = R0;
    val = ZERO;
    #ifdef VERBOSE_STATS
    #ifdef TRACE_STATS
    ctx->symcount[ZERO]++;
    ctx->fbitcount += 1.0;
    #endif
    #endif
  }
  else 
  {
    R = R + R - R0;
    V = V - R0s;
    val = ONE;
    #ifdef VERBOSE_STATS
    #ifdef TRACE_STATS
    ctx->symcount[ONE]++;
    ctx->fbitcount += 1.0;
    #endif
    #endif
  }
  
  if (--bits < ZERO) 
  {
    bits = TWO * BITS_IN_BYTE - ONE;
    V += buf->read_word();
  }
  
  #ifdef VERBOSE_STATS
  #ifdef TRACE_STATS
  ctx->bitcount[val] += 1;
  #endif
  #endif

  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode an unsigned integer
*
* Decodes one or more binary symbols and combines them into an unsigned integer.
* Binarization consists of a prefix and a suffix.
*
* \note The unsigned integer is larger than or equal to 1.
*
* \param[in,out] ctx Pointer to probability estimator used for prefix of binarization
* \param[in,out] ctx_suffix Pointer to probability estimator used for suffix of binarization
* \return Decoded unsigned integer
*/
//-----------------------------------------------------------------------------
int ArithDecoder::decode_symbol_unary(UnaryProbEst *ctx, FixedProbEst *ctx_suffix)
{
  int val, vals;
  int i;
  
  val = ONE;
  
  // Prefix decoding loop
  while (decode_symbol(ctx) == ONE) 
  {
    val += power2(ctx->suffix_size);
    #ifdef TRACE_STATS
    ctx->symcount[ONE] += power2(ctx->suffix_size) - ONE;
    #endif
    ctx->adjust_suffix();
  }
  
  vals = ZERO;
  // Suffix decoding loop  
  for (i = ZERO; i < ctx->suffix_size; i++)
    vals += vals + decode_symbol_half(ctx_suffix);
  
  #ifdef TRACE_STATS
  ctx->symcount[ONE] += vals;
  #endif  
  ctx->adjust_suffix();
  
  return val + vals;
}

#ifdef VERBOSE_STATS
//-----------------------------------------------------------------------------
/*!
* \brief Get number of bits decoded from the buffer
*
* \return Number of bits
*/
//-----------------------------------------------------------------------------
int ArithDecoder::get_nbits()
{
  return (8*buf->p_byte + buf->p_bit);
}
#endif
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
