/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ArithEncoder.cpp
*
* \brief Implementation of ArithEncoder class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "ArithEncoder.H"
#include "Param.H"
#ifdef TRACE_DEBUG
#include "Trace.H"
#endif
//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Sets output byte buffer
*
* \param[in] b Pointer to byte buffer containing arithmetic coded data
*/
//-----------------------------------------------------------------------------
void ArithEncoder::set_buffer(Buffer *b)
{
  buf = b;
}

//-----------------------------------------------------------------------------
/*!
* \brief Init encoder
*
* Initialize internal registers, required when starting coder.
*
*/
//-----------------------------------------------------------------------------
void ArithEncoder::init() 
{
  R = power2(AC_BITS) - TWO;  // range register
  L = ZERO;                   // low register
  bits_outstanding = ZERO;    // counter to deal with carry over
}

//-----------------------------------------------------------------------------
/*!
* \brief Write output bits
*
* Writes one or more bits, and solves carry over.
*
* \param[in] b value of output bit
*/
//-----------------------------------------------------------------------------
void ArithEncoder::output_bit(int b) 
{
  buf->write_bit(b);

  while (bits_outstanding > ZERO) 
  {
    buf->write_bit(ONE - b);
    bits_outstanding--;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode binary symbol
*
* Encodes binary symbol val using estimator ctx. The estimator ctx is updated
* according to the encoded symbol.
*
* \param[in] val Value of binary symbol to encode
* \param[in,out] ctx Pointer to probability estimator
*/
//-----------------------------------------------------------------------------
void ArithEncoder::encode_symbol(int val, SimpleProbEst *ctx)
{
  unsigned int R0;
#ifdef TRACE_DEBUG
  SVPRINT_ETRACE("R ", R);
  SVPRINT_ETRACE("p ", ctx->p);
#endif 
  // compute range for value 0, all values are 16-bit fixed point unsigned
  R0 = shift_right_round_us(R * ctx->p, PROB_BITS);

  // update registers
  if (val == ZERO)
    R = R0;
  else 
  {
    R = R - R0;
    L = L + R0;
  }

  // renormalize
  while (R < (unsigned int)power2(AC_BITS - ONE)) 
  {
    if (L >= (unsigned int)power2(AC_BITS)) 
    {
      output_bit(ONE);
      L -= power2(AC_BITS);
    }
    else if (L + R <= (unsigned int)power2(AC_BITS)) 
    {
      output_bit(ZERO);
    }
    else 
    {
      bits_outstanding++;
      L -= power2(AC_BITS - ONE);
    }

    R <<= ONE;
    L <<= ONE;

    ctx->bitcount[val]++;
  }

  ctx->update(val); // update probability
#ifdef TRACE_DEBUG
  SVPRINT_ETRACE("val ", val);
  RPRINT_ETRACE;
#endif  
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode binary symbol with 50/50 probability
*
* Encodes binary symbol val using estimator ctx
*
* \param[in] val Value of binary symbol to encode
* \param[in,out] ctx Pointer to probability estimator
*/
//-----------------------------------------------------------------------------
void ArithEncoder::encode_symbol_half(int val, FixedProbEst *ctx)
{
  unsigned int R0;

  // compute range for value 0, all values are 16-bit fixed point unsigned
  R0 = shift_right_round_us(R * ctx->p, PROB_BITS);

  // update registers
  if (val == ZERO)
    R = R0;
  else 
  {
    R = R - R0;
    L = L + R0;
  }

  // renormalize
  while (R < (unsigned int)power2(AC_BITS - ONE)) 
  {
    if (L >= (unsigned int)power2(AC_BITS)) 
    {
      output_bit(ONE);
      L -= power2(AC_BITS);
    }
    else if (L + R <= (unsigned int)power2(AC_BITS)) 
    {
      output_bit(ZERO);
    }
    else 
    {
      bits_outstanding++;
      L -= power2(AC_BITS - ONE);
    }

    R <<= ONE;
    L <<= ONE;

    ctx->bitcount[val]++;
  }

  ctx->symcount[val]++;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode unsigned integer symbol
*
* Encodes integer symbol val using estimators ctx and ctx_suffix.
*
* \param[in] val Value of integer symbol to encode (must be 1 or larger)
* \param[in,out] ctx Pointer to probability estimator used for prefix of binarization
* \param[in,out] ctx_suffix Pointer to probability estimator used for suffix of binarization
*/
//-----------------------------------------------------------------------------
void ArithEncoder::encode_symbol_unary(int val, UnaryProbEst *ctx, FixedProbEst *ctx_suffix)
{
  int i;

  // Prefix loop  
  for (i = ONE; i < val; i++) 
  {
    ctx->suffix++;

    if (ctx->suffix == power2(ctx->suffix_size)) 
    {
      encode_symbol(ONE, ctx);
      ctx->suffix = ZERO;
      ctx->adjust_suffix();
    }
  }

  encode_symbol(ZERO, ctx);

  // Suffix loop
  for (i = ZERO; i < ctx->suffix_size; i++)
    encode_symbol_half((ctx->suffix >> (ctx->suffix_size - ONE - i)) & ONE, ctx_suffix);

  ctx->suffix = ZERO;
  ctx->adjust_suffix();
}


//-----------------------------------------------------------------------------
/*!
* \brief Compute cost to encode binary symbol
*
* Computes estimated cost in bits of coding val using estimator ctx.
*
* \param[in] val value of binary symbol to encode
* \param[in] ctx Pointer to probability estimator
* \return Cost in bits (fixed point number with 10-bit fraction)
*/
//-----------------------------------------------------------------------------
int ArithEncoder::cost(int val, ProbEst *ctx)
{
  return param.bitsM[val][ctx->p];
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost to encode integer symbol
*
* Computes estimated cost in bits of coding val using estimator ctx
*
* \param[in] val value of binary symbol to encode
* \param[in] ctx Pointer to probability estimator used for prefix of binarization
* \param[in] ctx_suffix Pointer to probability estimator used for suffix of binarization
* \return Cost in bits (fixed point number with 10-bit fraction)
*/
//-----------------------------------------------------------------------------
int ArithEncoder::cost_unary(int val, UnaryProbEst *ctx, FixedProbEst *ctx_suffix)
{
  return cost(0, ctx) + ((val - 1) >> ctx->suffix_size) * cost(1, ctx) + ctx->suffix_size * cost(0, ctx_suffix);
}

//-----------------------------------------------------------------------------
/*!
* \brief End arithmetic code
*
* Needed at the end of each encode.
*
*/
//-----------------------------------------------------------------------------
void ArithEncoder::terminate()
{
  L += R >> ONE;
  output_bit((L >> AC_BITS) & ONE);
  buf->write_bit((L >> (AC_BITS - ONE)) & ONE);
  buf->write_bit((L >> (AC_BITS - TWO)) & ONE);
}


//-----------------------------------------------------------------------------
/*!
* \brief Copy the coding state
*
* Copies the entire coding state from the specified source object to the 
* current object
*
* \param[in,out] aenc Pointer to specified source object
*/
//-----------------------------------------------------------------------------
void ArithEncoder::copy_coding_state(ArithEncoder *aenc)
{

  R                = aenc->R;
  L                = aenc->L;
  bits_outstanding = aenc->bits_outstanding;

  buf->copy_coding_state(aenc->buf);
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
