/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ArithTranscoder.cpp
*
* \brief Implementation of ArithTranscoder class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"
#ifdef DEC_TRANSCODER
#include "ArithTranscoder.H"
#endif
#include "Param.H"
#ifdef TRACE_DEBUG
#include "Trace.H"
#endif
#include "ArithEncoder.H"
#include "EntropyModel.H"
//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------
#ifdef DEC_TRANSCODER
//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Initializes an encoder, entropymodel and buffer for each output substream.
*/
//-----------------------------------------------------------------------------
ArithTranscoder::ArithTranscoder()
{
    nthreads = param.decode_transcode_nthreads;
    if (nthreads <= 0)
        nthreads = 1;
    aenc = new ArithEncoder *[nthreads];
    cs = new EntropyModel *[nthreads];
    buf = new Buffer *[nthreads];
    for (int i = 0; i < nthreads; i++)
    {
        aenc[i] = new ArithEncoder();
        cs[i] = new EntropyModel(i == 0 ? NULL : cs[0]->synchro_state);
        buf[i] = new Buffer();
        buf[i]->reset();
    }
    decoding_unary = 0;
    thread_idx = 0;
    buf_out = NULL;
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory allocated by constructor.
*/
//-----------------------------------------------------------------------------
ArithTranscoder::~ArithTranscoder()
{
    for (int i = 0; i < nthreads; i++)
    {
        delete aenc[i];
        delete cs[i];
        delete buf[i];
    }
    delete[] aenc;
    delete[] cs;
    delete[] buf;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Return the number of bits used for just the indicated substreams, plus any final pad.
*
* The payload sizes of the nominated substreams are added, and any
* final pad where necessary (to byte-align the end of the frame).
*
* \param[in] first_ss The first substream index.
* \param[in] n_ss     The number of substreams.
*
* \return The number of bits to code the substreams and any final pad.
*/
//-----------------------------------------------------------------------------
int ArithTranscoder::substream_size(int first_ss, int n_ss)
{
    int result = 0;
    for (int i = first_ss; i < first_ss+n_ss; i++)
        result += buf[i]->payload_bits();
    // Do we add any pad bits?  They appear in between the last two streams,
    // or at the end of the last single stream.  Ie, at the end of the frame.
    if (first_ss+n_ss == nthreads)
      result += pad_bits;

    return result;
}

//-----------------------------------------------------------------------------
/*!
* \brief Return the number of bits required for coding the nominated substreams.
*
* The complete size of coding the substreams, associated sizes and padding.  This
* is a recursive routine.
*
* \param[in] first_emit The first substream index.
* \param[in] n_emit     The number of substreams.
* \param[out] last_bits Result is optionally stored here, too.
*
* \return The number of bits actually coded when writing the nominated substreams.
*/
//-----------------------------------------------------------------------------
int ArithTranscoder::calc_size_bits(int first_emit, int n_emit, int *sum_bits)
{
    if (n_emit <= 2)
      return 0; // No explicit size coded.

    int szb0 = 0;
    int szb1 = 0;
    int n_left;
    int n_right;
    Buffer::divide_substreams(n_emit, n_left, n_right);
    calc_size_bits(first_emit, n_left, &szb0);
    calc_size_bits(first_emit+n_left, n_right, &szb1);

    // Gotta code an explicit size.  In how many bits?
    int ss_size = substream_size(first_emit, n_emit);
    int bits = Buffer::get_coded_offset_size(ss_size+szb0+szb1);
    // Adding these bits might take us over a power-of-2 boundary, which is all the
    // decoder sees.
    bits = Buffer::get_coded_offset_size(ss_size+szb0+szb1+bits);

    if (sum_bits != NULL)
        *sum_bits += szb0 + szb1 + bits;

    return bits; // All the size bits we'll be coding.
}

//-----------------------------------------------------------------------------
/*!
* \brief Write a size field in the minimum number of bits.
*
* The complete size of coding the substreams, associated sizes and padding.  This
* is a recursive routine.
*
* \param[in] first_emit The first substream index.
* \param[in] n_emit     The number of substreams.
* \param[out] last_bits Result is optionally stored here, too.
*
* \return The number of bits actually coded when writing the nominated substreams.
*/
//-----------------------------------------------------------------------------
void ArithTranscoder::emit_size(int first_emit, int n_total_emit, int n_emit)
{
    int sum_size_bits = 0;
    int total_size_bits = calc_size_bits(first_emit, n_total_emit, NULL);
    calc_size_bits(first_emit, n_emit, &sum_size_bits);
    int ss_size = substream_size(first_emit, n_emit)+sum_size_bits;

    buf_out->write_bits(ss_size, total_size_bits);
}

//-----------------------------------------------------------------------------
/*!
* \brief Write substream buffers to buf_out.
*
* The substream buffers of the current frame are concatenated
* using a recursive process to reduce signalling.
*
* \param[in] first_emit Index of first substream to write.
* \param[in] n_emit     The number of substreams to write.
*/
//-----------------------------------------------------------------------------
void ArithTranscoder::emit_thread_bufs(int first_emit, int n_emit)
{
  if (n_emit == nthreads)
  {
    // Initialize buf_out.
    buf_out->reset();
  }
  if (n_emit >= 3)
  {
    int n_left;
    int n_right;
    Buffer::divide_substreams(n_emit, n_left, n_right);
    emit_size(first_emit, n_emit, n_left);
    emit_thread_bufs(first_emit, n_left);
    emit_thread_bufs(first_emit+n_left, n_right);
  }
  else if (n_emit == 2)
  {
      // We write the first buffer.
      // We reverse and write the second buffer.
      buf_out->append_bits_from(buf[first_emit]);
      // If we're writing two threads-worth of data at the end of the frame,
      // we use the pad_bits value to stuff some empty bits in for final byte
      // alignment.
      if (first_emit+n_emit == nthreads)
      {
          buf_out->write_bits(0, pad_bits);
      }
      buf_out->append_reversed_bits_from(buf[first_emit+1]);
  }
  else if (n_emit == 1)
  {
      buf_out->append_bits_from(buf[first_emit]);
      // If we're writing the last, we add some pad bits.
      if (first_emit+n_emit == nthreads)
        buf_out->write_bits(0, pad_bits);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Terminate all substream re-encoding, and write the result.
*
* All contexts used for re-encoding symbols are terminated, and
* the individual substreams are collected and written to the frame
* output buffer.
*/
//-----------------------------------------------------------------------------
void ArithTranscoder::terminate()
{
    // End of stream.  Terminate all.
    for (int i = 0; i < nthreads; i++)
        aenc[i]->terminate();

    // Concatenate buffers with a given pad, until we find
    // one that works.  With 2 substreams, we know
    // exactly the pad required.
    if (nthreads < 3)
    {
      pad_bits = 0;
      for (int i = 0; i < nthreads; i++)
        pad_bits += buf[i]->payload_bits()%8;
      pad_bits %= 8;
      pad_bits = 8-pad_bits;
      if (pad_bits == 8)
        pad_bits = 0;
      emit_thread_bufs(0, nthreads);
      if (buf_out->payload_bits()%8 != 0)
        abort();
    }
    else
    {
      // We encode sizes that depend on the remaining payload.
      // The final substream is right-byte-aligned to avoid encoding
      // a size for the final 2 substreams.
      // Start with a pad_bits of 0.
      pad_bits = 0;
      emit_thread_bufs(0, nthreads);
      if (buf_out->payload_bits()%8 != 0)
      {
        // Work out a likely pad_bits.
        pad_bits = buf_out->payload_bits()%8;
        pad_bits = 8-pad_bits;
        emit_thread_bufs(0, nthreads);
      }
      if (buf_out->payload_bits()%8 != 0)
      {
        // Changing the pad has changed some lengths.
        // Cycle through pad_bits until we find one that works.
        for (pad_bits = 1; pad_bits < 128; pad_bits++)
        {
          emit_thread_bufs(0, nthreads);
          if (buf_out->payload_bits()%8 == 0)
            break;
        }
      }
    }
}

//-----------------------------------------------------------------------------
/*!
* \brief resetSynchro on decoder, and on the specific re-encoder being used.
*
* In addition to the normal Decoder::resetSynchro, we invoke resetSynchro
* on the context currently being used to re-encode symbols.
*/
//-----------------------------------------------------------------------------
void ArithTranscoder::resetSynchro()
{
    if (!param.decode_inputref)
        ArithDecoder::resetSynchro();
    cs[thread_idx]->resetSynchro();
}

//-----------------------------------------------------------------------------
/*!
* \brief Indicate a specific context in which to re-encode symbols.
*
* When multiple re-encoding contexts are possible (ie, in the transcoder)
* a specific context in which to encode symbols can be indicated.  This is
* normally used to indicate '0' when decoding the trailing frame syntax
* after the macroblocks and superblocks have been decoded and encoded.
*
* \param[in] thrd The context number to use.
*/
//-----------------------------------------------------------------------------
void ArithTranscoder::setEncodeThread(int thrd)
{
    thread_idx = thrd;
}

//-----------------------------------------------------------------------------
/*!
* \brief initSynchro on decoder, and on the specific re-encoder being used.
*
* In addition to the normal Decoder::initSynchro, we invoke initSynchro
* on the context currently being used to re-encode symbols based on the
* line of superblocks being started.  The context being used is remembered.
*/
//-----------------------------------------------------------------------------
void ArithTranscoder::initSynchro(int iSB, int jSB)
{
    if (!param.decode_inputref)
        ArithDecoder::initSynchro(iSB, jSB);
    thread_idx = iSB%nthreads;
    cs[thread_idx]->initSynchro(iSB, jSB);
}

//-----------------------------------------------------------------------------
/*!
* \brief updateSynchro on decoder, and on the specific re-encoder being used.
*
* In addition to the normal Decoder::updateSynchro, we invoke updateSynchro
* on the context currently being used to re-encode symbols.
*/
//-----------------------------------------------------------------------------
void ArithTranscoder::updateSynchro(int iSB)
{
    if (!param.decode_inputref)
        ArithDecoder::updateSynchro(iSB);
    cs[thread_idx]->updateSynchro(iSB);
}

//-----------------------------------------------------------------------------
/*!
* \brief resetSynchro on decoder, and on the specific re-encoder being used.
*
* In addition to the normal Decoder::resetSynchro, we invoke resetSynchro
* on the context currently being used to re-encode symbols.
*/
//-----------------------------------------------------------------------------
void ArithTranscoder::init(Buffer *b_in, Buffer *b_out)
{
    // Give a dedicated buffer to each thread, which we'll join up and return later.
    buf_out = b_out;
    ArithDecoder::init(b_in, NULL);
    for (int i = 0; i < nthreads; i++)
    {
        aenc[i]->init();
        aenc[i]->set_buffer(buf[i]);
        cs[i]->reset();
        cs[i]->reset_counts();
        buf[i]->reset();
    }
}

//-----------------------------------------------------------------------------
/*!
* \brief decode_symbol on decoder, and re-encode the symbol.
*
* In addition to the normal Decoder::decode_symbol, we invoke encode_symbol
* on the context currently being used to re-encode symbols.
*/
//-----------------------------------------------------------------------------
int ArithTranscoder::decode_symbol(SimpleProbEst *ctx)
{
    int val = ArithDecoder::decode_symbol(ctx);
    if (decoding_unary)
        return val;
    if (ctx->idCtx < 0)
        abort();
    aenc[thread_idx]->encode_symbol(val, cs[thread_idx]->p_aCtx[ctx->idCtx]);
    return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief decode_symbol_half on decoder, and re-encode the symbol.
*
* In addition to the normal Decoder::decode_symbol_half, we invoke encode_symbol_half
* on the context currently being used to re-encode symbols.
*/
//-----------------------------------------------------------------------------
int ArithTranscoder::decode_symbol_half(FixedProbEst *ctx)
{
    int val = ArithDecoder::decode_symbol_half(ctx);
    if (decoding_unary)
        return val;
    if (ctx->idCtx < 0)
        abort();
    aenc[thread_idx]->encode_symbol_half(val, dynamic_cast<FixedProbEst *>(cs[thread_idx]->p_fCtx[ctx->idCtx]));
    return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief decode_symbol_unary on decoder, and re-encode the symbol.
*
* In addition to the normal Decoder::decode_symbol_unary, we invoke encode_symbol_unary
* on the context currently being used to re-encode symbols.
*/
//-----------------------------------------------------------------------------
int ArithTranscoder::decode_symbol_unary(UnaryProbEst *ctx, FixedProbEst *ctx_suffix)
{
    decoding_unary = TRUE;
    int val = ArithDecoder::decode_symbol_unary(ctx, ctx_suffix);
    if (ctx->idCtx < 0 || ctx_suffix->idCtx < 0)
        abort();
    aenc[thread_idx]->encode_symbol_unary(val, dynamic_cast<UnaryProbEst *>(cs[thread_idx]->p_aCtx[ctx->idCtx]),
                                               dynamic_cast<FixedProbEst *>(cs[thread_idx]->p_fCtx[ctx_suffix->idCtx]));
    decoding_unary = FALSE;
    return val;
}
#endif

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
