/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Block.cpp
*
* \brief Implementation of Block class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Block.H"

const int Block::bht[8] = { 4, 2, 8, 8, 0, 16, 8, 4 };
const int Block::bwt[8] = { 4, 8, 2, 8, 0, 16, 8, 4 };
const int Block::log_bht[8] = { 2, 1, 3, 3, 0, 4, 3, 2 };
const int Block::log_bwt[8] = { 2, 3, 1, 3, 0, 4, 3, 2 };


const int Block::ZigZag_Table_16x16[256]=
{
  0,  1,  16, 32, 17, 2,  3,  18, 33, 48, 64, 49, 34, 19, 4,  5,
  20, 35, 50, 65, 80, 96, 81, 66, 51, 36, 21, 6,  7,  22, 37, 52,
  67, 82, 97, 112,128,113,98, 83, 68, 53, 38, 23, 8,  9,  24, 39,
  54, 69, 84, 99, 114,129,144,160,145,130,115,100,85, 70, 55, 40,
  25, 10, 11, 26, 41, 56, 71, 86, 101,116,131,146,161,176,192,177,
  162,147,132,117,102,87, 72, 57, 42, 27, 12, 13, 28, 43, 58, 73,
  88, 103,118,133,148,163,178,193,208,224,209,194,179,164,149,134,
  119,104,89, 74, 59, 44, 29, 14, 15, 30, 45, 60, 75, 90, 105,120,
  135,150,165,180,195,210,225,240,241,226,211,196,181,166,151,136,
  121,106,91, 76, 61, 46, 31, 47, 62, 77, 92, 107,122,137,152,167,
  182,197,212,227,242,243,228,213,198,183,168,153,138,123,108,93,
  78, 63, 79, 94, 109,124,139,154,169,184,199,214,229,244,245,230,
  215,200,185,170,155,140,125,110,95, 111,126,141,156,171,186,201,
  216,231,246,247,232,217,202,187,172,157,142,127,143,158,173,188,
  203,218,233,248,249,234,219,204,189,174,159,175,190,205,220,235,
  250,251,236,221,206,191,207,222,237,252,253,238,223,239,254,255,
};

const int Block::ZigZag_Table_8x8[64]=
{
  0, 1, 8,16, 9, 2, 3,10,
  17,24,32,25,18,11, 4, 5,
  12,19,26,33,40,48,41,34,
  27,20,13, 6, 7,14,21,28,
  35,42,49,56,57,50,43,36,
  29,22,15,23,30,37,44,51,
  58,59,52,45,38,31,39,46,
  53,60,61,54,47,55,62,63
};

const int Block::ZigZag_Table_4x4[16]=
{
  0,1,4,8,
  5,2,3,6,
  9,12,13,10,
  7,11,14,15
};

//-----------------------------------------------------------------------------
// Public Methods
//-----------------------------------------------------------------------------
Block::Block()
{
  set_size(0, 0);
}

Block::Block(int type)
{
  set_type(type);
}

Block::Block(int h, int w)
{
  set_size(h, w);
}

void Block::set_type(int type)
{
  m_type = type;
  set_size(bht[type], bwt[type]);
}

void Block::set_size(int h, int w)
{
  height = h;
  width = w;
  TX_NCOEFF_MxN = h*w;
  
  assert(TX_NCOEFF_MxN <= 256);
}

Block::~Block()
{
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute sum of squared differences between current block and reference block
* 
* \param[in] x Pointer for block to refer
*
* \return Sum of squared differences between current block and reference block
*/
//-----------------------------------------------------------------------------
double Block::ssd(Block &x)
{
  int     i;
  double  r;
  double  tmp;
  
  r = ZERO;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++) 
  {
    tmp = (double) clip(shift_right_round(val[i], SAMPLE_ADD_PREC),-MID_SAMPLE_VAL, MAX_SAMPLE_VAL-MID_SAMPLE_VAL)
    - (double) clip(shift_right_round(x.val[i], SAMPLE_ADD_PREC),-MID_SAMPLE_VAL, MAX_SAMPLE_VAL-MID_SAMPLE_VAL);
    r += tmp*tmp; 
  }
  
  return (r);
}

const int* Block::get_zz_table()
{
  switch(TX_NCOEFF_MxN)
  {
  case 16:
    return ZigZag_Table_4x4;
  case 64:
    return ZigZag_Table_8x8;
  default:
    assert(TX_NCOEFF_MxN == 256);
    return ZigZag_Table_16x16;
  };
}

void Block::zigzagscan()
{
  assert(TX_NCOEFF_MxN <= 256);
  int tmp[256];
  const int * zz_table = get_zz_table();
  int i;

  for(i=0;i<TX_NCOEFF_MxN;i++)
    tmp[i] = val[zz_table[i]] ;    
  
  for(i=0;i<TX_NCOEFF_MxN;i++)
    val[i] = tmp[i] ;
}

void Block::inverse_zigzagscan()
{
  assert(TX_NCOEFF_MxN <= 256);
  int tmp[256];
  const int * zz_table = get_zz_table();
  int i;

  for(i=0;i<TX_NCOEFF_MxN;i++)
    tmp[i] = val[i];    
  
  for(i=0;i<TX_NCOEFF_MxN;i++)
    val[zz_table[i]] = tmp[i] ;
}

//-----------------------------------------------------------------------------
/*!
* \brief Subtract a block
*
* Subtracts a given block from current block
*
* \param[in] x Reference to block to subtract
*/
//-----------------------------------------------------------------------------
void Block::sub(Block &x)
{
  int i;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    val[i] -= x.val[i];
}

//-----------------------------------------------------------------------------
/*!
* \brief Add a block
*
* Adds a given block to current block
*
* \param[in] x Reference to block to add
*/
//-----------------------------------------------------------------------------
void Block::add(Block &x) 
{
  int i;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    val[i] += x.val[i];
}


void Block::insert(Block &blk, int i, int j)
{
  assert(i + blk.height <= height);
  assert(j + blk.width <= width);

  int k, l;
  for (k = 0; k < blk.height; k++)
  {
    for (l = 0; l < blk.width; l++)
    {
      val[width * (i + k) + (j + l)] = blk.val[blk.width * k + l];
    }
  }
}

void Block::extract(Block &blk, int i, int j)
{
  assert(i + blk.height <= height);
  assert(j + blk.width <= width);

  int k, l;
  for (k = 0; k < blk.height; k++)
  {
    for (l = 0; l < blk.width; l++)
    {
      blk.val[blk.width * k + l] = val[width * (i + k) + (j + l)];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Set all values in block to zero
*
*/
//-----------------------------------------------------------------------------
void Block::zero()
{
  set(ZERO);
}

//-----------------------------------------------------------------------------
/*!
* \brief Set all values in block to a given value
*
* \param[in] v Given value
*/
//-----------------------------------------------------------------------------
void Block::set(int v)
{
  int i;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    val[i] = v;
}

//-----------------------------------------------------------------------------
/*!
* \brief Determines if all values in block are zero
*
* \return TRUE if all values are zero, FALSE otherwise
*/
//-----------------------------------------------------------------------------
int Block::is_zero()
{
  int i;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    if (val[i] != ZERO)
      return FALSE;
  
  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute sum of values in block
*
* \return Sum of values in block
*/
//-----------------------------------------------------------------------------
int Block::sum() 
{
  int i;
  int r;
  
  r = ZERO;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    r += val[i];
  
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute sum of squared values in block
*
* \return Sum of squared values in block
*/
//-----------------------------------------------------------------------------
int Block::sum2() 
{
  int i;
  int r;
  
  r = ZERO;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    r += val[i] * val[i];
  
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Determines number of nonzero values in block
*
* \return Number of nonzero values in block
*/
//-----------------------------------------------------------------------------
int Block::nnz()
{
  int i;
  int r;
  
  r = ZERO;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    r += val[i] != ZERO;
  
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Subtract a constant
*
* Subtracts a constant from all values of current block
*
* \param[in] DC Value to be subtracted
*/
//-----------------------------------------------------------------------------
void Block::sub(int DC) 
{
  int i;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    val[i] -= DC;
}

//-----------------------------------------------------------------------------
/*!
* \brief Multiply by a constant
*
* Multiplies each element of the block by constant s.
*
* \param[in] s Constant
*/
//-----------------------------------------------------------------------------
void Block::mul(int s)
{
  int i;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    val[i] *= s;
}

//-----------------------------------------------------------------------------
/*!
* \brief Add a constant
*
* Adds a constant to all values in current block
*
* \param[in] DC Value to be added
*/
//------------------------------------------------------------------------------
void Block::add(int DC)
{
  int i;
  
  for (i = ZERO; i < TX_NCOEFF_MxN; i++)
    val[i] += DC;
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy a block
*
* Copy a given block to current block
*
* \param[in] x Reference to block to copy
*/
//-----------------------------------------------------------------------------
void Block::copy(Block& blk)
{
  assert(blk.TX_NCOEFF_MxN == TX_NCOEFF_MxN);
  
  memcpy(val, blk.val, sizeof(val[0]) * TX_NCOEFF_MxN);
}

const int Block::scan[6][TX_NCOEFF] = 
{
  { 0, 1, 4, 2, 8, 3, 5, 6, 9, 12, 7, 10, 13, 11, 14, 15
  },
  { 0, 8, 1, 9, 2, 3, 10, 4, 5, 11, 6, 7, 12, 13, 14, 15
  },
  { 0, 1, 2, 4, 3, 5, 6,  8, 10, 7, 9, 12, 11, 14, 13, 15
  },
  { 0, 1, 2, 3, 8, 9, 10, 11, 4, 5, 6, 7, 12, 13, 14, 15
  },
  { 0, 2, 4, 6, 1, 3, 5, 7, 8, 10, 12, 14, 9, 11, 13, 15
  },
  { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15
  },
};

void Block::rescan(int type1, int type2)
{
  int tmp[16];
  int i;
  
  assert(width * height == 16);
  
  if (type1 == type2)
    return;
  
  for (i=0; i<16; i++)
  {
    tmp[i] = val[scan[type1][i]];
  }
  
  for (i=0; i<16; i++)
  {
    val[scan[type2][i]] = tmp[i];
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
