/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Buffer.cpp
*
* \brief Implementation of Buffer class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Buffer.H"
#include "Global.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Allocates memory for buffer data.
*/
//-----------------------------------------------------------------------------
Buffer::Buffer()
{
  data = new unsigned char[MAX_BUFFER_SIZE];
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory.
*/
//-----------------------------------------------------------------------------
Buffer::~Buffer()
{
  delete[] data;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Reset buffer
*
* Resets position to first bit in buffer and clears buffer. 
* This should beinvoked before writing any bits or bytes to the buffer.
*
*/
//-----------------------------------------------------------------------------
void Buffer::reset()
{
  p_bit = ZERO;
  p_byte = ZERO;
  memset(data, ZERO, MAX_BUFFER_SIZE * sizeof(unsigned char));
}

//-----------------------------------------------------------------------------
/*!
* \brief Write a bit to buffer
*
* \param[in] b Bit value
*/
//-----------------------------------------------------------------------------
void Buffer::write_bit(int b)
{
  data[p_byte] |= b << (7 - p_bit);
  p_bit++;

  if (p_bit == 8) 
  {
    p_bit = ZERO;
    p_byte++;

    if (p_byte == MAX_BUFFER_SIZE) 
    {
      fprintf(stderr, "MAX_BUFFER_SIZE exceeded.\n");
      exit(EXIT_FAILURE);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Write an n-bit unsigned integer to buffer
*
* \note Bits are written starting from the MSB
*
* \param[in] b Integer value
* \param[in] n Size of integer
*
* \return Number of bits written
*/
//-----------------------------------------------------------------------------
int Buffer::write_bits(int b, int n)
{
  int i;

  for (i = ZERO; i < n; i++)
    write_bit((b >> (n-1 - i)) & 1);

  return n;
}

//-----------------------------------------------------------------------------
/*!
* \brief Read a bit from buffer
*
* \return Bit value
*/
//-----------------------------------------------------------------------------
int Buffer::read_bit()
{
  int b;

  b = data[p_byte];
  b = (b >> (7 - p_bit)) & ONE;
  p_bit++;
  if (p_bit == 8) 
  {
    p_bit = 0;
    p_byte++;
  }

  return b;
}

//-----------------------------------------------------------------------------
/*!
* \brief Read an n-bit unsigned integer from buffer
*
* \param[in] n Size of unsigned integer
*
* \return Integer value
*/
//-----------------------------------------------------------------------------
int Buffer::read_bits(int n)
{
  int i;
  int r;

  r = ZERO;

  for (i = ZERO; i < n; i++)
    r = (r << ONE) | read_bit();

  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Write an signed integer to buffer
*
* \return Number of bits written
*/
//-----------------------------------------------------------------------------
int Buffer::write_signed(int b) // cq_start
{
  int i;

  if (b==0)
  {
    write_bit(0);
    return 1;
  }
  else
  {
    for (i = ZERO; i < iabs(b); i++)
      write_bit(1);
    write_bit(0);
    if (b<0)
      write_bit(1);
    else
      write_bit(0);
    return (iabs(b) + 2);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Write an unsigned integer to buffer
*
* \return Number of bits written
*/
//-----------------------------------------------------------------------------
int Buffer::write_unsigned(int b)
{
  int i;

  if (b==0)
  {
    write_bit(0);
    return 1;
  }
  else
  {
    for (i = ZERO; i < iabs(b); i++)
      write_bit(1);
    write_bit(0);
    return (iabs(b) + 1);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Read an signed integer from buffer
*
* \return Signed integer value
*/
//-----------------------------------------------------------------------------
int Buffer::read_signed()
{
  int r;

  r = read_bit();
  if (r != 0)
  {
    while (read_bit())
      r++;
    if (read_bit() == NEGATIVE)
      r = -r;
  }
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Read an unsigned integer from buffer
*
* \return Unsigned integer value
*/
//-----------------------------------------------------------------------------
int Buffer::read_unsigned() // cq_end
{
  int r;

  r = read_bit();
  if (r != 0)
  {
    while (read_bit())
      r++;
  }
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Read a byte from buffer
*
* \note Position should be aligned to byte boundary. Bit position is ignored.
*
* \return Byte value
*/
//-----------------------------------------------------------------------------
int Buffer::read_byte()
{
  return data[p_byte++];
}

//-----------------------------------------------------------------------------
/*!
* \brief Read a word from buffer
*
* \note Position should be aligned to byte boundary. Bit position is ignored.
*
* \return Word value (16 bits)
*/
//-----------------------------------------------------------------------------
int Buffer::read_word()
{
  int result;

  result = (data[p_byte] << BITS_IN_BYTE) + data[p_byte + ONE];
  p_byte += sizeof(unsigned short);

  return result;
}

//-----------------------------------------------------------------------------
/*!
* \brief Write buffer to file
*
* Writes content of buffer to file, including payload type and size.
*
* \param[in,out] f File pointer
*/
//-----------------------------------------------------------------------------
void Buffer::write(FILE *f)
{
  payload_size = p_byte + ((p_bit + 7) >> 3);

  fwrite(&payload_type, sizeof(payload_type), ONE, f);
  fwrite(&payload_size, sizeof(payload_size), ONE, f);

  if (payload_size != ZERO)
    fwrite(data, payload_size, ONE, f);
}

//-----------------------------------------------------------------------------
/*!
* \brief Read buffer from file
*
* Reads content of buffer from file, including payload type and size. Position
* is reset to first bit in buffer.
*
* \param[in,out] f File pointer
*
* \return FALSE if an error occured, TRUE otherwise
*/
//-----------------------------------------------------------------------------
int Buffer::read(FILE *f)
{
  unsigned int bytes_read;

  bytes_read = (unsigned int)fread(&payload_type, ONE, sizeof(payload_type), f);
  if (bytes_read != sizeof(payload_type))
    return FALSE;

  bytes_read = (unsigned int)fread(&payload_size, ONE, sizeof(payload_size), f);
  if (bytes_read != sizeof(payload_size))
    return FALSE;

  if (payload_size > MAX_BUFFER_SIZE)
    return FALSE;
  
  if (payload_size != ZERO) 
  {
    bytes_read = (unsigned int)fread(data, ONE, payload_size, f);
    if (bytes_read != payload_size)
      return FALSE;
  }

  p_bit = ZERO;
  p_byte = ZERO;

  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy a buffer
*
* Copies content of buffer and current position. Payload type and size are
* not copied (!?!).
*
* \param[in] buf Byte buffer to copy from
*/
//-----------------------------------------------------------------------------
void Buffer::copy(Buffer *buf)
{
  p_bit = buf->p_bit;
  p_byte = buf->p_byte;
  memcpy(data, buf->data, MAX_BUFFER_SIZE * sizeof(unsigned char));
}

//-----------------------------------------------------------------------------
/*!
* \brief Store the state of bitstream
*
* \param[in] buf Byte buffer to copy from
*/
//-----------------------------------------------------------------------------
void Buffer::copy_coding_state(Buffer *buf)
{
  p_bit = buf->p_bit;
  p_byte = buf->p_byte;

  payload_type = buf->payload_type;
  payload_size = buf->payload_size;
}

#ifdef DEC_TRANSCODER
//-----------------------------------------------------------------------------
/*!
* \brief Append bytes from an external source.
*
* \param[in] p Byte buffer to copy from
* \param[in] sz How many bytes to copy
*/
//-----------------------------------------------------------------------------
void Buffer::append_bytes_from(unsigned char *p, int sz)
{
    if (p_bit != 0)
        abort();
    memcpy(data+p_byte, p, sz);
    p_byte += sz;
}
#endif

#ifdef DEC_TRANSCODER
//-----------------------------------------------------------------------------
/*!
* \brief Append bytes from a buffer that's been written to.
*
* \param[in] buf Byte buffer to copy from
*/
//-----------------------------------------------------------------------------
void Buffer::append_bytes_from(Buffer *buf)
{
    if (buf->p_bit != 0)
        abort();
    append_bytes_from(buf->data, buf->p_byte);
}
#endif

#ifdef DEC_TRANSCODER
//-----------------------------------------------------------------------------
/*!
* \brief Append bits from a buffer that's been written to.
*
* \param[in] buf Buffer to copy from
*/
//-----------------------------------------------------------------------------
void Buffer::append_bits_from(Buffer *buf)
{
    for (unsigned int i = 0; i < buf->p_byte; i++)
        write_bits(buf->data[i], 8);
    write_bits(buf->data[buf->p_byte]>>(8-buf->p_bit), buf->p_bit);
}
#endif

#ifdef DEC_TRANSCODER
unsigned char Buffer::rev_bits_tab[256];

//-----------------------------------------------------------------------------
/*!
* \brief Initialize the reversed-bits table.
*/
//-----------------------------------------------------------------------------
void Buffer::fill_rev_bits_tab()
{
    for (int i = 0; i <= 255; i++)
    {
        int normal = i;
        int rev = 0;
        for (int b = 0; b < 8; b++)
        {
            rev <<= 1;
            rev |= normal&1;
            normal >>= 1;
        }

        rev_bits_tab[i] = rev;
    }
}
#endif

#ifdef DEC_TRANSCODER
//-----------------------------------------------------------------------------
/*!
* \brief Append bits from a buffer that's been written to, but in reverse order.
*
* \param[in] buf Buffer to copy from
*/
//-----------------------------------------------------------------------------
void Buffer::append_reversed_bits_from(Buffer *buf)
{
    if (rev_bits_tab[0xff] != 0xff)
        fill_rev_bits_tab();
    for (unsigned int i = 0; i < buf->p_bit; i++)
        write_bit((buf->data[buf->p_byte]>>(7-buf->p_bit+i+1))&1);
    for (int i = (int)buf->p_byte-1; i >= 0; i--)
        write_bits(rev_bits_tab[buf->data[i]], 8);
}
#endif

#ifdef FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
* \brief Return the value of the specified number of bits from the given offset.
*
* \param[in] bit_offset
* \param[in] bit_length
*/
//-----------------------------------------------------------------------------
int Buffer::read_bits_at(int bit_offset, int bit_length)
{
    int r = 0;
    while (bit_length > 0)
    {
        r <<= 1;
        r |= read_bit_at(bit_offset);
        bit_offset++;
        bit_length--;
    }

    return r;
}
#endif

#ifdef FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
* \brief Initialize ourselves using bits from a nominated buffer.
*
* \param[in] buf buffer to copy from.
* \param[in] bit_offset
* \param[in] bit_length
*/
//-----------------------------------------------------------------------------
void Buffer::copy_bits_from(Buffer *buf, int bit_offset, int bit_length)
{
    p_byte = 0;
    p_bit = 0;
    memset(data, 0, (bit_length+7)/8);
    while (bit_offset%8 != 0 && bit_length > 0)
    {
        write_bit(buf->read_bit_at(bit_offset));
        bit_offset++;
        bit_length--;
    }
    while (bit_length >= 8)
    {
        write_bits(buf->data[bit_offset/8], 8);
        bit_offset += 8;
        bit_length -= 8;
    }
    while (bit_length > 0)
    {
        write_bit(buf->read_bit_at(bit_offset));
        bit_offset++;
        bit_length--;
    }
    // Reset ourselves to the start for later reading.
    p_byte = 0;
    p_bit = 0;
}
#endif

#if defined(FRAME_SUBSTREAMS) || defined(DEC_TRANSCODER)
//-----------------------------------------------------------------------------
/*!
* \brief Return the number of bits required to code a given value.
*
* \param[in] total_size_bits the max value wanted to be coded.
*
* \return The minimum number of bits required to code the given value.
*/
//-----------------------------------------------------------------------------
int Buffer::get_coded_offset_size(int total_size_bits)
{
    int bits = 0;
    while (total_size_bits > 0)
    {
        total_size_bits >>= 1;
        bits++;
    }
    return bits;
}
#endif

#if defined(FRAME_SUBSTREAMS) || defined(DEC_TRANSCODER)
//-----------------------------------------------------------------------------
/*!
* \brief Given a number of substreams to code, indicate the # left and right to minimize signalling.
*
* \param[in] n_ss Number of substreams being coded.  Should be 3 or larger.
* \param[out] n_left Number to code to the left.
* \param[out] n_right Number to code to the right.
*/
//-----------------------------------------------------------------------------
void Buffer::divide_substreams(int n_ss, int &n_left, int &n_right)
{
  // For the left side, we take the biggest power of 2 that fits.
  n_left = 1;
  while (n_left*2 < n_ss)
    n_left *= 2;
  n_right = n_ss - n_left;
}
#endif


#ifdef FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
* \brief Initialize ourselves using bits from a nominated buffer, copied in reverse order.
*
* The bits copied start from bit_offset-1, copying backward for bit_length bits.
*
* \param[in] buf buffer to copy from.
* \param[in] bit_offset
* \param[in] bit_length
*/
//-----------------------------------------------------------------------------
void Buffer::copy_bits_backwards_from(Buffer *buf, int bit_offset, int bit_length)
{
    p_byte = 0;
    p_bit = 0;
    memset(data, 0, (bit_length+7)/8);
    while (bit_length > 0)
    {
        bit_offset--;
        bit_length--;
        write_bit(buf->read_bit_at(bit_offset));
    }
    // Reset ourselves to the start for later reading.
    p_byte = 0;
    p_bit = 0;
}
#endif
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
