/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ChromaFrames8pel.cpp
*
* \brief Implementation of ChromaFrames8pel class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "ChromaFrames8pel.H"
#include "Param.H"
#include "Global.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Allocates memory.
*
* \param[in] h Height of reference frame (luma samples)
* \param[in] w Width of reference frame (luma samples)
* \param[in] b Size of added border (luma samples)
* \param[in] decoder_mode 0: Encoder mode, 1: Decoder mode
*/
//-----------------------------------------------------------------------------
ChromaFrames8pel::ChromaFrames8pel(int h, int w, int b, int decoder_mode)
{
  int i,j;

  this->border = b >> 1;
  width = (w + 2*b) >> 1;
  height = (h + 2*b) >> 1;

  this->decoder_mode = decoder_mode;
  if (decoder_mode)
  {
    data[0][0] = new unsigned char[TWO * width * height];
  }
  else
  {
    for (i=0; i<FRACTIONAL_PEL*TWO; i++)
    {
      for (j=0; j<FRACTIONAL_PEL*TWO; j++) 
      {
        data[i][j] = new unsigned char[TWO * width * height];
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory.
*
*/
//-----------------------------------------------------------------------------
ChromaFrames8pel::~ChromaFrames8pel()
{
  int i, j;

  if (decoder_mode)
  {
    delete[] data[0][0];
  }
  else
  {
    for (i=0; i<FRACTIONAL_PEL*TWO; i++) 
    {
      for (j=0; j<FRACTIONAL_PEL*TWO; j++) 
      {
        delete[] data[i][j];
      }   
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate interpolated frames for given reference frame
*
* \param[in] ref Pointer to reference frame
*/
//-----------------------------------------------------------------------------
void ChromaFrames8pel::makemap(Frame *ref)
{
  int i,j;
  int k,l;
  int u;
  int up;

  unsigned char *p;
  int round;
  int a[2],b[2],c[2],d[2];
  unsigned char *s[2], *t[2];
  int uv;

  int end = decoder_mode ? 1 : FRACTIONAL_PEL*TWO;
  for (k=0; k<end; k++) 
  {
    for (l=0; l<end; l++) 
    {
      round = 128 - (((k>>1) + (l>>1)) & param.mc_round);

      p = data[k][l];

      for (i=0; i<height; i++) 
      {
        u = i - border;
        up = u + ONE;
        u  = clip(u,  ZERO, ref->height2 - ONE);
        up = clip(up, ZERO, ref->height2 - ONE);

        for (uv=0; uv<2; uv++) 
        {
          s[uv] = &ref->crcb[uv][ref->width2*u ];
          t[uv] = &ref->crcb[uv][ref->width2*up];

          a[uv] = s[uv][0];
          b[uv] = a[uv];
          c[uv] = t[uv][0];
          d[uv] = c[uv];
        }

        for (j=0; j<border; j++) 
        {
          for (uv=0; uv<2; uv++)
            p[uv] = ((FRACTIONAL_PEL*TWO-k)*(FRACTIONAL_PEL*TWO-l)*a[uv] + (FRACTIONAL_PEL*TWO-k)*l*b[uv] + k*(FRACTIONAL_PEL*TWO-l)*c[uv] + k*l*d[uv] + round) >> 8;

          p += 2;
        }

        for (j=1; j<ref->width2; j++) 
        {
          for (uv=0; uv<2; uv++) 
          {
            b[uv] = s[uv][j];
            d[uv] = t[uv][j];

            p[uv] = ((FRACTIONAL_PEL*TWO-k)*(FRACTIONAL_PEL*TWO-l)*a[uv] + (FRACTIONAL_PEL*TWO-k)*l*b[uv] + k*(FRACTIONAL_PEL*TWO-l)*c[uv] + k*l*d[uv] + round) >> 8; 

            a[uv] = b[uv];
            c[uv] = d[uv];
          }

          p += 2;
        }

        for (j=0; j<=border; j++) 
        {
          for (uv=0; uv<2; uv++) 
            p[uv] = ((FRACTIONAL_PEL*TWO-k)*(FRACTIONAL_PEL*TWO-l)*a[uv] + (FRACTIONAL_PEL*TWO-k)*l*b[uv] + k*(FRACTIONAL_PEL*TWO-l)*c[uv] + k*l*d[uv] + round) >> 8;

          p += 2;
        }
      }
    }
  }

}

//-----------------------------------------------------------------------------
/*!
* \brief Generate interpolated frames for given reference frame in AIF mode
*
* \param[in] ref Pointer to reference frame
*/
//-----------------------------------------------------------------------------
void ChromaFrames8pel::AIF_makemap(Frame *ref)
{
  makemap(ref);
}

/*! 
* \brief Prints the interpolated values
* 
* Prints the interpolated values specified by the input arguments
*
* \param[in] si Starting position (same for both directions)
* \param[in] ei Ending position (same for both directions)
* \param[in] crcb Chroma index (U or V)
* \param[in] logpel Variable controlling the positions (full-pel, half-pel etc) printed
*/
void ChromaFrames8pel::print(int si, int ei, int crcb, int logpel)
{

  int i,j,k,l;

  printf("\nstart:%d end before:%d\n",si,ei);

  for (k=0;k<ei-si;k++)
  {
    for (i=0;i<FRACTIONAL_PEL*TWO;i=i+((FRACTIONAL_PEL*TWO)>>logpel))
    {
      for (l=0;l<ei-si;l++)
      {
        for (j=0;j<FRACTIONAL_PEL*TWO;j=j+((FRACTIONAL_PEL*TWO)>>logpel))
        {
          printf("%3d ",data[i][j][width*(si+k)+(si+l)+crcb]);
        }
        if (l==ei-si-1) printf("\n");
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Generate interpolated block
 *
 * \param[in] k Fractional vertical position
 * \param[in] l Fractional horizontal position
 * \param[in] u Integer vertical position
 * \param[in] v Integer horizontal position
 */
//-----------------------------------------------------------------------------
void ChromaFrames8pel::interpolate_chroma_on_the_fly(int k, int l, int u, int v, unsigned char *blk_data)
{
  int x, y;
  unsigned char *a, *c;
  int offset = FRACTIONAL_PEL * TWO;
  int offset_minus_ver = offset - k;
  int offset_minus_hor = offset - l;
  int round = 128 - (((k>>1) + (l>>1)) & param.mc_round);
  
  u = clip(u, 0, height - MOTION_BLK_SIZE/2 - 1);
  v = clip(v, 0, width - MOTION_BLK_SIZE/2 - 1);
  
  a = data[0][0] + 2 * (width * u + v);
  c = a + 2 * width;
  
  for (y = 0; y < MOTION_BLK_SIZE/2; y++) 
  {
    for (x = 0; x < MOTION_BLK_SIZE; x++) 
    {
      int val = round;
      val += offset_minus_ver * (offset_minus_hor * a[x] + l * a[x + 2]);
      val += k * (offset_minus_hor * c[x] + l * c[x + 2]);
      blk_data[MOTION_BLK_SIZE * y + x] = val >> 8;
    }
    
    a = c;
    c += 2 * width;
  }
}
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
