/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file ContextMap.cpp
*
* \brief Implementation of ContextMap class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "ContextMap.H"
#include "Global.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Allocates memory.
*
* \param[in] h Height of map
* \param[in] w Width of map
*/
//-----------------------------------------------------------------------------
ContextMap::ContextMap(int h, int w)
{
  width  = w + 1;
  height = h + 1;

  data = new int[width * height];
  base = data + width + 1;
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory.
*/
//-----------------------------------------------------------------------------
ContextMap::~ContextMap()
{
  delete[] data;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Reset map
*
* Set all elements' value to -1
*/
//-----------------------------------------------------------------------------
void ContextMap::reset()
{
  memset(data, 0xff, width * height * sizeof(int));
}

//-----------------------------------------------------------------------------
/*!
* \brief Set a value
*
* Set element at (i,j) to value v
*
* \param[in] i vertical position
* \param[in] j horizontal position
* \param[in] v value
*/
//-----------------------------------------------------------------------------
void ContextMap::set(int i, int j, int v)
{
  base[width * i + j] = v;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Set a value to a block of data
 *
 * Set element at (i,j) to value v
 *
 * \param[in] i vertical position
 * \param[in] j horizontal position
 * \param[in] v value
 * \param[in] h height of block
 * \param[in] w width of block
 */
//-----------------------------------------------------------------------------
void ContextMap::set(int i, int j, int v, int h, int w)
{
  int k, l;
  
  for (k = 0; k < h; k++)
  {
    for (l = 0; l < w; l++)
    {
      set(i+k, j+l, v);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Get a value
*
* \param[in] i Vertical position
* \param[in] j Horizontal position
*
* \return Value of element at (i,j)
*/
//-----------------------------------------------------------------------------
int ContextMap::get(int i, int j)
{
  return base[width * i + j];
}

//-----------------------------------------------------------------------------
/*!
* \brief Get context of an element
*
* Computes the context of element at (i,j). The value of the context is the value
* 'b' of the element at (i-1,j) plus the value 'a' of the element at (i,j-1). If
* both a and b are negative, the context is defined as 1. If either a or b is
* negative, the context is defined as twice the value of a or b that is positive.
*
* \param[in] i Vertical position
* \param[in] j Horizontal position
*
* \return Context value
*/
//-----------------------------------------------------------------------------
int ContextMap::getcontext(int i, int j)
{
  int result;

  int a = get(i, j - 1); // left
  int b = get(i - 1, j); // top

  if (a < 0 && b < 0) // Neither has a valid value
  {
    result = 1;
  }
  else
  {
    a = imin(a, 1);
    b = imin(b, 1);

    if (a < 0)
      a = b;
    if (b < 0)
      b = a;

    result = a + b;
  }

  return result;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode the value of the element at position (i,j)
*
* Decodes and stores value of element at position(i,j).
*
* \note This method does not use but duplicates code from ContextMap::getcontext()
* to speed up decoding.
*
* \param[in] i Vertical position
* \param[in] j Horizontal position
* \param[in,out] edec Pointer to entropy decoder
* \param[in] emodel Entropy model index (EntropyModel::I_luma, etc.)
*
* \return Decoded value
*/
//-----------------------------------------------------------------------------
int ContextMap::decode(int i, int j, EntropyDecoder *edec, int emodel)
{
  int ctx = getcontext(i, j);
  int result;

  result = edec->decode_zblk(ctx, emodel);

  set(i, j, result ^ 1);

  return result;
}


void ContextMap::convert_indices(int i, int j, int &i_map, int &j_map, int sip_type)
{
  switch (sip_type)
  {
  case SIP8x8:
    i_map = 2 * i;
    j_map = 2 * j;
    break;

  case SIP2x8:
    i_map = i>>1;
    j_map = 2*j + (i&1);
    break;

  case SIP8x2:
    i_map = 2*i + ((j>>1)&1);
    j_map = 2*(j>>2) + (j&1);
    break;

  default:
    i_map = i;
    j_map = j;
  }
}

void ContextMap::copy_from_inter(ContextMap *interMap, int i, int j)
{
  int k, l;
  
  for (k = 0; k < 2; k++)
  {
    for (l = 0; l < 2; l++)
    {
      set(i+k, j+l, interMap->get(i+k, j+l));
    }
  }
}

void ContextMap::copy_from_intra(ContextMap *intraMap, int i, int j, int sip_type)
{
  int k, l;

  if (sip_type == SIP8x2)
  {
    set(i, j,   intraMap->get(i,   j) + intraMap->get(i,   j+1), 2, 1);
    set(i, j+1, intraMap->get(i+1, j) + intraMap->get(i+1, j+1), 2, 1);
  }
  else if (sip_type == SIP2x8)
  {
    set(i,   j, intraMap->get(i,   j) + intraMap->get(i,   j+1), 1, 2);
    set(i+1, j, intraMap->get(i+1, j) + intraMap->get(i+1, j+1), 1, 2);
  }
  else
  {
    int size = (sip_type == SIP16x16) ? 4 : 2;

    for (k = 0; k < size; k++)
    {
      for (l = 0; l < size; l++)
      {
        set(i+k, j+l, intraMap->get(i+k, j+l));
      }
    }
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
