/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
 * \file Decoder.cpp
 *
 * \brief Implementation of Decoder class
 *
 *
 */

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Decoder.H"
#include "Param.H"
#include "IntraPrediction.H"       // All the intra prediction functions have been moved here.
#include "TransformAlternate.H"
#include "AIF_adaptive_filter.H"

#ifndef READ_QTABLE_FROM_FILE
#include "QuantTables.H"
#endif

#ifdef TRACE_DEBUG
#include "Trace.H"
#endif
#ifdef SUBSTREAM_MT
#include <semaphore.h>
#include <pthread.h> // mutex when running synchronously.
#endif

#ifdef SUBSTREAM_MT
#define edec_ss (reinterpret_cast<long>(pthread_getspecific(substream_number_key)))
#define thrd_no (reinterpret_cast<long>(pthread_getspecific(thread_number_key)))

pthread_key_t substream_number_key;
pthread_key_t thread_number_key;
#elif defined(FRAME_SUBSTREAMS)
static int edec_ss;
#define thrd_no 0
#endif

#ifdef FRAME_SUBSTREAMS
#define edec edecs[edec_ss]
#define intra_pred intra_preds[thrd_no]
#define mlc mlcs[thrd_no]
#endif

#include "Deblocking_Filter.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
 * \brief Constructor
 *
 * \param[in] h height of frame in sample units
 * \param[in] w width of frame in sample units
 */
//-----------------------------------------------------------------------------
Decoder::Decoder(int h, int w)
{
  int i;
  
  mf = new MotionField(h, w);
  
#ifdef FRAME_SUBSTREAMS
  edecs = new EntropyDecoder*[param.superblock_nsubstreams];
  for (i = 0; i < param.superblock_nsubstreams; i++)
    edecs[i] = new EntropyDecoder(i == 0 ? NULL : edecs[0]->GetEntropyModel()->synchro_state);
  substream_bufs = new Buffer *[param.superblock_nsubstreams];
  for (i = 0; i < param.superblock_nsubstreams; i++)
    substream_bufs[i] = new Buffer();
  ss_work_info = new SubstreamWorkInfo[param.superblock_nsubstreams];
  
#ifdef SUBSTREAM_MT
  substream_args = new SubstreamThreadArgs[param.decode_nthreads];
  pthread_key_create(&substream_number_key, NULL);
  pthread_key_create(&thread_number_key, NULL);
#endif
#else
  edec = new EntropyDecoder;
#endif
  
  zmap[ZERO] = new ContextMap(h>>2, w>>2);
  zmap[ONE ] = new ContextMap(h>>2, w>>2);
  zmap_chroma[ZERO][ZERO] = new ContextMap(h>>2, w>>2); 
  zmap_chroma[ZERO][ONE ] = new ContextMap(h>>2, w>>2); 
  zmap_chroma[ONE][ZERO] = new ContextMap(h>>2, w>>2); 
  zmap_chroma[ONE][ONE ] = new ContextMap(h>>2, w>>2);
  
#ifdef FRAME_SUBSTREAMS
  mlcs = new MLC_Stat[param.decode_nthreads];
#endif
  
  intra_map = new char[h * w >> 4];
  
  if (((param.P_subpel < 2 && param.B_subpel < 2) || !param.gen8pel_filter))
    ref_mgr = new RefFrameManager(h, w, 1); // invoke in decoder mode
  else
    ref_mgr = new RefFrameManager(h, w); // invoke in encoder mode
  
  trans[Transform::DCT] = new FastDCT;
  trans[Transform::AlternateIntra] = new TransformAlternate(ft_I_frame);
  trans[Transform::AlternateInter] = new TransformAlternate(ft_P_frame, param.alt8x8_trans_idx);
  
  wcur       = new WFrame(h, w);
  wcur_pred  = new WFrame(h, w);
  wcur_qpe   = new WFrame(h, w);
  
  // All the intra prediction functions have been moved here.
#ifdef FRAME_SUBSTREAMS
  intra_preds = new IntraPrediction*[param.decode_nthreads];
  for (i = 0; i < param.decode_nthreads; i++)
    intra_preds[i] = new IntraPrediction(h, w,
                                         i == 0 ? NULL : intra_preds[0]->get_NxN_type(),
                                         i == 0 ? NULL : intra_preds[0]->get_mode_y(),
                                         i == 0 ? NULL : intra_preds[0]->get_mode_u());
#else
  intra_pred = new IntraPrediction(h, w);       // pass the pointer of wcur to the IntraPrediction class.
#endif
  
  for (i = ZERO; i < 4; i++) 
  {
    quant[i] = new Quantizer(i);
  }
  
#ifdef READ_QTABLES_FROM_FILE
  FILE *f;
  f = fopen(param.qtabfile_4x4,"rt");
  if (f == NULL) 
  {
    printf("Could not open qtabfile_4x4: <%s>\n", param.qtabfile_4x4);
    exit(1);
  }
  for (i = ZERO; i < 4; i++) quant[i]->read_qtables(f, SIP4x4);
  fclose (f);
  
  f = fopen(param.qtabfile_2x8,"rt");
  if (f == NULL) 
  {
    printf("Could not open qtabfile_2x8: <%s>\n", param.qtabfile_2x8);
    exit(1);
  }
  for (i = ZERO; i < 4; i++) quant[i]->read_qtables(f, SIP2x8);
  fclose (f);
  
  f = fopen(param.qtabfile_8x2,"rt");
  if (f == NULL) 
  {
    printf("Could not open qtabfile_8x2: <%s>\n", param.qtabfile_8x2);
    exit(1);
  }
  for (i = ZERO; i < 4; i++) quant[i]->read_qtables(f, SIP8x2);
  fclose (f);
  
  if (param.trans8x8_enable) 
  {
    f = fopen(param.qtabfile_8x8,"rt");
    if (f == NULL) 
    {
      printf("Could not open qtabfile_8x8: <%s>\n", param.qtabfile_8x8);
      exit(1);
    }
    for (i = ZERO; i < 4; i++) quant[i]->read_qtables_8x8(f);
    fclose (f);
  }
#else
  for (i = 0; i < 4; i++)
  {
    quant[i]->copy_qtables(qtab4x4, SIP4x4);
    quant[i]->copy_qtables(qtab2x8, SIP2x8);
    quant[i]->copy_qtables(qtab8x2, SIP8x2);
    quant[i]->copy_qtables_8x8(qtab8x8);
  }
#endif
  
  lf = LoopFilter::create(h >> 2, w >> 2, param.ndf_mode);
  wlf = new WienerFilter(h, w, &param);
  dbf = new Deblocking_Filter(h, w);
  saif_half = new AIF_adaptive_filter (); 
  
  decoded_frames = ZERO;  
  reset_frame_type_stat();
  
}


#ifdef FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
 * \brief advance the superblock index of the substream we're working on.
 *
 * We've finished processing a superblock, we advance our counters.  We
 * signal that additional work is available if we notice we've just finished
 * a horizontal line of superblocks to account for our initial delay of 2
 * superblocks.
 *
 * \param[in] threadno what thread we are (unused).
 * \param[in] ss_no the substream we were working on.
 */
//-----------------------------------------------------------------------------
void Decoder::advance_ss_work(int threadno, int ss_no)
{
  (void)threadno; // unused parameter.
  if (!ss_work_info[ss_no].beingWorked)
    abort();
  // No lock required, the work-flag indicates this entry's not available.
  ss_work_info[ss_no].sb_j += 1;
  if (ss_work_info[ss_no].sb_j >= frame_width_in_sbs)
  {
    // We're starting another line -- we can post ss_down that there's
    // more available, corresponding to our initial 'delay'.  If there is
    // an ss_down.
    post_ss_work(threadno, ss_no, SUPERBLOCK_DELAY);
    ss_work_info[ss_no].sb_j = 0;
    ss_work_info[ss_no].sb_i += param.superblock_nsubstreams;
  }
  if (ss_work_info[ss_no].sb_i >= frame_height_in_sbs)
  {
    ss_work_info[ss_no].sb_avail = -1; // all done.
  }
  ss_work_info[ss_no].beingWorked = 0; // Data is no longer ours to play with.
}
#endif


#ifdef FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
 * \brief get some work to do, if we can.  We might wait if we're multithreading.
 *
 * We look at the work available and simply choose the first we find.  In a
 * multi-threading environment, we will wait if there's no work.
 *
 * \param[in] threadno what thread we are (unused).
 * \param[out] ss_no the substream that we will work on.
 *
 * \return 0 if no more work's available, and the thread should terminate.
 */
//-----------------------------------------------------------------------------
int Decoder::get_ss_work(int threadno, int &ss_no)
{
  (void)threadno; // unused parameter.
  // Want to find more work!
  while (1)
  {
    int work_avail = 0;
    int nworking = 0; //debug.
#ifdef SUBSTREAM_MT
    pthread_mutex_lock(&ss_work_info_mutex);
#endif
    // Find entry not being worked on, with work available.
    for (int ss = 0; ss < param.superblock_nsubstreams; ss++)
    {
      nworking += ss_work_info[ss].beingWorked;
      if (!ss_work_info[ss].beingWorked && ss_work_info[ss].sb_avail > 0)
      {
        ss_work_info[ss].beingWorked = 1;
        ss_work_info[ss].sb_avail -= 1;
        ss_no = ss;
#ifdef SUBSTREAM_MT
        pthread_mutex_unlock(&ss_work_info_mutex);
#endif
        return 1;
      }
      work_avail = work_avail || ss_work_info[ss].sb_avail >= 0;
    }
    
    // We wait, unless no work's available anywhere.  In that case, decoding's finished.
    if (ss_work_n_waiting_threads > param.decode_nthreads-nworking)
      abort();
    if (!work_avail)
    {
#ifdef SUBSTREAM_MT
      // Wake all threads for now.
      while (ss_work_n_waiting_threads > 0)
      {
        sem_post(&ss_work_waiting_threads);
        ss_work_n_waiting_threads--;
      }
      pthread_mutex_unlock(&ss_work_info_mutex);
#endif
      // Signal we're finished.
      return 0;
    }
#ifdef SUBSTREAM_MT
    // Possible future work... wait.
    ss_work_n_waiting_threads++;
    pthread_mutex_unlock(&ss_work_info_mutex);
    sem_wait(&ss_work_waiting_threads);
#else
    printf("wait-error!\n");
    abort();
#endif
  }
}
#endif


#ifdef FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
 * \brief Indicate to the 'next' substream that there is more work available.
 *
 * \param[in] threadno (unused).
 * \param[in] ss_src The index of the substream that was being processed.
 * \param[in] nwork The number of work-units to post -- normally 1.  2 at end-of-line.
 *
 * This is called just after a superblock has been initialized for processing,
 * and also called if we notice we've just completed a line of superblocks.
 * We permit a subsequent substream to decode another superblock.
 *
 * If we notice we're processing one of the first two superblocks in a line, no post
 * is done. This creates a 2-superblock 'delay' between subsequent substreams.
 * If, due to a post, we notice that work on the subsequent substream has
 * become available, we wake up a thread if we're in a multi-threading system.
 */
//-----------------------------------------------------------------------------
void Decoder::post_ss_work(int threadno, int ss_src, int nwork)
{
  (void)threadno; // unused parameter.
  int ss_down = ss_src == param.superblock_nsubstreams-1 ? 0 : ss_src+1;
  
  // We don't post for the initial 2 superblocks in a line.
  // We don't need to lock for accessing ss_src, it should be being worked on by us.
  if (ss_work_info[ss_src].sb_j < SUPERBLOCK_DELAY)
  {
    return;
  }
  
#ifdef SUBSTREAM_MT
  pthread_mutex_lock(&ss_work_info_mutex);
#endif
  // Post down if we have a line available below.
  if (ss_work_info[ss_src].sb_i+1 < frame_height_in_sbs)
    ss_work_info[ss_down].sb_avail += nwork;
  else
  {
  }
#ifdef SUBSTREAM_MT
  // If ss_down's come 'alive', see if a thread's available for working on it.
  if (!ss_work_info[ss_down].beingWorked && ss_work_info[ss_down].sb_avail == nwork && ss_work_n_waiting_threads > 0)
  {
    sem_post(&ss_work_waiting_threads);
    ss_work_n_waiting_threads--;
  }
#endif
  
#ifdef SUBSTREAM_MT
  pthread_mutex_unlock(&ss_work_info_mutex);
#endif
}
#endif

//-----------------------------------------------------------------------------
/*!
 * \brief Destructor
 */
//-----------------------------------------------------------------------------
Decoder::~Decoder() 
{
  int i;
  
#ifdef FRAME_SUBSTREAMS
  for (i = 0; i < param.decode_nthreads; i++)
    delete intra_preds[i];
  delete[] intra_preds;
#else
  delete intra_pred;
#endif
  
  delete mf;
#ifdef FRAME_SUBSTREAMS
  for (i = 0; i < param.superblock_nsubstreams; i++)
    delete edecs[i];
  delete[] edecs;
  for (i = 0; i < param.superblock_nsubstreams; i++)
    delete substream_bufs[i];
  delete[] substream_bufs;
  delete[] ss_work_info;
#else
  delete edec;
#endif
#ifdef SUBSTREAM_MT
  delete[] substream_args;
#endif
  delete zmap[ZERO];
  delete zmap[ONE ];
#ifdef FRAME_SUBSTREAMS
  delete[] mlcs;
#endif
  delete ref_mgr;
  delete lf;  
  delete wlf;
  delete dbf;
  delete saif_half; 
  delete trans[Transform::DCT];
  delete trans[Transform::AlternateIntra];
  delete trans[Transform::AlternateInter];
  delete wcur;
  delete wcur_pred;
  delete wcur_qpe;
  delete zmap_chroma[ZERO][ZERO]; 
  delete zmap_chroma[ZERO][ONE ]; 
  delete zmap_chroma[ONE][ZERO]; 
  delete zmap_chroma[ONE][ONE ]; 
  delete[] intra_map;
  
  for (i = ZERO; i < 4; i++)
    delete quant[i];
  
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

#ifdef SUBSTREAM_MT
void *launch_superblock_ss(void *p)
{
  // We're in a thread.
  // We invoke a decoder method to decode a substream.
  SubstreamThreadArgs *args = (SubstreamThreadArgs *)p;
  args->decoder->decode_ss(args->thr, args->cur);
  return NULL;
}
#endif

#ifdef FRAME_SUBSTREAMS
// NOTE: for the moment substream_buffer copies are reconstructed from the original frame_buf.
// LATER: we'll reuse the original framebuf, and have the ability to decode 'backwards from a buffer'.
void Decoder::find_substream_bufs(int first_ss, int n_ss, int bits_offset, int payload_bits, Buffer *frame_buf)
{
  int ss_bits_offset_coded_size; // # bits used to code the offset.
  int ss_bits_offset; // The offset.
  if (n_ss >= 3)
  {
    int n_left;
    int n_right;
    Buffer::divide_substreams(n_ss, n_left, n_right);
    ss_bits_offset_coded_size = Buffer::get_coded_offset_size(payload_bits);
    ss_bits_offset = frame_buf->read_bits_at(bits_offset, ss_bits_offset_coded_size);
    find_substream_bufs(first_ss, n_left, bits_offset+ss_bits_offset_coded_size, ss_bits_offset, frame_buf);
    find_substream_bufs(first_ss+n_left, n_right, bits_offset+ss_bits_offset_coded_size+ss_bits_offset, payload_bits-ss_bits_offset_coded_size-ss_bits_offset, frame_buf);
  }
  else if (n_ss == 2)
  {
    // Two substream buffers present, the second is reversed and concatenated with the first.
    substream_bufs[first_ss]->copy_bits_from(frame_buf, bits_offset, payload_bits);
    // If we're reading the final two substreams in the frame,
    // the final substream occupies the full final byte.
    // We copy the entire payload here -- we don't know the actual length.
    substream_bufs[first_ss+1]->copy_bits_backwards_from(frame_buf, bits_offset+payload_bits, payload_bits);
  }
  else if (n_ss == 1)
  {
    substream_bufs[first_ss]->copy_bits_from(frame_buf, bits_offset, payload_bits);
  }
}

void Decoder::init_substream_bufs(Buffer *frame_buf)
{
  find_substream_bufs(0, param.superblock_nsubstreams, 0, frame_buf->payload_size*8, frame_buf);
}
#endif

//-----------------------------------------------------------------------------
/*!
 * \brief Main decoding method
 *
 * Decodes one frame including header and updates reference frame buffer.
 *
 * \param[out] cur Reconstructed frame
 * \param[in] buf Byte buffer containing compressed frame
 */
//-----------------------------------------------------------------------------
#ifdef DEC_TRANSCODER
void Decoder::decode_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_filtered, Buffer *buf, Buffer *buf_out)
#else
void Decoder::decode_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_filtered, Buffer *buf)
#endif
{
  int i,j;
  
#ifdef FRAME_SUBSTREAMS
  // Find sub-buffers in main buffer.  For the moment, construct a new buffer for each one.
  init_substream_bufs(buf);
  for (i = 0; i < param.superblock_nsubstreams; i++)
#ifdef DEC_TRANSCODER
    edecs[i]->reset(substream_bufs[i], buf_out);
#else
  edecs[i]->reset(substream_bufs[i]);
#endif
#ifdef SUBSTREAM_MT
  pthread_setspecific(substream_number_key, 0); // Decode frame header with 1st substream.
#else
  edec_ss = 0;
#ifdef TRACE_DEBUG
  SET_TRACE_SS(0);
#endif
#endif
#else
#ifdef DEC_TRANSCODER
  edec->reset(buf, buf_out);
#else
  edec->reset(buf);
#endif
#endif
  
  while(edec->decode_rpps_type())
  {
    ref_mgr->temp_rpps_id[0]=edec->decode_rpps_id();
    for(i=0;i<LISTS;i++)
    {
      for(j=0;j<MAX_REF_FRAMES;j++)
      {
        ref_mgr->ref_process->inter_type[i][j]           = edec->decode_inter_type();
        ref_mgr->ref_process->filter_type[i][j]           = edec->decode_filter_type();
        ref_mgr->ref_process->ic_disable[i][j]=edec->decode_ic_disable();
        ref_mgr->ref_process->round_type[i][j]=edec->decode_round_type();
        
        ref_mgr->inter_type_set[ref_mgr->temp_rpps_id[0]][i][j]=ref_mgr->ref_process->inter_type[i][j];
        ref_mgr->filter_type_set[ref_mgr->temp_rpps_id[0]][i][j]=ref_mgr->ref_process->filter_type[i][j];
        ref_mgr->ic_disable_set[ref_mgr->temp_rpps_id[0]][i][j]=ref_mgr->ref_process->ic_disable[i][j];
        ref_mgr->round_type_set[ref_mgr->temp_rpps_id[0]][i][j]=ref_mgr->ref_process->round_type[i][j];
      }
    }
    ref_mgr->ref_pic_parameter_set_id_used[ref_mgr->temp_rpps_id[0]] = ONE;
    ref_mgr->ref_pic_parameter_set_id_shared[ref_mgr->temp_rpps_id[0]] = ONE;
  }
  ref_mgr->temp_rpps_id[0]=edec->decode_rpps_id();
  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      ref_mgr->ref_process->inter_type[i][j]=ref_mgr->inter_type_set[ref_mgr->temp_rpps_id[0]][i][j];
      ref_mgr->ref_process->filter_type[i][j]=ref_mgr->filter_type_set[ref_mgr->temp_rpps_id[0]][i][j];
      ref_mgr->ref_process->ic_disable[i][j]=ref_mgr->ic_disable_set[ref_mgr->temp_rpps_id[0]][i][j];
      ref_mgr->ref_process->round_type[i][j]=ref_mgr->round_type_set[ref_mgr->temp_rpps_id[0]][i][j];
    }
  }
  
  cur->frame_type          = edec->decode_frame_type();
  
  if (param.intra16x16_org)
  {
#ifdef I16_IN_P_FRAME
    param.intra16x16 = (cur->frame_type == ft_I_frame) || param.intra16x16p;
#else
    param.intra16x16 = (cur->frame_type == ft_I_frame);
#endif
  }
  
  if (cur->frame_type == ft_I_frame)
  {
    cur->intra_frame_type = edec->decode_I_frame_type();
  }
  cur->frame_order_counter = edec->decode_frame_counter();
  cur->reference_idc       = edec->decode_reference_idc();
  
  int new_frame_number;
  if (ref_mgr->deferred_refresh_flag)
    new_frame_number     = ref_mgr->deferred_refresh_frame_number + cur->frame_order_counter;
  else
    new_frame_number     = cur->frame_no + cur->frame_order_counter;
  
  //===============================
  //  Check for deferred refresh
  //===============================
  if (ref_mgr->deferred_refresh_flag)
  {
    if (cur->reference_idc==REFERENCE_PIC && new_frame_number>ref_mgr->deferred_refresh_frame_number)
    {
      /* !!! */
      ref_mgr->perform_deferred_refresh ();
      ref_mgr->deferred_refresh_flag=0;
      cur->qp = param.qp;
    }
  }
  
  if (cur->frame_type==ft_I_frame)
  {
    if (cur->intra_frame_type==ft_I_DDR)
    {
      ref_mgr->deferred_refresh_flag = 1;
      ref_mgr->deferred_refresh_frame_number = new_frame_number;
      cur->qp = param.qp;
    }
    else if (cur->intra_frame_type==ft_I_IDR)
    {
      ref_mgr->init ();
      cur->qp = param.qp;
    }
  }
  
  ref_mgr->process_list(cur, saif_half);
  
  cur->delta_qp            = edec->decode_deltaqp();
  
#ifdef Q_OFFSET
  cur->iTheta[0]           = edec->decode_q_offset_0();
  cur->iTheta[1]           = edec->decode_q_offset_1();
  cur->iTheta8x8[0]        = edec->decode_q_offset8x8_0();
  cur->iTheta8x8[1]        = edec->decode_q_offset8x8_1();
  for (i = 0; i < 4; i ++)
    quant[i]->setOffset(cur->iTheta, cur->iTheta8x8);
#endif
  
  edec->decode_quantization_set_id_half(); // cq_start
  edec->decode_qmatrix_half(0);
  edec->decode_qmatrix_half(1); // cq_end
  
  edec->decode_dummy();
  
  param.edge_prediction_ipd = edec->decode_edge_prediction_flag();
  if (param.edge_prediction_ipd)
  {
    param.threshold_edge_vector = edec->decode_threshold_edge();
  }
  
  // Lets update frame_no and qp 
  cur->frame_no = new_frame_number;
  cur->qp       += cur->delta_qp;
  cur->update_rd_params();
  
#ifdef TRACE_DEBUG
  INITFRAME_DTRACE(cur->frame_no);
  SVPRINT_DTRACE("FRAME TYPE", cur->frame_type);
  RPRINT_DTRACE;
#endif
  
  decode_one_frame(cur, cur_pred, cur_qpe_m, cur_qpe_s, cur_filtered);
  
#ifdef DEC_TRANSCODER
  edec->terminate();
#endif
  
#ifdef TRACE_STATS
  mf->write_stat_files(cur);
#endif
  
  frame_type_stat[cur->frame_type]++;
  total_frames++;
  decoded_frames++;
  
  param.intra16x16 = param.intra16x16_org;
  
  if (cur->reference_idc == FALSE)
  {
#ifndef MVCOMP_ZERORES
    ref_mgr->add_frame(cur);
#else
    ref_mgr->add_frame(cur, mf, &ref_mgr->ref_process->inter_pel[0][0], &ref_mgr->ref_process->log_inter_pel[0][0], NULL, NULL);
#endif
  }
  
}

#ifdef VERBOSE_STATS 
//-----------------------------------------------------------------------------
/*!
 * \brief Print entropy decoding statistics 
 *
 * \param[in,out] f File pointer
 */
//-----------------------------------------------------------------------------
void Decoder::print_stats_summary(FILE *f, int verbose)
{
#ifdef FRAME_SUBSTREAMS
  // We calculate the total # bits read here, for all substreams.
  int buffer_bits_read = 0;
  int buffer_payload_size = 0;
  for (int ss = 0; ss < param.superblock_nsubstreams; ss++)
  {
    buffer_bits_read += edecs[ss]->get_nbits();
    buffer_payload_size += edecs[ss]->get_payload_bits();
  }
  // Summed stats should always be in edecs[0] at the end of a frame.
  edecs[0]->print_stats_summary(f, verbose, buffer_payload_size, buffer_bits_read);
#else
  edec->print_stats_summary(f, verbose);
#endif
}
#else
//-----------------------------------------------------------------------------
/*!
 * \brief Print entropy decoding statistics 
 *
 * \param[in,out] f File pointer
 */
//-----------------------------------------------------------------------------
void Decoder::print_stats_summary(FILE *f)
{
#ifdef FRAME_SUBSTREAMS
  // Summed stats should always be in edecs[0] at the end of a frame.
  edecs[0]->print_stats_summary(f, verbose, buffer_payload_size, buffer_bits_read);
#else
  edec->print_stats_summary(f);
#endif
}
#endif

//-----------------------------------------------------------------------------
/*!
 * \brief Print entropy decoding statistics
 *
 * \param[in,out] f File pointer
 */
//-----------------------------------------------------------------------------
void Decoder::print_stats(FILE *f)
{
#ifdef FRAME_SUBSTREAMS
  // Summed stats should always be in edecs[0].
  edecs[0]->print_stats(f);
#else
  edec->print_stats(f);
#endif
}

//-----------------------------------------------------------------------------
/*!
 * \brief Reset frame type statistics
 *
 */
//-----------------------------------------------------------------------------
void Decoder::reset_frame_type_stat()
{
  int i;
  
  for (i = ZERO; i < 4; i++)
    frame_type_stat[i] = ZERO;
  
  total_frames = ZERO;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode motion data
 *
 * Decodes motion data including mode information for a target block in frame.
 *
 * \param[in] frame_type %Frame type (ft_I_frame, ft_P_frame, or fr_B_frame)
 * \param[in] i Vertical position of current 8x8 block
 * \param[in] j Horizontal position of current 8x8 block
 */
//-----------------------------------------------------------------------------
void Decoder::decodemvs_blk(int frame_type, int i, int j, bool largeIntra)
{
  int select;
  int mode;
  int ctx;
  int list;
  PredBlk pleft, ptop, ptleft;
  PredBlk *pred_left, *pred_top, *pred_tleft;
  PredBlk *curblk, *left, *top, *topleft;
#ifdef MVCOMP_ZERORES
  PredBlkSet predblkset;
  PredBlk **ptmp_preds, **ptmp_diff_preds;
  int n_diff_preds;
#endif
  
  int mvshare_size_ctx;
  int mvshare_type;
  
  curblk  = mf->get_pblk(i,   j  );
  left    = mf->get_pblk(i,   j-1);
  top     = mf->get_pblk(i-1, j  );
  topleft = mf->get_pblk(i-1, j-1);
  curblk->clear();
  
  if (frame_type == ft_I_frame || largeIntra)
  {
    mode = 0;
  }
  else
  {
    mode = edec->decode_mode(left, top, frame_type == ft_B_frame);
  }
  
  if (frame_type == ft_B_frame)
  {
    curblk->list = mode;
    
    if (mode != MODE_BI_PRED)
      curblk->list--;
    
    if (mode == MODE_INTRA)
      curblk->intra = &mf->blk_array_I[mf->width*i+j];
    intra_map[mf->width * i + j] = (mode == MODE_INTRA);    
  }
  else
  {  
    if (mode == MODE_P_INTRA)
    {
      curblk->list = -1;
      curblk->intra = &mf->blk_array_I[mf->width*i+j];
    }
    else
    {
      assert(mode == MODE_P_LIST_0_8x8);
      curblk->list = LIST_0;
    }
    
    intra_map[mf->width * i + j] = (mode == MODE_P_INTRA);
  }
  
  if (curblk->intra != NULL) 
  { // intra
    mlc.list_stat[4]++; 
  }
  else
  {      
    edec->decode_mv(curblk);
    
    list = curblk->list;
    mlc.list_stat[list]++; 
    
    if((curblk->mode()==MODE_LIST_0_8x8)||(curblk->mode()==BI_PRED))
    {
      curblk->ref_index[0] = edec->decode_refindex(ref_mgr->lsize[0], top, left, 0);
      mlc.ref_frame_stat[0][curblk->ref_index[0]]++;
    }
    if((curblk->mode()==MODE_LIST_1_8x8)||(curblk->mode()==BI_PRED))
    {
      curblk->ref_index[1] = edec->decode_refindex(ref_mgr->lsize[1], top, left, 1);
      mlc.ref_frame_stat[1][curblk->ref_index[1]]++;
    }
    
    left    = mf->get_pblk(i,  j-1);
    top     = mf->get_pblk(i-1,j  );
    topleft = mf->get_pblk(i-1,j-1);
    
#ifndef MVCOMP_ZERORES
#ifdef MV_CODING_ENHANCE
    
#ifdef PMV0_AT_OUT_OF_BOUNDARY
    pred_left = pleft.get_prediction(i,j,left, curblk, ref_mgr);
    pred_top = ptop.get_prediction(i,j,top,  curblk, ref_mgr);
#else
    pred_left = pleft.get_prediction(left, curblk, ref_mgr);
    pred_top = ptop.get_prediction(top,  curblk, ref_mgr);
#endif
    
#else
    pred_left = pleft.get_prediction(left, curblk);
    pred_top = ptop.get_prediction(top,  curblk);
#endif
#else
    ptmp_preds = predblkset.get_predictor_set_complete(ref_mgr, i, j, left, top, topleft, curblk);
#endif
#ifdef MV_CODING_ENHANCE
#ifdef PMV0_AT_OUT_OF_BOUNDARY
    pred_tleft = ptleft.get_prediction(i,j,topleft, curblk, ref_mgr);
#else
    pred_tleft = ptleft.get_prediction(topleft, curblk, ref_mgr);
#endif
    
#else
    pred_tleft = ptleft.get_prediction(topleft, curblk);
#endif
    
    
    if (pred_tleft == NULL)
      pred_tleft = &zero_mv;
    
#ifdef MVCOMP_ZERORES
    // 1st, reduce our set to those predictors that are different to those
    // that have come before.  NULL is not treated as a predictor.
    ptmp_diff_preds = predblkset.collapse_predictor_set(list, n_diff_preds);
    
    int extra_pred; // 1 indicates it's possible we want to signal 'temporal' (our third choice).
    predblkset.signal_treatment(n_diff_preds, extra_pred, list, edec->proba_mvp(2), edec->count_mvp(2, 0), edec->count_mvp(2, 1), 0);
    //predblkset.print(n_diff_preds, i, j);
    pred_left = n_diff_preds > 0 ? ptmp_diff_preds[0] : NULL;
    pred_top = n_diff_preds > 1 ? ptmp_diff_preds[1] : NULL;
#endif
    
#ifdef MV_CODING_ENHANCE
    if (pred_top == NULL && pred_left != NULL)
      curblk->add(pred_left, ref_mgr);
    else if (pred_top == NULL && pred_left == NULL)
      curblk->add(&zero_mv, ref_mgr);
    else if (pred_left == FALSE || pred_left->equal_pred(pred_top, list) == TRUE)
      curblk->add(pred_top, ref_mgr);
#else
    if (pred_top == NULL && pred_left != NULL)
      curblk->add(pred_left);
    else if (pred_top == NULL && pred_left == NULL)
      curblk->add(&zero_mv);
    else if (pred_left == FALSE || pred_left->equal_pred(pred_top, list) == TRUE)
      curblk->add(pred_top);
#endif
    else 
    {
      if (pred_tleft == NULL)
        pred_tleft = &zero_mv;
      
#ifndef MVCOMP_ZERORES
      ctx = pred_tleft->get_pred_context(pred_left, pred_top, list);
#else
      ctx = pred_tleft->get_pred_context(pred_left, pred_top, list, extra_pred);
#endif
      select = edec->decode_mvp(ctx);
#ifdef MV_CODING_ENHANCE
      curblk->add((select == ZERO) ? pred_left : pred_top, ref_mgr);
#else
      curblk->add((select == ZERO) ? pred_left : pred_top);
#endif
    }    
    
    // Decode information on motion sharing
    if (param.motion_sharing_enable)
    {
      mvshare_type = mf->mvshare_type(i, j);
      
      if (mvshare_type)
      {
        mvshare_size_ctx = mf->get_mvshare_size_ctx(mvshare_type, top, left);

        if (mvshare_type == 1)
        {
          int prev_val = (j == 0) ? 0 : left->mvshare_size[0];
          curblk->mvshare_size[0] = edec->decode_mvshare_size(mvshare_size_ctx, prev_val, 0, 1);
        }
        else if (mvshare_type == 2)
        {
          int prev_val = (i == 0) ? 0 : top->mvshare_size[1];
          curblk->mvshare_size[1] = edec->decode_mvshare_size(mvshare_size_ctx, 0, prev_val, 2);
        }
        else
        {
          int val = edec->decode_mvshare_size(mvshare_size_ctx, left->mvshare_size[0], top->mvshare_size[1], 4);
          curblk->mvshare_size[0] = (val > 10) ? 0 : val;
          curblk->mvshare_size[1] = (val > 10) ? val - 10 : 0;
        }
      }
    } // b_shfit_enable
  }
}


//-----------------------------------------------------------------------------
/*!
 * \brief Decode 4x4 luma block data with 4x4 inverse transform
 *
 * Decodes luma data including texture information for a target 4x4 block in frame.
 *
 * \param[in] i Vertical position of 8x8 block
 * \param[in] j Horizontal position of 8x8 block
 * \param[in] ii Vertical position of 4x4 sub-block in 8x8 block 
 * \param[in] jj Horizontal position of 4x4 sub-block in 8x8 block
 * \param[in] emodel Entropy model index (EntropyModel::I_luma, etc.)
 * \param[in] cur Current frame
 */
//-----------------------------------------------------------------------------
void Decoder::decode_luma_inter_block(int i0, int j0, int emodel)
{
  Block blk_pred(4, 4);
  Block blk(4, 4);
  int is_zero_blk;
  
  int i = i0>>2;
  int j = j0>>2;
  
  wcur_pred->get_blk_luma(&blk_pred, i, j);
  
#ifdef TRACE_DEBUG
  SPRINT_DTRACE("luma inter ");
  RPRINT_DTRACE;
#endif
  // residual decoding; blk
  is_zero_blk = zmap[ContextMap::inter]->decode(i, j, edec, emodel); 
  blk.zero();
  
  if (is_zero_blk == FALSE) 
  {
    edec->decode_coeffs(param.ztree[emodel], &blk, emodel);
  }
  
  zmap[ContextMap::inter]->set(i, j, blk.nnz());
  zmap[ContextMap::intra]->set(i, j, blk.nnz());
  
  quant[emodel]->loadmatrix(emodel, -1, 0, 0); // cq_start
  quant[emodel]->loadmatrix(emodel, -1, 0, 1); // cq_end
  
  int selected_transform = (param.transform_selection) ? blk.parity() : -1;

  quant[emodel]->iquant(&blk, 0);
  
  int transIdx = (selected_transform < 1) ? Transform::DCT : Transform::AlternateInter;
  trans[transIdx]->inverse(&blk);
  
  blk_pred.add(blk);                    
  
  wcur_qpe->put_blk_luma(&blk, i, j); 
  wcur->put_blk_luma(&blk_pred, i, j); // reconstructed signal
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode 8x8 luma block data with 8x8 inverse transform
 *
 * Decodes luma data including texture information for a target 8x8 block in frame.
 *
 * \param[in] i Vertical position of 8x8 block
 * \param[in] j Horizontal position of 8x8 block
 * \param[in] emodel Entropy model index (EntropyModel::I_luma, etc.)
 * \param[in] cur Current frame
 */
//-----------------------------------------------------------------------------
void Decoder::decode_luma_inter_block_8x8(int i, int j, int emodel)
{
  Block blk_pred8x8(8, 8);
  Block blk8x8(8, 8);
  int is_zero_blk;
  
  wcur_pred->get_blk_luma(&blk_pred8x8, i >> 1, j >> 1);
  
#ifdef TRACE_DEBUG
  SPRINT_DTRACE("luma inter8x8 ");
  RPRINT_DTRACE;
#endif
  
  is_zero_blk = zmap[ContextMap::inter]->decode(i, j, edec, emodel);    
  blk8x8.zero();
  
  if (is_zero_blk == FALSE) 
  {
    edec->decode_coeffs_8x8(param.ztree_64, &blk8x8);
  }

  zmap[ContextMap::inter]->set(i, j, blk8x8.nnz(), 2, 2);  
  zmap[ContextMap::intra]->set(i, j, blk8x8.nnz(), 2, 2);  
  
  quant[emodel]->loadmatrix(emodel, -1, 0, 0); // cq_start
  quant[emodel]->loadmatrix(emodel, -1, 0, 1); // cq_end
  
  int selected_transform = -1;
  if (param.transform_selection)
    selected_transform = blk8x8.parity();
  
  quant[emodel]->iquant(&blk8x8, 0);
  
  int transIdx = (selected_transform < 1) ? Transform::DCT : Transform::AlternateInter;
  trans[transIdx]->inverse(&blk8x8);
  
  wcur_qpe->put_blk_luma (&blk8x8, i>>1, j>>1);
  blk_pred8x8.add(blk8x8);
  wcur->put_blk_luma(&blk_pred8x8, i>>1, j>>1); // reconstructed signal
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode 4x4 chroma block data
 *
 * Decodes chroma data including texture information for a target 4x4 block in frame.
 *
 * \param[in] i Vertical position of 8x8 block
 * \param[in] j Horizontal position of 8x8 block
 * \param[in] c chroma component indicator
 * \param[in] emodel Entropy model index (EntropyModel::I_luma, etc.)
 */
//-----------------------------------------------------------------------------
void Decoder::decode_chroma_inter_block(int i, int j, int c, int emodel)
{
  Block blk_pred(4, 4);
  Block blk(4, 4);
  int is_zero_blk;
  
  blk.zero();
  
  // prediction; pred_blk
  wcur_pred->get_blk_chroma(&blk_pred, i, j, c);
  
#ifdef TRACE_DEBUG
  SPRINT_DTRACE("chroma inter ");
  RPRINT_DTRACE;
#endif
  // residual decoding
  is_zero_blk = zmap_chroma[c][ContextMap::inter]->decode(i, j, edec, emodel);
  
  if (is_zero_blk == FALSE) 
  {  
    edec->decode_coeffs(param.ztree[emodel], &blk, emodel);
  }
  
  zmap_chroma[c][ContextMap::inter]->set(i, j, blk.nnz());
  zmap_chroma[c][ContextMap::intra]->set(i, j, blk.nnz());
  
  quant[emodel]->loadmatrix(emodel, c, 0, 0); // cq_start
  quant[emodel]->loadmatrix(emodel, c, 0, 1); // cq_end
  
  quant[emodel]->iquant(&blk, 0);
  trans[Transform::DCT]->inverse(&blk);
  blk_pred.add(blk);                
  
  wcur_qpe->put_blk_chroma(&blk, i, j, c); 
  wcur->put_blk_chroma(&blk_pred, i, j, c);  // reconstructed signal 
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode one frame
 *
 * Includes the following steps:
 * - Construct reference lists
 * - Decode motion data
 * - Decode luma data
 * - Decode chroma data
 * - Generate motion predicted frame
 * - Add motion prediction
 * - Do inverse lapping
 * - Convert to 8 bits
 * - Apply loop filter
 *
 * \param[out] cur Reconstructed frame
 */
//-----------------------------------------------------------------------------
void Decoder::decode_one_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_filtered)
{
  int i;

  if(param.aif)
  {
    saif_half->ResetAdaptiveFilter_DEC();
  }
  
#ifdef FRAME_SUBSTREAMS
  for (int t = 0; t < param.decode_nthreads; t++)
    intra_preds[t]->reset_mode();
#else
  intra_pred->reset_mode();
#endif
  
  for (i = ZERO; i < FOUR; i++)
  {
    quant[i]->set_qp(cur->qp_matrix[i]);
    if (param.trans8x8_enable) 
    {
      quant[i]->set_qp8x8(cur->qp_matrix8x8[i]);
    }
    if (param.intra16x16)
      quant[i]->set_qp16x16(cur->qp_matrix16x16[i]);
  }
  
  if (cur->frame_type != ft_I_frame) 
  {
    ref_mgr->single_ref_frame = edec->decode_number_of_references();
    ref_mgr->prepare_list(cur->frame_type, LIST_0, cur->frame_no);
#ifdef MVCOMP_ZERORES
    param.mv_cur_comp_config_idx = edec->decode_mvcomp_config();
#endif
  }
  
  if (cur->frame_type == ft_B_frame)
  {
    ref_mgr->prepare_list(cur->frame_type, LIST_1, cur->frame_no);
  }
  
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  if (cur->frame_type != ft_I_frame) 
  {
    ref_mgr->standard_interpolation();
  }
#endif
  
  // AIF Filter creator
  if(param.aif)
  {
    if (cur->frame_type != ft_I_frame) 
    {
      saif_half->decode_filter_information (edec);
      if(saif_half->AIF_FilterFlag == 2)
        ref_mgr->AIF_replace_frame(cur, saif_half); 
    ref_mgr->getAIFcoef(saif_half);
    }    
  }
  // end; AIF Filter creator
  
  wcur->clear();
  wcur_pred->clear();
  wcur_qpe->clear();
  
  // do luma
  zmap[ContextMap::inter]->reset();
  zmap[ContextMap::intra]->reset();
  
  zmap_chroma[ZERO][ContextMap::inter]->reset(); 
  zmap_chroma[ZERO][ContextMap::intra]->reset(); 
  zmap_chroma[ONE][ContextMap::inter]->reset(); 
  zmap_chroma[ONE][ContextMap::intra]->reset(); 
  
#ifdef CABAC_SYNC_POINTS
#ifdef FRAME_SUBSTREAMS
  // reset_synchro on contexts
  for (int ss = 0; ss < param.superblock_nsubstreams; ss++)
    edecs[ss]->resetSynchro();
  // initialize substream info.
  for (int ss = 0; ss < param.superblock_nsubstreams; ss++)
  {
    ss_work_info[ss].beingWorked = 0;
    ss_work_info[ss].sb_avail = ss == 0 ? (cur->width4/ZHV+param.sb_width-1)/param.sb_width : 0;
    ss_work_info[ss].sb_i = ss;
    ss_work_info[ss].sb_j = 0;
    if (ss_work_info[ss].sb_i >= (cur->height4/ZHV+param.sb_height-1)/param.sb_height)
      ss_work_info[ss].sb_avail = -1; // Nothing to do in this substream.
  }
#ifdef SUBSTREAM_MT
  pthread_mutex_init(&ss_work_info_mutex, NULL);
  sem_init(&ss_work_waiting_threads, 0, 0);
  ss_work_n_waiting_threads = 0;
#endif
  frame_width_in_sbs = (cur->width8+param.sb_width-1)/param.sb_width;
  frame_height_in_sbs = (cur->height8+param.sb_height-1)/param.sb_height;
#else
  // reset_synchro on contexts
  edec->resetSynchro();
#endif
#endif // CABAC_SYNC_POINTS
  
#ifdef SUBSTREAM_MT
  // Launch the threads.
  pthread_t *tids = new pthread_t[param.decode_nthreads];
  for (int thr = 0; thr < param.decode_nthreads; thr++)
  {
    substream_args[thr].decoder = this;
    substream_args[thr].thr = thr;
    substream_args[thr].cur = cur;
    pthread_create(&tids[thr], NULL, ::launch_superblock_ss, &substream_args[thr]);
  }
  // Wait for threads to terminate.
  for (int thr = 0; thr < param.decode_nthreads; thr++)
  {
    pthread_join(tids[thr], NULL);
  }
  // Destroy mutexes and tids.
  pthread_mutex_destroy(&ss_work_info_mutex);
  sem_destroy(&ss_work_waiting_threads);
  delete[] tids;
  
  // We use the 1st substream and '1st thread' for decoding any remaining frame syntax.
  pthread_setspecific(thread_number_key, 0);
  pthread_setspecific(substream_number_key, 0);
#else
  // Run the original code, migrated to decode_ss.
  decode_ss(0, cur);
#ifdef FRAME_SUBSTREAMS
  // Use substream 0 for decoding end of frame.
  edec_ss = 0;
#endif
#ifdef TRACE_DEBUG
  SET_TRACE_SS(0);
#endif
#endif
#ifdef FRAME_SUBSTREAMS
  // Combine stream stats into edecs[0] on a per-frame basis.
  for (int ss = 1; ss < param.superblock_nsubstreams; ss++)
    edecs[0]->sum_stats_from(edecs[ss]);
#endif
  
#ifdef DEC_TRANSCODER
  // Use 1st substream to encode final frame syntax.
  edec->decodingFrameSyntax();
#endif
    
  if (param.wienerfilter)
  {
#ifdef TRACE_DEBUG
    DISABLE_DTRACE;
#endif
    
    wlf->decode_params(edec);
    
#ifdef TRACE_DEBUG
    ENABLE_DTRACE;
#endif
  }
  
  lf->set_luma(mf, zmap[ContextMap::inter]);
  lf->set_chroma(mf, zmap_chroma[0][ContextMap::inter], 0);  
  lf->set_chroma(mf, zmap_chroma[1][ContextMap::inter], 1);
  
  wcur->collapse(cur);  
  wcur_pred->collapse       (cur_pred);  
  wcur_qpe->collapse_modulus(cur_qpe_m);
  wcur_qpe->collapse_sign   (cur_qpe_s);
  
  if ((param.deblockingfilter&3) == 3)
  {
    dbf->filter_data_copy(lf, mf);
    
    dbf->init_alpha_values(cur->qp, cur->qpC);
    if (param.deblockingfilter&8)
    {
      dbf->set_filter_modes();
    }
    dbf->fixed_deblocking_filter( mf, cur->luma, cur->luma, 0);
    dbf->fixed_deblocking_filter( mf, cur->crcb[0], cur->crcb[0], 1);
    dbf->fixed_deblocking_filter( mf, cur->crcb[1], cur->crcb[1], 2);
  }
  
  lf->filter(cur);
  
  // Wiener fiter
  if (param.wienerfilter)
  {
    wlf->filter(cur, cur_filtered, cur_pred, cur_qpe_m, cur_qpe_s);
    
#ifdef WIENER_FILTER_INLOOP_SWITCH
    if (param.wienerfilter_inloop)
      cur->copy_samples(cur_filtered);
#endif
  }
  
  if ((param.deblockingfilter&3)==1)
  {
    dbf->filter_data_copy(lf, mf);

    dbf->init_alpha_values(cur->qp, cur->qpC);
    if (param.deblockingfilter&8)
    {
      dbf->set_filter_modes();
    }
    dbf->fixed_deblocking_filter( mf, cur->luma, cur->luma, 0);
    dbf->fixed_deblocking_filter( mf, cur->crcb[0], cur->crcb[0], 1);
    dbf->fixed_deblocking_filter( mf, cur->crcb[1], cur->crcb[1], 2);
    cur_filtered->copy(cur);
  }
}

//{{FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
 * \brief Decode macroblocks from substreams.
 * This contains the original macroblock decode code of a superblock.
 *
 * \param[in] thr Our thread number.
 * \param[in] cur The current frame.
 */
//-----------------------------------------------------------------------------
void Decoder::decode_ss(int thr, Frame *cur)
{  
#ifdef SUBSTREAM_MT
  pthread_setspecific(thread_number_key, (void *)thr);
#endif
#ifndef FRAME_SUBSTREAMS
  (void)thr; // unused parameter
#endif

  int q8i = ZHV * param.sb_height;
  int p8i = ZHV * param.sb_width;
  
#ifdef FRAME_SUBSTREAMS
  // We operate on a superblock basis.
  int ss = -1; // The substream we're working on.  Can change per superblock.
  while (1)
  {
    if (!get_ss_work(thr, ss))
      break;
#ifdef SUBSTREAM_MT
    pthread_setspecific(substream_number_key, (void *)ss);
#else
    edec_ss = ss;
#if defined(TRACE_DEBUG)
    SET_TRACE_SS(ss);
#endif
#endif
    
    // The entry ss_work_info[ss] is marked 'beingWorked' now, it's ours.
    // We can use and update its other fields except 'avail' without a lock.
    int q8 = ss_work_info[ss].sb_i * q8i;
    int p8 = ss_work_info[ss].sb_j * p8i;
    
#else // !FRAME_SUBSTREAMS
    // Original non-substream loops.  We decode all the superblocks in the frame.

    for (int q8=0; q8<cur->height4; q8+=q8i) 
    {
      
      for (int p8=0; p8<cur->width4; p8+=p8i) 
      {
#endif
        
#ifdef CABAC_SYNC_POINTS
        // init_synchro on contexts
        edec->initSynchro(q8/q8i, p8/p8i);
#endif
#ifdef FRAME_SUBSTREAMS
        // allow next thread to use data we're just leaving.
        post_ss_work(thr, ss, 1);
#endif
        
        decode_superblock(q8, p8, cur);
        
#ifdef CABAC_SYNC_POINTS
        // update_synchro on contexts
        edec->updateSynchro(p8/p8i);
#endif
#ifdef FRAME_SUBSTREAMS
        advance_ss_work(thr, ss);
#endif
#ifndef FRAME_SUBSTREAMS
      }
    }//superblock
#else
  } // FRAME_SUBSTREAMS: while
#endif
}
//FRAME_SUBSTREAMS}}

void Decoder::decode_superblock(int q8, int p8, Frame *cur)
{
  int i8s = q8;
  int i8l = q8 + ZHV * param.sb_height;
  int i8i = ZHV;
  
  int j8s = p8;
  int j8l = p8+ZHV*param.sb_width;
  int j8i = ZHV;
  int i8, j8;
  int ii, jj;
  int sip_type;
  
  if (param.intra16x16)
  {
    i8i = ZHV*2;
    j8i = ZHV*2;
  }
  
#ifdef TRACE_DEBUG
  INITBLOCK_DTRACE(p8, q8); 
#endif
  
  if (param.intra16x16)
  {
    int width_scale  = 2;
    int height_scale = 2;
    
    for (i8 = i8s; i8 < imin(i8l, cur->height4); i8 += i8i) 
    {
      for (j8 = j8s; j8 < imin(j8l, cur->width4); j8 += j8i) 
      {
        // decode luma
        intra_pred->avail_up_right_16x16 = (i8 == i8s) || (j8 != (j8l - j8i)); // top row or not last column

        // decode mvs_blk
        
        bool largeIntra = false;
        
        for (ii = 0; ii < height_scale; ii++)
        {
          for (jj = 0; jj < width_scale; jj++)
          {
            int u8 = (i8>>1) + ii;
            int v8 = (j8>>1) + jj;
            
            if (u8 >= cur->height8 || v8 >= cur->width8)
              continue;
            
            decodemvs_blk(cur->frame_type, u8, v8, largeIntra);
            
            if (mf->get_pblk(u8, v8)->intra)
            {
              if (largeIntra)
                continue;
              
#ifdef TRACE_DEBUG
              SPRINT_DTRACE("luma intra");
              RPRINT_DTRACE;
#endif
              sip_type = edec->decode_sip_type(intra_pred->get_sip_ctx(u8, v8));
              
              if (sip_type == SIP16x16)
              {                
                assert(ii == 0 && jj == 0);
                
                intra_pred->set_sip_type(u8, v8, sip_type, height_scale, width_scale);
                decode_intra_luma_blk16x16(u8>>1, v8>>1);
                zmap[ContextMap::inter]->copy_from_intra(zmap[ContextMap::intra], 2*u8, 2*v8, sip_type);
                
                if (cur->frame_type != ft_I_frame)
                {
                  decode_intra_chroma_blk(u8, v8, cur, SIP8x8);
                }
                
                largeIntra = true;
              }
              else
              {
                intra_pred->avail_up_right_8x8 = (jj == 0 || (intra_pred->avail_up_right_16x16 && ii == 0));
                intra_pred->set_sip_type(u8, v8, sip_type);
                
                int u, v;
                for (u = 0; u < 8; u += Block::bht[sip_type])
                {
                  for (v = 0; v < 8; v += Block::bwt[sip_type])
                  {
                    intra_pred->avail_up_right_4x4 = (v == 0 || (intra_pred->avail_up_right_8x8 && u == 0));
                    decode_intra_luma_blk(8*u8 + u, 8*v8 + v, sip_type);
                  }
                }
                
                zmap[ContextMap::inter]->copy_from_intra(zmap[ContextMap::intra], 2*u8, 2*v8, sip_type);
                
                if (cur->frame_type != ft_I_frame)
                {
                  decode_intra_chroma_blk(u8, v8, cur, SIP4x4);
                }
              }
            }
            else
            {
              mf->predict_blk(wcur_pred, ref_mgr, u8, v8);
              
              if (param.motion_sharing_enable && (mf->get_pblk_unsafe(u8, v8)->mvshare_size[0] || mf->get_pblk_unsafe(u8, v8)->mvshare_size[1]))
                mf->predict_blk_mvshare(wcur_pred, ref_mgr, u8, v8);
              
              sip_type = SIP_INTER;
              intra_pred->set_sip_type(u8, v8, sip_type);
              
              if (mf->get_transform_blocksize(u8, v8))
              {
                decode_luma_inter_block_8x8(2*u8, 2*v8,  EntropyModel::P_luma);
              }
              else
              {
                int u, v;
                for (u = 0; u < 8; u += 4)
                {
                  for (v = 0; v < 8; v += 4)
                  {
                    decode_luma_inter_block(8*u8 + u, 8*v8 + v, EntropyModel::P_luma);
                  }
                }
              }
              
              int c;              
              for (c = ZERO; c < TWO; c++) 
              {
                decode_chroma_inter_block(u8, v8, c, EntropyModel::P_chroma);
              }
            }
          }
        }
        
        if (cur->frame_type == ft_I_frame)
        {
          if (largeIntra)
          {
            decode_intra_chroma_blk(i8>>1, j8>>1, cur, SIP8x8);
          }
          else
          {
            for (ii = 0; ii < height_scale; ii++)
            {
              for (jj = 0; jj < width_scale; jj++)
              {
                if (((i8>>1)+ii >= cur->height8) || ((j8>>1)+jj >= cur->width8))
                  continue;
                
                decode_intra_chroma_blk((i8>>1)+ii, (j8>>1)+jj, cur, SIP4x4);
              }
            }
          }
        }
      } // j8
    } // i8
  }
  else
  {
    for (i8=i8s; i8<i8l; i8+=i8i) 
    {    
      for (j8=j8s; j8<j8l; j8+=j8i) 
      {
        if (i8 >= cur->height4 || j8 >= cur->width4)
          continue;
        
#ifdef TRACE_DEBUG
        INITBLOCK_DTRACE(j8, i8); 
#endif
        decodemvs_blk(cur->frame_type, (i8>>1), (j8>>1), false);
        
        int i = i8;
        int j = j8;        
        int intra  = intra_map[mf->width * ((i) >> ONE) + ((j) >> ONE)];
        
        if (intra == TRUE) 
        {
#ifdef TRACE_DEBUG
          SPRINT_DTRACE("luma intra");
          RPRINT_DTRACE;
#endif
          sip_type = edec->decode_sip_type(intra_pred->get_sip_ctx((i>>1),(j>>1)));
        }
        else 
        {
          mf->predict_blk(wcur_pred, ref_mgr, i>>1, j>>1);
          
          if (param.motion_sharing_enable && (mf->get_pblk_unsafe((i>>1), (j>>1))->mvshare_size[0] || mf->get_pblk_unsafe((i>>1), (j>>1))->mvshare_size[1]))
            mf->predict_blk_mvshare(wcur_pred, ref_mgr, i>>1, j>>1);
          
          sip_type = SIP_INTER;
        }

        intra_pred->set_sip_type((i>>1),(j>>1), sip_type);
        
        
        if (intra == FALSE && mf->get_transform_blocksize(i>>1, j>>1) == 0)
          sip_type = SIP_INTER+1;
        
        int blk_height = Block::bht[sip_type];
        int blk_width = Block::bwt[sip_type];
        
        for (ii = 4*i; ii < 4*i+8; ii += blk_height)
        {
          for (jj = 4*j ; jj < 4*j+8; jj += blk_width)
          {
            if (sip_type == SIP_INTER)
            {
              decode_luma_inter_block_8x8(ii >> 2, jj >> 2,  EntropyModel::P_luma);
            }
            else if (sip_type == SIP_INTER+1)
            {
              decode_luma_inter_block(ii, jj, EntropyModel::P_luma);
            }
            else
            {
              decode_intra_luma_blk(ii, jj, sip_type); 
            }
          }
        }
                     
        if (intra == TRUE && cur->frame_type!=ft_I_frame) 
        {
          zmap[ContextMap::inter]->copy_from_intra(zmap[ContextMap::intra], i, j, sip_type);
        }
        
        i =i8>>1; j=j8>>1;
        
        // chroma
        if (intra == TRUE)
        {
          decode_intra_chroma_blk(i, j, cur, SIP4x4);
        }
        else
        {
          int c;
          
          for (c = ZERO; c < TWO; c++) 
          {
            decode_chroma_inter_block(i, j, c, EntropyModel::P_chroma);
          }
        }          
      }
    }
  }
}

#ifdef FRAME_SUBSTREAMS
void Decoder::collect_mlc_stats()
{
  // Combine stats collected by threads.
  for (int t = 1; t < param.decode_nthreads; t++)
    mlcs[0].add(&mlcs[t]);
}
#endif

//-----------------------------------------------------------------------------
/*!
 * \brief Get frame type statistics
 *
 * Returns number of occurrences of a given frame type.
 *
 * \param[in] type %Frame type
 * \return Number of occurrences
 */
//-----------------------------------------------------------------------------
int Decoder::get_frame_type_stat(int type)
{
  return frame_type_stat[type];
}

//-----------------------------------------------------------------------------
/*!
 * \brief Get number of decoded frames
 *
 * \return number of decoded frames
 */
//-----------------------------------------------------------------------------
int Decoder::get_total_frames()
{
  return total_frames;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode luma block data
 *
 * Decodes luma data including texture information for a target 4x4 block in frame.
 *
 * \param[in] i0 Vertical position (sample unit)
 * \param[in] j0 Horizontal position (sample unit)
 * \param[in] cur Current frame
 * \param[in] sip_type Block type
 */
//-----------------------------------------------------------------------------
void Decoder::decode_intra_luma_blk(int i0, int j0, int sip_type) 
{
  int blk_height = Block::bht[sip_type];
  int blk_width = Block::bwt[sip_type];
  int blkIdx;
  int i, j;
  int iy = i0>>2, jy = j0>>2;
  int quantSipMode = sip_type;
  int size4 = 1;
  
  switch (sip_type)
  {
    case SIP2x8:
      i = i0>>1;
      j = j0>>3;
      blkIdx = i & 3;
      iy = i >> 1;
      jy = 2*j + (i & 1);
      break;
      
    case SIP8x2:
      i = i0>>3;
      j = j0>>1;
      blkIdx = j & 3;
      iy = 2*i + ((j&3)>>1);
      jy = 2*(j>>2) + (j&1);
      break;
      
    case SIP8x8:
      i = i0>>3;
      j = j0>>3;
      blkIdx = 0;
      quantSipMode = 0;
      size4 = 2;
      break;

    default: // SIP4x4
      assert(sip_type == SIP4x4);
      i = i0>>2;
      j = j0>>2;
      blkIdx = 2*(i&1) + (j&1);
      break;
  }
  
  Block blk_pred(blk_height, blk_width);
  Block blk(blk_height, blk_width);
  
  blk.zero();
  
  intra_pred->set_L_mode_pred_luma(blkIdx, sip_type);  
  
  int sip_mode = edec->decode_spatial_intra_prediction_luma(intra_pred->L_mode_pred_luma[blkIdx]);
  assert(sip_mode >= 0 && sip_mode < intra_pred->getNumPredModes(sip_type));
  
  if (sip_type == SIP8x8)
    intra_pred->put_mode_y_MxN(i0>>3, j0>>3, sip_mode, 1, 1);
  else
    intra_pred->put_mode_y(i0>>3, j0>>3, blkIdx, sip_mode);
  
  // get the prediction block
  intra_pred->generate_prediction(i, j, sip_type, sip_mode, &blk_pred, wcur);
  
  // Start decoding the block
  
  int is_zero_blk = zmap[ContextMap::intra]->decode(iy, jy, edec, EntropyModel::I_luma); 
  
  if (is_zero_blk == FALSE) 
  {
    if (sip_type == SIP8x8)
      edec->decode_coeffs_8x8(param.ztree_64, &blk);
    else
      edec->decode_coeffs_intra_luma_4x4(param.ztree[EntropyModel::I_luma], &blk, EntropyModel::I_luma, sip_type);
  }
  
  zmap[ContextMap::intra]->set(iy, jy, blk.nnz(), size4, size4);
  zmap[ContextMap::inter]->set(iy, jy, blk.nnz(), size4, size4);
  
  int selected_transform = (param.transform_selection) ? blk.parity() : -1;
  int transIdx = (selected_transform < 1) ? Transform::DCT : Transform::AlternateIntra;
  
  quant[EntropyModel::I_luma]->loadmatrix(EntropyModel::I_luma, -1, quantSipMode, 0); // cq_start
  quant[EntropyModel::I_luma]->loadmatrix(EntropyModel::I_luma, -1, quantSipMode, 1); // cq_end
  
  quant[EntropyModel::I_luma]->iquant(&blk, quantSipMode);  
  trans[transIdx]->inverse(&blk);
  
  wcur_pred->put_blk_luma(&blk_pred, i, j);
  wcur_qpe->put_blk_luma(&blk, i, j);
  blk.add(blk_pred);
  wcur->put_blk_luma(&blk, i, j);
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode 8x8 luma block data
 *
 * Decodes a target 8x8 luma block depending on the sip_type.
 *
 * \param[in] i Vertical position of 8x8 block
 * \param[in] j Horizontal position of 8x8 block
 * \param[in] sip_type the partition inside the 8x8 block
 * \param[in] cur Current frame 
 */
//-----------------------------------------------------------------------------
void Decoder::decode_luma_blocks_8x8(int i, int j, int sip_type, Frame *cur)
{
  int k, l;
  int ii, jj;
  int iii, jjj;
  
  // Loop over 8x8 blocks within 16x16 block
  for (k = 0; k < 2; k++)
  {
    iii = i + 2 * k;
    
    for (l = 0; l < 2; l++)
    {
      jjj = j + 2 * l;
      
      if (iii >= cur->height4 || jjj >= cur->width4)
        continue;
      
      // upper right block is available if either:
      // - first column (l == 0)
      // - first row and upper right block is available at 16x16 level
      intra_pred->avail_up_right_8x8 = (l == 0 || (intra_pred->avail_up_right_16x16 && k == 0));
      
      if (k + l != 0) 
      {
#ifdef TRACE_DEBUG
        SPRINT_DTRACE("luma intra");
        RPRINT_DTRACE;
#endif
        sip_type = edec->decode_sip_type(intra_pred->get_sip_ctx(iii >> 1, jjj >> 1));
      }
      
      intra_pred->set_sip_type(iii >> 1, jjj >> 1, sip_type);
      
      // decoding the 8x8 block
      int blk_height = Block::bht[sip_type];
      int blk_width = Block::bwt[sip_type];

      for (ii = 4*iii; ii < 4*iii+8; ii += blk_height)
      {
        for (jj = 4*jjj; jj < 4*jjj+8; jj += blk_width)
        {
          intra_pred->avail_up_right_4x4 = ((jj&7) == 0 || (intra_pred->avail_up_right_8x8 && (ii&7)==0));
          decode_intra_luma_blk(ii, jj, sip_type);
        }
      }      
    } // l
  } // k
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode 16x16 luma block data
 *
 * Decodes luma data including texture information for a target 16x16 block in frame.
 *
 * \param[in] i Vertical position of the 16x16 block in 16x16 unit
 * \param[in] j Horizontal position of the 16x16 block in 16x16 unit
 * \param[in] cur Current frame 
 */
//-----------------------------------------------------------------------------
void Decoder::decode_intra_luma_blk16x16(int i, int j) 
{
  int is_zero_blk, bsize = 16;
  Block blk16x16(bsize, bsize);
  Block blk_pre16x16(bsize, bsize);  
  
  intra_pred->set_L_mode_pred_luma(0, SIP16x16);  
  int sip_mode = edec->decode_spatial_intra_prediction_luma(intra_pred->L_mode_pred_luma[0]);
  assert(sip_mode >= 0 && sip_mode < intra_pred->getNumPredModes(SIP16x16));
  intra_pred->put_mode_y_MxN(2*i, 2*j, sip_mode, 2, 2);
  
  intra_pred->generate_prediction(i, j, SIP16x16, sip_mode, &blk_pre16x16, wcur);

  int selected_transform = -1;
    
  is_zero_blk = zmap[ContextMap::intra]->decode(4*i, 4*j, edec,EntropyModel::I_luma);
  
  blk16x16.zero();
  if (is_zero_blk == FALSE) 
  {
    if (param.transform_selection_16x16)
      selected_transform = edec->decode_tsfm_flag();
    edec->decode_coeffs_16x16(&param.ztree[NZTREES-NSUBZTREES], &blk16x16);
  }   
  
  zmap[ContextMap::intra]->set(4*i, 4*j, blk16x16.nnz(), 4, 4);
  
  quant[EntropyModel::I_luma]->loadmatrix(EntropyModel::I_luma, -1, 0, 0); // cq_start
  quant[EntropyModel::I_luma]->loadmatrix(EntropyModel::I_luma, -1, 0, 1); // cq_end
  
  quant[EntropyModel::I_luma]->iquant16x16(&blk16x16);
  
  int transIdx = (selected_transform < 1) ? Transform::DCT : Transform::AlternateIntra;
  trans[transIdx]->inverse_transform(&blk16x16);
  
  wcur_pred->put_blk_luma(&blk_pre16x16, i, j);
  wcur_qpe->put_blk_luma(&blk16x16, i, j);
  blk16x16.add(blk_pre16x16);
  wcur->put_blk_luma(&blk16x16, i, j);
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode intra chroma block data
 *
 * Decodes chroma data including texture information for a target 4x4 block in frame.
 *
 * \param[in] i Vertical position of 4x4 block
 * \param[in] j Horizontal position of 4x4 block
 * \param[in] cur Current frame 
 */
//-----------------------------------------------------------------------------
void Decoder::decode_intra_chroma_blk(int i, int j, Frame *cur, int sip_type)
{
  int blk_width = Block::bwt[sip_type];
  int blk_height = Block::bht[sip_type];
  int c;
  int is_zero_blk;
  int sip_mode = INT_INF;
  int up_mode = INT_INF;
  int left_mode = INT_INF;
  int up_is_inter = INT_INF;
  int left_is_inter = INT_INF;
  Block blk_pred(blk_height, blk_width);
  Block blk(blk_height, blk_width);
  int qpC = (sip_type == SIP8x8) ? cur->qpC : 0;
  int bsize4 = (sip_type == SIP8x8) ? 2 : 1;
  int in = (sip_type == SIP8x8) ? i >> 1 : i;
  int jn = (sip_type == SIP8x8) ? j >> 1 : j;
#ifdef TRACE_DEBUG
  SPRINT_DTRACE("chroma intra");
  RPRINT_DTRACE;
#endif 
  
  assert(intra_map[mf->width * i + j] == TRUE);
  
  if (i > 0) 
  {
    up_mode = intra_pred->get_mode_u(i-1, j);     //  mode_u[(i - 1) * cur->width8 + j];
    up_is_inter = !intra_map[cur->width8*(i - 1) + j];
  }
  if (j > 0) 
  {
    left_mode = intra_pred->get_mode_u(i, j-1);    // mode_u[i * cur->width8 + (j - 1)];
    left_is_inter = !intra_map[cur->width8*i + (j-1)]; 
  }
  
  sip_mode = edec->decode_spatial_intra_prediction_chroma(up_mode, left_mode, up_is_inter, left_is_inter);
  assert(sip_mode >= 0);
  
  intra_pred->put_mode_u(i, j, sip_mode);    // mode_u[i * cur->width8 + j] = sip_mode;
  if (sip_type == SIP8x8)
  {
    intra_pred->put_mode_u(i, j+1, sip_mode);
    intra_pred->put_mode_u(i+1, j, sip_mode);
    intra_pred->put_mode_u(i+1, j+1, sip_mode);
  }
  int emodel = EntropyModel::I_chroma;
  
  for (c = ZERO; c < TWO; c++) 
  {
    blk.zero();
    
    intra_pred->generate_prediction_chroma(in, jn, c, sip_type, sip_mode, &blk_pred, qpC, wcur);
    
    is_zero_blk = zmap_chroma[c][ContextMap::intra]->decode(i, j, edec, emodel);
    
    if (is_zero_blk == FALSE)
    {
      if (sip_type == SIP8x8)
        edec->decode_coeffs_8x8(param.ztree_64_CT8, &blk, 1); 
      else
        edec->decode_coeffs(param.ztree[emodel], &blk, emodel); 
    }
    
    quant[emodel]->loadmatrix(emodel, c, 0, 0); // cq_start
    quant[emodel]->loadmatrix(emodel, c, 0, 1); // cq_end
    
    zmap_chroma[c][ContextMap::intra]->set(i, j, blk.nnz(), bsize4, bsize4);
    zmap_chroma[c][ContextMap::inter]->set(i, j, blk.nnz(), bsize4, bsize4);
    
    quant[emodel]->iquant(&blk, 0);
    trans[Transform::DCT]->inverse(&blk);
    
    wcur_pred->put_blk_chroma(&blk_pred, in, jn, c);
    wcur_qpe->put_blk_chroma(&blk, in, jn, c);
    blk.add(blk_pred);
    wcur->put_blk_chroma(&blk, in, jn, c);
  } // c       
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode 8x8 chroma block data
 *
 * Decodes chroma data including texture information for a target 8x8 block depending on the luma sip_type.
 *
 * \param[in] i Vertical position of 8x8 block in 4x4 unit
 * \param[in] j Horizontal position of 8x8 block in 4x4 unit
 * \param[in] cur Current frame 
 */
//-----------------------------------------------------------------------------
void Decoder::decode_intra_chroma_blk8x8(int i, int j, Frame *cur)
{
  int sip_type = intra_pred->get_sip_type(i, j);
  int chroma_sip_type = (param.chroma8x8 && sip_type == SIP16x16) ? SIP8x8 : SIP4x4;
  int k, l;
  
  for (k = 4*i; k < imin(4*i+8, cur->height2); k += Block::bht[chroma_sip_type])
  {
    for (l = 4*j; l < imin(4*j+8, cur->width2); l += Block::bwt[chroma_sip_type])
    {
      decode_intra_chroma_blk(k>>2, l>>2, cur, chroma_sip_type);
    }
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
