/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
 * \file DenoiseThreshold.cpp
 *
 * \brief Block thresholding methods
 *
 */
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "DenoiseThreshold.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------------------
/*!
 * \brief Create threshold object based on desired attributes
 *
 * \param[in] size          Size of block to threshold (e.g. 8 for an 8x8 block)
 * \param[in] use_weighting Flag to enable weighting of blocks based on number of 0 coeffs
 * \param[in] use_oracle    Flag to enable optimal thresholding w.r.t target
 *
 * \return Threshold object
 */
//-----------------------------------------------------------------------------------------
DenoiseThreshold *DenoiseThreshold::create(int size, bool use_weighting, bool use_oracle)
{
  assert(size > 0);
  
  DenoiseThreshold *result = NULL;
  
  if (use_oracle)
  {
    result = new DenoiseThresholdOracle(size);
  }
  else if (use_weighting)
  {
    result = new DenoiseThresholdWeighted(size);
  }
  else
  {
    result = new DenoiseThreshold(size);
  }
  
  assert(result != NULL);
  
  return result;
}


//-----------------------------------------------------------------------------------------
/*!
 * \brief Constructor
 *
 * \param[in] size          Size of block to threshold (e.g. 8 for an 8x8 block)
 */
//-----------------------------------------------------------------------------------------
DenoiseThreshold::DenoiseThreshold(int size)
{
  assert(size > 0);
  
  m_size2 = size * size;
}

DenoiseThreshold::~DenoiseThreshold()
{
}

//-----------------------------------------------------------------------------------------
/*!
 * \brief Populate weight table
 *
 * Weight table is fixed-point representation of weight[0..i..n-1] = 1 / (n-i)
 *
 * \param[in] bits          Precision used for weight values
 */
//-----------------------------------------------------------------------------------------
void DenoiseThreshold::setWeights(int bits)
{
  assert(bits > 0);
  
  int i;
  int max_num = (1 << bits) - 1;
  
  for (i = 0; i < m_size2; i++) 
  {
    int val = (max_num + (m_size2 - i) / 2) / (m_size2 - i);
    val = imax(val, 1); // ensure weight is nonzero
    
    m_weights[i] = val;
  }
}

//-----------------------------------------------------------------------------------------
/*!
 * \brief Threshold block of data
 *
 * Set to 0 samples in block of data that are smaller than or equal to a threshold
 * Note: first sample is skiped (DC)
 *
 * \param[in,out] data    Block of data
 * \param[in]     th      Threshold
 *
 * \return weight = 1
 */
//-----------------------------------------------------------------------------------------
int DenoiseThreshold::go(int *data, int th)
{
  assert(data != NULL);
  assert(th >= 0);
  
  int i;
  
  for (i = 1; i < m_size2; i++)
  {
    if (abs(data[i]) <= th)
    {
      data[i] = 0;
    }
  }
  
  return 1;
}

//-----------------------------------------------------------------------------------------
/*!
 * \brief Threshold block of data
 *
 * Set to 0 samples in block of data that are smaller than or equal to a threshold
 * Note: first sample is skiped (DC)
 *
 * \param[out] data    Block of data
 * \param[in]  ref     Reference block of data
 * \param[in]  th      Threshold
 *
 * \return weight = 1
 */
//-----------------------------------------------------------------------------------------
int DenoiseThreshold::go(int *data, const int *ref, int th)
{
  int i;
  
  for (i = 1; i < m_size2; i++)
  {
    if (abs(ref[i]) <= th)
    {
      data[i] = 0;
    }
  }
  
  return 1;
}


//-----------------------------------------------------------------------------------------
/*!
 * \brief Constructor
 *
 * \param[in] size          Size of block to threshold (e.g. 8 for an 8x8 block)
 */
//-----------------------------------------------------------------------------------------
DenoiseThresholdWeighted::DenoiseThresholdWeighted(int size) : DenoiseThreshold(size)
{
}

int DenoiseThresholdWeighted::go(int *data, int th)
{
  int i;
  
#if 1
  int count = m_size2 - 1;
  for (i = 1; i < m_size2; i++)
  {
    int val = data[i];
    int sign = val >> 31;
    int delta = (th + sign - (val ^ sign)) >> 31;
    // int delta = (th - abs(val)) >> 31;
    val &= delta;
    count += delta;
    data[i] = val;
  }
#else
  int count = 0;
  for (i = 1; i < m_size2; i++)
    {
    if (abs(data[i]) <= th)
    {
      data[i] = 0;
      count++;
    }
  }
#endif
  
  int weight = m_weights[count];
  
  for (i = 0; i < m_size2; i++)
  {
    data[i] *= weight;
  }
  
  return weight;
}

int DenoiseThresholdWeighted::go(int *data, const int *ref, int th)
{
  int i;
  int count = 0;
  
  for (i = 1; i < m_size2; i++)
  {
    if (abs(ref[i]) <= th)
    {
      data[i] = 0;
      count++;
    }
  }

  int weight = m_weights[count];
  
  for (i = 0; i < m_size2; i++)
  {
    data[i] *= weight;
  }
  
  return weight;
}



DenoiseThresholdOracle::DenoiseThresholdOracle(int size) : DenoiseThreshold(size)
{
}

int DenoiseThresholdOracle::go(int *data, const int *ref, int /*th*/)
{
  int i;
  int count = 0;
  
  for (i = 1; i < m_size2; i++)
  {
    if (abs(ref[i] - 0) < abs(ref[i] - data[i]))
    {
      data[i] = 0;
      count++;
    }
  }
  
  return count;
}

