/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
 * \file DenoiseTransform.cpp
 *
 * \brief Implementation of DenoiseTransform class
 *
 */
//-----------------------------------------------------------------------------

#include "DenoiseTransform.H"
#include "DenoiseTransformHadamard4.H"
#include "DenoiseTransformHadamard8.H"
#include "DenoiseTransformDCT4.H"
#include "DenoiseTransformDCT8.H"


DenoiseTransform::DenoiseTransform(int width, int size, int norm, int energy, bool has_alternate)
{
  m_size = size;
  m_norm = norm;
  m_has_alternate = has_alternate;
  m_basis_energy = energy;
  m_width = width;

  m_buffer_fwd = new int[width * size];
  m_buffer_inv = new int[width * size];
  m_block = new int[size * size];
  
  if (has_alternate)
  {
    m_buffer_alt = new int[width * size];
    m_block_alt = new int[size * size];
  }

  m_mask = new int[width];
}

DenoiseTransform::~DenoiseTransform()
{
  delete[] m_buffer_fwd;
  delete[] m_buffer_inv;
  delete[] m_block;
  
  if (m_has_alternate)
  {
    delete[] m_buffer_alt;
    delete[] m_block_alt;
  }

  delete[] m_mask;
}

int DenoiseTransform::getSize()
{
  return m_size;
}

int DenoiseTransform::getNorm()
{
  return m_norm;
}

int DenoiseTransform::getBasisEnergy()
{
  return m_basis_energy;
}

int *DenoiseTransform::getBlock(bool use_alternate)
{
  return (use_alternate) ? m_block_alt : m_block;
}

void DenoiseTransform::resetAccumBuffer(int col_begin, int col_end)
{
  memset(m_buffer_inv + col_begin * m_size, 0, (col_end - col_begin + 1) * m_size * sizeof(*m_buffer_inv));
}

void DenoiseTransform::scaleBlock(int scale)
{
  int i;
  
  for (i = 0; i < m_size * m_size; i++)
  {
    m_block[i] *= scale;
  }
}

bool DenoiseTransform::isZeroBlock()
{
  int i;
  
  for (i = 1; i < m_size * m_size; i++)
  {
    if (m_block[i] != 0)
      return false;
  }
  
  return true;
}

bool DenoiseTransform::setMask(const int *th_row)
{
  int i;
  int count = 0;
  
  for (i = 0; i < m_width; i++)
  {
    if (th_row[i])
      count = -m_size;

    m_mask[i] = count >> 31;
    count++;
  }
  
  return count == m_width;
}

DenoiseTransform *DenoiseTransform::create(int tsize, bool hadamard, int width)
{
  if (tsize == 4 && hadamard)
  {
    return new DenoiseTransformHadamard4(width, true);
  }
  else if (tsize == 4)
  {
    return new DenoiseTransformDCT4(width, true);
  }
  else if (tsize == 8 && hadamard)
  {
    return new DenoiseTransformHadamard8(width, true);
  }
  else
  {
    return new DenoiseTransformDCT8(width, true);
  }
}
