/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Encoder.cpp
*
* \brief Implementation of Encoder class
*
*
*/


//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Encoder.H"
#include "MLC_Stat.H"
#include "ContextMap.H"
#include "Param.H"
#include "Global.H"
#include "FastDCT.H"
#include "TransformAlternate.H"
#include "IEncoder6.H"
#include "WFrame.H"
#include "MotionEst.H"
#include "MotionEstUZSearchBiPred.H"
#include "MotionEstRefine.H"
#include "MotionEstRegion.H"
#include "MotionEstTrellis.H"
#include "AIF_adaptive_filter.H"

#ifdef TRACE_DEBUG
#include "Trace.H"
#endif

#include "Deblocking_Filter.H"


#if defined(SUBSTREAM_MT) && defined(TRACE_DEBUG)
#include <pthread.h>
pthread_key_t substream_number_key;
#endif

#if defined(SUBSTREAM_MT) || defined(DENOISE_MT) || defined(WIENER_MT) || defined(DEBLOCK_MT)
// Not used in the encoder, but required for linking.
#include <pthread.h>
pthread_key_t thread_number_key;
#endif

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Allocates memory for entropy encoder, motion field, reference frame manager,
* zero block maps, intra maps, loop filter, lapped transform, transform,
* quantizers, I-block encoder, byte buffers and frames.
*
* \param[in] h Height of frame in luma sample units
* \param[in] w Width of frame in luma sample units
*/
//-----------------------------------------------------------------------------
Encoder::Encoder(int h, int w) 
{
  int i, j, k;
  FILE *f;

  frame_height = h;
  frame_width = w;

  eenc = new EntropyEncoder;
  mf = new MotionField(h, w);

  pred_frame = new WFrame(h, w);

  ref_mgr = new RefFrameManager(h, w);

  buf_tmp = new Buffer;
  frm_tmp = new Frame(h, w);
  frm_tmp_filtered = new Frame(h, w);
  frm_tmp_pred     = new Frame(h, w);
  frm_tmp_qpe_m    = new Frame(h, w);
  frm_tmp_qpe_s    = new Frame(h, w);  
  tmp_MLC = new MLC_Stat;
  buf_best[0] = new Buffer;                // For storing I
  frm_best[0] = new Frame(h, w);
#ifdef MVCOMP_ZERORES
  mf_best[0]  = new MotionField(h, w);
#endif
  frm_best_filtered  [0] = new Frame(h, w);
  frm_best_pred      [0] = new Frame(h, w);
  frm_best_qpe_m     [0] = new Frame(h, w);
  frm_best_qpe_s     [0] = new Frame(h, w);
  best_MLC[0] = new MLC_Stat;
  buf_best[1] = new Buffer;                // For storing P & B
  frm_best[1] = new Frame(h, w);
#ifdef MVCOMP_ZERORES
  mf_best[1]  = new MotionField(h, w);
#endif
  eenc_best[0] = new EntropyEncoder;
  eenc_best[1] = new EntropyEncoder;
  frm_best_filtered  [1] = new Frame(h, w);
  frm_best_pred      [1] = new Frame(h, w);
  frm_best_qpe_m     [1] = new Frame(h, w);
  frm_best_qpe_s     [1] = new Frame(h, w);
  best_MLC[1] = new MLC_Stat;

  // create zero block maps
  zmap[ZERO] = new ContextMap(frame_height>>2, frame_width>>2);
  zmap[ONE ] = new ContextMap(frame_height>>2, frame_width>>2);

  zmap_chroma[ZERO][ZERO] = new ContextMap(frame_height>>3, frame_width>>3); 
  zmap_chroma[ZERO][ONE ] = new ContextMap(frame_height>>3, frame_width>>3); 
  zmap_chroma[ONE][ZERO] = new ContextMap(frame_height>>3, frame_width>>3); 
  zmap_chroma[ONE][ONE ] = new ContextMap(frame_height>>3, frame_width>>3); 

  for (i = 0; i < 3; i++)
  {
    stored_eenc[i] = new EntropyEncoder;
    stored_buff[i] = new Buffer;
    stored_eenc[i]->reset(stored_buff[i]);

    stored_ztree_luma_P[i] = new ZeroTree(param.ztree_pattern[EntropyModel::P_luma]);
    if (param.trans8x8_enable) 
    {
      stored_ztree_luma_P_8x8[i] = new ZeroTree(param.ztree_pattern_64);
    }
  }
  
  for (i = 0; i < 3; i++)
  {
    for (j = 0; j < 2; j++)
    {
      for (k = 0; k < 2; k++)
      {
        stored_eencF[i][j][k] = new EntropyEncoder;
        stored_ztreeF_P[i][j][k] = new ZeroTree(param.ztree_pattern[4]);
      }
      stored_ztreeF_B[i][j] = new ZeroTree(param.ztree_pattern[5]);
    }
  }

  stored_ztree_luma_I = new ZeroTree(param.ztree_pattern[EntropyModel::I_luma]);
  stored_ztree_chroma_I = new ZeroTree(param.ztree_pattern[EntropyModel::I_chroma]);
  stored_ztree_chroma_P = new ZeroTree(param.ztree_pattern[EntropyModel::P_chroma]);
  stored_ztree_mv_P = new ZeroTree(param.ztree_pattern[4]);
  stored_ztree_mv_B = new ZeroTree(param.ztree_pattern[5]);

  lf = LoopFilter::create(frame_height>>2, frame_width>>2, param.ndf_mode);
  wlf = new WienerFilter(frame_height, frame_width, &param);  

  dbf = new Deblocking_Filter(frame_height, frame_width);
  saif_half = new AIF_adaptive_filter (); 

  trans[Transform::DCT] = new FastDCT;
  trans[Transform::AlternateIntra] = new TransformAlternate(ft_I_frame);
  trans[Transform::AlternateInter] = new TransformAlternate(ft_P_frame, param.alt8x8_trans_idx);
  
  coded_frames = 0;
  coded_inter_frames = 0;
  htotal_bits = 0;
  total_bits = 0;
  total_sad = 0;
  total_motion_bits = 0;
  intra_cost_update=0;
  num_loop_check=0;
  mlc_tot.reset();

  last_coded_frame = NULL;

  // create quantizer objects
  for (i=0; i<4; i++)
    quant[i] = new Quantizer(i);

  // read the entire inverse quant and forward quant table, where coeff norm is needed it is also read.
  f = fopen(param.qtabfile_4x4,"rt");
  if (f == NULL) 
  {
    printf("Could not open qtabfile_4x4: <%s>\n", param.qtabfile_4x4);
    exit(1);
  }
  for (i = ZERO; i < 4; i++) quant[i]->read_qtables(f, SIP4x4);
  fclose (f);

  f = fopen(param.qtabfile_2x8,"rt");
  if (f == NULL) 
  {
    printf("Could not open qtabfile_2x8: <%s>\n", param.qtabfile_2x8);
    exit(1);
  }
  for (i = ZERO; i < 4; i++) quant[i]->read_qtables(f, SIP2x8);
  fclose (f);

  f = fopen(param.qtabfile_8x2,"rt");
  if (f == NULL) 
  {
    printf("Could not open qtabfile_8x2: <%s>\n", param.qtabfile_8x2);
    exit(1);
  }
  for (i = ZERO; i < 4; i++) quant[i]->read_qtables(f, SIP8x2);
  fclose (f);

  if (param.trans8x8_enable) 
  {
    f = fopen(param.qtabfile_8x8,"rt");
    if (f == NULL) 
    {
      printf("Could not open qtabfile_8x8: <%s>\n", param.qtabfile_8x8);
      exit(1);
    }
    for (i = ZERO; i < 4; i++) quant[i]->read_qtables_8x8(f);
    fclose (f);
  }

  m_ienc = new IEncoder6(frame_height, frame_width, quant[0], quant[1], zmap[1], zmap_chroma[0][1], zmap_chroma[1][1]);
  m_ienc->setMotionField(mf);

#if defined(SUBSTREAM_MT) && defined(TRACE_DEBUG)
  // Just set the global substream number to 0 for tracing -- we only encode one stream.
  pthread_key_create(&substream_number_key, 0);
#endif
  MotionEstUZSearchBiPred::initPrevMf(h, w);
  
  m_cur2 = new WFrame(h, w);
  m_target = new WFrame(h, w);
  m_cur2_pred = new WFrame(h, w);
  m_cur2_qpe  = new WFrame(h, w);

  m_me[0] = new MotionEstUZSearchBiPred(mf, ref_mgr);
  m_me[1] = new MotionEstRefine(mf, ref_mgr);
  m_me[2] = new MotionEstTrellis(mf, ref_mgr);
  m_me[3] = new MotionEstRegion(mf, ref_mgr);
  
#ifdef I16_IN_P_FRAME
  for (i = 0; i < 6; i++)
  {
    m_tmpZeroTree16x16[i] = new ZeroTree(param.ztree_pattern[i]);
  }
  m_tmpZeroTree16x16[6] = new ZeroTree(param.ztree_pattern_64);
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory allocated by constructor.
*/
//-----------------------------------------------------------------------------
Encoder::~Encoder() 
{
  MotionEstUZSearchBiPred::freePrevMf();
  int i, j, k;

  delete buf_tmp;
  delete frm_tmp;
  delete frm_tmp_filtered;
  delete frm_tmp_pred;
  delete frm_tmp_qpe_m;
  delete frm_tmp_qpe_s;
  delete tmp_MLC;
  delete buf_best[0];
  delete frm_best[0];
#ifdef MVCOMP_ZERORES
  delete mf_best[0];
#endif
  delete best_MLC[0];
  delete buf_best[1];
  delete frm_best[1];
#ifdef MVCOMP_ZERORES
  delete mf_best[1];
#endif
  delete best_MLC[1];  
  delete eenc_best[0];
  delete eenc_best[1];
  delete frm_best_filtered  [0];
  delete frm_best_filtered  [1];
  delete frm_best_pred      [0];
  delete frm_best_pred      [1];
  delete frm_best_qpe_m     [0];
  delete frm_best_qpe_m     [1];
  delete frm_best_qpe_s     [0];
  delete frm_best_qpe_s     [1];

  delete mf;
  delete eenc;
  delete zmap[ZERO];
  delete zmap[ONE ];
  delete zmap_chroma[ZERO][ZERO]; 
  delete zmap_chroma[ZERO][ONE ]; 
  delete zmap_chroma[ONE][ZERO]; 
  delete zmap_chroma[ONE][ONE ]; 

  delete this->stored_ztree_luma_I;
  delete this->stored_ztree_chroma_I;
  delete this->stored_ztree_chroma_P;
  delete this->stored_ztree_mv_P;
  delete this->stored_ztree_mv_B;

  for (i = 0; i < 3; i++)
  {
    delete stored_eenc[i];
    delete stored_buff[i];
    delete stored_ztree_luma_P[i];
    if (param.trans8x8_enable) 
    {
      delete stored_ztree_luma_P_8x8[i];
    }
  }

  for (i = 0; i < 3; i++)
  {
    for (j = 0; j < 2; j++)
    {
      for (k = 0; k < 2; k++)
      {
        delete stored_eencF[i][j][k];
        delete stored_ztreeF_P[i][j][k];
      }
      delete stored_ztreeF_B[i][j];
    }
  }
  

  delete m_ienc;
  delete ref_mgr;
  delete lf;
  delete wlf;
  delete dbf;
  delete saif_half;

  delete trans[Transform::DCT];
  delete trans[Transform::AlternateIntra];
  delete trans[Transform::AlternateInter];

  delete pred_frame;

  for (i=0; i<4; i++)
    delete quant[i];
  
  delete m_cur2;
  delete m_cur2_qpe;
  delete m_cur2_pred;
  delete m_target;
  
  for (i = 0; i < 4; i++)
    delete m_me[i];
  
#ifdef I16_IN_P_FRAME
  for (i = 0; i < 7; i++)
  {
    delete m_tmpZeroTree16x16[i];
  }
#endif
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Main encoding function for one frame
*
* This main encoding function encode blocks in a frame and selects 
* the best encoding based on a R-D criterion.
*
* \param[in,out] cur Pointer to current frame. Initially contains original frame. Finally contains reconstructed frame.
* \param[out] buf Pointer to current byte buffer. Finally buffer contains coded frame.
*/
//-----------------------------------------------------------------------------
void Encoder::encode_frame(Frame   *cur, Frame   *cur_pred, Frame   *cur_qpe_m, Frame   *cur_qpe_s, Frame *cur_orig, Frame *cur_filtered, Buffer  *buf)
{
  int      tmp_bits[MAX_P_ENC];
  int      tmp_inter_type[MAX_P_ENC];
  int      tmp_AIF_type[MAX_P_ENC];
  double   tmp_cost[MAX_P_ENC];
  double   tmp_intra_cost=FLT_INF;
  int pel=fit_OneQuad_frame, pel_init=0, i, tmp_cnt;
  int count=0, increment=0, counter;
  int best_b=0, best_ref;  
  Frame *fr_filtered = frm_tmp_filtered;
  Frame *fr_pred     = frm_tmp_pred;
  Frame *fr_qpe_m    = frm_tmp_qpe_m;
  Frame *fr_qpe_s    = frm_tmp_qpe_s;
  Frame *fr=frm_tmp;
  double min_cost;
  MLC_Stat *mlc;

  change_mv_precision = 0;
  skip_me_loop_on = 0;

  int j;
  int P_AIF_FilterFlag=0;
  int B_AIF_FilterFlag=0;

  //===============================
  //  Check for deferred refresh
  //===============================
  if (cur->frame_no==0 || 
    (param.onlyI && param.idr_interval 
    && (cur->frame_no%(param.onlyI*param.idr_interval)==0)))
  {
    cur->reference_idc = REFERENCE_PIC;    // Is reference frame.
  }

  if (ref_mgr->deferred_refresh_flag)
  {
    // Reset frame_order_counter
    cur->frame_order_counter = cur->frame_no - ref_mgr->deferred_refresh_frame_number;
    if (cur->reference_idc==REFERENCE_PIC && cur->frame_no>ref_mgr->deferred_refresh_frame_number)
    {
      ref_mgr->perform_deferred_refresh ();
      ref_mgr->deferred_refresh_flag=0;
      cur->delta_qp = cur->qp - param.qp;        // change the refrence instead of last_qp 
    }
  }

  //-------------------------------------------
  //  Sets frame_type_idr flag and initialize
  //  reference frame buffer.
  //-------------------------------------------
  if (cur->frame_no==0 || 
    (param.onlyI && param.idr_interval 
    && (cur->frame_no%(param.onlyI*param.idr_interval)==0)))
  {
    if (param.deferred_decoder_refresh && cur->frame_no)
    {
      cur->intra_frame_type = ft_I_DDR;               // Is DDR
      ref_mgr->deferred_refresh_flag = 1;
      ref_mgr->deferred_refresh_frame_number = cur->frame_no;
      cur->delta_qp = cur->qp - param.qp;        // change the refrence instead of last_qp 
    }
    else
    {
      cur->intra_frame_type = ft_I_IDR;               // Is IDR
      ref_mgr->init ();
      cur->delta_qp = cur->qp - param.qp;        // change the refrence instead of last_qp
    }
  }
  else
  {
    cur->intra_frame_type = ft_I_noRefresh;
  }

  // reset some variables
  best_ref = ZERO;
  tmp_cnt = ZERO;

  fr = frm_tmp;
  fr->copy(cur);
  fr_filtered = frm_tmp_filtered;
  fr_pred     = frm_tmp_pred;
  fr_qpe_m    = frm_tmp_qpe_m;
  fr_qpe_s    = frm_tmp_qpe_s;
  fr_filtered->copy(cur_filtered);
  fr_pred->copy(cur_pred);
  fr_qpe_m->copy(cur_qpe_m);
  fr_qpe_s->copy(cur_qpe_s);
  fr->clear_stats();
  ref_mgr->lsize[LIST_0] = ZERO;
  ref_mgr->lsize[LIST_1] = ZERO;
  fr->list_size[LIST_0] = ref_mgr->lsize[LIST_0];
  fr->list_size[LIST_1] = ref_mgr->lsize[LIST_1];
  fr->frame_type = ft_I_frame;
  
  int framePos = frame_pos_intermediate;
  if (param.pyramid_coding && cur->intermed_idx == 0)
    framePos = frame_pos_edge;
  else if (cur->reference_idc == DISPOSABLE_PIC)
    framePos = frame_pos_disposable;
  
  int forcedIFrame = (cur->frame_no == 0) || (param.onlyI > 0 && cur->frame_no % param.onlyI == 0);
  
  // check if this is an onlyI picture
  if (forcedIFrame)    // check if it is the onlyI coded picture and use the appropriate qp_offset
  {
    fr->qp += param.qp_off_onlyI;    // set up the intra qp offset before the qp parameters are updated in update_rd_params.
    fr->delta_qp += param.qp_off_onlyI;    // update the delta as well.
  }
  else
  {
    if (framePos == frame_pos_disposable)
    {
      fr->qp += param.qp_off_intra_disposable_pic;    // set up the intra qp offset before the qp parameters are updated in update_rd_params.
      fr->delta_qp += param.qp_off_intra_disposable_pic;    // update the delta as well.
    }
    else
    {
      fr->qp += param.qp_off_intra;    // set up the intra qp offset before the qp parameters are updated in update_rd_params.
      fr->delta_qp += param.qp_off_intra;    // update the delta as well.
    }
  }

  fr->update_rd_params();      //Lets update parameters for RD decisions
  for (i=0; i<4; i++)
  {
    quant[i]->set_qp(fr->qp_matrix[i]);
    if (param.trans8x8_enable)
      quant[i]->set_qp8x8(fr->qp_matrix8x8[i]);
    if (param.intra16x16)
      quant[i]->set_qp16x16(fr->qp_matrix16x16[i]);
  }

  quant[0]->intra_frame=1;
  quant[1]->intra_frame=1;

  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      ref_mgr->ref_process->inter_type[i][j]=QUAD_PEL;
      ref_mgr->ref_process->filter_type[i][j]=S_FIF;
      ref_mgr->ref_process->ic_disable[i][j]=IC_ON;
      ref_mgr->ref_process->round_type[i][j]=ROUND_LASTSTEP;
    }
  }
  ref_mgr->process_list(cur, saif_half);
  ref_mgr->temp_rpps_id[tmp_cnt]=ref_mgr->find_rpps_id();

#ifdef TRACE_DEBUG
  OPEN_ETRACETMP;
#endif

  if (param.intra16x16_org)
  {
    param.intra16x16 = forcedIFrame;
  }
  tmp_bits[tmp_cnt] = encode_one_frame(fr,fr_pred,fr_qpe_m,fr_qpe_s,cur_orig,fr_filtered,buf_tmp, tmp_MLC, 0, -1, fit_OneQuad_frame, tmp_cnt);

#ifdef TRACE_DEBUG
  CLOSE_ETRACETMP;
#endif
  tmp_cost[tmp_cnt] = I_PIC_LAMBDA*(eenc->nbits_wlf_coeffs) + frm_tmp->compute_distortion(cur);
  tmp_inter_type[tmp_cnt] = fit_OneQuad_frame;

  tmp_intra_cost = INTRA_OFFSET_LAMBDA*(double)(eenc->nbits_wlf_coeffs) + frm_tmp->compute_distortion(cur);
  keep_best_buf_frm_mlc(0);          // keep a copy of the intra buffer and frame for fallback to intra mode.
  min_cost = 1e100;                 // do this once instead

  if (tmp_cost[tmp_cnt] < min_cost) 
  {
    min_cost = tmp_cost[tmp_cnt];

    if ((param.deblockingfilter&3)!=3)
    {
      dbf->filter_data_copy(lf, mf);
    }
  }

  tmp_cnt++;

#ifdef I16_IN_P_FRAME
  if (param.intra16x16_org)
  {
    param.intra16x16 = param.intra16x16p;
  }
#endif
  
  best_b = ZERO;

  if(param.aif)
  {
    for(j=0;j<4;j++)
    {
      quant[j]->save();
    }
  }

  num_loop_check=0;

  if (!forcedIFrame) 
  {
    int testB = param.enableB && framePos != frame_pos_edge;  // skip B for edge frames
    if (testB)
    {
      if(param.aif)
      {
        B_AIF_FilterFlag=1;
        for(j=0;j<4;j++)
        {
          quant[j]->B_save_1();
        }
      }
      switch (param.B_subpel)
      {
      case 1: //for 1/4 pel
        pel_init = 0; count = 1; increment = 1;
        break;
      case 2: //for 1/8 pel
        pel_init = 1; count = 1; increment = 1;
        break;
      case 3: //for 1/4 followed by 1/8 pel switching in B pictures
        pel_init = 0; count = 2; increment = 1;
        break;
      case 4: //for 1/8 pel followed by 1/4 pel switching in B pictures
        pel_init = 3; count = 2; increment = -1;
        break;
      default:
        printf("Unknown option for parameter B_subpel %d",param.B_subpel);
        exit (1);
      }

      for (counter=0, pel=pel_init;counter<count;counter++) 
      {
        change_mv_precision = counter != 0;
        skip_me_loop_on = counter != 0;
        pel += increment;

#ifdef TRACE_DEBUG
        OPEN_ETRACETMP;
#endif
        tmp_bits[tmp_cnt] = get_bit_cost_one_frame(cur, cur_pred, cur_qpe_m, cur_qpe_s, cur_orig, cur_filtered, tmp_cnt, ft_B_frame, BI_PRED_ONLY, best_ref, pel, -1);
#ifdef TRACE_DEBUG
        CLOSE_ETRACETMP;
#endif
        tmp_cost[tmp_cnt] = B_PIC_LAMBDA*(eenc->nbits_wlf_coeffs) + frm_tmp->compute_distortion(cur);

        tmp_inter_type[tmp_cnt] = pel*fit_OneQuad_frame;
        tmp_AIF_type[tmp_cnt] = 1;

        if (tmp_cost[tmp_cnt] < min_cost) 
        {
          min_cost = tmp_cost[tmp_cnt];
          keep_best_buf_frm_mlc(1);
          store_prob(cur, ft_B_frame, pel);
          if(param.aif)
          {
            for(j=0;j<4;j++)
            { 
              quant[j]->B_save_1();
            }
            B_AIF_FilterFlag=1;
          }
          if ((param.deblockingfilter&3)!=3)
          {
            dbf->filter_data_copy(lf, mf) ;
          }

        }
        tmp_cnt++;
      }
    } // if (param.enableB)

    best_b = ZERO;

    for (i=best_b+ONE; i<tmp_cnt; i++)
    {      
      if (tmp_cost[i] < tmp_cost[best_b])
        best_b = i;
    }

    mlc = best_MLC[best_b != 0];

    num_loop_check = param.enableB && framePos == frame_pos_edge;

    if (param.adapt_frame_type || !testB)
    {
      // Now lets do P frame coding
      if(param.aif)
      {
        P_AIF_FilterFlag=1;
        for(j=0;j<4;j++)
        {
          quant[j]->P_save_1();
        }
      }
      switch (param.P_subpel)
      {
      case 1: //for 1/4 pel
        pel_init = 0; count = 1; increment = 1;
        break;
      case 2: //for 1/8 pel
        pel_init = 1; count = 1; increment = 1;
        break;
      case 3: //for 1/4 followed by 1/8 pel switching in P pictures
        pel_init = 0; count = 2; increment = 1;
        break;
      case 4: // for 1/8 pel followed by 1/4 pel switching in P pictures
        pel_init = 3; count = 2; increment = -1;
        break;
      default:
        printf("Unknown option for parameter P_subpel %d",param.P_subpel);
        exit (1);
      }

      for (counter = 0, pel = pel_init; counter < count; counter++) 
      {
        change_mv_precision = counter != 0;
        skip_me_loop_on = counter != 0;
        pel += increment;
        
        if ((pel == 1 && framePos != frame_pos_disposable) || (pel == 2 && framePos == frame_pos_edge) || !param.enableB)
        {
#ifdef TRACE_DEBUG
          OPEN_ETRACETMP;
#endif
          //STANDARD_P START
          tmp_bits[tmp_cnt] = get_bit_cost_one_frame(cur, cur_pred, cur_qpe_m, cur_qpe_s, cur_orig, cur_filtered, tmp_cnt, ft_P_frame, LIST_0, 0, pel, -1);
#ifdef TRACE_DEBUG
          CLOSE_ETRACETMP;
#endif
          tmp_cost[tmp_cnt] = P_PIC_LAMBDA*(eenc->nbits_wlf_coeffs) + frm_tmp->compute_distortion(cur);

          tmp_inter_type[tmp_cnt] = pel*fit_OneQuad_frame;
          tmp_AIF_type[tmp_cnt] = 1;

          store_prob(cur, ft_P_frame, pel);
          if (tmp_cost[tmp_cnt] < min_cost) 
          {
            min_cost = tmp_cost[tmp_cnt];
            best_ref = ref_mgr->single_ref_frame;
            keep_best_buf_frm_mlc(1);
            if(param.aif)
            {
              for(j=0;j<4;j++)
              {
                quant[j]->P_save_1();
              }
              P_AIF_FilterFlag=1;
            }
            if ((param.deblockingfilter&3)!=3)
            {
              dbf->filter_data_copy(lf, mf) ;
            }
          }

          tmp_cnt++;
          //STANDARD_P END
        }
        if(param.aif)
        {
          change_mv_precision = 0;
          skip_me_loop_on = 1;
          //SAIF_P START
          if (framePos != frame_pos_disposable && (pel == 1 || framePos == frame_pos_edge || !param.enableB))
          {
#ifdef TRACE_DEBUG
            OPEN_ETRACETMP;
#endif
            tmp_bits[tmp_cnt] = get_bit_cost_one_frame(cur, cur_pred, cur_qpe_m, cur_qpe_s, cur_orig, cur_filtered,tmp_cnt, ft_P_frame, LIST_0, 0, pel, 0);
#ifdef TRACE_DEBUG
            CLOSE_ETRACETMP;
#endif
            tmp_cost[tmp_cnt] = P_PIC_LAMBDA*(eenc->nbits_wlf_coeffs) + frm_tmp->compute_distortion(cur); 

            tmp_inter_type[tmp_cnt] = pel*fit_OneQuad_frame;
            tmp_AIF_type[tmp_cnt] = 2;

            store_prob(cur, ft_P_frame, pel);
            
            if (tmp_cost[tmp_cnt] < min_cost) 
            {
              min_cost = tmp_cost[tmp_cnt];
              best_ref = ref_mgr->single_ref_frame;
              keep_best_buf_frm_mlc(1);

              for(j=0;j<4;j++)
              {
                quant[j]->P_save_2();
              }
              P_AIF_FilterFlag=2;
              if ((param.deblockingfilter&3)!=3)
              {
                dbf->filter_data_copy(lf, mf) ;
              }

            }
            tmp_cnt++;              
          } //SAIF_P END
        } // if(param.aif){
      } // for (counter=0;counter<count;counter++) { 
    } // if (param.adapt_frame_type || !(param.enableB && fr->intermed_idx != 0))
  } //if (coded_frames != 0 && param.onlyI == 0) {    

  for (i=best_b+ONE; i<tmp_cnt; i++)
  {
    if (tmp_cost[i] < tmp_cost[best_b])    
      best_b = i;
  }


  // Scene cut detection
  if (best_b != 0)
  {
    if(best_MLC[1]->iblks > best_MLC[0]->iblks*0.75)
    {
      best_b = 0;
    }
  }

  param.intra16x16 = param.intra16x16_org;

  if (best_b == 0) 
  {
    if (!forcedIFrame && param.intra16x16) 
    {
      fr = frm_tmp;
      fr->copy(cur);
      fr_filtered = frm_tmp_filtered;
      fr_pred     = frm_tmp_pred;
      fr_qpe_m    = frm_tmp_qpe_m;
      fr_qpe_s    = frm_tmp_qpe_s;
      fr_filtered->copy(cur_filtered);
      fr_pred->copy(cur_pred);
      fr_qpe_m->copy(cur_qpe_m);
      fr_qpe_s->copy(cur_qpe_s);
      fr->clear_stats();
      fr->frame_type = ft_I_frame;
      fr->qp += param.qp_off_intra_scene_cut;    // set up the intra qp offset before the qp parameters are updated in update_rd_params.
      fr->delta_qp += param.qp_off_intra_scene_cut;    // update the delta as well.

      fr->update_rd_params();      //Lets update parameters for RD decisions
      for (i=0; i<4; i++)
      {
        quant[i]->set_qp(fr->qp_matrix[i]);
        if (param.trans8x8_enable) quant[i]->set_qp8x8(fr->qp_matrix8x8[i]);
        quant[i]->set_qp16x16(fr->qp_matrix16x16[i]);
      }
      quant[0]->intra_frame=1;
      quant[1]->intra_frame=1;
#ifdef TRACE_DEBUG
      OPEN_ETRACETMP;
#endif
      best_ref = ZERO;
      tmp_cnt = ZERO;

      tmp_bits[tmp_cnt] = encode_one_frame(fr,fr_pred,fr_qpe_m,fr_qpe_s,cur_orig,fr_filtered,buf_tmp, tmp_MLC, 0, -1, fit_OneQuad_frame, tmp_cnt);

#ifdef TRACE_DEBUG
      CLOSE_ETRACETMP;
#endif


      tmp_cost[tmp_cnt] = I_PIC_LAMBDA*(eenc->nbits_wlf_coeffs) + frm_tmp->compute_distortion(cur);
      tmp_inter_type[tmp_cnt] = fit_OneQuad_frame;

      tmp_intra_cost = INTRA_OFFSET_LAMBDA*(double)(eenc->nbits_wlf_coeffs) + frm_tmp->compute_distortion(cur);


      keep_best_buf_frm_mlc(0);          // keep a copy of the intra buffer and frame for fallback to intra mode.
      
      if ((param.deblockingfilter&3)!=3)
      {
        dbf->filter_data_copy(lf, mf) ;
      }

    }

    if (param.open_loop)
      frm_best[0]->copy(cur);
    cur->copy(frm_best[0]);
    buf->copy(buf_best[0]);
    if (param.open_loop)
      frm_best_filtered[0]->copy(cur_filtered);
    cur_filtered->copy(frm_best_filtered[0]);

    if (param.open_loop)
      frm_best_pred[0]->copy(cur_pred);
    cur_pred->copy(frm_best_pred[0]);

    if (param.open_loop)
      frm_best_qpe_m[0]->copy(cur_qpe_m);
    cur_qpe_m->copy(frm_best_qpe_m[0]);

    if (param.open_loop)
      frm_best_qpe_s[0]->copy(cur_qpe_s);
    cur_qpe_s->copy(frm_best_qpe_s[0]);
    mlc_cur  = *best_MLC[0];
    eenc->copy_for_wlf(eenc_best[0]);
    eenc->set_buffer(buf);
    wlf->store_low_zeromv_rdoq=low_zeromv_rdoq_best[0];
  }
  else 
  {
    if (param.open_loop)
      frm_best[1]->copy(cur);
    cur->copy(frm_best[1]);
    buf->copy(buf_best[1]);
    if (param.open_loop)
      frm_best_filtered[1]->copy(cur_filtered);
    cur_filtered->copy(frm_best_filtered[1]);

    if (param.open_loop)
      frm_best_pred[1]->copy(cur_pred);
    cur_pred->copy(frm_best_pred[1]);

    if (param.open_loop)
      frm_best_qpe_m[1]->copy(cur_qpe_m);
    cur_qpe_m->copy(frm_best_qpe_m[1]);

    if (param.open_loop)
      frm_best_qpe_s[1]->copy(cur_qpe_s);
    cur_qpe_s->copy(frm_best_qpe_s[1]);
    mlc_cur  = *best_MLC[1];
    eenc->copy_for_wlf(eenc_best[1]);
    eenc->set_buffer(buf);
    wlf->store_low_zeromv_rdoq=low_zeromv_rdoq_best[1];
  }

  if (param.wienerfilter)
  {
    double lambdaY = calc_lambda_mode(cur->qp , cur->reference_idc, 0, wlf->store_low_zeromv_rdoq);
    double lambdaC = calc_lambda_mode(cur->qpC, cur->reference_idc, 2, wlf->store_low_zeromv_rdoq);
    
    wlf->estimate(cur, cur_orig, cur_pred, cur_qpe_m, cur_qpe_s, lambdaY, lambdaC, eenc, &param);
    wlf->encode_params(eenc);
    wlf->filter(cur, cur_filtered, cur_pred, cur_qpe_m, cur_qpe_s);

#ifdef WIENER_FILTER_INLOOP_SWITCH
    if (param.wienerfilter_inloop)
    {
      cur->copy_samples(cur_filtered);
    }
#endif
  }
  
  if ((param.deblockingfilter&3)==1)
  {
    dbf->init_alpha_values( cur->qp, cur->qpC );
    if (param.deblockingfilter&8)
    {
      dbf->set_filter_modes();
    }           
    dbf->fixed_deblocking_filter( dbf->str_mf, cur->luma, cur->luma, 0);
    dbf->fixed_deblocking_filter( dbf->str_mf, cur->crcb[0], cur->crcb[0], 1 );
    dbf->fixed_deblocking_filter( dbf->str_mf, cur->crcb[1], cur->crcb[1], 2 ); 
    memcpy(cur_filtered->luma,cur->luma,cur->height*cur->width*sizeof(unsigned char));
    memcpy(cur_filtered->crcb[0],cur->crcb[0],cur->height2*cur->width2*sizeof(unsigned char));
    memcpy(cur_filtered->crcb[1],cur->crcb[1],cur->height2*cur->width2*sizeof(unsigned char));
  }

  eenc->terminate();
  tmp_bits[best_b]=eenc->nbits;

  bits = tmp_bits[best_b];

  last_coded_frame = cur;

  ref_mgr->ref_pic_parameter_set_id_shared[ref_mgr->temp_rpps_id[best_b]] = ONE;

  if (last_coded_frame->reference_idc == FALSE)
  {
#ifndef MVCOMP_ZERORES
    ref_mgr->add_frame(last_coded_frame);
#else
    ref_mgr->add_frame(last_coded_frame, mf_best[!!best_b], &inter_pel_best[!!best_b][0][0], &log_inter_pel_best[!!best_b][0][0], &ref_list_best[!!best_b][0][0], &frame_no_best[!!best_b][0]);
#endif
  }
  if(param.aif)
  {
    if(cur->frame_type==ft_P_frame)
    {
      saif_half->AIF_FilterFlag=P_AIF_FilterFlag;
      for(j=0;j<4;j++)
      {
        quant[j]->P_save_all(P_AIF_FilterFlag);
      }
    }
    else if (cur->frame_type==ft_B_frame)
    {
      saif_half->AIF_FilterFlag=B_AIF_FilterFlag;
      for(j=0;j<4;j++)
      {
        quant[j]->B_save_all(B_AIF_FilterFlag);
      }
    }
    else
    {
      for(j=0;j<4;j++)
      {
        quant[j]->load();
      }
    }
  }
  mlc_tot.add(&mlc_cur);
  total_bits  += bits;
  htotal_bits += ((bits + 7)>>3) * 8;
  total_sad   += sad;

  coded_frames += 1;
  if(cur->frame_type!=ft_I_frame) coded_inter_frames+=1;         

}

//-----------------------------------------------------------------------------
/*!
* \brief Store probability for ME 
*
*
* \param[in] cur Pointer to current frame. 
* \param[in] frame_type %Frame type 
* \parem[in] pel subpel size
*
*/
//-----------------------------------------------------------------------------
void Encoder::store_prob(Frame *cur, int frame_type, int pel)
{
  int frameTypeIdx = (frame_type == ft_P_frame);
  int pelIdx = (pel == 1);
  int framePosIdx = frame_pos_intermediate; // Intermediate
  if (param.pyramid_coding && cur->intermed_idx == 0)
    framePosIdx = frame_pos_edge; // Edge
  else if (cur->reference_idc == DISPOSABLE_PIC)
    framePosIdx = frame_pos_disposable; // Disposable
  
  stored_eencF[framePosIdx][pelIdx][frameTypeIdx]->copy_coding_state_cs(eenc);
  stored_ztreeF_P[framePosIdx][pelIdx][frameTypeIdx]->copy_coding_state(param.ztree[4]);
  if (frameTypeIdx == 0)
  {
    stored_ztreeF_B[framePosIdx][pelIdx]->copy_coding_state(param.ztree[5]);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Calculate actual bit cost of one frame
*
* To select the best encoding based on a R-D criterion from multiple frame types/modes
*
* \param[in] cur Pointer to current frame. 
* \param[in] frame type ID
* \param[in] frame_type %Frame type 
* \param[in] me_mode  ME mode used in a frame
* \param[in] single_ref_frame flag indicating if number of references is 1 
* \param[in] pel subpel size
* \param[in] AIF_type type of AIF (2D Standard, FULL, HALF, PRED, SAIF)
*
* \return number of bits to encode a frame
*/
//-----------------------------------------------------------------------------
int Encoder::get_bit_cost_one_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame *cur_orig, Frame *cur_filtered, int tmp_cnt, int frame_type, int me_mode, int single_ref_frame, int pel, int AIF_type)
{

  int i, j, mode;
  Frame *fr;
  Frame *fr_filtered;
  Frame *fr_pred;
  Frame *fr_qpe_m;
  Frame *fr_qpe_s;

  if(param.aif)
  {
    for(j=0;j<4;j++)
    {
      quant[j]->load();
    }
  }

  fr = frm_tmp;
  fr->copy(cur);
  fr_filtered = frm_tmp_filtered;
  fr_pred     = frm_tmp_pred;
  fr_qpe_m    = frm_tmp_qpe_m;
  fr_qpe_s    = frm_tmp_qpe_s;
  fr_filtered->copy(cur_filtered);
  fr_pred->copy(cur_pred);
  fr_qpe_m->copy(cur_qpe_m);
  fr_qpe_s->copy(cur_qpe_s);
  fr->clear_stats();
  fr->frame_type = frame_type;

  if (frame_type == ft_P_frame) 
  {
    ref_mgr->single_ref_frame = single_ref_frame;
    ref_mgr->prepare_list(fr->frame_type, LIST_0, cur->frame_no);
    ref_mgr->lsize[LIST_1] = ZERO;

    fr->list_size[LIST_0] = ref_mgr->lsize[LIST_0];
    fr->list_size[LIST_1] = ref_mgr->lsize[LIST_1];
    fr->me_mode = me_mode;
    mode = LIST_0;
  }
  else if (frame_type == ft_B_frame) 
  {

    ref_mgr->single_ref_frame = single_ref_frame;
    ref_mgr->prepare_list(fr->frame_type, LIST_0, cur->frame_no);
    ref_mgr->prepare_list(fr->frame_type, LIST_1, cur->frame_no);
    ref_mgr->get_lRelation();
    //ref_mgr->print_lists();


    fr->list_size[LIST_0] = ref_mgr->lsize[LIST_0];
    fr->list_size[LIST_1] = ref_mgr->lsize[LIST_1];
    fr->me_mode = me_mode;
    mode = BI_PRED;

  }
  else 
  {
    mode = LIST_0;
    printf ("Unknown Path?\n");
  }

  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      if(pel==fit_HALFPEL_frame)
      { 
        ref_mgr->ref_process->inter_type[i][j]=HALF_PEL;
      }
      else if(pel==fit_OneQuad_frame)
      {
        ref_mgr->ref_process->inter_type[i][j]=QUAD_PEL;
      }
      else if(pel==fit_OneEighth_frame)
      {
        ref_mgr->ref_process->inter_type[i][j]=EIGHTH_PEL;
      }
      if(AIF_type==-1)
      {
        ref_mgr->ref_process->filter_type[i][j]=S_FIF;
      }
      else
      {
        if(AIF_type==0)
        {
          ref_mgr->ref_process->filter_type[i][j]=AIF_SAIF;
        }
      }
      ref_mgr->ref_process->ic_disable[i][j]=IC_ON;
      ref_mgr->ref_process->round_type[i][j]=ROUND_LASTSTEP;
    }
  }
  ref_mgr->process_list(cur, saif_half); 
  ref_mgr->temp_rpps_id[tmp_cnt]=ref_mgr->find_rpps_id();

#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  ref_mgr->standard_interpolation();
#endif

  ref_mgr->prev_rpps_id = ref_mgr->target_rpps_id;
  ref_mgr->target_rpps_id = ref_mgr->temp_rpps_id[tmp_cnt];

  fr->update_rd_params(); //Lets update parameters for RD decisions

  for (i=0; i<4; i++)   
  {
    quant[i]->set_qp(fr->qp_matrix[i]);
    if (param.trans8x8_enable) quant[i]->set_qp8x8(fr->qp_matrix8x8[i]);
  }

  if(num_loop_check==0 && frame_type == ft_B_frame)
  {
    intra_cost_update=ICOST_STORED;
  }else if(num_loop_check==0 && frame_type == ft_P_frame && param.enableB)
  {
    intra_cost_update=ICOST_COPY;
  }else
  {
    intra_cost_update=ICOST_UPDATE;
  }
  num_loop_check++;

  quant[0]->intra_frame=0;
  quant[1]->intra_frame=0;
#ifdef TRACE_DEBUG
  OPEN_ETRACETMP;
#endif
  return encode_one_frame(fr, fr_pred, fr_qpe_m, fr_qpe_s, cur_orig, fr_filtered, buf_tmp, tmp_MLC, mode, AIF_type, pel*fit_OneQuad_frame, tmp_cnt);
}

//-----------------------------------------------------------------------------
/*!
* \brief Set reference frame
*
* Copies input frame into last_coded_frame. This may be used in the case
* where a sequence is coded in open loop mode (e.g., use original as input).
*
* \note This seems to be nonfunctional as the frame is not inserted into the
* reference frame buffer.
*
* \param[in] cur Pointer to a frame.
*/
//-----------------------------------------------------------------------------
void Encoder::set_ref_frame(Frame *cur) 
{
  if (last_coded_frame != NULL)
    last_coded_frame->copy(cur);
}

//-----------------------------------------------------------------------------
/*!
* \brief Count zero mv in a frame
*
* This value is used in Quantizer
*
* \return flag indicating if the frequency of zero MV in a frame is smaller than 50%
*/
//-----------------------------------------------------------------------------
int Encoder::mv_count() 
{
  int i,j;
  int num_zeromv = 0;

  if (mf->frame_type == ft_I_frame)
    return 0;

  for (i=0; i<mf->height; i++) 
  {
    for (j=0; j<mf->width; j++) 
    {
      PredBlk *curblk = mf->get_pblk(i, j); 
      if (curblk->intra == NULL) 
      {
        int list = curblk->list;
        if (list == BI_PRED)
        {
          num_zeromv += ((curblk->mv[LIST_0][0] == 0) && (curblk->mv[LIST_0][1] == 0)) || ((curblk->mv[LIST_1][0] == 0) && (curblk->mv[LIST_1][1] == 0));
        }
        else
        {
          num_zeromv += (curblk->mv[list][0] == 0) && (curblk->mv[list][1] == 0);
        }
      }
    }
  }

  return ((float)num_zeromv / (mf->height*mf->width) < 0.5);
}

//-----------------------------------------------------------------------------
/*!
* \brief Select lambda
*
* Select lambeda value for block mode decision and RDO-Q 
* based on QP, frame_type, reference
*
* \param[in] qp quantization parameter
* \param[in] reference_idc reference frame indicator
* \param[in] indicator to switch the type of lambda (luma or chroma and block mode decision or RDO-Q)
* \param[in] low_zeromv_rdoq flag indicating if the frequency of zero MV in a frame is smaller than 50%
*
* \return lambda
*/
//-----------------------------------------------------------------------------
double Encoder::calc_lambda_mode(int qp, int reference_idc, int rdoq, int low_zeromv_rdoq) 
{
  if(param.enable_simple_lambdas){
    double qp_jm = 6.0/16 * qp + 4.0;
    double k;
    int chroma = rdoq >> 1;
    rdoq = rdoq & 1;
  
    if (chroma)
      k = (rdoq) ? param.sl_lambda_P_chroma_rdoq : param.sl_lambda_P_chroma;
    else
      k = (rdoq) ? param.sl_lambda_P_luma_rdoq : param.sl_lambda_P_luma;
  
    return k * pow(2.0, (qp_jm - 12.0) / 3);
  }else{  
    double  lambda_mode;
    double  qp_jm;

    qp_jm = 6 * (double) qp / 16 + 4;


    if(rdoq==3 && reference_idc==REFERENCE_PIC)
    {  // luma
      if(low_zeromv_rdoq==0)
        lambda_mode = 0.68 * pow(2.0, (double) (qp_jm - 12) / 3);
      else
        lambda_mode = 0.57 * pow(2.0, (double) (qp_jm - 12) / 3);
    }
    else if(rdoq==3)
    {
      if(low_zeromv_rdoq==0)
        lambda_mode = 0.85 * pow(2.0, (double) (qp_jm - 12) / 3);
      else
        lambda_mode = 0.68 * pow(2.0, (double) (qp_jm - 12) / 3);
    }
    else if(rdoq==1 && reference_idc==REFERENCE_PIC)
    {  // luma
      if(low_zeromv_rdoq==0)
        lambda_mode = 0.68 * pow(2.0, (double) (qp_jm - 12) / 3);
      else
        lambda_mode = 0.57 * pow(2.0, (double) (qp_jm - 12) / 3);
    }
    else if(rdoq==1)
    {
      if(low_zeromv_rdoq==0)
        lambda_mode = 0.85 * pow(2.0, (double) (qp_jm - 12) / 3);
      else
        lambda_mode = 0.68 * pow(2.0, (double) (qp_jm - 12) / 3);
    }
    else if(rdoq==2 && reference_idc==REFERENCE_PIC)
    {  // luma
      if(low_zeromv_rdoq==0)
        lambda_mode = 0.68 * pow(2.0, (double) (qp_jm - 12) / 3);
      else
        lambda_mode = 0.57 * pow(2.0, (double) (qp_jm - 12) / 3);
    }
    else if(rdoq==2)
    {
      if(low_zeromv_rdoq==0)
        lambda_mode = 0.85 * pow(2.0, (double) (qp_jm - 12) / 3);
      else
        lambda_mode = 0.68 * pow(2.0, (double) (qp_jm - 12) / 3);
    }
    else
      lambda_mode = 0.85 * pow(2.0, (double) (qp_jm - 12) / 3);

    return (lambda_mode);
  }
}


#ifdef MVCOMP_ZERORES
//-----------------------------------------------------------------------------
/*!
* \brief Initialize a 'skip' mb -- ref from neighbours, mv from predictor, residue 0.
*
* \param[in] me Pointer to motion estimator
* \param[in,out] mlc Pointer to statistics for Motion/Luma/Chroma
* \param[in] i Vertical position of current 8x8 block
* \param[in] j Horizontal position of current 8x8 block
* \param[in] type flag to switch outputting bits or counting bits
* \param[in] skip_type indicates which predictor to use.
*/
//-----------------------------------------------------------------------------        
int Encoder::skip_initialize(MotionEst *me, MLC_Stat * /*mlc*/, int i, int j, int skip_type)
{
  PredBlk *curblk;
  PredBlk *top, *left, *topleft;
  //PredBlk diff, tmp;
  int ctx;
  //int unresolved_mv_prediction;
  int list;
  PredBlkSet predblkset;
  PredBlk **ptmp_preds;
  PredBlk *pred;

  ctx = ZERO; // dummy initialization to avoid compiler warning

  curblk  = mf->get_pblk(i,  j  ); 
  left    = mf->get_pblk(i,  j-1);
  top     = mf->get_pblk(i-1,j  );
  topleft = mf->get_pblk(i-1,j-1);

  curblk->clear();
  
  // Imagine we're not transmitting anything for the skipped mb,
  // look at neighbours to determine a good list & ref to use.
  switch (skip_type)
  {
  case MV_PRED_LEFT:
    // predicted from 'a', coded L0, L1 or Bi as a is.
    if (left == NULL || left->intra != NULL)
    {
      return FALSE;
    }
    else
    {
      curblk->list = left->list;
      curblk->ref_index[LIST_0] = left->ref_index[LIST_0];
      curblk->ref_index[LIST_1] = left->ref_index[LIST_1];
    }
    break;
  case MV_PRED_TOP:
    // predicted from b, coded L0, L1 or Bi as b is.
    if (top == NULL || top->intra != NULL)
    {
      return FALSE;
    }
    else
    {
      curblk->list = top->list;
      curblk->ref_index[LIST_0] = top->ref_index[LIST_0];
      curblk->ref_index[LIST_1] = top->ref_index[LIST_1];
    }
    break;
  case MV_PRED_TEMPORAL:
    // We set our refs from the neighbours (minimum valid ref found for each
    // list), and then our list (LIST_0, LIST_1, BI_PRED).
    if (!curblk->get_best_temporal_refs(me->ref_mgr, i, j, left, top, topleft))
    {
      return FALSE;
    }
    break;
  case MV_PRED_ZERO:
    curblk->list = LIST_0;
    curblk->ref_index[LIST_0] = 0;
    break;
  case MV_PRED_NONE:
    return FALSE;
  }

  list = curblk->list;

  ptmp_preds = predblkset.get_predictor_set_complete(ref_mgr, i, j, left, top, topleft, curblk);
  pred = ptmp_preds[skip_type];

  if (pred == NULL)
    return FALSE;

  if (curblk->list == BI_PRED)
  {
    curblk->mv[LIST_0][0] = pred->mv[LIST_0][0];
    curblk->mv[LIST_0][1] = pred->mv[LIST_0][1];
    curblk->mv[LIST_1][0] = pred->mv[LIST_1][0];
    curblk->mv[LIST_1][1] = pred->mv[LIST_1][1];
    curblk->BP_scale[0] = pred->list == BI_PRED ? pred->BP_scale[0] : 0;
    curblk->BP_scale[1] = pred->list == BI_PRED ? pred->BP_scale[1] : 0;
    curblk->BP_DC = pred->list == BI_PRED ? pred->BP_DC : 0;
  }
  else
  {
    curblk->mv[curblk->list][0] = pred->mv[curblk->list][0];
    curblk->mv[curblk->list][1] = pred->mv[curblk->list][1];

    // contrast and brightness as well.
    curblk->scale[curblk->list] = pred->list == BI_PRED ? 0 : pred->scale[curblk->list];
    curblk->DC[curblk->list] = pred->list == BI_PRED ? 0 : pred->DC[curblk->list];
  }

  curblk->mv_zero_res = 1; // indicates residue forced to zero, later.

  return TRUE;
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Encode motion data
*
* Encodes the motion data for a 8x8 block. This includes the prediction
* mode (Intra, LIST_0, LIST_1, BI_PRED).
*
* \note Pointer to motion estimator is required as it is used to decide whether
* to use horizontal or vertical prediction of motion data. This operation
* requires access to cost functions (only if Param::mvs_realcost is false).
* It doesn't matter which estimator is used.
*
* \param[in] me Pointer to motion estimator
* \param[in,out] mlc Pointer to statistics for Motion/Luma/Chroma
* \param[in] i Vertical position of current 8x8 block
* \param[in] j Horizontal position of current 8x8 block
* \param[in] isFinal flag to enable stats
*/
//-----------------------------------------------------------------------------        
void Encoder::codemvsbi_blk(MotionEst *me, MLC_Stat *mlc, int i, int j, bool isFinal, EntropyEncoder *tmp_eenc) 
{
  PredBlk *curblk;
  PredBlk *top, *left, *topleft;
  PredBlk diff, tmp;
  int ctx = 0; // dummy initialization to avoid compiler warning
  int c0,c1;
  int unresolved_mv_prediction;
  int mv_pred_dir;
  int list;
#ifndef MVCOMP_ZERORES
  PredBlk pleft, ptop, ptleft;
#else
  PredBlk ptleft;
#endif
  PredBlk *pred_left, *pred_top, *pred_tleft;
#ifdef MVCOMP_ZERORES
  PredBlkSet predblkset;
  PredBlk **ptmp_preds, **ptmp_diff_preds;
  int n_diff_preds;
#endif
  
  int mvshare_size_ctx;
  int mvshare_type;
  
  curblk  = mf->get_pblk(i,  j  ); 
  left    = mf->get_pblk(i,  j-1);
  top     = mf->get_pblk(i-1,j  );
  topleft = mf->get_pblk(i-1,j-1);
  
#ifdef TRACE_DEBUG
  if (isFinal)
    ENABLE_ETRACE;
#endif
  
  tmp_eenc->encode_mode(curblk, left, top, mf->frame_type == ft_B_frame);
  
  if (curblk->intra == NULL) 
  {
    diff = *curblk;
    list = curblk->list;
    if (isFinal)
      mlc->list_stat[list]++; // Collect list statistics;
    
#ifndef MVCOMP_ZERORES
#ifdef MV_CODING_ENHANCE
    
#ifdef PMV0_AT_OUT_OF_BOUNDARY
    pred_left = pleft.get_prediction(i,j,left, curblk, ref_mgr);
    pred_top = ptop.get_prediction(i,j,top,  curblk, ref_mgr);
#else
    pred_left = pleft.get_prediction(left, curblk, ref_mgr);
    pred_top = ptop.get_prediction(top,  curblk, ref_mgr);
#endif
    
#else
    pred_left = pleft.get_prediction(left, curblk);
    pred_top = ptop.get_prediction(top,  curblk);
#endif
#else
    ptmp_preds = predblkset.get_predictor_set_complete(ref_mgr, i, j, left, top, topleft, curblk);
#endif
#ifdef MV_CODING_ENHANCE
    
#ifdef PMV0_AT_OUT_OF_BOUNDARY
    pred_tleft = ptleft.get_prediction(i,j,topleft, curblk, ref_mgr);
#else
    pred_tleft = ptleft.get_prediction(topleft, curblk, ref_mgr);
#endif
    
    
#else
    pred_tleft = ptleft.get_prediction(topleft, curblk);
#endif
    
    if (pred_tleft == NULL)
      pred_tleft = &zero_mv;
    
#ifdef MVCOMP_ZERORES
    ptmp_diff_preds = predblkset.collapse_predictor_set(list, n_diff_preds);
    int extra_pred;
    predblkset.signal_treatment(n_diff_preds, extra_pred, list,
                                tmp_eenc->proba_mvp(2), tmp_eenc->count_mvp(2, 0), tmp_eenc->count_mvp(2, 1), 0);
    pred_left = n_diff_preds > 0 ? ptmp_diff_preds[0] : NULL;
    pred_top = n_diff_preds > 1 ? ptmp_diff_preds[1] : NULL;
    unresolved_mv_prediction = n_diff_preds > 1;
#else
    unresolved_mv_prediction = (pred_left != NULL) && (pred_top != NULL) && (pred_left->equal_pred(pred_top, list) == FALSE);
#endif
    
    if (unresolved_mv_prediction == TRUE) 
    {
#ifndef MVCOMP_ZERORES
      ctx = pred_tleft->get_pred_context(pred_left, pred_top, list);
#else
      ctx = pred_tleft->get_pred_context(pred_left, pred_top, list, extra_pred);
#endif
      
      if (param.mvs_realcost == FALSE) 
      {
        diff.list = list;
        c0 = me->bit_cost_inter1(&diff, pred_left) + me->c_mvp[ctx][0];
        c1 = me->bit_cost_inter1(&diff, pred_top)  + me->c_mvp[ctx][1];
      }
      else 
      {
        if (curblk->list == BI_PRED) 
        {
          tmp = diff;
#ifdef MV_CODING_ENHANCE
          tmp.sub(pred_left, ref_mgr);
#else
          tmp.sub(pred_left);
#endif
          
          c0 =  tmp_eenc->cost_mv(tmp.mv[LIST_0][0], 0);
          c0 += tmp_eenc->cost_mv(tmp.mv[LIST_0][1], 1);
          c0 += tmp_eenc->cost_mv(tmp.mv[LIST_1][0], 0);
          c0 += tmp_eenc->cost_mv(tmp.mv[LIST_1][1], 1);
          
          c0 += tmp_eenc->cost_scale(tmp.BP_scale[0], 1);
          c0 += tmp_eenc->cost_scale(tmp.BP_scale[1], 2);
          c0 += tmp_eenc->cost_offset(tmp.BP_DC, 1);
          c0 += tmp_eenc->cost_mvp(0, ctx);
          
          tmp = diff;
#ifdef MV_CODING_ENHANCE
          tmp.sub(pred_top, ref_mgr);
#else
          tmp.sub(pred_top);
#endif
          
          c1 =  tmp_eenc->cost_mv(tmp.mv[LIST_0][0], 0);
          c1 += tmp_eenc->cost_mv(tmp.mv[LIST_0][1], 1);
          c1 += tmp_eenc->cost_mv(tmp.mv[LIST_1][0], 0);
          c1 += tmp_eenc->cost_mv(tmp.mv[LIST_1][1], 1);
          
          c1 += tmp_eenc->cost_scale(tmp.BP_scale[0], 1);
          c1 += tmp_eenc->cost_scale(tmp.BP_scale[1], 2);
          c1 += tmp_eenc->cost_offset(tmp.BP_DC, 1);
          c1 += tmp_eenc->cost_mvp(1, ctx);
        } 
        else 
        {
          tmp = diff;
#ifdef MV_CODING_ENHANCE
          tmp.sub(pred_left, ref_mgr);
#else
          tmp.sub(pred_left);
#endif
          c0 = tmp_eenc->cost_mv(tmp.mv[list][0], 0);
          c0 += tmp_eenc->cost_mv(tmp.mv[list][1], 1);
          c0 += tmp_eenc->cost_scale(tmp.scale[list], 0);
          c0 += tmp_eenc->cost_offset(tmp.DC[list], 0);
          c0 += tmp_eenc->cost_mvp(0, ctx);
          
          tmp = diff;
#ifdef MV_CODING_ENHANCE
          tmp.sub(pred_top, ref_mgr);
#else
          tmp.sub(pred_top);
#endif
          c1 = tmp_eenc->cost_mv(tmp.mv[list][0], 0);
          c1 += tmp_eenc->cost_mv(tmp.mv[list][1], 1);
          c1 += tmp_eenc->cost_scale(tmp.scale[list], 0);
          c1 += tmp_eenc->cost_offset(tmp.DC[list], 0);
          c1 += tmp_eenc->cost_mvp(1, ctx);
        }
      } 
      mv_pred_dir = (c0 > c1);
    }
    else 
    {
      mv_pred_dir = pred_top != NULL;
    }
    
#ifdef MV_CODING_ENHANCE
    if (pred_top == NULL && pred_left == NULL)
      diff.sub(&zero_mv, ref_mgr);
    else if (mv_pred_dir == 0)
      diff.sub(pred_left, ref_mgr);
    else
      diff.sub(pred_top, ref_mgr);
#else
    if (pred_top == NULL && pred_left == NULL)
      diff.sub(&zero_mv);
    else if (mv_pred_dir == 0)
      diff.sub(pred_left);
    else
      diff.sub(pred_top);
#endif
    
    ZeroTree *tmp_ztree = (curblk->list == BI_PRED) ? stored_ztree_mv_B : stored_ztree_mv_P;
    tmp_eenc->encode_mv(&diff, isFinal, (isFinal) ? NULL : tmp_ztree);
    
    if (curblk->list == BI_PRED) 
    {
      tmp_eenc->encode_refindex(curblk->ref_index[LIST_0], ref_mgr->lsize[LIST_0], top, left, LIST_0);
      tmp_eenc->encode_refindex(curblk->ref_index[LIST_1], ref_mgr->lsize[LIST_1], top, left, LIST_1);
      
      if(isFinal)
      {
        mlc->ref_frame_stat[LIST_0][curblk->ref_index[LIST_0]] ++;
        mlc->ref_frame_stat[LIST_1][curblk->ref_index[LIST_1]] ++;
      }
    }
    else 
    {
      tmp_eenc->encode_refindex(curblk->ref_index[list], ref_mgr->lsize[list], top, left, list);
      
      if (isFinal)
        mlc->ref_frame_stat[list][curblk->ref_index[list]] ++;
    }
    
    if (unresolved_mv_prediction == TRUE)
    {
      tmp_eenc->encode_mvp(mv_pred_dir, ctx); 
    }
    
    if (param.motion_sharing_enable)
    {
      mvshare_type = mf->mvshare_type(i, j);
      
      if (mvshare_type)
      {
        mvshare_size_ctx = mf->get_mvshare_size_ctx(mvshare_type, top, left);
        
        if (mvshare_type == 1)
        {
          int prev_val = (j == 0) ? 0 : left->mvshare_size[0];
          tmp_eenc->encode_mvshare_size(curblk->mvshare_size[0], prev_val, mvshare_size_ctx, 1);
        }
        else if (mvshare_type == 2)
        {
          int prev_val = (i == 0) ? 0 : top->mvshare_size[1];
          tmp_eenc->encode_mvshare_size(curblk->mvshare_size[1], prev_val, mvshare_size_ctx, 2);
        }
        else
        {
          int idx = (curblk->mvshare_size[0]) ? 0 : 1;
          int prev_val = (curblk->mvshare_size[0]) ? left->mvshare_size[0] : top->mvshare_size[1];
          tmp_eenc->encode_mvshare_size(curblk->mvshare_size[idx], prev_val, mvshare_size_ctx, 5+idx);
        }
      }
    }  // param.motion_sharing_enable
  }
  else if (isFinal)
  {
    mlc->list_stat[4]++; 
  }
  
#ifdef TRACE_DEBUG
  if(isFinal)
    DISABLE_ETRACE;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Refresh qp and quantizer
*
* This function will be called when qp is changed during encoding
*
* \param[in,out] cur Pointer to current frame.
* \param[in] ienc Pointer to entropy encoder
*/
//----------------------------------------------------------------------------- 
void Encoder::set_dqp(Frame *cur, IEncoder6  *ienc)
{
  int ii;

  cur->update_rd_params(); 

  for (ii=0; ii<4; ii++)  
  {
    quant[ii]->set_qp(cur->qp_matrix[ii]);
    if (param.trans8x8_enable)
      quant[ii]->set_qp8x8(cur->qp_matrix8x8[ii]);
    if (param.intra16x16)
      quant[ii]->set_qp16x16(cur->qp_matrix16x16[ii]);
  }

  quant[0]->intra_frame = cur->frame_type == ft_I_frame;
  quant[1]->intra_frame = cur->frame_type == ft_I_frame;
  
  ienc->iqp = cur->qp;
  ienc->iqpC = cur->qpC; 
}


//-----------------------------------------------------------------------------
/*!
* \brief Encode one luma 4x4 block
*
* \param[in] i Vertical position of 8x8 block
* \param[in] j Horizontal position of 8x8 block
* \param[in] ii Vertical position of 4x4 sub-block in 8x8 block 
* \param[in] jj Horizontal position of 4x4 sub-block in 8x8 block
* \param[in] emodel Entropy model index (EntropyModel::I_luma, etc.)
* \param[in,out] cur2 Pointer to current frame (WFrame). Initially contains original frame. Finally contains reconstructed frame.
* \param[out] mlc Pointer to statistics for Motion/Luma/Chroma
* \param[in] type flag to switch outputting bits or counting bits
* \param[in] lambda_rdoq lambda for RDO-Q 
* \param[in] cur Pointer to current frame (Frame).
*/
//----------------------------------------------------------------------------- 
void Encoder::encode_luma_inter_block(int i, int j, int ii, int jj, int emodel, MLC_Stat  *mlc,
                                      int type, double lambda_rdoq, double lambda_mode, int &transformIdx)
{
  // type 0; 1st coding loop
  // type 1; mode selection
  // type 2; 2nd coding loop
  int is_zero_blk;
  Block blk(4, 4);
  Block blk_pred(4, 4);
  Block org_blk4x4(4, 4);
  Block min_blk4x4(4, 4);
  Block residue_blk4x4(4, 4);
  int rate = 0;
  double distortion;
  double rdcost;
  double min_rdcost = FLT_INF;
  
  m_target->get_blk_luma(&blk, i+ii, j+jj);
  org_blk4x4.copy(blk);
  
  if (ii == 0 && jj == 0)
  {
    mf->predict_blk(pred_frame, ref_mgr, i>>1, j>>1);       
    
    if (param.motion_sharing_enable && (mf->get_pblk_unsafe((i>>1), (j>>1))->mvshare_size[0] || mf->get_pblk_unsafe((i>>1), (j>>1))->mvshare_size[1]))
      mf->predict_blk_mvshare(pred_frame, ref_mgr, i>>1, j>>1);
  }
  
  pred_frame->get_blk_luma(&blk_pred, i+ii, j+jj);
  
#ifndef MVCOMP_ZERORES
  blk.sub(blk_pred); 
#else
  if (mf->get_pblk_unsafe(i>>1, j>>1)->mv_zero_res)
    blk.zero();
  else
    blk.sub(blk_pred); 
#endif
  
  int start_type = -1, end_type = 0;
  int TransType;
  
  if (param.transform_selection == 1)
  {
    if (type == 1)
    {
      start_type = 0;
      end_type = 3;
      residue_blk4x4.copy(blk);

      for (TransType = start_type+1; TransType < end_type; TransType++)
      {
        stored_eenc[TransType]->copy_coding_state(stored_eenc[start_type]);
        stored_ztree_luma_P[TransType]->copy_coding_state(stored_ztree_luma_P[start_type]);
      }
    }
    else
    {
      start_type = transformIdx;
      end_type = transformIdx + 1;
    }
  }
  
  for (TransType = start_type; TransType < end_type; TransType ++)
  {
    EntropyEncoder *tmp_eenc = eenc;
    ZeroTree *tmp_ztree = param.ztree[emodel];
    
    if (type == 1)
    {
      tmp_eenc = stored_eenc[imax(TransType, 0)]; // cover TransType == -1 case
      tmp_ztree = stored_ztree_luma_P[imax(TransType, 0)];
    }
    
    if (param.transform_selection == 1)
    {
      if (type == 1)
      {
        if (TransType < 2)
          blk.copy(residue_blk4x4);
        
        rate = tmp_eenc->get_nbits();
      }
      
      if (TransType == 2)
        blk.zero();
    }
    
    int transIdx = (TransType < 1) ? Transform::DCT : Transform::AlternateInter;
    trans[transIdx]->forward(&blk);
    
    quant[emodel]->loadmatrix(emodel, -1, 0, 0); // cq_start
    quant[emodel]->loadmatrix(emodel, -1, 0, 1); // cq_end

    if (param.q_rdoffset)
    {
      quant[emodel]->quant_rdoq(&blk, tmp_eenc, emodel, lambda_rdoq, zmap[ZERO]->getcontext(i+ii,j+jj), tmp_ztree, 0, TransType);
    }
    else
    {
      quant[emodel]->quant(&blk, 0, TransType);
    }
    
    is_zero_blk = blk.is_zero();
    zmap[ZERO]->set(i+ii,j+jj, blk.nnz());
#ifdef TRACE_DEBUG
    if(type==0)
    {
      ENABLE_ETRACE;
    }
    SPRINT_ETRACE("luma inter ");
    RPRINT_ETRACE;
#endif
    tmp_eenc->encode_zblk(is_zero_blk, zmap[ZERO]->getcontext(i+ii,j+jj), emodel);
    
    if (is_zero_blk == FALSE) 
    {
      tmp_eenc->encode_coeffs(tmp_ztree, &blk, emodel);
    }
    
    if (type != 1)
    {
      if (is_zero_blk)
        mlc->zblks++;
      else
        mlc->nzblks++;
    }
    
#ifdef TRACE_DEBUG
    if(type==0)
    {
      DISABLE_ETRACE;
    }
#endif

    
    quant[emodel]->iquant(&blk, 0);    
    trans[transIdx]->inverse(&blk);
    
    if (type == 1 && param.transform_selection == 1)
    {
      rate = tmp_eenc->get_nbits() - rate;
      blk.add(blk_pred);
      
      distortion = blk.ssd(org_blk4x4);

      if(param.sl_trans_sel)
        rdcost = distortion + lambda_rdoq * rate;
      else
        rdcost = distortion + lambda_mode * rate / 1.6;
      
      if (rdcost < min_rdcost)
      {
        transformIdx = TransType;
        min_rdcost = rdcost;
        min_blk4x4.copy(blk);
      }
    }
  }
  
  if (type == 1 && param.transform_selection == 1)
  {
    if (transformIdx >= 1)
    {
      stored_eenc[0]->copy_coding_state(stored_eenc[transformIdx]);
      stored_ztree_luma_P[0]->copy_coding_state(stored_ztree_luma_P[transformIdx]);
    }

    m_cur2->put_blk_luma(&min_blk4x4, i+ii, j+jj);
  }
  else
  {
    m_cur2_pred->put_blk_luma(&blk_pred, i+ii, j+jj);
    m_cur2_qpe->put_blk_luma(&blk, i+ii, j+jj);
    blk.add(blk_pred);
    m_cur2->put_blk_luma(&blk, i+ii, j+jj);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode one luma 8x8 block with 8x8 transform
*
* \param[in] i Vertical position of 8x8 block
* \param[in] j Horizontal position of 8x8 block
* \param[in] emodel Entropy model index (EntropyModel::I_luma, etc.)
* \param[in,out] cur2 Pointer to current frame (WFrame). Initially contains original frame. Finally contains reconstructed frame.
* \param[in,out] pred_frame Pointer to motion compensated frame
* \param[out] mlc Pointer to statistics for Motion/Luma/Chroma
* \param[in] type flag to switch outputting bits or counting bits
* \param[in] cur Pointer to current frame (Frame).
*/
//----------------------------------------------------------------------------- 
void Encoder::encode_luma_inter_block_8x8(int i, int j, int emodel, WFrame *pred_frame, MLC_Stat  *mlc, int type , double lambda_rdoq, double lambda_mode, int &transformIdx)
{
  // type 0; 1st coding loop
  // type 1; mode selection
  // type 2; 2nd coding loop
  int is_zero_blk;
  Block blk8x8(8, 8);
  Block blk_pred8x8(8, 8);
  int TransType = -1;
  Block org_blk8x8(8, 8);
  Block min_blk8x8(8, 8);
  Block residue_blk8x8(8, 8);
  int rate = 0;
  double distortion;
  double rdcost;
  double min_rdcost = FLT_INF;
  PredBlk *cur_blk = mf->get_pblk_unsafe(i>>1, j>>1);
  
  m_target->get_blk_luma(&org_blk8x8, i>>1, j>>1);
  
  mf->predict_blk(pred_frame, ref_mgr, i>>1, j>>1);
  if (param.motion_sharing_enable && (cur_blk->mvshare_size[0] != 0 || cur_blk->mvshare_size[1] != 0))
    mf->predict_blk_mvshare(pred_frame, ref_mgr, i>>1, j>>1);

  m_target->get_blk_luma(&blk8x8, i>>1, j>>1);
  pred_frame->get_blk_luma(&blk_pred8x8, i>>1, j>>1);
  
#ifndef MVCOMP_ZERORES
  blk8x8.sub(blk_pred8x8);
#else
  if (cur_blk->mv_zero_res)
    blk8x8.zero();
  else
    blk8x8.sub(blk_pred8x8);
#endif
  
  int start_type = -1, end_type = 0;
  
  if (param.transform_selection == 1)
  {
    if (type == 1 && transformIdx == -1)
    {
      start_type = 0;
      end_type = 3;
    }
    else
    {
      assert(transformIdx >= 0 && transformIdx < 3);
      start_type = transformIdx;
      end_type = transformIdx + 1;
    }
    
    if (type == 1)
    {
      residue_blk8x8.copy(blk8x8);
      for (TransType = imax(start_type, 1); TransType < end_type; TransType++)
      {
        stored_eenc[TransType]->copy_coding_state(stored_eenc[0]);
        stored_ztree_luma_P_8x8[TransType]->copy_coding_state(stored_ztree_luma_P_8x8[0]);
      }
    }
  }
  
  quant[emodel]->loadmatrix(emodel, -1, 0, 0); // cq_start
  quant[emodel]->loadmatrix(emodel, -1, 0, 1); // cq_end

  for (TransType = start_type; TransType < end_type; TransType ++)
  {
    EntropyEncoder *tmp_eenc = eenc;
    ZeroTree *tmp_ztree = param.ztree_64;

    if (type == 1)
    {
      tmp_eenc = stored_eenc[imax(0, TransType)];
      tmp_ztree = stored_ztree_luma_P_8x8[imax(0, TransType)];
    }

    if (param.transform_selection == 1)
    {
      if (type == 1)
      {
        if (TransType < 2)
          blk8x8.copy(residue_blk8x8);
        
        rate = tmp_eenc->get_nbits();
      }
      
      if (TransType == 2)
        blk8x8.zero();
    }

    int transIdx = (TransType < 1) ? Transform::DCT : Transform::AlternateInter;
    trans[transIdx]->forward(&blk8x8);
    
    if(param.q_rdoffset8x8)
    {
      quant[emodel]->quant_rdoq(&blk8x8, tmp_eenc, emodel, lambda_rdoq, zmap[ZERO]->getcontext(i,j), tmp_ztree, TransType);  
    }
    else
    {
      quant[emodel]->quant(&blk8x8, 0, TransType);
    }
    
    zmap[ContextMap::inter]->set(i, j, blk8x8.nnz(), 2, 2);
    
#ifdef TRACE_DEBUG
    if(type==0)
    {
      ENABLE_ETRACE;
    }
    SPRINT_ETRACE("luma inter8x8 ");
    RPRINT_ETRACE; 
#endif
    
    is_zero_blk = blk8x8.is_zero();    
    tmp_eenc->encode_zblk(is_zero_blk, zmap[ZERO]->getcontext(i,j), emodel);
    
    if (is_zero_blk == FALSE) 
    {        
      tmp_eenc->encode_coeffs_8x8(tmp_ztree, &blk8x8);
    }

    if (type==0 || type==2)
    {
      if (is_zero_blk)
        mlc->zblks += 4;
      else
        mlc->nzblks += 4;
    }
    
#ifdef TRACE_DEBUG
    if(type==0)
      DISABLE_ETRACE;
#endif
    
    quant[emodel]->iquant(&blk8x8, 0);

    transIdx = (TransType < 1) ? Transform::DCT : Transform::AlternateInter;
    trans[transIdx]->inverse(&blk8x8);
    
    m_cur2_pred->put_blk_luma(&blk_pred8x8, i>>1, j>>1);
    m_cur2_qpe->put_blk_luma(&blk8x8, i>>1, j>>1);  
    blk8x8.add(blk_pred8x8);
    
    if (type == 1 && param.transform_selection == 1)
    {
      rate = tmp_eenc->get_nbits() - rate;
      
      distortion = blk8x8.ssd(org_blk8x8);

      if(param.sl_trans_sel)
        rdcost = distortion + lambda_rdoq * rate;
      else
        rdcost = distortion + lambda_mode * rate / 1.6;
      
      if (rdcost < min_rdcost)
      {
        transformIdx = TransType;
        min_rdcost = rdcost;
        min_blk8x8.copy(blk8x8);
      }
    }
  }
  
  if (type == 1 && param.transform_selection == 1)
  {
    if (transformIdx >= 1)
    {
      stored_eenc[0]->copy_coding_state(stored_eenc[transformIdx]);
      stored_ztree_luma_P_8x8[0]->copy_coding_state(stored_ztree_luma_P_8x8[transformIdx]);
    }

    m_cur2->put_blk_luma(&min_blk8x8, i>>1, j>>1);
  }
  else
    m_cur2->put_blk_luma(&blk8x8, i>>1, j>>1);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode one chroma 4x4 block
*
* \param[in] i Vertical position of 4x4 block
* \param[in] j Horizontal position of 4x4 block
* \param[in] c chroma component indicator
* \param[in] emodel Entropy model index (EntropyModel::I_luma, etc.)
* \param[in,out] cur2 Pointer to current frame (WFrame). Initially contains original frame. Finally contains reconstructed frame.
* \param[in] type flag to switch outputting bits or counting bits
* \param[in] lambda_rdoq lambda for RDO-Q 
*/
//----------------------------------------------------------------------------- 
void Encoder::encode_chroma_inter_blocks(int i, int j, int emodel, int type, double lambda_chroma)
{
  // type 0; 1st coding loop
  // type 1; mode selection
  // type 2; 2nd coding loop
  int is_zero_blk;
  Block blk(4, 4);
  Block blk_pred(4, 4);
  ZeroTree *tmp_ztree = (type == 1) ? stored_ztree_chroma_P : param.ztree[emodel];
  EntropyEncoder *tmp_eenc = (type == 1) ? stored_eenc[0] : eenc;
  int c;

  for (c = 0; c < 2; c++)
  {
    m_target->get_blk_chroma(&blk, i, j, c);
    
    pred_frame->get_blk_chroma(&blk_pred, i, j, c); 
    
#ifndef MVCOMP_ZERORES
    blk.sub(blk_pred);   
#else
    if (mf->get_pblk_unsafe(i, j)->mv_zero_res)
      blk.zero();
    else
      blk.sub(blk_pred);   
#endif
    
    trans[Transform::DCT]->forward(&blk);
    
    quant[emodel]->loadmatrix(emodel, -1, 0, 0); // cq_start
    quant[emodel]->loadmatrix(emodel, -1, 0, 1); // cq_end
    
    if(param.q_rdoffset)
    {
      quant[emodel]->quant_rdoq(&blk, tmp_eenc, emodel, lambda_chroma,zmap_chroma[c][ZERO]->getcontext(i,j), tmp_ztree, 0);     
    }
    else
    {
      quant[emodel]->quant(&blk, 0);
    }
    
    is_zero_blk = blk.is_zero();
    zmap_chroma[c][ZERO]->set(i,j,blk.nnz());
#ifdef TRACE_DEBUG
    if(type==0)
    {
      ENABLE_ETRACE;
    }
    SPRINT_ETRACE("chroma inter ");
    RPRINT_ETRACE;
#endif
    
    tmp_eenc->encode_zblk(is_zero_blk, zmap_chroma[c][ZERO]->getcontext(i,j), emodel);
    
    if (is_zero_blk == FALSE)
    {
      tmp_eenc->encode_coeffs(tmp_ztree, &blk, emodel); 
    }
    
#ifdef TRACE_DEBUG
    if(type==0)
      DISABLE_ETRACE;
#endif
    
    quant[emodel]->loadmatrix(emodel, c, 0, 0); // cq_start
    quant[emodel]->loadmatrix(emodel, c, 0, 1); // cq_end
    
    quant[emodel]->iquant(&blk, 0);
    trans[Transform::DCT]->inverse(&blk);       
    m_cur2_pred->put_blk_chroma(&blk_pred, i, j, c);
    m_cur2_qpe->put_blk_chroma (&blk, i, j, c);
    blk.add(blk_pred);   
    m_cur2->put_blk_chroma(&blk, i, j, c);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Block mode decision for 8x8 block 
*
* 8x8 block includes 4 of 4x4 luma block and 2 of 4x4 chroma blocks
*
* \param[in] cur Pointer to current frame (Frame).
* \param[in] i8 Vertical position of 4x4 block
* \param[in] j8 Horizontal position of 4x4 block
* \param[out] mlc Pointer to statistics for Motion/Luma/Chroma
* \param[out] ienc Pointer to I-block encoder
* \param[in] me Pointer to motion estimator 
* \param[in] low_zeromv_rdoq flag indicating if the frequency of zero MV in a frame is smaller than 50%
* \param[in] lambda_rdoq lambda for RDO-Q 
*/
//-----------------------------------------------------------------------------         
void Encoder::block_mode_decision(Frame *cur, int i8, int j8, MLC_Stat *mlc, MotionEst *me, int low_zeromv_rdoq, int &trans8x8Idx, int *trans4x4Idx)
{
  int i,j,c,ii,jj; // loop counters
  //  int is_zero_blk;
  int intra;
  
  PredBlk stored_predblk;
  
  int tmp_offset8x8[4][4][TX_NCOEFF8];
  int tmp_offset[4][4][TX_NCOEFF]; 
  int mode_loop;
#ifndef MVCOMP_ZERORES
  int icost_list[10];
  double rcost_list[10];
#else
  int const n_skip_modes_max = MV_MAX_NPREDICTORS;
  int icost_list[10+n_skip_modes_max];
  double rcost_list[10+n_skip_modes_max];
  PredBlk stored_skip[n_skip_modes_max];
#endif
  int n, min_cost_list, no_cost_list;
  double distortion;
  
  int mvshare_type;
  
#ifndef MVCOMP_ZERORES
  int mvshare_start=2;
  int number_of_modes=6;
#else
  int const mvshare_start=2+n_skip_modes_max;
  int       number_of_modes=6+n_skip_modes_max;
  int const mv_skip_start = 2;
  int const mv_skip_end = mv_skip_start+n_skip_modes_max;
#endif
  int const INTRA_COST_IDX = mvshare_start - 1;
  
  Block blk_org0(8, 8);
  Block blk_org4(4, 4);
  Block blk_org5(4, 4);
  Block blk_rec0(8, 8);
  Block blk_rec4(4, 4);
  Block blk_rec5(4, 4);
  int mode_number;
  
  Block blk(4, 4);
  
  if (!param.motion_sharing_enable)
  {
    number_of_modes = mvshare_start;
  }
  
  double lambda_mode = calc_lambda_mode(cur->qp, cur->reference_idc, 0, low_zeromv_rdoq);  
  double lambda_chroma = calc_lambda_mode(cur->qpC, cur->reference_idc, (param.enable_simple_lambdas ? 3 : 2), low_zeromv_rdoq); 
  double lambda_rdoq = calc_lambda_mode(cur->qp, cur->reference_idc, 1, low_zeromv_rdoq);  
  double lambda_rdoq8x8 = calc_lambda_mode(cur->qp, cur->reference_idc, (param.enable_simple_lambdas ? 1: 3), low_zeromv_rdoq);   
  double lambda_mode_chroma = calc_lambda_mode(cur->qpC, cur->reference_idc, (param.enable_simple_lambdas ? 2 : 0), low_zeromv_rdoq);
  
  // store  block data
  m_target->get_blk_luma(&blk_org0, i8>>1  , j8>>1  );
  m_target->get_blk_chroma(&blk_org4, i8>>1, j8>>1, 0);
  m_target->get_blk_chroma(&blk_org5, i8>>1, j8>>1, 1);
  
  PredBlk *cur_pblk = mf->get_pblk_unsafe(i8 >> 1, j8 >> 1);
  stored_predblk.copy(cur_pblk);
  
  
  // Save offsets
  for(i=0;i<4;i++)
  {
    for(j=0;j<4;j++)
    {
      if (param.trans8x8_enable) 
      {
        for(c=0;c<TX_NCOEFF8;c++)
        {
          tmp_offset8x8[i][j][c] = quant[i]->offset8x8[j][c];
        }
      }
      for(c=0;c<TX_NCOEFF;c++)
      {
        tmp_offset[i][j][c] = quant[i]->offset[j][c];
      }
    }
  }
  
  // end store block data
  
  for (n = 0; n < number_of_modes; n++)
    rcost_list[n] = FLT_INF;
  
  intra = 0;
  
  if (param.motion_sharing_enable)
  {
    mvshare_type = mf->mvshare_type(i8>>1, j8>>1);
  }
  else
  {
    mvshare_type = 0;
  }
  
  
  for (mode_number=0; mode_number< number_of_modes; mode_number++)
  {
#ifdef MVCOMP_ZERORES
    // intra
    // mode detected during ME
    // skip using info from a.
    // skip using info from b.
    // ... other skips.
    // 8x8 inter mvshare ver 2
    // 8x8 inter mvshare ver 4
    // 8x8 inter mvshare hor 2
    // 8x8 inter mvshare hor 4
#endif
    if (mode_number>=mvshare_start)
    {
      if(mvshare_type==0) continue;
      if(mvshare_type==1 && (mode_number-mvshare_start)>1) continue;
      if(mvshare_type==2 && ((mode_number-mvshare_start)<2 || (mode_number-mvshare_start)>3)) continue;
      if(mvshare_type==4 && (mode_number-mvshare_start)>3) continue;
      if(!param.mvshare_two_enable && mvshare_type && ((mode_number-mvshare_start)%2==0)) continue;
    }
    
    mode_loop = (mode_number == 0) ? -1 : 0;
    intra = mode_loop < 0;
    
    // Rotate indices to place intra right before motion sharing modes
    if (mode_number >= mvshare_start)
      no_cost_list = mode_number;
    else if (mode_number == 0)
      no_cost_list = INTRA_COST_IDX;
    else
      no_cost_list = mode_number - 1;
    
    
    cur_pblk->mvshare_size[0] = 0;
    cur_pblk->mvshare_size[1] = 0;
#ifdef MVCOMP_ZERORES
    cur_pblk->mv_zero_res = 0;
#endif
    
    if (intra)
    {
      cur_pblk->clear();
      cur_pblk->intra = &mf->blk_array_I[mf->width*(i8>>1)+(j8>>1)];
    }
    else
    {
      cur_pblk->intra = NULL;
#ifdef MVCOMP_ZERORES
      if (cur_pblk->list < 0)
        cur_pblk->list = 0;
#endif
      if (mode_number >= mvshare_start && mode_number < mvshare_start+4)
      {
        int mvshare_mode = mode_number - mvshare_start;
        cur_pblk->mvshare_size[mvshare_mode >> 1] = 2 * (1 + (mvshare_mode & 1));
      }
#ifdef MVCOMP_ZERORES
      else if (mode_number>=mv_skip_start && mode_number < mv_skip_end)
      {
        // Get appropriate mv predictor, and set our mv to it.
        // If it's not possible, we 'continue' the loop to skip that attempt.
        if (mode_number-mv_skip_start >= param.mv_cur_comp_config()->mv_npredictors)
          continue;
        if (!param.mv_cur_comp_config()->mv_zero_residue)
          continue; // No skip processing wanted.
        if (!skip_initialize(me, mlc, i8>>1, j8>>1, param.mv_cur_comp_config()->mv_predictors[mode_number-mv_skip_start]))
          continue;
        
        // save it.
        stored_skip[mode_number-mv_skip_start].copy(cur_pblk);
        stored_skip[mode_number-mv_skip_start].mv_zero_res = 1;
      }
#endif
    }     
    
    // initialize block data
    
    stored_eenc[0]->copy_coding_state(eenc);
    
    if (intra)
    {
      stored_ztree_luma_I->copy_coding_state(param.ztree[EntropyModel::I_luma]);
      stored_ztree_chroma_I->copy_coding_state(param.ztree[EntropyModel::I_chroma]);
    }
    else
    {
      stored_ztree_luma_P[0]->copy_coding_state(param.ztree[EntropyModel::P_luma]);
      stored_ztree_chroma_P->copy_coding_state(param.ztree[EntropyModel::P_chroma]);
    }
    
    if (param.trans8x8_enable)
      stored_ztree_luma_P_8x8[0]->copy_coding_state(param.ztree_64);
    stored_ztree_mv_P->copy_coding_state(param.ztree[4]);
    stored_ztree_mv_B->copy_coding_state(param.ztree[5]);
    
    // end: initialize block data
    
    // end: setting of block parameter
    
    icost_list[no_cost_list] = stored_eenc[0]->get_nbits();
    
    // Code lists / MV
    codemvsbi_blk(me, mlc, (i8>>1), (j8>>1), false, stored_eenc[0]);
    
    // Code luma
    i = i8;
    j = j8;
    intra = (cur_pblk->intra != NULL);

    if (intra)
    {
      m_ienc->encode_luma_block_NxN(i,j, stored_eenc[0], NULL, 0, stored_ztree_luma_I, stored_ztree_luma_P_8x8[0]);
      selected_transform_per_mode[no_cost_list] = m_ienc->get_transform();
    }
    else
    {
      m_ienc->intra_pred->set_sip_type(i>>1, j>>1, SIP_INTER);  //// set the type

      if (mf->get_transform_blocksize(i>>1, j>>1) == 1)
      {
        selected_transform_per_mode[no_cost_list] = -1;
        encode_luma_inter_block_8x8(i, j, EntropyModel::P_luma, pred_frame, mlc, 1, lambda_rdoq8x8, lambda_mode, selected_transform_per_mode[no_cost_list]);
      }
      else
      {
        for (ii = 0; ii < 2; ii++)
        {
          for (jj = 0; jj < 2; jj++)
          {
            selected_transform4x4_per_mode[no_cost_list][ii+jj*2] = -1;
            encode_luma_inter_block(i, j, ii, jj, EntropyModel::P_luma, mlc, 1, lambda_rdoq, lambda_mode, selected_transform4x4_per_mode[no_cost_list][ii+jj*2]);
          }
        }
      }
    }
    
    // Code chroma
    i = i8>>1;
    j = j8>>1;
    
    if (intra)
    {
      m_ienc->encode_chroma_blocks(i, j, stored_eenc[0], NULL, 0, stored_ztree_chroma_I, SIP4x4);
    }
    else
    {
      encode_chroma_inter_blocks(i, j, EntropyModel::P_chroma, 1, lambda_chroma);
    }

    // Consider cost for neighboring blocks
    if (mf->get_pblk((i8>>1), (j8>>1)+1) != NULL)
      codemvsbi_blk(me, mlc, (i8>>1), (j8>>1)+1, false, stored_eenc[0]);
    
    if (mf->get_pblk((i8>>1)+1, (j8>>1)) != NULL)
      codemvsbi_blk(me, mlc, (i8>>1)+1, (j8>>1), false, stored_eenc[0]);
    
    // Determine rate
    stored_eenc[0]->terminate();
    icost_list[no_cost_list] = stored_eenc[0]->get_nbits() - icost_list[no_cost_list];
        
    // Fetch blocks to compute distortion
    m_cur2->get_blk_luma(&blk_rec0, i8>>1  , j8>>1  );
    m_cur2->get_blk_chroma(&blk_rec4, i8>>1, j8>>1, 0);
    m_cur2->get_blk_chroma(&blk_rec5, i8>>1, j8>>1, 1);
    
    distortion = blk_rec0.ssd(blk_org0);
    
    if (param.lambda_mode_chroma)
    {
      distortion += blk_rec4.ssd(blk_org4) * lambda_mode / lambda_mode_chroma;
      distortion += blk_rec5.ssd(blk_org5) * lambda_mode / lambda_mode_chroma;
    }
    else
    {
      distortion += blk_rec4.ssd(blk_org4);
      distortion += blk_rec5.ssd(blk_org5);
    }
    
    rcost_list[no_cost_list] = distortion + lambda_mode * icost_list[no_cost_list];
    
    //--------- Restore data

    cur_pblk->copy(&stored_predblk);

    // Restore offsets
    for (i=0;i<4;i++)
    {
      for (j=0;j<4;j++)
      {
        if (param.trans8x8_enable) 
        {
          for(c=0;c<TX_NCOEFF8;c++)
          {
            quant[i]->offset8x8[j][c] = tmp_offset8x8[i][j][c];
          }
        }
        for(c=0;c<TX_NCOEFF;c++)
        {
          quant[i]->offset[j][c] = tmp_offset[i][j][c];
        }
      }
    }
  } // mode_number
  
  // mode selection
  min_cost_list = 0;
  for (n = 1; n < number_of_modes; n++)
  {
    if (rcost_list[n] < rcost_list[min_cost_list])
      min_cost_list = n;
  }
  
  trans8x8Idx = selected_transform_per_mode[min_cost_list];
  for (n = 0; n < 4; n ++)
    trans4x4Idx[n] = selected_transform4x4_per_mode[min_cost_list][n];
  // end: mode selection
    
#ifdef MVCOMP_ZERORES
  if (min_cost_list+1 >= mv_skip_start && min_cost_list+1 < mv_skip_end)
  {
    cur_pblk->copy(&stored_skip[min_cost_list+1-mv_skip_start]);
    cur_pblk->mv_zero_res = 1;
  }
#endif
  
  if (param.motion_sharing_log)
  {
    if (mvshare_type==0)
      mlc->mvshare_off++;
    else if(min_cost_list>=mvshare_start && (min_cost_list-mvshare_start)%2==0)
      mlc->mvshare_two++;
    else if(min_cost_list>=mvshare_start && (min_cost_list-mvshare_start)%2==1)
      mlc->mvshare_four++;
    else
      mlc->mvshare_zero++;
  }
  
  cur_pblk->mvshare_size[0] = 0;
  cur_pblk->mvshare_size[1] = 0;
  
  if (min_cost_list==mvshare_start)
  {
    cur_pblk->mvshare_size[0]=2;
  }
  else if(min_cost_list==mvshare_start+1)
  {
    cur_pblk->mvshare_size[0]=4;
  }
  else if(min_cost_list==mvshare_start+2)
  {
    cur_pblk->mvshare_size[1]=2;
  }
  else if(min_cost_list==mvshare_start+3)
  {
    cur_pblk->mvshare_size[1]=4;
  }
  
  
  intra = min_cost_list == INTRA_COST_IDX;
  
  if (intra)
  {
    mf->set_block_intra(i8>>1, j8>>1);
    cur_pblk->list = -1;
  }
  else
  {
    cur_pblk->intra = NULL;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode frame header
*
* \param[in] cur Pointer to current frame
*/
//-----------------------------------------------------------------------------         
void Encoder::frame_header(Frame *cur)
{
#ifdef TRACE_DEBUG
  ENABLE_ETRACE;
#endif
  // Write frame type. 
  // This happens to be 01 for P, and 11 for B (written from LSB to MSB)
  // Also write intra type when frame_type == ft_I_frame.
  eenc->encode_frame_type (cur->frame_type);
  if (cur->frame_type==ft_I_frame)
  {
    eenc->encode_I_frame_type (cur->intra_frame_type);
  }
  eenc->encode_frame_counter(cur->frame_order_counter);
  eenc->encode_reference_idc(cur->reference_idc);
  eenc->encode_deltaqp(cur->delta_qp);

#ifdef Q_OFFSET
  eenc->encode_q_offsets(param.iTheta, param.iTheta8x8); // define q_offsets here!
#endif

  eenc->encode_quantization_set_id_half(); // cq_start
  eenc->encode_qmatrix_half(0);
  eenc->encode_qmatrix_half(1); // cq_end

  eenc->encode_dummy();

  eenc->encode_edge_prediction_flag(param.edge_prediction_ipd);
  if(param.edge_prediction_ipd == TRUE)
  {
    eenc->encode_threshold_edge(param.threshold_edge_vector);
  }

  if (cur->frame_type != ft_I_frame)
    eenc->encode_number_of_references(ref_mgr->single_ref_frame);
#ifdef MVCOMP_ZERORES
  if (cur->frame_type != ft_I_frame)
    eenc->encode_mvcomp_config(param.mv_cur_comp_config_idx);
#endif
#ifdef TRACE_DEBUG
  DISABLE_ETRACE;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode one frame
*
*
* \param[in,out] cur Pointer to current frame. Initially contains original frame. Finally contains reconstructed frame.
* \param[out] buf Pointer to current byte buffer.
* \param[out] mlc Pointer to statistics for Motion/Luma/Chroma
* \param[out] ienc Pointer to I-block encoder
* \param[in] type %Frame type (LIST_0, BI_PRED)
* \param[in] aif_type type of AIF
* \param[in] inter_type subpel size
* \param[in] tmp_cnt temporal counter
*
* \return number of bits
*/
//-----------------------------------------------------------------------------
int Encoder::encode_one_frame(Frame *cur, Frame *cur_pred, Frame *cur_qpe_m, Frame *cur_qpe_s, Frame * /*cur_orig*/, Frame * /*cur_filtered*/, Buffer *buf, MLC_Stat *mlc, int type, int aif_type, int /*inter_type*/, int tmp_cnt) 
{
  int i,j,ii,jj; // loop counters
  Block blk(4, 4);
  //  int is_zero_blk;
  int intra;

  zmap[ZERO]->reset();
  zmap[ONE ]->reset();
  zmap_chroma[ZERO][ZERO]->reset(); 
  zmap_chroma[ZERO][ONE ]->reset(); 
  zmap_chroma[ONE][ZERO]->reset(); 
  zmap_chroma[ONE][ONE ]->reset(); 

  int i8, j8;

  double lambda_rdoq;
  int low_zeromv_rdoq;
  double lambda_chroma;
  double lambda_rdoq8x8;

#ifdef SIPNxN_FRAME_STAT
  ienc->reset_sip_mode_count();  
  for (i=0; i<NSIP; i++)
    ienc->sip_type_count[i]=0;
#endif

  mlc->reset();
  mf->frame_type = cur->frame_type;
#ifdef TRACE_DEBUG
  INITFRAME_ETRACE(cur->frame_no);
  ENABLE_ETRACE;
  SVPRINT_ETRACE("FRAME TYPE", cur->frame_type);
  RPRINT_ETRACE;
  DISABLE_ETRACE;
#endif

  if(param.aif && aif_type==-1)
  {
    saif_half->ResetAdaptiveFilter(cur);
  }
  cur->melq += param.mel_offset;
  m_me[0]->reset(cur);
  cur->melq -= param.mel_offset;

  m_me[1]->reset(cur);
  m_me[2]->reset(cur);
  m_me[3]->reset(cur);
  m_me[1]->nloops = param.me_loops1;    
  m_me[2]->nloops = param.me_loops2;
  m_me[3]->nloops = param.me_loops3;
  m_me[0]->intra_cost_update = intra_cost_update; 
  m_me[1]->intra_cost_update = intra_cost_update;    
  m_me[2]->intra_cost_update = intra_cost_update;
  m_me[3]->intra_cost_update = intra_cost_update;

  m_ienc->setFrames(m_cur2, m_cur2_pred, m_cur2_qpe, m_target); // Make sure IEncoder also has access to these frames
  
  lf->reset(ZERO);

#ifdef MVCOMP_ZERORES
  // Choose config.
  param.mv_cur_comp_config_idx = param.mv_default_comp_config_idx;
  int will_reset = 0;
#endif

  //LP
  if(cur->frame_type!=ft_I_frame)
  {  // H.264 filter
#ifndef MVCOMP_ZERORES
    if(param.pyramid_coding==0 && (aif_type==-1 || param.aif==FALSE))
    {
      eenc->reset(buf);
    }
    else if(cur->frame_type==ft_B_frame)
#else
    // Reset moved to after initialization code, in order
    // to reset all but movement if required.
    if(param.pyramid_coding==0 && (aif_type==-1 || param.aif==FALSE))
    {
      will_reset = 1;
    }
    if(cur->frame_type==ft_B_frame)
#endif
    {
      int pelIdx = (ref_mgr->ref_process->inter_type[0][0] == 2); // is 1/4th pel
      int frameTypeIdx = 0;
      int framePosIdx = frame_pos_intermediate;
      if (param.pyramid_coding && cur->intermed_idx == 0)
        framePosIdx = frame_pos_edge;
      else if (cur->reference_idc == DISPOSABLE_PIC)
        framePosIdx = frame_pos_disposable;
      
      eenc->copy_coding_state_cs(stored_eencF[framePosIdx][pelIdx][frameTypeIdx]);
      
#ifdef MVCOMP_ZERORES
      if (!will_reset)
#endif
      {
        param.ztree[4]->copy_coding_state(stored_ztreeF_P[framePosIdx][pelIdx][frameTypeIdx]);
        param.ztree[5]->copy_coding_state(stored_ztreeF_B[framePosIdx][pelIdx]);

#ifdef FRAME_SUBSTREAMS
        param.ztree[5]->update_bit_costs(param.ztree[5]->get_pe(eenc->GetEntropyModel()));
        param.ztree[4]->update_bit_costs(param.ztree[4]->get_pe(eenc->GetEntropyModel()));
#else
        param.ztree[5]->update_bit_costs();
        param.ztree[4]->update_bit_costs();
#endif
      }
    }
    else
    {
      int pelIdx = (ref_mgr->ref_process->inter_type[0][0] == 2); // is 1/4th pel
      int frameTypeIdx = 1;
      int framePosIdx = 2;
      if (param.pyramid_coding && cur->intermed_idx == 0)
        framePosIdx = 0;
      else if (cur->reference_idc == DISPOSABLE_PIC)
        framePosIdx = 1;
      
      eenc->copy_coding_state_cs(stored_eencF[framePosIdx][pelIdx][frameTypeIdx]);
      
#ifdef MVCOMP_ZERORES
      if (!will_reset)
#endif
      {
        param.ztree[4]->copy_coding_state(stored_ztreeF_P[framePosIdx][pelIdx][frameTypeIdx]);
        
#ifdef FRAME_SUBSTREAMS
        param.ztree[4]->update_bit_costs(param.ztree[4]->get_pe(eenc->GetEntropyModel()));
#else
        param.ztree[4]->update_bit_costs();
#endif
      }
    }
#ifdef MVCOMP_ZERORES
    if(will_reset)
      eenc->reset(buf, param.mv_cur_comp_config()->mv_keep_learned_movement);
    else if (param.mv_cur_comp_config()->mv_keep_learned_movement)
      eenc->recalc_movement(); // Not reset, but reset to global image probas.
#endif
  }
  else
  {
    eenc->reset(buf);
    
    if (cur->frame_no == 0)
    {
      int i, j, k;
      
      for (i = 0; i < 3; i++)
      {
        for (j = 0; j < 2; j++)
        {
          for (k = 0; k < 2; k++)
          {
            stored_eencF[i][j][k]->copy_coding_state_cs(eenc);
            stored_ztreeF_P[i][j][k]->copy_coding_state(param.ztree[4]);
          }
          stored_ztreeF_B[i][j]->copy_coding_state(param.ztree[5]);
        }
      }
    }
    
    if(ref_mgr->temp_rpps_id[tmp_cnt]>=0)
    {
      if(ref_mgr->ref_pic_parameter_set_id_shared[ref_mgr->temp_rpps_id[tmp_cnt]]==0)
      {
        eenc->encode_rpps_type(1);
        eenc->encode_rpps_id(ref_mgr->temp_rpps_id[tmp_cnt]);
        for(i=0;i<LISTS;i++)
        {
          for(j=0;j<MAX_REF_FRAMES;j++)
          {
            eenc->encode_inter_type(ref_mgr->ref_process->inter_type[i][j]);
            eenc->encode_filter_type(ref_mgr->ref_process->filter_type[i][j]);
            eenc->encode_ic_disable(ref_mgr->ref_process->ic_disable[i][j]);
            eenc->encode_round_type(ref_mgr->ref_process->round_type[i][j]);
          }
        }
      }
      eenc->encode_rpps_type(0);
      eenc->encode_rpps_id(ref_mgr->temp_rpps_id[tmp_cnt]);
    }
    else
    {
      printf("Error Exceed max_rpps_id");
      return INT_INF;
    }
    frame_header(cur);
  }

  if (cur->frame_type == ft_I_frame) 
  {
    mf->set_all_intra();
  }
  else 
  {  
    // Inter Frame Set-Up
    mf->fix_lists();

    if (change_mv_precision)
    {
      mf->change_resolution_motion_field(ref_mgr, cur->frame_type);
    }
    else if (skip_me_loop_on==0)
    {
      mf->reset_motion_field();
    }

    if(param.aif)
    {
      saif_half->AIF_FilterFlag=aif_type+2;
      if(aif_type!=-1)
        ref_mgr->AIF_replace_frame(cur, saif_half);
    }

    for (i=0; i<4; i++) 
    { // do each type of ME
      if (skip_me_loop_on == 0 || i >= 2)
      {
        m_me[i]->intra_init(m_ienc, eenc); // set intra block values
        m_me[i]->estimate(cur, type); // ME search
      }
    }  // ME LOOP

    if (skip_me_loop_on==0 && change_mv_precision==0 && ((cur->reference_idc==DISPOSABLE_PIC && (cur->frame_type==ft_B_frame ? param.me_repeat_disposable_b : param.me_repeat_disposable_p)) || (cur->intermed_idx==0 && param.me_repeat_edge) || (cur->reference_idc!=DISPOSABLE_PIC && cur->intermed_idx!=0 && (cur->frame_type==ft_B_frame ? param.me_repeat_intref_b : param.me_repeat_intref_p))))
    {
      m_me[2]->nloops = (cur->reference_idc==DISPOSABLE_PIC ? param.me_loops2_disposable : (cur->intermed_idx==0 ? param.me_loops2_edge : param.me_loops2_intref));
      m_me[3]->nloops = (cur->reference_idc==DISPOSABLE_PIC ? param.me_loops3_disposable : (cur->intermed_idx==0 ? param.me_loops3_edge : param.me_loops3_intref));

      for (j = 0; j < (cur->reference_idc==DISPOSABLE_PIC ? (cur->frame_type==ft_B_frame ? param.me_repeat_disposable_b : param.me_repeat_disposable_p) : (cur->intermed_idx==0 ? param.me_repeat_edge : (cur->frame_type==ft_B_frame ? param.me_repeat_intref_b : param.me_repeat_intref_p))); j++)
      {
#ifdef SIXSTEPME_LOG
        printf("6step ME on; ptype %d reference_idc %d inter_type %d aif_type %d\n", cur->frame_type, cur->reference_idc, inter_type, AIF_FilterFlag-1);
#endif
        for (i = 2; i < 4; i++)
        {
          m_me[i]->estimate(cur, type); // ME search
        }
      }
    }

    if(cur->frame_type!=ft_I_frame)
    {
      eenc->reset(buf);
      if(ref_mgr->temp_rpps_id[tmp_cnt]>=0)
      {
        if(ref_mgr->ref_pic_parameter_set_id_shared[ref_mgr->temp_rpps_id[tmp_cnt]]==0)
        {
          eenc->encode_rpps_type(1);
          eenc->encode_rpps_id(ref_mgr->temp_rpps_id[tmp_cnt]);
          for(i=0;i<LISTS;i++)
          {
            for(j=0;j<MAX_REF_FRAMES;j++)
            {
              eenc->encode_inter_type(ref_mgr->ref_process->inter_type[i][j]);
              eenc->encode_filter_type(ref_mgr->ref_process->filter_type[i][j]);
              eenc->encode_ic_disable(ref_mgr->ref_process->ic_disable[i][j]);
              eenc->encode_round_type(ref_mgr->ref_process->round_type[i][j]);
            }
          }
        }
        eenc->encode_rpps_type(0);
        eenc->encode_rpps_id(ref_mgr->temp_rpps_id[tmp_cnt]);
      }
      else
      {
        printf("Error Exceed max_rpps_id");
        return INT_INF;
      }
      frame_header(cur);
    }
#ifdef TRACE_DEBUG
    ENABLE_ETRACE;
#endif   
    // AIF Filter creator
    if(param.aif)
    {
      if(saif_half->AIF_FilterFlag==1)
      { 
        if(!cur->reference_idc)
        {
          saif_half->AIF_calculate_coef_SAIF(cur, mf, ref_mgr);
        }
      } else if(saif_half->AIF_FilterFlag==2)
      {
         // from AIF_encode_one_frame
         mlc->aif_bits = -eenc->get_nbits();
         saif_half->encode_filter_information(eenc);
         mlc->aif_bits += eenc->get_nbits();
    }
    }
  ref_mgr->getAIFcoef(saif_half);
    // end: AIF Filter creator
    // end: Inter Frame Set-Up
  }   
#ifdef TRACE_DEBUG
  DISABLE_ETRACE;
#endif
  // setting for quantizetion
  int low_zeromv = 0;
  if (param.q_adaptoffset == 4 && cur->frame_type != ft_I_frame && cur->reference_idc != DISPOSABLE_PIC)
    low_zeromv = mv_count();
      
  for (i = 0; i < 4; i++)
  {
    quant[i]->low_zeromv = low_zeromv;
    quant[i]->reference_idc = cur->reference_idc;
    quant[i]->frame_type = cur->frame_type;
  }

  low_zeromv_rdoq = mv_count();

  // end: setting for quantization


  m_cur2->expand(cur); // convert frame to 16-bit representation  
  m_target->copy(m_cur2);
  m_cur2_qpe->clear();
  m_cur2_pred->clear();

  m_ienc->iqp = cur->qp;
  m_ienc->iqpC = cur->qpC;
  m_ienc->frame_type = cur->frame_type; 
  m_ienc->reset_zmap();

  // do luma  

  double lambda_mode = calc_lambda_mode(cur->qp, cur->reference_idc, 0, low_zeromv_rdoq);  
  lambda_chroma = calc_lambda_mode(cur->qpC, cur->reference_idc, (param.enable_simple_lambdas ? 3: 2), low_zeromv_rdoq); 
  lambda_rdoq = calc_lambda_mode(cur->qp, cur->reference_idc, 1, low_zeromv_rdoq);  
  lambda_rdoq8x8 = calc_lambda_mode(cur->qp, cur->reference_idc, (param.enable_simple_lambdas ? 1: 3), low_zeromv_rdoq);

#ifdef CABAC_SYNC_POINTS
  // reset_synchro on contexts
  if (param.superblock_synchro)
    eenc->GetEntropyModel()->resetSynchro();
#endif
  int q8l,p8l,i8l,j8l,q8i,p8i,i8i,j8i,i8s,j8s;
      
  q8l = cur->height4;
  q8i = ZHV*param.sb_height;
  p8l = cur->width4;
  p8i = ZHV*param.sb_width;

  i8l = 0; // avoid compiler warning
  j8l = 0; // avoid compiler warning

  for (int q8=0; q8<q8l; q8+=q8i) 
  {

    for (int p8=0; p8<p8l; p8+=p8i) 
    {
      i8s = q8;
      i8l = q8+ZHV*param.sb_height;
      i8i = ZHV;

      j8s = p8;
      j8l = p8+ZHV*param.sb_width;
      j8i = ZHV;

      if(param.intra16x16)
      {
        i8i = ZHV*2;
        j8i = ZHV*2;
      }
#ifdef CABAC_SYNC_POINTS
      // init_synchro on contexts
      if (param.superblock_synchro)
        eenc->GetEntropyModel()->initSynchro(q8/q8i, p8/p8i);
#endif
#ifdef TRACE_DEBUG
      ENABLE_ETRACE;
      INITBLOCK_ETRACE(p8, q8); //position of superblock
      DISABLE_ETRACE;
#endif

      if (param.intra16x16)
      {
        int width_scale  = 2;
        int height_scale = 2;

        m_ienc->best_coeff_16x16 = new Block(height_scale*8, width_scale*8);

        for (i8 = i8s; i8 < imin(i8l, cur->height4); i8 += i8i) 
        {
          for (j8 = j8s; j8 < imin(j8l, cur->width4); j8 += j8i) 
          {
            m_ienc->intra_pred->avail_up_right_16x16 = (i8 == i8s) || (j8 < j8l - j8i);

#ifdef I16_IN_P_FRAME
            if (cur->frame_type != ft_I_frame)
            {
              encode_16x16_blocks(cur, i8, j8, mlc, eenc, low_zeromv_rdoq);
            }
            else
#endif
            {
              mlc->luma_bits -= eenc->get_nbits();
              i = i8; j = j8;
              m_ienc->encode_luma_block_16x16(i>>1, j>>1, eenc, mlc, 1, NULL, NULL, NULL);   // index change to use 8x8 unit
              mlc->bit_count_luma += eenc->bits_luma;              
              mlc->iblks += min(cur->height4 - i, 4) * min(cur->width4 - j, 4);
              mlc->luma_bits += eenc->get_nbits();

              i = i8 >> 1; j = j8 >> 1;
              mlc->chroma_bits -= eenc->get_nbits();
              m_ienc->encode_chroma_block_8x8(i, j, eenc, mlc);
              mlc->chroma_bits += eenc->get_nbits();         
            }
          } // j8
        } // i8
      }
      else
      {
        for (i8=i8s; i8<i8l; i8+=i8i) 
        {    
          for (j8=j8s; j8<j8l; j8+=j8i) 
          {
            if(!(i8<cur->height4&&j8<cur->width4)) continue; //super block
      // NOTE; block coding is performed in this function to calculate the RD cost using the actual bit cost

            int trans8x8Idx;
            int trans4x4Idx[4];
      if(cur->frame_type != ft_I_frame)
      {
        block_mode_decision(cur,i8, j8, mlc, m_me[2], low_zeromv_rdoq, trans8x8Idx, trans4x4Idx);
      }

      mlc->motion_bits -= eenc->get_nbits();
      if (cur->frame_type != ft_I_frame)
        codemvsbi_blk(m_me[2], mlc, (i8>>1), (j8>>1), true, eenc);
      mlc->motion_bits += eenc->get_nbits();

      i=i8; j=j8;

      mlc->luma_bits -= eenc->get_nbits();  

            intra = mf->get_pblk_unsafe(i >> 1, j >> 1)->intra != NULL;
            
            if (intra)
            {
              m_ienc->encode_luma_block_NxN(i, j, eenc, mlc, 1, NULL, NULL);
              mlc->bit_count_luma += eenc->bits_luma;
              mlc->iblks += 4;
            }
            else
            {
              m_ienc->intra_pred->set_sip_type(i>>1, j>>1, SIP_INTER);  //// set the type

              if (mf->get_transform_blocksize(i>>1, j>>1) == 1)
              {
                encode_luma_inter_block_8x8(i, j, EntropyModel::P_luma, pred_frame, mlc, 0, lambda_rdoq8x8, (param.enable_simple_lambdas ? lambda_mode : lambda_rdoq8x8), trans8x8Idx);
              }
              else
              {
                for (ii = 0; ii < 2; ii++)
                {
                  for (jj = 0; jj < 2; jj++)
                  {
                    encode_luma_inter_block(i, j, ii, jj, EntropyModel::P_luma, mlc, 0 ,lambda_rdoq,  (param.enable_simple_lambdas ? lambda_mode : lambda_rdoq), trans4x4Idx[ii+jj*2]);
                  }
                }
              }
            }
            

            mlc->luma_bits += eenc->get_nbits();
            mlc->chroma_bits -= eenc->get_nbits();
            i = i8>>1; j= j8>>1;

            if (intra)
            {
              m_ienc->encode_chroma_blocks(i, j, eenc, mlc, 1, NULL, SIP4x4);
              mlc->bit_count_chroma += eenc->bits_chroma;
            }
            else
            {
              encode_chroma_inter_blocks(i, j, EntropyModel::P_chroma, 0, lambda_chroma);
            }

      mlc->chroma_bits += eenc->get_nbits();

      if(cur->frame_type != ft_I_frame)
      {    
        if (mf->get_pblk(i8>>1, j8>>1)->intra)
        {
          int sip_type = m_ienc->intra_pred->get_sip_type(i8>>1, j8>>1);
          zmap[ContextMap::inter]->copy_from_intra(zmap[ContextMap::intra], i8, j8, sip_type);
          zmap_chroma[0][ContextMap::inter]->set(i8>>1, j8>>1, zmap_chroma[0][ContextMap::intra]->get(i8>>1, j8>>1));
          zmap_chroma[1][ContextMap::inter]->set(i8>>1, j8>>1, zmap_chroma[1][ContextMap::intra]->get(i8>>1, j8>>1));
        }
        else
        {
          zmap[ContextMap::intra]->copy_from_inter(zmap[ContextMap::inter], i8, j8);
          zmap_chroma[0][ContextMap::intra]->set(i8>>1, j8>>1, zmap_chroma[0][ContextMap::inter]->get(i8>>1, j8>>1));
          zmap_chroma[1][ContextMap::intra]->set(i8>>1, j8>>1, zmap_chroma[1][ContextMap::inter]->get(i8>>1, j8>>1));
        }
      }
    }
  }
}

if (param.intra16x16)
{
  delete m_ienc->best_coeff_16x16;
}

#ifdef CABAC_SYNC_POINTS
// update_synchro on contexts
if (param.superblock_synchro)
eenc->GetEntropyModel()->updateSynchro(p8/p8i);
#endif
    }
  }//superblock

  lf->set_luma(mf, zmap[ZERO]);
  lf->set_chroma(mf, zmap_chroma[0][ZERO], 0); 
  lf->set_chroma(mf, zmap_chroma[1][ZERO], 1);     

#ifdef SIPNxN_FRAME_STAT  
  ienc->print_sip_mode_count();   
  for (i=0; i<NSIP; i++ )
    printf ("sip type = %d, count = %d\n", i, ienc->sip_type_count[i]); 
#endif

  m_cur2->collapse(cur); 
  m_cur2_pred->collapse       (cur_pred); 
  m_cur2_qpe->collapse_modulus(cur_qpe_m); 
  m_cur2_qpe->collapse_sign   (cur_qpe_s);

  if ((param.deblockingfilter&3) == 3)
  {
    dbf->filter_data_copy(lf, mf);

    dbf->init_alpha_values( cur->qp, cur->qpC );
    if (param.deblockingfilter&8)
    {
      dbf->set_filter_modes();
    }
    dbf->fixed_deblocking_filter( dbf->str_mf, cur->luma, cur->luma, 0);
    dbf->fixed_deblocking_filter( dbf->str_mf, cur->crcb[0], cur->crcb[0], 1 );
    dbf->fixed_deblocking_filter( dbf->str_mf, cur->crcb[1], cur->crcb[1], 2 ); 
  }

  lf->filter(cur);

  eenc->nbits_wlf_coeffs = eenc->nbits;

  //printf("%d\n", eenc->nbits);
  
  wlf->store_low_zeromv_rdoq=low_zeromv_rdoq;
  return eenc->nbits;
}

#ifdef I16_IN_P_FRAME
void Encoder::encode_16x16_blocks(Frame *cur, int i, int j, MLC_Stat *mlc, EntropyEncoder *eenc, int low_zeromv_rdoq)
{
  int i4, j4;
  int c;
  int i0, j0;
  int ii, jj;
  double lambda_mode = calc_lambda_mode(cur->qp, cur->reference_idc, 0, low_zeromv_rdoq);  
  double lambda_chroma = calc_lambda_mode(cur->qpC, cur->reference_idc, (param.enable_simple_lambdas ? 3: 2), low_zeromv_rdoq); 
  double lambda_rdoq = calc_lambda_mode(cur->qp, cur->reference_idc, 1, low_zeromv_rdoq);  
  double lambda_rdoq8x8 = calc_lambda_mode(cur->qp, cur->reference_idc, (param.enable_simple_lambdas ? 1: 3), low_zeromv_rdoq);   
  double lambda_mode_chroma = calc_lambda_mode(cur->qpC, cur->reference_idc, (param.enable_simple_lambdas ? 2: 0), low_zeromv_rdoq);
  int blkIdx;
  int trans8x8Idx[4];
  int trans4x4Idx[4][4];
  PredBlk *pblk;

  // Block decision for individual 8x8 blocks
  blkIdx = 0;
  for (i4 = i; i4 < imin(i + 4, m_cur2->height4); i4 += 2)
  {
    for (j4 = j; j4 < imin(j + 4, m_cur2->width4); j4 += 2, blkIdx++)
    {
      // need to set availability of 8x8
      m_ienc->intra_pred->avail_up_right_8x8 = (j == j4) || (m_ienc->intra_pred->avail_up_right_16x16 && i == i4);
      trans8x8Idx[blkIdx] = -1;
      block_mode_decision(cur, i4, j4, mlc, m_me[2], 0, trans8x8Idx[blkIdx], trans4x4Idx[blkIdx]);
      if (mf->get_pblk(i4>>1, j4>>1)->intra)
      {
        int sip_type = m_ienc->intra_pred->get_sip_type(i4>>1, j4>>1);
        zmap[ContextMap::inter]->copy_from_intra(zmap[ContextMap::intra], i4, j4, sip_type);
        zmap_chroma[0][ContextMap::inter]->set(i4>>1, j4>>1, zmap_chroma[0][ContextMap::intra]->get(i4>>1, j4>>1));
        zmap_chroma[1][ContextMap::inter]->set(i4>>1, j4>>1, zmap_chroma[1][ContextMap::intra]->get(i4>>1, j4>>1));
      }
      else
      {
        zmap[ContextMap::intra]->copy_from_inter(zmap[ContextMap::inter], i4, j4);
        zmap_chroma[0][ContextMap::intra]->set(i4>>1, j4>>1, zmap_chroma[0][ContextMap::inter]->get(i4>>1, j4>>1));
        zmap_chroma[1][ContextMap::intra]->set(i4>>1, j4>>1, zmap_chroma[1][ContextMap::inter]->get(i4>>1, j4>>1));        
      }
    }
  }

  stored_eenc[0]->copy_coding_state(eenc);
  if (param.trans8x8_enable)
    stored_ztree_luma_P_8x8[0]->copy_coding_state(param.ztree_64);
  stored_ztree_luma_I->copy_coding_state(param.ztree[EntropyModel::I_luma]);
  stored_ztree_chroma_I->copy_coding_state(param.ztree[EntropyModel::I_chroma]);
  stored_ztree_luma_P[0]->copy_coding_state(param.ztree[EntropyModel::P_luma]);
  stored_ztree_chroma_P->copy_coding_state(param.ztree[EntropyModel::P_chroma]);  
  stored_ztree_mv_P->copy_coding_state(param.ztree[4]);
  stored_ztree_mv_B->copy_coding_state(param.ztree[5]);
  
  int tmp_offset8x8[4][4][TX_NCOEFF8];
  int tmp_offset[4][4][TX_NCOEFF]; 

  // Save offsets
  for(i0=0;i0<4;i0++)
  {
    for(j0=0;j0<4;j0++)
    {
      if (param.trans8x8_enable) 
      {
        for(c=0;c<TX_NCOEFF8;c++)
        {
          tmp_offset8x8[i0][j0][c] = quant[i0]->offset8x8[j0][c];
        }
      }
      for(c=0;c<TX_NCOEFF;c++)
      {
        tmp_offset[i0][j0][c] = quant[i0]->offset[j0][c];
      }
    }
  }
  
  int rate = stored_eenc[0]->get_nbits();

  // Compute cost for all 8x8 blocks
  blkIdx = 0;
  for (i4 = i; i4 < imin(i + 4, m_cur2->height4); i4 += 2)
  {
    for (j4 = j; j4 < imin(j + 4, m_cur2->width4); j4 += 2, blkIdx++)
    {
      assert(blkIdx < 4);
      codemvsbi_blk(m_me[2], mlc, i4>>1, j4>>1, false, stored_eenc[0]);
      
      pblk = mf->get_pblk(i4>>1, j4>>1);
      if (pblk->intra)
      {
        m_ienc->intra_pred->avail_up_right_8x8 = (j == j4) || (m_ienc->intra_pred->avail_up_right_16x16 && i == i4);
        m_ienc->encode_luma_block_NxN(i4, j4, stored_eenc[0], NULL, 0, stored_ztree_luma_I, stored_ztree_luma_P_8x8[0]);
        m_ienc->encode_chroma_blocks(i4>>1, j4>>1, stored_eenc[0], NULL, 0, stored_ztree_chroma_I, SIP4x4);
        
        int sip_type = m_ienc->intra_pred->get_sip_type(i4>>1, j4>>1);
        zmap[ContextMap::inter]->copy_from_intra(zmap[ContextMap::intra], i4, j4, sip_type);
        zmap_chroma[0][ContextMap::inter]->set(i4>>1, j4>>1, zmap_chroma[0][ContextMap::intra]->get(i4>>1, j4>>1));
        zmap_chroma[1][ContextMap::inter]->set(i4>>1, j4>>1, zmap_chroma[1][ContextMap::intra]->get(i4>>1, j4>>1));
      }
      else
      {
        m_ienc->intra_pred->set_sip_type(i4>>1, j4>>1, SIP_INTER);  //// set the type
        if (mf->get_transform_blocksize(i4>>1, j4>>1))
        {
          encode_luma_inter_block_8x8(i4, j4, EntropyModel::P_luma, pred_frame, mlc, 1, lambda_rdoq8x8, lambda_mode, trans8x8Idx[blkIdx]);
        }
        else
        {
          for (ii = 0; ii < 2; ii++)
          {
            for (jj = 0; jj < 2; jj++)
            {
              encode_luma_inter_block(i4, j4, ii, jj, EntropyModel::P_luma, mlc, 1, lambda_rdoq, lambda_mode, trans4x4Idx[blkIdx][2*ii+jj]);
            }
          }
        }
        encode_chroma_inter_blocks(i4>>1, j4>>1, EntropyModel::P_chroma, 1, lambda_chroma);

        zmap[ContextMap::intra]->copy_from_inter(zmap[ContextMap::inter], i4, j4);
        zmap_chroma[0][ContextMap::intra]->set(i4>>1, j4>>1, zmap_chroma[0][ContextMap::inter]->get(i4>>1, j4>>1));
        zmap_chroma[1][ContextMap::intra]->set(i4>>1, j4>>1, zmap_chroma[1][ContextMap::inter]->get(i4>>1, j4>>1));        
      }
    }
  }
  
  // Consider cost of mvs for neighboring blocks
  pblk = mf->get_pblk((i>>1), (j>>1)+2);
  if (pblk)
    codemvsbi_blk(m_me[2], mlc, (i>>1), (j>>1)+2, false, stored_eenc[0]);

  pblk = mf->get_pblk((i>>1)+1, (j>>1)+2);
  if (pblk)
    codemvsbi_blk(m_me[2], mlc, (i>>1)+1, (j>>1)+2, false, stored_eenc[0]);

  pblk = mf->get_pblk((i>>1)+2, (j>>1));
  if (pblk)
    codemvsbi_blk(m_me[2], mlc, (i>>1)+2, (j>>1), false, stored_eenc[0]);

  pblk = mf->get_pblk((i>>1)+2, (j>>1)+1);
  if (pblk)
    codemvsbi_blk(m_me[2], mlc, (i>>1)+2, (j>>1)+1, false, stored_eenc[0]);

  stored_eenc[0]->terminate();
  rate = stored_eenc[0]->get_nbits() - rate;
  
  Block blkOriginal(16, 16);
  Block blkReconstructed(16, 16);
  m_target->get_blk_luma(&blkOriginal, i>>4, j>>4);
  m_cur2->get_blk_luma(&blkReconstructed, i>>4, j>>4);
  double distortion = blkReconstructed.ssd(blkOriginal);
  blkOriginal.set_size(8, 8);
  for (c = 0; c < 2; c++)
  {
    m_target->get_blk_chroma(&blkOriginal, i>>4, j>>4, c);
    m_cur2->get_blk_chroma(&blkReconstructed, i>>4, j>>4, c);
    if (param.lambda_mode_chroma)
      distortion += blkReconstructed.ssd(blkOriginal) * lambda_mode / lambda_mode_chroma;
    else
      distortion += blkReconstructed.ssd(blkOriginal);
  }
  double cost8 = lambda_mode * rate + distortion;

  // Save motion field
  PredBlk myMotion[2][2];
  for (ii = 0; ii < 2; ii++)
  {
    for (jj = 0; jj < 2; jj++)
    {
      PredBlk *pblk = mf->get_pblk((i>>1)+ii, (j>>1)+jj);
      if (pblk)
        myMotion[ii][jj] = *pblk;
    }
  }
  
  // Compute cost for intra 16x16
  
  double cost16 = DBL_MAX;
  Block coeff16(SIP16x16);
  int mode16 = 0;
  if (i + 4 <= cur->height4 && j + 4 <= cur->width4)
  {
    stored_eenc[0]->copy_coding_state(eenc);
    rate = stored_eenc[0]->get_nbits();
    
    mf->set_block_intra((i>>1), (j>>1));
    mf->set_block_intra((i>>1), (j>>1)+1);
    mf->set_block_intra((i>>1)+1, (j>>1));
    mf->set_block_intra((i>>1)+1, (j>>1)+1);
    codemvsbi_blk(m_me[2], mlc, i>>1, j>>1, false, stored_eenc[0]);

    stored_eenc[0]->entropycode_sip_type(SIP16x16, m_ienc->intra_pred->get_sip_ctx(i>>1, j>>1));
    
    int subtree_i;
    for (subtree_i=0; subtree_i<NSUBZTREES; subtree_i++)     
      m_ienc->rd_subztrees_luma[subtree_i]->copy_coding_state(param.ztree[NZTREES-NSUBZTREES+subtree_i]);

    m_ienc->get_rdcost_luma_blk16x16(&coeff16, i>>1, j>>1, stored_eenc[0]); // return bogus R-D cost
    mode16 = m_ienc->intra_pred->get_mode_y(i>>1, j>>1, 0);

    stored_ztree_luma_P_8x8[0]->copy_coding_state(param.ztree_64_CT8); // Use tmp luma ztree as tmp chroma ztree
    m_ienc->encode_chroma_blocks(i>>2, j>>2, stored_eenc[0], NULL, 0, stored_ztree_luma_P_8x8[0], SIP8x8);

    // Consider cost of mvs for neighboring blocks
    pblk = mf->get_pblk((i>>1), (j>>1)+2);
    if (pblk)
      codemvsbi_blk(m_me[2], mlc, (i>>1), (j>>1)+2, false, stored_eenc[0]);

    pblk = mf->get_pblk((i>>1)+1, (j>>1)+2);
    if (pblk)
      codemvsbi_blk(m_me[2], mlc, (i>>1)+1, (j>>1)+2, false, stored_eenc[0]);

    pblk = mf->get_pblk((i>>1)+2, (j>>1));
    if (pblk)
      codemvsbi_blk(m_me[2], mlc, (i>>1)+2, (j>>1), false, stored_eenc[0]);

    pblk = mf->get_pblk((i>>1)+2, (j>>1)+1);
    if (pblk)
      codemvsbi_blk(m_me[2], mlc, (i>>1)+2, (j>>1)+1, false, stored_eenc[0]);

    stored_eenc[0]->terminate();
    rate = stored_eenc[0]->get_nbits() - rate;

    Block blkOriginal(16, 16);
    Block blkReconstructed(16, 16);
    m_target->get_blk_luma(&blkOriginal, i>>4, j>>4);
    m_cur2->get_blk_luma(&blkReconstructed, i>>4, j>>4);
    double distortion = blkReconstructed.ssd(blkOriginal);
    blkOriginal.set_size(8, 8);
    for (c = 0; c < 2; c++)
    {
      m_target->get_blk_chroma(&blkOriginal, i>>4, j>>4, c);
      m_cur2->get_blk_chroma(&blkReconstructed, i>>4, j>>4, c);
      if (param.lambda_mode_chroma)
        distortion += blkReconstructed.ssd(blkOriginal) * lambda_mode / lambda_mode_chroma;
      else
        distortion += blkReconstructed.ssd(blkOriginal);
    }
    
    cost16 = lambda_mode * rate + distortion;
  }

  // Restore offsets
  for (i0=0;i0<4;i0++)
  {
    for (j0=0;j0<4;j0++)
    {
      if (param.trans8x8_enable) 
      {
        for(c=0;c<TX_NCOEFF8;c++)
        {
          quant[i0]->offset8x8[j0][c] = tmp_offset8x8[i0][j0][c];
        }
      }
      for(c=0;c<TX_NCOEFF;c++)
      {
        quant[i0]->offset[j0][c] = tmp_offset[i0][j0][c];
      }
    }
  }
  
  // Encode all 8x8 blocks
  // or encoder intra 16x16 block
  
  if (cost8 < cost16)
  {
    // Restore motion field
    for (ii = 0; ii < 2; ii++)
    {
      for (jj = 0; jj < 2; jj++)
      {
        pblk = mf->get_pblk((i>>1)+ii, (j>>1)+jj);
        if (pblk)
          *pblk = myMotion[ii][jj];
      }
    }
    
    int rate2 = eenc->get_nbits();
    blkIdx = 0;
    for (i4 = i; i4 < imin(i + 4, m_cur2->height4); i4 += 2)
    {
      for (j4 = j; j4 < imin(j + 4, m_cur2->width4); j4 += 2, blkIdx++)
      {      
        assert(blkIdx < 4);
        mlc->motion_bits -= eenc->get_nbits();
        codemvsbi_blk(m_me[2], mlc, i4>>1, j4>>1, true, eenc);
        mlc->motion_bits += eenc->get_nbits();
        
        pblk = mf->get_pblk(i4>>1, j4>>1);
        if (pblk->intra)
        {
          m_ienc->intra_pred->avail_up_right_8x8 = (j == j4) || (m_ienc->intra_pred->avail_up_right_16x16 && i == i4);
          mlc->luma_bits -= eenc->get_nbits();
          m_ienc->encode_luma_block_NxN(i4, j4, eenc, mlc, 1, NULL, NULL);
          mlc->luma_bits += eenc->get_nbits();

          mlc->chroma_bits -= eenc->get_nbits();
          m_ienc->encode_chroma_blocks(i4>>1, j4>>1, eenc, mlc, 1, NULL, SIP4x4);
          mlc->chroma_bits += eenc->get_nbits();

          int sip_type = m_ienc->intra_pred->get_sip_type(i4>>1, j4>>1);
          zmap[ContextMap::inter]->copy_from_intra(zmap[ContextMap::intra], i4, j4, sip_type);
          zmap_chroma[0][ContextMap::inter]->set(i4>>1, j4>>1, zmap_chroma[0][ContextMap::intra]->get(i4>>1, j4>>1));
          zmap_chroma[1][ContextMap::inter]->set(i4>>1, j4>>1, zmap_chroma[1][ContextMap::intra]->get(i4>>1, j4>>1));
          mlc->iblks += 4;
        }
        else
        {
          m_ienc->intra_pred->set_sip_type(i4>>1, j4>>1, SIP_INTER);  //// set the type
          mlc->luma_bits -= eenc->get_nbits();
          if (mf->get_transform_blocksize(i4>>1, j4>>1))
          {
            encode_luma_inter_block_8x8(i4, j4, EntropyModel::P_luma, pred_frame, mlc, 0, lambda_rdoq8x8, lambda_mode, trans8x8Idx[blkIdx]);
          }
          else
          {
            for (ii = 0; ii < 2; ii++)
            {
              for (jj = 0; jj < 2; jj++)
              {
                encode_luma_inter_block(i4, j4, ii, jj, EntropyModel::P_luma, mlc, 0, lambda_rdoq, lambda_mode, trans4x4Idx[blkIdx][2*ii+jj]);
              }
            }
          }
          mlc->luma_bits += eenc->get_nbits();

          mlc->chroma_bits -= eenc->get_nbits();
          encode_chroma_inter_blocks(i4>>1, j4>>1, EntropyModel::P_chroma, 0, lambda_chroma);
          mlc->chroma_bits += eenc->get_nbits();

          zmap[ContextMap::intra]->copy_from_inter(zmap[ContextMap::inter], i4, j4);
          zmap_chroma[0][ContextMap::intra]->set(i4>>1, j4>>1, zmap_chroma[0][ContextMap::inter]->get(i4>>1, j4>>1));
          zmap_chroma[1][ContextMap::intra]->set(i4>>1, j4>>1, zmap_chroma[1][ContextMap::inter]->get(i4>>1, j4>>1));
        }
      }
    }
    rate2 = eenc->get_nbits() - rate2;
    if (rate2 != rate && 0)
    {
      printf("%d %d %d ", rate, rate2, eenc->get_nbits());
      printf("%d ", mf->get_pblk(i4>>1, j4>>1) != NULL);
      printf("%d ", mf->get_pblk((i4>>1), (j4>>1)+1) != NULL);
      printf("%d ", mf->get_pblk((i4>>1)+1, j4>>1) != NULL);
      printf("%d ", mf->get_pblk((i4>>1)+1, (j4>>1)+1) != NULL);
      printf("\n");
    }
  }
  else
  {
    for (ii = 0; ii < 2; ii++)
    {
      for (jj = 0; jj < 2; jj++)
      {
        mf->set_block_intra((i>>1)+ii, (j>>1)+jj);
        mf->get_pblk((i>>1)+ii, (j>>1)+jj)->list = -1;
      }
    }
    
    mlc->motion_bits -= eenc->get_nbits();
    codemvsbi_blk(m_me[2], mlc, i>>1, j>>1, true, eenc);
    mlc->motion_bits += eenc->get_nbits();
    
    m_ienc->intra_pred->set_sip_type(i>>1, j>>1, SIP16x16, 2, 2);
    m_ienc->intra_pred->put_mode_y_MxN(i>>1, j>>1, mode16, 2, 2);
    m_ienc->intra_pred->set_L_mode_pred_luma(0, SIP16x16);
    
    mlc->luma_bits -= eenc->get_nbits();
    m_ienc->entropycode_luma_blk16x16(&coeff16, i>>1, j>>1, eenc, 1, NULL, SIP16x16);
    mlc->luma_bits += eenc->get_nbits();
    
    mlc->chroma_bits -= eenc->get_nbits();
    m_ienc->encode_chroma_blocks(i>>2, j>>2, eenc, mlc, 1, NULL, SIP8x8);
    mlc->chroma_bits += eenc->get_nbits();

    zmap[ContextMap::inter]->copy_from_intra(zmap[ContextMap::intra], i, j, SIP16x16);
    zmap_chroma[0][ContextMap::inter]->copy_from_intra(zmap_chroma[0][ContextMap::intra], i>>1, j>>1, SIP8x8);
    zmap_chroma[1][ContextMap::inter]->copy_from_intra(zmap_chroma[1][ContextMap::intra], i>>1, j>>1, SIP8x8);
    mlc->iblks += 16;
  }
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Store buffer and MLC
*
* Keep the best encoding based on R-D cost
*
* \param[in] i flag indicating whether frame_type is intra
*/
//-----------------------------------------------------------------------------
void Encoder::keep_best_buf_frm_mlc(int i)
{
  Buffer *buf_swap;
  Frame *frm_swap;
  Frame *frm_swap_filtered;
  Frame *frm_swap_pred;
  Frame *frm_swap_qpe_m;
  Frame *frm_swap_qpe_s;
  MLC_Stat *MLC_swap;

  buf_swap = buf_best[i];
  frm_swap = frm_best[i];
  buf_best[i] = buf_tmp;
  frm_best[i] = frm_tmp;
  buf_tmp = buf_swap;
  frm_tmp = frm_swap;
#ifdef MVCOMP_ZERORES
  // We do a copy for the moment, rather than swapping.  We can probably swap.
  mf_best[i]->copy_field_from(mf);
  // Copy the inter_pel fields in effect, to.
  memcpy(inter_pel_best[i], ref_mgr->ref_process->inter_pel, sizeof(ref_mgr->ref_process->inter_pel));
  memcpy(log_inter_pel_best[i], ref_mgr->ref_process->log_inter_pel, sizeof(ref_mgr->ref_process->log_inter_pel));

  // Copy the reference list info.
  memcpy(ref_list_best[i], ref_mgr->list, sizeof(ref_mgr->list));
  for (int r = ZERO; r < MAX_REF_FRAMES; r++)
    frame_no_best[i][r] = ref_mgr->ref[r]->empty ? -1 : ref_mgr->ref[r]->frame_no;
#endif
#ifdef TRACE_DEBUG
  SAVE_ETRACEBEST;
#endif
  frm_swap_filtered = frm_best_filtered[i];
  frm_swap_pred     = frm_best_pred    [i];
  frm_swap_qpe_m    = frm_best_qpe_m   [i];
  frm_swap_qpe_s    = frm_best_qpe_s   [i];
  frm_best_filtered[i] = frm_tmp_filtered;
  frm_best_pred    [i] = frm_tmp_pred;
  frm_best_qpe_m   [i] = frm_tmp_qpe_m;
  frm_best_qpe_s   [i] = frm_tmp_qpe_s;
  frm_tmp_filtered = frm_swap_filtered;
  frm_tmp_pred     = frm_swap_pred;
  frm_tmp_qpe_m    = frm_swap_qpe_m;
  frm_tmp_qpe_s    = frm_swap_qpe_s;
  (eenc_best[i])->copy_for_wlf(eenc);
  (eenc_best[i])->set_buffer(buf_best[i]);
  low_zeromv_rdoq_best[i]=wlf->store_low_zeromv_rdoq;

  MLC_swap = best_MLC[i];
  best_MLC[i] = tmp_MLC;
  tmp_MLC = MLC_swap;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute distortion against original
*
* Computes SSE, MSE and PSNR.
*
* \param[in] f Reference frame
* \return R-D cost
*/
//-----------------------------------------------------------------------------
double Encoder::compute_distortion_filtered_against_orig(Frame *cur, Frame *cur_orig, Frame *cur_filtered)
{
  int i;
  int d;
  int c;
  int sum;

  sum = ZERO;

  for (i = ZERO; i < cur->width * cur->height; i++)
  {
    d = cur_filtered->luma[i] - cur_orig->luma[i];
    sum += d * d;
  }

  cur->sse_luma = sum;
  cur->mse_luma = 1.0 * sum / (cur->width * cur->height);
  cur->snr_luma = psnr(cur->mse_luma);

  for (c=0; c<2; c++) 
  {
    sum = ZERO;

    for (i=0; i<cur->width2*cur->height2; i++) 
    {
      d = cur_filtered->crcb[c][i] - cur_orig->crcb[c][i];
      sum += d * d;
    }

    cur->sse_chroma[c] = sum;
    cur->mse_chroma[c] = 1.0 * sum / (cur->width2 * cur->height2);
    cur->snr_chroma[c] = psnr(cur->mse_chroma[c]);
  }

  return cur->sse_luma * cur->lambda + (cur->sse_chroma[0] + cur->sse_chroma[1]) * cur->lambdaC;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
