/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file EntropyDecoder.cpp
* \brief Implementation of EntropyDecoder class
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "EntropyDecoder.H"
#include "Param.H"
#include "Global.H"
#include "WienerFilter.H"
#include "IntraPrediction.H"

#ifdef TRACE_DEBUG
#include "Trace.H"
#endif

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] share Non-null if this decoder is sharing its synchrostate with other decoders.
*/
//-----------------------------------------------------------------------------
#ifdef CABAC_SYNC_POINTS
EntropyDecoder::EntropyDecoder(SynchroState *share)
#else
EntropyDecoder::EntropyDecoder()
#endif
{
#ifdef DEC_TRANSCODER
  if (param.decode_transcode)
    adec = new ArithTranscoder();
  else
    adec = new ArithDecoder;
#else
  adec = new ArithDecoder;
#endif
#ifdef CABAC_SYNC_POINTS
  cs = new EntropyModel(share);
#else
  cs = new EntropyModel;
#endif
  
  cs->reset_counts();
  //{{FRAME_SUBSTREAMS
  // Moved from param.
  int ztree_size[NZTREES64] = 
  {16, 16, 16, 16
  };
  for (int i=0; i<NZTREES64; i++)
  {   
    int add_node;
    add_node = (i<NZTREES64-1);
    zt64[i].nleaves = ztree_size[i];
    zt64[i].s_nleaves = ztree_size[i]+add_node;
    zt64[i].iSoN = (i==0)?0:(zt64[i-1].iSoN+zt64[i-1].nleaves);
    zt64[i].stats = new int[power2(zt64[i].s_nleaves)];
    for(int j = 0;j<zt64[i].nleaves;j++)
    {
      zt_idx[j+zt64[i].iSoN][0] = i;
      zt_idx[j+zt64[i].iSoN][1] = j;
    }
  }

  for(int j=0; j<NZTREES64; j++)
    for (int i=0; i<power2(zt64[j].s_nleaves); i++)
    {
      zt64[j].stats[i] =0;
    }
    //FRAME_SUBSTREAMS}}
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*/
//-----------------------------------------------------------------------------
EntropyDecoder::~EntropyDecoder()
{
  delete cs;
  delete adec;
  
  //{{FRAME_SUBSTREAMS
  // Moved from param.
  if (param.trans8x8_enable) 
  {
    for (int i=0; i<NZTREES64; i++)
    {
      delete[] zt64[i].stats;
    }
  }
  //FRAME_SUBSTREAMS}}
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Print statistics
*
* \param[in,out] f File pointer
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::print_stats(FILE *f)
{
  if (f != NULL)
    cs->print_stat(f);
}

#ifdef VERBOSE_STATS
//-----------------------------------------------------------------------------
/*!
* \brief Print summarized statistics
*
* \param[in,out] f File pointer
*/
//-----------------------------------------------------------------------------
#ifdef FRAME_SUBSTREAMS
void EntropyDecoder::print_stats_summary(FILE *f, int verbose, int buffer_payload_bits, int buffer_read_bits)
#else
void EntropyDecoder::print_stats_summary(FILE *f, int verbose)
#endif
{
  if (f != NULL)
    if (verbose) 
    {
      cs->print_by_category_full(f);
#ifdef FRAME_SUBSTREAMS
      fprintf(f, "Buffer bits (read)%9d\n\n\n", buffer_read_bits);
      fprintf(f, "Buffer bits (orig)%9d\n", buffer_payload_bits);
#else
      fprintf(f, "Buffer bits (read)%9d\n",  adec->get_nbits());
      fprintf(f, "Buffer bits (orig)%9d\n\n\n", adec->buf->payload_size*8);
#endif
    }
    else
      cs->print_by_category();
}
#else
//-----------------------------------------------------------------------------
/*!
* \brief Print summarized statistics
*
* \param[in,out] f File pointer
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::print_stats_summary(FILE *f)
{
  if (f != NULL)
    cs->print_by_category();
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Reset
*
* Reset statistics and initialize arithmetic decoder.
*
* \param[in] b Byte buffer containing coded data
*/
//-----------------------------------------------------------------------------
#ifdef DEC_TRANSCODER
void EntropyDecoder::reset(Buffer *b, Buffer *b_out)
#else
void EntropyDecoder::reset(Buffer *b)
#endif
{
  // reset probabilities to initial values
  cs->reset(); 
  cs->reset_counts();
  
  // initialize arithmetic decoder
#ifdef DEC_TRANSCODER
  adec->init(b, b_out);
#else
  adec->init(b);
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode frame type
*
* \return %Frame type (ft_I_frame, ft_P_frame, or ft_B_frame)
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_frame_type()
{
  int frame_type = adec->decode_symbol_half(&cs->p_frame_type);
  
  if (frame_type != ft_I_frame)
    frame_type += 2 * adec->decode_symbol_half(&cs->p_frame_type);
  
  return frame_type;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decodes INTRA frame type
*
* Used to decode INTRA frame type when frame_type == ft_I_frame.
*
* \return %Frame type (ft_I_noRefresh {0}, ft_I_IDR {10}, or ft_I_DDR {11})
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_I_frame_type ()
{
  int I_frame_type = adec->decode_symbol_half(&cs->p_frame_type);  // note that this used p_frame_type, p_I_frame_type is not defined at the moment.
  
  if (I_frame_type!=ft_I_noRefresh) // not zero.
  {
    I_frame_type += 2 * adec->decode_symbol_half(&cs->p_frame_type);  // note that this used p_frame_type, p_I_frame_type is not defined at the moment.
  }
  return I_frame_type;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode reference frame indicator
*
* \return Flag indicating whether frame is a reference frame
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_reference_idc()
{ 
  return adec->decode_symbol_half(&cs->p_reference_idc);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode edge prediction indicator
*
* \return Flag indicating whether edge prediction for intra is used or not
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_edge_prediction_flag()
{ 
  return adec->decode_symbol_half(&cs->p_edge_prediction_flag);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode threshold for edge prediction
*
* \param[in] val Threshold value
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_threshold_edge()
{
  int i;
  int val=0;

  for(i=0;i<16;i++)
    val = (val<<1)+adec->decode_symbol_half(&cs->p_threshold_edge);

  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode frame counter
*
* \return %Frame counter
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_frame_counter()
{
  int v;
  
  // This is currently needed for first frame only. 
  v = adec->decode_symbol_half(&cs->p_frm_zero);  
  
  if (v) 
  {
    while (adec->decode_symbol_half(&cs->p_framecount_sign))
      v++;
    
    if (adec->decode_symbol_half(&cs->p_framecount_sign) == NEGATIVE)
      v = -v;
  }
  
  return v;  
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode delta QP
*
* \return Delta QP
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_deltaqp()
{
  int v;
  
  v = adec->decode_symbol_half(&cs->p_dqp_zero);
  
  if (v) 
  {
    while (adec->decode_symbol_half(&cs->p_deltaqp))
      v++;
    
    if (adec->decode_symbol_half(&cs->p_dqp_sign) == NEGATIVE)
      v = -v;
  }
  
  return v;  
}

#ifdef Q_OFFSET
//-----------------------------------------------------------------------------
/*!
* \brief Decode q_offset for 4x4 intra
*
* \return Delta q_offset
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_q_offset_0()
{
  int v;
  
  v = adec->decode_symbol_half(&cs->p_q_offset_zero[0]);
  
  if (v) 
  {
    while (adec->decode_symbol_half(&cs->p_q_offset[0]))
      v++;
    
    if (adec->decode_symbol_half(&cs->p_q_offset_sign[0]) == NEGATIVE)
      v = -v;
  }
  
  return v;  
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode q_offset for 4x4 inter
*
* \return Delta q_offset
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_q_offset_1()
{
  int v;
  
  v = adec->decode_symbol_half(&cs->p_q_offset_zero[1]);
  
  if (v) 
  {
    while (adec->decode_symbol_half(&cs->p_q_offset[1]))
      v++;
    
    if (adec->decode_symbol_half(&cs->p_q_offset_sign[1]) == NEGATIVE)
      v = -v;
  }
  
  return v;  
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode q_offset for 8x8 intra  // RESERVED
*
* \return Delta q_offset
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_q_offset8x8_0()
{
  int v;
  
  v = adec->decode_symbol_half(&cs->p_q_offset8x8_zero[0]);
  
  if (v) 
  {
    while (adec->decode_symbol_half(&cs->p_q_offset8x8[0]))
      v++;
    
    if (adec->decode_symbol_half(&cs->p_q_offset8x8_sign[0]) == NEGATIVE)
      v = -v;
  }
  
  return v;  
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode q_offset for 8x8 inter
*
* \return Delta q_offset
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_q_offset8x8_1()
{
  int v;
  
  v = adec->decode_symbol_half(&cs->p_q_offset8x8_zero[1]);
  
  if (v) 
  {
    while (adec->decode_symbol_half(&cs->p_q_offset8x8[1]))
      v++;
    
    if (adec->decode_symbol_half(&cs->p_q_offset8x8_sign[1]) == NEGATIVE)
      v = -v;
  }
  
  return v;  
}

#endif

//-----------------------------------------------------------------------------
/*!
* \brief Decode the value of quantization set id, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_quantization_set_id_half() // cq_start
{
  int v;

  v = adec->decode_symbol_half(&cs->quantization_set_id_value_zero);
  if (v) 
  {
    while (adec->decode_symbol_half(&cs->quantization_set_id_value))
      v++;
  }
  param.quantization_set_id = v;
  assert(param.quantization_set_id <= param.max_quantization_set_num_minus_one);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode values of a q matrix, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::read_qmatrix_list_half(int *qmatrixList, int sizeOfQMatrixList, int *useDefaultQMatrixFlag, int matrix_type)
{
  int j;
  int v;
  int delta_scale;
  int lastScale = 8;
  int nextScale = 8;

  for (j = 0; j < sizeOfQMatrixList; j++)
  {
    if (nextScale != 0)
    {
      v = adec->decode_symbol_half(&cs->qmatrix_value_zero[matrix_type]);
      if (v) 
      {
        while (adec->decode_symbol_half(&cs->qmatrix_value[matrix_type]))
          v++;
        if (adec->decode_symbol_half(&cs->qmatrix_value_sign[matrix_type]) == NEGATIVE)
          v = -v;
      }
      delta_scale = v;
      nextScale = ( lastScale + delta_scale + 256 ) % 256;
      *useDefaultQMatrixFlag = ((j==0) && (nextScale==0));
    }
    qmatrixList[j] = (nextScale == 0) ? lastScale : nextScale;
    lastScale = qmatrixList[j];
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode update values of a q matrix, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::read_qmatrix_update_value_half(int *qmatrixList, int matrix_type)
{
  int i;
  int v;

  //! scale ratio, slope ratio, number of delta values, matrix values
  for(i = 1; i < 260; i++)
  {
    if( (i >= 4) && ((i - 4) >= qmatrixList[3]) ) 
      break;
    v = adec->decode_symbol_half(&cs->qmatrix_value_zero[matrix_type]);
    if (v) 
    {
      while (adec->decode_symbol_half(&cs->qmatrix_value[matrix_type]))
        v++;
      if (i != 3 && adec->decode_symbol_half(&cs->qmatrix_value_sign[matrix_type]) == NEGATIVE)
        v = -v;
    }
    qmatrixList[i] = v;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode values of a q matrix set, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::read_qmatrix_set_half(int quantization_set_id, int matrix_type)
{
  int j;
  int k;
  int v;
  int qmatrix_list_present_flag;
  int *qmatrixList;
  int qmatrix_modetype;
  int matrix_size;

  for (j = 0; j < 16; j++)
  {
    if (j<8) //!< 4x4
      matrix_size = 16;
    else if (j<14) //!< 8x8
      matrix_size = 64;
    else //!< 16x16
      matrix_size = 256;
        
    qmatrix_list_present_flag = adec->decode_symbol_half(&cs->qmatrix_list_present_flag[matrix_type]);
    param.qmatrix_update_check[matrix_type][quantization_set_id][j] = qmatrix_list_present_flag;
    qmatrixList = param.qmatrix_tmp[matrix_type][quantization_set_id][j];
    qmatrix_modetype = 1;
    if (qmatrix_list_present_flag)
    {
      //! mode type
      v = adec->decode_symbol_half(&cs->qmatrix_value_zero[matrix_type]);
      if (v) 
      {
        while (adec->decode_symbol_half(&cs->qmatrix_value[matrix_type]))
          v++;
      }
      qmatrixList[0] = v;
      qmatrix_modetype = qmatrixList[0];

      if (qmatrix_modetype == 0) //!< update mode
      {
        read_qmatrix_update_value_half(qmatrixList, matrix_type);
        param.matrix_inv_scan(qmatrixList+4, j, matrix_size);
      }
      else if (qmatrix_modetype == 2 && matrix_type == 1) //!< flat mode
      {
        v = adec->decode_symbol_half(&cs->qmatrix_value_zero[matrix_type]);
        if (v) 
        {
          while (adec->decode_symbol_half(&cs->qmatrix_value[matrix_type]))
            v++;
          if (adec->decode_symbol_half(&cs->qmatrix_value_sign[matrix_type]) == NEGATIVE)
            v = -v;
        }
        qmatrixList[4] = v;
      }
    }
    if (qmatrix_modetype == 1) //!< matrix mode
    {
      qmatrix_list_present_flag = adec->decode_symbol_half(&cs->qmatrix_list_present_flag[matrix_type]);
      param.qmatrix_check[matrix_type][quantization_set_id][j] = qmatrix_list_present_flag;
      if (qmatrix_list_present_flag)
      {
        read_qmatrix_list_half(qmatrixList+4, matrix_size, &(param.UseDefaultQMatrixFlag[matrix_type][j]), matrix_type);
        param.matrix_inv_scan(qmatrixList+4, j, matrix_size);
        if (param.UseDefaultQMatrixFlag[matrix_type][j])
        {
          for (k = 0; k < matrix_size; k++)
            qmatrixList[k+4] = 16;
        }
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode values of an updated q matrix set, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_qmatrix_half(int matrix_type) // cq_end
{
  int qmatrix_overwrite_flag;

  qmatrix_overwrite_flag = adec->decode_symbol_half(&cs->qmatrix_overwrite_flag[matrix_type]);
  if (qmatrix_overwrite_flag)
    read_qmatrix_set_half(param.quantization_set_id, matrix_type);
  param.update_qmatrix(param.quantization_set_id, matrix_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode number of references used in lists
*
* \return Value indicating whether one or more references may be in list
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_number_of_references()
{
  return adec->decode_symbol_half(&cs->p_num_of_refs);
}

#ifdef MVCOMP_ZERORES
//-----------------------------------------------------------------------------
/*!
* \brief Decode mvcomp configuration choice.
*
* \return Value indicating whether we use config #0 or #1.
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_mvcomp_config()
{
  return adec->decode_symbol_half(&cs->p_mvcomp_config);
}
#endif


//-----------------------------------------------------------------------------
/*!
* \brief Decode zero block flag
*
* \param[in] past Context derived from past decodes of this flag
* \param[in] chroma Flag indicating whether this is for a chroma block
* \return Flag indicating whether a block has nonzero coefficients
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_zblk(int past, int chroma)
{
#ifdef TRACE_DEBUG
  SVPRINT_DTRACE("zblk ctx ",past);
#endif
  return adec->decode_symbol(&cs->p_zblk[chroma][past]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode motion vector prediction direction
*
* \note Context can be obtained using PredBlk::get_pred_context()
*
* \param[in] ctx Context derived from neighbouring motion vectors
* \return Prediction (horizontal or vertical)
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_mvp(int ctx)
{
#ifdef TRACE_DEBUG
  SVPRINT_DTRACE("mvp ctx ", ctx);
#endif
  return adec->decode_symbol(&cs->p_mvp[ctx]);
}

//-----------------------------------------------------------------------------
/*!
 * \brief Decode fixed-length code
 *
 * \param[in] numbits Length of code
 * \param[in] pe Probabiliy estimator (50/50)
 *
 * \return value for mv resolution
 */
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_FLC(int numbits, FixedProbEst *pe)
{
  int val = 0;
  int i;
  
  for (i = 0; i < numbits; i++)
    val = 2*val + adec->decode_symbol_half(pe);
  
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode interpolation type
*
* \return value for mv resolution
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_inter_type()
{
  return decode_FLC(LOG_MAX_INTER_TYPE, &cs->p_inter_type);
}


//-----------------------------------------------------------------------------
/*!
* \brief Decode rpps type
*
* \return Flag indicating whether following data is RPPS or not
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_rpps_type()
{
  return decode_FLC(LOG_MAX_RPPS_TYPE, &cs->p_rpps_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode rpps id
*
* \return RPPS id
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_rpps_id()
{
  return decode_FLC(LOG_MAX_RPPS_ID, &cs->p_rpps_id);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode interpolation filter type
*
* \return value for interpolation filter
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_filter_type()
{
  return decode_FLC(LOG_MAX_FILTER_TYPE, &cs->p_filter_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode intensity compensation disabling flag
*
* \return flag indicating for disabling intensity compensation
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_ic_disable()
{
  return decode_FLC(LOG_MAX_IC_DISABLE, &cs->p_ic_disable);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode reounding type in interpolation process
*
* \return value for rounding type in interpolation process
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_round_type()
{
  return decode_FLC(LOG_MAX_ROUND_TYPE, &cs->p_round_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode Spatial Intra Prediction type
*
* \return Flag indicating the sip type
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_sip_type(int ctx) 
{
#ifdef TRACE_DEBUG
  SVPRINT_DTRACE("sip_type ctx ", ctx);
  RPRINT_DTRACE;
#endif
  if (param.intra16x16)
  {
    if( adec->decode_symbol(&cs->p_sip_type_b1[ctx]) == 0 )
    {
      return 3;
    }
    else
    {
      if( adec->decode_symbol(&cs->p_sip_type_b2[ctx]) == 0 )
      {
        return 5;// INTRA16x16
      }
      else
      {
        if( adec->decode_symbol( &cs->p_sip_type_b3[ctx] ) == 0 )
        {
          return 0;
        }
        else
        {
          if ( adec->decode_symbol( &cs->p_sip_type_b4[ctx] )==0)
          {
            return 1;
          }
          else
          {
            return 2;
          }
        }
      }
    }
  }
  else
  {
    if (adec->decode_symbol(&cs->p_sip_type_b1[ctx]) == 0) 
    {
      return 3;
    }
    else 
    {
      if (adec->decode_symbol(&cs->p_sip_type_b2[ctx]) == 0) 
      {
        return 0;
      }
      else 
      {
        if (adec->decode_symbol(&cs->p_sip_type_b3[ctx]) == 0) 
        {
          return 1;
        }
        else
        {
          return 2;
        }
      }    
    }
  }    
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode Spatial Intra Prediction mode for luma
*
* \param[in] L_mode_pred_luma the value of the L-Mode predictor
* \return value indicating the luma intra prediction mode
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_spatial_intra_prediction_luma(int L_mode_pred_luma)
{
  int min_mode;
  int prev_intra_pred_mode_flag;
  int rem_intra_pred_mode;
  
  min_mode = L_mode_pred_luma;
#ifdef TRACE_DEBUG
    SPRINT_DTRACE("MPM");
#endif 
  prev_intra_pred_mode_flag = this->adec->decode_symbol(&this->cs->p_prev_intra_pred_mode_flag);

  if (prev_intra_pred_mode_flag == 1) 
  {
    return (min_mode);
  }
#ifdef TRACE_DEBUG
    SPRINT_DTRACE("rem_intra_pred");
    RPRINT_DTRACE;
#endif
  if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[0]) == 0) 
  {
    if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[1]) == 0) 
    {
      if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[2]) == 0) 
      {
        rem_intra_pred_mode = 0;  // Bin string : 000
      } 
      else 
      {
        rem_intra_pred_mode = 1;  // Bin string : 001
      }
    } 
    else 
    {
      if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[2]) == 0) 
      {
        rem_intra_pred_mode = 2;  // Bin string : 010
      } 
      else 
      {
        rem_intra_pred_mode = 3;  // Bin string : 011
      }
    }
  } 
  else 
  {
    if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[1]) == 0) 
    {
      if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[2]) == 0) 
      {
        rem_intra_pred_mode = 4;  // Bin string : 100
      } 
      else 
      {
        rem_intra_pred_mode = 5;  // Bin string : 101
      }
    } 
    else 
    {
      if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[2]) == 0) 
      {
        rem_intra_pred_mode = 6;  // Bin string : 110
      } 
      else 
      {
        if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[3]) == 0) 
        {
          rem_intra_pred_mode = 7;  // Bin string : 1110
        } 
        else 
        {
          if (this->adec->decode_symbol(&this->cs->p_rem_intra_pred_mode_flag[4]) == 0) 
          {
            rem_intra_pred_mode = 8;  // Bin string : 11110
          } 
          else 
          {
            rem_intra_pred_mode = 9;  // Bin string : 11111
          }
        }
      }
    }
  }

  return rem_intra_pred_mode + (rem_intra_pred_mode >= min_mode);
}



//-----------------------------------------------------------------------------
/*!
* \brief Decode Spatial Intra Prediction mode for chroma
*
* \param[in] up_mode mode of block above
* \param[in] left_mode mode of block to the left
* \param[in] up_is_inter flag indicating if the above block is inter
* \param[in] left_is_inter flag indicating if the left block is inter
* \return value indicating the chroma intra prediction mode
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_spatial_intra_prediction_chroma(int up_mode, int left_mode, int up_is_inter, int left_is_inter)
{
  int sip_mode;
  
  if (up_mode == INT_INF || up_is_inter) 
  {
    up_mode = 0;
  }
  if (left_mode == INT_INF || left_is_inter) 
  {
    left_mode = 0;
  }
  
  int flag_count = (up_mode != 0) + (left_mode != 0);

#ifdef TRACE_DEBUG
  SVPRINT_DTRACE("intra_pred ctx", flag_count);
  RPRINT_DTRACE;
#endif
  if (this->adec->decode_symbol(&this->cs->p_intra_chroma_pred_mode[flag_count]) == 0) 
  {
    sip_mode = 0;     // Bin string : 0
  } 
  else 
  {
    if (this->adec->decode_symbol(&this->cs->p_intra_chroma_pred_mode_1) == 0) 
    {
      sip_mode = 1;   // Bin string : 10
    } 
    else 
    {
      if (this->adec->decode_symbol(&this->cs->p_intra_chroma_pred_mode_2) == 0) 
      {
        sip_mode = 2; // Bin string : 110
      }
      else 
      {
        if (this->adec->decode_symbol(&this->cs->p_intra_chroma_pred_mode_2) == 0) 
        {
          sip_mode = 3;  // Bin string : 1110
        }
        else 
        {
          if (this->adec->decode_symbol(&this->cs->p_intra_chroma_pred_mode_2) == 0) 
          {
            sip_mode = 4;  // Bin string : 11110
          } 
          else 
          {
            sip_mode = 5;  // Bin string : 11111
          }
        }
      }
    }
  }
  
  return (sip_mode);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode LSB of reference index
*
* This is used when MSB = 0.
*
* \param[in] ctx Context derived from neighbouring reference indices
* \param[in] list List (LIST_0 or LIST_1)
* \return Value of LSB
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_refindex_lsb0(int ctx, int list)
{
  return adec->decode_symbol(&cs->p_refidx_lsb0[list][ctx]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode LSB of reference index
*
* This is used when MSB = 1.
*
* \param[in] ctx Context derived from neighbouring reference indices
* \param[in] list List (LIST_0 or LIST_1)
* \return Value of LSB
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_refindex_lsb1(int ctx, int list)
{
  return adec->decode_symbol(&cs->p_refidx_lsb1[list][ctx]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode MSB of reference index
*
* \param[in] ctx Context derived from neighbouring reference indices
* \param[in] list List (LIST_0 or LIST_1)
* \return Value of MSB
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_refindex_msb(int ctx, int list)
{
  return adec->decode_symbol(&cs->p_refidx_msb[list][ctx]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode reference index
*
* \param[in] n_ref_frames Size of list for which to decode index
* \param[in] top Motion data from block above
* \param[in] left Motion data from block to the left
* \param[in] list List (LIST_0 or LIST_1)
* \return Reference index
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_refindex(int n_ref_frames, PredBlk *top, PredBlk *left, int list)
{
  int x0, x1;
  int val;
  
  val = ZERO;
  
  if (n_ref_frames < TWO)
    return val;
#ifdef TRACE_DEBUG
  SPRINT_DTRACE("ref index");
  RPRINT_DTRACE;
#endif
  x0 = (left && !left->intra) ? left->ref_index[list] : (top  && !top->intra ) ?  top->ref_index[list] : 0;
  x1 = (top  && !top->intra ) ?  top->ref_index[list] : (left && !left->intra) ? left->ref_index[list] : 0;

  if (n_ref_frames > TWO) 
  {
    val += 2 * decode_refindex_msb((x0>>1)+(x1>>1), list);
    
    if (val < TWO)
      val += decode_refindex_lsb0((x0 &1)+(x1 &1), list);
    else if (n_ref_frames > THREE)
      val += decode_refindex_lsb1((x0 &1)+(x1 &1), list);
  }
  else 
    val += decode_refindex_lsb0(x0+x1, list);
  
  return val;
}


//-----------------------------------------------------------------------------
/*!
* \brief Decode shift value
*
* \param[in] ctx Context derived from neighbouring reference indices
* \param[in] top shift data from block above
* \param[in] left shift data from block the left
* \param[in] type of shift direction
* \return Value of shift
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_mvshare_size(int past, int ref1, int ref2, int type)
{
  int val;
#ifdef TRACE_DEBUG
  SVPRINT_DTRACE("shift value ctx ", past);
  RPRINT_DTRACE;
#endif
  val = adec->decode_symbol(&cs->p_mvshare_type_zero[past]);
  if(val)
  {
    if(type==1)
    {
      if(!param.mvshare_two_enable)
      {
        val=4;
      }
      else
      {
        if(ref1==0)
        {
          if(adec->decode_symbol_half(&cs->p_mvshare_four))
            val=4;
          else
            val=2;
        }
        else
        {
          if(adec->decode_symbol(&cs->p_mvshare_pred)==0)
          {
            val=ref1;
          }
          else
          {
            if(ref1==2) val=4;
            else val=2;
          }
        }
      }
    }
    else if(type==2)
    {
      if(!param.mvshare_two_enable)
      {
        val=4;
      }
      else
      {
        if(ref2==0)
        {
          if(adec->decode_symbol_half(&cs->p_mvshare_four))
            val=4;
          else
            val=2;
        }
        else
        {
          if(adec->decode_symbol(&cs->p_mvshare_pred)==0)
          {
            val=ref2;
          }
          else
          {
            if(ref2==2) val=4;
            else val=2;
          }
        }
      }
    }
    else
    {
      if(adec->decode_symbol_half(&cs->p_mvshare_four))
      {
        if(!param.mvshare_two_enable)
        {
          val=14;
        }
        else
        {
          if(ref2==0)
          {
            if(adec->decode_symbol_half(&cs->p_mvshare_four))
              val=14;
            else
              val=12;
          }
          else
          {
            if(adec->decode_symbol(&cs->p_mvshare_pred)==0)
            {
              val=ref2+10;
            }
            else
            {
              if(ref2==2) val=14;
              else val=12;
            }
          }
        }
      }
      else
      {
        if(!param.mvshare_two_enable)
        {
          val=4;
        }
        else
        {
          if(ref1==0)
          {
            if(adec->decode_symbol_half(&cs->p_mvshare_four))
              val=4;
            else
              val=2;
          }
          else
          {
            if(adec->decode_symbol(&cs->p_mvshare_pred)==0)
            {
              val=ref1;
            }
            else
            {
              if(ref1==2) val=4;
              else val=2;
            }
          }
        }
      }
    }
     
  }
#ifdef TRACE_DEBUG
  RPRINT_DTRACE;
#endif
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode mode for a block
*
* Decodes mode information for a block (intra, list0, list1, or bipred)
*
* \param[in] left Motion data from block to the left
* \param[in] top Motion data from block above
* \param[in] Bframe Flag indicating whether current frame is a B frame
* \return Mode (0=intra, 1=LIST_0, 2=LIST_1, 3=BI_PRED)
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_mode(PredBlk *left, PredBlk *top, int Bframe)
{
  int ctx;
  int mode, l, t;
  
  if (Bframe == TRUE) 
  {
    l = left ? left->mode_B() : MODE_DEF;
    t = top  ?  top->mode_B() : MODE_DEF;
    
    ctx = (l & 1) + (t & 1);
#ifdef TRACE_DEBUG
    SVPRINT_DTRACE("p_mode ctx", ctx);
#endif
    mode = adec->decode_symbol(&cs->p_mode[ctx]);

    ctx = ((l >> 1)&1) + ((t >> 1)&1);
#ifdef TRACE_DEBUG
    SVPRINT_DTRACE("p_mode ctx", ctx);
#endif
    mode |= 2 * adec->decode_symbol(&cs->p_mode[ctx]);
  }
  else 
  {
    l = left ? left->mode_P() : MODE_P_DEF;
    t = top  ?  top->mode_P() : MODE_P_DEF;

    ctx = (l & 1) + (t & 1);
#ifdef TRACE_DEBUG
    SVPRINT_DTRACE("p_mode_P1 ctx", ctx);
#endif
    mode = adec->decode_symbol(&cs->p_mode_P1[ctx]);     
  }
#ifdef TRACE_DEBUG
    SVPRINT_DTRACE("block mode", mode);
    RPRINT_DTRACE;
#endif 
  return mode;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode motion data difference
*
* \param[in,out] pblk Motion data difference of current block
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_mv(PredBlk *pblk)
{
  int i, v;
  int lists;
  ZeroTree *ztree;
  int cur_list;
  int zpatt;
  
#ifdef TRACE_DEBUG
  SPRINT_DTRACE("motion vector");
  RPRINT_DTRACE;
#endif

  v = adec->decode_symbol(&cs->p_mv_zero[0]);
  if (v == TRUE)
  {
    return;
  }
  
  cur_list = pblk->list;
  
  if (cur_list == BI_PRED) 
  {
    ztree = param.ztree[5];
    zpatt = decode_ztree(ztree);
    #ifdef TRACE_STATS
    ztree->stats[zpatt]++;
    #endif
    
    for (lists=0; lists<LISTS; lists++) 
    {
      for (i=0; i<2; i++) 
      {
        if ((zpatt & power2(2*lists+i)) != ZERO)
          pblk->mv[lists][i] = decode_mvd(i);
      }
    }
    
    if ((zpatt & power2(4)) != ZERO)
      pblk->BP_scale[0] = decode_scale(1);
    
    if ((zpatt & power2(5)) != ZERO)
      pblk->BP_scale[1] = decode_scale(2);
    
    if ((zpatt & power2(6)) != ZERO)
      pblk->BP_DC = decode_offset(1);   
  }
  else 
  {
    ztree = param.ztree[4];
    zpatt = decode_ztree(ztree);
    #ifdef TRACE_STATS
    ztree->stats[zpatt]++;
    #endif
    
    for (i=0; i<2; i++) 
    {
      if ((zpatt & power2(i)) != ZERO)
        pblk->mv[cur_list][i] = decode_mvd(i);
    }
    
    if ((zpatt & power2(2)) != ZERO)
      pblk->scale[cur_list] = decode_scale(0);
    
    if ((zpatt & power2(3)) != ZERO)
      pblk->DC[cur_list] = decode_offset(0); 
    
  }
  
  return;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode zero tree
*
* \param[in] ztree Zero tree structure
* \param[out] rep pointer to bitpattern of non zero leaves
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_ztree_64(ZeroTree *ztree,unsigned char* rep)
{
  int val;
  int stack;
  int *ztree_structure;
  int ztree_node_both;
  int ztree_node_left;
  int rep_idx;
  int not_leaf;
  int *idx;
  int i;
  SimpleProbEst *pe;
  
  val    = TRUE;
  stack  = ZERO;
  
  ztree_structure = ztree->m_structure;
  idx             = ztree->leaf_index;
#ifdef FRAME_SUBSTREAMS
  pe = ztree->get_pe(cs);
#else
  pe              = ztree->pe;
#endif
  rep_idx = 0;
  
  for(i=0;i<NZTREES64;i++) 
  {
    zt64[i].zpatt = 0; // FRAME_SUBSTREAMS: zt64 no longer in param.
    //param.zt64[i].zpatt = 0;
  }
  
  for(i=0;i<ZTsize64;i++)
    rep[i] = 0;
  
  
  while (TRUE) 
  {
    STATE_1:
    not_leaf = *ztree_structure++;
    
    if (not_leaf == ZERO) 
    {
      
      rep[*idx] = 1;
      #ifdef TRACE_STATS
      zt64[zt_idx[*idx][0]].zpatt += power2(zt_idx[*idx][1]); // FRAME_SUBSTREAMS: zt64 no longer in param.
      //param.zt64[param.zt_idx[*idx][0]].zpatt += power2(param.zt_idx[*idx][1]);
      #endif
      
      if (stack == ZERO)
        break;     
      idx++;
      val = stack & ONE;
      stack >>= ONE;
      
      if (val == ZERO)
        goto STATE_0;
    }
    else 
    {
      stack <<= ONE;
      
      ztree_node_both = adec->decode_symbol(pe);
      if (ztree_node_both)
        stack++;
      else 
      {
        ztree_node_left = adec->decode_symbol(pe + ONE);
        if (ztree_node_left == FALSE) 
        {
          stack++;
          pe += TWO;
          goto STATE_0;
        }
      }
      
      pe += TWO;
    }
    
    goto STATE_1;
    
    
    STATE_0:
    not_leaf = *ztree_structure++;
    
    if (not_leaf == ZERO) 
    {
      if (stack == ZERO)
        break;
      
      idx++;
      val = stack & ONE;
      stack >>= ONE;
      
      if (val != ZERO)
        goto STATE_1;
      
    }
    else 
    {
      stack <<= ONE;
      pe += TWO;
    }
    
    goto STATE_0;
  }
  
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode zero tree
*
* \param[in] ztree Zero tree structure
* \return Bit pattern of leaves of tree that have nonzero values
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_ztree(ZeroTree *ztree)
{
  int val;
  int stack;
  int *ztree_structure;
  int ztree_node_both;
  int ztree_node_left;
  int result;
  int not_leaf;
  int *idx;
  SimpleProbEst *pe;
  
  val    = TRUE;
  result = ZERO;
  stack  = ZERO;
  
  ztree_structure = ztree->m_structure;
  idx             = ztree->leaf_index;
#ifdef FRAME_SUBSTREAMS
  pe = ztree->get_pe(cs);
#else
  pe              = ztree->pe;
#endif
  
  while (TRUE) 
  {
    STATE_1:
    not_leaf = *ztree_structure++;
    
    if (not_leaf == ZERO) 
    {
      result += power2(*idx);
      
      if (stack == ZERO)
        break;
      
      idx++;
      val = stack & ONE;
      stack >>= ONE;
      
      if (val == ZERO)
        goto STATE_0;
    }
    else 
    {
      stack <<= ONE;
      
      ztree_node_both = adec->decode_symbol(pe);
      if (ztree_node_both)
        stack++;
      else 
      {
        ztree_node_left = adec->decode_symbol(pe + ONE);
        if (ztree_node_left == FALSE) 
        {
          stack++;
          pe += TWO;
          goto STATE_0;
        }
      }
      
      pe += TWO;
    }
    
    goto STATE_1;
    
    
    STATE_0:
    not_leaf = *ztree_structure++;
    
    if (not_leaf == ZERO) 
    {
      if (stack == ZERO)
        break;
      
      idx++;
      val = stack & ONE;
      stack >>= ONE;
      
      if (val != ZERO)
        goto STATE_1;
      
    }
    else 
    {
      stack <<= ONE;
      pe += TWO;
    }
    
    goto STATE_0;
  }
  
  return result;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode coefficient value
*
* Decodes the signed value of a coefficient known to be nonzero.
*
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] k Coefficient index
* \return Coefficient value
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_coeff_level(int type, int k)
{
  int val;
  
  val = adec->decode_symbol_unary(&cs->p_coeff_level[type][k], &cs->p_coeff_suffix[type]);
  if (adec->decode_symbol_half(&cs->p_coeff_sign[type]))
    val = -val;
  
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode coefficient value for 8x8 block
*
* Decodes the signed value of a coefficient known to be nonzero.
*
* \param[in] k Coefficient index
* \return Coefficient value
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_coeff_level_64(int k)
{
  int val;
  
  val = adec->decode_symbol_unary(&cs->p_coeff_level_64[k], &cs->p_coeff_suffix_64);
  if (adec->decode_symbol_half(&cs->p_coeff_sign_64))
    val = -val;
  
  return val;
}

//-----------------------------------C------------------------------------------
/*!
* \brief Decode coefficient value for 8x8 chroma block
*
* Decodes the signed value of a coefficient known to be nonzero.
*
* \param[in] k Coefficient index
* \return Coefficient value
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_coeff_level_64_CT8(int k)
{
  int val;
  
  val = adec->decode_symbol_unary(&cs->p_coeff_level_64_CT8[k], &cs->p_coeff_suffix_64_CT8);
  if (adec->decode_symbol_half(&cs->p_coeff_sign_64_CT8))
    val = -val;
  
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode block of coefficients
*
* Decodes a block of coefficients knowing that at least one coefficient is
* nonzero.
*
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[out] blk Block containing coefficients
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_coeffs_intra_luma_4x4(ZeroTree *ztree, Block *blk, int type, int sip_type)
{
  assert(blk->get_width() * blk->get_height() == 16);

  int zpatt = decode_ztree(ztree);
  
#ifdef TRACE_STATS
  ztree->stats[zpatt]++;
#endif

  int ctx = init_coeff_ctx_4x4();
  int k;
  
  for (k = TX_NCOEFF-1; k >= ZERO ; k--) 
  {
    if ((zpatt & power2(Block::ZigZag_Table_4x4[k])) != ZERO ) 
    {
      blk->val[Block::ZigZag_Table_4x4[k]] = decode_coeff_level(type, Block::ZigZag_Table_4x4[ctx]);
      ctx = update_coeff_ctx_4x4(ctx, abs(blk->val[Block::ZigZag_Table_4x4[k]]), TRUE);
    }    
  }
  
  blk->rescan(0, sip_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode block of coefficients
*
* Decodes a block of coefficients knowing that at least one coefficient is
* nonzero.
*
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[out] blk Block containing coefficients
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_coeffs(ZeroTree *ztree, Block *blk, int type)
{
  int zpatt = decode_ztree(ztree);
  
#ifdef TRACE_STATS
  ztree->stats[zpatt]++;
#endif
  
  int ctx = init_coeff_ctx_4x4();
  int k;
  
  for (k = TX_NCOEFF-1; k >= 0 ; k--) 
  {
    if (zpatt & power2(k)) 
    {
      blk->val[k] = decode_coeff_level(type, ctx);
      ctx = update_coeff_ctx_4x4(ctx, abs(blk->val[k]), FALSE);
    }    
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode block of coefficients for 8x8 block
*
* Decodes a block of coefficients knowing that at least one coefficient is
* nonzero.
*
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[out] blk Block containing coefficients
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_coeffs_8x8(ZeroTree *ztree, Block *blk8x8, int chroma)
{
  int k;
  int i;
  unsigned char zp[ZTsize64];  
  
  decode_ztree_64(ztree,zp);
  
  #ifdef TRACE_STATS
  int j,p,zflag;
  
  for(j=0;j<NZTREES64;j++)
  {
    if(j<NZTREES64-1)  
    {
      zflag = 0;
      for (p=j+1;p<NZTREES64;p++) 
      {
        zflag |= (zt64[p].zpatt!=0); // FRAME_SUBSTREAMS: zt64 no longer in param.
        //zflag |= (param.zt64[p].zpatt!=0);
      }
      zt64[j].zpatt |= zflag<<(zt64[j].nleaves); // FRAME_SUBSTREAMS: zt64 no longer in param.
      //param.zt64[j].zpatt |= zflag<<(param.zt64[j].nleaves);
    }
    zt64[j].stats[zt64[j].zpatt]++; // FRAME_SUBSTREAMS: zt64 no longer in param.
    //param.zt64[j].stats[param.zt64[j].zpatt]++;
  }
  #endif
  
  i = init_coeff_ctx_8x8();
  
  for (k = ZTsize64-1; k >= ZERO ; k--) 
  {
    if (zp[k] != ZERO) 
    {
      if (param.chroma8x8 && chroma)
        blk8x8->val[k] = decode_coeff_level_64_CT8(i);
      else
        blk8x8->val[k] = decode_coeff_level_64(i);
      
      i = update_coeff_ctx_8x8(i, abs(blk8x8->val[k]));
    }
  }

  blk8x8->inverse_zigzagscan();
  //blk8x8->blk8to4_Inverse_partition();
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode motion vector difference
*
* \param[in] type Motion vector component index (horizontal or vertical)
* \return Motion vector difference
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_mvd(int type)
{
  int val,mv_sign;
  
  val = adec->decode_symbol_unary(&cs->p_mv_level[type], &cs->p_mv_suffix);

  mv_sign = adec->decode_symbol_half(&cs->p_mv_sign);
  if (mv_sign == NEGATIVE)
    val = -val;
  
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode scale value
*
* Decodes a scale value used for intensity compensation
*
* \param[in] type Scale type (single-list=0, first scale for bipred=1, second scale for bipred=2)
* \return Scale value
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_scale(int type)
{
  int val;
  
  val = adec->decode_symbol_unary(&cs->p_scale[type], &cs->p_scale_suffix);
  if (adec->decode_symbol_half(&cs->p_scale_sign) == NEGATIVE)
    val = -val;
  
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode offset value
*
* Decodes an offset value used for intensity compensation
*
* \param[in] type Offset type (single-list=0, bipred=1)
* \return Offset value
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_offset(int type)
{
  int val;
  
  val = adec->decode_symbol_unary(&cs->p_offset[type], &cs->p_offset_suffix);
  if (adec->decode_symbol_half(&cs->p_offset_sign) == NEGATIVE)
    val = -val;
  
  return val;
}

void EntropyDecoder::decode_wlf_flag(int *wlf_flag)
{
  *wlf_flag =  adec->decode_symbol_half(&cs->p_WLF_suffix);
}

void EntropyDecoder::decode_wlf_filtertype(int *filtertype)
{
  *filtertype =  adec->decode_symbol_half(&cs->p_WLF_suffix);
  *filtertype+=2*adec->decode_symbol_half(&cs->p_WLF_suffix);
}

void EntropyDecoder::decode_val(int *p)
{
  *p=adec->decode_symbol(&cs->p_WLF_coef_zero);
  
  if (*p != ZERO) 
  {
    *p = adec->decode_symbol_unary(&cs->p_WLF_coef, &cs->p_WLF_coef_sign);
    if (adec->decode_symbol_half(&cs->p_WLF_coef_sign) == NEGATIVE) 
    {
      *p = -(*p);
    }
  }
}

void EntropyDecoder::decode_wlf_hint_2D(int **coeffs, int *size_rec_x, int *size_rec_y, int *size_pred_x, int *size_pred_y, int *size_qpe_x, int *size_qpe_y, int component, int *filter_precision, int *filter_precision_table)
{
  int m,n;
  int *i_p;
  int tmp=0;
  
  size_rec_x [component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_rec_y [component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_pred_x[component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_pred_y[component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_qpe_x [component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_qpe_y [component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  
  decode_val(&(filter_precision [component]));
  
  i_p=coeffs[component];
  
//Hints for reconstructed signal    
  for (m=0;m<size_rec_y [component];m++)
  {
    for (n=0;n<size_rec_x [component];n++)
    {
      decode_val(i_p);
      if (m==(size_rec_y [component]-1)/2 && n==(size_rec_x [component]-1)/2)
      {
        tmp=(*i_p); 
        (*i_p)+=filter_precision_table[filter_precision[component]];
      }
      i_p++;
    }
  }
  
//Hints for prediction signal    
  for (m=0;m<size_pred_y [component];m++)
  {
    for (n=0;n<size_pred_x [component];n++)
    {
      decode_val(i_p);
      if (m==(size_pred_y [component]-1)/2 && n==(size_pred_x [component]-1)/2)
      {
        (*i_p)-=tmp;
        tmp=(*i_p);
      }
      i_p++;
    }
  }
  
//Hints for quantized prediction error signal    
  for (m=0;m<size_qpe_y [component];m++)
  {
    for (n=0;n<size_qpe_x [component];n++)
    {
      decode_val(i_p);
      if (m==(size_qpe_y [component]-1)/2 && n==(size_qpe_x [component]-1)/2)
      {
        (*i_p)+=tmp;
      }
      i_p++;
    }
  }

//Offset
  decode_val(i_p);
  
}  


void EntropyDecoder::decode_wlf_hint_1D(int **coeffs, int *size_rec_x, int *size_rec_y, int *size_pred_x, int *size_pred_y, int *size_qpe_x, int *size_qpe_y, int component, int *filter_precision, int *filter_precision_table)
{
  int m,n;
  int tmp_x=0;
  int tmp_y=0;
  int *i_p;
  
  size_rec_x [component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_rec_y [component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_pred_x[component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_pred_y[component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_qpe_x [component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  size_qpe_y [component]=adec->decode_symbol_unary (&cs->p_WLF, &cs->p_WLF_suffix);
  
  decode_val(&(filter_precision [component]));
  
  i_p=coeffs[component];
  
//Hints for reconstructed signal    
  for (n=0;n<size_rec_x [component];n++)
  {
    decode_val(i_p);
    if (n==(size_rec_x [component]-1)/2)
    {
      tmp_x=(*i_p);
      (*i_p)+=filter_precision_table[filter_precision[component]];
    }
    i_p++;
  }
  for (m=0;m<size_rec_y [component];m++)
  {
    decode_val(i_p);
    if (m==(size_rec_y [component]-1)/2)
    {
      tmp_y=(*i_p);
      (*i_p)+=filter_precision_table[filter_precision[component]];
    }
    i_p++;
  }
  
//Hints for prediction signal    
  for (n=0;n<size_rec_x [component];n++)
  {
    decode_val(i_p);
    if (n==(size_pred_x [component]-1)/2)
    {
      (*i_p)-=tmp_x;
      tmp_x=(*i_p);    
    }
    i_p++;
  }
  for (m=0;m<size_rec_y [component];m++)
  {
    decode_val(i_p);
    if (m==(size_pred_y [component]-1)/2)
    {
      (*i_p)-=tmp_y;
      tmp_y=(*i_p);
    }    
    i_p++;
  }
  
//Hints for quantized prediction error signal    
  for (n=0;n<size_rec_x [component];n++)
  {
    decode_val(i_p);
    if (n==(size_qpe_x [component]-1)/2)
    {
      (*i_p)+=tmp_x;
    }    
    i_p++;
  }
  for (m=0;m<size_rec_y [component];m++)
  {
    decode_val(i_p);
    if (m==(size_qpe_y [component]-1)/2)
    {
      (*i_p)+=tmp_y;
    }
    i_p++;
  }

//Offset
  decode_val(i_p);  
}  

//-----------------------------------------------------------------------------
/*!
* \brief Decode dummy flag
*
* \return Dummy flag
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_dummy()
{
  return adec->decode_symbol_half(&cs->p_dummy);
}
//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF subpel position used
*
* \return Flag indicating if all the AIF subpel position are used
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_AIF_use_all_subpel_positions()
{
  return adec->decode_symbol(&cs->p_AIF_use_all_subpel_positions);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF subpel position pattern
*
* \return Flag indicating the AIF subpel position pattern
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_AIF_subpel_positions_pattern()
{
  int val;
  
  val=adec->decode_symbol(&cs->p_AIF_subpel_positions_pattern_zero);
  
  if (val != ZERO) 
  {
    val = adec->decode_symbol_unary(&cs->p_AIF_subpel_positions_pattern,&cs->p_AIF_subpel_positions_pattern_suffix);
  }
  
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF filter pattern
*
* \return Flag indicating the AIF filter pattern
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_AIF_filter_pattern()
{
  int val;
  
  val=adec->decode_symbol_unary(&cs->p_AIF_filter_pattern,&cs->p_AIF_filter_pattern_suffix);
  
  return val;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF differential quantized filter coefficient
*
* \return Flag indicating the AIF differential quantized filter coefficient
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_AIF_diff_qfilter_coef()
{
  int val;
  
  val=adec->decode_symbol(&cs->p_AIF_diff_qfilter_coef_zero);
  
  if (val != ZERO) 
  {
    val = adec->decode_symbol_unary(&cs->p_AIF_diff_qfilter_coef, &cs->p_AIF_diff_qfilter_coef_sign);
    if (adec->decode_symbol_half(&cs->p_AIF_diff_qfilter_coef_sign) == NEGATIVE) 
    {
      val = -val;
    }
  }
  
  return val;
}
//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF mode flag1
*
* \return Flag indicating AIF mode
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_AIF_mode_flag1()
{
  return adec->decode_symbol(&cs->p_AIF_mode_flag1);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF mode flag2
*
* \return Flag indicating AIF mode
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_AIF_mode_flag2()
{
  return adec->decode_symbol(&cs->p_AIF_mode_flag2);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode transform selection flag
*
* \return decoded transform selection flag
*/
//-----------------------------------------------------------------------------
int EntropyDecoder::decode_tsfm_flag()
{
#ifdef TRACE_DEBUG
    SPRINT_DTRACE("T16FLAG");
#endif 
  return adec->decode_symbol_half(&cs->p_tsfm_flag);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode block of coefficients for 16x16 block
*
* Decodes a block of coefficients knowing that at least one coefficient is
* nonzero.
*
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[out] blk Block containing coefficients
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_coeffs_16x16(ZeroTree **ztrees, Block *blk)
{
  int j, k, ind;
  unsigned char zp[256];  
  
  decode_subztrees(ztrees, zp);

  for(j=3; j>=0; j--)
  {
    int ctx = init_coeff_ctx_8x8();
    
    for (k = ZTsize64-1; k >= ZERO ; k--) 
    {
      ind = ZTsize64*j + k;
      if (zp[ind] != ZERO) 
      {
        blk->val[ind] = decode_coeff_level_64(ctx);

        ctx = update_coeff_ctx_8x8(ctx, abs(blk->val[ind]));
      }
    }
  }
  blk->inverse_zigzagscan();
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode sub zero tree
*
* \param[in] ztree Zero tree structure
* \param[in, out] rep Indicator for each leaf indicating if it is zero
*/
//-----------------------------------------------------------------------------
void EntropyDecoder::decode_subztrees(ZeroTree **ztree, unsigned char *rep)
{
  int i, k;

  for(i=0; i<256; i++)
    rep[i] = 0;

  for(i=0; i<16; i++)
  {
    int tree_indx = imin(i, 3);
    int zpatt = decode_ztree(ztree[tree_indx]);

    for (k = 0; k < TX_NCOEFF; k++) 
    {
      if ((zpatt & power2(k)) != ZERO )
        rep[i*16+k] = 1;
    }

    if( (zpatt & power2(TX_NCOEFF)) == ZERO )
      break;
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
