/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file EntropyEncoder.cpp
* \brief Implementation of EntropyEncoder class
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "EntropyEncoder.H"
#include "Global.H"
#include "Param.H"
#ifdef TRACE_DEBUG
#include "Trace.H"
#endif

#include "IntraPrediction.H"


//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*/
//-----------------------------------------------------------------------------
EntropyEncoder::EntropyEncoder()
{
  aenc = new ArithEncoder;
  cs = new EntropyModel;
  
  cs->reset_counts();
  
  bits_luma   = 0;
  bits_chroma = 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* \note Prints statistics to file "stats_enc.txt". This should be done
* elsewhere.
*/
//-----------------------------------------------------------------------------
EntropyEncoder::~EntropyEncoder()
{  
#if 0
  FILE *f;
  f = fopen("stats_enc.txt","wt");
  cs->print_stat(f);
  fclose(f);
#endif
  
  delete aenc;
  delete cs;
}

//-----------------------------------------------------------------------------
// Public methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Reset
*
* Reset statistics and initialize arithmetic encoder.
*
* \param[in,out] b Byte buffer containing coded data
*/
//-----------------------------------------------------------------------------
#ifndef MVCOMP_ZERORES
void EntropyEncoder::reset(Buffer *b)
#else
void EntropyEncoder::reset(Buffer *b, int mv_keep_movement)
#endif
{
  buf = b;
  buf->reset();
  aenc->set_buffer(buf);
  aenc->init();
#ifndef MVCOMP_ZERORES
  cs->reset();
#else
  cs->reset(mv_keep_movement);
#endif
  cs->reset_counts();
  
  nbits = ZERO;
  nbits_wlf_coeffs = ZERO;
}

#ifdef MVCOMP_ZERORES
//-----------------------------------------------------------------------------
/*!
* \brief Recalculate mv_pred_dir stats from previous symbol counts if possible.
*
* Recalculate mv_pred_dir stats from previous symbol counts if possible.
*
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::recalc_movement()
{
  cs->recalc_movement();
}
#endif

void EntropyEncoder::set_buffer(Buffer *b)
{
  buf=b;
  aenc->set_buffer(b);
}

void EntropyEncoder::copy_for_wlf(EntropyEncoder *e)
{
#ifdef CABAC_SYNC_POINTS
  SynchroState *saved_synchro_state = cs->synchro_state;
  int saved_allocated = cs->synchro_state_allocated;
#endif
  memcpy(this->aenc, e->aenc, sizeof (ArithEncoder));
  memcpy(this->cs  , e->cs  , sizeof (EntropyModel));
  
  this->cs->copy_coding_state(e->cs);
  this->aenc->copy_coding_state(e->aenc);
  
#ifdef CABAC_SYNC_POINTS
  cs->synchro_state = saved_synchro_state;
  cs->synchro_state_allocated = saved_allocated;
  memcpy(cs->synchro_state, e->cs->synchro_state, sizeof(e->cs->synchro_state));
#endif

  nbits            = e->nbits;
  nbits_wlf_coeffs = e->nbits_wlf_coeffs;
  bits_luma        = e->bits_luma;
  bits_chroma      = e->bits_chroma;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get number of bits written to buffer
*
* \return Number of bits
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::get_nbits()
{
  nbits = 8*buf->p_byte + buf->p_bit + aenc->get_bits_outstanding();
  return nbits;
}

//-----------------------------------------------------------------------------
/*!
* \brief Terminates arithmetic code
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::terminate()
{
  aenc->terminate();
  nbits = 8*buf->p_byte + buf->p_bit;
}

//-----------------------------------------------------------------------------
/*!
* \brief Computes entropy of coded symbol strings
*
* \return Entropy
*/
//-----------------------------------------------------------------------------
double EntropyEncoder::entropy()
{
  return cs->entropy();
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode frame type
*
* \param[in] val %Frame type (ft_I_frame, ft_P_frame, or ft_B_frame)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_frame_type(int val)
{
  aenc->encode_symbol_half(val & ONE, &cs->p_frame_type);
  if (val != ft_I_frame)
    aenc->encode_symbol_half(val >> ONE, &cs->p_frame_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode INTRA frame type
*
* \param[in] val IFrame type (ft_I_noRefresh, ft_I_IDR, or ft_I_DDR)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_I_frame_type (int val)
{
  aenc->encode_symbol_half(val & ONE, &cs->p_frame_type);  // note that this used p_frame_type, p_I_frame_type is not defined at the moment.
  if (val != ft_I_noRefresh)
  {
    aenc->encode_symbol_half((val>>ONE) & ONE, &cs->p_frame_type);  // note that this used p_frame_type, p_I_frame_type is not defined at the moment.
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode reference frame indicator
*
* \param[in] val Value indicating whether frame is a reference frame
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_reference_idc(int val)
{
  aenc->encode_symbol_half(val, &cs->p_reference_idc);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode frame counter
*
* \param[in] val %Frame counter
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_frame_counter(int val)
{
  aenc->encode_symbol_half(val != 0, &cs->p_frm_zero);
  encode_unary2(val, &cs->p_framecount, &cs->p_framecount_sign);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode number of references used in lists
*
* \param[in] val Value indicating whether one or more references may be in list
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_number_of_references(int val)
{
  aenc->encode_symbol_half(val, &cs->p_num_of_refs);
}

#ifdef MVCOMP_ZERORES
//-----------------------------------------------------------------------------
/*!
* \brief Encode mv competition config.
*
* \param[in] val Value indicating which config is being used.
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_mvcomp_config(int val)
{
  aenc->encode_symbol_half(val, &cs->p_mvcomp_config);
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Encode delta QP
*
* \param[in] val Delta QP
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_deltaqp(int val)
{
  aenc->encode_symbol_half(val != 0, &cs->p_dqp_zero);
  encode_unary2(val, &cs->p_deltaqp, &cs->p_dqp_sign);
}

#ifdef Q_OFFSET
//-----------------------------------------------------------------------------
/*!
* \brief Encode q_offsets
*
* \param[in] val Values indicating q_offsets for 4x4 and 8x8
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_q_offsets(int iTheta[2], int iTheta8x8[2])
{
  int i;
  
  for (i = 0; i < 2; i++)
  {
    aenc->encode_symbol_half(iTheta[i] != 0, &cs->p_q_offset_zero[i]);
    encode_unary2(iTheta[i], &cs->p_q_offset[i], &cs->p_q_offset_sign[i]);
  }

  for (i = 0; i < 2; i++)
  {
    aenc->encode_symbol_half(iTheta8x8[i] != 0, &cs->p_q_offset8x8_zero[i]);        
    encode_unary2(iTheta8x8[i], &cs->p_q_offset8x8[i], &cs->p_q_offset8x8_sign[i]); 
  }
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Encode the value of quantization set id, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_quantization_set_id_half() // cq_start
{
  aenc->encode_symbol_half(param.quantization_set_id != 0, &cs->quantization_set_id_value_zero);
  encode_unary3(param.quantization_set_id, &cs->quantization_set_id_value);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode values of a q matrix, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::write_qmatrix_list_half(int *qmatrixList, int sizeOfQMatrixList, int *useDefaultQMatrixFlag, int matrix_type)
{
  int j;
  int k;
  int lastScale;
  int nextScale;
  int delta_scale;

  *useDefaultQMatrixFlag = 0;
  if (qmatrixList[0] == 0)
  {
    aenc->encode_symbol_half(1, &cs->qmatrix_value_zero[matrix_type]);
    encode_unary2(-8, &cs->qmatrix_value[matrix_type], &cs->qmatrix_value_sign[matrix_type]);
    *useDefaultQMatrixFlag = 1;
    return;
  }
  else
  {
    lastScale = 8;
    for (j = 0; j < sizeOfQMatrixList; j++)
    {
      nextScale = qmatrixList[j];
      delta_scale = nextScale - lastScale;
      aenc->encode_symbol_half(delta_scale != 0, &cs->qmatrix_value_zero[matrix_type]);
      encode_unary2(delta_scale, &cs->qmatrix_value[matrix_type], &cs->qmatrix_value_sign[matrix_type]);
      lastScale = nextScale;
      if (nextScale == 0)
      {
        for (k = j; k < sizeOfQMatrixList; k++)
          qmatrixList[k] = qmatrixList[j-1];
        return;
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode update values of a q matrix, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::write_qmatrix_update_value_half(int *qmatrixList, int matrix_type)
{
  int i;

  //! scale ratio, slope ratio, number of delta values, matrix values
  for(i = 1; i < qmatrixList[3] + 4; i++)
  {
    aenc->encode_symbol_half(qmatrixList[i] != 0, &cs->qmatrix_value_zero[matrix_type]);
    if (i != 3)
      encode_unary2(qmatrixList[i], &cs->qmatrix_value[matrix_type], &cs->qmatrix_value_sign[matrix_type]);
    else
      encode_unary3(qmatrixList[i], &cs->qmatrix_value[matrix_type]);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode values of a q matrix set, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::write_qmatrix_set_half(int quantization_set_id, int matrix_type)
{
  int j;
  int k;
  int qmatrix_list_present_flag;
  int *qmatrixList;
  int qmatrix_modetype;
  int matrix_size;

  for (j = 0; j < 16; j++)
  {
    if (j<8) //!< 4x4
      matrix_size = 16;
    else if (j<14) //!< 8x8
      matrix_size = 64;
    else //!< 16x16
      matrix_size = 256;
      
    qmatrix_list_present_flag = param.qmatrix_update_check[matrix_type][quantization_set_id][j];
    aenc->encode_symbol_half(qmatrix_list_present_flag, &cs->qmatrix_list_present_flag[matrix_type]);
    qmatrixList = param.qmatrix_tmp[matrix_type][quantization_set_id][j];
    qmatrix_modetype = 1;

    if (qmatrix_list_present_flag)
    {
      qmatrix_modetype = qmatrixList[0];

      //! mode type
      aenc->encode_symbol_half(qmatrix_modetype != 0, &cs->qmatrix_value_zero[matrix_type]);
      encode_unary3(qmatrix_modetype, &cs->qmatrix_value[matrix_type]);

      if (qmatrix_modetype == 0) //!< update mode
      {
        param.matrix_scan(qmatrixList+4, j, matrix_size);
        write_qmatrix_update_value_half(qmatrixList, matrix_type);
        param.matrix_inv_scan(qmatrixList+4, j, matrix_size);
      }
      else if (qmatrix_modetype == 2 && matrix_type == 1) //!< flat mode
      {
        aenc->encode_symbol_half(qmatrixList[4] != 0, &cs->qmatrix_value_zero[matrix_type]);
        encode_unary2(qmatrixList[4], &cs->qmatrix_value[matrix_type], &cs->qmatrix_value_sign[matrix_type]);
      }
    }
    if (qmatrix_modetype == 1) //!< matrix mode
    {
      qmatrix_list_present_flag = param.qmatrix_check[matrix_type][quantization_set_id][j];
      aenc->encode_symbol_half(qmatrix_list_present_flag, &cs->qmatrix_list_present_flag[matrix_type]);
      if (qmatrix_list_present_flag)
      {
        param.matrix_scan(qmatrixList+4, j, matrix_size);
        write_qmatrix_list_half(qmatrixList+4, matrix_size, &(param.UseDefaultQMatrixFlag[matrix_type][j]), matrix_type);
        param.matrix_inv_scan(qmatrixList+4, j, matrix_size);
        if (param.UseDefaultQMatrixFlag[matrix_type][j])
        {
          for (k = 0; k < matrix_size; k++)
            qmatrixList[k+4] = 16;
        }
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode values of an updated q matrix set, 50/50 probability
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_qmatrix_half(int matrix_type) // cq_end
{
  int qmatrix_overwrite_flag;
  
  qmatrix_overwrite_flag = param.qmatrix_overwrite_flag[matrix_type][param.quantization_set_id];
  aenc->encode_symbol_half(qmatrix_overwrite_flag, &cs->qmatrix_overwrite_flag[matrix_type]);
  if (qmatrix_overwrite_flag)
    write_qmatrix_set_half(param.quantization_set_id, matrix_type);
  if(!param.already_update[matrix_type])
  {
    param.update_qmatrix(param.quantization_set_id, matrix_type);
    param.already_update[matrix_type] = 1;
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Encode fix-length string
 *
 * \param[in] val value
 * \param[in] size size of value in bits
 * \param[in] ctx context
 */
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_bits(int val, int size, FixedProbEst *pe)
{
  int i;
  
  for (i = size - 1; i >= 0; i--)
  {
    aenc->encode_symbol_half((val >> i) & 1, pe);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode interploation type
*
* \param[in] val mv accuracy 
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_inter_type(int val) 
{
  encode_bits(val, LOG_MAX_INTER_TYPE, &cs->p_inter_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode edge prediction flag
*
* \param[in] val Value indicating whether edge prediction is used
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_edge_prediction_flag(int val)
{
  aenc->encode_symbol_half(val, &cs->p_edge_prediction_flag);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode threshold for edge prediction
*
* \param[in] val Threshold value
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_threshold_edge(int val)
{
  encode_bits(val, 16, &cs->p_threshold_edge);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode RPPS type
*
* \param[in] val RPPS type
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_rpps_type(int val) 
{
  encode_bits(val, LOG_MAX_RPPS_TYPE, &cs->p_rpps_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode RPPS id
*
* \param[in] val rpps id 
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_rpps_id(int val) 
{
  encode_bits(val, LOG_MAX_RPPS_ID, &cs->p_rpps_id);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode interpolation filter type
*
* \param[in] val interpolation filter type
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_filter_type(int val) 
{
  encode_bits(val, LOG_MAX_FILTER_TYPE, &cs->p_filter_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode ic disable
*
* \param[in] val ic disable (1bit)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_ic_disable(int val) 
{
  encode_bits(val, LOG_MAX_IC_DISABLE, &cs->p_ic_disable);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode rounding type in interpolation
*
* \param[in] val ic rounding type in interpolation
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_round_type(int val) 
{
  encode_bits(val, LOG_MAX_ROUND_TYPE, &cs->p_round_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode zero block flag
*
* \param[in] val Flag indicating whether a block has coefficients
* \param[in] past Context derived from past decodes of this flag
* \param[in] chroma Flag indicating whether this is for a chroma block
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_zblk(int val, int past, int chroma)
{
#ifdef TRACE_DEBUG
  SVPRINT_ETRACE("zblk ctx ",past);
#endif
  aenc->encode_symbol(val, &cs->p_zblk[chroma][past]);
}

//-----------------------------------------------------------------------------
/*!
 * \brief Cost of zero block flag
 *
 * \param[in] val Flag indicating whether a block has coefficients
 * \param[in] past Context derived from past decodes of this flag
 * \param[in] chroma Flag indicating whether this is for a chroma block
 * \return bit cost
 */
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_zblk(int val, int past, int chroma)
{
  return aenc->cost(val, &cs->p_zblk[chroma][past]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode motion vector prediction direction
*
* \param[in] val Prediction (horizontal or vertical)
* \param[in] ctx Context derived from neighbouring motion vectors
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_mvp(int val, int ctx) 
{
#ifdef TRACE_DEBUG
  SVPRINT_ETRACE("mvp ctx ",ctx);
#endif
  aenc->encode_symbol(val, &cs->p_mvp[ctx]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode shift value
*
* \param[in] val shift (horizontal or vertical)
* \param[in] shift data from block above or the lesft
* \param[in] ctx Context derived from neighbouring shift values
* \param[in] type of shift direction
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_mvshare_size(int val, int ref, int past, int type)
{
#ifdef TRACE_DEBUG
  SVPRINT_ETRACE("shift value ctx ", past);
  RPRINT_ETRACE;
#endif
  aenc->encode_symbol(val != 0, &cs->p_mvshare_type_zero[past]);
  
  if (val != 0)
  {
    if(type >= 4)
    { // Encode shift direction if both are possible 
      aenc->encode_symbol_half(type != 5, &cs->p_mvshare_four);
    }

    if(param.mvshare_two_enable)
    {
      if (ref == 0)
      {
        aenc->encode_symbol_half(val != 2, &cs->p_mvshare_four);
      }
      else
      {
        aenc->encode_symbol(val != ref, &cs->p_mvshare_pred);
      }
    }
  }
#ifdef TRACE_DEBUG
  RPRINT_ETRACE;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode reference index
*
* \param[in] val Reference index
* \param[in] n_ref_frames Size of list for which to encode index
* \param[in] top Motion data from block above
* \param[in] left Motion data from block to the left
* \param[in] list List (LIST_0 or LIST_1)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_refindex(int val, int n_ref_frames, PredBlk *top, PredBlk *left, int list)
{
  int x0, x1;

  if (n_ref_frames < TWO)
    return;
#ifdef TRACE_DEBUG
  SPRINT_ETRACE("ref index");
  RPRINT_ETRACE;
#endif
  x0 = (left && !left->intra) ? left->ref_index[list] : (top  && !top->intra ) ?  top->ref_index[list] : 0;
  x1 = (top  && !top->intra ) ?  top->ref_index[list] : (left && !left->intra) ? left->ref_index[list] : 0;
  if (n_ref_frames > TWO) 
  {
    encode_refindex_msb(val>>1, (x0>>1)+(x1>>1), list);
    
    if (val < TWO)
      encode_refindex_lsb0(val &1, (x0 &1)+(x1 &1), list);
    else if (n_ref_frames > THREE)
      encode_refindex_lsb1(val &1, (x0 &1)+(x1 &1), list);
  }
  else 
    encode_refindex_lsb0(val, x0+x1, list);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode the Spatial Intra Prediction Type 
*
* \param[in] val sip_type (0: 4x4, 1: 2x8, 2: 8x2)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::entropycode_sip_type(int val, int ctx) 
{
#ifdef TRACE_DEBUG
  SVPRINT_ETRACE("sip_type ctx ", ctx);
  RPRINT_ETRACE;
#endif
  
  int max_val = (param.intra16x16 ? 4 : 3);

  if (param.intra16x16)
  {
    switch (val)
    {
    case 3:
      val = 0;
      break;
    case 5:
      val = 1;
      break;
    case 0:
      val = 2;
      break;
    case 1:
      val = 3;
      break;
    case 2:
      val = 4;
      break;
    }
  }
  else
  {
    val = (val + 1) & 3;    
  }
  
  aenc->encode_symbol(val > 0, &cs->p_sip_type_b1[ctx]);

  if (val > 0 && max_val > 1)
  {
    aenc->encode_symbol(val > 1, &cs->p_sip_type_b2[ctx]); 

    if (val > 1 && max_val > 2)
    {
      aenc->encode_symbol(val > 2, &cs->p_sip_type_b3[ctx]); 
      if (val > 2 && max_val > 3)
      {
        aenc->encode_symbol(val > 3, &cs->p_sip_type_b4[ctx]); 
      }
    }
  }  
}

//-----------------------------------------------------------------------------
/*!
* \brief Estimate the cost of the Spatial Intra Prediction Type 
*
* \param[in] val sip_type (0: 4x4, 1: 2x8, 2: 8x2)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_sip_type(int val, int ctx) 
{
  int cost;
  int max_val = (param.intra16x16 ? 4 : 3);
  
  if (param.intra16x16)
  {
    switch (val)
    {
    case 3:
      val = 0;
      break;
    case 5:
      val = 1;
      break;
    case 0:
      val = 2;
      break;
    case 1:
      val = 3;
      break;
    case 2:
      val = 4;
      break;
    }
  }
  else
  {
    val = (val + 1) & 3;
  }
  
  cost = aenc->cost(val > 0, &cs->p_sip_type_b1[ctx]);
  
  if (val > 0 && max_val > 1)
  {
    cost += aenc->cost(val > 1, &cs->p_sip_type_b2[ctx]); 
    
    if (val > 1 && max_val > 2)
    {
      cost += aenc->cost(val > 2, &cs->p_sip_type_b3[ctx]); 
      if (val > 2 && max_val > 3)
      {
        cost += aenc->cost(val > 3, &cs->p_sip_type_b4[ctx]); 
      }
    }
  }
  
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode Spatial Intra Prediction mode for luma
*
* \param[in] sip_mode intra prediction mode of the block
* \param[in] up_mode mode of block above
* \param[in] left_mode mode of block to the left
* \param[in] L_mode_pred_luma the value of the L-Mode predictor
* \return value indicating the luma L-Mode predictor
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::encode_spatial_intra_prediction_luma(int sip_mode, int L_mode_pred)
{
  int mode_ = sip_mode;
  int min_mode = L_mode_pred; 
#ifdef TRACE_DEBUG
    SPRINT_ETRACE("MPM");
#endif
  aenc->encode_symbol(min_mode == mode_, &cs->p_prev_intra_pred_mode_flag);

  if (min_mode == mode_) 
    return (min_mode);
  
  if (min_mode < mode_) 
  {
    mode_--;
  }
#ifdef TRACE_DEBUG
  SPRINT_ETRACE("rem_intra_pred");
  RPRINT_ETRACE;
#endif
  switch (mode_) 
  {
  case 0:
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[2]);
      break;
    case 1:
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[2]);
      break;
    case 2:
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[2]);
      break;
    case 3:
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[2]);
      break;
    case 4:
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[2]);
      break;
    case 5:
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[2]);
      break;
    case 6:
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[2]);
      break;
    case 7:
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[2]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[3]);
      break;
    case 8:
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[2]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[3]);
      aenc->encode_symbol(0, &cs->p_rem_intra_pred_mode_flag[4]);
      break;
    case 9:
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[0]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[1]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[2]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[3]);
      aenc->encode_symbol(1, &cs->p_rem_intra_pred_mode_flag[4]);
      break;
    default:
      fprintf(stderr, "EntropyEncoder::encode_spatial_intra_prediction: No such SIP mode: %d\n", mode_);
      break;
  }
  
  return (min_mode);
}

int EntropyEncoder::cost_spatial_intra_prediction_luma(int sip_mode, int L_mode_pred)
{
  int cost = aenc->cost(L_mode_pred == sip_mode, &cs->p_prev_intra_pred_mode_flag);

  if (L_mode_pred != sip_mode)
  {
    int rem = sip_mode - (L_mode_pred < sip_mode);    
    int val = imin(rem, 7);
    cost += aenc->cost((val >> 2) & 1, &cs->p_rem_intra_pred_mode_flag[0]);
    cost += aenc->cost((val >> 1) & 1, &cs->p_rem_intra_pred_mode_flag[1]);
    cost += aenc->cost( val       & 1, &cs->p_rem_intra_pred_mode_flag[2]);
    
    if (rem >= 7)
    {
      cost += aenc->cost(rem >= 8, &cs->p_rem_intra_pred_mode_flag[3]);
      
      if (rem >= 8)
      {
        cost += aenc->cost(rem >= 9, &cs->p_rem_intra_pred_mode_flag[4]);
      }
    }
  }

  return cost;  
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode Spatial Intra Prediction mode for chroma
*
* \param[in] sip_mode intra prediction mode of the block
* \param[in] up_mode mode of block above
* \param[in] left_mode mode of block to the left
* \param[in] up_is_inter flag indicating if the above block is inter
* \param[in] left_is_inter flag indicating if the left block is inter
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_spatial_intra_prediction_chroma(int sip_mode, int up_mode, int left_mode, int up_is_inter, int left_is_inter)
{
  int flag_count;
  
  if (up_mode == INT_INF || up_is_inter)
    up_mode = 0;
  
  if (left_mode == INT_INF || left_is_inter) 
    left_mode = 0;
  
  flag_count = (up_mode != 0) + (left_mode != 0);

#ifdef TRACE_DEBUG
  SVPRINT_ETRACE("intra_pred ctx", flag_count);
  RPRINT_ETRACE;
#endif
  assert(sip_mode >= 0 && sip_mode <= 5);
  
  aenc->encode_symbol(sip_mode > 0, &cs->p_intra_chroma_pred_mode[flag_count]);

  if (sip_mode > 0)
  {
    aenc->encode_symbol(sip_mode > 1, &cs->p_intra_chroma_pred_mode_1);

    if (sip_mode > 1)
    {
      aenc->encode_symbol(sip_mode > 2, &cs->p_intra_chroma_pred_mode_2);

      if (sip_mode > 2)
      {
        aenc->encode_symbol(sip_mode > 3, &cs->p_intra_chroma_pred_mode_2);

        if (sip_mode > 3)
        {
          aenc->encode_symbol(sip_mode > 4, &cs->p_intra_chroma_pred_mode_2);
        }
      }
    }
  }  
}

int EntropyEncoder::cost_spatial_intra_prediction_chroma(int sip_mode, int up_mode, int left_mode, int up_is_inter, int left_is_inter)
{
  int flag_count;
  
  if (up_mode == INT_INF || up_is_inter)
    up_mode = 0;
  
  if (left_mode == INT_INF || left_is_inter) 
    left_mode = 0;
  
  flag_count = (up_mode != 0) + (left_mode != 0);
  
  int cost = aenc->cost(sip_mode > 0, &cs->p_intra_chroma_pred_mode[flag_count]);
  
  if (sip_mode > 0)
  {
    cost += aenc->cost(sip_mode > 1, &cs->p_intra_chroma_pred_mode_1);
    
    if (sip_mode > 1)
    {
      cost += aenc->cost(sip_mode > 2, &cs->p_intra_chroma_pred_mode_2);
      
      if (sip_mode > 2)
      {
        cost += aenc->cost(sip_mode > 3, &cs->p_intra_chroma_pred_mode_2);
        
        if (sip_mode > 3)
        {
          cost += aenc->cost(sip_mode > 4, &cs->p_intra_chroma_pred_mode_2);
        }
      }
    }
  }  
  
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode mode for a block
*
* Encodes mode information for a block (intra, list0, list1, or bipred)
*
* \param[in] pblk Motion data from current block containing mode
* \param[in] left Motion data from block to the left
* \param[in] top Motion data from block above
* \param[in] Bframe Flag indicating whether current frame is a B frame
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_mode(PredBlk *pblk, PredBlk *left, PredBlk *top, int Bframe)
{
  int ctx;
  int mode, l, t;
  
  if (Bframe == TRUE) 
  {
    mode = pblk->mode_B();  
    
    l = left ? left->mode_B() : MODE_DEF;
    t = top  ?  top->mode_B() : MODE_DEF;
    
    ctx = (l & 1) + (t & 1);
#ifdef TRACE_DEBUG
    SVPRINT_ETRACE("p_mode ctx", ctx);
#endif
    aenc->encode_symbol(mode & 1, &cs->p_mode[ctx]);
    
    ctx = ((l >> 1)&1) + ((t >> 1)&1);
#ifdef TRACE_DEBUG
    SVPRINT_ETRACE("p_mode ctx", ctx);
#endif
    aenc->encode_symbol(((mode >> 1)&1), &cs->p_mode[ctx]);
  }
  else 
  {
    mode = pblk->mode_P();
    
    l = left ? left->mode_P() : MODE_P_DEF;
    t = top  ?  top->mode_P() : MODE_P_DEF;
    
    ctx = (l & 1) + (t & 1);
#ifdef TRACE_DEBUG
    SVPRINT_ETRACE("p_mode_P1 ctx", ctx);
#endif
    aenc->encode_symbol(mode & 1, &cs->p_mode_P1[ctx]);
  }
#ifdef TRACE_DEBUG
  SVPRINT_ETRACE("block mode", mode);
  RPRINT_ETRACE;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode motion data
*
* \param[in] pblk Motion data of current block
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] ztree Zero tree structure
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_mv(PredBlk *pblk, int type, ZeroTree *tmp_ztree)
{
  Block blk;
  ZeroTree *ztree;
  int v;
#ifdef TRACE_DEBUG
    SPRINT_ETRACE("motion vector");
    RPRINT_ETRACE;
#endif
  if (pblk->mode() != BI_PRED) 
  {
    blk.val[0] = pblk->mv[pblk->list][0];
    blk.val[1] = pblk->mv[pblk->list][1];
    blk.val[2] = pblk->scale[pblk->list];
    blk.val[3] = pblk->DC[pblk->list];
    ztree = (type) ? param.ztree[4] : tmp_ztree;
  }
  else 
  {
    blk.val[0] = pblk->mv[LIST_0][0];
    blk.val[1] = pblk->mv[LIST_0][1];
    blk.val[2] = pblk->mv[LIST_1][0];
    blk.val[3] = pblk->mv[LIST_1][1];
    blk.val[4] = pblk->BP_scale[0];
    blk.val[5] = pblk->BP_scale[1];
    blk.val[6] = pblk->BP_DC;
    ztree = (type) ? param.ztree[5] : tmp_ztree;
  }
  
  ztree->update_values(&blk);
  v = ztree->root->value == 0;

  aenc->encode_symbol(v, &cs->p_mv_zero[0]);

  if (v == FALSE) 
  {
    encode_ztree(ztree, ztree->root);
    
    if (pblk->mode() == BI_PRED) 
    { // BI_PRED
      encode_unary(blk.val[0], &cs->p_mv_level[0], &cs->p_mv_sign);
      encode_unary(blk.val[1], &cs->p_mv_level[1], &cs->p_mv_sign);
      encode_unary(blk.val[2], &cs->p_mv_level[0], &cs->p_mv_sign);
      encode_unary(blk.val[3], &cs->p_mv_level[1], &cs->p_mv_sign);
      encode_unary(blk.val[4], &cs->p_scale[1], &cs->p_mv_sign); // encode scales
      encode_unary(blk.val[5], &cs->p_scale[2], &cs->p_mv_sign);
      encode_unary(blk.val[6], &cs->p_offset[1], &cs->p_mv_sign); // encode offset
    }
    else 
    {
      encode_unary(blk.val[0], &cs->p_mv_level[0], &cs->p_mv_sign);
      encode_unary(blk.val[1], &cs->p_mv_level[1], &cs->p_mv_sign);
      encode_unary(blk.val[2], &cs->p_scale[0], &cs->p_mv_sign); // encode scale
      encode_unary(blk.val[3], &cs->p_offset[0], &cs->p_mv_sign); // encode offset
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode block of coefficients
*
* Encodes a block of coefficients knowing that at least one coefficient is
* nonzero.
*
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[in] blk Block containing coefficients
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_coeffs(ZeroTree *ztree, Block *blk, int type, int sip_type)
{
  int k;
  
  blk->rescan(sip_type, 0);
  ztree->update_values(blk); // update zero tree values based on coeffs
  encode_ztree(ztree, ztree->root);

  int ctx = init_coeff_ctx_4x4();
  
  for (k=TX_NCOEFF-1; k >=0 ; k--) 
  {
    int val = blk->val[Block::ZigZag_Table_4x4[k]];
    encode_unary(val, &cs->p_coeff_level[type][Block::ZigZag_Table_4x4[ctx]], &cs->p_coeff_sign[type]);
    ctx = update_coeff_ctx_4x4(ctx, abs(val), TRUE);
  }
  blk->rescan(0, sip_type);
}

int EntropyEncoder::cost_coeffs(ZeroTree *ztree, Block *blk, int type, int sip_type)
{
  int k;
  
  blk->rescan(sip_type, 0);
#ifdef FRAME_SUBSTREAMS
  int cost = ztree->cost_nodes(blk, aenc, cs);
#else
  int cost = ztree->cost_nodes(blk, aenc);
#endif
  
  int ctx = init_coeff_ctx_4x4();
  for (k = TX_NCOEFF-1; k >=0 ; k--) 
  {
    int val = blk->val[Block::ZigZag_Table_4x4[k]];
    cost += cost_unary(val, &cs->p_coeff_level[type][Block::ZigZag_Table_4x4[ctx]], &cs->p_coeff_sign[type]);
    ctx = update_coeff_ctx_4x4(ctx, abs(val), TRUE);
  }

  blk->rescan(0, sip_type);
  
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode block of coefficients
*
* Encodes a block of coefficients knowing that at least one coefficient is
* nonzero.
*
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[in] blk Block containing coefficients
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_coeffs(ZeroTree *ztree, Block *blk, int type)
{
  int k;
  
  ztree->update_values(blk); // update zero tree values based on coeffs
  
  encode_ztree(ztree, ztree->root);
  
  int ctx = init_coeff_ctx_4x4();
  for (k=TX_NCOEFF-1; k >=0 ; k--) 
  {
    encode_unary(blk->val[k], &cs->p_coeff_level[type][ctx], &cs->p_coeff_sign[type]);
    ctx = update_coeff_ctx_4x4(ctx, abs(blk->val[k]), FALSE);
  }
}

int EntropyEncoder::cost_coeffs(ZeroTree *ztree, Block *blk, int type)
{
  int k;
  
#ifdef FRAME_SUBSTREAMS
  int cost = ztree->cost_nodes(blk, aenc, cs);
#else
  int cost = ztree->cost_nodes(blk, aenc);
#endif
  
  int ctx = init_coeff_ctx_4x4();
  for (k = TX_NCOEFF-1; k >=0 ; k--) 
  {
    int val = blk->val[k];
    cost += cost_unary(val, &cs->p_coeff_level[type][ctx], &cs->p_coeff_sign[type]);
    ctx = update_coeff_ctx_4x4(ctx, abs(blk->val[k]), FALSE);
  }
  
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode block of coefficients for 8x8 block
*
* Encodes a block of coefficients knowing that at least one coefficient is
* nonzero.
*
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[in] blk Block containing coefficients
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_coeffs_8x8(ZeroTree *ztree, Block *blk8x8, int chroma)
{
  int k;
  
  blk8x8->zigzagscan();
  
  ztree->update_values(blk8x8); // update zero tree values based on coeffs
  encode_ztree(ztree, ztree->root);

  int ctx = init_coeff_ctx_8x8();
  for (k = ZTsize64-1; k >=0 ; k--) 
  {
    if (param.chroma8x8 && chroma)
    {
      encode_unary(blk8x8->val[k], &cs->p_coeff_level_64_CT8[ctx], &cs->p_coeff_sign_64_CT8);
    }
    else
    {
      encode_unary(blk8x8->val[k], &cs->p_coeff_level_64[ctx], &cs->p_coeff_sign_64);
    }

    ctx = update_coeff_ctx_8x8(ctx, abs(blk8x8->val[k]));
  }
  
  blk8x8->inverse_zigzagscan();
}

int EntropyEncoder::cost_coeffs_8x8(ZeroTree *ztree, Block *blk8x8, int chroma)
{
  int k;
  
  blk8x8->zigzagscan();
  
  ztree->update_values(blk8x8); // update zero tree values based on coeffs  
  int cost = cost_ztree(ztree, ztree->root);
  
  int ctx = init_coeff_ctx_8x8();
  for (k=ZTsize64-1; k >=0 ; k--) 
  {
    if (param.chroma8x8 && chroma)
    {
      cost += cost_unary(blk8x8->val[k], &cs->p_coeff_level_64_CT8[ctx], &cs->p_coeff_sign_64_CT8);
    }
    else
    {
      cost += cost_unary(blk8x8->val[k], &cs->p_coeff_level_64[ctx], &cs->p_coeff_sign_64);
    }
    
    ctx = update_coeff_ctx_8x8(ctx, abs(blk8x8->val[k]));
  }

  blk8x8->inverse_zigzagscan();

  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding a zerotree
*
* \param[in] node Zero tree node
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_ztree(ZTNode *node, int type) 
{
  int val;
  int lval, rval;
  int cost;
  
  cost = 0;
  
  if (!node->leaf_node()) 
  {
    lval = node->left->value;
    rval = node->right->value;
    
    val = lval & rval;
    cost += param.bitsM[val][param.ztree[type]->i_prob[2*(node->id)+0]];
    
    if (val == 0)
      cost += param.bitsM[lval][param.ztree[type]->i_prob[2*(node->id)+1]];
    
    if (lval != 0)
      cost += cost_ztree(node->left, type);
    if (rval != 0)
      cost += cost_ztree(node->right, type);
  }
  
  return cost;
}

int EntropyEncoder::cost_ztree(ZeroTree *ztree, ZTNode *node) 
{
  int val;
  int lval, rval;
  int cost;
  
  cost = 0;
  
  if (!node->leaf_node()) 
  {
    lval = node->left->value;
    rval = node->right->value;
    
    val = lval & rval;
#ifdef FRAME_SUBSTREAMS
    cost += aenc->cost(val, &ztree->get_pe(cs)[2*(node->id)+0]);
#else
    cost += aenc->cost(val, &ztree->pe[2*(node->id)+0]);
#endif
    
    if (val == 0)
#ifdef FRAME_SUBSTREAMS
      cost += aenc->cost(lval, &ztree->get_pe(cs)[2*(node->id)+1]);
#else
      cost += aenc->cost(lval, &ztree->pe[2*(node->id)+1]);
#endif
    
    if (lval != 0)
      cost += cost_ztree(ztree, node->left);
    if (rval != 0)
      cost += cost_ztree(ztree, node->right);
  }
  
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding a motion vector difference
*
* \param[in] val Difference
* \param[in] i Motion vector component index (vertical or horizontal)
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_mv(int val, int i)
{
  return cost_unary(val, &cs->p_mv_level[i], &cs->p_mv_sign);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding a reference index
*
* \param[in] val Reference index of current block
* \param[in] n_ref_frames Size of list for which to encode index
* \param[in] x0 Reference index of block to the left of current block
* \param[in] x1 Reference index of block above current block
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_refindex(int val, int n_ref_frames, int x0, int x1)
{
  int r;
  
  r = ZERO;
  
  if (n_ref_frames < TWO)
    return r;
  
  if (n_ref_frames > TWO) 
  {
    r += cost_refindex_msb(val>>1, (x0>>1)+(x1>>1));
    
    if (val < TWO)
      r += cost_refindex_lsb0(val &1, (x0 &1)+(x1 &1));
    else if (n_ref_frames > THREE)
      r += cost_refindex_lsb1(val &1, (x0 &1)+(x1 &1));
  }
  else 
    r += cost_refindex_lsb0(val, x0+x1);
  
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding mode
*
* \param[in] mode Mode of current block
* \param[in] left Mode of block to the left of current block
* \param[in] top Mode of block above current block
* \param[in] Bframe Flag indicating whether current frame is a B frame
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_mode(int mode, int left, int top, int Bframe)
{
  int ctx;
  int r;
  
  if (Bframe == TRUE) 
  {
    ctx = (left & 1) + (top & 1);
    r = aenc->cost(mode & 1, &cs->p_mode[ctx]);
    ctx = (left >> 1) + (top >> 1);
    r += aenc->cost(mode >> 1, &cs->p_mode[ctx]); 
  }
  else 
  {
    ctx = (left & 1) + (top & 1);
    r = aenc->cost(mode & 1, &cs->p_mode_P1[ctx]);
    
    if (!param.bugfix94)
    {
      if (mode != MODE_P_INTRA)
      {
        ctx = (left == MODE_P_LIST_0_8x8 && top == MODE_P_LIST_0_8x8) ? 0 : 1;
        r += aenc->cost(((mode>>1) & 1), &cs->p_mode_P2[ctx]);
      }
      else
      {
        ctx = (left == MODE_P_INTRA && top == MODE_P_INTRA) ? 0 : 1;
        r += aenc->cost(((mode>>1) & 1), &cs->p_mode_TP1[ctx]);
      }
    }
  }
  
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding shift value
*
* \param[in] val shift (horizontal or vertical)
* \param[in] shift data from block above or the lesft
* \param[in] ctx Context derived from neighbouring shift values
* \param[in] type of shift direction
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_mvshare_size(int val, int ref, int past, int type)
{
  int r;
     
  r = aenc->cost(val != 0, &cs->p_mvshare_type_zero[past]);
  
  if (val != 0)
  {
    if (type >= 4)
    {
      r += aenc->cost(type != 5, &cs->p_mvshare_four);
    }

    if (param.mvshare_two_enable)
    {
      if (ref == 0)
      {
        r += aenc->cost(val != 2, &cs->p_mvshare_four);
      }
      else
      {
        r += aenc->cost(val != ref, &cs->p_mvshare_pred);
      }
    }
  }
     
  return r;
}


//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding motion data prediction direction
*
* \param[in] val Prediction direction (horizontal or vertical)
* \param[in] ctx Context derived from neighbouring motion vectors
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_mvp(int val, int ctx) 
{
  return aenc->cost(val, &cs->p_mvp[ctx]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of encoding a one coefficients
*
* \param[in] blk Block containing coefficients
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] pos Position of the coefficient in the block
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
double EntropyEncoder::cost_one_coeff(int val, int type, int pos) 
{
  int cost = cost_unary(val, &cs->p_coeff_level[type][pos], &cs->p_coeff_sign[type]);
  
  assert(cost >= 0);
  
  return cost * (1.0 / power2(RATE_BITS));
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding a zerotree with a different zero tree from default
* 
* Calls cost_ztree_rdo_base()
*
* \param[in] node Zero tree node
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] ztree The zero tree to use.
* \return Cost in double
*/
//-----------------------------------------------------------------------------
double EntropyEncoder::cost_ztree_rdo(Block *blk, int type, int context, ZeroTree *ztree) 
{
#ifdef FRAME_SUBSTREAMS
  int cost = ztree->cost_nodes(blk, aenc, cs);
#else
  int cost = ztree->cost_nodes(blk, aenc);
#endif
  int val = ztree->root->value == 0;
  cost += aenc->cost(val, &cs->p_zblk[type][context]);
  return cost * (1.0 / power2(RATE_BITS));
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of encoding a one coefficients for 8x8 Inter block
*
* \param[in] blk Block containing coefficients
* \param[in] pos Position of the coefficient in the block
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
double EntropyEncoder::cost_one_coeff8x8(int val, int pos) 
{
  int cost = cost_unary(val, &cs->p_coeff_level_64[pos], &cs->p_coeff_sign_64);
  assert(cost >= 0);
  
  return cost * (1.0 / power2(RATE_BITS));
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding a zerotree with a different zero tree from default for 8x8 inter block
* 
* Calls cost_ztree_rdo_base8x8()
*
* \param[in] node Zero tree node
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] ztree The zero tree to use.
* \return Cost in double
*/
//-----------------------------------------------------------------------------
double EntropyEncoder::cost_ztree_rdo8x8(Block *blk, int type, int context, ZeroTree *ztree) 
{
  blk->zigzagscan();

  ztree->update_values(blk); // update zero tree values based on coeffs
  
  int val = ztree->root->value == 0;
  double cost = (double)aenc->cost(val, &cs->p_zblk[type][context]) * (1.0 / power2(RATE_BITS));
  if (val == FALSE)
    cost += (double)(cost_ztree_rdo_base(ztree, ztree->root)) * (1.0 / power2(RATE_BITS)); // NOTE TYPE HERE
  
  blk->inverse_zigzagscan();

  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of a node in a zerotree
*
* \param[in] ztree The zero tree to use.
* \param[in] node Zero tree node
* \return Cost in double
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_ztree_rdo_base(ZeroTree *ztree, ZTNode *node)
{
  int i;
  int cost = 0;
  
  for (i = 0; i < ztree->nleaves - 1; i++)
  {
    if (node[i].value)
    {
      int lval = node[i].left->value;
      int rval = node[i].right->value;
      int val = lval & rval;
      
#ifdef FRAME_SUBSTREAMS
      cost += aenc->cost(val, &ztree->get_pe(cs)[2 * i + 0]);
#else
      cost += aenc->cost(val, &ztree->pe[2 * i + 0]);
#endif
      
      if (val == 0)
#ifdef FRAME_SUBSTREAMS
        cost += aenc->cost(lval, &ztree->get_pe(cs)[2 * i + 1]);
#else
        cost += aenc->cost(lval, &ztree->pe[2 * i + 1]);
#endif
    }
  }
  
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of encoding scale parameter
*
* \param[in] val Scale parameter
* \param[in] type Scale type (single-list=0, first scale for bipred=1, second scale for bipred=2)
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_scale(int val, int type)
{
  return cost_unary(val, &cs->p_scale[type], &cs->p_mv_sign);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of encoding offset parameter
*
* \param[in] val Offset parameter
* \param[in] type Offset type (single-list=0, bipred=1)
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_offset(int val, int type)
{
  return cost_unary(val, &cs->p_offset[type], &cs->p_mv_sign);
}

//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Encode a signed nonzero integer
*
* \param[in] val Integer
* \param[in,out] pe_level Probability estimator for coding level
* \param[in,out] pe_sign Probability estimator for coding sign
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_unary(int val, UnaryProbEst *pe_level, FixedProbEst *pe_sign)
{
  if (val == ZERO)
    return;
  
  aenc->encode_symbol_unary(iabs(val), pe_level, pe_sign);
  aenc->encode_symbol_half(val < ZERO, pe_sign);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode a signed nonzero integer
*
* \note This method does not use adaptive suffixes to encode integer.
*
* \param[in] val Integer
* \param[in,out] pe_level Probability estimator for coding level
* \param[in,out] pe_sign Probability estimator for coding sign
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_unary2(int val, FixedProbEst *pe_level, FixedProbEst *pe_sign)
{
  int i;
  
  if (val == 0)
    return;
  
  for (i = 1; i < iabs(val); i++)
    aenc->encode_symbol_half(1, pe_level);
  
  aenc->encode_symbol_half(0, pe_level);
  aenc->encode_symbol_half(val < 0, pe_sign);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode a unsigned nonzero integer
*
* \note This method does not use adaptive suffixes to encode integer.
*
* \param[in] val Integer
* \param[in,out] pe_level Probability estimator for coding level
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_unary3(int val, FixedProbEst *pe_level) // cq_start, cq_end
{
  int i;
  
  if (val == 0)
    return;
  for (i = 1; i < iabs(val); i++)
    aenc->encode_symbol_half(1, pe_level);
  aenc->encode_symbol_half(0, pe_level);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost for coding a signed nonzero integer
*
* \param[in] val Integer
* \param[in] pe_level Probability estimator for coding level
* \param[in] pe_sign Probability estimator for coding sign
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_unary(int val, UnaryProbEst *pe_level, FixedProbEst *pe_sign)
{
  int r = 0;
  
  if (val != ZERO) 
  {
    r += aenc->cost_unary(iabs(val), pe_level, pe_sign);
    r += aenc->cost(val < ZERO, pe_sign);
  }
  
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode zerotree
*
* \param[in] ztree Pointer to zerotree
* \param[in] node Pointer to zerotree node
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_ztree(ZeroTree *ztree, ZTNode *node)
{
  int val, lval, rval;
  
  if (node->leaf_node() == FALSE) 
  {
    lval = node->left->value;
    rval = node->right->value;
    
    val = lval & rval;
#ifdef FRAME_SUBSTREAMS
    aenc->encode_symbol(val, &ztree->get_pe(cs)[2*node->id+0]);
#else
    aenc->encode_symbol(val, &ztree->pe[2*node->id+0]);
#endif
    
    if (val == 0)
#ifdef FRAME_SUBSTREAMS
      aenc->encode_symbol(lval, &ztree->get_pe(cs)[2*node->id+1]);
#else
      aenc->encode_symbol(lval, &ztree->pe[2*node->id+1]);
#endif
    
    if (lval != 0)
      encode_ztree(ztree, node->left);
    if (rval != 0)
      encode_ztree(ztree, node->right);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode LSB of reference index
*
* This is used when MSB = 0.
*
* \param[in] val Value of LSB
* \param[in] ctx Context derived from neighbouring reference indices
* \param[in] list List (LIST_0 or LIST_1)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_refindex_lsb0(int val, int ctx, int list)
{
  aenc->encode_symbol(val, &cs->p_refidx_lsb0[list][ctx]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode LSB of reference index
*
* This is used when MSB = 1.
*
* \param[in] val Value of LSB
* \param[in] ctx Context derived from neighbouring reference indices
* \param[in] list List (LIST_0 or LIST_1)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_refindex_lsb1(int val, int ctx, int list)
{
  aenc->encode_symbol(val, &cs->p_refidx_lsb1[list][ctx]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode MSB of reference index
*
* \param[in] val Value of MSB
* \param[in] ctx Context derived from neighbouring reference indices
* \param[in] list List (LIST_0 or LIST_1)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_refindex_msb(int val, int ctx, int list)
{
  aenc->encode_symbol(val, &cs->p_refidx_msb[list][ctx]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost for coding LSB of reference index
*
* This is used when MSB = 0.
*
* \param[in] val Value of LSB
* \param[in] ctx Context derived from neighbouring reference indices
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_refindex_lsb0(int val, int ctx)
{
  return param.bitsM[val][cs->p_refidx_lsb0[0][ctx].p];
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost for coding LSB of reference index
*
* This is used when MSB = 1.
*
* \param[in] val Value of LSB
* \param[in] ctx Context derived from neighbouring reference indices
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_refindex_lsb1(int val, int ctx)
{
  return param.bitsM[val][cs->p_refidx_lsb1[0][ctx].p];
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost for coding MSB of reference index
*
* \param[in] val Value of LSB
* \param[in] ctx Context derived from neighbouring reference indices
* \return Cost in bits (fixed-point number with 10 fractional bits)
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_refindex_msb(int val, int ctx)
{
  return param.bitsM[val][cs->p_refidx_msb[0][ctx].p];
}


//-----------------------------------------------------------------------------
/*!
* \brief Copy the coding state
*
* Note for speed, the buffer content is not copied, just the state.
*
* \param[in] eenc The EntropyEncoder to copy from
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::copy_coding_state(EntropyEncoder *eenc)
{
  aenc->copy_coding_state(eenc->aenc);
  cs->copy_coding_state(eenc->cs);
  buf->copy_coding_state(eenc->buf);
  
  //  memcpy(cs, eenc->cs, sizeof (EntropyModel));
  
  nbits       = eenc->nbits;
  bits_luma   = eenc->bits_luma;
  bits_chroma = eenc->bits_chroma;
}

void EntropyEncoder::encode_wlf_flag(int *wlf_flag)
{
  aenc->encode_symbol_half(*wlf_flag!=0?1:0,&cs->p_WLF_suffix);
}

void EntropyEncoder::encode_wlf_filtertype(int *filtertype)
{
  aenc->encode_symbol_half((*filtertype)&1,&cs->p_WLF_suffix);
  aenc->encode_symbol_half((*filtertype)&2,&cs->p_WLF_suffix);
}

void EntropyEncoder::encode_val(int p)
{
  aenc->encode_symbol(p != 0, &cs->p_WLF_coef_zero);
  encode_unary(p, &cs->p_WLF_coef, &cs->p_WLF_coef_sign);
}

void EntropyEncoder::encode_wlf_coeffs(int **coeffs, int *size_rec_x, int *size_rec_y, int *size_pred_x, int *size_pred_y, int *size_qpe_x, int *size_qpe_y, int component, int *filter_precision, int *filter_precision_table)
{
  int m,n;
  int *i_p;
  int tmp=0, tmp2=0;
  
  aenc->encode_symbol_unary(size_rec_x [component], &cs->p_WLF,&cs->p_WLF_suffix);
  aenc->encode_symbol_unary(size_rec_y [component], &cs->p_WLF,&cs->p_WLF_suffix);
  aenc->encode_symbol_unary(size_pred_x[component], &cs->p_WLF,&cs->p_WLF_suffix);
  aenc->encode_symbol_unary(size_pred_y[component], &cs->p_WLF,&cs->p_WLF_suffix);
  aenc->encode_symbol_unary(size_qpe_x [component], &cs->p_WLF,&cs->p_WLF_suffix);
  aenc->encode_symbol_unary(size_qpe_y [component], &cs->p_WLF,&cs->p_WLF_suffix);
   
  encode_val(filter_precision [component]);
  
  i_p=coeffs[component];
  
//Coefficients for reconstructed signal    
  for (m=0;m<size_rec_y [component];m++)
  {
    for (n=0;n<size_rec_x [component];n++)
    {
      if (m==(size_rec_y [component]-1)/2 && n==(size_rec_x [component]-1)/2)
      {
        tmp=(*i_p)-filter_precision_table[filter_precision[component]];
        encode_val(tmp);
      }
      else
      {
        encode_val(*i_p);
      }
      i_p++;
    }
  }
  
//Coefficients for prediction signal    
  for (m=0;m<size_pred_y [component];m++)
  {
    for (n=0;n<size_pred_x [component];n++)
    {
      if (m==(size_pred_y [component]-1)/2 && n==(size_pred_x [component]-1)/2)
      {
        encode_val( (*i_p) + tmp);
        tmp2=(*i_p);
      }
      else
      {
        encode_val(*i_p);
      }
      i_p++;
    }
  }
  
//Coefficients for quantized prediction error signal    
  for (m=0;m<size_qpe_y [component];m++)
  {
    for (n=0;n<size_qpe_x [component];n++)
    {
      if (m==(size_qpe_y [component]-1)/2 && n==(size_qpe_x [component]-1)/2)
      {
        encode_val((*i_p)-tmp2);
      }
      else
      {
        encode_val(*i_p);
      }
      i_p++;
    }
  }

//Offset
  encode_val(*i_p);
}  

//-----------------------------------------------------------------------------
/*!
* \brief Encode dummy flag
*
* \param[in] 
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_dummy()
{
    aenc->encode_symbol_half(0, &cs->p_dummy);
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy the coding state for cs
*
* Note for speed, the buffer content is not copied, just the state.
*
* \param[in] eenc The EntropyEncoder to copy from
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::copy_coding_state_cs(EntropyEncoder *eenc)
{
  cs->copy_coding_state(eenc->cs);
}

//-----------------------------------------------------------------------------
/*!
* \brief Get entropy model for mv zeroc
*
* \param[in] ctx Context derived from neighbouring reference indices
* \return Probability estimate 
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::output_mv_zero(int val)
{
  return (int)cs->p_mv_zero[val].p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get entropy model for prev intra pred mode
*
* \return Probability estimate 
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::output_prev_intra_pred_mode_flag()
{
  return (int)cs->p_prev_intra_pred_mode_flag.p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get entropy model for rem intra pred mode
*
* \param[in] ctx Context derived from neighbouring reference indices
* \return Probability estimate 
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::output_rem_intra_pred_mode_flag(int ctx)
{
  return (int)cs->p_rem_intra_pred_mode_flag[ctx].p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get entropy model for intra chroma pred mode
*
* \param[in] ctx Context derived from neighbouring reference indices
* \return Probability estimate 
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::output_intra_chroma_pred_mode(int ctx)
{
  return (int)cs->p_intra_chroma_pred_mode[ctx].p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get entropy model for intra chroma pred mode 1
*
* \return Probability estimate 
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::output_intra_chroma_pred_mode_1()
{
  return (int)cs->p_intra_chroma_pred_mode_1.p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get entropy model for intra chroma pred mode 2
*
* \return Probability estimate 
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::output_intra_chroma_pred_mode_2()
{
  return (int)cs->p_intra_chroma_pred_mode_2.p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get entropy model for coeff_level
*
* \param[in] val Level of coefficient
* \param[in] type Index of YUV component
* \param[in] ctx Context derived from neighbouring reference indices
* \return Probability estimate 
*/
//-----------------------------------------------------------------------------
int EntropyEncoder::cost_coeff_level(int val, int type, int ctx)
{
  return cost_unary(val, &cs->p_coeff_level[type][ctx], &cs->p_coeff_sign[type]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF subpel position used
*
* \param[in] Flag indicating if all the AIF subpel position are used
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_AIF_use_all_subpel_positions(int val) 
{
  aenc->encode_symbol(val, &cs->p_AIF_use_all_subpel_positions);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode AIF subpel position pattern
*
* \param[in] Flag indicating the AIF subpel position pattern
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_AIF_subpel_positions_pattern(int val) 
{
  aenc->encode_symbol(val != 0, &cs->p_AIF_subpel_positions_pattern_zero);
  
  if (val != 0) 
  {
    aenc->encode_symbol_unary(val, &cs->p_AIF_subpel_positions_pattern,&cs->p_AIF_subpel_positions_pattern_suffix);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode AIF filter pattern
*
* \param[in] Flag indicating the AIF filter pattern
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_AIF_filter_pattern(int val) 
{
  aenc->encode_symbol_unary(val, &cs->p_AIF_filter_pattern,&cs->p_AIF_filter_pattern_suffix);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF differential quantized filter coefficient
*
* \param[in] Flag indicating the AIF differential quantized filter coefficient
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_AIF_diff_qfilter_coef(int val) 
{
  aenc->encode_symbol(val != 0, &cs->p_AIF_diff_qfilter_coef_zero);
  encode_unary(val, &cs->p_AIF_diff_qfilter_coef,&cs->p_AIF_diff_qfilter_coef_sign);
}
//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF mode flag1
*
* \param[in] Flag indicating AIF mode
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_AIF_mode_flag1(int val) 
{
  aenc->encode_symbol(val, &cs->p_AIF_mode_flag1);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decode AIF mode flag2
*
* \param[in] Flag indicating AIF mode
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_AIF_mode_flag2(int val) 
{
  aenc->encode_symbol(val, &cs->p_AIF_mode_flag2);
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode transform selection flag
*
* \param[in] val transform selection flag
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_tsfm_flag( int val )
{
  aenc->encode_symbol_half(val, &cs->p_tsfm_flag);
}

int EntropyEncoder::cost_tsfm_flag( int val )
{
  return aenc->cost(val, &cs->p_tsfm_flag);
}

//-----------------------------------------------------------------------------
/*!
* \brief Check whether a sub tree is all-zero
*
* \param[in out] significance Indicate whether each sub-tree is all-zero
* \param[in]     val          Leaf values
* \param[in]     nsubsets     The number of sub-trees
* \param[in]     datasize     The number of leaves in the tree
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::subset_significance(Block *significance, Block *val, int nsubsets, int data_size)
{
  int i, j;
  int step, index=0;

  step = data_size/nsubsets;
  for( i=0; i < nsubsets; i++)
  {
    significance->val[i] = 0;
    for( j = index; j < index + step; j++)
      if(val->val[j]!=0)
        significance->val[i] = 1;
    index += step;
  }

  for(i=nsubsets-2; i>=0; i--)
    significance->val[i] = significance->val[i] || significance->val[i+1];
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy the values of a large block and the significance to a subtree
*
* \param[in out] dst          A subtree with (size) leaves and a connecting node
* \param[in]     src          Coefficient values of a large block
* \param[in]     size         The number of leaves of a sub-trees
* \param[in]     significance Indicate whether there are more subtrees
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::copy_vals(Block *dst, Block *src, int offset, int size, int significance)
{
  int i;
  for(i=0; i<size; i++)
    dst->val[i] = (src->val[i+offset]!=ZERO);
  dst->val[size] = significance;
}
//-----------------------------------------------------------------------------
/*!
* \brief Encode block of coefficients for 16x16 block
*
* Encodes a block of coefficients knowing that at least one coefficient is
* nonzero.
*
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[in] blk Block containing coefficients
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
*/
//-----------------------------------------------------------------------------
void EntropyEncoder::encode_coeffs_16x16(ZeroTree **ztree, Block *blk)
{
  int i, j, k, ind;
  Block significance(1, 17);
  Block vals(1, 17);

  blk->zigzagscan();
  subset_significance(&significance, blk, 16, 256);  // Subset significance
  significance.val[16] = 0;

  for(i=0; i<16; i++)
  {
    if( significance.val[i]==0 )
      break;

    copy_vals(&vals, blk, 16*i, 16, significance.val[i+1]);

    int idx = imin(i, 3);
    ztree[idx]->update_values(&vals);
    encode_ztree(ztree[idx], ztree[idx]->root);
  }

  for(j=3; j>=0; j--)
  {
    int ctx = init_coeff_ctx_8x8();
    
    for (k=ZTsize64-1; k >=0 ; k--) 
    {
      ind = ZTsize64*j + k;
      encode_unary(blk->val[ind], &cs->p_coeff_level_64[ctx], &cs->p_coeff_sign_64);

      ctx = update_coeff_ctx_8x8(ctx, abs(blk->val[ind]));
    }
  }
  
  blk->inverse_zigzagscan();
}

int EntropyEncoder::cost_coeffs_16x16(ZeroTree **ztree, Block *blk)
{
  int i, j, k, ind;
  Block significance(1, 17);
  Block vals(1, 17);
  
  int cost = 0;
  blk->zigzagscan();
  subset_significance(&significance, blk, 16, 256);  // Subset significance
  significance.val[16] = 0;
  
  for(i=0; i<16; i++)
  {
    if( significance.val[i]==0 )
      break;
    
    copy_vals(&vals, blk, 16*i, 16, significance.val[i+1]);
    
    int idx = imin(i, 3);
    ztree[idx]->update_values(&vals);
    cost += cost_ztree(ztree[idx], ztree[idx]->root);
  }
  
  for(j=3; j>=0; j--)
  {
    int ctx = init_coeff_ctx_8x8();

    for (k=ZTsize64-1; k >=0 ; k--) 
    {
      ind = ZTsize64*j + k;
      cost += cost_unary(blk->val[ind], &cs->p_coeff_level_64[ctx], &cs->p_coeff_sign_64);
      
      ctx = update_coeff_ctx_8x8(ctx, abs(blk->val[ind]));
    }
  }
  
  blk->inverse_zigzagscan();
  
  return cost;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
