/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file EntropyModel.cpp
* \brief Implementation of EntropyModel class
*/
//-----------------------------------------------------------------------------

#include "EntropyModel.H"
#include "Param.H"
#include "Global.H"

// Assign various print counters to zblks, coeffs and ztrees.
static unsigned int zblk_pbcf_classes[] = {CTXT_CLASS_PBCF_ZBLK0,
                                           CTXT_CLASS_PBCF_ZBLK1,
                                           CTXT_CLASS_PBCF_ZBLK2,
                                           CTXT_CLASS_PBCF_ZBLK3 };
static unsigned int coeff_pbcf_classes[] = {CTXT_CLASS_PBCF_COEFF0,
                                            CTXT_CLASS_PBCF_COEFF1,
                                            CTXT_CLASS_PBCF_COEFF2,
                                            CTXT_CLASS_PBCF_COEFF3,
                                            CTXT_CLASS_PBCF_COEFF4,
                                            CTXT_CLASS_PBCF_COEFF5 };
static unsigned int ztree_pbcf_classes[] = {CTXT_CLASS_PBCF_ZTREE0,
                                            CTXT_CLASS_PBCF_ZTREE1,
                                            CTXT_CLASS_PBCF_ZTREE2,
                                            CTXT_CLASS_PBCF_ZTREE3,
                                            CTXT_CLASS_PBCF_ZTREE4,
                                            CTXT_CLASS_PBCF_ZTREE5,
                                            CTXT_CLASS_PBCF_ZTREE6,
                                            CTXT_CLASS_PBCF_ZTREE7,
#ifdef STATS_16x16_FIX
                                            CTXT_CLASS_PBCF_ZTREE8,
                                            CTXT_CLASS_PBCF_ZTREE9};
#else
                                            CTXT_CLASS_PBCF_ZTREE8};
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] share Non-null if this entropymodel is sharing synchronisation state with others.
*
* Resets all bit usage counters
*/
//-----------------------------------------------------------------------------
#ifdef CABAC_SYNC_POINTS
EntropyModel::EntropyModel(SynchroState *share) 
#else
EntropyModel::EntropyModel() 
#endif
{
  int i;

  for (i=0; i<4; i++)
    bstat[i] = 0.0;

  bstat_intra_block  = 0.0;
  bstat_refindex_msb = 0.0;
  bstat_refindex_lsb = 0.0;
  bstat_intcomp = 0.0;
  bstat_mvpdir  = 0.0;
  bstat_mvdsign = 0.0;
  bstat_mvdzero = 0.0;
  bstat_mvd     = 0.0;

#ifdef CABAC_SYNC_POINTS
  if (share != NULL)
  {
      synchro_state = share;
      synchro_state_allocated = 0;
  }
  else
  {
      synchro_state = new SynchroState();
      synchro_state_allocated = 1;
  }
#endif
  //------------------------------------------
  //--------- initialize ctx ID
  for(int i=0; i<MAX_ID_CTX; i++)
  {
    p_fCtx[i]= 0;
    p_aCtx[i]= 0;
  }
  max_idFCtx = max_idACtx = -1;

  int j,k;

  // Changes: we add a name and set of classifications to each proba.

  // We have added a tag (the name of the
  // proba) to each line of the .dat file.  This disconnects the
  // push order here from the order in the .dat file, so that the orders can
  // be different.
  // In this case we can change the order of pushes here to reflect other
  // external documentation, such as the encoder specification.

  // A p_DUMMY that is pushed here is a placeholder for an unused
  // tool to ensure other proba ids are correct for documentation.
  // Only non-DUMMY probas are written to and read from external files,
  // and only non-DUMMY probs are read from the external .dat files.

  // These probas should be pushed in an order agreeing with
  // external documentation, to be assigned correct id numbers.

  // Encoded streams only contain probas used by tools currently defined, and
  // probas are written and read in the order that they're pushed here.

  // idACtx: xxx
  for (j=0; j<4; j++) 
  {
    char const *zblk_names[] = {"p_zblk0", "p_zblk1", "p_zblk2", "p_zblk3"};
    for (i=0; i<NCTX_ZBLK; i++) 
    {
      pushAProbCtx(p_zblk[j][i], zblk_names[j], CTXT_CLASS_PBCS_SUMA|zblk_pbcf_classes[j]);
    }
  }

  // idACtx: xxx
  for (j=0; j<4; j++)
  {
    char const *coeff_level_names[] = {"p_coeff_level0", "p_coeff_level1", "p_coeff_level2", "p_coeff_level3"};
    for (i=0; i<16; i++)
    {
      pushAProbCtx(p_coeff_level[j][i], coeff_level_names[j], CTXT_CLASS_PBCS_SUMC|coeff_pbcf_classes[j]);
    }
  }
  for (i=0; i<NCTX_ZMV; i++)
    pushAProbCtx(p_mv_zero[i], "p_mv_zero", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_MVDZERO|CTXT_CLASS_PBCF_ZMV);
  // idACtx: xxx
  for (i=0; i<2; i++)
    pushAProbCtx(p_mv_level[i], "p_mv_level", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_MVD|CTXT_CLASS_PBCF_MVVAL);

#ifndef MVCOMP_ZERORES
  // 79 to 80
  for (i=0; i<2; i++)
    pushAProbCtx(p_mvp[i], "p_mvp", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_MVPDIR|CTXT_CLASS_PBCF_MVP);
#else
  // 79 to 81
  for (i=0; i<3; i++)
  {
    int class_type = CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_MVPDIR|CTXT_CLASS_PBCF_MVP;
#ifndef MVP_CTX_READ
    if (i == 2)
      class_type |= CTXT_CLASS_NORW;
#endif // MVP_CTX_READ
    pushAProbCtx(p_mvp[i], "p_mvp", class_type);
  }
#endif
#if 0 // was pushed twice!
  // idACtx: xxx
  for (i=0; i<2; i++)
    pushAProbCtx(p_mvp[i], "p_mvp", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_MVPDIR|CTXT_CLASS_PBCF_MVP);
#endif

  // idACtx: unused
  pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);  // was intra_blk.

  // idACtx: xxx
  pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);   

  // idACtx: xxx
  for (j=0; j<LISTS; j++) 
  {
    for (i=0; i<3; i++)
      pushAProbCtx(p_refidx_lsb0[j][i], "p_refidx_lsb0", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_REFINDEX_LSB|CTXT_CLASS_PBCF_REFINDEX);
    for (i=0; i<3; i++)
      pushAProbCtx(p_refidx_lsb1[j][i], "p_refidx_lsb1", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_REFINDEX_LSB|CTXT_CLASS_PBCF_REFINDEX);
    for (i=0; i<3; i++)
      pushAProbCtx(p_refidx_msb[j][i], "p_refidx_msb", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_REFINDEX_MSB|CTXT_CLASS_PBCF_REFINDEX);
  }

  // idACtx: xxx
  for (i=0; i<NCTX_SCALE; i++)
    pushAProbCtx(p_scale[i], "p_scale", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTCOMP|CTXT_CLASS_PBCF_IC);
  // idACtx: xxx
  for (i=0; i<2; i++)
    pushAProbCtx(p_offset[i], "p_offset", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTCOMP|CTXT_CLASS_PBCF_IC);
  // idACtx: xxx
  for (i=0; i<3; i++)
    pushAProbCtx(p_mode[i], "p_mode", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_PMODE);
  // idACtx: xxx
  for (i=0; i<3; i++)
    pushAProbCtx(p_mode_P1[i], "p_mode_P1", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_PMODE);
  // idACtx: xxx
  for (i=0; i<3; i++)
    pushAProbCtx(p_mode_P2[i], "p_mode_P2", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_PMODE);

  // The first 7 ztrees (0..6) are pushed here, the others (if necessary) are pushed later.  This currently
  // reflects the old .dat order.
  for (i=0; i<NZTREES-NSUBZTREES; i++) 
  {
    char const *ztree_names[] = {"p_ztree0", "p_ztree1", "p_ztree2", "p_ztree3",
                                 "p_ztree4", "p_ztree5", "p_ztree6", "p_ztree7",
                                 "p_ztree8", "p_ztree9", "p_ztree10", "p_ztree11"};
    for (j=0; j<param.ztree[i]->nleaves-1; j++) 
    {
      for (k=0; k<2; k++) 
      {
        unsigned int ctxt_class = ztree_pbcf_classes[i];
        if (i < 4)
          ctxt_class |= CTXT_CLASS_PBCS_SUMB;
        else if (i < 6)
          ctxt_class |= CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_MVDZERO;
        else
          ctxt_class = 0;
#ifdef FRAME_SUBSTREAMS
        pushAProbCtx(param.ztree[i]->get_pe(this)[2*j+k], ztree_names[i], ctxt_class);
#else
        pushAProbCtx(param.ztree[i]->pe[2*j+k], ztree_names[i], ctxt_class);
#endif
      }
    }
  }

  // idACtx: xxx
  pushAProbCtx(p_prev_intra_pred_mode_flag, "p_prev_intra_pred_mode_flag", CTXT_CLASS_PBCF_SIPLUMA);
  // idACtx: xxx
  for (i = 0; i < 5; i++) 
  {
    pushAProbCtx(p_rem_intra_pred_mode_flag[i], "p_rem_intra_pred_mode_flag", CTXT_CLASS_PBCF_SIPLUMA);
  }
  // idACtx: xxx
  for (i = 0; i < 3; i++) 
  {
    pushAProbCtx(p_intra_chroma_pred_mode[i], "p_intra_chroma_pred_mode", CTXT_CLASS_PBCF_SIPCHROMA);
  }
  // idACtx: xxx
  pushAProbCtx(p_intra_chroma_pred_mode_1, "p_intra_chroma_pred_mode_1", CTXT_CLASS_PBCF_SIPCHROMA);
  // idACtx: xxx
  pushAProbCtx(p_intra_chroma_pred_mode_2, "p_intra_chroma_pred_mode_2", CTXT_CLASS_PBCF_SIPCHROMA);

  // idACtx: xxx
  if (param.aif) 
  {
    pushAProbCtx(p_AIF_use_all_subpel_positions, "p_AIF_use_all_subpel_positions", CTXT_CLASS_PBCF_AIF);
    pushAProbCtx(p_AIF_subpel_positions_pattern_zero, "p_AIF_subpel_positions_pattern_zero", CTXT_CLASS_PBCF_AIF);
    pushAProbCtx(p_AIF_subpel_positions_pattern, "p_AIF_subpel_positions_pattern", CTXT_CLASS_PBCF_AIF);
    pushAProbCtx(p_AIF_filter_pattern, "p_AIF_filter_pattern", CTXT_CLASS_PBCF_AIF);
    pushAProbCtx(p_AIF_diff_qfilter_coef_zero, "p_AIF_diff_qfilter_coef_zero", CTXT_CLASS_PBCF_AIF);
    pushAProbCtx(p_AIF_diff_qfilter_coef, "p_AIF_diff_qfilter_coef", CTXT_CLASS_PBCF_AIF);
    // AIF mode flags are pushed later.
  }
  else
  {
    for (int i = 0; i < 6; i++)
      pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);
  }
  pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);
  // NLT probas are still in the prob.dat for READ_OLD_DAT.
  pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);
  pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);

  // idACtx: xxx
  for (i=0; i<3; i++)
    pushAProbCtx(p_mode_TP1[i], "p_mode_TP1", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_PMODE);
  // idACtx: xxx
  for (i=0; i<3; i++)
    pushAProbCtx(p_mode_TP2[i], "p_mode_TP2", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_PMODE);
  // idACtx: xxx
  for (i=0; i<3; i++)
    pushAProbCtx(p_mode_TB1[i], "p_mode_TB1", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_PMODE);
  // idACtx: xxx
  for (i=0; i<3; i++)
    pushAProbCtx(p_mode_TB2[i], "p_mode_TB2", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_PMODE);

  // idACtx: xxx
  for (i=0; i<3; i++)
    pushAProbCtx(p_mvshare_type_zero[i], "p_mvshare_type_zero", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_MVSHARE);
  // idACtx: xxx
  pushAProbCtx(p_mvshare_pred, "p_mvshare_pred", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_INTRA_BLOCK|CTXT_CLASS_PBCF_MVSHARE);

  // idACtx: xxx
  if (param.wienerfilter)
  {
    pushAProbCtx(p_WLF, "p_WLF", CTXT_CLASS_PBCF_WLF);
    pushAProbCtx(p_WLF_coef_zero, "p_WLF_coef_zero", CTXT_CLASS_PBCF_WLF);
    pushAProbCtx(p_WLF_coef, "p_WLF_coef", CTXT_CLASS_PBCF_WLF);
  }
  else
  {
    for (i = 0; i < 3; i++)
      pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);
  }

  // idACtx: xxx
  if (param.deblockingfilter)
  {
    pushAProbCtx(p_Dummy1, "p_dummy", CTXT_CLASS_PBCF_DUMMY1);
    pushAProbCtx(p_Dummy2, "p_dummy", CTXT_CLASS_PBCF_DUMMY1);
    pushAProbCtx(p_Dummy3, "p_dummy", CTXT_CLASS_PBCF_DUMMY1);
  }
  else
  {
    for (i = 0; i < 3; i++)
      pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);
  }

  for (i=0; i<NCTX_SIP_TYPE; i++)
  {
    pushAProbCtx(p_sip_type_b1[i], "p_sip_type_b1", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBCF_SIP); 
    pushAProbCtx(p_sip_type_b2[i], "p_sip_type_b2", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBCF_SIP); 
    pushAProbCtx(p_sip_type_b3[i], "p_sip_type_b3", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBCF_SIP); 
  int i16x16 = param.intra16x16_org;
    if (i16x16)
#ifdef SIP_CTX_READ
      pushAProbCtx(p_sip_type_b4[i], "p_sip_type_b4", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBCF_SIP);
#else
      pushAProbCtx(p_sip_type_b4[i], "p_sip_type_b4", CTXT_CLASS_NORW|CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBCF_SIP);
#endif //SIP_CTX_READ
  }

  pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);

    pushAProbCtx(p_DUMMY, "p_dummy", CTXT_CLASS_DUMMY);

  // intra16 ztrees.
  int i16x16 = param.intra16x16_org;
  for (i=NZTREES-NSUBZTREES; i<(i16x16 ? NZTREES : NZTREES-NSUBZTREES); i++) 
  {
    char const *ztree_names[] = {"p_ztree0", "p_ztree1", "p_ztree2", "p_ztree3",
                                 "p_ztree4", "p_ztree5", "p_ztree6", "p_ztree7",
                                 "p_ztree8", "p_ztree9", "p_ztree10", "p_ztree11"};
    for (j=0; j<param.ztree[i]->nleaves-1; j++) 
    {
      for (k=0; k<2; k++) 
      {
#ifdef STATS_16x16_FIX
        unsigned int ctxt_class = CTXT_CLASS_PBCS_SUMB | CTXT_CLASS_PBCF_ZTREE9;
#else
        unsigned int ctxt_class = ztree_pbcf_classes[i];
        if (i < 4)
          ctxt_class |= CTXT_CLASS_PBCS_SUMB;
        else if (i < 6)
          ctxt_class |= CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_MVDZERO;
        else
          ctxt_class = 0;
#endif // STATS_16x16_FIX
#ifdef FRAME_SUBSTREAMS
        pushAProbCtx(param.ztree[i]->get_pe(this)[2*j+k], ztree_names[i], ctxt_class);
#else
        pushAProbCtx(param.ztree[i]->pe[2*j+k], ztree_names[i], ctxt_class);
#endif
      }
    }
  }
  pushAProbCtx(p_AIF_mode_flag1, "p_AIF_mode_flag1", CTXT_CLASS_PBCF_AIF);
  pushAProbCtx(p_AIF_mode_flag2, "p_AIF_mode_flag2", CTXT_CLASS_PBCF_AIF);

  if (param.trans8x8_enable) 
  {
    for (i=0; i<ZTsize64; i++)
      pushAProbCtx(p_coeff_level_64[i], "p_coeff_level_64", CTXT_CLASS_8X8|CTXT_CLASS_PBCS_SUMC|CTXT_CLASS_PBCF_COEFF4);

    for (j=0; j<param.ztree_64->nleaves-1; j++) 
    {
      for (k=0; k<2; k++) 
      {
#ifdef FRAME_SUBSTREAMS
        pushAProbCtx(param.ztree_64->get_pe(this)[2*j+k], "p_ztree_64", CTXT_CLASS_8X8|CTXT_CLASS_PBCS_SUMB|CTXT_CLASS_PBCF_ZTREE7);
#else
        pushAProbCtx(param.ztree_64->pe[2*j+k], "p_ztree_64", CTXT_CLASS_8X8|CTXT_CLASS_PBCS_SUMB|CTXT_CLASS_PBCF_ZTREE7);
#endif
      }
    } 
    if (param.chroma8x8)
    {
      for (i=0; i<ZTsize64; i++)
          pushAProbCtx(p_coeff_level_64_CT8[i], "p_coeff_level_64_CT8", CTXT_CLASS_8X8|CTXT_CLASS_PBCS_SUMC|CTXT_CLASS_PBCF_COEFF5);
      for (j=0; j<param.ztree_64_CT8->nleaves-1; j++) 
      {
        for (k=0; k<2; k++) 
#ifdef ZTREE_64_CT8_CTX_READ
#ifdef FRAME_SUBSTREAMS
          pushAProbCtx(param.ztree_64_CT8->get_pe(this)[2*j+k], "p_ztree_64_CT8", CTXT_CLASS_8X8|CTXT_CLASS_PBCS_SUMB|CTXT_CLASS_PBCF_ZTREE8);
#else
          pushAProbCtx(param.ztree_64_CT8->pe[2*j+k], "p_ztree_64_CT8", CTXT_CLASS_8X8|CTXT_CLASS_PBCS_SUMB|CTXT_CLASS_PBCF_ZTREE8);
#endif
#else
#ifdef FRAME_SUBSTREAMS
          pushAProbCtx(param.ztree_64_CT8->get_pe(this)[2*j+k], "p_ztree_64_CT8", CTXT_CLASS_NORW|CTXT_CLASS_8X8|CTXT_CLASS_PBCS_SUMB|CTXT_CLASS_PBCF_ZTREE8);
#else
          pushAProbCtx(param.ztree_64_CT8->pe[2*j+k], "p_ztree_64_CT8", CTXT_CLASS_NORW|CTXT_CLASS_8X8|CTXT_CLASS_PBCS_SUMB|CTXT_CLASS_PBCF_ZTREE8);
#endif
#endif //ZTREE_64_CT8_CTX_READ
      }
    }
  }

  // idFCtx: 0
  pushFProbCtx(p_inter_type, "p_inter_type", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_mvshare_four, "p_mvshare_four", CTXT_CLASS_PBCF_MVSHARE);
  pushFProbCtx(p_AIF_subpel_positions_pattern_suffix, "p_AIF_subpel_positions_pattern_suffix", CTXT_CLASS_PBCF_AIF);
  pushFProbCtx(p_AIF_filter_pattern_suffix, "p_AIF_filter_pattern_suffix", CTXT_CLASS_PBCF_AIF);
  pushFProbCtx(p_AIF_diff_qfilter_coef_sign, "p_AIF_diff_qfilter_coef_sign", CTXT_CLASS_PBCF_AIF);
  
  pushFProbCtx(p_WLF_suffix, "p_WLF_suffix", CTXT_CLASS_PBCF_WLF);
  pushFProbCtx(p_WLF_coef_sign, "p_WLF_coef_sign", CTXT_CLASS_PBCF_WLF);

  pushFProbCtx(p_dummy, "p_dummy", 0);
  pushFProbCtx(p_dummy, "p_dummy", 0);

  pushFProbCtx(p_dummy, "p_dummy", 0);
  pushFProbCtx(p_tsfm_flag, "p_tsfm_flag", CTXT_CLASS_PBCF_TSFM);

  for (int i = 0; i < 4; i++)
    pushFProbCtx(p_coeff_sign[i], "p_coeff_sign", CTXT_CLASS_PBCS_SUMC|coeff_pbcf_classes[i]);
  for (int i = 0; i < 4; i++)
    pushFProbCtx(p_coeff_suffix[i], "p_coeff_suffix", coeff_pbcf_classes[i]);
  pushFProbCtx(p_mv_suffix, "p_mv_suffix", CTXT_CLASS_PBCF_MVVAL);
  pushFProbCtx(p_coeff_suffix_64, "p_coeff_suffix_64", CTXT_CLASS_PBCF_COEFF4);
  pushFProbCtx(p_coeff_sign_64, "p_coeff_sign_64", CTXT_CLASS_PBCF_COEFF4);
  pushFProbCtx(p_coeff_suffix_64_CT8, "p_coeff_suffix_64_CT8", CTXT_CLASS_PBCF_COEFF5);
  pushFProbCtx(p_coeff_sign_64_CT8, "p_coeff_sign_64_CT8", CTXT_CLASS_PBCF_COEFF5);
  pushFProbCtx(p_offset_suffix, "p_offset_suffix", CTXT_CLASS_PBCF_IC);
  pushFProbCtx(p_scale_suffix, "p_scale_suffix", CTXT_CLASS_PBCF_IC);
  pushFProbCtx(p_offset_sign, "p_offset_sign", CTXT_CLASS_PBCF_IC);
  pushFProbCtx(p_scale_sign, "p_scale_sign", CTXT_CLASS_PBCF_IC);
  pushFProbCtx(p_mv_sign, "p_mv_sign", CTXT_CLASS_PBCS_SUMD|CTXT_CLASS_PBC_MVDSIGN|CTXT_CLASS_PBCF_MVVAL);
  pushFProbCtx(p_frame_type, "p_frame_type", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_reference_idc, "p_reference_idc", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_frm_zero, "p_frm_zero", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_framecount_sign, "p_framecount_sign", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_framecount, "p_framecount", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_num_of_refs, "p_num_of_refs", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_dqp_zero, "p_dqp_zero", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_dqp_sign, "p_dqp_sign", CTXT_CLASS_PBCF_HEADER);
  pushFProbCtx(p_deltaqp, "p_deltaqp", CTXT_CLASS_PBCF_HEADER);
#ifdef MVCOMP_ZERORES
  pushFProbCtx(p_mvcomp_config, "p_mvcomp_config", CTXT_CLASS_PBCF_HEADER);
#endif

#ifdef Q_OFFSET
  for (int i = 0; i < 2; i++)
    pushFProbCtx(p_q_offset_zero[i], "p_q_offset_zero", 0);
  for (int i = 0; i < 2; i++)
    pushFProbCtx(p_q_offset8x8_zero[i], "p_q_offset8x8_zero", 0);
  for (int i = 0; i < 2; i++)
    pushFProbCtx(p_q_offset_sign[i], "p_q_offset_sign", 0);
  for (int i = 0; i < 2; i++)
    pushFProbCtx(p_q_offset8x8_sign[i], "p_q_offset8x8_sign", 0);
  for (int i = 0; i < 2; i++)
    pushFProbCtx(p_q_offset[i], "p_q_offset", 0);
  for (int i = 0; i < 2; i++)
    pushFProbCtx(p_q_offset8x8[i], "p_q_offset8x8", 0);
#endif

  pushFProbCtx(quantization_set_id_value_zero, "quantization_set_id_value_zero", CTXT_CLASS_PBCF_QMATRIX); // cq_start
  pushFProbCtx(quantization_set_id_value, "quantization_set_id_value", CTXT_CLASS_PBCF_QMATRIX);
  for (int i = 0; i < 2; i++)
  {
    pushFProbCtx(qmatrix_overwrite_flag[i], "qmatrix_overwrite_flag[i]", CTXT_CLASS_PBCF_QMATRIX);
    pushFProbCtx(qmatrix_value_zero[i], "qmatrix_value_zero[i]", CTXT_CLASS_PBCF_QMATRIX);
    pushFProbCtx(qmatrix_value_sign[i], "qmatrix_value_sign[i]", CTXT_CLASS_PBCF_QMATRIX);
    pushFProbCtx(qmatrix_value[i], "qmatrix_value[i]", CTXT_CLASS_PBCF_QMATRIX);
    pushFProbCtx(qmatrix_list_present_flag[i], "qmatrix_list_present_flag[i]", CTXT_CLASS_PBCF_QMATRIX);
  } // cq_end

  pushFProbCtx(p_rpps_type, "p_rpps_type", CTXT_CLASS_PBCF_RPPS);
  pushFProbCtx(p_rpps_id, "p_rpps_id", CTXT_CLASS_PBCF_RPPS);
  pushFProbCtx(p_filter_type, "p_filter_type", CTXT_CLASS_PBCF_RPPS);
  pushFProbCtx(p_ic_disable, "p_ic_disable", CTXT_CLASS_PBCF_RPPS);
  pushFProbCtx(p_round_type, "p_round_type", CTXT_CLASS_PBCF_RPPS);

  pushFProbCtx(p_edge_prediction_flag, "p_edge_prediction_flag", 0);
  pushFProbCtx(p_threshold_edge, "p_threshold_edge", 0);

  // Copy default (initial) probas from param.
  if (param.initialProbEntropyModel != NULL)
    copyInitialProbasFrom(param.initialProbEntropyModel);
}

//-----------------------------------------------------------------------------
/*!
* \brief copy initial proba settings of all adaptive contexts.
*
* \param[in] cs The source EntropyModel from which to copy.
*/
//-----------------------------------------------------------------------------
void EntropyModel::copyInitialProbasFrom(EntropyModel *cs)
{
  for (int c = 0; c <= max_idACtx; c++)
    p_aCtx[c]->copyInitialProba(cs->p_aCtx[c]);
}

//-----------------------------------------------------------------------------
/*!
* \brief print some basic proba info for adaptive contexts.
*
* \param[in] msg Optional message to prepend to the output.
*/
//-----------------------------------------------------------------------------
void EntropyModel::printProbas(char const *msg)
{
  static int cnt;
  printf("adaptive %s %d: %d contexts\n", msg != NULL ? msg : "", cnt++, max_idACtx);
  for (int c = 0; c <= max_idACtx; c++)
    printf("%d(%s): p=%d i_p=%d\n", c, p_aCtx[c]->idName, p_aCtx[c]->p, p_aCtx[c]->i_p);
  fflush(stdout);
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*/
//-----------------------------------------------------------------------------
EntropyModel::~EntropyModel() 
{
#ifdef CABAC_SYNC_POINTS
    if (synchro_state_allocated)
        delete synchro_state;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Save pointer to a FixedProbEst with the specified context number.
*
* A pointer to the ProbEst is saved, and the ProbEst's external data is set to NULL.
*
* \param[in] p The base object of a FixedProbEst
* \param[in] idCtx A context identifier
* \param[in] name Associated name
* \param[in] classes Classification bits
*/
//-----------------------------------------------------------------------------
void EntropyModel::storeFProbCtx(ProbEst &p, int idCtx, char const *name, unsigned int classes)
{
  p.idCtx = idCtx;
  p.idName = name;
  p.idClasses = classes|CTXT_CLASS_ALL;
  if (idCtx >= MAX_ID_CTX)
    abort();
  if (p_fCtx[idCtx] != 0)
  {
    printf("Already used idCtx!!\n"); abort();
  }
  p_fCtx[idCtx] = &p;
#ifdef CABAC_SYNC_POINTS
  p.setExternalData(0, 0);
#endif
  if (idCtx > max_idFCtx)
    max_idFCtx = idCtx;
}

//-----------------------------------------------------------------------------
/*!
* \brief Save pointer to a SimpleProbEst with the specified context number.
*
* A pointer to the SimpleProbEst is saved, and the SimpleProbEst's external data
* pointers are initialized to point into our synchro_state object.
*
* \param[in] p The SimpleProbEst address to save
* \param[in] idCtx A context identifier
* \param[in] name Associated name
* \param[in] classes Classification bits
*/
//-----------------------------------------------------------------------------
void EntropyModel::storeAProbCtx(SimpleProbEst &p, int idCtx, char const *name, unsigned int classes)
{
  p.idCtx = idCtx;
  p.idName = name;
  p.idClasses = classes|CTXT_CLASS_ALL;
  if (idCtx >= MAX_ID_CTX)
    abort();
  if (p_aCtx[idCtx] != 0)
  {
    printf("Already used idCtx!!\n"); abort();
  }
  p_aCtx[idCtx] = &p;
#ifdef CABAC_SYNC_POINTS
  p.setExternalData(synchro_state->p2d[idCtx], synchro_state->mnb[idCtx]);
#endif
  if (idCtx > max_idACtx)
    max_idACtx = idCtx;
}

//-----------------------------------------------------------------------------
/*!
* \brief Save pointer to a UnaryProbEst with the specified context number.
*
* A pointer to the UnaryProbEst is saved, and its external data
* pointers are initialized to point into our synchro_state object.
*
* \param[in] p The UnaryProbEst address to save
* \param[in] idCtx A context identifier
* \param[in] name Associated name
* \param[in] classes Classification bits
*/
//-----------------------------------------------------------------------------
void EntropyModel::storeAProbCtx(UnaryProbEst &p, int idCtx, char const *name, unsigned int classes)
{
  p.idCtx = idCtx;
  p.idName = name;
  p.idClasses = classes|CTXT_CLASS_ALL;
  if (idCtx >= MAX_ID_CTX)
    abort();
  if (p_aCtx[idCtx] != 0)
  {
    printf("Already used idCtx!!\n"); abort();
  }
  p_aCtx[idCtx] = &p;
#ifdef CABAC_SYNC_POINTS
  p.setExternalData(synchro_state->p2d[idCtx], synchro_state->mnb[idCtx], synchro_state->s2d[idCtx]);
#endif
  if (idCtx > max_idACtx)
    max_idACtx = idCtx;
}

#ifdef CABAC_SYNC_POINTS
//-----------------------------------------------------------------------------
/*!
* \brief Apply resetSynchro to all saved adaptive contexts.
*
* This is called at the start of a frame.
*/
//-----------------------------------------------------------------------------
void EntropyModel::resetSynchro()
{
  for(int i=0; i<MAX_ID_CTX; i++)
    if (p_aCtx[i] != NULL && (p_aCtx[i]->idClasses&CTXT_CLASS_DUMMY) == 0)
      p_aCtx[i]->resetSynchro();
}

//-----------------------------------------------------------------------------
/*!
* \brief Apply initSynchro to all saved adaptive contexts.
*
* This is called prior to the processing of each superblock.
*/
//-----------------------------------------------------------------------------
void EntropyModel::initSynchro(int iSB, int jSB)
{
  (void)iSB; // unused parameter.
  for(int i=0; i<MAX_ID_CTX; i++)
  {
    if (p_aCtx[i] != NULL && (p_aCtx[i]->idClasses&CTXT_CLASS_DUMMY) == 0)
      p_aCtx[i]->initSynchro(jSB);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Apply updateSynchro to all saved adaptive contexts.
*
* This is called as the final step of processing each superblock.
*/
//-----------------------------------------------------------------------------
void EntropyModel::updateSynchro(int iSB)
{
  for(int i=0; i<MAX_ID_CTX; i++)
    if (p_aCtx[i] != NULL && (p_aCtx[i]->idClasses&CTXT_CLASS_DUMMY) == 0)
      p_aCtx[i]->updateSynchro(iSB);
}

#endif // CABAC_SYNC_POINTS

//-----------------------------------------------------------------------------
/*!
* \brief Print summary of bit usage
*/
//-----------------------------------------------------------------------------
void EntropyModel::print_summary() 
{
  int i;
  double sum_all;
  FILE *f;

  f = stdout;

  sum_all = 0.0;
  for (i=0; i<4; i++)
    sum_all += bstat[i];

  fprintf(f, "%9.2f (%5.2f%%) ", bstat[0], 100.0 * bstat[0] / sum_all);
  fprintf(f, "%9.2f (%5.2f%%) ", bstat[1], 100.0 * bstat[1] / sum_all);
  fprintf(f, "%9.2f (%5.2f%%) ", bstat[2], 100.0 * bstat[2] / sum_all);
  fprintf(f, "%9.2f (%5.2f%%)\n", bstat[3], 100.0 * bstat[3] / sum_all);

  fprintf(f, "Mode      :%9.2f (%5.2f%%)\n", bstat_intra_block,
    100.0 * bstat_intra_block / bstat[3]);
  fprintf(f, "RefIdxMSB: %9.2f (%5.2f%%)\n", bstat_refindex_msb,
    100.0 * bstat_refindex_msb / bstat[3]);
  fprintf(f, "RefIdxLSB: %9.2f (%5.2f%%)\n", bstat_refindex_lsb,
    100.0 * bstat_refindex_lsb / bstat[3]);
  fprintf(f, "Int. comp: %9.2f (%5.2f%%)\n", bstat_intcomp,
    100.0 * bstat_intcomp / bstat[3]);
  fprintf(f, "MVP dir. : %9.2f (%5.2f%%)\n", bstat_mvpdir,
    100.0 * bstat_mvpdir / bstat[3]);
  fprintf(f, "MVD sign : %9.2f (%5.2f%%)\n", bstat_mvdsign,
    100.0 * bstat_mvdsign / bstat[3]);
  fprintf(f, "MVD delta: %9.2f (%5.2f%%)\n", bstat_mvd,
    100.0 * bstat_mvd / bstat[3]);
  fprintf(f, "MVD zero : %9.2f (%5.2f%%)\n", bstat_mvdzero,
    100.0 * bstat_mvdzero / bstat[3]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Read initial probability settings from a file.
*
* \param[in] f The file from which to read.
* \param[in] p8x8 Whether we're reading CTXT_CLASS_8X8 probabilities or not.
*/
//-----------------------------------------------------------------------------
void EntropyModel::read_prob_file(FILE *f)
{
  for (int c = 0; c <= max_idACtx; c++)
  {
    if ((p_aCtx[c]->idClasses&(CTXT_CLASS_DUMMY|CTXT_CLASS_NORW)) != 0)
      continue;
    p_aCtx[c]->set_default_prob(param.read_prob(f, p_aCtx[c]->idName, param.qp));
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief A set of initial probabilities is read from a coded stream.
*
* The order in which probabilities are read is the same order in which
* they have been pushed in EntropyModel::EntropyModel.  Only non-dummy
* and non-NORW probabilities are read, so the set of probabilities depends
* on the tools that are activated.
*
* \param[in] buf The buffer from which to read probas.
*/
//-----------------------------------------------------------------------------
void EntropyModel::read_stream_probs(Buffer *buf)
{
  for (int c = 0; c <= max_idACtx; c++)
    if ((p_aCtx[c]->idClasses&(CTXT_CLASS_DUMMY|CTXT_CLASS_NORW)) == 0)
      p_aCtx[c]->set_default_prob(buf->read_bits(PROB_BITS));
}

//-----------------------------------------------------------------------------
/*!
* \brief A set of initial probabilities is written to a coded stream.
*
* The order in which probabilities are written is the same order in which
* they have been pushed in EntropyModel::EntropyModel.  Only non-dummy
* and non-NORW probabilities are written, so the set of probabilities depends
* on the tools that are activated.
*
* \param[in] buf The buffer to which to write probas.
*
* \return number of bits written.
*/
//-----------------------------------------------------------------------------
int EntropyModel::write_stream_probs(Buffer *buf)
{
  int r = 0;
  //printProbas("writing");

  for (int c = 0; c <= max_idACtx; c++)
    if ((p_aCtx[c]->idClasses&(CTXT_CLASS_DUMMY|CTXT_CLASS_NORW)) == 0)
      r += buf->write_bits(p_aCtx[c]->i_p, PROB_BITS);

  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Reset probability estimators
*/
//-----------------------------------------------------------------------------
#ifndef MVCOMP_ZERORES
void EntropyModel::reset() 
#else
void EntropyModel::reset(int mv_keep_movement)
#endif
{
#ifdef MVCOMP_ZERORES
#ifndef MVP_CTX_READ
  // temporal proba from current config.
  p_mvp[2].set_default_prob(0x10000-param.mv_cur_comp_config()->mv_temporal_proba);
#endif // MVP_CTX_READ
#endif
  for (int c = 0; c <= max_idACtx; c++)
      p_aCtx[c]->restoreDefaultState();

#ifdef MVCOMP_ZERORES
  if (mv_keep_movement)
    recalc_movement();
#endif
}

#ifdef MVCOMP_ZERORES
//-----------------------------------------------------------------------------
/*!
* \brief Recalculate mv_pred_dir stats from previous symbol counts if possible.
*/
//-----------------------------------------------------------------------------
void EntropyModel::recalc_movement()
{
  for (int i=0; i<3; i++)
  {
    // Fabricate our P from the symbol counts of the previous image.
    // TRACE_STATS has to be enabled.
    int cnt0 = p_mvp[i].symcount[0];
    int cnt1 = p_mvp[i].symcount[1];
    if (cnt0+cnt1 > 9) // Arbitrary value to ensure we've coded a symbol, and to
                       // ensure we've coded a 'few' to get at least something
                       // resembling a stat.  Otherwise we use the initial proba.
    {
        double p = (double)cnt0/(cnt0+cnt1);
        int r = (int)(power2(PROB_BITS) * p + 0.5);
        r = clip(r, 32, 65504);
        p_mvp[i].set_prob(r);
        // printf("cnts: %d, %d\n", cnt0, cnt1);
        fflush(stdout);
    }
    // Update from default initial value already done as part
    // of normal processing.
  }
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Reset statistics gathered by probability estimators
*/
//-----------------------------------------------------------------------------
void EntropyModel::reset_counts() 
{
  for (int c = 0; c <= max_idACtx; c++)
    p_aCtx[c]->reset_counts();
  for (int c = 0; c <= max_idFCtx; c++)
    p_fCtx[c]->reset_counts();

  // NOTE! this includes the previously unreset:
  // p_inter_type (if RPPS is defined, p_inter_type is not reset).
  // p_AIF_subpel_positions_pattern_suffix
  // p_AIF_filter_pattern_suffix
  // p_AIF_diff_qfilter_coef_sign
}

#ifdef VERBOSE_STATS
//-----------------------------------------------------------------------------
/*!
* \brief print pretty graph of summary by category
*
* \param[in,out] f File pointer
* \param[in] caption string of caption
* \param[in] sum_val sum of bits belonging to the category 
* \param[in] sum_val sum of all the bits 
*/
//-----------------------------------------------------------------------------
void EntropyModel::print_bar(FILE *f, const char *caption, double sum_val, double sum_all) 
{
  int i;

  fprintf(f, "%s  %9.2f (%5.2f%%)|", caption, sum_val, 100.0 * sum_val/sum_all); 
  for (i=0; i<int(100.0 * sum_val/sum_all + 0.5); i++) 
    printf("*"); 
  printf("\n");
}

#ifdef FRAME_SUBSTREAMS
//-----------------------------------------------------------------------------
/*!
* \brief Sum bit usage counts from one substream to another.
*
* \param[in] src
*/
//-----------------------------------------------------------------------------
void EntropyModel::sum_stats_from(EntropyModel *src)
{
  for (int p = 0; p <= max_idACtx; p++)
  {
    p_aCtx[p]->fbitcount += src->p_aCtx[p]->fbitcount;
    p_aCtx[p]->symcount[0] += src->p_aCtx[p]->symcount[0];
    p_aCtx[p]->symcount[1] += src->p_aCtx[p]->symcount[1];
  }
  for (int p = 0; p <= max_idFCtx; p++)
  {
    p_fCtx[p]->fbitcount += src->p_fCtx[p]->fbitcount;
    p_fCtx[p]->symcount[0] += src->p_fCtx[p]->symcount[0];
    p_fCtx[p]->symcount[1] += src->p_fCtx[p]->symcount[1];
  }
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Accumulate bit usage counts and print to file summary by category
*
* \param[in,out] f File pointer
*/
//-----------------------------------------------------------------------------
void EntropyModel::print_by_category_full(FILE *f) 
{
  double sum_all = 0.0;
#define CLASS_TO_CTR(val) (((val)&CTXT_CLASS_PBCF_MASK)>>CTXT_CLASS_PBCF_RSHIFT)
  double sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_LASTCLASS)+1];

  for (int i = 0; i <= CLASS_TO_CTR(CTXT_CLASS_PBCF_LASTCLASS); i++)
    sums[i] = 0.0;

  for (int typ = 0; typ < 2; typ++)
  for (int c = 0; c <= (typ == 0 ? max_idACtx : max_idFCtx); c++)
  {
    ProbEst *p = (typ == 0) ? p_aCtx[c] : p_fCtx[c];
    // counters are exclusive.
    int ctr = CLASS_TO_CTR(p->idClasses&CTXT_CLASS_PBCF_MASK);
    sums[ctr] += p->fbitcount;
    sum_all += p->fbitcount;
  }

  fprintf(f, "------------------------------------------------------------------------------\n");
  print_bar(f, "header bits            ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_HEADER)], sum_all);
  print_bar(f, "zblk intra luma        ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZBLK0)], sum_all);
  print_bar(f, "zblk intra chroma      ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZBLK1)], sum_all);
  print_bar(f, "zblk inter luma        ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZBLK2)], sum_all);
  print_bar(f, "zblk inter chroma      ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZBLK3)], sum_all);
  print_bar(f, "ztree intra luma       ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE0)], sum_all);
  print_bar(f, "ztree intra chroma     ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE1)], sum_all);
  print_bar(f, "ztree inter luma       ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE2)], sum_all);
  print_bar(f, "ztree inter chroma     ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE3)], sum_all);
  print_bar(f, "ztree inter luma 64    ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE7)], sum_all);
  if (param.chroma8x8)
    print_bar(f, "ztree intra chroma 64  ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE8)], sum_all);
#ifdef STATS_16x16_FIX
  if (param.intra16x16)
    print_bar(f, "ztree intra luma 16x16 ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE9)], sum_all);
#endif // STATS_16x16_FIX
  print_bar(f, "coeff intra luma       ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_COEFF0)], sum_all);
  print_bar(f, "coeff intra chroma     ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_COEFF1)], sum_all);
  print_bar(f, "coeff inter luma       ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_COEFF2)], sum_all);
  print_bar(f, "coeff inter chroma     ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_COEFF3)], sum_all);
  print_bar(f, "coeff inter luma 64    ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_COEFF4)], sum_all);
  if (param.chroma8x8)
    print_bar(f, "coeff intra chroma 64  ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_COEFF5)], sum_all);
  print_bar(f, "zmv                    ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZMV)], sum_all);
  print_bar(f, "ztree mv pred          ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE4)], sum_all);
  print_bar(f, "ztree mv bi-pred       ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE5)], sum_all);
  print_bar(f, "ztree mv 4x4           ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_ZTREE6)], sum_all);
  print_bar(f, "mv value               ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_MVVAL)], sum_all);
  print_bar(f, "mv pred bits           ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_MVP)], sum_all);
  print_bar(f, "mvshare_type_zero      ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_MVSHARE)], sum_all);
  print_bar(f, "reference index        ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_REFINDEX)], sum_all);
  print_bar(f, "prediction mode        ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_PMODE)], sum_all);
  print_bar(f, "ic (offset & scale)    ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_IC)], sum_all);
  print_bar(f, "sip type               ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_SIP)], sum_all);
  print_bar(f, "sip mode luma          ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_SIPLUMA)], sum_all);
  print_bar(f, "sip mode chroma        ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_SIPCHROMA)], sum_all);
  print_bar(f, "AIF info               ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_AIF)], sum_all);
  print_bar(f, "RPPS info              ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_RPPS)], sum_all);
  print_bar(f, "Wiener Filter          ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_WLF)], sum_all);
  print_bar(f, "q matrix               ", sums[CLASS_TO_CTR(CTXT_CLASS_PBCF_QMATRIX)], sum_all); // cq_start, cq_end
  print_bar(f, "uncounted bits         ", sums[0], sum_all);

  fprintf(f, "------------------------------------------------------------------------------\n");
  fprintf(f, "Total bits           %9.2f\n",  sum_all);
}

#endif

//-----------------------------------------------------------------------------
/*!
* \brief Accumulate bit usage counts and print to file summary by category
*
*/
//-----------------------------------------------------------------------------
void EntropyModel::print_by_category() 
{
  double sumA = 0.0, sumB = 0.0, sumC = 0.0, sumD = 0.0, sum_all;
  for (int typ = 0; typ < 2; typ++)
  for (int c = 0; c <= (typ == 0 ? max_idACtx : max_idFCtx); c++)
  {
    ProbEst *p = (typ == 0) ? p_aCtx[c] : p_fCtx[c];
    if ((p->idClasses&CTXT_CLASS_PBCS_MASK) == CTXT_CLASS_PBCS_SUMA)
      sumA += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBCS_MASK) == CTXT_CLASS_PBCS_SUMB)
      sumB += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBCS_MASK) == CTXT_CLASS_PBCS_SUMC)
      sumC += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBCS_MASK) == CTXT_CLASS_PBCS_SUMD)
      sumD += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBC_MASK) == CTXT_CLASS_PBC_REFINDEX_LSB)
      bstat_refindex_lsb += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBC_MASK) == CTXT_CLASS_PBC_REFINDEX_MSB)
      bstat_refindex_msb += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBC_MASK) == CTXT_CLASS_PBC_INTRA_BLOCK)
      bstat_intra_block += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBC_MASK) == CTXT_CLASS_PBC_INTCOMP)
      bstat_intcomp += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBC_MASK) == CTXT_CLASS_PBC_MVPDIR)
      bstat_mvpdir += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBC_MASK) == CTXT_CLASS_PBC_MVDSIGN)
      bstat_mvdsign += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBC_MASK) == CTXT_CLASS_PBC_MVDZERO)
      bstat_mvdzero += p->fbitcount;
    if ((p->idClasses&CTXT_CLASS_PBC_MASK) == CTXT_CLASS_PBC_MVD)
      bstat_mvd += p->fbitcount;
  }

  sum_all = sumA + sumB + sumC + sumD;

  bstat[0] += sumA;
  bstat[1] += sumB;
  bstat[2] += sumC;
  bstat[3] += sumD;
}

//-----------------------------------------------------------------------------
/*!
* \brief Print to file statistics for each estimator
*
* \param[in,out] f File pointer
*/
//-----------------------------------------------------------------------------
void EntropyModel::print_stat(FILE *f) 
{
  // We write all stats.
  // We don't write dummy entries as lines are now prefixed with the ctxt name.
  for (int c = 0; c <= max_idACtx; c++)
    if ((p_aCtx[c]->idClasses&CTXT_CLASS_DUMMY) == 0)
      p_aCtx[c]->print(f);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute entropy based on statistics of estimators
*
* \return Entropy
*/
//-----------------------------------------------------------------------------
double EntropyModel::entropy() 
{
  double r = 0.0;
  for (int c = 0; c <= max_idACtx; c++)
    r += p_aCtx[c]->entropy();
  for (int c = 0; c <= max_idFCtx; c++)
    r += p_fCtx[c]->entropy();

  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy the coding state of the Entropy Model
*
* \param[in] eenc The EntropyModel to copy from
*/
//-----------------------------------------------------------------------------
void EntropyModel::copy_coding_state(EntropyModel  *cs)
#ifdef FAST_STATE_COPY
{
  int i,j;
  
  memcpy(this->bstat, cs->bstat, 4 * sizeof (double));
  this->bstat_intra_block  = cs->bstat_intra_block;
  this->bstat_refindex_lsb = cs->bstat_refindex_lsb;
  this->bstat_refindex_msb = cs->bstat_refindex_msb;
  this->bstat_intcomp      = cs->bstat_intcomp;
  this->bstat_mvpdir       = cs->bstat_mvpdir;
  this->bstat_mvdsign      = cs->bstat_mvdsign;
  this->bstat_mvdzero      = cs->bstat_mvdzero;
  this->bstat_mvd          = cs->bstat_mvd;
  
  memcpy(&this->quantization_set_id_value_zero,  &cs->quantization_set_id_value_zero,  sizeof (FixedProbEst));  // cq_start
  memcpy(&this->quantization_set_id_value,       &cs->quantization_set_id_value,       sizeof (FixedProbEst));  
  for (i = 0; i < 2; i++)
  {
    memcpy(&this->qmatrix_overwrite_flag[i],          &cs->qmatrix_overwrite_flag[i],          sizeof (FixedProbEst));
    memcpy(&this->qmatrix_value_zero[i],              &cs->qmatrix_value_zero[i],              sizeof (FixedProbEst));
    memcpy(&this->qmatrix_value_sign[i],              &cs->qmatrix_value_sign[i],              sizeof (FixedProbEst));
    memcpy(&this->qmatrix_value[i],                   &cs->qmatrix_value[i],                   sizeof (FixedProbEst));  
    memcpy(&this->qmatrix_list_present_flag[i],       &cs->qmatrix_list_present_flag[i],       sizeof (FixedProbEst));  
  } // cq_end

#ifndef MVCOMP_ZERORES
  for (j=0; j<2; j++) p_mvp[j].copy_est(cs->p_mvp[j]);
#else
  for (j=0; j<3; j++) p_mvp[j].copy_est(cs->p_mvp[j]);
#endif
  for (j=0; j<NCTX_ZMV; j++) p_mv_zero[j].copy_est(cs->p_mv_zero[j]);
  for (i = 0; i < LISTS; i++) 
  {
    for (j=0; j<3; j++) p_refidx_lsb0[i][j].copy_est(cs->p_refidx_lsb0[i][j]);
    for (j=0; j<3; j++) p_refidx_lsb1[i][j].copy_est(cs->p_refidx_lsb1[i][j]);
    for (j=0; j<3; j++) p_refidx_msb [i][j].copy_est(cs->p_refidx_msb [i][j]);
  }
  for (i = 0; i < 4; i++) 
  {
    for (j=0; j<NCTX_ZBLK; j++) p_zblk[i][j].copy_est(cs->p_zblk[i][j]);
  }  
  p_intra_blk.copy_est(cs->p_intra_blk);
  for (j=0; j<NCTX_SIP_TYPE; j++)
  {
    p_sip_type_b1[j].copy_est(cs->p_sip_type_b1[j]);
    p_sip_type_b2[j].copy_est(cs->p_sip_type_b2[j]);
    p_sip_type_b3[j].copy_est(cs->p_sip_type_b3[j]);
    if (param.intra16x16)
      p_sip_type_b4[j].copy_est(cs->p_sip_type_b4[j]);
  }
  
  for (j=0; j<3; j++) p_mode[j].copy_est(cs->p_mode[j]);
  for (j=0; j<3; j++) p_mode_P1[j].copy_est(cs->p_mode_P1[j]);
  for (j=0; j<3; j++) p_mode_P2[j].copy_est(cs->p_mode_P2[j]);

  for (j=0; j<3; j++) p_mvshare_type_zero[j].copy_est(cs->p_mvshare_type_zero[j]);
  p_mvshare_pred.copy_est(cs->p_mvshare_pred);

  for (j=0; j<3; j++) p_mode_TP1[j].copy_est(cs->p_mode_TP1[j]);
  for (j=0; j<3; j++) p_mode_TP2[j].copy_est(cs->p_mode_TP2[j]);
  for (j=0; j<3; j++) p_mode_TB1[j].copy_est(cs->p_mode_TB1[j]);
  for (j=0; j<3; j++) p_mode_TB2[j].copy_est(cs->p_mode_TB2[j]);
  
  p_prev_intra_pred_mode_flag.copy_est(cs->p_prev_intra_pred_mode_flag);
  for (j=0; j<5; j++) p_rem_intra_pred_mode_flag[j].copy_est(cs->p_rem_intra_pred_mode_flag[j]);
  for (j=0; j<3; j++) p_intra_chroma_pred_mode[j].copy_est(cs->p_intra_chroma_pred_mode[j]);
  p_intra_chroma_pred_mode_1.copy_est(cs->p_intra_chroma_pred_mode_1);
  p_intra_chroma_pred_mode_2.copy_est(cs->p_intra_chroma_pred_mode_2);
  
  for (i = 0; i < 4; i++) 
  {
    for (j=0; j<16; j++) p_coeff_level[i][j].copy_est(cs->p_coeff_level[i][j]);
  }
  if (param.trans8x8_enable) 
  {
    for (j=0; j<ZTsize64; j++) p_coeff_level_64[j].copy_est(cs->p_coeff_level_64[j]);
    if (param.chroma8x8)
      for (j=0; j<ZTsize64; j++) p_coeff_level_64_CT8[j].copy_est(cs->p_coeff_level_64_CT8[j]);
  }
  
  for (j=0; j<2; j++) p_mv_level[j].copy_est(cs->p_mv_level[j]);
  for (j=0; j<NCTX_SCALE; j++) p_scale[j].copy_est(cs->p_scale[j]);
  for (j=0; j<2; j++) p_offset[j].copy_est(cs->p_offset[j]);
  
  p_WLF.copy_est(cs->p_WLF);
  p_WLF_coef.copy_est(cs->p_WLF_coef);
  p_WLF_coef_zero.copy_est(cs->p_WLF_coef_zero);
}
#else // FAST_STATE_COPY
{
  int i;

  memcpy(this->bstat, cs->bstat, 4 * sizeof (double));
  this->bstat_intra_block  = cs->bstat_intra_block;
  this->bstat_refindex_lsb = cs->bstat_refindex_lsb;
  this->bstat_refindex_msb = cs->bstat_refindex_msb;
  this->bstat_intcomp      = cs->bstat_intcomp;
  this->bstat_mvpdir       = cs->bstat_mvpdir;
  this->bstat_mvdsign      = cs->bstat_mvdsign;
  this->bstat_mvdzero      = cs->bstat_mvdzero;
  this->bstat_mvd          = cs->bstat_mvd;

  memcpy(&this->quantization_set_id_value_zero,  &cs->quantization_set_id_value_zero,  sizeof (FixedProbEst)); // cq_start 
  memcpy(&this->quantization_set_id_value,       &cs->quantization_set_id_value,       sizeof (FixedProbEst));  
  for (i = 0; i < 2; i++)
  {
    memcpy(&this->qmatrix_overwrite_flag[i],          &cs->qmatrix_overwrite_flag[i],          sizeof (FixedProbEst));
    memcpy(&this->qmatrix_value_zero[i],              &cs->qmatrix_value_zero[i],              sizeof (FixedProbEst));
    memcpy(&this->qmatrix_value_sign[i],              &cs->qmatrix_value_sign[i],              sizeof (FixedProbEst));
    memcpy(&this->qmatrix_value[i],                   &cs->qmatrix_value[i],                   sizeof (FixedProbEst));  
    memcpy(&this->qmatrix_list_present_flag[i],       &cs->qmatrix_list_present_flag[i],       sizeof (FixedProbEst));  
  } // cq_end

#ifndef MVCOMP_ZERORES
  memcpy(this->p_mvp,                         cs->p_mvp,                          2 * sizeof (SimpleProbEst));
#else
  memcpy(this->p_mvp,                         cs->p_mvp,                          3 * sizeof (SimpleProbEst));
#endif
  memcpy(this->p_mv_zero,                     cs->p_mv_zero,               NCTX_ZMV * sizeof (SimpleProbEst));
  for (i = 0; i < LISTS; i++) 
  {
    memcpy(this->p_refidx_lsb0[i],            cs->p_refidx_lsb0[i],               3 * sizeof (SimpleProbEst));
    memcpy(this->p_refidx_lsb1[i],            cs->p_refidx_lsb1[i],               3 * sizeof (SimpleProbEst));
    memcpy(this->p_refidx_msb[i],             cs->p_refidx_msb[i],                3 * sizeof (SimpleProbEst));
  }
  for (i = 0; i < 4; i++) 
  {
    memcpy(this->p_zblk[i], cs->p_zblk[i], NCTX_ZBLK * sizeof (SimpleProbEst));
  }
  memcpy(&this->p_intra_blk,                 &cs->p_intra_blk,                        sizeof (SimpleProbEst));
  memcpy(this->p_sip_type_b1,                 cs->p_sip_type_b1,        NCTX_SIP_TYPE*sizeof (SimpleProbEst));
  memcpy(this->p_sip_type_b2,                 cs->p_sip_type_b2,        NCTX_SIP_TYPE*sizeof (SimpleProbEst));
  memcpy(this->p_sip_type_b3,                 cs->p_sip_type_b3,        NCTX_SIP_TYPE*sizeof (SimpleProbEst));
  if (param.intra16x16)
    memcpy(this->p_sip_type_b4,                 cs->p_sip_type_b4,        NCTX_SIP_TYPE*sizeof (SimpleProbEst));

  memcpy( this->p_mode,                       cs->p_mode,                         3 * sizeof (SimpleProbEst));
  memcpy( this->p_mode_P1,                    cs->p_mode_P1,                      3 * sizeof (SimpleProbEst));
  memcpy( this->p_mode_P2,                    cs->p_mode_P2,                      3 * sizeof (SimpleProbEst));

  memcpy( this->p_mvshare_type_zero,                     cs->p_mvshare_type_zero,                      3 * sizeof (SimpleProbEst));
  memcpy(&this->p_mvshare_four,               &cs->p_mvshare_four,                 sizeof (FixedProbEst));
  memcpy(&this->p_mvshare_pred,           &cs->p_mvshare_pred,             sizeof (SimpleProbEst));

  memcpy( this->p_mode_TP1,                   cs->p_mode_TP1,                     3 * sizeof (SimpleProbEst));
  memcpy( this->p_mode_TP2,                   cs->p_mode_TP2,                     3 * sizeof (SimpleProbEst));
  memcpy( this->p_mode_TB1,                   cs->p_mode_TB1,                     3 * sizeof (SimpleProbEst));
  memcpy( this->p_mode_TB2,                   cs->p_mode_TB2,                     3 * sizeof (SimpleProbEst));

  memcpy(&this->p_prev_intra_pred_mode_flag, &cs->p_prev_intra_pred_mode_flag,        sizeof (SimpleProbEst));
  memcpy( this->p_rem_intra_pred_mode_flag,   cs->p_rem_intra_pred_mode_flag,     5 * sizeof (SimpleProbEst));
  memcpy( this->p_intra_chroma_pred_mode,     cs->p_intra_chroma_pred_mode,       3 * sizeof (SimpleProbEst));
  memcpy(&this->p_intra_chroma_pred_mode_1,  &cs->p_intra_chroma_pred_mode_1,         sizeof (SimpleProbEst));
  memcpy(&this->p_intra_chroma_pred_mode_2,  &cs->p_intra_chroma_pred_mode_2,         sizeof (SimpleProbEst));

  for (i = 0; i < 4; i++) 
  {
    memcpy(this->p_coeff_level[i],            cs->p_coeff_level[i],              16 * sizeof (UnaryProbEst));
  }
  if (param.trans8x8_enable) 
  {
    memcpy(this->p_coeff_level_64,            cs->p_coeff_level_64,        ZTsize64 * sizeof (UnaryProbEst));
    if (param.chroma8x8)
      memcpy(this->p_coeff_level_64_CT8,        cs->p_coeff_level_64_CT8,    ZTsize64 * sizeof (UnaryProbEst));
  }
  memcpy(this->p_mv_level,                    cs->p_mv_level,                     2 * sizeof (UnaryProbEst));

  memcpy(this->p_scale,                       cs->p_scale,               NCTX_SCALE * sizeof (UnaryProbEst));
  memcpy(this->p_offset,                      cs->p_offset,                       2 * sizeof (UnaryProbEst));

  memcpy(&this->p_mv_sign,                   &cs->p_mv_sign,                          sizeof (FixedProbEst));
  for (i = 0; i < 4; i++) 
  {
    memcpy(&this->p_coeff_sign[i],           &cs->p_coeff_sign[i],                    sizeof (FixedProbEst));
    memcpy(&this->p_coeff_suffix[i],         &cs->p_coeff_suffix[i],                  sizeof (FixedProbEst));
  }
  memcpy(&this->p_mv_suffix,                 &cs->p_mv_suffix,                        sizeof (FixedProbEst));
  memcpy(&this->p_offset_suffix,             &cs->p_offset_suffix,                    sizeof (FixedProbEst));
  memcpy(&this->p_scale_suffix,              &cs->p_scale_suffix,                     sizeof (FixedProbEst));
  memcpy(&this->p_offset_sign,               &cs->p_offset_sign,                      sizeof (FixedProbEst));
  memcpy(&this->p_scale_sign,                &cs->p_scale_sign,                       sizeof (FixedProbEst));
  memcpy(&this->p_coeff_suffix_64,           &cs->p_coeff_suffix_64,                  sizeof (FixedProbEst));
  memcpy(&this->p_coeff_sign_64,             &cs->p_coeff_sign_64,                    sizeof (FixedProbEst));
  if (param.chroma8x8)
  {
    memcpy(&this->p_coeff_suffix_64_CT8,       &cs->p_coeff_suffix_64_CT8, sizeof (FixedProbEst));
    memcpy(&this->p_coeff_sign_64_CT8,         &cs->p_coeff_sign_64_CT8, sizeof (FixedProbEst));
  }
  memcpy(&this->p_frame_type,                &cs->p_frame_type,                       sizeof (FixedProbEst));
  memcpy(&this->p_reference_idc,             &cs->p_reference_idc,                    sizeof (FixedProbEst));
  memcpy(&this->p_inter_type,                &cs->p_inter_type,                       sizeof (FixedProbEst));
  memcpy(&this->p_frm_zero,                  &cs->p_frm_zero,                         sizeof (FixedProbEst));
  memcpy(&this->p_framecount_sign,           &cs->p_framecount_sign,                  sizeof (FixedProbEst));
  memcpy(&this->p_framecount,                &cs->p_framecount,                       sizeof (FixedProbEst));
  memcpy(&this->p_num_of_refs,               &cs->p_num_of_refs,                      sizeof (FixedProbEst));
#ifdef MVCOMP_ZERORES
  memcpy(&this->p_mvcomp_config,             &cs->p_mvcomp_config,                    sizeof (FixedProbEst));
#endif
  memcpy(&this->p_dqp_zero,                  &cs->p_dqp_zero,                         sizeof (FixedProbEst));
  memcpy(&this->p_dqp_sign,                  &cs->p_dqp_sign,                         sizeof (FixedProbEst));
  memcpy(&this->p_deltaqp,                   &cs->p_deltaqp,                          sizeof (FixedProbEst));
  memcpy(&this->p_WLF_suffix,                &cs->p_WLF_suffix,                       sizeof (FixedProbEst));
  memcpy(&this->p_WLF,                       &cs->p_WLF,                              sizeof (UnaryProbEst));  
  memcpy(&this->p_WLF_coef,                  &cs->p_WLF_coef,                         sizeof (UnaryProbEst));
  memcpy(&this->p_WLF_coef_sign,             &cs->p_WLF_coef_sign,                    sizeof (FixedProbEst));
  memcpy(&this->p_WLF_coef_zero,             &cs->p_WLF_coef_zero,                    sizeof (SimpleProbEst));
#ifdef Q_OFFSET
  memcpy(&this->p_q_offset_zero[0],          &cs->p_q_offset_zero[0],                 sizeof (FixedProbEst));
  memcpy(&this->p_q_offset_sign[0],          &cs->p_q_offset_sign[0],                 sizeof (FixedProbEst));
  memcpy(&this->p_q_offset[0],               &cs->p_q_offset[0],                      sizeof (FixedProbEst));
  memcpy(&this->p_q_offset_zero[1],          &cs->p_q_offset_zero[1],                 sizeof (FixedProbEst));
  memcpy(&this->p_q_offset_sign[1],          &cs->p_q_offset_sign[1],                 sizeof (FixedProbEst));
  memcpy(&this->p_q_offset[1],               &cs->p_q_offset[1],                      sizeof (FixedProbEst));
  memcpy(&this->p_q_offset8x8_zero[0],       &cs->p_q_offset8x8_zero[0],              sizeof (FixedProbEst));  
  memcpy(&this->p_q_offset8x8_sign[0],       &cs->p_q_offset8x8_sign[0],              sizeof (FixedProbEst));  
  memcpy(&this->p_q_offset8x8[0],            &cs->p_q_offset8x8[0],                   sizeof (FixedProbEst));  
  memcpy(&this->p_q_offset8x8_zero[1],       &cs->p_q_offset8x8_zero[1],              sizeof (FixedProbEst));
  memcpy(&this->p_q_offset8x8_sign[1],       &cs->p_q_offset8x8_sign[1],              sizeof (FixedProbEst));
  memcpy(&this->p_q_offset8x8[1],            &cs->p_q_offset8x8[1],                   sizeof (FixedProbEst));
#endif

  if (param.intra16x16 && param.transform_selection_16x16)
    memcpy(&this->p_tsfm_flag,         &cs->p_tsfm_flag ,            sizeof (FixedProbEst));
}
#endif
