/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file ExplicitGop.cpp
*
* \brief Implementation of ExplicitGop class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "ExplicitGop.H"
#include "Global.H"
#include "Param.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] int_fr Number of intermediate frames in GOP
*/
//-----------------------------------------------------------------------------
ExplicitGop::ExplicitGop(int int_fr)
{ 
  intermed_frame = int_fr;
  
  gop_structure = new GOP_DATA[imax(1, intermed_frame)]; // At least one entry
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
ExplicitGop::~ExplicitGop()
{
  delete[] gop_structure;
}

//-----------------------------------------------------------------------------
// Public methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Create pyramid
*
* Creates a pyramid of intermediate frames. 
* Defines coding order as breadth-first traversal of pyramid.
*
*/
//-----------------------------------------------------------------------------
void ExplicitGop::create_pyramid()
{
  int num_levels;
  int i, j;
  
  if (intermed_frame == 0) // Nothing to create if no intermediate frames
    return;
  
  num_levels = 1;
  
  if (param.pyramid_coding)
  {
    while (((intermed_frame + 1) >> num_levels) > 1)
      num_levels++;
    
    if (param.pyramid_coding == EGOP_TWO_LAYER)
      num_levels = imin(num_levels, 2);
  }
    
  for (i = 0; i < intermed_frame; i++) 
  {
    int layer = num_levels - 1;
    gop_structure[i].frame_type     = ft_X_frame;
    gop_structure[i].display_no     = i;
    gop_structure[i].pyramid_layer  = layer;
    gop_structure[i].reference_idc  = param.intermed_disp;
    gop_structure[i].frame_qp       = param.qp + layer * param.qp_off_intref + (param.intermed_disp ? param.qp_off_intdisp - param.qp_off_intref : param.qp_off_intref);
  }
  
  for (j = 1; j < num_levels; j++)
  {
    for (i = power2(j) - 1; i < intermed_frame + 1 - power2(j); i += power2(j)) 
    {
      int layer = num_levels - 1 - j;
      gop_structure[i].pyramid_layer  = layer;
      gop_structure[i].reference_idc  = REFERENCE_PIC;
      gop_structure[i].frame_qp       = param.qp + (layer + 1) * param.qp_off_intref;
    }
  }
  
  // Reorder frames to encode lower layers first
  if (!param.no_reordering) 
  {
    for (i = intermed_frame - 1; i > 1; i--)
    {
      j = i;
      
      while (j > 0 && gop_structure[j].pyramid_layer < gop_structure[j-1].pyramid_layer) 
      {
        GOP_DATA tmp = gop_structure[j-1];
        gop_structure[j-1] = gop_structure[j];
        gop_structure[j] = tmp;
        j--;
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Parses explicit GOP structure parameter
*
* \param[in] p Pointer to parameter string. "expPrmdFrmt" or "expPrmdFrmtIdr"
*
*/
//-----------------------------------------------------------------------------
void ExplicitGop::interpret_gop_structure(char *p)
{
  int i, k;
  int v;
  
  if (intermed_frame == 0)
    return;
  
  for (i = ZERO; i < intermed_frame; i++) 
  {
    if (p[ZERO] == ZERO) 
    {
      printf("%s is too short. Please check configuration file.\n",
        p==param.expPrmdFrmt?"expPrmdFrmt":"expPrmdFrmtIdr");
      exit(EXIT_FAILURE);
    }
    
    v = read_symbol(&p);
    
    switch (v) 
    {
      case 'P':
        gop_structure[i].frame_type = ft_P_frame;
        break;
      case 'B':
        gop_structure[i].frame_type = ft_B_frame;
        break;
      case 'I':
        gop_structure[i].frame_type = ft_I_frame;
        break;
      case 'X':
        gop_structure[i].frame_type = ft_X_frame;
        break;
      default:
        printf("Frame Type invalid in %s param. Please check configuration file.\n",
          p==param.expPrmdFrmt?"expPrmdFrmt":"expPrmdFrmtIdr");
        exit(EXIT_FAILURE);
        break;
    }
    
    v = read_number(&p);
    
    if (v == INT_INF) 
    {
      printf("Frame Type needs to be followed by Display Order. Please check configuration file.\n");
      exit(EXIT_FAILURE);
    }
    else if (v < ZERO || v >= intermed_frame) 
    {
      printf("Invalid Frame Order value. Frame position needs to be in [0,%d] range.\n",intermed_frame - 1);
      exit(EXIT_FAILURE);
    }
    else 
    {
      for (k = ZERO; k < i; k++) 
      {
        if (gop_structure[k].display_no == v) 
        {
          printf("Frame Order value %d in frame %d already used for enhancement frame %d.\n", v, i, k);
          exit(EXIT_FAILURE);
        }
      }
      gop_structure[i].display_no = v;
    }
    
    v = read_symbol(&p);
    
    switch (v) 
    {
      case 'E':
        gop_structure[i].reference_idc = DISPOSABLE_PIC;
        break;
      case 'R':
        gop_structure[i].reference_idc = REFERENCE_PIC;
        break;
      default:
        printf("Reference_IDC invalid in %s param. Please check configuration file.\n",
          p==param.expPrmdFrmt?"expPrmdFrmt":"expPrmdFrmtIdr");
        exit(EXIT_FAILURE);
        break;
    }
    
    v = read_number(&p);
    
    if (v == INT_INF) 
    {
      printf("Reference_IDC needs to be followed by lambdaq_offset. Please check configuration file.\n");
      exit(EXIT_FAILURE);
    }
    else 
    {
      gop_structure[i].frame_qp = v;
    }
  }
  
  if (p[ZERO] != ZERO) 
  {
    printf("%s is too long. Please check configuration file.\n",
      p==param.expPrmdFrmt?"expPrmdFrmt":"expPrmdFrmtIdr");
    exit(EXIT_FAILURE);
  }
}

//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Parses number in explicit GOP structure parameter
*
* \param[in,out] p Pointer to parameter string
*
* \return Number
*/
//-----------------------------------------------------------------------------
int ExplicitGop::read_number(char **p)
{
  int r;
  char c, *q;
  
  r = ZERO;
  q = *p;
  
  while (TRUE) 
  {
    c = *q;
    if (c >= '0' && c <= '9') 
    {
      r = r * 10 + (c - '0');
      q++;
    }
    else
      break;
  }
  
  if (q == *p)
    return INT_INF;
  
  *p = q;
  
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Parses alphabetic symbol in explicit GOP structure parameter
*
* \param[in,out] p Pointer to parameter string
*
* \return Symbol
*/
//-----------------------------------------------------------------------------
int ExplicitGop::read_symbol(char **p)
{
  int r;
  char c, *q;
  
  r = ZERO;
  q = *p;
  
  c = *q;
  if (c >= 'a' && c <= 'z') 
  {
    r = c - 'a' + 'A';
    q++;
  }
  else if (c >= 'A' && c <= 'Z') 
  {
    r = c;
    q++;
  }
  
  if (q == *p)
    return INT_INF;
  
  *p = q;
  
  return r;
}

void ExplicitGop::resize(int s)
{
  while (s < intermed_frame)
  {
    intermed_frame--;
    int no = gop_structure[intermed_frame].display_no;

    int i;
    for (i = 0; i < intermed_frame; i++)
    {
      if (gop_structure[i].display_no > no)
        gop_structure[i].display_no--;
    }
  }
}
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
