/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file FastDCT.cpp
*
* \brief Implementation of FastDCT class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "FastDCT.H"

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Forward transform
*
* \param[in,out] blk Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::forward(Block *blk)
{
  if (blk->get_height() == 2)
    forward_2x8(blk);
  else if (blk->get_width() == 2)
    forward_8x2(blk);
  else if (blk->get_width() == 8)
    forward_8x8(blk);
  else
    forward_4x4(blk);
}

void FastDCT::forward_4x4(Block *blk)
{
  int i;
  int tmp[TX_NCOEFF];
  int a0,a1,a2,a3;
  int b0,b1,b2,b3;
  
  for (i=0; i<TX_SIZE; i++) 
  {
    a0 = blk->val[TX_SIZE*i+0];
    a1 = blk->val[TX_SIZE*i+1];
    a2 = blk->val[TX_SIZE*i+2];
    a3 = blk->val[TX_SIZE*i+3];
    b0 = (a0+a3+2)>>2;
    b1 = (a1+a2+2)>>2;
    b2 = (a1-a2+2)>>2;
    b3 = (a0-a3+2)>>2;
    tmp[i+TX_SIZE*0] = b0+b1;
    tmp[i+TX_SIZE*1] = b3+((b3+2)>>2)+(b2>>1);
    tmp[i+TX_SIZE*2] = b0-b1;
    tmp[i+TX_SIZE*3] = (b3>>1)-b2-((b2+2)>>2);
  }
  
  for (i=0; i<TX_SIZE; i++) 
  {
    a0 = tmp[TX_SIZE*i+0];
    a1 = tmp[TX_SIZE*i+1];
    a2 = tmp[TX_SIZE*i+2];
    a3 = tmp[TX_SIZE*i+3];
    b0 = (a0+a3+2)>>2;
    b1 = (a1+a2+2)>>2;
    b2 = (a1-a2+2)>>2;
    b3 = (a0-a3+2)>>2;
    blk->val[i+TX_SIZE*0] = b0+b1;
    blk->val[i+TX_SIZE*1] = ((b3+2)>>2)+b3+(b2>>1);
    blk->val[i+TX_SIZE*2] = b0-b1;
    blk->val[i+TX_SIZE*3] = (b3>>1)-((b2+2)>>2)-b2;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Inverse transform
*
* \param[in,out] blk Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::inverse(Block *blk)
{
  int blk_height = blk->get_height();
  int blk_width = blk->get_width();
  
  if (blk_height == 2)
    inverse_2x8(blk);
  else if (blk_width == 2)
    inverse_8x2(blk);
  else if (blk_width == 4)
    inverse_4x4(blk);
  else
    inverse_8x8(blk);
}

void FastDCT::inverse_4x4(Block *blk)
{
  int i;
  int tmp[TX_NCOEFF];
  int a0,a1,a2,a3;
  int b0,b1,b2,b3;
  
  for (i=0; i<TX_SIZE; i++) 
  {
    a0 = blk->val[TX_SIZE*i+0];
    a1 = blk->val[TX_SIZE*i+1];
    a2 = blk->val[TX_SIZE*i+2];
    a3 = blk->val[TX_SIZE*i+3];
    b0 = a0+a2;
    b1 = a0-a2;
    b2 = a1+((a1+2)>>2)+(a3>>1);
    b3 = (a1>>1)-a3-((a3+2)>>2);
    tmp[i+TX_SIZE*0] = b0+b2;
    tmp[i+TX_SIZE*1] = b1+b3;
    tmp[i+TX_SIZE*2] = b1-b3;
    tmp[i+TX_SIZE*3] = b0-b2;
  }
  
  for (i=0; i<TX_SIZE; i++) 
  {
    a0 = tmp[TX_SIZE*i+0];
    a1 = tmp[TX_SIZE*i+1];
    a2 = tmp[TX_SIZE*i+2];
    a3 = tmp[TX_SIZE*i+3];
    b0 = a0+a2;
    b1 = a0-a2;
    b2 = a1+((a1+2)>>2)+(a3>>1);
    b3 = (a1>>1)-a3-((a3+2)>>2);
    blk->val[i+TX_SIZE*0] = b0 + b2;
    blk->val[i+TX_SIZE*1] = b1 + b3;
    blk->val[i+TX_SIZE*2] = b1 - b3;
    blk->val[i+TX_SIZE*3] = b0 - b2;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Forward transform
*
* \param[in,out] blk Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::forward_2x8(Block *blk)
{
  int i;
  int tmp[16];
  int a0,a1,a2,a3;
  int p0, p1, p2, p3, p4, p5 ,p6, p7;  
  int b0, b1, b2, b3, b4, b5, b6, b7;
  
  for (i=0; i<2; i++) 
  {
    p0 = blk->val[8*i+0];
    p1 = blk->val[8*i+1];
    p2 = blk->val[8*i+2];
    p3 = blk->val[8*i+3];
    p4 = blk->val[8*i+4];
    p5 = blk->val[8*i+5];
    p6 = blk->val[8*i+6];
    p7 = blk->val[8*i+7];
    
    a0 = p0 + p7;
    a1 = p1 + p6;
    a2 = p2 + p5;
    a3 = p3 + p4;
    
    b0 = a0 + a3;
    b1 = a1 + a2;
    b2 = a0 - a3;
    b3 = a1 - a2;
    
    a0 = p0 - p7;
    a1 = p1 - p6;
    a2 = p2 - p5;
    a3 = p3 - p4;
    
    b4 = a1 + a2 + ((a0 >> 1) + a0);
    b5 = a0 - a3 - ((a2 >> 1) + a2);
    b6 = a0 + a3 - ((a1 >> 1) + a1);
    b7 = a1 - a2 + ((a3 >> 1) + a3);
    
    tmp[8*i+0] =  b0 + b1;
    tmp[8*i+1] =  b4 + (b7 >> 2);
    tmp[8*i+2] =  b2 + (b3 >> 1);
    tmp[8*i+3] =  b5 + (b6 >> 2);
    tmp[8*i+4] =  b0 - b1;
    tmp[8*i+5] =  b6 - (b5 >> 2);
    tmp[8*i+6] = (b2 >> 1) - b3;                 
    tmp[8*i+7] = (b4 >> 2) - b7;
  }
  
  for (i=0; i<8; i++) 
  {
    a0 = tmp[i+8*0] >> 4;
    a1 = tmp[i+8*1] >> 4;
    blk->val[i+8*0] = (a0+a1);
    blk->val[i+8*1] = (a0-a1);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Inverse transform
*
* \param[in,out] blk Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::inverse_2x8(Block *blk)
{
  int i;
  int tmp[TX_NCOEFF];
  int a0,a1,a2,a3;
  int p0, p1, p2, p3, p4, p5 ,p6, p7;  
  int b0, b1, b2, b3, b4, b5, b6, b7;
  
  for (i=0; i<8; i++) 
  {
    a0 = blk->val[i+8*0];
    a1 = blk->val[i+8*1];
    tmp[i+8*0] = a0+a1;
    tmp[i+8*1] = a0-a1;
  }
  
  for (i=0; i<2; i++) 
  {
    p0 = tmp[8*i+0];
    p1 = tmp[8*i+1];
    p2 = tmp[8*i+2];
    p3 = tmp[8*i+3];
    p4 = tmp[8*i+4];
    p5 = tmp[8*i+5];
    p6 = tmp[8*i+6];
    p7 = tmp[8*i+7];
    
    a0 = p0 + p4;
    a1 = p0 - p4;
    a2 = -p6 + (p2 >> 1);
    a3 = p2 + (p6 >> 1);
    
    b0 =  a0 + a3;
    b2 =  a1 + a2;
    b4 =  a1 - a2;
    b6 =  a0 - a3;
    
    a0 = -p3 + p5 - p7 - (p7 >> 1);    
    a1 =  p1 + p7 - p3 - (p3 >> 1);    
    a2 = -p1 + p7 + p5 + (p5 >> 1);    
    a3 =  p3 + p5 + p1 + (p1 >> 1);
    
    b1 =  a0 + (a3>>2);    
    b3 =  a1 + (a2>>2);    
    b5 =  -a2 + (a1>>2);
    b7 =  a3 - (a0>>2);                
    
    blk->val[8*i+0] = b0 + b7;
    blk->val[8*i+1] = b2 + b5;
    blk->val[8*i+2] = b4 + b3;
    blk->val[8*i+3] = b6 + b1;
    blk->val[8*i+4] = b6 - b1;
    blk->val[8*i+5] = b4 - b3;
    blk->val[8*i+6] = b2 - b5;
    blk->val[8*i+7] = b0 - b7;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Forward transform
*
* \param[in,out] blk Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::forward_8x2(Block *blk)
{
  int i;
  int tmp[16];
  int a0,a1,a2,a3;
  int p0, p1, p2, p3, p4, p5 ,p6, p7;  
  int b0, b1, b2, b3, b4, b5, b6, b7;
  
  for (i=0; i<2; i++) 
  {
    p0 = blk->val[i];
    p1 = blk->val[2+i];
    p2 = blk->val[4+i];
    p3 = blk->val[6+i];
    p4 = blk->val[8+i];
    p5 = blk->val[10+i];
    p6 = blk->val[12+i];
    p7 = blk->val[14+i];
    
    a0 = p0 + p7;
    a1 = p1 + p6;
    a2 = p2 + p5;
    a3 = p3 + p4;
    
    b0 = a0 + a3;
    b1 = a1 + a2;
    b2 = a0 - a3;
    b3 = a1 - a2;
    
    a0 = p0 - p7;
    a1 = p1 - p6;
    a2 = p2 - p5;
    a3 = p3 - p4;
    
    b4 = a1 + a2 + ((a0 >> 1) + a0);
    b5 = a0 - a3 - ((a2 >> 1) + a2);
    b6 = a0 + a3 - ((a1 >> 1) + a1);
    b7 = a1 - a2 + ((a3 >> 1) + a3);
    
    tmp[i+0] =  b0 + b1;
    tmp[i+2] =  b4 + (b7 >> 2);
    tmp[i+4] =  b2 + (b3 >> 1);
    tmp[i+6] =  b5 + (b6 >> 2);
    tmp[i+8] =  b0 - b1;
    tmp[i+10] =  b6 - (b5 >> 2);
    tmp[i+12] = (b2 >> 1) - b3;                 
    tmp[i+14] = (b4 >> 2) - b7;
  }
  
  for (i=0; i<8; i++) 
  {
    a0 = tmp[2*i] >> 4;
    a1 = tmp[2*i+1] >> 4;
    blk->val[2*i] = (a0+a1);
    blk->val[2*i+1] = (a0-a1);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Inverse transform
*
* \param[in,out] blk Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::inverse_8x2(Block *blk)
{
  int i;
  int tmp[TX_NCOEFF];
  int a0,a1,a2,a3;
  int p0, p1, p2, p3, p4, p5 ,p6, p7;  
  int b0, b1, b2, b3, b4, b5, b6, b7;
  
  for (i=0; i<8; i++) 
  {
    a0 = blk->val[2*i];
    a1 = blk->val[2*i+1];
    tmp[2*i] = a0+a1;
    tmp[2*i+1] = a0-a1;
  }
  
  for (i=0; i<2; i++) 
  {
    p0 = tmp[i+0];
    p1 = tmp[i+2];
    p2 = tmp[i+4];
    p3 = tmp[i+6];
    p4 = tmp[i+8];
    p5 = tmp[i+10];
    p6 = tmp[i+12];
    p7 = tmp[i+14];
    
    a0 = p0 + p4;
    a1 = p0 - p4;
    a2 = -p6 + (p2 >> 1);
    a3 = p2 + (p6 >> 1);
    
    b0 =  a0 + a3;
    b2 =  a1 + a2;
    b4 =  a1 - a2;
    b6 =  a0 - a3;
    
    a0 = -p3 + p5 - p7 - (p7 >> 1);    
    a1 =  p1 + p7 - p3 - (p3 >> 1);    
    a2 = -p1 + p7 + p5 + (p5 >> 1);    
    a3 =  p3 + p5 + p1 + (p1 >> 1);
    
    b1 =  a0 + (a3>>2);    
    b3 =  a1 + (a2>>2);    
    b5 =  -a2 + (a1>>2);
    b7 =  a3 - (a0>>2);                
    
    blk->val[i+0] = b0 + b7;
    blk->val[i+2] = b2 + b5;
    blk->val[i+4] = b4 + b3;
    blk->val[i+6] = b6 + b1;
    blk->val[i+8] = b6 - b1;
    blk->val[i+10] = b4 - b3;
    blk->val[i+12] = b2 - b5;
    blk->val[i+14] = b0 - b7;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Forward transform
*
* \param[in,out] blk8 Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::forward_8x8(Block *blk8)
{
  int i, ii,k;  
  int BLOCK_SIZE_8x8 = 8;
  int pos_y =0 ,  pos_x =0;
  
  int tmp[64];
  int *pTmp = tmp, *pblock;
  int block[8][8];
  int tblock[8][8];
  
  static int a0, a1, a2, a3;
  static int p0, p1, p2, p3, p4, p5 ,p6, p7;
  static int b0, b1, b2, b3, b4, b5, b6, b7;
  
  for(i=0;i<8;i++)
    for(ii=0;ii<8;ii++)
      block[i][ii]=blk8->val[i*8+ii];
  
  #define dct8x8offset1 0
  #define dct8x8offset2 0
  #define dct8x8offset3 0
  #define dct8x8offset4 0
  #define dct8x8offset5 0
  #define dct8x8offset6 0
  #define dct8x8offset7 0
  #define dct8x8offset8 0
  #define dct8x8offset9 0
  
  #define dct8x8shift1 0
  #define dct8x8shift2 0
  #define dct8x8shift3 0
  #define dct8x8shift4 0
  #define dct8x8shift5 0
  #define dct8x8shift6 0
  #define dct8x8shift7 0
  #define dct8x8shift8 0
  #define dct8x8shift9 0
  
  #define endoffset 4
  #define endshift 3
  
  for (i=pos_y; i < pos_y + BLOCK_SIZE_8x8; i++)
  {
    pblock = &block[i][pos_x];
    p0 = *(pblock++);
    p1 = *(pblock++);
    p2 = *(pblock++);
    p3 = *(pblock++);
    p4 = *(pblock++);
    p5 = *(pblock++);
    p6 = *(pblock++);
    p7 = *(pblock  );
    
    a0 = ((p0 + p7)+dct8x8offset1)>>dct8x8shift1;
    a1 = ((p1 + p6)+dct8x8offset1)>>dct8x8shift1;
    a2 = ((p2 + p5)+dct8x8offset1)>>dct8x8shift1;
    a3 = ((p3 + p4)+dct8x8offset1)>>dct8x8shift1;
    
    b0 = ((a0 + a3)+dct8x8offset2)>>dct8x8shift2;
    b1 = ((a1 + a2)+dct8x8offset2)>>dct8x8shift2;
    b2 = ((a0 - a3)+dct8x8offset2)>>dct8x8shift2;
    b3 = ((a1 - a2)+dct8x8offset2)>>dct8x8shift2;
    
    a0 = (p0 - p7+dct8x8offset3)>>dct8x8shift3;
    a1 = (p1 - p6+dct8x8offset3)>>dct8x8shift3;
    a2 = (p2 - p5+dct8x8offset3)>>dct8x8shift3;
    a3 = (p3 - p4+dct8x8offset3)>>dct8x8shift3;
    
    b4 = (a1 + a2 + ((a0 >> 1) + a0)+dct8x8offset4)>>dct8x8shift4;
    b5 = (a0 - a3 - ((a2 >> 1) + a2)+dct8x8offset4)>>dct8x8shift4;
    b6 = (a0 + a3 - ((a1 >> 1) + a1)+dct8x8offset4)>>dct8x8shift4;
    b7 = (a1 - a2 + ((a3 >> 1) + a3)+dct8x8offset4)>>dct8x8shift4;
    
    *(pTmp++) = ((b0 + b1)+dct8x8offset5)>>dct8x8shift5;
    *(pTmp++) = ((b4 + (b7 >> 2))+dct8x8offset5)>>dct8x8shift5;
    *(pTmp++) = ((b2 + (b3 >> 1))+dct8x8offset5)>>dct8x8shift5;
    *(pTmp++) = ((b5 + (b6 >> 2))+dct8x8offset5)>>dct8x8shift5;
    *(pTmp++) = ((b0 - b1)+dct8x8offset5)>>dct8x8shift5;
    *(pTmp++) = ((b6 - (b5 >> 2))+dct8x8offset5)>>dct8x8shift5;
    *(pTmp++) = (((b2 >> 1) - b3)+dct8x8offset5)>>dct8x8shift5;                 
    *(pTmp++) = (((b4 >> 2) - b7)+dct8x8offset5)>>dct8x8shift5;
  }
  
  for (i=0; i < BLOCK_SIZE_8x8; i++)
  {
    pTmp = tmp + i;
    p0 = *pTmp;
    p1 = *(pTmp += BLOCK_SIZE_8x8);
    p2 = *(pTmp += BLOCK_SIZE_8x8);
    p3 = *(pTmp += BLOCK_SIZE_8x8);
    p4 = *(pTmp += BLOCK_SIZE_8x8);
    p5 = *(pTmp += BLOCK_SIZE_8x8);
    p6 = *(pTmp += BLOCK_SIZE_8x8);
    p7 = *(pTmp += BLOCK_SIZE_8x8);
    
    a0 = ((p0 + p7)+dct8x8offset7)>>dct8x8shift7;
    a1 = ((p1 + p6)+dct8x8offset7)>>dct8x8shift7;
    a2 = ((p2 + p5)+dct8x8offset7)>>dct8x8shift7;
    a3 = ((p3 + p4)+dct8x8offset7)>>dct8x8shift7;
    
    b0 = ((a0 + a3)+dct8x8offset8)>>dct8x8shift8;
    b1 = ((a1 + a2)+dct8x8offset8)>>dct8x8shift8;
    b2 = ((a0 - a3)+dct8x8offset8)>>dct8x8shift8;
    b3 = ((a1 - a2)+dct8x8offset8)>>dct8x8shift8;
    
    a0 = ((p0 - p7)+dct8x8offset9)>>dct8x8shift9;
    a1 = ((p1 - p6)+dct8x8offset9)>>dct8x8shift9;
    a2 = ((p2 - p5)+dct8x8offset9)>>dct8x8shift9;
    a3 = ((p3 - p4)+dct8x8offset9)>>dct8x8shift9;
    
    b4 = a1 + a2 + ((a0 >> 1) + a0);
    b5 = a0 - a3 - ((a2 >> 1) + a2);
    b6 = a0 + a3 - ((a1 >> 1) + a1);
    b7 = a1 - a2 + ((a3 >> 1) + a3);
    
    ii = pos_x + i;
    tblock[pos_y    ][ii] =  (b0 + b1)>>6;
    tblock[pos_y + 1][ii] =  (b4 + (b7 >> 2))>>6;
    tblock[pos_y + 2][ii] =  (b2 + (b3 >> 1))>>6;
    tblock[pos_y + 3][ii] =  (b5 + (b6 >> 2))>>6;
    tblock[pos_y + 4][ii] =  (b0 - b1)>>6;
    tblock[pos_y + 5][ii] =  (b6 - (b5 >> 2))>>6;
    tblock[pos_y + 6][ii] = ((b2 >> 1) - b3)>>6;
    tblock[pos_y + 7][ii] = ((b4 >> 2) - b7)>>6;
  }
  
  k=0;
  
  for(i=0;i<8;i++)
  {
    for(ii=0;ii<8;ii++)
    {
      blk8->val[i*8+ii] = tblock[pos_y+i][pos_x+ii];
    }
    k++;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Inverse transform
*
* \param[in,out] blk8 Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::inverse_8x8(Block *blk8)
{
  int i, ii;
  int BLOCK_SIZE_8x8 = 8;
  int pos_y =0 ,  pos_x =0;
  int tmp[64];
  int *pTmp = tmp, *pblock;
  int block[8][8];
  int tblock[8][8];
  
#ifdef SUBSTREAM_MT
  // remove statics!!!!!
  /*static*/ int a0, a1, a2, a3;
  /*static*/ int p0, p1, p2, p3, p4, p5 ,p6, p7;  
  /*static*/ int b0, b1, b2, b3, b4, b5, b6, b7;
#else
  static int a0, a1, a2, a3;
  static int p0, p1, p2, p3, p4, p5 ,p6, p7;  
  static int b0, b1, b2, b3, b4, b5, b6, b7;
#endif
  
  for(i=0;i<8;i++)
  {
    for(ii=0;ii<8;ii++)
    {
      tblock[i][ii]=blk8->val[i*8+ii];
    }
  }
  
  // Horizontal  
  for (i=pos_y; i < pos_y + BLOCK_SIZE_8x8; i++)
  {
    pblock = &tblock[i][pos_x];
    p0 = *(pblock++);
    p1 = *(pblock++);
    p2 = *(pblock++);
    p3 = *(pblock++);
    p4 = *(pblock++);
    p5 = *(pblock++);
    p6 = *(pblock++);
    p7 = *(pblock  );
    
    a0 = p0 + p4;
    a1 = p0 - p4;
    a2 = p6 - (p2 >> 1);
    a3 = p2 + (p6 >> 1);
    
    b0 =  a0 + a3;
    b2 =  a1 - a2;
    b4 =  a1 + a2;
    b6 =  a0 - a3;
    
    a0 = -p3 + p5 - p7 - (p7 >> 1);    
    a1 =  p1 + p7 - p3 - (p3 >> 1);    
    a2 = -p1 + p7 + p5 + (p5 >> 1);    
    a3 =  p3 + p5 + p1 + (p1 >> 1);
    
    b1 =  a0 + (a3>>2);    
    b3 =  a1 + (a2>>2);    
    b5 =  a2 - (a1>>2);
    b7 =  a3 - (a0>>2);                
    
    *(pTmp++) = b0 + b7;
    *(pTmp++) = b2 - b5;
    *(pTmp++) = b4 + b3;
    *(pTmp++) = b6 + b1;
    *(pTmp++) = b6 - b1;
    *(pTmp++) = b4 - b3;
    *(pTmp++) = b2 + b5;
    *(pTmp++) = b0 - b7;
  }
  
  for (i=0; i < BLOCK_SIZE_8x8; i++)
  {
    pTmp = tmp + i;
    p0 = *pTmp;
    p1 = *(pTmp += BLOCK_SIZE_8x8);
    p2 = *(pTmp += BLOCK_SIZE_8x8);
    p3 = *(pTmp += BLOCK_SIZE_8x8);
    p4 = *(pTmp += BLOCK_SIZE_8x8);
    p5 = *(pTmp += BLOCK_SIZE_8x8);
    p6 = *(pTmp += BLOCK_SIZE_8x8);
    p7 = *(pTmp += BLOCK_SIZE_8x8);
    
    a0 =  p0 + p4;
    a1 =  p0 - p4;
    a2 =  p6 - (p2>>1);
    a3 =  p2 + (p6>>1);
    
    b0 = a0 + a3;
    b2 = a1 - a2;
    b4 = a1 + a2;
    b6 = a0 - a3;
    
    a0 = -p3 + p5 - p7 - (p7 >> 1);
    a1 =  p1 + p7 - p3 - (p3 >> 1);
    a2 = -p1 + p7 + p5 + (p5 >> 1);
    a3 =  p3 + p5 + p1 + (p1 >> 1);
    
    b1 =  a0 + (a3 >> 2);
    b7 =  a3 - (a0 >> 2);
    b3 =  a1 + (a2 >> 2);
    b5 =  a2 - (a1 >> 2);
    
    ii = i + pos_x;
    block[pos_y    ][ii] = b0 + b7;
    block[pos_y + 1][ii] = b2 - b5;
    block[pos_y + 2][ii] = b4 + b3;
    block[pos_y + 3][ii] = b6 + b1;
    block[pos_y + 4][ii] = b6 - b1;
    block[pos_y + 5][ii] = b4 - b3;
    block[pos_y + 6][ii] = b2 + b5;
    block[pos_y + 7][ii] = b0 - b7;
  }
  for(i=0;i<8;i++)
  {
    for(ii=0;ii<8;ii++)
    {
      blk8->val[i*8+ii] = block[pos_y+i][pos_x+ii];
    }
    //    printf("\n");
  }
  //  printf("\n");
}


// DCT16x16 = int8(dct(eye(16))) * (2^7)
int DCT16x16[] = {
   32,   32,   32,   32,   32,   32,   32,   32,   32,   32,   32,   32,   32,   32,   32,   32,
   45,   43,   40,   35,   29,   21,   13,    4,   -4,  -13,  -21,  -29,  -35,  -40,  -43,  -45,
   44,   38,   25,    9,   -9,  -25,  -38,  -44,  -44,  -38,  -25,   -9,    9,   25,   38,   44,
   43,   29,    4,  -21,  -40,  -45,  -35,  -13,   13,   35,   45,   40,   21,   -4,  -29,  -43,
   42,   17,  -17,  -42,  -42,  -17,   17,   42,   42,   17,  -17,  -42,  -42,  -17,   17,   42,
   40,    4,  -35,  -43,  -13,   29,   45,   21,  -21,  -45,  -29,   13,   43,   35,   -4,  -40,
   38,   -9,  -44,  -25,   25,   44,    9,  -38,  -38,    9,   44,   25,  -25,  -44,   -9,   38,
   35,  -21,  -43,    4,   45,   13,  -40,  -29,   29,   40,  -13,  -45,   -4,   43,   21,  -35,
   32,  -32,  -32,   32,   32,  -32,  -32,   32,   32,  -32,  -32,   32,   32,  -32,  -32,   32,
   29,  -40,  -13,   45,   -4,  -43,   21,   35,  -35,  -21,   43,    4,  -45,   13,   40,  -29,
   25,  -44,    9,   38,  -38,   -9,   44,  -25,  -25,   44,   -9,  -38,   38,    9,  -44,   25,
   21,  -45,   29,   13,  -43,   35,    4,  -40,   40,   -4,  -35,   43,  -13,  -29,   45,  -21,
   17,  -42,   42,  -17,  -17,   42,  -42,   17,   17,  -42,   42,  -17,  -17,   42,  -42,   17,
   13,  -35,   45,  -40,   21,    4,  -29,   43,  -43,   29,   -4,  -21,   40,  -45,   35,  -13,
    9,  -25,   38,  -44,   44,  -38,   25,   -9,   -9,   25,  -38,   44,  -44,   38,  -25,    9,
    4,  -13,   21,  -29,   35,  -40,   43,  -45,   45,  -43,   40,  -35,   29,  -21,   13,   -4
};

//-----------------------------------------------------------------------------
/*!
* \brief Forward transform
*
* \param[in,out] blk Block of data to transform
*/
//-----------------------------------------------------------------------------
void FastDCT::forward_transform(Block *blk)
{
  int i, j;
  int a[256];
  int tmp[256];
  int tsize;
  int scale;
  int shift2;

  tsize = 16;
  assert(blk->get_height() == tsize && blk->get_width() == tsize);

  scale = 3;
  shift2 = (1 << (scale - 1));

  int b[8];
  for (i = 0; i < tsize; i ++) 
  {
    for (j = 0; j < tsize; j ++ )
      a[j] = blk->val[tsize * i + j];

    b[0] = a[0] + a[3] + a[4] + a[7] + a[8] + a[11] + a[12] + a[15];
    b[1] = a[1] + a[2] + a[5] + a[6] + a[9] + a[10] + a[13] + a[14];

    tmp[i] = (b[0] + b[1]) << 5;
    tmp[i + tsize * 8] = (b[0] - b[1]) << 5;

    b[0] = a[0] - a[15];
    b[1] = a[1] - a[14];
    b[2] = a[2] - a[13];
    b[3] = a[3] - a[12];
    b[4] = a[4] - a[11];
    b[5] = a[5] - a[10];
    b[6] = a[6] - a[9];
    b[7] = a[7] - a[8];
    
    tmp[i + tsize]      = 45*b[0] + 43*b[1] + 40*b[2] + 35*b[3] + 29*b[4] + 21*b[5] + 13*b[6] + (b[7]<<2); 
    tmp[i + tsize * 3]  = 43*b[0] + 29*b[1] + (b[2]<<2) - 21*b[3] - 40*b[4] - 45*b[5] - 35*b[6] - 13*b[7]; 
    tmp[i + tsize * 5]  = 40*b[0] + (b[1]<<2) - 35*b[2] - 43*b[3] - 13*b[4] + 29*b[5] + 45*b[6] + 21*b[7]; 
    tmp[i + tsize * 7]  = 35*b[0] - 21*b[1] - 43*b[2] + (b[3]<<2) + 45*b[4] + 13*b[5] - 40*b[6] - 29*b[7]; 
    tmp[i + tsize * 9]  = 29*b[0] - 40*b[1] - 13*b[2] + 45*b[3] - (b[4]<<2) - 43*b[5] + 21*b[6] + 35*b[7]; 
    tmp[i + tsize * 11] = 21*b[0] - 45*b[1] + 29*b[2] + 13*b[3] - 43*b[4] + 35*b[5] + (b[6]<<2) - 40*b[7];
    tmp[i + tsize * 13] = 13*b[0] - 35*b[1] + 45*b[2] - 40*b[3] + 21*b[4] + (b[5]<<2) - 29*b[6] + 43*b[7];
    tmp[i + tsize * 15] = (b[0]<<2) - 13*b[1] + 21*b[2] - 29*b[3] + 35*b[4] - 40*b[5] + 43*b[6] - 45*b[7];

    b[0] = a[0] - a[7] - a[8]  + a[15];
    b[1] = a[1] - a[6] - a[9]  + a[14];
    b[2] = a[2] - a[5] - a[10] + a[13];
    b[3] = a[3] - a[4] - a[11] + a[12];

    tmp[i + tsize * 2]  = 44*b[0] + 38*b[1] + 25*b[2] +  9*b[3];
    tmp[i + tsize * 6]  = 38*b[0] -  9*b[1] - 44*b[2] - 25*b[3];
    tmp[i + tsize * 10] = 25*b[0] - 44*b[1] +  9*b[2] + 38*b[3];
    tmp[i + tsize * 14] =  9*b[0] - 25*b[1] + 38*b[2] - 44*b[3];

    b[0] = a[0] - a[3] - a[4] + a[7] + a[8] - a[11] - a[12] + a[15];
    b[1] = a[1] - a[2] - a[5] + a[6] + a[9] - a[10] - a[13] + a[14];
    tmp[i + tsize * 4]  = 42*b[0] + 17*b[1];
    tmp[i + tsize * 12] = 17*b[0] - 42*b[1];
  }

  for (i = 0; i < tsize; i ++) 
  {
    for (j = 0; j < tsize; j ++ )
      a[j] = (tmp[tsize * i + j] + shift2) >> scale;

    b[0] = a[0] + a[3] + a[4] + a[7] + a[8] + a[11] + a[12] + a[15];
    b[1] = a[1] + a[2] + a[5] + a[6] + a[9] + a[10] + a[13] + a[14];

    blk->val[i] = (b[0] + b[1]) << 5;
    blk->val[i + tsize * 8] = (b[0] - b[1]) << 5;

    b[0] = a[0] - a[15];
    b[1] = a[1] - a[14];
    b[2] = a[2] - a[13];
    b[3] = a[3] - a[12];
    b[4] = a[4] - a[11];
    b[5] = a[5] - a[10];
    b[6] = a[6] - a[9];
    b[7] = a[7] - a[8];

    blk->val[i + tsize]      = 45*b[0] + 43*b[1] + 40*b[2] + 35*b[3] + 29*b[4] + 21*b[5] + 13*b[6] + (b[7]<<2); 
    blk->val[i + tsize * 3]  = 43*b[0] + 29*b[1] + (b[2]<<2) - 21*b[3] - 40*b[4] - 45*b[5] - 35*b[6] - 13*b[7]; 
    blk->val[i + tsize * 5]  = 40*b[0] + (b[1]<<2) - 35*b[2] - 43*b[3] - 13*b[4] + 29*b[5] + 45*b[6] + 21*b[7]; 
    blk->val[i + tsize * 7]  = 35*b[0] - 21*b[1] - 43*b[2] + (b[3]<<2) + 45*b[4] + 13*b[5] - 40*b[6] - 29*b[7]; 
    blk->val[i + tsize * 9]  = 29*b[0] - 40*b[1] - 13*b[2] + 45*b[3] - (b[4]<<2) - 43*b[5] + 21*b[6] + 35*b[7]; 
    blk->val[i + tsize * 11] = 21*b[0] - 45*b[1] + 29*b[2] + 13*b[3] - 43*b[4] + 35*b[5] + (b[6]<<2) - 40*b[7];
    blk->val[i + tsize * 13] = 13*b[0] - 35*b[1] + 45*b[2] - 40*b[3] + 21*b[4] + (b[5]<<2) - 29*b[6] + 43*b[7];
    blk->val[i + tsize * 15] = (b[0]<<2) - 13*b[1] + 21*b[2] - 29*b[3] + 35*b[4] - 40*b[5] + 43*b[6] - 45*b[7];

    b[0] = a[0] - a[7] - a[8]  + a[15];
    b[1] = a[1] - a[6] - a[9]  + a[14];
    b[2] = a[2] - a[5] - a[10] + a[13];
    b[3] = a[3] - a[4] - a[11] + a[12];

    blk->val[i + tsize * 2]  = 44*b[0] + 38*b[1] + 25*b[2] +  9*b[3];
    blk->val[i + tsize * 6]  = 38*b[0] -  9*b[1] - 44*b[2] - 25*b[3];
    blk->val[i + tsize * 10] = 25*b[0] - 44*b[1] +  9*b[2] + 38*b[3];
    blk->val[i + tsize * 14] =  9*b[0] - 25*b[1] + 38*b[2] - 44*b[3];

    b[0] = a[0] - a[3] - a[4] + a[7] + a[8] - a[11] - a[12] + a[15];
    b[1] = a[1] - a[2] - a[5] + a[6] + a[9] - a[10] - a[13] + a[14];
    blk->val[i + tsize * 4]  = 42*b[0] + 17*b[1];
    blk->val[i + tsize * 12] = 17*b[0] - 42*b[1];

    for (j = 0; j < tsize; j ++)
      blk->val[i + tsize * j] = (blk->val[i + tsize * j] + shift2) >> scale; 
  }
}
//-----------------------------------------------------------------------------
/*!
* \brief Inverse transform
*
* \param[in,out] blk Block of data to transform
* \param[in] TransType transform type
*/
//-----------------------------------------------------------------------------
void FastDCT::inverse_transform(Block *blk)
{
  int i, j;
  int tmp[256];
  int a[256];
  int T[256];
  int tsize;
  int scale, shift2;
  
  tsize = 16;
  assert(blk->get_height() == tsize && blk->get_width() == tsize);
  
  scale = 7;
  shift2 = (1 << (scale - 1));

  int aa[2], b[8], c[4], d[8], e[2], f[8], g[4], h[8];

  for (i = 0; i < tsize; i ++) 
  {
    for (j = 0; j < tsize; j ++ )   
      a[j] = blk->val[tsize * i + j];

    aa[0] = (a[0] + a[8]) << 5;
    aa[1] = (a[0] - a[8]) << 5;

    b[0] = 45*a[1];
    b[1] = b[0] - a[1] - a[1];
    b[2] = 40*a[1];
    b[3] = 35*a[1];
    b[4] = 29*a[1];
    b[5] = 21*a[1];
    b[6] = 13*a[1];
    b[7] =  a[1]<<2;

    c[0] = 44*a[2];
    c[1] = 38*a[2];
    c[2] = 25*a[2];
    c[3] = 9*a[2];

    d[0] = 43*a[3];
    d[1] = 29*a[3];
    d[2] = a[3]<<2;
    d[3] = 21*a[3];
    d[4] = 40*a[3];
    d[5] = d[0] + a[3] + a[3];
    d[6] = 35*a[3];
    d[7] = 13*a[3];

    e[0] = 42*a[4];
    e[1] = 17*a[4];

    f[0] = 40*a[5];
    f[1] = a[5]<<2;
    f[2] = 35*a[5];
    f[3] = 43*a[5];
    f[4] = 13*a[5];
    f[5] = 29*a[5];
    f[6] = f[3] + a[5] + a[5];
    f[7] = 21*a[5];

    g[0] = 38*a[6];
    g[1] = 9*a[6];
    g[2] = 44*a[6];
    g[3] = 25*a[6];

    h[0] = 35*a[7];
    h[1] = 21*a[7];
    h[2] = 43*a[7];
    h[3] = a[7]<<2;
    h[4] = h[2] + a[7] + a[7];
    h[5] = 13*a[7];
    h[6] = 40*a[7];
    h[7] = 29*a[7];

    tmp[i            ]  =  aa[0] + b[0] + c[0] + d[0] + e[0] + f[0] + g[0] + h[0];
    tmp[i + tsize    ]  =  aa[1] + b[1] + c[1] + d[1] + e[1] + f[1] - g[1] - h[1];
    tmp[i + tsize * 2]  =  aa[1] + b[2] + c[2] + d[2] - e[1] - f[2] - g[2] - h[2];
    tmp[i + tsize * 3]  =  aa[0] + b[3] + c[3] - d[3] - e[0] - f[3] - g[3] + h[3];
    tmp[i + tsize * 4]  =  aa[0] + b[4] - c[3] - d[4] - e[0] - f[4] + g[3] + h[4];
    tmp[i + tsize * 5]  =  aa[1] + b[5] - c[2] - d[5] - e[1] + f[5] + g[2] + h[5];
    tmp[i + tsize * 6]  =  aa[1] + b[6] - c[1] - d[6] + e[1] + f[6] + g[1] - h[6];
    tmp[i + tsize * 7]  =  aa[0] + b[7] - c[0] - d[7] + e[0] + f[7] - g[0] - h[7];
    tmp[i + tsize * 8]  =  aa[0] - b[7] - c[0] + d[7] + e[0] - f[7] - g[0] + h[7];
    tmp[i + tsize * 9]  =  aa[1] - b[6] - c[1] + d[6] + e[1] - f[6] + g[1] + h[6];
    tmp[i + tsize * 10] =  aa[1] - b[5] - c[2] + d[5] - e[1] - f[5] + g[2] - h[5];
    tmp[i + tsize * 11] =  aa[0] - b[4] - c[3] + d[4] - e[0] + f[4] + g[3] - h[4];
    tmp[i + tsize * 12] =  aa[0] - b[3] + c[3] + d[3] - e[0] + f[3] - g[3] - h[3];
    tmp[i + tsize * 13] =  aa[1] - b[2] + c[2] - d[2] - e[1] + f[2] - g[2] + h[2];
    tmp[i + tsize * 14] =  aa[1] - b[1] + c[1] - d[1] + e[1] - f[1] - g[1] + h[1];
    tmp[i + tsize * 15] =  aa[0] - b[0] + c[0] - d[0] + e[0] - f[0] + g[0] - h[0];

    b[0] = 29*a[9];
    b[1] = 40*a[9];
    b[2] = 13*a[9];
    b[3] = 45*a[9];
    b[4] = a[9]<<2;
    b[5] = b[3] - a[9] - a[9];
    b[6] = 21*a[9];
    b[7] = 35*a[9];

    c[0] = 25*a[10];
    c[1] = 44*a[10];
    c[2] = 9*a[10];
    c[3] = 38*a[10];

    d[0] = 21*a[11];
    d[1] = 45*a[11];
    d[2] = 29*a[11];
    d[3] = 13*a[11];
    d[4] = d[1] - a[11] - a[11];
    d[5] = 35*a[11];
    d[6] = a[11]<<2;
    d[7] = 40*a[11];

    e[0] = 17*a[12];
    e[1] = 42*a[12];

    f[0] = 13*a[13];
    f[1] = 35*a[13];
    f[2] = 45*a[13];
    f[3] = 40*a[13];
    f[4] = 21*a[13];
    f[5] = a[13]<<2;
    f[6] = 29*a[13];
    f[7] = f[2] - a[13]- a[13];

    g[0] = 9*a[14];
    g[1] = 25*a[14];
    g[2] = 38*a[14];
    g[3] = 44*a[14];

    h[0] = a[15]<<2;
    h[1] = 13*a[15];
    h[2] = 21*a[15];
    h[3] = 29*a[15];
    h[4] = 35*a[15];
    h[5] = 40*a[15];
    h[6] = 43*a[15];
    h[7] = h[6] + a[15] + a[15];

    tmp[i            ]  +=    b[0] + c[0] + d[0] + e[0] + f[0] + g[0] + h[0];
    tmp[i + tsize    ]  +=  - b[1] - c[1] - d[1] - e[1] - f[1] - g[1] - h[1];
    tmp[i + tsize * 2]  +=  - b[2] + c[2] + d[2] + e[1] + f[2] + g[2] + h[2];
    tmp[i + tsize * 3]  +=  + b[3] + c[3] + d[3] - e[0] - f[3] - g[3] - h[3];
    tmp[i + tsize * 4]  +=  - b[4] - c[3] - d[4] - e[0] + f[4] + g[3] + h[4];
    tmp[i + tsize * 5]  +=  - b[5] - c[2] + d[5] + e[1] + f[5] - g[2] - h[5];
    tmp[i + tsize * 6]  +=  + b[6] + c[1] + d[6] - e[1] - f[6] + g[1] + h[6];
    tmp[i + tsize * 7]  +=  + b[7] - c[0] - d[7] + e[0] + f[7] - g[0] - h[7];
    tmp[i + tsize * 8]  +=  - b[7] - c[0] + d[7] + e[0] - f[7] - g[0] + h[7];
    tmp[i + tsize * 9]  +=  - b[6] + c[1] - d[6] - e[1] + f[6] + g[1] - h[6];
    tmp[i + tsize * 10] +=  + b[5] - c[2] - d[5] + e[1] - f[5] - g[2] + h[5];
    tmp[i + tsize * 11] +=  + b[4] - c[3] + d[4] - e[0] - f[4] + g[3] - h[4];
    tmp[i + tsize * 12] +=  - b[3] + c[3] - d[3] - e[0] + f[3] - g[3] + h[3];
    tmp[i + tsize * 13] +=  + b[2] + c[2] - d[2] + e[1] - f[2] + g[2] - h[2];
    tmp[i + tsize * 14] +=  + b[1] - c[1] + d[1] - e[1] + f[1] - g[1] + h[1];
    tmp[i + tsize * 15] +=  - b[0] + c[0] - d[0] + e[0] - f[0] + g[0] - h[0];
  }

  for (i = 0; i < tsize; i ++) 
  {
    for (j = 0; j < tsize; j ++ )
      a[j] = (tmp[tsize * i + j] + shift2) >> scale;

    aa[0] = (a[0] + a[8]) << 5;
    aa[1] = (a[0] - a[8]) << 5;

    b[0] = 45*a[1];
    b[1] = b[0] - a[1] - a[1];
    b[2] = 40*a[1];
    b[3] = 35*a[1];
    b[4] = 29*a[1];
    b[5] = 21*a[1];
    b[6] = 13*a[1];
    b[7] =  a[1]<<2;

    c[0] = 44*a[2];
    c[1] = 38*a[2];
    c[2] = 25*a[2];
    c[3] = 9*a[2];

    d[0] = 43*a[3];
    d[1] = 29*a[3];
    d[2] = a[3]<<2;
    d[3] = 21*a[3];
    d[4] = 40*a[3];
    d[5] = d[0] + a[3] + a[3];
    d[6] = 35*a[3];
    d[7] = 13*a[3];

    e[0] = 42*a[4];
    e[1] = 17*a[4];

    f[0] = 40*a[5];
    f[1] = a[5]<<2;
    f[2] = 35*a[5];
    f[3] = 43*a[5];
    f[4] = 13*a[5];
    f[5] = 29*a[5];
    f[6] = f[3] + a[5] + a[5];
    f[7] = 21*a[5];

    g[0] = 38*a[6];
    g[1] = 9*a[6];
    g[2] = 44*a[6];
    g[3] = 25*a[6];

    h[0] = 35*a[7];
    h[1] = 21*a[7];
    h[2] = 43*a[7];
    h[3] = a[7]<<2;
    h[4] = h[2] + a[7] + a[7];
    h[5] = 13*a[7];
    h[6] = 40*a[7];
    h[7] = 29*a[7];

    T[i            ]  =  aa[0] + b[0] + c[0] + d[0] + e[0] + f[0] + g[0] + h[0];
    T[i + tsize    ]  =  aa[1] + b[1] + c[1] + d[1] + e[1] + f[1] - g[1] - h[1];
    T[i + tsize * 2]  =  aa[1] + b[2] + c[2] + d[2] - e[1] - f[2] - g[2] - h[2];
    T[i + tsize * 3]  =  aa[0] + b[3] + c[3] - d[3] - e[0] - f[3] - g[3] + h[3];
    T[i + tsize * 4]  =  aa[0] + b[4] - c[3] - d[4] - e[0] - f[4] + g[3] + h[4];
    T[i + tsize * 5]  =  aa[1] + b[5] - c[2] - d[5] - e[1] + f[5] + g[2] + h[5];
    T[i + tsize * 6]  =  aa[1] + b[6] - c[1] - d[6] + e[1] + f[6] + g[1] - h[6];
    T[i + tsize * 7]  =  aa[0] + b[7] - c[0] - d[7] + e[0] + f[7] - g[0] - h[7];
    T[i + tsize * 8]  =  aa[0] - b[7] - c[0] + d[7] + e[0] - f[7] - g[0] + h[7];
    T[i + tsize * 9]  =  aa[1] - b[6] - c[1] + d[6] + e[1] - f[6] + g[1] + h[6];
    T[i + tsize * 10] =  aa[1] - b[5] - c[2] + d[5] - e[1] - f[5] + g[2] - h[5];
    T[i + tsize * 11] =  aa[0] - b[4] - c[3] + d[4] - e[0] + f[4] + g[3] - h[4];
    T[i + tsize * 12] =  aa[0] - b[3] + c[3] + d[3] - e[0] + f[3] - g[3] - h[3];
    T[i + tsize * 13] =  aa[1] - b[2] + c[2] - d[2] - e[1] + f[2] - g[2] + h[2];
    T[i + tsize * 14] =  aa[1] - b[1] + c[1] - d[1] + e[1] - f[1] - g[1] + h[1];
    T[i + tsize * 15] =  aa[0] - b[0] + c[0] - d[0] + e[0] - f[0] + g[0] - h[0];

    b[0] = 29*a[9];
    b[1] = 40*a[9];
    b[2] = 13*a[9];
    b[3] = 45*a[9];
    b[4] = a[9]<<2;
    b[5] = b[3] - a[9] - a[9];
    b[6] = 21*a[9];
    b[7] = 35*a[9];

    c[0] = 25*a[10];
    c[1] = 44*a[10];
    c[2] = 9*a[10];
    c[3] = 38*a[10];

    d[0] = 21*a[11];
    d[1] = 45*a[11];
    d[2] = 29*a[11];
    d[3] = 13*a[11];
    d[4] = d[1] - a[11] - a[11];
    d[5] = 35*a[11];
    d[6] = a[11]<<2;
    d[7] = 40*a[11];

    e[0] = 17*a[12];
    e[1] = 42*a[12];

    f[0] = 13*a[13];
    f[1] = 35*a[13];
    f[2] = 45*a[13];
    f[3] = 40*a[13];
    f[4] = 21*a[13];
    f[5] = a[13]<<2;
    f[6] = 29*a[13];
    f[7] = f[2] - a[13]- a[13];

    g[0] = 9*a[14];
    g[1] = 25*a[14];
    g[2] = 38*a[14];
    g[3] = 44*a[14];

    h[0] = a[15]<<2;
    h[1] = 13*a[15];
    h[2] = 21*a[15];
    h[3] = 29*a[15];
    h[4] = 35*a[15];
    h[5] = 40*a[15];
    h[6] = 43*a[15];
    h[7] = h[6] + a[15] + a[15];

    T[i            ]  +=    b[0] + c[0] + d[0] + e[0] + f[0] + g[0] + h[0];
    T[i + tsize    ]  +=  - b[1] - c[1] - d[1] - e[1] - f[1] - g[1] - h[1];
    T[i + tsize * 2]  +=  - b[2] + c[2] + d[2] + e[1] + f[2] + g[2] + h[2];
    T[i + tsize * 3]  +=    b[3] + c[3] + d[3] - e[0] - f[3] - g[3] - h[3];
    T[i + tsize * 4]  +=  - b[4] - c[3] - d[4] - e[0] + f[4] + g[3] + h[4];
    T[i + tsize * 5]  +=  - b[5] - c[2] + d[5] + e[1] + f[5] - g[2] - h[5];
    T[i + tsize * 6]  +=    b[6] + c[1] + d[6] - e[1] - f[6] + g[1] + h[6];
    T[i + tsize * 7]  +=    b[7] - c[0] - d[7] + e[0] + f[7] - g[0] - h[7];
    T[i + tsize * 8]  +=  - b[7] - c[0] + d[7] + e[0] - f[7] - g[0] + h[7];
    T[i + tsize * 9]  +=  - b[6] + c[1] - d[6] - e[1] + f[6] + g[1] - h[6];
    T[i + tsize * 10] +=    b[5] - c[2] - d[5] + e[1] - f[5] - g[2] + h[5];
    T[i + tsize * 11] +=    b[4] - c[3] + d[4] - e[0] - f[4] + g[3] - h[4];
    T[i + tsize * 12] +=  - b[3] + c[3] - d[3] - e[0] + f[3] - g[3] + h[3];
    T[i + tsize * 13] +=    b[2] + c[2] - d[2] + e[1] - f[2] + g[2] - h[2];
    T[i + tsize * 14] +=    b[1] - c[1] + d[1] - e[1] + f[1] - g[1] + h[1];
    T[i + tsize * 15] +=  - b[0] + c[0] - d[0] + e[0] - f[0] + g[0] - h[0];

    for (j = 0; j < tsize; j ++)
      blk->val[i + tsize * j] = (T[i + tsize * j] + shift2) >> scale;
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
