/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Frame.cpp
*
* \brief Implementation of Frame class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Frame.H"
#include "Global.H"
#include "Param.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in luma sample units
* \param[in] w width of frame in luma sample units
*/
//-----------------------------------------------------------------------------
Frame::Frame(int h, int w)
{
  height = h;
  width = w;
  
  height2 = height >> ONE;
  width2  = width  >> ONE;
  height4 = height >> TWO;
  width4  = width  >> TWO;
  height8 = height >> THREE;
  width8  = width  >> THREE;
  
  size = width * height + TWO * height2 * width2;
  
  data = new unsigned char[size];
  luma = data;
  
  crcb[ZERO] = data + (width * height);
  crcb[ONE ] = data + (width * height) + (width2 * height2);
  
  
  sse_luma = ZERO;
  sse_chroma[ZERO] = ZERO;
  sse_chroma[ONE ] = ZERO;
  snr_luma = 0.0;
  snr_chroma[ZERO] = 0.0;
  snr_chroma[ONE ] = 0.0;
  mse_luma = 0.0;
  mse_chroma[ZERO] = 0.0;
  mse_chroma[ONE ] = 0.0;
  
  frame_no = ZERO;
  intermed_idx = 0;
  list_size[LIST_0] = ZERO;
  list_size[LIST_1] = ZERO;
  frame_order_counter = ZERO;
  qp = param.qp;
  delta_qp = ZERO;
  delta_qpC = ZERO;
  lq_offset = ZERO;
  frame_type = ft_I_frame;
  reference_idc = REFERENCE_PIC;
  me_mode = LIST_0;
  
  //  update_rd_params();
  clear_stats();
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
Frame::~Frame()
{
  delete[] data;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Clear frame
*
* Sets all luma samples to 0 and all chroma samples to MID_SAMPLE_VAL.
*/
//-----------------------------------------------------------------------------
void Frame::clear()
{
  memset(luma, ZERO, width * height * sizeof(unsigned char));
  memset(crcb[ZERO], MID_SAMPLE_VAL, TWO * width2 * height2 * sizeof(unsigned char)); // This works only since crcb is unsigned char
}

//-----------------------------------------------------------------------------
/*!
* \brief Clear statistics
*
* Clear stats included in ref_frame_stat array.
*/
//-----------------------------------------------------------------------------
void Frame::clear_stats()
{
  memset(ref_frame_stat, ZERO, TWO * MAX_REF_FRAMES * sizeof(int));
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy frame
*
* Copies data of frame f into current frame. Note that both frames should have
* the same size for this to work properly.
*
* \param[in] f %Frame from which data is copied
*/
//-----------------------------------------------------------------------------
void Frame::copy(Frame *f)
{
  if (size != f->size)
    fprintf(stderr, "Warning: trying to copy frames of different sizes.\n");
  
  mse_luma            = f->mse_luma;
  mse_chroma[ZERO]    = f->mse_chroma[ZERO];
  mse_chroma[ONE ]    = f->mse_chroma[ONE ];
  sse_luma            = f->sse_luma;
  sse_chroma[ZERO]    = f->sse_chroma[ZERO];
  sse_chroma[ONE ]    = f->sse_chroma[ONE ];
  snr_luma            = f->snr_luma;
  snr_chroma[ZERO]    = f->snr_chroma[ZERO];
  snr_chroma[ONE ]    = f->snr_chroma[ONE ];
  list_size[LIST_0]   = f->list_size[LIST_0];
  list_size[LIST_1]   = f->list_size[LIST_1];
  frame_type          = f->frame_type;
  intra_frame_type    = f->intra_frame_type;
  frame_no            = f->frame_no;
  intermed_idx        = f->intermed_idx;
  lq_offset           = f->lq_offset;
  lambdaq             = f->lambdaq;
  me_mode             = f->me_mode;
  qp                  = f->qp;
  qpC                 = f->qpC;
  qpDC                = f->qpDC;
  delta_qp            = f->delta_qp;
  delta_qpC           = f->delta_qpC;
  frame_order_counter = f->frame_order_counter;
  reference_idc       = f->reference_idc;
  melq                = f->melq;
  
  memcpy(data,           f->data,           size                 * sizeof(unsigned char));
  memcpy(qp_matrix,      f->qp_matrix,      4 * 16               * sizeof(int          ));
  if (param.trans8x8_enable) 
  {
    memcpy(qp_matrix8x8,      f->qp_matrix8x8,      4 * 64               * sizeof(int          ));
  }  
  if (param.intra16x16)
    memcpy(qp_matrix16x16,    f->qp_matrix16x16,    4 * 256 * sizeof(int));
  memcpy(ref_frame_stat, f->ref_frame_stat, TWO * MAX_REF_FRAMES * sizeof(int          ));
}

void Frame::copy_samples(Frame *src)
{
  memcpy(data, src->data, size * sizeof(unsigned char));
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute distortion
*
* Computes SSE, MSE and PSNR.
*
* \param[in] f Reference frame
* \return R-D cost
*/
//-----------------------------------------------------------------------------
double Frame::compute_distortion(Frame *f)
{
  int i;
  int d;
  int c;
  double sum = 0.0;
  
  for (i = ZERO; i < width * height; i++) 
  {
    d = luma[i] - f->luma[i];
    sum += d * d;
  }
  
  sse_luma = sum;
  mse_luma = sum / (width * height);
  snr_luma = psnr(mse_luma);
  
  for (c=0; c<2; c++) 
  {
    sum = 0.0;
    
    for (i=0; i<width2*height2; i++) 
    {
      d = crcb[c][i] - f->crcb[c][i];
      sum += d * d;
    }
    
    sse_chroma[c] = sum;
    mse_chroma[c] = sum / (width2 * height2);
    snr_chroma[c] = psnr(mse_chroma[c]);
  }
  
  return sse_luma * lambda + (sse_chroma[0] + sse_chroma[1]) * lambdaC;
}

//-----------------------------------------------------------------------------
/*!
* \brief Update rate-distortion optimization parameters
*
* Sets values of lambdaq, melq, qpC, qpDC, qp_matrix, lambdaq, lambdaqC,
* lambda, lambdaC based on the values of Param::lambdaq_base, Param::melq_base,
* Param::better_color, Param::lumaDC, qp, and lq_offset.
*
*/
//-----------------------------------------------------------------------------
void Frame::update_rd_params()
{
  int i,j;

#ifdef BETTER_CHROMA_DC
  int qpDC_C = qp; // Initialize to avoid compiler warning
  int intra_qpDC_C = qp;
#endif

  lambdaq = 2 * qp + param.lambdaq_base;
  
  melq = qp + param.melq_base;   
  
  // Lets keep this for the time being since we can later add support of QPs.  
#ifdef BETTER_CHROMA_DC
  if (param.better_color == 7) 
  {
    if(frame_type==ft_I_frame)
    { 
      qpC  = adjust_qp(param.better_color-1, 0) + param.bc_offset - 2;  
    }
    else if (reference_idc==DISPOSABLE_PIC) 
    { 
      qpC  = adjust_qp(param.better_color-1, 0) + param.bc_offset + 4;      
    }
    else 
    {
      qpC  = adjust_qp(param.better_color-1, 0) + param.bc_offset;      
    }
    qpDC = adjust_qp(param.lumaDC, 0);
    if (qpC > param.chroma_dc_th)
    {
      qpDC_C = qpC - (((qpC - param.chroma_dc_th) * param.chroma_dc_scale)>> 8);
    }
    else
    {
      qpDC_C = qpC;
    }

    //printf ("qpC =%d, qpDC_C=%d\n", qpC, qpDC_C);

  #ifdef INTRA_QPC_OFF
    intra_qpC = qpC;   
    intra_qpDC_C = qpDC_C;
  #else
    intra_qpC = adjust_qp(param.better_color-1, 0) + param.bc_offset - 2; 
    if (intra_qpC > param.chroma_dc_th)
    {
      intra_qpDC_C = intra_qpC - (((intra_qpC - param.chroma_dc_th) * param.chroma_dc_scale)>> 8);
    }
    else
    {
      intra_qpDC_C = intra_qpC;
    }
  #endif
  }
  else
  {
    if(frame_type==ft_I_frame)
    { 
      qpC  = adjust_qp(param.better_color, 0) + param.bc_offset - 2;  
    }
    else if (reference_idc==DISPOSABLE_PIC) 
    { 
      qpC  = adjust_qp(param.better_color, 0) + param.bc_offset + 4;      
    }
    else 
    {
      qpC  = adjust_qp(param.better_color, 0) + param.bc_offset;      
    }
    qpDC = adjust_qp(param.lumaDC, 0);

  #ifdef INTRA_QPC_OFF
    intra_qpC = qpC;
  #else
    intra_qpC = adjust_qp(param.better_color, 0) + param.bc_offset - 2; 
  #endif
  }
#else
  if(frame_type==ft_I_frame)
  { 
    qpC  = adjust_qp(param.better_color, 0) + param.bc_offset - 2;  
  }
  else if (reference_idc==DISPOSABLE_PIC) 
  { 
    qpC  = adjust_qp(param.better_color, 0) + param.bc_offset + 4;      
  }
  else 
  {
    qpC  = adjust_qp(param.better_color, 0) + param.bc_offset;      
  }
  qpDC = adjust_qp(param.lumaDC, 0);

#ifdef INTRA_QPC_OFF
  intra_qpC = qpC;
#else
  intra_qpC = adjust_qp(param.better_color, 0) + param.bc_offset - 2; 
#endif
#endif
  // Compute quantization matrices
  for (i=0; i<4; i++) 
  {
    for (j=0; j<4; j++) 
    {
      qp_matrix[0][4*i+j] = qp;
      qp_matrix[1][4*i+j] = intra_qpC; 
      qp_matrix[2][4*i+j] = qp;
      qp_matrix[3][4*i+j] = qpC;
    }
  }
  
  qp_matrix[0][0] = qpDC;
#ifdef BETTER_CHROMA_DC
  if (param.better_color == 7)
  {
    qp_matrix[1][0] = intra_qpDC_C;
    qp_matrix[3][0] = qpDC_C;
  }
#endif

  if (param.trans8x8_enable) 
  {
    for (i=0;i<64;i++)
    {
      qp_matrix8x8[0][i] = qp;
      qp_matrix8x8[1][i] = intra_qpC; 
      qp_matrix8x8[2][i] = qp;
      qp_matrix8x8[3][i] = qpC;
    }
    qp_matrix8x8[0][0] = qpDC;
#ifdef BETTER_CHROMA_DC
    if (param.better_color == 7)
    {
      qp_matrix8x8[1][0] = intra_qpDC_C;
      qp_matrix8x8[3][0] = qpDC_C;
    }
#endif
  }

  if (param.intra16x16)
  {
    for (i = 0; i < TX_NCOEFF16; i ++)
    {
      qp_matrix16x16[0][i] = qp;
      qp_matrix16x16[1][i] = intra_qpC; 
      qp_matrix16x16[2][i] = qp;
      qp_matrix16x16[3][i] = qpC;
    }

    qp_matrix16x16[0][0] = qpDC;
#ifdef BETTER_CHROMA_DC
    if (param.better_color == 7)
    {
      qp_matrix16x16[1][0] = intra_qpDC_C;
      qp_matrix16x16[3][0] = qpDC_C;
    }
#endif
  }

  lambdaq  = lambdaq + lq_offset; 
  lambdaqC = lambdaq - 2*(qp-qpC);
  
  lambda  = (float)pow(0.5,lambdaq/16.0);
  lambdaC = (float)pow(0.5,lambdaqC/16.0);
#ifdef USE_SSD
  if (frame_type==ft_P_frame && (!param.ssd_disp_only || reference_idc==DISPOSABLE_PIC)) 
    param.use_ssd = (qp > param.ssd_qp_P);
  else 
  if (frame_type==ft_B_frame && (!param.ssd_disp_only || reference_idc==DISPOSABLE_PIC)) 
    param.use_ssd = (qp > param.ssd_qp_B);
  else 
    param.use_ssd = 0;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Computes an adjusted QP value
*
* Returns an adjusted value of QP based on the value of qp and the adjustment
* type v. See better_color parameter in configuration file for adjustment
* function types.
*
* \param[in] v Adjustment type
* \param[in] offset Offset to be added to value of qp before mapping
* \return Adjusted QP value
*/
//-----------------------------------------------------------------------------
int Frame::adjust_qp(int v, int offset) 
{  
  int r;
  int qp_temp = qp;  
  qp += offset;     
  
  
  switch (v) 
  {
    case 1:
      r = shift_right_round(3 * qp, 2);
      break;
    case 2:
      r = qp - shift_right_round(qp * qp, 9);
      break;
    case 3:
      r = qp>>1;
      break;
    case 4:
      r = 32;
      break;
    case 5:
      r = qp - shift_right_round(qp * qp, 8);
      break;
    case 6:
      r = qp - shift_right_round(qp * qp * qp, 16);
      break;
#ifndef BETTER_CHROMA_DC
    case 7:
      if (qp < 64)
        r = qp;
      else
        r = qp - ((qp - 64) >> 1);
#endif
      break;
    default:
      r = qp;
      break;
  }
  
  qp = qp_temp;  
  
  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Adjusts QP value based on frame content
*
* Computes the standard deviation of samples within each 4x4 block in the frame
* (luma and chroma) and adjusts qp accordingly. %Quantizer step is adjusted 
* proportionally to the average standard deviation.
*
* \note This has an effect only when Param::qp_frame_mod is set to TRUE.
*/
//-----------------------------------------------------------------------------
void Frame::E()
{
  int i,j,c;
  int x, e, delta;
  float en, sum;
  Block blk(4, 4);
  
  if (param.qp_frame_mod == FALSE)
  {
    qp = clip(qp, 0, 127);
    return;
  }
  
  sum = 0.0;
  
  for (i=0; i<height4; i++)
    for (j=0; j<width4; j++) 
    {
      get_blk_luma(&blk, i, j);
      x = blk.sum();
      blk.mul(16);
      blk.sub(x);
      e = blk.sum2();
      en = (float) (sqrt((float)e)/16.0);
      sum += (float) (16.0*log(1.0+en)/log(2.0));
    }
  
  for (c=0; c<2; c++) 
  {
    for (i=0; i<height8; i++) 
    {
      for (j=0; j<width8; j++) 
      {
        get_blk_chroma(&blk, i, j, c);
        x = blk.sum();
        blk.mul(16);
        blk.sub(x);
        e = blk.sum2();
        en = (float) (sqrt((float)e)/16.0);
        sum += (float) (16.0*log(1.0+en)/log(2.0));
      }
    }
  }
  
  sum /= (width4*height8*3);
  
  delta = (int)(sum+0.5)-7*(16>>1);
  delta -= shift_right_round(delta, 1); // be a bit less agressive for now (actually might be good compromise)

  qp += delta;
  qp = clip(qp, 0, 127);
}

//-----------------------------------------------------------------------------
/*!
* \brief Retrieve sub-luma block (block-aligned)
*
* \param[out] sub-block (dimensions are specified in width+height)
* \param[in] i Vertical position (in block units x height)
* \param[in] j Horizontal position (in block units x width)
*/
//-----------------------------------------------------------------------------
void Frame::get_blk_luma(Block *blk, int i, int j)
{
  int k,l;
  int u,v;
  int gridx = blk->get_width();
  int gridy = blk->get_height();
   
  for (k = 0; k < min(gridy, height - gridy * i); k ++)
  {
    for (l = 0; l < min(gridx, width - gridx * j); l ++) 
    {
      u = gridy * i + k;
      v = gridx * j + l;
      blk->val[blk->get_stride() * k + l] = luma[width * u + v];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Retrieve sub-luma block (block-aligned)
 *
 * \param[out] sub-block (dimensions are specified in width+height)
 * \param[in] i Vertical position (in block units x height)
 * \param[in] j Horizontal position (in block units x width)
 * \param[in] c Chroma component index
 */
//-----------------------------------------------------------------------------
void Frame::get_blk_chroma(Block *blk, int i, int j, int c)
{
  int k,l;
  int u,v;
  int gridx = blk->get_width();
  int gridy = blk->get_height();
  
  for (k = 0; k < min(gridy, height2 - gridy * i); k ++)
  {
    for (l = 0; l < min(gridx, width2 - gridx * j); l ++) 
    {
      u = gridy * i + k;
      v = gridx * j + l;
      blk->val[blk->get_stride() * k + l] = crcb[c][width2 * u + v];
    }
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
