/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file IEncoder6.cpp
* \brief Implementation of IEncoder6 class
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "IEncoder6.H"
#include "Quantizer.H"
#include "Transform.H"
#include "FastDCT.H"
#include "TransformAlternate.H"
#include "Param.H"
#include "EntropyModel.H"
#include "ContextMap.H"
#include "Global.H"
#include "WFrame.H"
#include "IntraPrediction.H"
#ifdef TRACE_DEBUG
#include "Trace.H"
#endif

//-----------------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] height Height of picture in sample units
* \param[in] width Width of picture in sample units
* \param[in] q_luma Pointer to luma quantizer
* \param[in] q_chroma Pointer to chroma quantizer
*/
//-----------------------------------------------------------------------------
IEncoder6::IEncoder6(int height, int width, Quantizer *q_luma, Quantizer *q_chroma, ContextMap *zmap_luma, ContextMap *zmap_chroma0, ContextMap *zmap_chroma1)
: best_coeff_8x8(8, 8)
{
  m_cur2 = NULL;
  m_cur2_pred = NULL;
  m_cur2_qpe = NULL;
  m_target = NULL;  
  store = new PredBlkI[width*height>>6];
  store_ipred_img = new WFrame(height, width); 
  ipic_store = new PredBlkI[width*height>>6];
  tmp_mode_y = new int[width*height>>4];
  intra_pred = new IntraPrediction(height, width);
#ifdef SIPNxN_FRAME_STAT
  int kk, ll;
  for (ll=0; ll<11; ll++) 
  {
    for (kk=0;kk<16;kk++) ss_4x4[ll][kk] = 0.0;
    ss_4x4_count[ll]=0;
  }
  for (ll=0; ll<11; ll++) 
  { 
    for (kk=0;kk<16;kk++) ss_2x8[ll][kk] = 0.0;
    ss_2x8_count[ll]=0;
  }
  for (ll=0; ll<11; ll++) 
  {
    for (kk=0;kk<16;kk++) ss_8x2[ll][kk] = 0.0;
    ss_8x2_count[ll]=0;
  }
  for (ll=0; ll<11; ll++) 
  {
    for (kk=0;kk<64;kk++) ss_8x8[ll][kk] = 0.0;
    ss_8x8_count[ll]=0;
  }
#endif

  // create objects
  m_trans[Transform::DCT] = new FastDCT;
  m_trans[Transform::AlternateIntra] = new TransformAlternate(ft_I_frame);
  quant_luma = q_luma;
  quant_chroma = q_chroma;

  zmap = zmap_luma;
  zmap_chroma[0] = zmap_chroma0; 
  zmap_chroma[1] = zmap_chroma1; 

  stored_eenc.reset(&stored_buff);

  stored_ztree_luma = new ZeroTree(param.ztree_pattern[EntropyModel::I_luma]);
  stored_ztree_chroma = new ZeroTree(param.ztree_pattern[EntropyModel::I_chroma]);
  stored_ztree_luma_8x8 = new ZeroTree(param.ztree_pattern_64);

  rd_eenc.reset(&rd_buff);

  rd_ztree_luma = new ZeroTree(param.ztree_pattern[EntropyModel::I_luma]);
  rd_ztree_chroma = new ZeroTree(param.ztree_pattern[EntropyModel::I_chroma]);

  rd_ztree_luma_8x8 = new ZeroTree(param.ztree_pattern_64);
  
  int i;
  if (param.intra16x16)
  {
    rd_eenc_tmp16.reset(&rd_buff_tmp16);
    for(i=0; i<NSUBZTREES; i++)  
    {
      rd_subztrees_luma[i] = new ZeroTree(param.ztree_pattern[NZTREES-NSUBZTREES+i]);
      stored_subztrees_luma[i] = new ZeroTree(param.ztree_pattern[NZTREES-NSUBZTREES+i]);
    }
  }
  if (param.chroma8x8)
  {
    stored_ztree_chroma_8x8 = new ZeroTree(param.ztree_pattern_64);
    rd_ztree_chroma_8x8 = new ZeroTree(param.ztree_pattern_64);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*/
//-----------------------------------------------------------------------------
IEncoder6::~IEncoder6() 
{
  // delete objects
  delete m_trans[Transform::DCT];
  delete m_trans[Transform::AlternateIntra];

  delete this->stored_ztree_luma;
  delete this->stored_ztree_chroma;
  delete this->stored_ztree_luma_8x8;

  delete this->rd_ztree_luma;
  delete this->rd_ztree_chroma;
  delete this->rd_ztree_luma_8x8;
  if (param.chroma8x8)
  {
    delete this->stored_ztree_chroma_8x8;
    delete this->rd_ztree_chroma_8x8;
  }

  if (param.intra16x16)
  {
    int i;
    for(i=0; i<NSUBZTREES; i++)  
    {
      delete rd_subztrees_luma[i];
      delete stored_subztrees_luma[i];
    }
  }
  
  delete intra_pred;
  delete [] store; 
  delete store_ipred_img;   
  delete [] tmp_mode_y; 
  delete[] ipic_store;  
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

void IEncoder6::setFrames(WFrame *cur2, WFrame *cur2_pred, WFrame *cur2_qpe, WFrame *target)
{
  m_cur2 = cur2;
  m_cur2_pred = cur2_pred;
  m_cur2_qpe = cur2_qpe;
  m_target = target;
}

void IEncoder6::setMotionField(MotionField *motionField)
{
  m_motionField = motionField;
}

//-----------------------------------------------------------------------------
/*!
* \brief Reset zero block maps.
*/
//-----------------------------------------------------------------------------
void IEncoder6::reset_zmap()
{
  zmap->reset();
  zmap_chroma[0]->reset();  
  zmap_chroma[1]->reset(); 
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode chroma 8x8 block based on luma sip_type
*
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in,out] eenc Pointer to entropy encoder
* \param[in,out] mlc Pointer to MLC statistics
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] tmp_ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
void IEncoder6::encode_chroma_block_8x8(int i, int j, EntropyEncoder *eenc, MLC_Stat *mlc)
{
  assert(m_cur2 != NULL);
  int blk_height = 2;  // in 4x4 unit = 2x4 = 8 pixels
  int blk_width  = 2;  // in 4x4 unit = 2x4 = 8 pixels
  if (param.chroma8x8 && (intra_pred->get_sip_type(i,j) == SIP16x16))
  {
    if ((i <= m_cur2->height8-2) && (j <= m_cur2->width8-2))
    {
      encode_chroma_blocks(i>>1, j>>1, eenc, mlc, 1, NULL, SIP8x8);
      mlc->bit_count_chroma += eenc->bits_chroma;
    }
  }
  else
  {
    for(int ii=0; ii<blk_height; ii++)
    {
      int iii = i + ii;
      for(int jj=0; jj<blk_width; jj++)
      {
        int jjj = j + jj;

        if ((iii > m_cur2->height8-1) || (jjj > m_cur2->width8-1))
          continue;

        encode_chroma_blocks(iii, jjj, eenc, mlc, 1, NULL, SIP4x4);
        mlc->bit_count_chroma += eenc->bits_chroma;
      } // jj
    } // ii
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode a chroma 4x4 block
*
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in] c Chroma component index
* \param[in,out] eenc Pointer to entropy encoder
* \param[in,out] mlc Pointer to mlc stats
* \param[in] type for RDOQ
* \param[in] tmp_ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
void IEncoder6::encode_chroma_blocks(int i, int j, EntropyEncoder *eenc, MLC_Stat *mlc, int type, ZeroTree *tmp_ztree, int sip_type)
{
  assert(m_cur2 != NULL);
  int count1, count2;
  int c, c_temp;
  int mode;
  int is_zero_blk;
  int emodel = EntropyModel::I_chroma;
  int up_mode = INT_INF;
  int left_mode = INT_INF;
  int up_is_inter = INT_INF; 
  int left_is_inter = INT_INF;  
  Block blk[2];
  Block blk_org[2];
  double  lambda_mode;

  if (type)
    tmp_ztree = (sip_type == SIP8x8) ? param.ztree_64_CT8 : param.ztree[emodel];
  
  if (sip_type == SIP8x8)
    rd_ztree_chroma_8x8->copy_coding_state(tmp_ztree);

  for (count1 = 0; count1 < 2; count1++) 
  {
    for (count2 = 0; count2 < IntraPrediction::NO_INTRA_PMODE_CHROMA; count2++) 
    {
      intra_pred->use_mode_chroma[count1][count2] = 0;
    }
  }

  for (c_temp = 0; c_temp < 2; c_temp++) 
  {
    blk[c_temp].set_type(sip_type);
    blk_org[c_temp].set_type(sip_type);
    intra_pred->generate_prediction_chroma(i, j, c_temp, sip_type, -1, NULL, iqpC, m_cur2);
    m_target->get_blk_chroma(&blk_org[c_temp], i, j, c_temp); 
  }

  int i4 = (sip_type == SIP8x8) ? 2 * i : i;
  int j4 = (sip_type == SIP8x8) ? 2 * j : j;
  
  if (i4 > 0) 
  {
    up_mode   = intra_pred->get_mode_u(i4-1, j4);     // mode_u[(i - 1) * cur2->width8 + j]; 
    up_is_inter = m_motionField->get_pblk(i4-1, j4)->intra == NULL; 
  }
  if (j4 > 0) 
  {
    left_mode = intra_pred->get_mode_u(i4, j4-1);     // mode_u[i * cur2->width8 + (j - 1)]; 
    left_is_inter = m_motionField->get_pblk(i4, j4-1)->intra == NULL; 
  }

  lambda_mode = this->calc_lambda_mode(1,0);
  
  mode = this->rd_mode_decision_chroma(eenc, blk, blk_org, i, j, up_mode, left_mode, up_is_inter, left_is_inter, lambda_mode, tmp_ztree); 

  intra_pred->put_mode_u(i4, j4, mode);
  if (sip_type == SIP8x8)
  {
    intra_pred->put_mode_u(i4, j4+1, mode);
    intra_pred->put_mode_u(i4+1, j4, mode);
    intra_pred->put_mode_u(i4+1, j4+1, mode);
  }

  if (type) 
  {
    if (mode == IntraPrediction::TMA_L5_PRED_C)
      mlc->intra_tma_L5c += blk_org[0].get_num_elements() >> 4;
    else if (mode == IntraPrediction::TMA_L9_PRED_C)
      mlc->intra_tma_L9c += blk_org[0].get_num_elements() >> 4;
  }
  {
    store[i4*m_cur2->width8 + j4].mode_chroma = mode;
    if (sip_type == SIP8x8)
    {
      store[i4*m_cur2->width8 + j4+1].mode_chroma = mode;
      store[(i4+1)*m_cur2->width8 + j4].mode_chroma = mode;
      store[(i4+1)*m_cur2->width8 + j4+1].mode_chroma = mode;
    }
  }
  
  for (c = 0; c < 2; c++)
  {
    sip_mode_count[0][c+1][mode] += blk_org[0].get_num_elements() >> 4;       

    is_zero_blk = blk[c].is_zero();
    zmap_chroma[c]->set(i4, j4, blk[c].nnz(), 1+(sip_type==SIP8x8), 1+(sip_type==SIP8x8)); 
  #ifdef TRACE_DEBUG
    if(type==1) 
    {
      ENABLE_ETRACE;
    }
      SPRINT_ETRACE("chroma intra");
  #endif
    eenc->bits_chroma = -eenc->get_nbits();
    if (c == 0) 
    { 
      eenc->encode_spatial_intra_prediction_chroma(mode, up_mode, left_mode, up_is_inter, left_is_inter);
    } 
    eenc->bits_chroma += eenc->get_nbits();

    // encode data
    eenc->encode_zblk(is_zero_blk, zmap_chroma[c]->getcontext(i4,j4), 1); 

    if (is_zero_blk == FALSE) 
    {
      if (sip_type == SIP8x8)
        eenc->encode_coeffs_8x8(tmp_ztree, &blk[c], emodel);  
      else        
        eenc->encode_coeffs(tmp_ztree, &blk[c], emodel);  
    }
  #ifdef TRACE_DEBUG
    if(type==1) 
    {
      DISABLE_ETRACE;
    }
  #endif

    quant_chroma->loadmatrix(emodel, c, 0, 0); // cq_start
    quant_chroma->loadmatrix(emodel, c, 0, 1); // cq_end

    quant_chroma->iquant(&blk[c], 0); 
    m_trans[Transform::DCT]->inverse(&blk[c]);

    Block &blk_pre = intra_pred->mprr_c_blk[c][sip_type==SIP8x8][mode];
    m_cur2_pred->put_blk_chroma(&blk_pre, i, j, c);
    m_cur2_qpe->put_blk_chroma(&blk[c], i, j, c);
    blk[c].add(blk_pre);
    m_cur2->put_blk_chroma(&blk[c], i, j, c);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decide the intra prediction mode to use for a luma 4x4 block
*
* \param[in,out] eenc Pointer to entropy encoder
* \param[out] min_blk Pointer to the prediction sample block
* \param[in] org_blk Pointer to the block to be predicted
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in] lambda for rd decision
* \param[in] ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
int IEncoder6::rd_mode_decision_luma(EntropyEncoder *eenc, Block *min_blk, Block *org_blk, int i, int j, double lambda, ZeroTree *ztree, int sip_type, double &min_rdcost)
{
  double distortion;
  double rdcost;
  min_rdcost = DBL_MAX;
  int min_mode = -1;
  int mode_count;
  Block perr(sip_type);
  Block coef(sip_type);
  Block trans(sip_type);
  int TransType;
  double lambda_rdoq;
  int izmap, jzmap;

  zmap->convert_indices(i, j, izmap, jzmap, sip_type);

  int doRDOQ = (sip_type == SIP8x8) ? param.q_rdoffset8x8 : param.q_rdoffset;
  int q_sip_type = (sip_type == SIP8x8) ? 0 : sip_type;

  lambda_rdoq = this->calc_lambda_mode(0, 1);  

  quant_luma->loadmatrix(EntropyModel::I_luma, -1, q_sip_type, 0); // cq_start
  quant_luma->loadmatrix(EntropyModel::I_luma, -1, q_sip_type, 1); // cq_end

  for (mode_count = 0; mode_count < IntraPrediction::NO_INTRA_PMODE_LUMA_MxN; mode_count++) 
  {
    if (intra_pred->use_mode_luma[mode_count] == 0) 
      continue; // skip mode for which no prediction is available

    for (TransType = 0; TransType < (param.transform_selection ? 2 : 1); TransType++)
    {
      Block &pred = intra_pred->mprr_blk[sip_type][mode_count];
      perr.copy(*org_blk);
      perr.sub(pred);

      int transIdx = (TransType < 1) ? Transform::DCT : Transform::AlternateIntra;
      m_trans[transIdx]->forward(&perr);

      if (doRDOQ == 1) 
      {
        quant_luma->quant_rdoq(&perr, eenc, EntropyModel::I_luma, lambda_rdoq, zmap->getcontext(izmap,jzmap), ztree, sip_type, TransType); 
      }
      else 
      {
        trans.copy(perr);
        quant_luma->quant_loop(&perr, q_sip_type, TransType); 
      }

      coef.copy(perr);
      quant_luma->iquant(&perr, q_sip_type);
      m_trans[transIdx]->inverse(&perr);

      perr.add(pred);
      distortion = perr.ssd(*org_blk);
      if (distortion > min_rdcost) 
        continue; // Don't bother computing rate is distortion already exceeds R-D cost

      rdcost = distortion + lambda * get_rate_luma(eenc, &coef, izmap, jzmap, mode_count, sip_type);

      if (min_rdcost > rdcost) 
      {
        selected_transform = TransType;
        min_rdcost = rdcost;
        min_mode = mode_count;
        min_blk->copy((doRDOQ == 1) ? coef : trans);
      }
    }
  }    

  if (doRDOQ == 0)
    quant_luma->quant(min_blk, q_sip_type, selected_transform); 
  else if (doRDOQ == 2)
    quant_luma->quant_rdoq(min_blk, eenc, EntropyModel::I_luma, lambda_rdoq, zmap->getcontext(izmap,jzmap), ztree, sip_type, selected_transform); 

  assert(min_mode != -1);
  return (min_mode);
}

//-----------------------------------------------------------------------------
/*!
* \brief Decide the intra prediction mode to use for a chroma 4x4 block
*
* \param[in,out] eenc Pointer to entropy encoder
* \param[out] min_blk Pointer to the prediction sample block
* \param[in] org_blk Pointer to the block to be predicted
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in] up_mode mode of block above
* \param[in] left_mode mode of block to the left
* \param[in] up_is_inter flag indicating if the above block is inter
* \param[in] left_is_inter flag indicating if the left block is inter
* \param[in] lambda for rd decision
* \param[in] ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
int IEncoder6::rd_mode_decision_chroma(EntropyEncoder *eenc,
                                       Block       min_blk[2],
                                       Block       org_blk[2],
                                       const int      i,
                                       const int      j,
                                       const int      up_mode,
                                       const int      left_mode,
                                       const int      up_is_inter, 
                                       const int      left_is_inter,  
                                       double         lambda,
                                       ZeroTree        *ztree)
{
  double    distortion;
  double    rdcost;
  double    min_rdcost = DBL_MAX;
  int       min_mode = -1;
  int       mode_count;
  int       uv;
  int sip_type = org_blk[0].get_type();
  Block perr(sip_type);
  Block coef[2];
  Block trans[2]; 
  double lambda_rdoq;
  int doRDOQ = (sip_type == SIP8x8) ? param.q_rdoffset8x8 : param.q_rdoffset;
  int i4 = (sip_type == SIP8x8) ? 2*i : i;
  int j4 = (sip_type == SIP8x8) ? 2*j : j;

  lambda_rdoq = this->calc_lambda_mode(1, 1); 

  for (mode_count = 0; mode_count < IntraPrediction::NO_INTRA_PMODE_CHROMA; mode_count++) 
  {

    if (intra_pred->use_mode_chroma[0][mode_count] == 0) 
    { 
      continue;
    }
    
    rdcost = 0.0;

    for (uv = 0; uv < 2; uv++) 
    {
      coef[uv].set_type(sip_type);
      trans[uv].set_type(sip_type);
      
      perr.copy(org_blk[uv]);

      Block &pred = intra_pred->mprr_c_blk[uv][sip_type==SIP8x8][mode_count];

      perr.sub(pred);     
      m_trans[Transform::DCT]->forward(&perr);

      quant_chroma->loadmatrix(EntropyModel::I_chroma, uv, 0, 0); // cq_start
      quant_chroma->loadmatrix(EntropyModel::I_chroma, uv, 0, 1); // cq_end

      if (doRDOQ == 1) 
      {
        quant_chroma->quant_rdoq(&perr, eenc, EntropyModel::I_chroma, lambda_rdoq, zmap_chroma[uv]->getcontext(i4,j4), ztree, 0);
      }
      else 
      {
        trans[uv].copy(perr);
        quant_chroma->quant_loop(&perr, 0); 
      }

      coef[uv].copy(perr);

      quant_chroma->iquant(&perr, 0);
      m_trans[Transform::DCT]->inverse(&perr);
      perr.add(pred);
      distortion = perr.ssd(org_blk[uv]);
      rdcost += distortion; 
    }
    
    rdcost += lambda * get_rate_chroma(eenc, coef, i4, j4, mode_count, up_mode, left_mode, up_is_inter, left_is_inter); 
    
    if (min_rdcost > rdcost) 
    {
      min_rdcost = rdcost;
      min_mode = mode_count;
      for (uv = 0; uv < 2; uv++) 
      {
        if (doRDOQ == 1) 
        {
          min_blk[uv].copy(coef[uv]);
        }
        else 
        {
          min_blk[uv].copy(trans[uv]);
        }
      }
    }
  }

  for (uv = 0; uv < 2; uv++) 
  { 
    quant_luma->loadmatrix(EntropyModel::I_chroma, uv, 0, 0); // cq_start
    quant_luma->loadmatrix(EntropyModel::I_chroma, uv, 0, 1); // cq_end

    if (doRDOQ == 0)
      quant_chroma->quant(&min_blk[uv], 0); 
    else if (doRDOQ == 2)
      quant_chroma->quant_rdoq(&min_blk[uv], eenc, EntropyModel::I_chroma, lambda_rdoq, zmap_chroma[uv]->getcontext(i4,j4), ztree, 0);

  }

    for (uv = 0; uv < 2; uv++) 
    { 
      Block &pred = intra_pred->mprr_c_blk[uv][sip_type==SIP8x8][min_mode];
      perr.copy(org_blk[uv]);
      perr.sub(pred);
      store_ipred_img->put_blk_chroma(&perr, i, j, uv); 
    }

  return (min_mode);
}

//-----------------------------------------------------------------------------
/*!
* \brief get the rate of intra prediction for a luma 4x4 block
*
* \param[in,out] eenc Pointer to entropy encoder
* \param[in] coeff Pointer to the coefficient block
* \param[in] i Vertical position of block to encode (zmap coordinates)
* \param[in] j Horizontal position of block to encode (zmap coordinates)
* \param[in] cur_mode mode of block to get the rate for
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
double IEncoder6::get_rate_luma(EntropyEncoder *eenc, Block *coef, int i, int j, int cur_mode, int sip_type)
{
  int blkIdx = 2*(i&1) + (j&1);
  int is_zero_blk = coef->is_zero();

  int rate = eenc->cost_spatial_intra_prediction_luma(cur_mode, intra_pred->L_mode_pred_luma[blkIdx]);
  rate += eenc->cost_zblk(is_zero_blk, zmap->getcontext(i, j), EntropyModel::I_luma);
  
  if (!is_zero_blk)
  {
    if (sip_type == SIP8x8)
      rate += eenc->cost_coeffs_8x8(rd_ztree_luma_8x8, coef); 
    else
      rate += eenc->cost_coeffs(rd_ztree_luma, coef, EntropyModel::I_luma, sip_type);
  }
  
  return rate * (1.0 / power2(RATE_BITS)) + param.sl_rate_offset;
}

//-----------------------------------------------------------------------------
/*!
* \brief get the rate of intra prediction for a chroma 4x4 block
*
* \param[in,out] eenc Pointer to entropy encoder
* \param[in] uv_coef Pointer to the coefficient block of u and v
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in] cur_mode mode of block to get the rate for
* \param[in] up_mode mode of block above
* \param[in] left_mode mode of block to the left
* \param[in] up_is_inter flag indicating if the above block is inter
* \param[in] left_is_inter flag indicating if the left block is inter
*/
//-----------------------------------------------------------------------------
double IEncoder6::get_rate_chroma(EntropyEncoder *eenc, Block uv_coef[2], int i, int j, int cur_mode, int up_mode, int left_mode, int up_is_inter, int left_is_inter) 
{
  int rate;
  int emodel = EntropyModel::I_chroma;
  int is_zero_blk; 
  int uv; 

  rate = eenc->cost_spatial_intra_prediction_chroma(cur_mode, up_mode, left_mode, up_is_inter, left_is_inter); 

  for(uv=0; uv<2; uv++)
  { 
    is_zero_blk = uv_coef[uv].is_zero(); 
    
    rate += eenc->cost_zblk(is_zero_blk, zmap_chroma[uv]->getcontext(i, j), 1); 
    
    if (is_zero_blk == FALSE) 
    {
      if (uv_coef[0].get_type() == SIP8x8)
        rate += eenc->cost_coeffs_8x8(rd_ztree_chroma_8x8, uv_coef+uv, 1);
      else
        rate += eenc->cost_coeffs(param.ztree[emodel], uv_coef+uv, emodel);  
    }
  }
  
  return rate * (1.0 / power2(RATE_BITS)) + param.sl_rate_offset;
}


//-----------------------------------------------------------------------------
/*!
* \brief calculate the Lamda mode
*
* \param[in] chroma Flag for chroma
* \param[in] rdoq Flag for rdoq
*/
//-----------------------------------------------------------------------------
double IEncoder6::calc_lambda_mode(int chroma, int rdoq) 
{
  if(param.enable_simple_lambdas){
    int qp = (chroma) ? iqpC : iqp;
    double qp_jm = 6.0/16 * qp + 4.0;
    double k;
  
    if (frame_type == ft_I_frame)
    {
      if (chroma)
        k = (rdoq) ? param.sl_lambda_I_chroma_rdoq : param.sl_lambda_I_chroma;
      else
        k = (rdoq) ? param.sl_lambda_I_luma_rdoq : param.sl_lambda_I_luma;
    }
    else
    {
      if (chroma)
        k = (rdoq) ? param.sl_lambda_P_chroma_rdoq : param.sl_lambda_P_chroma;
      else
        k = (rdoq) ? param.sl_lambda_P_luma_rdoq : param.sl_lambda_P_luma;
    }
  
    return k * pow(2.0, (qp_jm - 12.0) / 3);
  
  }else{
    double  lambda_mode;
    double  qp_jm;
    double  offset;

    if (frame_type==ft_I_frame) 
    {
      if (rdoq) 
      {
        if (chroma) 
        { 
          offset = this->iqpC * this->iqpC * this->iqpC / 65536; 
          qp_jm = 6 * (double) (iqpC-offset) / 16 + 4; 
          if(param.onlyI)
            lambda_mode = 0.50 * pow(2.0, (double) (qp_jm - 12) / 3); 
          else
            lambda_mode = 0.3 * pow(2.0, (double) (qp_jm - 12) / 3); 
        }
        else 
        {
          offset = this->iqp * this->iqp * this->iqp / 65536; 
          qp_jm = 6 * (double) (iqp-offset) / 16 + 4;
          if (param.intra16x16)
            lambda_mode = 0.40 * pow(2.0, (double) (qp_jm - 12) / 3);
          else
            lambda_mode = 0.57 * pow(2.0, (double) (qp_jm - 12) / 3);
        } 
      }
      else 
      {
        if(chroma) 
        { 
          qp_jm = 6 * (double) iqpC / 16 + 4; 
          if(param.q_rdoffset)
            lambda_mode = 0.57 * pow(2.0, (double) (qp_jm - 12) / 3); 
          else
            lambda_mode = 0.85 * pow(2.0, (double) (qp_jm - 12) / 3); 
        }
        else 
        {
          qp_jm = 6 * (double) iqp / 16 + 4;
          if (param.intra16x16)
          {
            if (param.q_rdoffset)
              lambda_mode = 0.40 * pow(2.0, (double) (qp_jm - 12) / 3);
            else
              lambda_mode = 0.60 * pow(2.0, (double) (qp_jm - 12) / 3);
          }
          else
          {
            if (param.q_rdoffset)
              lambda_mode = 0.57 * pow(2.0, (double) (qp_jm - 12) / 3);
            else
              lambda_mode = 0.85 * pow(2.0, (double) (qp_jm - 12) / 3);
          }
        } 
      }
    }
    else 
    {
      if (rdoq) 
      {
        if (chroma) 
        { 
          qp_jm = 6 * (double) iqpC / 16 + 4; 
          lambda_mode = 0.30 * pow(2.0, (double) (qp_jm - 12) / 3); 
        }
        else 
        {
          qp_jm = 6 * (double) iqp / 16 + 4;
          lambda_mode = 0.57 * pow(2.0, (double) (qp_jm - 12) / 3);
        } 
      }
      else 
      {
        if (chroma) 
        { 
          qp_jm = 6 * (double) iqpC / 16 + 4; 
          lambda_mode = 0.85 * pow(2.0, (double) (qp_jm - 12) / 3); 
        }
        else 
        {
          qp_jm = 6 * (double) iqp / 16 + 4;
          lambda_mode = 0.85 * pow(2.0, (double) (qp_jm - 12) / 3);
        } 
      }
    }

    return (lambda_mode);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Reset SIP mode counters
*/
//-----------------------------------------------------------------------------
void IEncoder6::reset_sip_mode_count()
{
  int i, j;
  for (j=0; j<NSIP; j++ ) // number of SIP 
  { 
    for (i=0; i<IntraPrediction::NO_INTRA_PMODE_LUMA_MxN; i++ ) 
    {
      sip_mode_count[j][0][i] = 0;       
      sip_mode_count[j][1][i] = 0;       
      sip_mode_count[j][2][i] = 0;       
      sip_min_mode_count[j][0][i] = 0;       
      sip_min_mode_count[j][1][i] = 0;       
      sip_min_mode_count[j][2][i] = 0;       
      sip_sel_mode_count[j][0][i] = 0;      
      sip_sel_mode_count[j][1][i] = 0;       
      sip_sel_mode_count[j][2][i] = 0;  
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Print SIP mode counters
*/
//-----------------------------------------------------------------------------
void IEncoder6::print_sip_mode_count() const
{
  int i, sum=0, min=0, sel=0;

  printf ("sip mode count 4x4 luma\n");
  for (i=0; i<IntraPrediction::NO_INTRA_PMODE_LUMA_MxN; i++ ) 
  {
    printf ("%d\tsel %5d\tL-Mode\t%5d\tout of\t%5d\t(%.2f)\n", i, sip_sel_mode_count[SIP4x4][0][i], sip_min_mode_count[SIP4x4][0][i], sip_mode_count[SIP4x4][0][i], (float)sip_min_mode_count[SIP4x4][0][i]/sip_mode_count[SIP4x4][0][i]);
    sum += sip_mode_count[SIP4x4][0][i];
    min += sip_min_mode_count[SIP4x4][0][i];
    sel += sip_sel_mode_count[SIP4x4][0][i];
  }
  printf ("sel = %d\tL-Mode = %d out of %d\n", sel, min, sum);

  sum=0;
  printf ("sip mode count chroma\n");
  for (i=0; i < IntraPrediction::NO_INTRA_PMODE_CHROMA; i++ ) 
  {
    printf ("%d\t%d (%.2f)\n", i, sip_mode_count[SIP4x4][1][i], (float)sip_min_mode_count[SIP4x4][1][i]/sip_mode_count[SIP4x4][1][i]);
    sum += sip_mode_count[SIP4x4][1][i];
  }
  printf ("sum = %d\n", sum);

  sum=0;
  min=0;
  sel=0;
  printf ("sip mode count 2x8\n");
  for (i=0; i<IntraPrediction::NO_INTRA_PMODE_LUMA_MxN; i++ ) 
  {
    printf ("%d\tsel %5d\tL-Mode\t%5d\tout of\t%5d\t(%.2f)\n", i, sip_sel_mode_count[SIP2x8][0][i], sip_min_mode_count[SIP2x8][0][i], sip_mode_count[SIP2x8][0][i], (float)sip_min_mode_count[SIP2x8][0][i]/sip_mode_count[SIP2x8][0][i]);
    sum += sip_mode_count[SIP2x8][0][i];
    min += sip_min_mode_count[SIP2x8][0][i];
    sel += sip_sel_mode_count[SIP2x8][0][i];
  }
  printf ("sel = %d\tL-Mode = %d out of %d\n", sel, min, sum);

  sum=0;
  min=0;
  sel=0;
  printf ("sip mode count 8x2\n");
  for (i=0; i<IntraPrediction::NO_INTRA_PMODE_LUMA_MxN; i++ ) 
  {
    printf ("%d\tsel %5d\tL-Mode\t%5d\tout of\t%5d\t(%.2f)\n", i, sip_sel_mode_count[SIP8x2][0][i], sip_min_mode_count[SIP8x2][0][i], sip_mode_count[SIP8x2][0][i], (float)sip_min_mode_count[SIP8x2][0][i]/sip_mode_count[SIP8x2][0][i]);
    sum += sip_mode_count[SIP8x2][0][i];
    min += sip_min_mode_count[SIP8x2][0][i];
    sel += sip_sel_mode_count[SIP8x2][0][i];
  }
  printf ("sel = %d\tL-Mode = %d out of %d\n", sel, min, sum);

  sum=0;
  min=0;
  sel=0;
  printf ("sip mode count 8x8\n");
  for (i=0; i<IntraPrediction::NO_INTRA_PMODE_LUMA_MxN; i++ ) 
  {
    printf ("%d\tsel %5d\tL-Mode\t%5d\tout of\t%5d\t(%.2f)\n", i, sip_sel_mode_count[SIP8x8][0][i], sip_min_mode_count[SIP8x8][0][i], sip_mode_count[SIP8x8][0][i], (float)sip_min_mode_count[SIP8x8][0][i]/sip_mode_count[SIP8x8][0][i]);
    sum += sip_mode_count[SIP8x8][0][i];
    min += sip_min_mode_count[SIP8x8][0][i];
    sel += sip_sel_mode_count[SIP8x8][0][i];
  }
  printf ("sel = %d\tL-Mode = %d out of %d\n", sel, min, sum);

  sum=0;
  min=0;
  sel=0;
  if (param.intra16x16)
  {
    printf ("sip mode count 16x16\n");
    for (i=0; i<IntraPrediction::NO_INTRA_PMODE_LUMA_MxN; i++ ) 
    {
      printf ("%d\tsel %5d\tL-Mode\t%5d\tout of\t%5d\t(%.2f)\n", i, sip_sel_mode_count[SIP16x16][0][i], sip_min_mode_count[SIP16x16][0][i], sip_mode_count[SIP16x16][0][i], (float)sip_min_mode_count[SIP16x16][0][i]/sip_mode_count[SIP16x16][0][i]);
      sum += sip_mode_count[SIP16x16][0][i];
      min += sip_min_mode_count[SIP16x16][0][i];
      sel += sip_sel_mode_count[SIP16x16][0][i];
    }
    printf ("sel = %d\tL-Mode = %d out of %d\n", sel, min, sum);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode a luma 16x16 block
*
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in,out] eenc Pointer to entropy encoder
* \param[in,out] mlc Pointer to mlc stats
* \param[in] type for RDOQ
* \param[in] tmp_ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
void IEncoder6::encode_luma_block_16x16(int i, int j, EntropyEncoder *eenc, MLC_Stat *mlc, int type, ZeroTree *tmp_ztree, ZeroTree *tmp_ztree_64, ZeroTree **tmp_subztrees)
{
  assert(m_cur2 != NULL);
    int ii, jj, iii, jjj;
    int blk_height = 2;  // in 8x8 unit = 2x8 = 16 pixels
    int blk_width  = 2;  // in 8x8 unit = 2x8 = 16 pixels
    int log2_blksize = 1;
    int istep = blk_height>>1;
    int jstep = blk_width>>1;
    int sip_type16;
    double tot_rdcost8 = 0;

    int rate8 = 0;
    double dist8 = 0;

    int subtree_i;

    Block cur2_blk(blk_height<<3, blk_width<<3); //array to keep the reconstructed Block 
    Block cur2_blk_pre(blk_height<<3, blk_width<<3);
    Block cur2_blk_qpe(blk_height<<3, blk_width<<3);
    Block coeff(blk_height<<3, blk_width<<3);
    Block blk_pred_best(blk_height<<3, blk_width<<3);
    Block blk_org(blk_height<<3, blk_width<<3);

    m_target->get_blk_luma(&blk_org, i>>1, j>>1);

    quant_luma->I_save(QUANT_OFFSET_STARTING);      // save the quant offset at the starting of the 4 8x8 block coding

    rd_eenc_tmp16.copy_coding_state(eenc);

    ZeroTree *rd_ztree_luma_4;
    ZeroTree *rd_ztree_luma_8;
    rd_ztree_luma_4 = new ZeroTree(param.ztree_pattern[EntropyModel::I_luma]);
    rd_ztree_luma_8 = new ZeroTree(param.ztree_pattern_64);

    if(type)
    {
      rd_ztree_luma_4->copy_coding_state(param.ztree[EntropyModel::I_luma]);
      rd_ztree_luma_8->copy_coding_state(param.ztree_64);
    }
    else
    {
      rd_ztree_luma_4->copy_coding_state(tmp_ztree);
      rd_ztree_luma_8->copy_coding_state(tmp_ztree_64);
    } 

    rate8 = rd_eenc_tmp16.get_nbits();

    for(ii=0; ii<blk_height; ii+=istep)
    {
      iii = i + ii;
      for(jj=0; jj<blk_width; jj+=jstep)
      {
        jjj = j + jj;

        if ((iii > m_cur2->height8-1) || (jjj > m_cur2->width8-1))
          continue;

        if (intra_pred->avail_up_right_16x16)
        {
          intra_pred->avail_up_right_8x8 = (ii && jj)? 0 : 1;
        }
        else
        {
          intra_pred->avail_up_right_8x8 = ( jj )? 0 : 1;
        }

        encode_luma_block_NxN(iii<<1, jjj<<1, &rd_eenc_tmp16, mlc, 1, NULL, NULL);

        tot_rdcost8 += rdcost8;
        rd_eenc_tmp16.get_nbits();
        // store the coefficients for each 8x8 block
        best_coeff_16x16->insert(best_coeff_8x8, 8*ii, 8*jj);

        // store the L_mode_pred_luma to tmp_lmode_y
        intra_pred->copy_from_L_mode_pred_luma(ii, jj, istep, jstep);

      } // jj
    } // ii

    rd_eenc_tmp16.terminate();
    rate8 = rd_eenc_tmp16.get_nbits() - rate8;

    sip_type16 = 0;  // 8x8 and below types

    quant_luma->I_save(QUANT_OFFSET_BEST);

    intra_pred->copy_sip_type_from_NxN_type(i, j, i%blk_height, j%blk_width, blk_height, blk_width);
    intra_pred->copy_from_mode_y(intra_pred->Large_mode_y, i, j, blk_height, blk_width);

    m_cur2->get_blk_luma(&cur2_blk, i>>log2_blksize, j>>log2_blksize);
    m_cur2_pred->get_blk_luma  (&(cur2_blk_pre), i>>log2_blksize, j>>log2_blksize);
    m_cur2_qpe->get_blk_luma (&(cur2_blk_qpe), i>>log2_blksize, j>>log2_blksize);

      store_ipred_img->put_blk_luma(&blk_pred_best, i>>log2_blksize, j>>log2_blksize); 
    
    dist8 = cur2_blk.ssd(blk_org);
    tot_rdcost8 = dist8 + calc_lambda_mode(0, 0)*rate8;

    // Get the RD_Cost for 1 16x16 blocks.
    quant_luma->I_load(QUANT_OFFSET_STARTING);
    rd_eenc_tmp16.copy_coding_state(eenc);
    if(type)
    {
      param.ztree_64->copy_coding_state(rd_ztree_luma_8);
      rd_ztree_luma_8x8->copy_coding_state(param.ztree_64);
      for(subtree_i=0; subtree_i<NSUBZTREES; subtree_i++)
        rd_subztrees_luma[subtree_i]->copy_coding_state(param.ztree[NZTREES-NSUBZTREES+subtree_i]);
    }
    else
    {
      tmp_ztree_64->copy_coding_state(rd_ztree_luma_8);
      rd_ztree_luma_8x8->copy_coding_state(tmp_ztree_64);
      for(subtree_i=0; subtree_i<NSUBZTREES; subtree_i++)     
        rd_subztrees_luma[subtree_i]->copy_coding_state(tmp_subztrees[subtree_i]);
    } 

    intra_pred->set_sip_type(i, j, SIP16x16, blk_height, blk_width);

    rd_eenc_tmp16.entropycode_sip_type(SIP16x16, intra_pred->get_sip_ctx(i, j));

    if ((i > m_cur2->height8-2) || (j > m_cur2->width8-2))
      rdcost16 = DBL_MAX;
    else
    {
      if (type) 
      {
        rdcost16 = get_rdcost_luma_blk16x16(&coeff, i, j, &rd_eenc_tmp16);
      }
      else
      {
        rdcost16 = get_rdcost_luma_blk16x16(&coeff,i, j, &rd_eenc_tmp16);
      }
    }

    if (param.qmatrix_enable_flag[0] && param.qmatrix_enable_flag[1]) // cq_start
      rdcost16 *=INTRA16x16WEIGHTEDCOST/128.0; // cq_end

    // compare the rdcost of the whole 16x16 mode and the submode and update the best data to buffer
    if( rdcost16 <  tot_rdcost8 )
    {
      sip_type16 = 1;

      quant_luma->I_save(QUANT_OFFSET_BEST);  

      intra_pred->copy_sip_type_from_NxN_type(i, j, i%blk_height, j%blk_width, blk_height, blk_width);
      intra_pred->copy_from_mode_y(intra_pred->Large_mode_y, i, j, blk_height, blk_width);
      intra_pred->copy_from_L_mode_pred_luma(0, 0, blk_height, blk_width);

      best_coeff_16x16->copy( coeff );

      m_cur2->get_blk_luma (&(cur2_blk), i>>log2_blksize, j>>log2_blksize);

      m_cur2_pred->get_blk_luma  (&(cur2_blk_pre), i>>log2_blksize, j>>log2_blksize);
      m_cur2_qpe->get_blk_luma (&(cur2_blk_qpe), i>>log2_blksize, j>>log2_blksize);

        store_ipred_img->put_blk_luma(&blk_pred_best, i>>log2_blksize, j>>log2_blksize); 
    }

    rdcost16 = min(rdcost16, tot_rdcost8);

    // Restore everything to the best selected mode
    quant_luma->I_load(QUANT_OFFSET_BEST);
    intra_pred->copy_sip_type_to_NxN_type(i, j, i%blk_height, j%blk_width, blk_height, blk_width);
    intra_pred->copy_to_mode_y(intra_pred->Large_mode_y, i, j, blk_height, blk_width);

    m_cur2->put_blk_luma(&(cur2_blk), i>>log2_blksize, j>>log2_blksize);

    m_cur2_pred->put_blk_luma  (&cur2_blk_pre, i>>log2_blksize, j>>log2_blksize);          
    m_cur2_qpe->put_blk_luma (&cur2_blk_qpe, i>>log2_blksize, j>>log2_blksize);          

      store_ipred_img->put_blk_luma(&blk_pred_best, i>>log2_blksize, j>>log2_blksize); 

    if(type)
    {
      param.ztree[EntropyModel::I_luma]->copy_coding_state(rd_ztree_luma_4);
      param.ztree_64->copy_coding_state(rd_ztree_luma_8);
    }
    else
    {
      tmp_ztree->copy_coding_state(rd_ztree_luma_4);
      tmp_ztree_64->copy_coding_state(rd_ztree_luma_8);
    }

#ifdef TRACE_DEBUG
  if(type==1) 
  {
    ENABLE_ETRACE;
  }
    SPRINT_ETRACE("luma intra 16x16");
    RPRINT_ETRACE;
#endif

#ifdef INTRA_MODE_STATS
    for(ii=0; ii<blk_height; ii+=istep)
    {
      for(jj=0; jj<blk_width; jj+=jstep)
      {
        if( (i+ii)>(cur2->height8-1) || (j+jj)>(cur2->width8-1) )
          continue;
        int sip_type_tmp = intra_pred->get_sip_type(i+ii,j+jj);
        mlc->sip_type_count[sip_type_tmp] ++;

        int kk;
        for(kk=0; kk<4; kk++)
        {
          int mode_y_tmp   = intra_pred->get_mode_y(i+ii, j+jj, kk);
          mlc->sip_mode_count[sip_type_tmp][0][mode_y_tmp]++;
        }
      }
    }
#endif //INTRA_MODE_STATS

    // entropy encoding the sip_type, modes, and coeffs
    switch (sip_type16)
    {
    case 0:
      {
        Block coeff_8x8(SIP8x8);
        for(ii=0; ii<blk_height; ii+=istep)
        {
          for(jj=0; jj<blk_width; jj+=jstep)
          {
            if ((i+ii > m_cur2->height8-1) || (j+jj > m_cur2->width8-1))
              continue;
            best_coeff_16x16->extract(coeff_8x8, 8*ii, 8*jj);
            intra_pred->copy_to_L_mode_pred_luma((i+ii)%blk_height, (j+jj)%blk_width);
            entropycode_luma_sub_block_8x8(coeff_8x8, (i+ii)<<1, (j+jj)<<1, eenc, type, tmp_ztree, tmp_ztree_64, intra_pred->get_sip_type(i+ii,j+jj) );
          }
        }
        break;
      }
    case 1:
      {
        intra_pred->copy_to_L_mode_pred_luma(i%blk_height, j%blk_width);
        entropycode_luma_blk16x16(best_coeff_16x16, i, j, eenc, type, tmp_subztrees, intra_pred->get_sip_type(i,j));
        break;
      }
    default:
      {
        printf("No such sip_type in 16x16 level\n");
      }
    }
#ifdef TRACE_DEBUG
  if(type==1) 
    DISABLE_ETRACE;
#endif
    delete rd_ztree_luma_4;
    delete rd_ztree_luma_8;
}

//-----------------------------------------------------------------------------
/*!
* \brief Decide the intra prediction mode to use for a luma 16x16 block
*
* \param[in,out] eenc Pointer to entropy encoder
* \param[out] min_blk Pointer to the prediction sample block
* \param[in] org_blk Pointer to the block to be predicted
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in] lambda for rd decision
* \param[in] ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
int IEncoder6::rd_mode_decision_luma_16x16(EntropyEncoder *eenc, Block *min_blk, Block *org_blk, const int i, const int j, double lambda)
{
  int bsize = 16;
  double    distortion;
  double    rdcost;
  double    min_rdcost = DBL_MAX;
  double    min_distortion = DBL_MAX;
  int       min_mode = -1;
  int       mode_count;
  Block  perr(bsize, bsize);
  Block  pred(bsize, bsize);
  Block  coef(bsize, bsize);
  Block  trans(bsize, bsize); 

  int       TransType = -1;
  int       best_transform = -1;

  double    lambda_rdoq;

  lambda_rdoq = this->calc_lambda_mode(0, 1);  

  quant_luma->loadmatrix(0, -1, 0, 0); // cq_start
  quant_luma->loadmatrix(0, -1, 0, 1); // cq_end
 
  for (mode_count = 0; mode_count < IntraPrediction::NO_INTRA_PMODE_LUMA_MxN; mode_count++) 
  {
    if (intra_pred->use_mode_luma[mode_count] == 0) 
      continue;

    for (TransType = 0; TransType < (param.transform_selection_16x16 ? 2 : 1); TransType++)
    {
      // get the original block
      perr.copy( *org_blk );

      // intra prediction
      pred.copy(intra_pred->mprr_blk[SIP16x16][mode_count]);

      // get the residue
      perr.sub(pred);

      // encode with 8x8 transform/quantization
      // transform
      int transIdx = (TransType < 1) ? Transform::DCT : Transform::AlternateIntra;
      m_trans[transIdx]->forward_transform(&perr);

      trans.copy( perr ); 

      // quantization
      quant_luma->quant16x16(&perr, 0);        // 0: mode decision, no rounding offset update
      coef.copy( perr );

      // inverse quantization
      quant_luma->iquant16x16(&perr);

      // inverse transform
      m_trans[transIdx]->inverse_transform(&perr);

      pred.add(perr);

      distortion = pred.ssd(*org_blk);

      if (distortion > min_rdcost) 
        continue;

      if (param.transform_selection_16x16)
        selected_transform_16x16 = TransType;
      
      rdcost = distortion + lambda * get_rate_luma_16x16(eenc, &coef, i, j, mode_count) * param.sl_weight16x16;
 
      if (min_rdcost > rdcost) 
      {
        if (param.transform_selection_16x16)
          best_transform = TransType;
        min_rdcost = rdcost;
        min_mode = mode_count;
        min_distortion = distortion;

        min_blk->copy( trans );
      }    
    }
  }

  if (param.transform_selection_16x16)
    selected_transform_16x16 = best_transform;

  // quantization
  quant_luma->quant16x16(min_blk, 1); // 1 for encoding, to update the rounding offset

  rdcost16 = min_rdcost;  

#if 0
    for (k = 0; k < bsize*bsize; k++)
    {
      pred.val[k] = org_blk->val[k] - intra_pred->mprr_blk[SIP16x16][min_mode].val[k];
    }
    store_ipred_img->put_blk_luma(&pred, i>>1, j>>1); 
#endif
  
  return (min_mode);
}


//-----------------------------------------------------------------------------
/*!
* \brief get the rate of intra prediction for a luma 16x16 block
*
* \param[in,out] eenc Pointer to entropy encoder
* \param[in] coeff Pointer to the coefficient block
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in] cur_mode mode of block to get the rate for
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
double IEncoder6::get_rate_luma_16x16(EntropyEncoder *eenc, Block *coef, int i, int j, int cur_mode)
{
  int rate = 0;
  int is_zero_blk;
  int ii, jj;
  
  rate += eenc->cost_spatial_intra_prediction_luma(cur_mode, intra_pred->L_mode_pred_luma[0]);
  
  ii = i << 1;
  jj = j << 1;
  is_zero_blk = coef->is_zero();
  zmap->set(ii, jj, coef->nnz(), 4, 4);
  rate += eenc->cost_zblk(is_zero_blk, zmap->getcontext(ii, jj), 0);
  
  if (is_zero_blk == FALSE) 
  {
    if (param.transform_selection_16x16 == 1)
      rate += eenc->cost_tsfm_flag(selected_transform_16x16);
    rate += eenc->cost_coeffs_16x16(rd_subztrees_luma, coef); 
  }
  
  return rate * (1.0 / power2(RATE_BITS)) + param.sl_rate_offset;
}

//-----------------------------------------------------------------------------
// Start of all the SIP4x4, SIP8x8, SIP2x8 and SIP8x2 functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Encode a luma 4x4, 2x8 and 8x2 block, and choose the best NxN partition
*
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in,out] eenc Pointer to entropy encoder
* \param[in,out] mlc Pointer to mlc stats
* \param[in] type for RDOQ
* \param[in] tmp_ztree  Pointer to zero tree to be used
* \param[in] tmp_ztree_64  Pointer to zero tree to be used for 8x8
*/
//-----------------------------------------------------------------------------
void IEncoder6::encode_luma_block_NxN(int i, int j, EntropyEncoder *eenc, MLC_Stat *mlc, int type, ZeroTree *tmp_ztree, ZeroTree *tmp_ztree_64)
{
  assert(m_cur2 != NULL);
  
  double rdcost[NSIP];
  int ii,jj;
  double rdcost_min = DBL_MAX;
  int sip_type = 0;
  int blkIdx;

  Block cur2_blk(8, 8);
  Block cur2_blk_pred(8, 8);
  Block cur2_blk_qpe(8, 8);
  Block coeff_tmp[4];
  Block coeff(8, 8);
  int tmp_lmode_y[4];

  if (type)
  {
    tmp_ztree = param.ztree[EntropyModel::I_luma];
    tmp_ztree_64 = param.ztree_64;
  }

  int types[NSIP];
  int typeCount = 0;
  
  types[typeCount++] = SIP4x4;
  if (param.intra8x8)
    types[typeCount++] = SIP8x8;
  types[typeCount++] = SIP2x8;
  types[typeCount++] = SIP8x2;

  int weight[NSIP]; // cq_start
  weight[SIP4x4] = INTRA4x4WEIGHTEDCOST;
  weight[SIP2x8] = INTRA2x8WEIGHTEDCOST;
  weight[SIP8x2] = INTRA8x2WEIGHTEDCOST;
  weight[SIP8x8] = INTRA8x8WEIGHTEDCOST;  // cq_end
  
  quant_luma->I_save(QUANT_OFFSET_STARTING);      // save the quant offset at the starting of the 4 4x4 block coding

  int typeIdx, bestTypeIdx = -1;
  for (typeIdx = 0; typeIdx < typeCount; typeIdx++)
  {
    int sip_type = types[typeIdx];

    quant_luma->I_load(QUANT_OFFSET_STARTING);      // save the quant offset at the starting of the 4 4x4 block coding
    rd_eenc.copy_coding_state(eenc);
    
    if (sip_type == SIP8x8)
      rd_ztree_luma_8x8->copy_coding_state(tmp_ztree_64);
    else
      rd_ztree_luma->copy_coding_state(tmp_ztree);
    
    intra_pred->set_sip_type(i>>1, j>>1, sip_type);
    
    rdcost[sip_type] = 0.0;
    if (param.bugfix92)
    {
      double lambda_mode = this->calc_lambda_mode(0, 0);  
      rdcost[sip_type] += lambda_mode * rd_eenc.cost_sip_type(sip_type, intra_pred->get_sip_ctx(i>>1, j>>1)) * (1.0 / power2(RATE_BITS));
    }

    rd_eenc.entropycode_sip_type(sip_type, intra_pred->get_sip_ctx(i>>1, j>>1));

    blkIdx = 0;
    for (ii = 0; ii < 8; ii += Block::bht[sip_type])
    {
      for (jj = 0; jj < 8; jj += Block::bwt[sip_type])
      {
        if (param.intra16x16)
        {
          intra_pred->avail_up_right_4x4 = (jj == 0) || (intra_pred->avail_up_right_8x8 && ii == 0);
        }

        coeff_tmp[blkIdx].set_type(sip_type);
        ZeroTree *zt = (sip_type == SIP8x8) ? tmp_ztree_64 : tmp_ztree;
        rdcost[sip_type] += get_rdcost_luma(coeff_tmp[blkIdx], (4*i+ii) >> Block::log_bht[sip_type], (4*j+jj) >> Block::log_bwt[sip_type], &rd_eenc, zt, blkIdx);

        blkIdx++;
      }
    }
    
    if (param.qmatrix_enable_flag[0] && param.qmatrix_enable_flag[1]) // cq_start
      rdcost[sip_type] *= weight[sip_type]/128.0; // cq_end
    
    if (rdcost[sip_type] < rdcost_min)
    {
      bestTypeIdx = typeIdx;
      quant_luma->I_save(QUANT_OFFSET_BEST);

      rdcost_min = rdcost[sip_type];

      m_cur2->get_blk_luma(&cur2_blk, i>>1, j>>1);
      m_cur2_pred->get_blk_luma(&cur2_blk_pred, i>>1, j>>1);
      m_cur2_qpe->get_blk_luma(&cur2_blk_qpe, i>>1, j>>1);

      for (ii=0; ii<4; ii++) 
      {
        blkIdx = (sip_type == SIP8x8) ? 0 : ii;
        tmp_mode_y[ii] = intra_pred->get_mode_y(i>>1, j>>1, blkIdx); 
        tmp_lmode_y[ii] = intra_pred->L_mode_pred_luma[blkIdx];
      }
      
      blkIdx = 0;
      for (ii = 0; ii < 8; ii += Block::bht[sip_type])
      {
        for (jj = 0; jj < 8; jj += Block::bwt[sip_type])
        {
          coeff.insert(coeff_tmp[blkIdx++], ii, jj);
        }
      }
    }  
  }
  
  assert(bestTypeIdx != -1);
  sip_type = types[bestTypeIdx];

  if ( param.intra16x16 )
  {
    rdcost8 = rdcost_min;
    best_coeff_8x8.copy(coeff);
  }

  // Restore everything to the best selected mode
  quant_luma->I_load(QUANT_OFFSET_BEST);
  intra_pred->set_sip_type(i>>1, j>>1, sip_type);  // NxN_type[(i>>1)*cur2->width8+(j>>1)] = sip_type;  // set the sip type
#ifdef TRACE_DEBUG
  if(type==1) 
  {
    ENABLE_ETRACE;
  }
  SPRINT_ETRACE("luma intra");
  RPRINT_ETRACE;
#endif

  m_cur2->put_blk_luma(&cur2_blk, i>>1, j>>1);
  m_cur2_pred->put_blk_luma(&cur2_blk_pred, i>>1, j>>1);
  m_cur2_qpe->put_blk_luma(&cur2_blk_qpe, i>>1, j>>1);

  if (type)
  {
    for (ii = 0; ii < ((sip_type == SIP8x8) ? 1 : 4); ii++)
    {
      if (tmp_mode_y[ii] == IntraPrediction::TMA_L9_PRED_MxN)
        mlc->intra_tma_L5++;
      else if (tmp_mode_y[ii] == IntraPrediction::TMA_L17_PRED_MxN)
        mlc->intra_tma_L9++;
    }
  }
  
  for (ii=0; ii<2; ii++) 
  {
    for (jj=0; jj<2; jj++) 
    { 
      intra_pred->put_mode_y(i>>1, j>>1, 2*ii+jj, tmp_mode_y[2*ii+jj]);      
      intra_pred->L_mode_pred_luma[2*ii+jj] = tmp_lmode_y[2*ii+jj];

        store[(i>>1)*m_cur2->width8+(j>>1)].sip_type = sip_type;
        store[(i>>1)*m_cur2->width8+(j>>1)].mode_luma[ii][jj] = tmp_mode_y[2*ii + jj];
        store[(i>>1)*m_cur2->width8+(j>>1)].lmode_luma[ii][jj] = tmp_lmode_y[2*ii + jj];
    }
  }
    // Compute and store residual for use by ME
    Block blk_pred_best(8, 8);
    m_target->get_blk_luma(&blk_pred_best, i>>1, j>>1); // get original block
    blk_pred_best.sub(cur2_blk_pred);
    store_ipred_img->put_blk_luma(&blk_pred_best, i>>1, j>>1); 

  if (! param.intra16x16)
  {
    sip_type_count[sip_type]++;
    sip_sel_mode_count[sip_type][0][tmp_mode_y[0]]++;    
    sip_sel_mode_count[sip_type][0][tmp_mode_y[1]]++;    
    sip_sel_mode_count[sip_type][0][tmp_mode_y[2]]++;    
    sip_sel_mode_count[sip_type][0][tmp_mode_y[3]]++;    
  }

  entropycode_luma_sub_block_8x8(coeff, i, j, eenc, type, tmp_ztree, tmp_ztree_64, sip_type);
}

//-----------------------------------------------------------------------------
/*!
* \brief Get the rd cost for a luma block
*
* \param[out] coeff Pointer to the coefficient block
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in,out] eenc Pointer to entropy encoder
* \param[in] tmp_ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
double IEncoder6::get_rdcost_luma(Block &coeff, int i, int j, EntropyEncoder *eenc, ZeroTree *tmp_ztree, int blkIdx) 
{
  assert(m_cur2 != NULL);
  int sip_type = coeff.get_type();

  intra_pred->generate_prediction(i, j, sip_type, -1, NULL, m_cur2); // Generate all predictions
  intra_pred->set_L_mode_pred_luma(blkIdx, sip_type); // Set most probable mode

  double lambda_mode = this->calc_lambda_mode(0, 0);  
  double min_rdcost = DBL_MAX;
  Block blk(sip_type);
  Block blk_org(sip_type);
  m_target->get_blk_luma(&blk_org, i, j); // get original block

  int mode = this->rd_mode_decision_luma(eenc, &blk, &blk_org, i, j, lambda_mode, tmp_ztree, sip_type, min_rdcost);
  coeff.copy(blk); // Copy coeffs

  int is_zero_blk = blk.is_zero();
  int izmap, jzmap;
  zmap->convert_indices(i, j, izmap, jzmap, sip_type);
  zmap->set(izmap, jzmap, !is_zero_blk, 1+(sip_type==SIP8x8), 1+(sip_type==SIP8x8));

  if (sip_type == SIP8x8)
    intra_pred->put_mode_y_MxN(i, j, mode, 1, 1);
  else
    intra_pred->put_mode_y(izmap>>1, jzmap>>1, blkIdx, mode);    

  int min_mode = eenc->encode_spatial_intra_prediction_luma(mode, intra_pred->L_mode_pred_luma[blkIdx]);

  // encode data, will be discarded but this updates the probabilities
  eenc->encode_zblk(is_zero_blk, zmap->getcontext(izmap, jzmap), 0);

  if (is_zero_blk == FALSE) 
  {
    if (sip_type == SIP8x8)
      eenc->encode_coeffs_8x8(rd_ztree_luma_8x8, &blk);
    else
      eenc->encode_coeffs(rd_ztree_luma, &blk, EntropyModel::I_luma, sip_type);
  }

  int q_sip_type = (sip_type == SIP8x8) ? 0 : sip_type;

  quant_luma->loadmatrix(EntropyModel::I_luma, -1, q_sip_type, 0); // cq_start
  quant_luma->loadmatrix(EntropyModel::I_luma, -1, q_sip_type, 1); // cq_end

  quant_luma->iquant(&blk, q_sip_type);
  int transIdx = (this->selected_transform == 0) ? Transform::DCT : Transform::AlternateIntra;
  m_trans[transIdx]->inverse(&blk);

  Block &blk_pre = intra_pred->mprr_blk[sip_type][mode];
  m_cur2_pred->put_blk_luma(&blk_pre, i, j);
  m_cur2_qpe->put_blk_luma(&blk, i, j);
  blk.add(blk_pre);
  m_cur2->put_blk_luma(&blk, i, j);

  // Some stats
  sip_mode_count[sip_type][0][mode] += blk.get_num_elements() >> 4;
  if (mode == min_mode)
    sip_min_mode_count[sip_type][0][mode] += blk.get_num_elements() >> 4;       
  
  return min_rdcost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode luma 4x4 block
*
* \param[out] blk Pointer to the coefficient block
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in,out] eenc Pointer to entropy encoder
* \param[in] type ???
* \param[in] tmp_ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
void IEncoder6::entropycode_luma(Block &blk, int i, int j, EntropyEncoder *eenc, int type, ZeroTree *tmp_ztree, int blkIdx) 
{
  int sip_type = blk.get_type();

  int izmap, jzmap;
  zmap->convert_indices(i, j, izmap, jzmap, sip_type);

  int emodel = EntropyModel::I_luma;
  int is_zero_blk;
  int mode, min_mode;

  if (type)
    tmp_ztree = (sip_type == SIP8x8) ? param.ztree_64 : param.ztree[emodel];
  
  is_zero_blk = blk.is_zero();
  zmap->set(izmap, jzmap, blk.nnz(), 1+(sip_type == SIP8x8), 1+(sip_type == SIP8x8));

  eenc->bits_luma = -eenc->get_nbits();
  mode = intra_pred->get_mode_y(blk.get_height()*i, blk.get_width()*j);      // use pixel coordinates

  min_mode = eenc->encode_spatial_intra_prediction_luma(mode, intra_pred->L_mode_pred_luma[blkIdx]);

  eenc->bits_luma += eenc->get_nbits();

  // encode data
  eenc->encode_zblk(is_zero_blk, zmap->getcontext(izmap, jzmap), 0); 

  if (is_zero_blk == FALSE) 
  {
    if (sip_type == SIP8x8)
      eenc->encode_coeffs_8x8(tmp_ztree, &blk, emodel);
    else
      eenc->encode_coeffs(tmp_ztree, &blk, emodel, sip_type);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Get the rd cost for a luma 16x16 block
*
* \param[out] coeff Pointer to the coefficient block
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in,out] eenc Pointer to entropy encoder
* \param[in] tmp_ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
  double IEncoder6::get_rdcost_luma_blk16x16(Block *coeff, int i, int j, EntropyEncoder *eenc)
{
  assert(m_cur2 != NULL);
  int mode;
  int min_mode;
  int is_zero_blk;

  Block blk_16x16(SIP16x16);
  Block blk_org(SIP16x16);
  Block blk_pre(SIP16x16);
  double  lambda_mode;
  double  min_rdcost = DBL_MAX;

  intra_pred->generate_prediction(i >> 1, j >> 1, SIP16x16, -1, NULL, m_cur2);

  intra_pred->set_L_mode_pred_luma(0, SIP16x16);    // set the L_mode_predictor

  m_target->get_blk_luma(&blk_org, i>>1, j>>1);

  lambda_mode = this->calc_lambda_mode(0, 0);  

  mode = this->rd_mode_decision_luma_16x16(eenc, &blk_16x16, &blk_org, i, j, lambda_mode);

  coeff->copy( blk_16x16 );

  min_mode = eenc->encode_spatial_intra_prediction_luma(mode, intra_pred->L_mode_pred_luma[0]);
  sip_mode_count[SIP16x16][0][mode] += 16;

  if (mode == min_mode)
    sip_min_mode_count[SIP16x16][0][mode] += 16;

  int transform_16x16 = -1;

  zmap->set(2*i, 2*j, blk_16x16.nnz(), 4, 4);

  intra_pred->put_mode_y_MxN(i, j, mode, 2, 2);
 
  // encode data
  is_zero_blk = blk_16x16.is_zero();
  eenc->encode_zblk(is_zero_blk, zmap->getcontext(i<<1, j<<1), 0);
  if (is_zero_blk == FALSE)
  {
    if (param.transform_selection_16x16 == 1)
    {
      transform_16x16 = selected_transform_16x16;
      eenc->encode_tsfm_flag(transform_16x16);
    }

    eenc->encode_coeffs_16x16(rd_subztrees_luma, &blk_16x16);
  }

  quant_luma->loadmatrix(0, -1, 0, 0); // cq_start
  quant_luma->loadmatrix(0, -1, 0, 1); // cq_end

  // encode with 16x16 transform/quantization
  quant_luma->iquant16x16(&blk_16x16);

  int transIdx = (transform_16x16 < 1) ? Transform::DCT : Transform::AlternateIntra;
  m_trans[transIdx]->inverse_transform(&blk_16x16);

  blk_pre.copy(intra_pred->mprr_blk[SIP16x16][mode]);

  m_cur2_pred->put_blk_luma(&blk_pre, i>>1, j>>1);
  m_cur2_qpe->put_blk_luma(&blk_16x16, i>>1, j>>1);

  blk_pre.add(blk_16x16);
  m_cur2->put_blk_luma(&blk_pre, i>>1, j>>1);

  min_rdcost = rdcost16;

  return min_rdcost; 
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode luma 16x16 block
*
* \param[out] blk Pointer to the coefficient block
* \param[in] i Vertical position of block to encode
* \param[in] j Horizontal position of block to encode
* \param[in,out] eenc Pointer to entropy encoder
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] tmp_ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
void IEncoder6::entropycode_luma_blk16x16(Block *blk, int i, int j, EntropyEncoder *eenc, int type, ZeroTree **tmp_subztrees, int sip_type)
{
  int is_zero_blk;
  int mode, min_mode;

  if (type)
    tmp_subztrees = &param.ztree[NZTREES-NSUBZTREES];
  
  int ii = i << 1;
  int jj = j << 1; 

  if (sip_type == SIP16x16)
    eenc->entropycode_sip_type(SIP16x16, intra_pred->get_sip_ctx(i, j));

  is_zero_blk = blk->is_zero();
  zmap->set(ii, jj, blk->nnz(), 4, 4);

  eenc->bits_luma = -eenc->get_nbits();
  mode = intra_pred->get_mode_y(ii<<2, jj<<2);      // use pixel coordinates

  min_mode = eenc->encode_spatial_intra_prediction_luma(mode, intra_pred->L_mode_pred_luma[0]);

  eenc->bits_luma += eenc->get_nbits();

  // encode data
  eenc->encode_zblk(is_zero_blk, zmap->getcontext(ii,jj), 0);

  if (is_zero_blk == FALSE) 
  { 
    if (param.transform_selection_16x16 == 1)
      eenc->encode_tsfm_flag(selected_transform_16x16);

    eenc->encode_coeffs_16x16(tmp_subztrees, blk);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Encode luma 8x8 sub-block
*
* \param[out] blk Pointer to the coefficient block
* \param[in] i Vertical position of block to encode (4x4 unit)
* \param[in] j Horizontal position of block to encode (4x4 unit)
* \param[in,out] eenc Pointer to entropy encoder
* \param[in] type ???
* \param[in] tmp_ztree  Pointer to zero tree to be used
*/
//-----------------------------------------------------------------------------
void IEncoder6::entropycode_luma_sub_block_8x8(Block &coeff, int i, int j, EntropyEncoder *eenc, int type, ZeroTree *tmp_ztree, ZeroTree *tmp_ztree_64, int sip_type)
{
  int ii, jj;

  Block coeff_tmp(sip_type);

  eenc->entropycode_sip_type(sip_type, intra_pred->get_sip_ctx(i>>1, j>>1));

  ZeroTree *zt = (sip_type == SIP8x8) ? tmp_ztree_64 : tmp_ztree;
  if (type)
    zt = NULL;

  int blkIdx = 0;
  for (ii = 0; ii < 8; ii += Block::bht[sip_type])
  {
    for (jj = 0; jj < 8; jj += Block::bwt[sip_type])
    {
      coeff.extract(coeff_tmp, ii, jj);
      entropycode_luma(coeff_tmp, (4*i+ii) >> Block::log_bht[sip_type], (4*j+jj) >> Block::log_bwt[sip_type], eenc, type, zt, blkIdx);
      blkIdx++;
    }
  }

#ifdef TRACE_DEBUG
  if(type==1) 
    DISABLE_ETRACE;
#endif
  
#ifdef SIPNxN_FRAME_STAT
  switch (sip_type)
  {
    case SIP2x8:
      for (int kk=0;kk<64;kk++) 
      {
        ss_2x8[tmp_mode_y[ii>>4]][kk] += (double)coeff.val[kk] * coeff.val[kk];
        ss_2x8_count[tmp_mode_y[ii>>4]]++;
      }
      break;

    case SIP8x2:
      for (int kk=0;kk<64;kk++) 
      {
        ss_8x2[tmp_mode_y[ii>>4]][kk] += (double)coeff.val[kk] * coeff.val[kk];
        ss_8x2_count[tmp_mode_y[ii>>4]]++;
      }
      break;
      
    case SIP8x8:
      for (int kk=0;kk<64;kk++) 
      {
        ss_8x8[tmp_mode_y[0]][kk] += (double)coeff.val[kk] * coeff.val[kk];
        ss_8x8_count[tmp_mode_y[0]]++;
      }
      break;

    default:
      assert(sip_type == SIP4x4);
      for (int kk=0;kk<64;kk++) 
      {
        ss_4x4[tmp_mode_y[ii>>4]][kk] += (double)coeff.val[kk] * coeff.val[kk];
        ss_4x4_count[tmp_mode_y[ii>>4]]++;
      }
      break;
  }
#endif  
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
