/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file IntraPrediction.cpp
* \brief Implementation of IntraPrediction functions that are used by encoder and decoder
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "IntraPrediction.H"
#include "Param.H"

//-----------------------------------------------------------------------------
/*!
* \brief IntraPrediction constructor
*/
//-----------------------------------------------------------------------------
#ifdef FRAME_SUBSTREAMS
IntraPrediction::IntraPrediction(int h, int w, int *NxN_type_given, int *mode_y_given, int *mode_u_given)
#else
IntraPrediction::IntraPrediction(int h, int w)
#endif
{
  m_width4 = w >> 2;
  m_height4 = h >> 2;
  m_width8 = w >> 3;
  m_height8 = h >> 3;
  
  candidate_search_region = param.candidate_search_region;
  if (param.intra16x16)
    candidate_search_region_intra16x16 = param.candidate_search_region_intra16x16;
  num_candidate_averaged  = param.num_candidate_averaged;
  candidate_sad_threshold = param.candidate_sad_threshold;

  int sip_type, i;
  for (sip_type = 0; sip_type < 6; sip_type++)
  {
    for (i = 0; i < NO_INTRA_PMODE_LUMA_MxN; i++)
    {
      mprr_blk[sip_type][i].set_type(sip_type);
    }
  }
#ifdef FRAME_SUBSTREAMS
  if (NxN_type_given != NULL)
  {
      NxN_type = NxN_type_given;
      mode_y = mode_y_given;
      mode_u = mode_u_given;
      fields_allocated = 0;
  }
  else
  {
      NxN_type = new int[(h*w)>>6];
      mode_y = new int[(h * w) >> 4];
      mode_u = new int[(h * w) >> 6];
      fields_allocated = 1;
  }
#else
  NxN_type = new int[(h*w)>>6];
  mode_y = new int[(h * w) >> 4];
  mode_u = new int[(h * w) >> 6];
#endif

  init_tma_param_table();
  int max_search_range = imax(param.candidate_search_region_intra16x16, param.candidate_search_region);
  int search_area_width = max_search_range + 16 + 8;
  int search_area_height = max_search_range + 16;
  search_area = new int[search_area_width * search_area_height];
}

//-----------------------------------------------------------------------------
/*!
* \brief IntraPrediction destructor
*/
//-----------------------------------------------------------------------------
IntraPrediction::~IntraPrediction()
{
#ifdef FRAME_SUBSTREAMS
  if (fields_allocated)
  {
    delete[] NxN_type;
    delete[] mode_y;
    delete[] mode_u;
  }
#else
  delete[] NxN_type;
  delete[] mode_y;
  delete[] mode_u;
#endif
  
  delete[] search_area;
}

//-----------------------------------------------------------------------------
/*!
* \brief Reset array storing mode information
*/
//-----------------------------------------------------------------------------
void IntraPrediction::reset_mode()
{
  int i;

  for (i = 0; i < m_width4 * m_height4; i++)
  {
    mode_y[i] = INT_INF;
  }
  
  for (i = 0; i < m_width8 * m_height8; i++)
  {
    mode_u[i] = INT_INF;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy NxN_type to tmp_sip_type
*
* \param[in] fi       Vertical 8x8 index in frame
* \param[in] fj       Horizontal 8x8 index in frame
* \param[in] ti       Vertical 8x8 index in a 16x16 block
* \param[in] tj       Horizontal 8x8 index in a 16x16 block
* \param[in] h8       Height in 8x8 unit
* \param[in] w8       Width in 8x8 unit
*/
//-----------------------------------------------------------------------------
void IntraPrediction::copy_sip_type_from_NxN_type(int fi, int fj, int ti, int tj, int h8, int w8)
{
  int i, j;

  for(i=0; i<min(h8, m_height8-fi); i++)
  {
    for(j=0; j<min(w8, m_width8-fj); j++)
    {
      tmp_sip_type[ti+i][tj+j] = NxN_type[(fi+i)*m_width8 + fj+j];
    }
  }
}
//-----------------------------------------------------------------------------
/*!
* \brief Copy tmp_sip_type to NxN_type
*
* \param[in] ti       Vertical 8x8 index in frame
* \param[in] tj       Horizontal 8x8 index in frame
* \param[in] fi       Vertical 8x8 index in a 16x16 block
* \param[in] fj       Horizontal 8x8 index in a 16x16 block
* \param[in] h8       Height in 8x8 unit
* \param[in] w8       Width in 8x8 unit
*/
//-----------------------------------------------------------------------------
void IntraPrediction::copy_sip_type_to_NxN_type( int ti, int tj, int fi, int fj, int h8, int w8)
{
  int i, j;
  
  for(i=0; i<min(h8, m_height8-ti); i++)
  {
    for(j=0; j<min(w8, m_width8-tj); j++)
    {
      NxN_type[(ti+i)*m_width8 + tj+j] = tmp_sip_type[fi+i][fj+j];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Sets the sip_mode for the 8x8 block, index is in units of 8x8 blocks.
*
* \param[in] i Vertical 8x8 index
* \param[in] j Horizontal 8x8 index
* \param[in] sip_type Sip_type to set
*/
//-----------------------------------------------------------------------------
void IntraPrediction::set_sip_type(int i, int j, int sip_type, int h8, int w8)
{
  int ii, jj;

  h8 = min(h8, m_height8 - i); 
  w8 = min(w8, m_width8 - j);

  for(ii=0; ii<h8; ii++)
  {
    for(jj=0; jj<w8; jj++)
    {
      NxN_type[(i+ii)*m_width8 + j+jj] = sip_type;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Gets the sip_mode for the 8x8 block, index is in units of 8x8 blocks.
*
* \param[in] i Vertical 8x8 index
* \param[in] j Horizontal 8x8 index
* \return sip_type
*/
//-----------------------------------------------------------------------------
int IntraPrediction::get_sip_type(int i, int j)
{
  return NxN_type[i*m_width8+j]; // set the type
}

//-----------------------------------------------------------------------------
/*!
* \brief Reset array storing sip type information
*/
//-----------------------------------------------------------------------------
void IntraPrediction::reset_sip_type()
{
  int i;

  for (i = 0; i < m_width8 * m_height8; i++)
  {
    NxN_type[i] = INT_INF;
  }
}
//-----------------------------------------------------------------------------
/*!
* \brief Gets the sip_type context for the 8x8 block, index is in units of 8x8 blocks.
*
* \param[in] i Vertical 8x8 index
* \param[in] j Horizontal 8x8 index
* \return ctx
*/
//-----------------------------------------------------------------------------
int IntraPrediction::get_sip_ctx(int i, int j)
{
  int sip_type_left, sip_type_above;

  if (i==0 /*|| get_sip_type(i-1, j) == SIP_INTER */) // first row or block above is inter
  {
    if (j==0 /*|| get_sip_type(i, j-1) == SIP_INTER */)
      return 1;  // mix context. // first col or block left is inter
    else
    { // check the left only
      sip_type_left = get_sip_type(i, j-1);
      if (sip_type_left == SIP4x4)
        return 0;
      else if (sip_type_left == SIP8x8)
        return 2;
      else
        return 1;
    }
  }
  else
  {
    sip_type_above = get_sip_type(i-1, j);
    if (j==0 /*|| get_sip_type(i, j-1) == SIP_INTER */)
    { // check the above only
      if (sip_type_above == SIP4x4)
        return 0;
      else if (sip_type_above == SIP8x8)
        return 2;
      else
        return 1;
    }
    else
    {
      sip_type_left = get_sip_type(i, j-1);
      if (sip_type_above == SIP4x4 && sip_type_left == SIP4x4)
        return 0;
      else if (sip_type_above == SIP8x8 && sip_type_left == SIP8x8 )
        return 2;
      else if (sip_type_above == SIP16x16 && sip_type_left == SIP16x16 )
        return 2;
      else
        return 1;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Setting the luma mode, it is indexed by the 8x8 blocks followed by 4 subblocks.
*
* \param[in] i Vertical 8x8 index
* \param[in] j Horizontal 8x8 index
* \param[in] subblock Value of 0 to 3 depending onthe subblock
* \param[in] mode Intra Prediction mode
*/
//-----------------------------------------------------------------------------
void IntraPrediction::put_mode_y(const int i, const int j, int subblock, int mode)
{
  mode_y[4*(i*m_width8+j) + subblock] = mode ; // set the mode
}

void IntraPrediction::put_mode_y_MxN(int i, int j, int mode, int h, int w)
{
  int k, u, v;
  
  for (u = 0; u < h; u++)
  {
    for (v = 0; v < w; v++)
    {
      for (k = 0; k < 4; k++)
      {
        mode_y[4*((i+u)*m_width8+(j+v)) + k] = mode ; // set the mode
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief One way of getting the mode_y. It is indexed by the 8x8 blocks followed by 4 subblocks.
* \brief useful if the NxN_type is already known.
*
* \param[in] i Vertical 8x8 index
* \param[in] j Horizontal 8x8 index
* \param[in] subblock Value of 0 to 3 depending onthe subblock
*
* \return Luma Intra Prediction Mode
*/
//-----------------------------------------------------------------------------
int IntraPrediction::get_mode_y(const int i, const int j, int subblock)
{
  return (mode_y[4*(i*m_width8+j) + subblock]); // return the mode
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy mode_y to pDes
*
* \param[in,out] pDes     The destination of copy
* \param[in] i8       Vertical 8x8 index in frame
* \param[in] j8       Horizontal 8x8 index in frame
* \param[in] h8       Height in 8x8 unit
* \param[in] w8       Width in 8x8 unit
*/
//-----------------------------------------------------------------------------
void IntraPrediction::copy_from_mode_y(int* pDes, int i8, int j8, int h8, int w8)
{
  int i, j, ii, jj;
  int *pD, *pS;
  
  int *pSrc = mode_y;

  for(i=0; i<min(h8, m_height8-i8); i++)
  {
    ii = i8 + i;
    for(j=0; j<min(w8, m_width8-j8); j++)
    {
      jj = j8 + j;
      pD = pDes + ((ii%2)*2 + jj%2)*4;
      pS = pSrc + (ii*m_width8 + jj)*4;
      
      memcpy(pD, pS, 4*sizeof(int));
    }
  }
}
//-----------------------------------------------------------------------------
/*!
* \brief Copy pDes to mode_y
*
* \param[in,out] pDes     The source of copy
* \param[in] i8       Vertical 8x8 index in frame
* \param[in] j8       Horizontal 8x8 index in frame
* \param[in] h8       Height in 8x8 unit
* \param[in] w8       Width in 8x8 unit
*/
//-----------------------------------------------------------------------------
void IntraPrediction::copy_to_mode_y(int* pSrc, int i8, int j8, int h8, int w8)
{
  int i, j, ii, jj;
  int *pD, *pS;
  
  int *pDes = mode_y;

  for(i=0; i<min(h8, m_height8-i8); i++)
  {
    ii = i8 + i;
    for(j=0; j<min(w8, m_width8-j8); j++)
    {
      jj = j8 + j;
      pD = pDes + (ii*m_width8 + jj)*4;
      pS = pSrc + ((ii%2)*2 + jj%2)*4;
      
      memcpy(pD, pS, 4*sizeof(int));
    }
  }
}
//-----------------------------------------------------------------------------
/*!
* \brief Copy L_mode_pred_luma to Large_L_mode_pred_luma
*
* \param[in] i8       Vertical 8x8 index in 16x16 block
* \param[in] j8       Horizontal 8x8 index in 16x16 block
* \param[in] h8       Height in 8x8 unit
* \param[in] w8       Width in 8x8 unit
*/
//-----------------------------------------------------------------------------
void IntraPrediction::copy_from_L_mode_pred_luma(int i8, int j8, int h8, int w8)
{
  int i, j, ii, jj;
  
  for( i=0; i<h8; i++)
  {
    ii = i8 + i;
    for( j=0; j<w8; j++)
    {
      jj = j8 + j;
      memcpy( Large_L_mode_pred_luma[(ii*2+jj)], &L_mode_pred_luma[0], 4*sizeof(int));
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy Large_L_mode_pred_luma to L_mode_pred_luma
*
* \param[in] i8       Vertical 8x8 index in 16x16 block
* \param[in] j8       Horizontal 8x8 index in 16x16 block
*/
//-----------------------------------------------------------------------------
void IntraPrediction::copy_to_L_mode_pred_luma(int i8, int j8)
{
  memcpy(&L_mode_pred_luma[0], Large_L_mode_pred_luma[(i8*2+j8)], 4*sizeof(int));
}

//-----------------------------------------------------------------------------
/*!
* \brief Another way of getting the mode_y. It is indexed by pixels.
* \brief Useful if the NxN_type is not known. Example getting the mode of the block above or to the left.
*
* \param[in] i_p Vertical pixel index
* \param[in] j_p Horizontal pixel index
*
* \return Luma Intra Prediction Mode
*/
//-----------------------------------------------------------------------------
int IntraPrediction::get_mode_y(const int i_p, const int j_p)
{

  // First get the NxN_type
  int subblock;
  int index = (i_p>>3) * m_width8 + (j_p>>3);

  switch (NxN_type[index])
  {
  case SIP4x4:
    {
      subblock = 2 * ((i_p%8)>>2) + ((j_p%8)>>2);
      break;
    }
  case SIP2x8:
    {
      subblock = (i_p%8)>>1;
      break;
    }
  case SIP8x2:
    {
      subblock = (j_p%8)>>1;
      break;
    }
  case SIP8x8:
    {
      subblock = 2 * ((i_p%8)>>2) + ((j_p%8)>>2);
      break;
    }

  case SIP16x16:
    {
      subblock = 2 * ((i_p%8)>>2) + ((j_p%8)>>2);
      break;
    }
  case SIP_INTER:
    {
      return (INT_INF);      // return an infinite value to see the effect
      break;
    }
  default:
    {
      printf ("Unknown SIP type, %d, (%d,%d) => %d\n", NxN_type[index], i_p, j_p, index);
      exit(1);
    }
  }

  return (mode_y[4*index + subblock]); // return the mode
}

//-----------------------------------------------------------------------------
/*!
* \brief Setting the chroma u mode, it is indexed in units of 4x4 chroma blocks (8x8 luma blocks)
*
* \param[in] i Vertical 8x8 index
* \param[in] j Horizontal 8x8 index
* \param[in] mode Intra Prediction mode
*/
//-----------------------------------------------------------------------------
void IntraPrediction::put_mode_u(const int i, const int j, int mode)
{
  mode_u[i*m_width8 + j] = mode;
}

//-----------------------------------------------------------------------------
/*!
* \brief Getting the chroma u mode, it is indexed in units of 4x4 chroma blocks (8x8 luma blocks)
*
* \param[in] i Vertical 8x8 index
* \param[in] j Horizontal 8x8 index
* \param[in] mode Intra Prediction mode
*
* \return Chroma u Intra Prediction Mode
*/
//-----------------------------------------------------------------------------
int IntraPrediction::get_mode_u(const int i, const int j)
{
  return mode_u[i*m_width8 + j];
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute edge vectors in the neighborhood
*
* \param[in] i              Vertical pixel index
* \param[in] j              Horizontal pixel index
* \param[in] sip_type       SIP type of the block
*/
//-----------------------------------------------------------------------------
int IntraPrediction::intrapred_luma_edge(const int i, const int j, int sip_type, Block *blk)
{
  int grad_x, grad_y;
  int dir_x, dir_y;
  int dir_max_x = 0, dir_max_y = 0;
  int norm_max = 0;
  int ref_array[20][32]; //To store the reference pixels
  int a, b;
  int cross_block = 0;
  int edge_pred_computed;
  int a_start, a_end, b_start, b_end;
  
  //To determine if the neighbours are available
  int available_up = (i >= 4);
  int available_left = (j >= 4);
  int available_up_left = available_up && available_left;
  int available_up_right = available_up;
  int available_down_left = available_left;
  
  int sb_width = 8*param.sb_width;
  int sb_edge = (j & -sb_width) + sb_width;
  
  // Note: i and j are sample positions
  switch (sip_type)
  {
    case SIP4x4:
      available_up_right &= (j+4 != imin(sb_edge, cur2->width)); // outside frame or superblock no good
      available_up_right &= (i & j & 4) != 4; // lower-right within 8x8 no good
      if (param.intra16x16)
        available_up_right &= this->avail_up_right_4x4;
      available_down_left &= ((i | j) & 4) == 0; // top-left within 8x8 ok
      break;
      
    case SIP2x8:
      available_up_right = 0; // for simplicity
      if (available_down_left)
        available_down_left = (6 - (i & 6)) >> 1;
      break;
      
    case SIP8x2:
      if (available_up_right)
      {
        if (param.intra16x16)
        {
          if (this->avail_up_right_8x8)
          {
            available_up_right = imin(sb_edge, cur2->width) - (j+2);
            if (param.bugfix90)
              available_up_right = imin(available_up_right, 6);
            else
              available_up_right = (available_up_right >= 6) ? 6 : 0;
          }
          else
            available_up_right = 6 - (j & 6);
        }
        else
        {
          available_up_right = imin(sb_edge, cur2->width) - (j+2);
          available_up_right = imin(available_up_right, 6);
        }
        available_up_right = available_up_right >> 1;
      }
      available_down_left = 0;
      break;
      
    case SIP8x8:
      available_up_right &= (j+8 != imin(sb_edge, cur2->width));
      if (param.intra16x16)
        available_up_right &= this->avail_up_right_8x8;
      available_down_left = 0;
      break;
      
    default:
      assert(sip_type == SIP16x16);
      available_up_right = 0;
      available_down_left = 0;
      break;
  }
  
  available_up *= Block::bwt[sip_type];
  available_up_right *= Block::bwt[sip_type];
  available_left *= Block::bht[sip_type];
  available_down_left *= Block::bht[sip_type];
  
  if (available_up_left)
  {
    a_start = 1;
    b_start = 1;
  }
  else
  {
    b_start = (available_up) ? 5 : 0;
    a_start = (available_left) ? 5 : 0;
  }
  
  b_end = (available_up) ? imin(available_up + available_up_right, 16) + 3 : 0;
  a_end = (available_left) ? imin(available_left + available_down_left, 16) + 3 : 0;
  
  short *ptr = cur2->get_luma_ptr(i-4, j-4);
  
  if (available_up_left)
  {
    for (a = 0; a < 4; a++)
    {
      for (b = 0; b < 4; b++)
      {
        // Note: 127 clipping probably not needed (just in case sample value is 32640 or above, which it shouldn't be)
        ref_array[a][b] = imin(shift_right_round(ptr[a * cur2->width + b], SAMPLE_ADD_PREC+2), 127); 
      }
    }
  }
  
  if (available_up)
  {
    for (a = 0; a < 4; a++)
    {
      for (b = 4; b <= b_end; b++)
      {
        ref_array[a][b] = imin(shift_right_round(ptr[a * cur2->width + b], SAMPLE_ADD_PREC+2), 127);
      }
    }
  }
  
  if (available_left)
  {
    for (a = 4; a <= a_end; a++)
    {
      for (b = 0; b < 4; b++)
      {
        ref_array[a][b] = imin(shift_right_round(ptr[a * cur2->width + b], SAMPLE_ADD_PREC+2), 127);
      }
    }
  }
  
  int limit_up_low = imax(b_start-5, -1);
  int limit_up_high = b_end-4;
  int limit_left_low = imax(a_start-5, -1);
  int limit_left_high = a_end-4;
  
  //////////////////////////////////////////////
  //Compute edge vectors and find the maximum
  //////////////////////////////////////////////
  
  if (available_up && !available_left)
  {
    a_start = 1;
    a_end = 3;
  }
  
  for (a = a_start; a < a_end; a++)
  {
    int b_start_tmp = (a < 3) ? b_start : 1;
    int b_end_tmp = (a < 3) ? b_end : 3;
    
    for (b = b_start_tmp; b < b_end_tmp; b++)
    {
      // Compute gradiants using Sobel filter
      int diag1 = ref_array[a+1][b+1] - ref_array[a-1][b-1];
      int diag2 = ref_array[a-1][b+1] - ref_array[a+1][b-1];
      grad_x = diag1 + diag2 + 2 *(ref_array[a][b+1] - ref_array[a][b-1]);
      grad_y = diag1 - diag2 + 2 *(ref_array[a+1][b] - ref_array[a-1][b]);
      dir_x = (-grad_y);
      dir_y = grad_x;
      int norm = dir_x * dir_x + dir_y * dir_y;
      
      if (norm > norm_max)
      {
        cross_block = check_direction_edge_vector(b-4, a-4, &dir_x, &dir_y, sip_type);
        if(cross_block)
        {
          dir_max_x = dir_x;
          dir_max_y = dir_y;
          norm_max = norm;
        }
      }
    }
  }
  
  //If the direction is vertical or horizontal, the mode is not used
  //because the same prediction can be obtained with VERT_PRED or HOR_PRED
  edge_pred_computed = (norm_max >= param.threshold_edge_vector && dir_max_x != 0 && dir_max_y != 0);

  int edge_dir_x, edge_dir_y;

  if (edge_pred_computed)
  {
    int sign = dir_max_x >> 31;
    edge_dir_x = (dir_max_x ^ sign) - sign; // take absolute value
    edge_dir_y = (dir_max_y ^ sign) - sign; // swap sign if dir_max_x was negative
  }
  else
  {
    edge_dir_x = 0;
    edge_dir_y = 0;
  }
  
  if (edge_pred_computed)
  {
    int k, l;
    int size_x = Block::bwt[sip_type];
    int size_y = Block::bht[sip_type];

    // Need to fetch again, only kept 6-bit value last time...
    short *ptr = cur2->get_luma_ptr(i-1, j-1);

    if (available_up_left)
    {
      ref_array[0][0] = ptr[0];
    }
    
    if (available_up)
    {
      for (b = limit_up_low; b <= limit_up_high; b++)
      {
        ref_array[0][1+b] = ptr[1+b];
      }
    }
    
    if (available_left)
    {
      for (a = limit_left_low; a <= limit_left_high; a++)
      {
        ref_array[1+a][0] = ptr[(1+a) * cur2->width];
      }
    }
    
    for(l=0; l<size_y; l++)
    {
      for(k=0; k<size_x; k++)
      {
        int sum_weights = 0;
        int pred_pix = 0;
        int mask = power2(EDGE_ADD_PREC) - 1;
        
        if (available_up && edge_dir_y != 0)
        {
          int delta_x = ( (l+1) * (edge_dir_x << EDGE_ADD_PREC) + (edge_dir_y >> 1) ) / edge_dir_y;
          int weigh_left = delta_x & mask;
          int weigh_right = power2(EDGE_ADD_PREC) - weigh_left;
          int floor_dx = delta_x >> EDGE_ADD_PREC;
          int ceil_dx = floor_dx + (weigh_left != 0);
          
          //Check the position of the pixel
          if( (k-floor_dx <= limit_up_high) && (k-ceil_dx >= limit_up_low) )
          {
            pred_pix += shift_right_round(( weigh_left * ref_array[0][1+k-ceil_dx] + weigh_right * ref_array[0][1+k-floor_dx] ), EDGE_ADD_PREC);
            sum_weights += 1;
          }
        }           // end if avail_up
        if (available_left && edge_dir_x != 0)
        {
          int delta_y = ( (k+1) * (edge_dir_y << EDGE_ADD_PREC) + (edge_dir_x >> 1) ) / edge_dir_x;
          int weigh_up = delta_y & mask;
          int weigh_down = power2(EDGE_ADD_PREC) - weigh_up;
          int floor_dy = delta_y >> EDGE_ADD_PREC;
          int ceil_dy = floor_dy + (weigh_up != 0);
          
          //Check the position of the pixel
          if( (l-ceil_dy >= limit_left_low) && (l-floor_dy <= limit_left_high) )
          {
            pred_pix += shift_right_round(( weigh_up * ref_array[1+l-ceil_dy][0] + weigh_down * ref_array[1+l-floor_dy][0] ), EDGE_ADD_PREC);
            sum_weights += 1;
          }
        }
        
        //Store prediction
        if (sum_weights == 2)
          pred_pix = pred_pix / 2;
        else if (sum_weights == 0)
        {
          if (k==0 && l==0 && !available_up && available_left)
            pred_pix = ref_array[1][0];
          else if (k==0 && l==0 && available_up && !available_left)
            pred_pix = ref_array[0][1];
          else if (k==0 && l==0 && available_up && available_left)
            pred_pix = ref_array[0][0];
          else if (k==0)
            pred_pix = blk->val[(l-1)*size_x+k];
          else if (l==0)
            pred_pix = blk->val[l*size_x+(k-1)];
          else
            pred_pix = (blk->val[(l-1)*size_x+(k-1)] + blk->val[(l-1)*size_x+k] + blk->val[l*size_x+(k-1)]) / 3;
        }
        
        blk->val[l*size_x+k] = pred_pix;
      }
    }
  }
  
  return edge_pred_computed;
}

//-----------------------------------------------------------------------------
/*!
* \brief Check if the computed edge vector crosses the block
*
* \param[in] ref_x       Horizontal position of the reference
* \param[in] ref_y       Vertical position of the reference
* \param[in] dir_x       Horizontal coordinate of the vector
* \param[in] dir_y       Vertical coordinate of the vector
* \param[in] sip_type    SIP type of the block
*/
//-----------------------------------------------------------------------------
int IntraPrediction::check_direction_edge_vector(const int ref_x, const int ref_y, int *dir_x, int *dir_y, int sip_type)
{
  int dist_block, first_dist_block = 0;
  int delta_x, delta_y;
  int cross_block = 0;
  int tmp_dir_x = *dir_x;
  int tmp_dir_y = *dir_y;
  
  int size_x = Block::bwt[sip_type];
  int size_y = Block::bht[sip_type];
  
  //Check the position of the reference pixel
  //to determine if the vector is from the blocks up or left
  //For vectors up
  if (ref_y == -3 || ref_y == -2)
  {
    if (tmp_dir_y == 0)
      return 0;
    
    if (tmp_dir_y < 0)
    {
      tmp_dir_y = -tmp_dir_y;
      tmp_dir_x = -tmp_dir_x;
    }
    
    first_dist_block = -ref_y;
    
    for (dist_block = first_dist_block; (cross_block == 0) && (dist_block < first_dist_block + size_y); dist_block++)
    {
      delta_x = (dist_block * (tmp_dir_x << EDGE_ADD_PREC) + (tmp_dir_y >> 1)) / tmp_dir_y;
      delta_x = shift_right_round(delta_x, EDGE_ADD_PREC);
      cross_block = ((ref_x + delta_x) >= 0 && (ref_x + delta_x) < size_x);
    }
  }
  else    //For vectors left
  {
    assert(ref_x == -3 || ref_x == -2);
    if (tmp_dir_x == 0)
      return 0;
    
    if (tmp_dir_x < 0)
    {
      tmp_dir_x = -tmp_dir_x;
      tmp_dir_y = -tmp_dir_y;
    }
    
    first_dist_block = -ref_x;
    
    for (dist_block = first_dist_block; (cross_block == 0) && (dist_block < first_dist_block + size_x); dist_block++)
    {
      delta_y = (dist_block * (tmp_dir_y << EDGE_ADD_PREC) + (tmp_dir_x >> 1) ) / tmp_dir_x;
      delta_y = shift_right_round(delta_y, EDGE_ADD_PREC);
      cross_block = ((ref_y + delta_y) >= 0 && (ref_y + delta_y) < size_y);
    }    
  }
  
  return cross_block;
}

/*!
*************************************************************************************
* \brief
*    Prefiltering for Chroma prediction (8x8 or 16x16)
*
* Creates the intra prediction samples.
*
* \param[in] PredPel predicted pels
* \param[in] block_up, block_left 
* \param[in] bsize
 *************************************************************************************
 */
#ifdef CHROMA_INTRA_ADAPTIVE_LPF
  void IntraPrediction::LowPassForChromaPred(short *PredPel, int block_up, int block_left, int bsize, const int iqpC)
#else
  void IntraPrediction::LowPassForChromaPred(short *PredPel, int block_up, int block_left, int bsize)
#endif // CHROMA_INTRA_ADAPTIVE_LPF
{
  int i;
  assert(bsize * 2 <= 16);
  short LoopArray[16];
  
  memcpy(LoopArray,PredPel, bsize * 2 * sizeof(short));

#ifdef CHROMA_INTRA_ADAPTIVE_LPF
  int T = 10 * ( iqpC - 32 );

  if(block_up)
  {
    if( abs(PredPel[0]-PredPel[1]) < T )
    {
      LoopArray[0] = ((PredPel[0] + (PredPel[0]<<1) + PredPel[1] + 2)>>2);
    }

    for(i = 1; i <bsize-1; i++)
    {
      if( abs(PredPel[i]-PredPel[i+1]) < T && abs(PredPel[i]-PredPel[i-1]) < T )
      {
        LoopArray[i] = ((PredPel[i-1] + (PredPel[i]<<1) + PredPel[i+1] + 2)>>2);
      }
    }

    if( abs(PredPel[bsize-1]-PredPel[bsize-2]) < T )
    {
      LoopArray[bsize-1] = ((PredPel[bsize-1] + (PredPel[bsize-1]<<1) + PredPel[bsize-2] + 2)>>2);
    }

  }
  
  if(block_left)
  {
    if( abs(PredPel[bsize]-PredPel[bsize+1]) < T )
    {
      LoopArray[bsize] = ((PredPel[bsize]+ (PredPel[bsize]<<1) + PredPel[bsize+1] + 2)>>2);
    }

    for(i = bsize+1; i <2*bsize-2; i++)
    {
      if( abs(PredPel[i]-PredPel[i+1]) < T && abs(PredPel[i]-PredPel[i-1]) < T )
      {
        LoopArray[i] = ((PredPel[i-1] + (PredPel[i]<<1) + PredPel[i+1] + 2)>>2);
      }
    }
    if( abs(PredPel[2*bsize-1]-PredPel[2*bsize-2]) < T )
    {
      LoopArray[2*bsize-1] = ((PredPel[2*bsize-2] + (PredPel[2*bsize-1]<<1) + PredPel[2*bsize-1] + 2) >> 2);
    }

  }
#else // CHROMA_INTRA_ADAPTIVE_LPF
  if(block_up)
  {

    LoopArray[0] = ((PredPel[0] + (PredPel[0]<<1) + PredPel[1] + 2)>>2);

    for(i = 1; i <bsize-1; i++)
    {
      LoopArray[i] = ((PredPel[i-1] + (PredPel[i]<<1) + PredPel[i+1] + 2)>>2);
    }
    LoopArray[bsize-1] = ((PredPel[bsize-1] + (PredPel[bsize-1]<<1) + PredPel[bsize-2] + 2)>>2);
  }
  
  if(block_left)
  {
    LoopArray[bsize] = ((PredPel[bsize]+ (PredPel[bsize]<<1) + PredPel[bsize+1] + 2)>>2);

    for(i = bsize+1; i <2*bsize-2; i++)
    {
      LoopArray[i] = ((PredPel[i-1] + (PredPel[i]<<1) + PredPel[i+1] + 2)>>2);
    }
    LoopArray[2*bsize-1] = ((PredPel[2*bsize-2] + (PredPel[2*bsize-1]<<1) + PredPel[2*bsize-1] + 2) >> 2);
  }
#endif // CHROMA_INTRA_ADAPTIVE_LPF

  memcpy(PredPel, LoopArray, 2*bsize * sizeof(short));

}

//-----------------------------------------------------------------------------
/*!
* \brief Chooses the L-mode
*
* \param[in] count     index of the blk within the 8x8
* \param[in] sip_type  SIP type of the block
*/
//-----------------------------------------------------------------------------
void IntraPrediction::set_L_mode_pred_luma(int count, int sip_type)
{
  if (sip_type == SIP8x8 || sip_type == SIP16x16)
  {
    assert(count == 0);
    int i;
    for (i = 0; i < 4; i++)
      L_mode_pred_luma[i] = param.mpm_default;
  }
  else
  {
    L_mode_pred_luma[count] = param.mpm_default;   // default mode
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Init tma parameter table
*/
//-----------------------------------------------------------------------------
void IntraPrediction::init_tma_param_table()
{
  // 4x4 L5
  atma_base_param[0][0].blk_height = 4;
  atma_base_param[0][0].blk_width  = 4;
  atma_base_param[0][0].tmp_blk_height = 2;
  atma_base_param[0][0].tmp_blk_width  = 2;
  atma_base_param[0][0].tmp_height = 1;
  atma_base_param[0][0].tmp_width  = 1;
  atma_base_param[0][0].stride = atma_base_param[0][0].blk_width;
  atma_base_param[0][0].Lnum = 5;
  atma_base_param[0][0].search_range = param.candidate_search_region;
  atma_base_param[0][0].sx = 4;
  atma_base_param[0][0].sy = 4;

  // 4x4 L9
  atma_base_param[0][1].blk_height = 4;
  atma_base_param[0][1].blk_width  = 4;
  atma_base_param[0][1].tmp_blk_height = 4;
  atma_base_param[0][1].tmp_blk_width  = 4;
  atma_base_param[0][1].tmp_height = 1;
  atma_base_param[0][1].tmp_width  = 1;
  atma_base_param[0][1].stride = atma_base_param[0][1].blk_width;
  atma_base_param[0][1].Lnum = 9;
  atma_base_param[0][1].search_range = param.candidate_search_region;
  atma_base_param[0][1].sx = 4;
  atma_base_param[0][1].sy = 4;

  // 2x8 L5
  atma_base_param[1][0].blk_height = 2;
  atma_base_param[1][0].blk_width  = 8;
  atma_base_param[1][0].tmp_blk_height = 2;
  atma_base_param[1][0].tmp_blk_width  = 2;
  atma_base_param[1][0].tmp_height = 1;
  atma_base_param[1][0].tmp_width  = 1;
  atma_base_param[1][0].stride = atma_base_param[1][0].blk_width;
  atma_base_param[1][0].Lnum = 5;
  atma_base_param[1][0].search_range = param.candidate_search_region;
  atma_base_param[1][0].sx = 8;
  atma_base_param[1][0].sy = 2;

  // 8x2 L5
  atma_base_param[2][0].blk_height = 8;
  atma_base_param[2][0].blk_width  = 2;
  atma_base_param[2][0].tmp_blk_height = 2;
  atma_base_param[2][0].tmp_blk_width  = 2;
  atma_base_param[2][0].tmp_height = 1;
  atma_base_param[2][0].tmp_width  = 1;
  atma_base_param[2][0].stride = atma_base_param[2][0].blk_width;
  atma_base_param[2][0].Lnum = 5;
  atma_base_param[2][0].search_range = param.candidate_search_region;
  atma_base_param[2][0].sx = 2;
  atma_base_param[2][0].sy = 8;

 // 8x8 L5
  atma_base_param[3][0].blk_height = 8;
  atma_base_param[3][0].blk_width  = 8;
  atma_base_param[3][0].tmp_blk_height = 2;
  atma_base_param[3][0].tmp_blk_width  = 2;
  atma_base_param[3][0].tmp_height = 1;
  atma_base_param[3][0].tmp_width  = 1;
  atma_base_param[3][0].stride = atma_base_param[3][0].blk_width;
  atma_base_param[3][0].Lnum = 5;
  atma_base_param[3][0].search_range = param.candidate_search_region;
  atma_base_param[3][0].sx = 4;
  atma_base_param[3][0].sy = 4;

 // 8x8 L9
  atma_base_param[3][1].blk_height = 8;
  atma_base_param[3][1].blk_width  = 8;
  atma_base_param[3][1].tmp_blk_height = 4;
  atma_base_param[3][1].tmp_blk_width  = 4;
  atma_base_param[3][1].tmp_height = 1;
  atma_base_param[3][1].tmp_width  = 1;
  atma_base_param[3][1].stride = atma_base_param[3][1].blk_width;
  atma_base_param[3][1].Lnum = 9;
  atma_base_param[3][1].search_range = param.candidate_search_region;
  atma_base_param[3][1].sx = 4;
  atma_base_param[3][1].sy = 4;

 // 16x16 L9
  atma_base_param[4][0].blk_height = 16;
  atma_base_param[4][0].blk_width  = 16;
  atma_base_param[4][0].tmp_blk_height = 4;
  atma_base_param[4][0].tmp_blk_width  = 4;
  atma_base_param[4][0].tmp_height = 1;
  atma_base_param[4][0].tmp_width  = 1;
  atma_base_param[4][0].stride = atma_base_param[4][0].blk_width;
  atma_base_param[4][0].Lnum = 9;
  atma_base_param[4][0].search_range = param.candidate_search_region_intra16x16;
  atma_base_param[4][0].sx = 4;
  atma_base_param[4][0].sy = 4;

  // 16x16 L9
  atma_base_param[4][1].blk_height = 16;
  atma_base_param[4][1].blk_width  = 16;
  atma_base_param[4][1].tmp_blk_height = 8;
  atma_base_param[4][1].tmp_blk_width  = 8;
  atma_base_param[4][1].tmp_height = 1;
  atma_base_param[4][1].tmp_width  = 1;
  atma_base_param[4][1].stride = atma_base_param[4][1].blk_width;
  atma_base_param[4][1].Lnum = 17;
  atma_base_param[4][1].search_range = param.candidate_search_region_intra16x16;
  atma_base_param[4][1].sx = 4;
  atma_base_param[4][1].sy = 4;

}
//-----------------------------------------------------------------------------
/*!
* \brief Set tma parameters
* \param[in] sip_type block sip type
* \param[in] mode prediction mode
* \param[in] chroma_channel chroma component index
* \param[in] is_up_right_available indicates if up right block is available
* \param[in] blk predicted block
*/
//-----------------------------------------------------------------------------
void IntraPrediction::set_tma_param(int sip_type, int mode, int chroma_channel, int is_up_right_available, Block* blk)
{
  int access_mode = -1;
  int default_value = 0;
  Block* pDes = NULL;

  if ( chroma_channel >=0 )
  {
    if( mode == TMA_L5_PRED_C)
      access_mode = 0;
    if( mode == TMA_L9_PRED_C)
      access_mode = 1;
  }
  else
  {
    if( mode == TMA_L9_PRED_MxN)
      access_mode = 0;
    if( mode == TMA_L17_PRED_MxN)
      access_mode = 1;
  }

  stma_param = atma_base_param[min(sip_type, 4)][access_mode];
  stma_param.sip_type = sip_type;

  if( blk )
  {
    pDes = blk;
    default_value = pDes->val[0];
  }

  if (chroma_channel >= 0)
  {
    stma_param.is_chroma = 1;
    stma_param.chroma_channel = chroma_channel;
    if (!blk)
    {
      pDes = &mprr_c_blk[chroma_channel][sip_type == SIP8x8][mode];
      default_value = this->mprr_c_blk[chroma_channel][sip_type == SIP8x8][DC_PRED_C].val[0];
    }
  }
  else
  {
    stma_param.is_chroma = 0;
    stma_param.chroma_channel = -1;
 
    if (!blk)
    {
      pDes = &mprr_blk[sip_type][mode];
      default_value = mprr_blk[sip_type][DC_PRED_MxN].val[0];
    }
  }

  stma_param.psPred = pDes;
  stma_param.default_value = default_value;
  stma_param.is_up_right_available = is_up_right_available;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get right edge of the search area
* \param[in] i Vertical block index
* \param[in] j Horizontal block index
* \param[in] ii Vertical subblock index
* \param[in] jj Horizontal subblock index
* \param[in] ptma_param template matching parameter set
* \return value of 1
*/
//-----------------------------------------------------------------------------
int IntraPrediction::get_right_edge(int i, int j, int ii, int jj, tma_param* ptma_param)
{
  int right_edge = 0;
  int sip_type = ptma_param->sip_type;
  int tmp_blk_width = ptma_param->tmp_blk_width;
  int tmp_blk_height = ptma_param->tmp_blk_height;
  int blk_width  = ptma_param->blk_width;

  if ( !ptma_param->is_chroma )
  {
    switch (sip_type)
    {
    case SIP16x16:
      {
        if( ptma_param->is_up_right_available? (ii>=tmp_blk_height && jj==(blk_width-tmp_blk_width)) : jj==(blk_width-tmp_blk_width))
        {
          right_edge = j * 4 + jj  + TMA_SEARCH_WINDOW_RIGHT;
        }
        else
        {
          right_edge = j * 4 + jj  + tmp_blk_width;
        }

        break;
      }
    case SIP8x8:
      {
        right_edge = j * 4 + blk_width - tmp_blk_width;
        break;
      }
    case SIP4x4:
      {
        if ( ((param.intra16x16) && (this->avail_up_right_4x4 == 0)) || ((!param.intra16x16) && (i & j & 1)))
        {
          right_edge = j * 4 + (blk_width-tmp_blk_width) + TMA_SEARCH_WINDOW_RIGHT;
        }
        else
        {
          right_edge = j * 4 + blk_width + (blk_width-tmp_blk_width);
        }
        break;
      }
    case SIP2x8:
      {
        right_edge = j * 8 + 6;
        break;
      }
    case SIP8x2:
      {
        right_edge = ((j>>2)<<3) + 6;
        break;
      }
    default:
      {
        break;
      }
    }
  }
  else
  {
    switch (sip_type)
    {
    case SIP8x8:
      {
        if (ii >= tmp_blk_height && jj == blk_width-tmp_blk_width)
        {
          right_edge = j * 4 + jj + TMA_SEARCH_WINDOW_RIGHT;
        }
        else
        {
          right_edge = j * 4 + jj + tmp_blk_width;
        }
        break;
      }
    case SIP4x4:
      {
        if (i & j & 1) // i and j are both odd
        {
          right_edge = j * 4 + (blk_width-tmp_blk_width) + TMA_SEARCH_WINDOW_RIGHT;
        }
        else
        {
          right_edge = j * 4 + blk_width + (blk_width-tmp_blk_width);
        }
        break;
      }
    }
  }

  return right_edge;
}
//-----------------------------------------------------------------------------
/*!
* \brief Performs Template Matching
*
* Creates the intra prediction samples.  Samples may be shifted by the offset
*
* \param[in] i Vertical block index
* \param[in] j Horizontal block index
* \param[in] ptma_param template matching parameter set
* \return value of 1
*/
//-----------------------------------------------------------------------------
int IntraPrediction::template_matching(int i0, int j0, tma_param* ptma_param)
{
  int i, j;
  
  // Copy TMA parameters into local variables
  int search_range = ptma_param->search_range;
  int blk_width = ptma_param->blk_width;
  int blk_height = ptma_param->blk_height;
  int tmp_blk_width = ptma_param->tmp_blk_width;
  int tmp_blk_height = ptma_param->tmp_blk_height;
  
  int search_area_width = search_range + blk_width + 8;
  int search_area_height = search_range + blk_height;
  
  int *search_area_origin = search_area + search_range * search_area_width + search_range;
  
  int origin_x = j0 * ptma_param->sx; // position in picture (samples)
  int origin_y = i0 * ptma_param->sy;
  
  int sb_width = (ptma_param->is_chroma) ? 4 * param.sb_width : 8 * param.sb_width;
  int sb_edge = (origin_x & -sb_width) + sb_width; // Next superblock edge position
  
  int pic_width = (ptma_param->is_chroma) ? cur2->width2 : cur2->width;
  int limit_x = imin(sb_edge, pic_width);
  int limit_y = (ptma_param->is_chroma) ? cur2->height2 : cur2->height;
  
  short *pic = (ptma_param->is_chroma) ? cur2->get_chroma_ptr(0, 0, ptma_param->chroma_channel) : cur2->get_luma_ptr(0, 0);
  
  // Some modes are different...
  bool special = (ptma_param->is_chroma) ? (ptma_param->sip_type == SIP8x8) : (ptma_param->sip_type == SIP16x16);
  
  //-------------------------------------------------
  // Prepare area to be searched
  
  // Prepare search area by setting all values to default
  int default_value = ptma_param->default_value;
  for (i = 0; i < search_area_width * search_area_height; i++)
    search_area[i] = default_value;
  
  // Fetch top
  for (i = imax(0, origin_y - search_range); i < origin_y; i++)
  {
    for (j = imax(0, origin_x - search_range); j < imin(origin_x + search_area_width - search_range, limit_x); j++)
    {
      search_area_origin[(i-origin_y)*search_area_width+(j-origin_x)] = pic[i*pic_width+j];
    }
  }
  
  // Fetch left
  for (i = origin_y; i < imin(origin_y + search_area_height - search_range, limit_y); i++)
  {
    for (j = imax(0, origin_x - search_range); j < origin_x; j++)
    {
      search_area_origin[(i-origin_y)*search_area_width+(j-origin_x)] = pic[i*pic_width+j];
    }
  }
  
  //------------------------------------------------
  // Start matching

  int ii, jj;
  assert(num_candidate_averaged > 0 && num_candidate_averaged <= 16);
  int cand_pos[16];
  int cand_sad[16];
  
  for(ii=0; ii < blk_height; ii+=tmp_blk_height)
  {
    for(jj=0; jj < blk_width; jj+=tmp_blk_width)
    {
      // Set pattern we want to match
      int target_pos = (ii-1) * search_area_width + (jj-1);
      int target[17];
      int k;
      for (k = 0; k <= tmp_blk_width; k++)
        target[k] = search_area_origin[target_pos + k];
      for (k = 1; k <= tmp_blk_height; k++)
        target[k + tmp_blk_width] = search_area_origin[target_pos + k * search_area_width];
      
      int threshold = 17*65536;
      for (k = 0; k < num_candidate_averaged; k++)
        cand_sad[k] = threshold;
      
      int ti = 0, tj = 0;
      int search_offy = blk_height - tmp_blk_height;
      
      if (special)
      {
        ti = ii;
        tj = jj;
        search_offy = 0;
      }
      
      int x, y;
      for (y = ti - search_range; y < ti + search_offy; y++)
      {
        for (x = tj - search_range; x < tj - tmp_blk_width; x++)
        {
          int test_pos = y * search_area_width + x;
          
          int sad = 0;
          for (k = 0; k <= tmp_blk_width; k++)
            sad += abs(target[k] - search_area_origin[test_pos + k]);
          
          if (sad >= threshold)
            continue;
          
          for (k = 1; k <= tmp_blk_height; k++)
            sad += abs(target[k + tmp_blk_width] - search_area_origin[test_pos + k * search_area_width]);
          
          if (sad >= threshold)
            continue;
          
          int k = num_candidate_averaged-1;
          while (k > 0 && cand_sad[k-1] > sad)
          {
            k--;
            cand_sad[k+1] = cand_sad[k];
            cand_pos[k+1] = cand_pos[k];
          }
          cand_sad[k] = sad;
          cand_pos[k] = test_pos;
          
          threshold = imin(cand_sad[0] + (candidate_sad_threshold << SAMPLE_ADD_PREC), cand_sad[num_candidate_averaged-1]);
        }//x
      }//y
      
      int right_edge = get_right_edge(i0, j0, ii, jj, ptma_param);
      if (special)
        right_edge = imin(right_edge, sb_edge - tmp_blk_width);
      
      for (y = ti - search_range; y < ti - tmp_blk_height; y++)
      {
        for (x = tj - tmp_blk_width; x < right_edge - origin_x; x++)
        {
          int test_pos = y * search_area_width + x;
          
          int sad = 0;
          for (k = 0; k <= tmp_blk_width; k++)
            sad += abs(target[k] - search_area_origin[test_pos + k]);
          
          if (sad >= threshold)
            continue;
          
          for (k = 1; k <= tmp_blk_height; k++)
            sad += abs(target[k + tmp_blk_width] - search_area_origin[test_pos + k * search_area_width]);
          
          if (sad >= threshold)
            continue;
          
          int k = num_candidate_averaged-1;
          while (k > 0 && cand_sad[k-1] > sad)
          {
            k--;
            cand_sad[k+1] = cand_sad[k];
            cand_pos[k+1] = cand_pos[k];
          }
          cand_sad[k] = sad;
          cand_pos[k] = test_pos;
          
          threshold = imin(cand_sad[0] + (candidate_sad_threshold << SAMPLE_ADD_PREC), cand_sad[num_candidate_averaged-1]);
        }//x
      }//y
      
      // Determine number of candidates to average
      for (k = 1; (1<<k) <= this->num_candidate_averaged; k++)
      {
        if (cand_sad[(1<<k)-1] - (this->candidate_sad_threshold << SAMPLE_ADD_PREC) > cand_sad[0])
          break;
      }
      k--;
      
      // Average candidates
      int *src = search_area_origin + search_area_width + 1;
      int *dst = search_area_origin + ii * search_area_width + jj;
      
      for (i = 0; i < tmp_blk_height; i++)
      {
        for (j = 0; j < tmp_blk_width; j++)
        {
          int val = 0;
          int offset = i*search_area_width+j;
          int z;

          for (z = 0; z < (1<<k); z++)
          {
            int pos = cand_pos[z];
            val += src[pos + offset];
          }
          
          if (k != 0)
            val = shift_right_round(val, k);
          
          dst[offset] = val;
        } // j
      } // i
    } // jj
  } // ii
  
  //----------------------------------
  // Copy prediction
  
  for (ii = 0; ii < blk_height; ii++)
  {
    for (jj = 0; jj < blk_width; jj++)
    {
      ptma_param->psPred->val[ii*blk_width+jj] = search_area_origin[ii*search_area_width+jj];
    }
  }
  
  return 1;
}


void IntraPrediction::reset_use_mode_luma()
{
  int count1;
  
  for (count1 = 0; count1 < NO_INTRA_PMODE_LUMA_MxN; count1++) 
  {
    use_mode_luma[count1] = 0;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Performs Intra Prediction for luma block
*
* Creates the intra prediction samples.
*
* \param[in] predmode prediction mode to create
* \param[in] i Vertical index
* \param[in] j Horizontal index
* \param[in] sip_type block size
* \param[in] pred_mode prediction mode to be generated for decoder
* \param[out] pred_blk  Pointer to block where the intra prediction sample is placed
* \return value of 1 when OK.
*/
//-----------------------------------------------------------------------------
int IntraPrediction::intrapred_luma_MxN(const int i, const int j, int sip_type, int pred_mode, Block *pred_blk)
{
  int jj, tmp_val;
  int count1, count2, zVR;
  int s0;
  
  int blk_width = Block::bwt[sip_type];
  int blk_height = Block::bht[sip_type];
  int blk_log_width = Block::log_bwt[sip_type];
  int blk_log_height = Block::log_bht[sip_type];
  int sb_width = 8 * param.sb_width;
  int j_pic = blk_width * j;
  int sb_edge = (j_pic & -sb_width) + sb_width;
  int boundary = imin(sb_edge, cur2->width);

  int block_available_left     = (j != 0);
  int block_available_up       = (i != 0);
  int block_available_up_left  = (j != 0) && (i != 0);
  int block_available_up_right = (i != 0) && (j_pic + blk_width < boundary);

  switch (sip_type)
  {
    case SIP16x16:
      block_available_up_right &= this->avail_up_right_16x16;
      break;
    case SIP8x8:
      if (param.intra16x16)
        block_available_up_right &= this->avail_up_right_8x8;
      break;
    case SIP4x4:
      block_available_up_right &= (i & j & 1) != 1; // Not lower-right 4x4 block within 8x8
      if (param.intra16x16)
        block_available_up_right &= this->avail_up_right_4x4;
      break;
    default:
      block_available_up_right = 0;
      break;
  }
  
  Block blk(sip_type);
  assert(blk_height <= 16 && blk_width <= 16);
  short val[69];
  short *val_topleft = &val[34];
  short *val_top = &val[35];
  short *val_left = &val[33];
  short *ref_ptr = cur2->get_luma_ptr(blk_height*i, blk_width*j);
    
  if (block_available_up)
  {
    for (jj = 0; jj < blk_width; jj++)
      val_top[jj] = ref_ptr[-cur2->width+jj];
  }
  else
  {
    for (jj = 0; jj < blk_width; jj++)
      val_top[jj] = DC_PRED_VALUE_LUMA;
  }
  
  if (block_available_up_right)
  {
    for (jj = 0; jj < blk_width; jj++)
      val_top[jj+blk_width] = ref_ptr[-cur2->width+blk_width+jj];
  }
  else
  {
    for (jj = 0; jj < blk_width; jj++)
      val_top[jj+blk_width] = val_top[blk_width-1];
  }
  
  if (block_available_left)
  {
    for (jj = 0; jj < blk_height; jj++)
      val_left[-jj] = ref_ptr[jj*cur2->width-1];
  }
  else
  {
    for (jj = 0; jj < blk_height; jj++)
      val_left[-jj] = DC_PRED_VALUE_LUMA;
  }
  
  if (block_available_up_left)
  {
    val_topleft[0] = ref_ptr[-cur2->width-1];
  }
  else
  {
    val_topleft[0] = DC_PRED_VALUE_LUMA;
  }
  
  if (sip_type == SIP8x8 || sip_type == SIP16x16)
  {
    LowPassForLumaMxNPred(val_topleft, blk_height, blk_width, block_available_up, block_available_left, block_available_up_left);
  }
  
  // left-bottom padding
  for (jj = 0; jj < blk_height+2; jj++)
    val_left[-blk_height-jj] = val_left[-blk_height+1];
  
  // top-right padding
  for (jj = 0; jj < 2; jj++)
    val_top[2*blk_width+jj] = val_top[2*blk_width-1];
  
  ///////////////////////////////
  // make DC prediction
  ///////////////////////////////
  if ((DC_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && ((pred_mode < 0) || (pred_mode == DC_PRED_MxN)))
  {
    s0 = 0;
    if (block_available_up && block_available_left && blk_width == blk_height)
    {
      // no edge
      tmp_val = 0;
      for (jj = 0; jj < blk_width; jj++)
      {
        tmp_val += val_top[jj];
        tmp_val += val_left[-jj];
      }
      s0 = shift_right_round(tmp_val, blk_log_width + 1);
    }
    else if (block_available_left && blk_height >= blk_width)
    {
      // upper edge
      tmp_val = 0;
      for (jj = 0; jj < blk_height; jj++)
      {
        tmp_val += val_left[-jj];
      }
      s0 = shift_right_round(tmp_val, blk_log_height);
    }
    else if (block_available_up && blk_width >= blk_height)
    {
      // left edge
      tmp_val = 0;
      for (jj = 0; jj < blk_width; jj++)
      {
        tmp_val += val_top[jj];
      }
      s0 = shift_right_round(tmp_val, blk_log_width);
    }
    else
    { //if (!block_available_up && !block_available_left)
      // top left corner, nothing to predict from
      s0 = DC_PRED_VALUE_LUMA;
    }
    
    // store DC prediction
    if (pred_mode < 0)
      pred_blk = &mprr_blk[sip_type][DC_PRED_MxN];

    for (count1 = 0; count1 < blk_height; count1++)
    {
      for (count2 = 0; count2 < blk_width; count2++)
      {
        pred_blk->val[count1*blk_width + count2] = s0;
      }
    }
    use_mode_luma[DC_PRED_MxN] = 1;
    
    ///////////////////////////////
    // make high precision edge prediction
    ///////////////////////////////
    if(param.edge_prediction_ipd == TRUE)
    {
      if (block_available_up || block_available_left)
      {
        if (pred_mode < 0)
          pred_blk = &mprr_blk[sip_type][DC_PRED_MxN];
        intrapred_luma_edge(i*blk_height, j*blk_width, sip_type, pred_blk);
      }
    }
  }
  
  ///////////////////////////////
  // make horiz and vert prediction
  ///////////////////////////////
  if ((VERT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && block_available_up && ((pred_mode < 0) || (pred_mode == VERT_PRED_MxN)))
  {
    if (pred_mode < 0)
      pred_blk = &mprr_blk[sip_type][VERT_PRED_MxN];
    
    for (count1 = 0; count1 < blk_height; count1 ++)
    {
      for (count2 = 0; count2 < blk_width; count2 ++)
      {
        pred_blk->val[count1*blk_width + count2] = val_top[count2];
      }
    }
    use_mode_luma[VERT_PRED_MxN] = 1;
  }
  
  if ((HOR_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && block_available_left && ((pred_mode < 0) || (pred_mode == HOR_PRED_MxN)))
  {
    if (pred_mode < 0)
      pred_blk = &mprr_blk[sip_type][HOR_PRED_MxN];
    
    for (count1 = 0; count1 < blk_height; count1 ++)
    {
      for (count2 = 0; count2 < blk_width; count2 ++)
      {
        pred_blk->val[count1*blk_width + count2] = val_left[-count1];
      }
    }
    
    use_mode_luma[HOR_PRED_MxN] = 1;
  }
  
  ///////////////////////////////////
  // make diagonal down left prediction
  ///////////////////////////////////
  if (DIAG_DOWN_LEFT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN && ((blk_height > blk_width) ? block_available_left : block_available_up))
  {
    if ((DIAG_DOWN_LEFT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && ((pred_mode < 0) || (pred_mode == DIAG_DOWN_LEFT_PRED_MxN)))
    {
      // Mode DIAG_DOWN_LEFT_PRED
      if (pred_mode < 0)
        pred_blk = &mprr_blk[sip_type][DIAG_DOWN_LEFT_PRED_MxN];

      for (count1 = 0; count1 < blk_height; count1 ++)
      {
        for (count2 = 0; count2 < blk_width; count2 ++)
        {
          jj = count1 + count2;
          if (blk_height > blk_width)
            pred_blk->val[count1*blk_width + count2] = shift_right_round(val_left[-jj] + 2*val_left[-jj-1] + val_left[-jj-2], 2);
          else
            pred_blk->val[count1*blk_width + count2] = shift_right_round(val_top[jj] + 2*val_top[jj+1] + val_top[jj+2], 2);
        }
      }

      use_mode_luma[DIAG_DOWN_LEFT_PRED_MxN] = 1;
    }
  }

    ///////////////////////////////////
    // make vertical left prediction
    ///////////////////////////////////
  if (VERT_LEFT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN && ((blk_height > blk_width) ? block_available_left : block_available_up))
  {
    if ((VERT_LEFT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && ((pred_mode < 0) || (pred_mode == VERT_LEFT_PRED_MxN)))
    {
      if (pred_mode < 0)
        pred_blk = &mprr_blk[sip_type][VERT_LEFT_PRED_MxN];
      
      for (count1 = 0; count1 < blk_height; count1 ++)
      {
        for (count2 = 0; count2 < blk_width; count2++)
        {
          if (blk_height > blk_width)
          {
            jj = 2*count2 + count1;
            pred_blk->val[count1*blk_width + count2] = shift_right_round(val_left[-jj-1] + 2*val_left[-jj-2] + val_left[-jj-3], 2);
          }
          else
          {
            jj = count2 + (count1 >> 1);
            if (count1 & 1)
            {
              pred_blk->val[count1*blk_width + count2] = shift_right_round(val_top[jj] + 2*val_top[jj+1] + val_top[jj+2], 2);
            }
            else
            {
              pred_blk->val[count1*blk_width + count2] = shift_right_round(val_top[jj] + val_top[jj+1], 1);
            }
          }
        }
      }

      use_mode_luma[VERT_LEFT_PRED_MxN] = 1;
    }
  }
  
  ///////////////////////////////////
  // make diagonal down right prediction
  ///////////////////////////////////
  if (DIAG_DOWN_RIGHT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN && VERT_RIGHT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN && HOR_DOWN_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN &&
      block_available_up && block_available_left && block_available_up_left)
  {
    if ((DIAG_DOWN_RIGHT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && ((pred_mode < 0) || (pred_mode == DIAG_DOWN_RIGHT_PRED_MxN)))
    {
      // Mode DIAG_DOWN_RIGHT_PRED
      if (pred_mode < 0)
        pred_blk = &mprr_blk[sip_type][DIAG_DOWN_RIGHT_PRED_MxN];

      for (count1 = 0; count1 < blk_height; count1 ++)
      {
        for (count2 = 0; count2 < blk_width; count2 ++)
        {
          jj = count2 - count1;
          pred_blk->val[count1*blk_width + count2] = shift_right_round(val_topleft[jj-1] + 2*val_topleft[jj] + val_topleft[jj+1], 2);
        }
      }
      use_mode_luma[DIAG_DOWN_RIGHT_PRED_MxN] = 1;
    }
    
    ///////////////////////////////////
    // make vertical right prediction
    ///////////////////////////////////
    if ((VERT_RIGHT_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && ((pred_mode < 0) || (pred_mode == VERT_RIGHT_PRED_MxN)))
    {
      if (pred_mode < 0)
        pred_blk = &mprr_blk[sip_type][VERT_RIGHT_PRED_MxN];
      
      for (count1 = 0; count1 < blk_height; count1 ++)
      {
        for (count2 = 0; count2 < blk_width; count2 ++)
        {
          zVR = 2*count2 - count1;
          
          if (zVR >= 0)
          {
            jj = zVR >> 1;
            if (zVR & 1)
              pred_blk->val[count1*blk_width + count2] = shift_right_round(val_topleft[jj] + 2*val_topleft[1+jj] + val_topleft[2+jj], 2);
            else
              pred_blk->val[count1*blk_width + count2] = shift_right_round(val_topleft[jj] + val_topleft[1+jj], 1);
          }
          else
          {
            pred_blk->val[count1*blk_width + count2] = shift_right_round(val_topleft[zVR] + 2*val_topleft[zVR+1] + val_topleft[zVR+2], 2);
          }
        }
      }
      use_mode_luma[VERT_RIGHT_PRED_MxN] = 1;
    }
    
    ///////////////////////////////////
    // make horizontal down prediction
    ///////////////////////////////////
    if ((HOR_DOWN_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && ((pred_mode < 0) || (pred_mode == HOR_DOWN_PRED_MxN)))
    {
      if (pred_mode < 0)
        pred_blk = &mprr_blk[sip_type][HOR_DOWN_PRED_MxN];

      for (count1 = 0; count1 < blk_height; count1 ++)
      {
        for (count2 = 0; count2 < blk_width; count2 ++)
        {
          zVR = 2*count1 - count2;
          
          if (zVR >= 0)
          {
            jj = zVR >> 1;
            if (zVR & 1)
            {
              pred_blk->val[count1*blk_width + count2] = shift_right_round(val_topleft[-jj] + 2*val_topleft[-jj-1] + val_topleft[-jj-2], 2);
            }
            else
            {
              pred_blk->val[count1*blk_width + count2] = shift_right_round(val_topleft[-jj] + val_topleft[-jj-1], 1);
            }
          }
          else
          {
            pred_blk->val[count1*blk_width + count2] = shift_right_round(val_topleft[-zVR] + 2*val_topleft[-zVR-1] + val_topleft[-zVR-2], 2);
          }
        }
      }
      use_mode_luma[HOR_DOWN_PRED_MxN] = 1;
    }
  }
  
  ///////////////////////////////////
  // make horizontal up prediction
  ///////////////////////////////////
  if ((HOR_UP_PRED_MxN < NO_INTRA_PMODE_LUMA_MxN) && ((blk_width > blk_height) ? block_available_up : block_available_left) && ((pred_mode < 0) || (pred_mode == HOR_UP_PRED_MxN)))
  {
    if (pred_mode < 0)
      pred_blk = &mprr_blk[sip_type][HOR_UP_PRED_MxN];
    
    for (count1 = 0; count1 < blk_height; count1 ++)
    {
      for (count2 = 0; count2 < blk_width; count2 ++)
      {
        if (blk_width > blk_height)
        {
          jj = (param.bugfix88) ? 2*count1 + count2 : count1 + count2;
          pred_blk->val[count1*blk_width + count2] = shift_right_round(val_top[1+jj] + 2*val_top[2+jj] + val_top[3+jj], 2);
        }
        else
        {
          jj = count1 + (count2 >> 1);
          if (count2 & 1)
          {
            pred_blk->val[count1*blk_width + count2] = shift_right_round(val_left[-jj] + 2*val_left[-jj-1] + val_left[-jj-2], 2);
          }
          else
          {
            pred_blk->val[count1*blk_width + count2] = shift_right_round(val_left[-jj] + val_left[-jj-1], 1);
          }
        }
      }
    }
    use_mode_luma[HOR_UP_PRED_MxN] = 1;
  }
  
  return 1;
}

//-----------------------------------------------------------------------------
/*!
* \brief Performs Intra Prediction for chroma block
*
* Creates the intra prediction samples.
*
* \param[in] i Vertical index
* \param[in] j Horizontal index
* \param[in] c chroma component
* \param[in] sip_type chroma block sip type
* \param[in] iqpC chroma QP
* \param[in] pred_mode prediction mode to create
* \param[out] blk_pred  Pointer to block where the intra prediction sample is placed
* \return value of 1 when OK. -1 or -2 when not.
*/
//-----------------------------------------------------------------------------
int IntraPrediction::intrapred_chroma_MxN(const int i, const int j, const int c, int sip_type, const int iqpC, int pred_mode, Block *pred_blk)
{
  int count1, count2, bsize = 0, log2bsize = 0, tmp_val, s, jj;
  int block_available_up;
  int block_available_left;

  bsize = Block::bht[sip_type];
  log2bsize = Block::log_bht[sip_type];

  Block  blk(bsize, bsize);
  assert(2 * bsize <= 16);
  short val[17];

  block_available_left     = (j != 0);
  block_available_up       = (i != 0);

  // form predictor pels
  if (block_available_up)
  {
    cur2->get_blk_chroma(&blk, i-1, j, c);
    for (jj = 0; jj < bsize; jj++)
      val[jj] = blk.val[bsize*(bsize-1)+jj];
  }
  else
  {
    for (jj = 0; jj < bsize; jj++)
      val[jj] = DC_PRED_VALUE_CHROMA;
  }

  if (block_available_left)
  {
    cur2->get_blk_chroma(&blk, i, j-1, c);
    for (jj = 0; jj < bsize; jj++)
      val[jj+bsize] = blk.val[bsize*jj+bsize-1];
  }
  else
  {
    for (jj = 0; jj < bsize; jj++)
      val[jj+bsize] = DC_PRED_VALUE_CHROMA;
  }

  if (block_available_up && block_available_left)
  {
    cur2->get_blk_chroma(&blk, i-1, j-1, c);
    val[2*bsize] = blk.val[bsize*bsize-1];
  }
  else
  {
    val[2*bsize] = DC_PRED_VALUE_CHROMA;
  }


  if (sip_type == SIP8x8)
#ifdef CHROMA_INTRA_ADAPTIVE_LPF
    LowPassForChromaPred(val, block_available_up, block_available_left, bsize, iqpC);
#else
    LowPassForChromaPred(val, block_available_up, block_available_left, bsize);
#endif // CHROMA_INTRA_ADAPTIVE_LPF

  // compute all chroma intra prediction modes for both U and V
  // DC prediction
  if ((pred_mode < 0) || (pred_mode == DC_PRED_C))
  {
    s = DC_PRED_VALUE_CHROMA;

    //===== get prediction value =====
    if (!block_available_up && !block_available_left)
    {
      ;
    }
    else if (!block_available_up && block_available_left)
    {
      tmp_val = 0;
      for (jj = 0; jj < bsize; jj++)
        tmp_val += val[jj+bsize];
      s = shift_right_round(tmp_val, log2bsize);
    }
    else if (block_available_up && !block_available_left)
    {
      tmp_val = 0;
      for (jj = 0; jj < bsize; jj++)
        tmp_val += val[jj];
      s = shift_right_round(tmp_val, log2bsize);
    }
    else if (block_available_up && block_available_left)
    {
      tmp_val = 0;
      for (jj = 0; jj < bsize; jj++)
      {
        tmp_val += val[jj];
        tmp_val += val[jj+bsize];
      }
      s = shift_right_round(tmp_val, log2bsize+1);
    }

    //===== prediction =====
    if (pred_mode < 0)
      pred_blk = &mprr_c_blk[c][sip_type == SIP8x8][DC_PRED_C];
    for (count1 = 0; count1 < bsize; count1++)
    {
      for (count2 = 0; count2 < bsize; count2++)
      {
        pred_blk->val[count1*bsize+count2] = s;
      }
    }
    use_mode_chroma[c][DC_PRED_C] = 1;
  }

  // vertical prediction
  if ((pred_mode < 0) || (pred_mode == VERT_PRED_C))
  {
    if (block_available_up)
    {
      if (pred_mode < 0)
        pred_blk = &mprr_c_blk[c][sip_type == SIP8x8][VERT_PRED_C];
      for (count1 = 0; count1 < bsize; count1 ++)
        for (count2 = 0; count2 < bsize; count2 ++)
          pred_blk->val[count1*bsize+count2] = val[count2];

      use_mode_chroma[c][VERT_PRED_C] = 1;
    }
  }
  // horizontal prediction
  if ((pred_mode < 0) || (pred_mode == HOR_PRED_C))
  {
    if (block_available_left)
    {
      if (pred_mode < 0)
        pred_blk = &mprr_c_blk[c][sip_type == SIP8x8][HOR_PRED_C];
      for (count1 = 0; count1 < bsize; count1 ++)
        for (count2 = 0; count2 < bsize; count2 ++)
          pred_blk->val[count1*bsize+count2] = val[count1+bsize];

      use_mode_chroma[c][HOR_PRED_C] = 1;
    }
  }

  // plane prediction
  if ((pred_mode < 0) || (pred_mode == PLANE_PRED_C))
  {
    s = 0;
    if (block_available_up && block_available_left)
    {
      if (pred_mode < 0)
        pred_blk = &mprr_c_blk[c][sip_type == SIP8x8][PLANE_PRED_C];

      if (param.bugfix89)
      {
        int bsize2 = power2(log2bsize - 1);

        int H = -bsize2 * val[2*bsize];
        for (count2 = 0; count2 < bsize; count2++)
          H += (count2 - bsize2 + 1) * val[count2];

        int V = -bsize2 * val[2*bsize];
        for (count1 = 0; count1 < bsize; count1++)
          V += (count1 - bsize2 + 1) * val[bsize+count1];

        int a = 16 * (val[bsize-1] + val[2*bsize-1]);
        int b = (bsize == 8) ? (34 * H + 32) >> 6 : (205 * H + 32) >> 6;
        int c = (bsize == 8) ? (34 * V + 32) >> 6 : (205 * V + 32) >> 6;

        for (count1 = 0; count1 < bsize; count1++)
        {
          for (count2 = 0; count2 < bsize; count2++)
          {
            pred_blk->val[count1*bsize+count2] = clip((a + b *(count2-bsize2+1) + c * (count1-bsize2+1) + 16) >> 5, -MID_SAMPLE_VAL << SAMPLE_ADD_PREC, (MAX_SAMPLE_VAL - MID_SAMPLE_VAL) << SAMPLE_ADD_PREC);
          }
        }
      }
      else
      {
        for (count1 = 0; count1 < bsize; count1++)
        {
          for (count2 = 0; count2 < bsize; count2++)
          {
            if (count1 + count2 <= (bsize-3))
            {
              s = shift_right_round(val[count1+count2+1] + val[bsize+count1+count2+1], 1);
            } 
            else
            {
              s = shift_right_round(val[bsize-1] + val[2*bsize-1], 1);
            }
            pred_blk->val[count1*bsize+count2] = s;
          }
        }
      }
      use_mode_chroma[c][PLANE_PRED_C] = 1;
    }
  }

  return 1;
}
/*!
 *************************************************************************************
 * \brief
 *    Prefiltering for Intra prediction
 *
 * Creates the intra prediction samples.
 *
 * \param[in] PredPel neighboring predictor pels
 * \param[in] block_up_left, block_up, block_left 
 *************************************************************************************
 */
void IntraPrediction::LowPassForLumaMxNPred(short *PredPel, int bheight, int bwidth, int available_top, int available_left, int available_topleft)
{
  assert(bheight <= 16 && bwidth <= 16);
  short tmp[51];
  short *p = &tmp[bheight+1];
  
  if (!available_top && !available_left && !available_topleft)
    return;
  
  memcpy(tmp+1, PredPel-bheight, (bheight+2*bwidth+1)*sizeof(short));
  tmp[0] = tmp[1];
  tmp[bheight+2*bwidth+2] = tmp[bheight+2*bwidth+1];
  
  if (!available_topleft)
  {
    p[0] = (available_top) ? p[1] : p[-1];
  }

  int start = (available_left) ? -bheight : 1;
  int end = (available_top) ? 2*bwidth : -1;
  
  int i;
  for (i = start; i <= end; i++)
  {
    PredPel[i] = shift_right_round(p[i-1] + 2*p[i] + p[i+1], 2);
  }
}

void IntraPrediction::generate_prediction(int i, int j, int sip_type, int pred_mode, Block *blk_pre, WFrame *ref)
{
  this->cur2 = ref;
  
  int iy, jy;
  switch (sip_type)
  {
  case SIP16x16:
    iy = 4 * i;
    jy = 4 * j;
    break;
  case SIP8x8:
    iy = 2 * i;
    jy = 2 * j;
    break;
  default:
    iy = i;
    jy = j;
    break;
  }

  int tmp_avail_up_right;
  switch (sip_type)
  {
  case SIP16x16:
    tmp_avail_up_right = avail_up_right_16x16;
    break;
  case SIP4x4:
    tmp_avail_up_right = avail_up_right_4x4;
    break;
  default:
    tmp_avail_up_right = avail_up_right_8x8;
    break;
  }

  // Large template not valid for 8x2 and 2x8
  assert(pred_mode != TMA_L17_PRED_MxN || (sip_type != SIP8x2 && sip_type != SIP2x8));

  reset_use_mode_luma();
  
  bool isTMA = (pred_mode == TMA_L9_PRED_MxN || pred_mode == TMA_L17_PRED_MxN);
  int pred_mode2 = (isTMA) ? DC_PRED_MxN : pred_mode;

  intrapred_luma_MxN(i, j, sip_type, pred_mode2, blk_pre);

  if (pred_mode == TMA_L9_PRED_MxN || pred_mode == -1)
  {
    set_tma_param(sip_type, TMA_L9_PRED_MxN, -1, tmp_avail_up_right, blk_pre);
    use_mode_luma[TMA_L9_PRED_MxN] = template_matching(iy, jy, &stma_param);
  }

  if (pred_mode == TMA_L17_PRED_MxN || (pred_mode == -1 && (sip_type != SIP8x2 && sip_type != SIP2x8)))
  {
    set_tma_param(sip_type, TMA_L17_PRED_MxN, -1, tmp_avail_up_right, blk_pre);
    use_mode_luma[TMA_L17_PRED_MxN] = template_matching(iy, jy, &stma_param);
  }
}

void IntraPrediction::generate_prediction_chroma(int i, int j, int c, int sip_type, int pred_mode, Block *blk_pre, int qpC, WFrame *ref)
{
  this->cur2 = ref;
  
  int iy, jy;
  switch (sip_type)
  {
    case SIP8x8:
      iy = 2 * i;
      jy = 2 * j;
      break;
    default:
      iy = i;
      jy = j;
      break;
  }
  
  int tmp_avail_up_right;
  switch (sip_type)
  {
    case SIP4x4:
      tmp_avail_up_right = avail_up_right_8x8;
      break;
    default:
      tmp_avail_up_right = avail_up_right_16x16;
      break;
  }
  
  bool isTMA = (pred_mode == TMA_L5_PRED_C || pred_mode == TMA_L9_PRED_C);
  int pred_mode2 = (isTMA) ? DC_PRED_C : pred_mode;
  intrapred_chroma_MxN(i, j, c, sip_type, qpC, pred_mode2, blk_pre);
  
  if (pred_mode == TMA_L5_PRED_C || pred_mode == -1)
  {
    set_tma_param(sip_type, TMA_L5_PRED_C, c, tmp_avail_up_right, blk_pre);
    use_mode_chroma[c][TMA_L5_PRED_C] = template_matching(iy, jy, &stma_param);
  }
  
  if (pred_mode == TMA_L9_PRED_C || pred_mode == -1)
  {
    set_tma_param(sip_type, TMA_L9_PRED_C, c, tmp_avail_up_right, blk_pre);
    use_mode_chroma[c][TMA_L9_PRED_C] = template_matching(iy, jy, &stma_param);
  }
}
