/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file LoopFilter.cpp
*
* \brief Implementation of LoopFilter class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "LoopFilter.H"
#include "LoopFilterNull.H"
#include "LoopFilterOvercomplete3.H"
#include "Param.H"

#define MVTH 0 //!< Threshold for MV difference

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in 4x4 luma block units
* \param[in] w width of frame in 4x4 luma block units
*/
//-----------------------------------------------------------------------------
LoopFilter::LoopFilter(int h, int w) 
{
  height = h;
  width = w;
  height2 = h>>1;
  width2 = w>>1;
  
  lumaH = new int[height*width];
  lumaV = new int[height*width];
  chromaH[0] = new int[height2*width2];
  chromaH[1] = new int[height2*width2];
  chromaV[0] = new int[height2*width2];
  chromaV[1] = new int[height2*width2];
  
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
LoopFilter::~LoopFilter() 
{
  delete[] lumaH;
  delete[] lumaV;
  delete[] chromaH[0];
  delete[] chromaH[1];
  delete[] chromaV[0];
  delete[] chromaV[1];
  
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Reset boundary mode arrays
*
* Sets all boundary mode values to v
*
* \param[in] v Reset mode value
*/
//-----------------------------------------------------------------------------
void LoopFilter::reset(int v) 
{
  int i,j;
  int c;
  
  for (i=0; i<height; i++)
  {
    for (j=0; j<width; j++) 
    {
      lumaH[width*i+j] = v;
      lumaV[width*i+j] = v;
    }
  }
  
  for (c=0; c<2; c++)
  {
    for (i=0; i<height2; i++)
    {
      for (j=0; j<width2; j++) 
      {
        chromaH[c][width2*i+j] = v;
        chromaV[c][width2*i+j] = v;
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Create loop filter instance
*
* Creates an instance based on provided type
*
* \param[in] h height of frame in 4x4 luma block units
* \param[in] w width of frame in 4x4 luma block units
* \param[in] type Type of filter. See ndf_mode parameter in configuration file.
* \return Instance
*/
//-----------------------------------------------------------------------------
LoopFilter *LoopFilter::create(int h, int w, int type) 
{
  LoopFilter *lf;
  
  
  switch (type) 
  {
      
    case 3:
      lf = new LoopFilterOvercomplete3(h, w);
      break;
      
    default:
      lf = new LoopFilterNull(h, w);
      break;
  }
  
  return lf;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
