/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file LoopFilterOvercomplete3.cpp
*
* \brief Implementation of LoopFilterOvercomplete3 class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"
#include "LoopFilterOvercomplete3.H"
#include "Param.H"

#ifdef DENOISE_MT

#include <pthread.h>
pthread_key_t denoise_thread_number_key;
#define denoise_thrd_no (reinterpret_cast<long>(pthread_getspecific(denoise_thread_number_key)))

// minimize code change.
#define trans_luma trans_luma_thrd[denoise_thrd_no]
#define trans_chroma trans_chroma_thrd[denoise_thrd_no]
#define thres_luma thres_luma_thrd[denoise_thrd_no]
#define thres_chroma thres_chroma_thrd[denoise_thrd_no]
#define m_window_luma m_window_luma_thrd[denoise_thrd_no]
#define m_window_chroma m_window_chroma_thrd[denoise_thrd_no]

#endif

//-----------------------------------------------------------------------------
// Definitions
//-----------------------------------------------------------------------------

const int BSIZE=4; //!< Block size
const int LG_EXT=BSIZE/2; //!< Large Extent
const int SM_EXT=BSIZE/4; //!< Small Extent

const int CLG_EXT=BSIZE/2; //!< Large Extent for Chroma
const int CSM_EXT=BSIZE/4; //!< Small Extent for Chroma

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in 4x4 luma block units
* \param[in] w width of frame in 4x4 luma block units
*/
//-----------------------------------------------------------------------------
LoopFilterOvercomplete3::LoopFilterOvercomplete3(int h, int w) : LoopFilter(h, w) 
{
  assert(h > 0);
  assert(w > 0);
  
  m_use_weighting = param.ndf_weighting != 0;
  m_do_two_rounds = param.ndf_two_rounds != 0;
  m_aggressive    = param.ndf_aggressive != 0;
  m_signif_cnt = param.ndf_signif_cnt;
  m_mvd_th = param.ndf_mvd_th;
  
  // Get strength override parameters
  m_override_strength[DCM_SKIP          ] = INVALID_THRES;
  m_override_strength[DCM_INTRA_QUANT   ] = param.ndf_s_intra_quant;
  m_override_strength[DCM_PRED_SIGNIF   ] = param.ndf_s_pred_signif;
  m_override_strength[DCM_PRED_QUANT    ] = param.ndf_s_pred_quant;
  m_override_strength[DCM_PRED_MOT      ] = param.ndf_s_pred_mot;
  m_override_strength[DCM_MOT_DISC      ] = param.ndf_s_mot_disc;
  m_override_strength[DCM_IC_STRONG     ] = param.ndf_s_ic_strong;
  m_override_strength[DCM_IC_INTERMED   ] = param.ndf_s_ic_intermed;
  m_override_strength[DCM_IC_WEAK       ] = param.ndf_s_ic_weak;

  // Get extent override parameters
  m_override_extent[DCM_SKIP          ] = INVALID_EXTENT;
  m_override_extent[DCM_INTRA_QUANT   ] = param.ndf_x_intra_quant;
  m_override_extent[DCM_PRED_SIGNIF   ] = param.ndf_x_pred_signif;
  m_override_extent[DCM_PRED_QUANT    ] = param.ndf_x_pred_quant;
  m_override_extent[DCM_PRED_MOT      ] = param.ndf_x_pred_mot;
  m_override_extent[DCM_MOT_DISC      ] = param.ndf_x_mot_disc;
  m_override_extent[DCM_IC_STRONG     ] = param.ndf_x_ic_strong;
  m_override_extent[DCM_IC_INTERMED   ] = param.ndf_x_ic_intermed;
  m_override_extent[DCM_IC_WEAK       ] = param.ndf_x_ic_weak;
  
  strength_chroma = param.ndf_s_chroma;
    
  int i;
  m_frame_height = 4 * height;
  m_frame_width  = 4 * width;
  
  effected   = new int *[m_frame_height];
  thres_mask = new int *[m_frame_height];
  
  for (i = 0; i < m_frame_height; i++)
  {
    effected[i]   = new int[m_frame_width];
    thres_mask[i] = new int[m_frame_width];
  }
  
#ifdef DENOISE_MT
  // We split the image by rows for now, so these arrays are per-thread.
  // If we split per-column we can allocate less?
  m_data_accum_thrd = new DataLinesThread[param.decode_ndenoisethreads];
  m_norm_accum_thrd = new DataLinesThread[param.decode_ndenoisethreads];
  m_norm_accum2_thrd = new int*[param.decode_ndenoisethreads];
  for (int t = 0; t < param.decode_ndenoisethreads; t++)
  {
    for (i = 0; i < 8; i++)
    {
      m_data_accum_thrd[t].lines[i] = new int[m_frame_width];
      //memset(m_data_accum_thrd[t].lines[i], 0, m_frame_width*sizeof(int));
      m_norm_accum_thrd[t].lines[i] = new int[m_frame_width];
      //memset(m_norm_accum_thrd[t].lines[i], 0, m_frame_width*sizeof(int));
    }
    m_norm_accum2_thrd[t] = new int[m_frame_width];
  }
#else
  for (i = 0; i < 8; i++)
  {
    m_data_accum[i] = new int[m_frame_width];
    m_norm_accum[i] = new int[m_frame_width];
  }
  m_norm_accum2 = new int[m_frame_width];
#endif
  
#ifdef DENOISE_MT
  m_indata = new int[m_frame_width * m_frame_height];
  m_outdata = new int[m_frame_width * m_frame_height];
#else
  m_data = new int[m_frame_width * m_frame_height];
#endif
  
  if (m_do_two_rounds)
  {
    m_est  = new int[m_frame_width * m_frame_height];
  }
  
  bool use_oracle = false;
  int luma_tsize = (param.ndf_trsf_8x8) ? 8 : 4;
  int chroma_tsize = 4;

#ifdef DENOISE_MT
  pthread_key_create(&denoise_thread_number_key, NULL);

  trans_luma_thrd = new DenoiseTransform *[param.decode_ndenoisethreads];
  trans_chroma_thrd = new DenoiseTransform *[param.decode_ndenoisethreads];
  thres_luma_thrd = new DenoiseThreshold *[param.decode_ndenoisethreads];
  thres_chroma_thrd = new DenoiseThreshold *[param.decode_ndenoisethreads];
  for (i = 0; i < param.decode_ndenoisethreads; i++)
  {
    trans_luma_thrd[i]   = DenoiseTransform::create(luma_tsize, false, m_frame_width);
    trans_chroma_thrd[i] = DenoiseTransform::create(chroma_tsize, false, m_frame_width >> 1);
    thres_luma_thrd[i]   = DenoiseThreshold::create(luma_tsize, m_use_weighting, use_oracle);
    thres_chroma_thrd[i] = DenoiseThreshold::create(chroma_tsize, m_use_weighting, use_oracle);  
  }
  m_window_luma_thrd = new FilterWindow[param.decode_ndenoisethreads];
  m_window_chroma_thrd = new FilterWindow[param.decode_ndenoisethreads];
#else
  trans_luma   = DenoiseTransform::create(luma_tsize, false, m_frame_width);
  trans_chroma = DenoiseTransform::create(chroma_tsize, false, m_frame_width >> 1);
  thres_luma   = DenoiseThreshold::create(luma_tsize, m_use_weighting, use_oracle);
  thres_chroma = DenoiseThreshold::create(chroma_tsize, m_use_weighting, use_oracle);  
#endif
}

LoopFilterOvercomplete3::~LoopFilterOvercomplete3()
{
  int i;

  for (i = 0; i < m_frame_height; i++)
  {
    delete[] effected[i];
    delete[] thres_mask[i];
  }
  
  delete[] effected;
  delete[] thres_mask;
  
#ifdef DENOISE_MT
  for (int t = 0; t < param.decode_ndenoisethreads; t++)
  {
    for (i = 0; i < 8; i++)
    {
      delete[] m_data_accum_thrd[t].lines[i];
      delete[] m_norm_accum_thrd[t].lines[i];
    }
    delete[] m_norm_accum2_thrd[t];
  }
  delete[] m_data_accum_thrd;
  delete[] m_norm_accum_thrd;
  delete[] m_norm_accum2_thrd;
#else
  for (i = 0; i < 8; i++)
  {
    delete[] m_data_accum[i];
    delete[] m_norm_accum[i];
  }
  delete[] m_norm_accum2;
#endif
  
#ifdef DENOISE_MT
  delete[] m_indata;
  delete[] m_outdata;
#else
  delete[] m_data;
#endif
  
  if (m_do_two_rounds)
  {
    delete[] m_est;
  }
  
#ifdef DENOISE_MT
  for (i = 0; i < param.decode_ndenoisethreads; i++)
  {
    delete trans_luma_thrd[i];
    delete trans_chroma_thrd[i];
    delete thres_luma_thrd[i];
    delete thres_chroma_thrd[i];
  }
  delete[] trans_luma_thrd;
  delete[] trans_chroma_thrd;
  delete[] thres_luma_thrd;
  delete[] thres_chroma_thrd;
  delete[] m_window_luma_thrd;
  delete[] m_window_chroma_thrd;
#else
  delete trans_luma;
  delete trans_chroma;
  delete thres_luma;
  delete thres_chroma;  
#endif
}

/*!
* Lookup table for 1024 * pow(2.0, cur->qp / 16.0)
*/
int LoopFilterOvercomplete3::pow_table[POW_TABLE_LIMIT] =
{
  0x000400, 0x00042d, 0x00045d, 0x00048e, 0x0004c2, 0x0004f8, 0x000530, 0x00056b, 0x0005a8, 0x0005e8, 0x00062b, 0x000671, 0x0006ba, 0x000706, 0x000756, 0x0007a9, 
  0x000800, 0x00085b, 0x0008b9, 0x00091c, 0x000983, 0x0009ef, 0x000a60, 0x000ad6, 0x000b50, 0x000bd1, 0x000c56, 0x000ce2, 0x000d74, 0x000e0d, 0x000eac, 0x000f52, 
  0x001000, 0x0010b5, 0x001173, 0x001238, 0x001307, 0x0013df, 0x0014c0, 0x0015ab, 0x0016a1, 0x0017a1, 0x0018ad, 0x0019c5, 0x001ae9, 0x001c1a, 0x001d58, 0x001ea5, 
  0x002000, 0x00216b, 0x0022e5, 0x002471, 0x00260e, 0x0027bd, 0x002980, 0x002b56, 0x002d41, 0x002f42, 0x00315a, 0x003389, 0x0035d1, 0x003833, 0x003ab0, 0x003d49, 
  0x004000, 0x0042d5, 0x0045cb, 0x0048e2, 0x004c1c, 0x004f7b, 0x0052ff, 0x0056ac, 0x005a82, 0x005e84, 0x0062b4, 0x006712, 0x006ba2, 0x007066, 0x007560, 0x007a93, 
  0x008000, 0x0085ab, 0x008b96, 0x0091c4, 0x009838, 0x009ef5, 0x00a5ff, 0x00ad58, 0x00b505, 0x00bd09, 0x00c567, 0x00ce25, 0x00d745, 0x00e0cd, 0x00eac1, 0x00f525, 
  0x010000, 0x010b56, 0x01172c, 0x012388, 0x013070, 0x013dea, 0x014bfe, 0x015ab0, 0x016a0a, 0x017a11, 0x018ace, 0x019c49, 0x01ae8a, 0x01c19a, 0x01d582, 0x01ea4b, 
  0x020000, 0x0216ab, 0x022e57, 0x02470f, 0x0260e0, 0x027bd5, 0x0297fb, 0x02b561, 0x02d414, 0x02f423, 0x03159d, 0x033892, 0x035d14, 0x038333, 0x03ab03, 0x03d496
};

//-----------------------------------------------------------------------------
// Tuned thresholds
//-----------------------------------------------------------------------------
int LoopFilterOvercomplete3::lum_thres_adpround[NUM_LOOP_MODES3] = {0, 36, 32, 18, 32, 16, 22, 18,  6};
int LoopFilterOvercomplete3::chr_thres_adpround[NUM_LOOP_MODES3] = {0, 30, 16, 14, 16,  8, 16, 14,  4};
int LoopFilterOvercomplete3::lum_thres_adpround_aggr[NUM_LOOP_MODES3] = {0, 36, 32, 18, 32, 16, 22, 18,  6};
int LoopFilterOvercomplete3::chr_thres_adpround_aggr[NUM_LOOP_MODES3] = {0, 30, 16, 14, 16,  8, 16, 14,  4};
 
//-----------------------------------------------------------------------------
// Extents
//-----------------------------------------------------------------------------
int LoopFilterOvercomplete3::lum_extent[NUM_LOOP_MODES3] =      {0, LG_EXT, LG_EXT, SM_EXT, LG_EXT, SM_EXT, LG_EXT, SM_EXT, SM_EXT};
int LoopFilterOvercomplete3::lum_extent_aggr[NUM_LOOP_MODES3] = {0, LG_EXT, LG_EXT, LG_EXT, LG_EXT, LG_EXT, LG_EXT, LG_EXT, LG_EXT};
int LoopFilterOvercomplete3::chr_extent[NUM_LOOP_MODES3] =      {0, CLG_EXT, CLG_EXT, CSM_EXT, CLG_EXT, CSM_EXT, CLG_EXT, CSM_EXT, CSM_EXT};
int LoopFilterOvercomplete3::chr_extent_aggr[NUM_LOOP_MODES3] = {0, CLG_EXT, CLG_EXT, CLG_EXT, CLG_EXT, CLG_EXT, CLG_EXT, CLG_EXT, CLG_EXT};

int LoopFilterOvercomplete3::lum_threshold_table[NUM_LOOP_MODES3][2]; //!< Luma threshold table for the current QP
int LoopFilterOvercomplete3::lum_extent_table[NUM_LOOP_MODES3]; //!< Luma extent table for the current QP
int LoopFilterOvercomplete3::chr_threshold_table[NUM_LOOP_MODES3][2]; //!< Chroma threshold table for the current QP
int LoopFilterOvercomplete3::chr_extent_table[NUM_LOOP_MODES3]; //!< Chroma extent table for the current QP

//-----------------------------------------------------------------------------
// General Functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*! 
* \brief Sets the luma and chroma thresholds
*/
//-----------------------------------------------------------------------------
void LoopFilterOvercomplete3::set_thresholds()
{
  int (*tmp_lum)[NUM_LOOP_MODES3],(*tmp_chr)[NUM_LOOP_MODES3];
  int i;

  if(m_aggressive) 
  {
    
    tmp_lum=&lum_thres_adpround_aggr;
    tmp_chr=&chr_thres_adpround_aggr;
  }
  else 
  {
    
    tmp_lum=&lum_thres_adpround;
    tmp_chr=&chr_thres_adpround;
  }

  for(i=0;i<NUM_LOOP_MODES3;i++) 
  {
    lum_threshold_table[i][0]=(*tmp_lum)[i];
    lum_threshold_table[i][1]=DENOMINATOR;

    chr_threshold_table[i][0]=(*tmp_chr)[i];
    chr_threshold_table[i][1]=DENOMINATOR;
  }
}

//-----------------------------------------------------------------------------
/*! 
* \brief Sets the luma and chroma extents
*/
//-----------------------------------------------------------------------------
void LoopFilterOvercomplete3::set_extents()
{
  int i;

  if(m_aggressive) 
  {
    for(i=0;i<NUM_LOOP_MODES3;i++) 
    {
      lum_extent_table[i]=lum_extent_aggr[i];
      chr_extent_table[i]=chr_extent_aggr[i];
    }
  }
  else 
  {
    for(i=0;i<NUM_LOOP_MODES3;i++) 
    {
      lum_extent_table[i]=lum_extent[i];
      chr_extent_table[i]=chr_extent[i];
    }
  }

}

//-----------------------------------------------------------------------------
/*! 
* \brief Sets the threshold and extent values for a given mode based on input parameters
* 
* \param[in] strength_val Strength value
* \param[in] extent_val Extent value
* \param[in] mode Number of the mode
* \param[in] mode_name Name of the mode
* \param[in] debug Flag when set, causes an output reflecting the operations performed
*/
//-----------------------------------------------------------------------------
void LoopFilterOvercomplete3::set_param_from_config_file()
{
  int modeIdx;
  
  for (modeIdx = 0; modeIdx < NUM_LOOP_MODES3; modeIdx++)
  {
    if (m_override_strength[modeIdx] != INVALID_THRES)
    {
      lum_threshold_table[modeIdx][0] = m_override_strength[modeIdx];
    }
    
    if (m_override_extent[modeIdx] != INVALID_EXTENT)
    {
      lum_extent_table[modeIdx] = m_override_extent[modeIdx];
    }
  }  
}

//-----------------------------------------------------------------------------
// Methods for the current class
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Determine luma boundary modes
*
* \param[in] mf Motion field of current frame
* \param[in] zmap Zero block map of current frame
*/
//-----------------------------------------------------------------------------
void LoopFilterOvercomplete3::set_luma(MotionField *mf, ContextMap *zmap) 
{
  assert(mf != NULL);
  assert(zmap != NULL);
  
  int i, j;
  int ncoeffA, ncoeffB;
  PredBlk *mvA, *mvB, *mv_cur;
  int ia, ja, ib, jb;
  
  for (i = 0; i < height; i++) 
  {
    lumaH[width * i + 0] = DCM_SKIP;
    
    for (j = 1; j < width; j++) 
    {  
      ia = i >> 1; ja = (j - 1) >> 1;
      ib = i >> 1; jb = j >> 1;

      mvA = mf->get_pblk_4x4(i, j - 1);
      mvB = mf->get_pblk_4x4(i, j);

      mv_cur = mf->get_pblk_unsafe(i>>1, j>>1);
      // motion sharing parallel to boundary
      if ( (j%2 == 0) && (mv_cur->mvshare_size[1] == 2) )
        mvB = mvA;
      
      ncoeffA = zmap->get(i, j - 1);
      
      if (mf->get_pblk(ia, ja)->intra != NULL)
        ncoeffA = -1;
      if (ncoeffA != 0) 
      {
        if (mf->get_transform_blocksize(ia, ja) == 1)
        {
          ncoeffA = zmap->get(2 * ia, 2 * ja);
        }
        else 
        {
          ncoeffA = param.ndf_4x4_mult * zmap->get(i, j - 1);
        }
      }
      
      ncoeffB = zmap->get(i, j);
      if(mf->get_pblk(ib, jb)->intra != NULL)
        ncoeffB = -1;
      if (ncoeffB != ZERO) 
      {
        if(mf->get_transform_blocksize(ib, jb)==1)
        {
          ncoeffB = zmap->get(2*ib, 2*jb);
        }
        else 
        {
          ncoeffB = param.ndf_4x4_mult * zmap->get(i, j);
        }
      }
  
      lumaH[width * i + j] = mode_decision_4x4(mvA, mvB, ncoeffA, ncoeffB, m_mvd_th, j, mv_cur->mvshare_size[1], 0);
    }
  }
  
  for (j=0; j<width; j++) 
  {
    lumaV[0 + j] = DCM_SKIP;
    
    for (i=1; i<height; i++) 
    {
      ia = (i - 1) >> 1; ja = j >> 1;
      ib = i >> 1; jb = j >> 1;

      mvA = mf->get_pblk_4x4(i - 1, j);
      mvB = mf->get_pblk_4x4(i, j);

      mv_cur = mf->get_pblk_unsafe(i>>1, j>>1);
      // motion sharing parallel to boundary
      if ( (i%2 == 0) && (mv_cur->mvshare_size[0] == 2) )
        mvB = mvA;

      ncoeffA = zmap->get(i - ONE, j);
      if(mf->get_pblk(ia, ja)!=NULL && mf->get_pblk(ia, ja)->intra!=NULL)
        ncoeffA = -1;
      if (ncoeffA != ZERO) 
      {
        if(mf->get_transform_blocksize(ia, ja)==1)
        {
          ncoeffA = zmap->get(2*ia, 2*ja);
        }
        else 
        {
          ncoeffA = param.ndf_4x4_mult * zmap->get(i - ONE, j);
        }
      }
      
      ncoeffB = zmap->get(i, j);
      if(mf->get_pblk(ib, jb)->intra!=NULL)
        ncoeffB = -1;
      if (ncoeffB != ZERO) 
      {
        if(mf->get_transform_blocksize(ib, jb)==1)
        {
          ncoeffB = zmap->get(2*ib, 2*jb);
        }
        else 
        {
          ncoeffB = param.ndf_4x4_mult * zmap->get(i, j);
        }
      }
      
      lumaV[width * i + j] = mode_decision_4x4(mvA, mvB, ncoeffA, ncoeffB, m_mvd_th, i, mv_cur->mvshare_size[0], 0);
    }
  }
  
  for (i=0; i<height; i++) 
  {  
    for (j=0; j<width; j++) 
    {  
      ia = i>>ONE; ja = j>>ONE;
      
      ncoeffA = zmap->get(i, j);
      if(mf->get_pblk(ia, ja)->intra!=NULL)
        ncoeffA = -1;
      if (ncoeffA != ZERO) 
      {
        if(mf->get_transform_blocksize(ia, ja)==1)
        {
          ncoeffA = zmap->get(2*ia, 2*ja);
        }
        else 
        {
          ncoeffA = param.ndf_4x4_mult * zmap->get(i, j);
        }
      }
      if (ncoeffA > m_signif_cnt || mf->get_pblk(i >> ONE, j >> ONE)->intra != NULL) 
      {    
        lumaH[width * i + j] |= DCM_DENOISE_THIS_BLOCK;
        lumaV[width * i + j] |= DCM_DENOISE_THIS_BLOCK;
        if (i == 0)
          lumaV[j] = lumaV[width + j];
        if (j == 0)
          lumaH[width * i] = lumaH[width * i + 1];
      }    
    }
  }
  
}

//-----------------------------------------------------------------------------
/*!
* \brief Determine chroma boundary modes
*
* \param[in] mf Motion field of current frame
* \param[in] zmap Zero block map of current frame
* \param[in] c Chroma component index
*/
//-----------------------------------------------------------------------------
void LoopFilterOvercomplete3::set_chroma(MotionField *mf, ContextMap *zmap, int c) 
{
  int i, j;
  int ncoeffA, ncoeffB;
  PredBlk *mvA, *mvB, *mv_cur;
  int ia, ja, ib, jb;
  
  for (i=0; i<height2; i++) 
  {  
    chromaH[c][width2 * i + 0] = DCM_SKIP;
    
    for (j=1; j<width2; j++) 
    {      
      ia = i; ja = j-ONE;
      ib = i; jb = j;

      mvA = mf->get_pblk(ia, ja);
      mvB = mf->get_pblk(ib, jb);
      
      mv_cur = mf->get_pblk_unsafe(i, j);
      // motion sharing parallel to boundary
      if (mv_cur->mvshare_size[1] > 0)
        mvB = mvA;

      ncoeffA = zmap->get(i, j - ONE);
      if(mf->get_pblk(ia, ja)!=NULL && mf->get_pblk(ia, ja)->intra!=NULL)
        ncoeffA = -1;
      if (ncoeffA != ZERO) 
      {
        ncoeffA = param.ndf_4x4_mult * zmap->get(i, j - ONE);
      }
      
      ncoeffB = zmap->get(i, j);
      if(mf->get_pblk(ib, jb)->intra!=NULL)
        ncoeffB = -1;
      if (ncoeffB != ZERO) 
      {
        ncoeffB = param.ndf_4x4_mult * zmap->get(i, j);
      }
  
      chromaH[c][width2 * i + j] = mode_decision_4x4(mvA, mvB, ncoeffA, ncoeffB, TWO * m_mvd_th, j*2, mv_cur->mvshare_size[1], 1);
    }
  }
  
  for (j=0; j<width2; j++) 
  {  
    chromaV[c][0 + j] = DCM_SKIP;
    for (i=1; i<height2; i++) 
    {      
      ia = i-ONE; ja = j;
      ib = i; jb = j;
      
      mvA = mf->get_pblk(ia, ja);
      mvB = mf->get_pblk(ib, jb);
      
      mv_cur = mf->get_pblk_unsafe(i, j);
      // motion sharing parallel to boundary
      if (mv_cur->mvshare_size[0] > 0)
        mvB = mvA;

      ncoeffA = zmap->get(i - ONE, j);
      if(mf->get_pblk(ia, ja)!=NULL && mf->get_pblk(ia, ja)->intra!=NULL)
        ncoeffA = -1;
      if (ncoeffA != ZERO) 
      {
        ncoeffA = param.ndf_4x4_mult * zmap->get(i - ONE, j);
      }
      
      ncoeffB = zmap->get(i, j);
      if(mf->get_pblk(ib, jb)->intra!=NULL)
        ncoeffB = -1;
      if (ncoeffB != ZERO) 
      {
        ncoeffB = param.ndf_4x4_mult * zmap->get(i, j);
      }
      chromaV[c][width2 * i + j] = mode_decision_4x4(mvA, mvB, ncoeffA, ncoeffB, TWO * m_mvd_th, i*2, mv_cur->mvshare_size[0], 1);
    }
  }
  
  for (i=0; i<height2; i++)
  {  
    for (j=0; j<width2; j++) 
    {    
      ncoeffA = zmap->get(i, j);
      if(mf->get_pblk(i, j)->intra!=NULL)
        ncoeffA = -1;
      if (ncoeffA != ZERO) 
      {
        ncoeffA = param.ndf_4x4_mult * zmap->get(i, j);
      }
      if (ncoeffA > m_signif_cnt || mf->get_pblk(i, j)->intra != NULL) 
      {        
        chromaH[c][width2 * i + j] |= DCM_DENOISE_THIS_BLOCK;
        chromaV[c][width2 * i + j] |= DCM_DENOISE_THIS_BLOCK;
        if (i == 0)
          chromaV[c][j] = chromaV[c][width2 + j];
        if (j == 0)
          chromaH[c][width2 * i] = chromaH[c][width2 * i + 1];
      }
    }
  }
}

void LoopFilterOvercomplete3::calc_thres_values(int threshold, bool is_luma)
{
  int i;
  
  int (* threshold_table)[2] = (is_luma) ? lum_threshold_table : chr_threshold_table;
  int *threshold_values = (is_luma) ? lum_threshold_values : chr_threshold_values;
  
  for (i = 0; i < NUM_LOOP_MODES3; i++)
  {
    int val = threshold_table[i][0];
    int den = threshold_table[i][1];
    
    assert(val != INVALID_THRES);    
    assert(den >= 1);
    
    threshold_values[i] = (val * threshold + den / 2) / den;
  }  
}

void LoopFilterOvercomplete3::calc_extent_values(bool is_luma)
{
  int i;
  int *table = (is_luma) ? lum_extent_table : chr_extent_table;
  int *values = (is_luma) ? lum_extent_values : chr_extent_values;
  
  for (i = 0; i < NUM_LOOP_MODES3; i++)
  {
    int val = table[i];
    assert(val != INVALID_EXTENT);
    values[i] = val;
  }
}


void LoopFilterOvercomplete3::set_effected(int k_begin, int k_end, int l_begin, int l_end)
{
  assert(k_begin >= 0);
  assert(l_begin >= 0);
  assert(k_end <= m_frame_height);
  assert(l_end <= m_frame_width);

  int k, l;
  
  for (k = k_begin; k < k_end; k++)
  {
    for (l = l_begin; l < l_end; l++)
    {
      effected[k][l] = 1;
    }
  }  
}

void LoopFilterOvercomplete3::set_threshold(int k_begin, int k_end, int l_begin, int l_end, int thres)
{
  assert(k_begin >= 0);
  assert(l_begin >= 0);
  assert(k_end < m_frame_height);
  assert(l_end < m_frame_width);
  assert(thres > 0); // Shouldn't be here if thres is no larger than 0

  int k, l;

  for (k = k_begin; k <= k_end; k++)
  {
    for (l = l_begin; l <= l_end; l++)
    {
      thres_mask[k][l] = imax(thres_mask[k][l], thres);     
    }
  }
}

bool LoopFilterOvercomplete3::set_thres_extent(int mode, int *thres, int *extent, int *bs_inside, bool is_luma)
{
  assert(thres != NULL);
  assert(extent != NULL);
  assert(bs_inside != NULL);
  
  bool denoise_this_block = false;
  
  // handle the modifier
  if (mode & DCM_DENOISE_THIS_BLOCK) 
  {
    mode &= ~DCM_DENOISE_THIS_BLOCK;
    denoise_this_block = true;
  }

  // handle the modifier
  if (mode & DCM_BS_QUARTER)
  {
    mode &= ~DCM_BS_QUARTER;
    *bs_inside=1;
  } else if (mode & DCM_BS_HALF)
  {
    mode &= ~DCM_BS_HALF;
    *bs_inside=2;
  } else
    *bs_inside=0;

  assert(mode >= 0 && mode < NUM_LOOP_MODES3);

  *thres = (is_luma) ? lum_threshold_values[mode] : chr_threshold_values[mode];
  *extent = (is_luma) ? lum_extent_values[mode] : chr_extent_values[mode];

  assert(*thres >= 0);
  assert(*extent <= 2);
  
  return denoise_this_block;
}


void LoopFilterOvercomplete3::set_thres_mask(const int *lcH, const int *lcV, int bsize, bool is_luma)
{
  assert(lcH != NULL);
  assert(lcV != NULL);
  
  FilterWindow window = (is_luma) ? m_window_luma : m_window_chroma;
  int i,j,fi,fj;
  int thres=0, extent;
  int bs_inside, bs_thres, bs_extent;
  bool denoise_this_block;
  int blk_index_4x4;
  int k_begin, k_end;
  int l_begin, l_end;
  int row;
  int tsize = (is_luma) ? trans_luma->getSize() : trans_chroma->getSize();
  int pitch = (is_luma) ? this->width : this->width2;
  int row_max = (is_luma) ? 4 * this->height - 1 : 4 * this->height2 - 1;
  int col_max = (is_luma) ? 4 * this->width  - 1 : 4 * this->width2  - 1;

  assert((window.row_begin & 3) == 0);
  assert((window.row_end   & 3) == 3);
  assert((window.col_begin & 3) == 0);
  assert((window.col_end   & 3) == 3);
    
  // Clear arrays
  for (row = window.row_begin; row <= window.row_end; row++) 
  {    
    memset(thres_mask[row] + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(int));
    memset(effected[row]   + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(int));
  }

  // Load threshold and extent for motion sharing edges
  bs_thres  = (is_luma) ? lum_threshold_values[DCM_MOT_DISC] : chr_threshold_values[DCM_MOT_DISC];
  bs_extent = (is_luma) ? lum_extent_values[DCM_MOT_DISC] : chr_extent_values[DCM_MOT_DISC];

  // H
  for (fi = window.row_begin; fi <= window.row_end; fi += 4) 
  {    
    for (fj = window.col_begin; fj <= window.col_end; fj += 4) 
    {
      i = (fi + window.row_offset) >> 2;
      j = (fj + window.col_offset) >> 2;
      blk_index_4x4 = i * pitch + j;
      
      if (lcH[blk_index_4x4] != DCM_SKIP) 
      {
        denoise_this_block = set_thres_extent(lcH[blk_index_4x4], &thres, &extent, &bs_inside, is_luma);
        
        if (j > 0) 
        {
          // We are now processing the top row and left col.
          // This is to make sure we denoise those blocks completely
          // if denoise_this_block=1. Without j>0 condition, it is
          // sometimes the case that effected is set but nothing
          // gets thresholded in top/left leading to warnings...
          
          k_begin = imax(fi - extent, 0);
          k_end   = imin(fi + bsize + extent, row_max + 1);
          l_begin = imax(fj - extent, 0);
          l_end   = imin(fj + extent, col_max + 1);
          if (thres > 0) set_effected(k_begin, k_end, l_begin, l_end); // Do nothing if thres <= 0

          // Set extent for motion sharing edges
          if (bs_inside && (bs_thres > 0) )
          {
            k_begin = imax(fi - bs_extent, 0);
            k_end   = imin(fi + bsize + bs_extent, row_max + 1);
            l_begin = imax(fj + bs_inside - bs_extent, 0);
            l_end   = imin(fj + bs_inside + bs_extent, col_max + 1);
            set_effected(k_begin, k_end, l_begin, l_end);
          }
        }
        
        //set thresholds due to the boundary condition
        k_begin = imax(fi - tsize + 1, 0);
        k_end   = imin(fi + bsize - 1, row_max);
        l_begin = imax(fj - tsize + 1, 0);
        l_end   = imin(fj - 1, col_max);
        if (thres > 0) set_threshold(k_begin, k_end, l_begin, l_end, thres); // Do nothing if thres <= 0

        // Set threshold for motion sharing edges
        if (bs_inside && (bs_thres > 0) )
        {
          k_begin = imax(fi - tsize + 1, 0);
          k_end   = imin(fi + bsize - 1, row_max);
          l_begin = imax(fj + bs_inside - tsize + 1, 0);
          l_end   = imin(fj + bs_inside - 1, col_max);
          set_threshold(k_begin, k_end, l_begin, l_end, bs_thres);
        }
        
        //set thresholds due to condition inside the block
        if (denoise_this_block) 
        {
          if (thres > 0) set_threshold(fi, fi + bsize - tsize, fj, fj + bsize - tsize, thres); // Do nothing if thres <= 0
          set_effected(fi, fi + bsize, fj, fj + bsize);
        }
      }
    }
  }  
  
  // V  
  for (fj = window.col_begin; fj <= window.col_end; fj += 4) 
  {
    for (fi = window.row_begin; fi <= window.row_end; fi += 4) 
    {
      i = (fi + window.row_offset) >> 2;
      j = (fj + window.col_offset) >> 2;
      blk_index_4x4 = i * pitch + j;

      if (lcV[blk_index_4x4] != DCM_SKIP) 
      {
        denoise_this_block = set_thres_extent(lcV[blk_index_4x4], &thres, &extent, &bs_inside, is_luma);

        if (i > 0) 
        {
          k_begin = imax(fi - extent, 0);
          k_end   = imin(fi + extent, row_max + 1);
          l_begin = imax(fj - extent, 0);
          l_end   = imin(fj + bsize + extent, col_max + 1);
          if (thres > 0) set_effected(k_begin, k_end, l_begin, l_end); // Do nothing if thres <= 0

          // Set extent for motion sharing edges
          if (bs_inside && (bs_thres > 0) )
          {
            k_begin = imax(fi + bs_inside - bs_extent, 0);
            k_end   = imin(fi + bs_inside + bs_extent, row_max + 1);
            l_begin = imax(fj - bs_extent, 0);
            l_end   = imin(fj + bsize + bs_extent, col_max + 1);
            set_effected(k_begin, k_end, l_begin, l_end);
          }
        }
        
        k_begin = imax(fi - tsize + 1, 0);
        k_end   = imin(fi - 1, row_max);
        l_begin = imax(fj - tsize + 1, 0);
        l_end   = imin(fj + bsize - 1, col_max);
        if (thres > 0) set_threshold(k_begin, k_end, l_begin, l_end, thres); // Do nothing if thres <= 0
        
        // Set threshold for motion sharing edges
        if (bs_inside && (bs_thres > 0) )
        {
          k_begin = imax(fi + bs_inside - tsize + 1, 0);
          k_end   = imin(fi + bs_inside - 1, row_max);
          l_begin = imax(fj - tsize + 1, 0);
          l_end   = imin(fj + bsize - 1, col_max);
          set_threshold(k_begin, k_end, l_begin, l_end, bs_thres);
        }

        if (denoise_this_block) 
        {
          if (thres > 0) set_threshold(fi, fi + bsize - tsize, fj, fj + bsize - tsize, thres); // Do nothing if thres <= 0
          set_effected(fi, fi + bsize, fj, fj + bsize);
        }
      }
    }
  }

  if (is_luma)
    skip_transforms(param.ndf_lc_luma);
  else
    skip_transforms(param.ndf_lc_chroma);
}

void LoopFilterOvercomplete3::skip_transforms(int skip_pattern)
{
  int i, j;

  switch(skip_pattern)
  {
    case 0:
      break;
      
    case 1: //skip 50% of transforms - option 1
      for (i = 0; i < m_frame_height; i++)
      {
        for (j = (i&1); j < m_frame_width; j=j+2)
        {
          thres_mask[i][j] = 0;
        }
      }
      break;

    case 2: //skip 50% of transforms - option 2
      for (i = 0; i < m_frame_height; i++)
      {
        for (j = ((i+1)&1); j < m_frame_width; j=j+2)
        {
          thres_mask[i][j] = 0;
        }
      }
      break;

    case 3: //skip 25% of transforms - option 1
      for (i = 0; i < m_frame_height; i++)
      {
        for (j = 0; j < m_frame_width; j++)
        {
          if ((i & j & 1) == 0)
            thres_mask[i][j] = 0;
        }
      }
      break;

    case 4: //skip 25% of transforms - option 2
      for (i = 0; i < m_frame_height; i++)
      {
        for (j = 0; j < m_frame_width; j++)
        {
          if ((i | j) & 1)
            thres_mask[i][j] = 0;
        }
      }
      break;

    default:
      printf("Option for ndf_lc not supported. No transforms skipped\n");
      break;
  }
}

int LoopFilterOvercomplete3::s_qptable[32] = 
{
  1,  2,  2,  3,  3,  3,  4,  4, 
  6,  6,  7,  8,  8,  9, 22, 20, 
  32, 32, 34, 34, 36, 36, 38, 38,
  32, 32, 34, 34, 36, 36, 38, 38
};

void LoopFilterOvercomplete3::rotate(int **handle, int size)
{
  assert(handle != NULL);
  assert(size > 0);
  
  int i;
  int *tmp = handle[0];
  
  for (i = 1; i < size; i++)
  {
    handle[i - 1] = handle[i];
  }
  
  handle[size - 1] = tmp;
}

void LoopFilterOvercomplete3::denoise_row(const int *src, const int *est, int src_row, int src_pitch, int col_begin, int col_end, DenoiseTransform *t, DenoiseThreshold *h, int *th)
{
  int tsize_minus1 = t->getSize() - 1;
  
  int col;
  int col_loop_end = col_end - tsize_minus1;
  int j;
  
  memset(m_data_accum[tsize_minus1] + col_begin, 0, (col_end - col_begin + 1) * sizeof(m_data_accum[0][0]));
  
  int *accum_ptr = m_norm_accum[tsize_minus1];
  memset(accum_ptr + col_begin, 0, (col_end - col_begin + 1) * sizeof(accum_ptr[0]));

  bool skip_row = t->setMask(th);
  
  if (skip_row)
    return;
  
  t->doColumnsForward(src, src_row, src_pitch, col_begin, col_end);
  if (est)
  {
    t->doColumnsForward(est, src_row, src_pitch, col_begin, col_end, true);
  }
  
  t->resetAccumBuffer(col_begin, col_end);
  
  for (col = col_begin; col <= col_loop_end; col++)
  {
    int my_th = th[col];
    
    if (my_th == 0)
      continue;
    
    int scale;
    
    t->doRowsForward(col);
    if (est)
    {
      t->doRowsForward(col, true);
      scale = h->go(t->getBlock(), t->getBlock(true), (my_th + 1) >> 1);
    }
    else
    {
      scale = h->go(t->getBlock(), my_th);
    }
    
    t->doRowsInverse(col);
    
    scale *= t->getNorm();
    
    for (j = 0; j <= tsize_minus1; j++)
    {
      accum_ptr[col + j] += scale;
    }    
  }

  for (col = col_begin; col <= col_end; col++)
  {
    m_norm_accum2[col] += accum_ptr[col];
  }
  
  t->doColumnsInverse(m_data_accum, col_begin, col_end);
}

//-----------------------------------------------------------------------------
/*!
 * \brief Normalize one row of accumulated data
 *
 * \param[in]  row       Row number
 * \param[in]  col_begin First column to process
 * \param[in]  col_end   Last column to process (inclusive)
 * \param[in]  orig      Original (noisy) image
 * \param[in]  pitch     Pitch to address orig and out
 * \param[out] out       Output (denoised) image
 */
//-----------------------------------------------------------------------------
void LoopFilterOvercomplete3::normalize_row(int row, int col_begin, int col_end, const int *orig, int pitch, int *out)
{
  const int *orig_ptr = orig + row * pitch;
  int *data_ptr = m_data_accum[0];
  int *out_ptr = out + row * pitch;
  int *norm_ptr = m_norm_accum[0];
  int *effected_ptr = effected[row];
  int col;
  
  for (col = col_begin; col <= col_end; col++)
  {
    int norm = m_norm_accum2[col];
    m_norm_accum2[col] = norm - norm_ptr[col];
    
    if (effected_ptr[col] && norm > 0)
    {
      int tmp = (norm + 1) >> 1;
      int val = data_ptr[col];
      int sign = val >> 31;
      val = (val ^ sign) - sign;
      val = (unsigned int)(val + tmp) / (unsigned int)norm;
      val = (val ^ sign) - sign;
      val = clip(val, -NDF_SAMPLE_OFFSET, 255-NDF_SAMPLE_OFFSET);
      out_ptr[col] = val;
    }
    else
    {
      out_ptr[col] = orig_ptr[col];
    }
  }
}

void LoopFilterOvercomplete3::adjust_thres(int scalar)
{
  int k;
  for(k=0; k<NUM_LOOP_MODES3; k++)
  {
    lum_threshold_values[k] = int(lum_threshold_values[k]*scalar>>3);
    chr_threshold_values[k] = int(chr_threshold_values[k]*scalar>>3);
  }
}

#ifdef DENOISE_MT
void LoopFilterOvercomplete3::init_filter_plane(unsigned char *img, int /*height*/, int width, int threshold, int *H, int *V, int /*qp*/, bool is_luma)
{
  int my_threshold;
  DenoiseTransform *t = (is_luma) ? trans_luma : trans_chroma;
  int bsize = 4;
  int tsize = t->getSize();
  DenoiseThreshold *h = (is_luma) ? thres_luma : thres_chroma;
  int i, j;
  FilterWindow window = (is_luma) ? m_window_luma : m_window_chroma;
  
  assert(t->getBasisEnergy() < 0x7fffffff / threshold); // Make sure we don't overflow
  my_threshold = shift_right_round(threshold * t->getBasisEnergy(), 10);
  
  calc_thres_values(my_threshold, is_luma);
  calc_extent_values(is_luma);
  
  h->setWeights(6);

  set_thres_mask(H, V, bsize, is_luma);

  //int qp_thres = param.lf_spthres ? s_qptable[qp/6] : 255;
  for (i = window.row_begin; i <= window.row_end; i++)
  {
    for (j = window.col_begin; j <= window.col_end; j++)
    {
      m_indata[width * i + j] = img[width * i + j] - NDF_SAMPLE_OFFSET;
    }
  }
}
#endif

void LoopFilterOvercomplete3::filter_plane(unsigned char *img, int /*height*/, int width, int threshold, int *H, int *V, int /*qp*/, bool is_luma)
{
  int my_threshold;
  DenoiseTransform *t = (is_luma) ? trans_luma : trans_chroma;
  int bsize = 4;
  int tsize = t->getSize();
  DenoiseThreshold *h = (is_luma) ? thres_luma : thres_chroma;
#ifdef DENOISE_MT
  int i;
#else
  int i, j;
#endif
  FilterWindow window = (is_luma) ? m_window_luma : m_window_chroma;
  
  assert(t->getBasisEnergy() < 0x7fffffff / threshold); // Make sure we don't overflow
  my_threshold = shift_right_round(threshold * t->getBasisEnergy(), 10);
  
#ifdef DENOISE_MT
    // If we're not processing the first part of the image, we back ourselves up
    // by tsize-1 lines to initialize our rolling buffers correctly.
    // We don't write results for these lines.
    int write_data = 1;
    if (window.row_begin >= tsize)
    {
      window.row_begin -= (tsize-1);
      write_data = 0;
    }
#endif

#ifdef DENOISE_MT
    // done prior for the whole frame.
    h->setWeights(6);
#else
  calc_thres_values(my_threshold, is_luma);
  calc_extent_values(is_luma);
  
  h->setWeights(6);
  
  set_thres_mask(H, V, bsize, is_luma);

  for (i = window.row_begin; i <= window.row_end; i++)
  {
    for (j = window.col_begin; j <= window.col_end; j++)
    {
      m_data[width * i + j] = img[width * i + j] - NDF_SAMPLE_OFFSET;
    }
  }
#endif
  
  if (m_do_two_rounds)
  {
    for (i = 1; i < tsize; i++)
    {
      memset(m_data_accum[i] + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_data_accum[0][0]));
      memset(m_norm_accum[i] + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_norm_accum[0][0]));
    }
    memset(m_norm_accum2 + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_norm_accum2[0]));

    for (i = window.row_begin; i <= window.row_end; i++)
    {
      rotate(m_data_accum, tsize);
      rotate(m_norm_accum, tsize);
      
      if (i <= window.row_end - tsize + 1)
      {
#ifdef DENOISE_MT
        denoise_row(m_indata, NULL, i, width, window.col_begin, window.col_end, t, h, thres_mask[i]);
#else
        denoise_row(m_data, NULL, i, width, window.col_begin, window.col_end, t, h, thres_mask[i]);
#endif
      }
      
#ifdef DENOISE_MT
#else
      normalize_row(i, window.col_begin, window.col_end, m_data, width, m_est);
#endif
    }

    for (i = 1; i < tsize; i++)
    {
      memset(m_data_accum[i] + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_data_accum[0][0]));
      memset(m_norm_accum[i] + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_norm_accum[0][0]));
    }
    memset(m_norm_accum2 + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_norm_accum2[0]));

    for (i = window.row_begin; i <= window.row_end; i++)
    {
      rotate(m_data_accum, tsize);
      rotate(m_norm_accum, tsize);
      
      if (i <= window.row_end - tsize + 1)
      {
#ifdef DENOISE_MT
        denoise_row(m_indata, m_est, i, width, window.col_begin, window.col_end, t, h, thres_mask[i]);
#else
        denoise_row(m_data, m_est, i, width, window.col_begin, window.col_end, t, h, thres_mask[i]);
#endif
      }
      
#ifdef DENOISE_MT
      normalize_row(i, window.col_begin, window.col_end, m_indata, width, m_outdata);
#else
      normalize_row(i, window.col_begin, window.col_end, m_data, width, m_data);
#endif
    }
  }
  else
  {
    for (i = 1; i < tsize; i++)
    {
      memset(m_data_accum[i] + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_data_accum[0][0]));
      memset(m_norm_accum[i] + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_norm_accum[0][0]));
    }
    memset(m_norm_accum2 + window.col_begin, 0, (window.col_end - window.col_begin + 1) * sizeof(m_norm_accum2[0]));
    
#ifdef DENOISE_MT
    if (!write_data)
    {
      for (i = window.row_begin-(tsize-1); i < window.row_begin; i++)
      {
        rotate(m_data_accum, tsize);
        rotate(m_norm_accum, tsize);
        
        if (i <= window.row_end - tsize + 1)
        {
          denoise_row(m_indata, NULL, i, width, window.col_begin, window.col_end, t, h, thres_mask[i]);
        }
      }
    }
#endif
    for (i = window.row_begin; i <= window.row_end; i++)
    {
      rotate(m_data_accum, tsize);
      rotate(m_norm_accum, tsize);

#ifdef DENOISE_MT
      if (i < (is_luma ? m_frame_height : m_frame_height/2)-tsize+1)
#else
      if (i <= window.row_end - tsize + 1)
#endif
      {
#ifdef DENOISE_MT
        denoise_row(m_indata, NULL, i, width, window.col_begin, window.col_end, t, h, thres_mask[i]);
#else
        denoise_row(m_data, NULL, i, width, window.col_begin, window.col_end, t, h, thres_mask[i]);
#endif
      }
      
#ifdef DENOISE_MT
      normalize_row(i, window.col_begin, window.col_end, m_indata, width, m_outdata);
#else
      normalize_row(i, window.col_begin, window.col_end, m_data, width, m_data);
#endif
    }
  }

#ifdef DENOISE_MT
  // Results are transferred when all threads are finished.
#else
  for (i = window.row_begin; i <= window.row_end; i++)
  {
    for (j = window.col_begin; j <= window.col_end; j++)
    {
      img[width * i + j] = m_data[width * i + j] + NDF_SAMPLE_OFFSET;
    }
  }
#endif
}


#ifdef DENOISE_MT
struct DenoiseThreadArgs
{
  pthread_t tid;
  int tno;
  LoopFilterOvercomplete3 *parent;
  Frame *cur;
  int luma_threshold;
  int chroma_threshold;
  int crcb;
};
static void *launch_filter_luma_plane(void *arg)
{
  struct DenoiseThreadArgs *targ = (struct DenoiseThreadArgs *)arg;
  pthread_setspecific(denoise_thread_number_key, (void *)targ->tno);
  targ->parent->filter_plane(targ->cur->luma,
                             targ->cur->height,
                             targ->cur->width,
                             targ->luma_threshold,
                             targ->parent->lumaH,
                             targ->parent->lumaV,
                             targ->cur->qp,
                             true);
  return NULL;
}

static void *launch_filter_chroma_plane(void *arg)
{
  struct DenoiseThreadArgs *targ = (struct DenoiseThreadArgs *)arg;
  pthread_setspecific(denoise_thread_number_key, (void *)targ->tno);
  targ->parent->filter_plane(targ->cur->crcb[targ->crcb],
                             targ->cur->height2,
                             targ->cur->width2,
                             targ->chroma_threshold,
                             targ->parent->chromaH[targ->crcb],
                             targ->parent->chromaV[targ->crcb],
                             targ->cur->qpC,
                             false);
  return NULL;
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Filter frame
*
* This method assumes that boundary modes have been set for luma and chroma
* by invoking set_luma and set_chroma.
*
* \param[in,out] cur Current frame to be filtered
*/
//-----------------------------------------------------------------------------
void LoopFilterOvercomplete3::filter(Frame *cur) 
{
  assert(cur != NULL);
  assert(cur->qp  >= 0 && cur->qp  < POW_TABLE_LIMIT);
  assert(cur->qpC >= 0 && cur->qpC < POW_TABLE_LIMIT);

  int i;
    
#ifdef DENOISE_MT
  // split by row.
  int lumas_per_thread = (cur->height+param.decode_ndenoisethreads-1)/param.decode_ndenoisethreads;
  // Round up to xform size.
  lumas_per_thread += trans_luma_thrd[0]->getSize()-1;
  lumas_per_thread /= trans_luma_thrd[0]->getSize();
  lumas_per_thread *= trans_luma_thrd[0]->getSize();

  for (int t = 0; t < param.decode_ndenoisethreads; t++)
  {
    m_window_luma_thrd[t].row_begin = lumas_per_thread*t;
    m_window_luma_thrd[t].row_end   = m_window_luma_thrd[t].row_begin+lumas_per_thread-1;
    //if (t == 0)
    //  m_window_luma_thrd[t].row_end += 10*8; // for laffs.
    if (m_window_luma_thrd[t].row_end > cur->height-1)
      m_window_luma_thrd[t].row_end = cur->height-1;

    m_window_luma_thrd[t].col_begin = 0;
    m_window_luma_thrd[t].col_end   = cur->width - 1;

    m_window_luma_thrd[t].row_offset = 0;
    m_window_luma_thrd[t].col_offset = 0;

    m_window_chroma_thrd[t].row_begin  = m_window_luma_thrd[t].row_begin  >> 1;
    m_window_chroma_thrd[t].row_end    = m_window_luma_thrd[t].row_end    >> 1;
    m_window_chroma_thrd[t].col_begin  = m_window_luma_thrd[t].col_begin  >> 1;
    m_window_chroma_thrd[t].col_end    = m_window_luma_thrd[t].col_end    >> 1;
    m_window_chroma_thrd[t].row_offset = m_window_luma_thrd[t].row_offset >> 1;
    m_window_chroma_thrd[t].col_offset = m_window_luma_thrd[t].col_offset >> 1;
  }
#else
  m_window_luma.row_begin = 0;
  m_window_luma.row_end   = cur->height - 1;
  m_window_luma.col_begin = 0;
  m_window_luma.col_end   = cur->width - 1;

  m_window_luma.row_offset = 0;
  m_window_luma.col_offset = 0;
  
  m_window_chroma.row_begin  = m_window_luma.row_begin  >> 1;
  m_window_chroma.row_end    = m_window_luma.row_end    >> 1;
  m_window_chroma.col_begin  = m_window_luma.col_begin  >> 1;
  m_window_chroma.col_end    = m_window_luma.col_end    >> 1;
  m_window_chroma.row_offset = m_window_luma.row_offset >> 1;
  m_window_chroma.col_offset = m_window_luma.col_offset >> 1;
#endif

  for (i = 0; i < NUM_LOOP_MODES3; i++) 
  {
    lum_threshold_table[i][0] = lum_threshold_table[i][1] = INVALID_THRES;
    chr_threshold_table[i][0] = chr_threshold_table[i][1] = INVALID_THRES;
    
    lum_extent_table[i] = INVALID_EXTENT;
    chr_extent_table[i] = INVALID_EXTENT;
  }
  
  set_thresholds();
  set_extents();
  set_param_from_config_file();
    
  int threshold;

  threshold = pow_table[cur->qp];
  
#ifdef DENOISE_MT
  struct DenoiseThreadArgs *targs = new DenoiseThreadArgs[param.decode_ndenoisethreads];
  int chroma_threshold = shift_right_round(strength_chroma * pow_table[cur->qpC], 3);

  {
    FilterWindow oldw0 = m_window_luma_thrd[0];
    m_window_luma_thrd[0].row_begin = 0;
    m_window_luma_thrd[0].row_end = cur->height-1;
    m_window_luma_thrd[0].col_begin = 0;
    m_window_luma_thrd[0].col_end = cur->width-1;
    pthread_setspecific(denoise_thread_number_key, (void *)0);
    init_filter_plane(cur->luma, cur->height, cur->width, threshold, lumaH, lumaV, cur->qp, true);
    m_window_luma_thrd[0] = oldw0;
  }
  for (int t = 0; t < param.decode_ndenoisethreads; t++)
  {
    if (m_window_luma_thrd[t].row_begin > m_window_luma_thrd[t].row_end)
      break;
    targs[t].tno = t;
    targs[t].parent = this;
    targs[t].cur = cur;
    targs[t].luma_threshold = threshold;
    targs[t].chroma_threshold = chroma_threshold;
  }

  if (param.decode_ndenoisethreads == 1)
  {
    // direct call.
    launch_filter_luma_plane(&targs[0]);
  }
  else
  {
    for (int t = 0; t < param.decode_ndenoisethreads; t++)
    {
      if (m_window_luma_thrd[t].row_begin > m_window_luma_thrd[t].row_end)
        break;
      pthread_create(&targs[t].tid, NULL, launch_filter_luma_plane, &targs[t]);
    }
    for (int t = 0; t < param.decode_ndenoisethreads; t++)
    {
      if (m_window_luma_thrd[t].row_begin > m_window_luma_thrd[t].row_end)
        break;
      pthread_join(targs[t].tid, NULL);
    }
  }
  // Transfer results.
  for (int i = 0; i < cur->height; i++)
  {
    for (int j = 0; j < cur->width; j++)
    {
      cur->luma[cur->width * i + j] = m_outdata[cur->width * i + j] + NDF_SAMPLE_OFFSET;
    }
  }

  for (int crcb = 0; crcb < 2; crcb++)
  {
    {
      FilterWindow oldw0 = m_window_chroma_thrd[0];
      m_window_chroma_thrd[0].row_begin = 0;
      m_window_chroma_thrd[0].row_end = cur->height2-1;
      m_window_chroma_thrd[0].col_begin = 0;
      m_window_chroma_thrd[0].col_end = cur->width2-1;
      pthread_setspecific(denoise_thread_number_key, (void *)0);
      init_filter_plane(cur->crcb[crcb], cur->height2, cur->width2, chroma_threshold, chromaH[crcb], chromaV[crcb], cur->qpC, false);
      m_window_chroma_thrd[0] = oldw0;
    }
    for (int t = 0; t < param.decode_ndenoisethreads; t++)
    {
      if (m_window_luma_thrd[t].row_begin > m_window_luma_thrd[t].row_end)
        break;
      targs[t].crcb = crcb;
    }
    if (param.decode_ndenoisethreads == 1)
    {
      // direct call.
      launch_filter_chroma_plane(&targs[0]);
    }
    else
    {
      for (int t = 0; t < param.decode_ndenoisethreads; t++)
      {
        if (m_window_luma_thrd[t].row_begin > m_window_luma_thrd[t].row_end)
          break;
        pthread_create(&targs[t].tid, NULL, launch_filter_chroma_plane, &targs[t]);
      }
      for (int t = 0; t < param.decode_ndenoisethreads; t++)
      {
        if (m_window_luma_thrd[t].row_begin > m_window_luma_thrd[t].row_end)
          break;
        pthread_join(targs[t].tid, NULL);
      }
    }
    // Transfer results.
    for (int i = 0; i < cur->height2; i++)
    {
      for (int j = 0; j < cur->width2; j++)
      {
        cur->crcb[crcb][cur->width2 * i + j] = m_outdata[cur->width2 * i + j] + NDF_SAMPLE_OFFSET;
      }
    }
  }
  delete [] targs;

#else
  filter_plane(cur->luma, cur->height, cur->width, threshold, lumaH, lumaV, cur->qp, true);
  
  threshold = shift_right_round(strength_chroma * pow_table[cur->qpC], 3);
  
  for (i = 0; i < 2; i++)
  {
    filter_plane(cur->crcb[i], cur->height2, cur->width2, threshold, chromaH[i], chromaV[i], cur->qpC, false);
  }
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Determine filtering mode
*
* Determines which filter mode should be used for a boundary between two blocks.
*
* \param[in] mvA Motion data of first block
* \param[in] mvB Motion data of second block
* \param[in] ncoeffA Number of coded coefficients in first block
* \param[in] ncoeffB Number of coded coefficients in second block
* \param[in] mth Threshold for motion vector differences
* \param[in] i Index affecting mode decision based on odd/even
* \return Filtering mode
*/
//-----------------------------------------------------------------------------
int LoopFilterOvercomplete3::mode_decision_4x4(PredBlk *mvA, PredBlk *mvB, int ncoeffA, int ncoeffB, int mth, int i, int mvshare_type, int chroma) 
{
  int mode = DCM_SKIP;
  int bs_inside=0;
  int ic_scale_diff,ic_dc_diff,mot_diff;    // to help refine for IC modes
  
  ic_dc_diff=!(mvA->equal_ic_DC(mvB));    // IC offset difference across the block boundary
  ic_scale_diff=!(mvA->equal_ic_scale(mvB));  // IC scale difference across the block boundary
  mot_diff=!(mvA->equal_mot(mvB));      // Motion difference across the block boundary
  
  if (mvA->intra != NULL || mvB->intra != NULL)
  {
    mode = DCM_INTRA_QUANT; // due to intra
  }
  else if (ncoeffA + ncoeffB != ZERO) 
  {
    if (ncoeffA + ncoeffB >= m_signif_cnt)
    {
      mode = DCM_PRED_SIGNIF;
    }
    else if (mvA->soft_equal(mvB, mth) == FALSE)
    {
      mode = DCM_PRED_MOT;
    }
    else
    {
      mode = DCM_PRED_QUANT; // due to pred. quantization
    }
  }
  else if((param.ndf_ic_refine==TRUE)&&(ic_dc_diff&&ic_scale_diff))
  {
    if(mot_diff)
      mode = DCM_IC_STRONG;  
    else
      mode = DCM_IC_INTERMED; 
  }
  else if ((param.ndf_ic_refine==TRUE)&&((ic_dc_diff||ic_scale_diff)&&mot_diff)) 
  {
    
    mode = DCM_IC_WEAK;
  }
  else if (mvA->soft_equal(mvB, mth) == FALSE)
    mode = DCM_MOT_DISC;
  else
    mode = DCM_SKIP;
  
  if(mvshare_type==2 && chroma==0 && (i&1)==0)
  {
    bs_inside = 2;
  }
  else if(mvshare_type>2 && chroma==1)
  {
    bs_inside = 2;
  }
  else if(mvshare_type>0 && chroma==1)
  {
    bs_inside = 1;
  }

  if (bs_inside == 1)
    mode |= DCM_BS_QUARTER;
  else if (bs_inside == 2)
    mode |= DCM_BS_HALF;

  return mode;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
