/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file LumaFrames8pel.cpp
*
* \brief Implementation of LumaFrames8pel class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "LumaFrames8pel.H"
#include "Global.H"
#include "Param.H"
#include "AIF_adaptive_filter.H"

//-----------------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------------

#define COEFF0 5
#define COEFF1 20
#define SHIFT 5
#define COEFF0_8TAP 3
#define COEFF1_8TAP 6
#define COEFF2_8TAP 20

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Allocates memory.
*
* \param[in] h Height of reference frame (luma samples)
* \param[in] w Width of reference frame (luma samples)
* \param[in] b Size of added border (luma samples)
* \param[in] decoder_mode 0: Encoder mode, 1: Decoder mode
*/
//-----------------------------------------------------------------------------
LumaFrames8pel::LumaFrames8pel(int h, int w, int b, int decoder_mode) 
{
  int i,j, step;
  
  if (param.B_subpel > 1 || param.P_subpel> 1) 
  { // allocate the memory only if 1/8 pel is used.
    step = 1;
  }
  else 
  {
    step = 2;
  }
  
  this->border = b;
  width = w + TWO * b;
  height = h + TWO * b;
  
  this->decoder_mode = decoder_mode;
  if (decoder_mode)
  {
    data[0][0] = new unsigned char[width*height];
  }
  else
  {
    for (i=ZERO; i<FRACTIONAL_PEL; i+=step)
    {
      for (j=ZERO; j<FRACTIONAL_PEL; j+=step)
      {
        data[i][j] = new unsigned char[width*height];
      }
    }
  }  
}
  
//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory.
*
*/
//-----------------------------------------------------------------------------
LumaFrames8pel::~LumaFrames8pel() 
{
  int i,j, step;
  
  if (param.B_subpel > 1 || param.P_subpel> 1) 
  { // allocate the memory only if 1/8 pel is used.
    step = 1;
  }
  else 
  {
    step = 2;
  }
  
  if (decoder_mode)
    delete[] data[0][0];
  else
  {
    for (i=ZERO; i<FRACTIONAL_PEL; i+=step)
    {
      for (j=ZERO; j<FRACTIONAL_PEL; j+=step)
      {
        delete[] data[i][j];
      }
    }
  }  
}

//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Generate quarter-pel interpolation
*
* Generates quarter-pel positions and stores them in data (after rounding at 
* the normal precision) and in intermediate (at a higher precision)
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::generate_qp()
{
  int i, j;
  int a=0, b=0, v;
  
  avg(intermediate[FPEL][FPEL], intermediate[FPEL][HPEL], intermediate[FPEL][QPEL], data[FPEL][QPEL]);
  avg(intermediate[FPEL][FPEL], intermediate[HPEL][FPEL], intermediate[QPEL][FPEL], data[QPEL][FPEL]);
  avg(intermediate[FPEL][HPEL], intermediate[HPEL][FPEL], intermediate[QPEL][QPEL], data[QPEL][QPEL]);
  avg(intermediate[FPEL][HPEL], intermediate[HPEL][HPEL], intermediate[QPEL][HPEL], data[QPEL][HPEL]);
  avg(intermediate[HPEL][FPEL], intermediate[HPEL][HPEL], intermediate[HPEL][QPEL], data[HPEL][QPEL]);
  
  
  avg_r(intermediate[FPEL][HPEL], intermediate[FPEL][FPEL], intermediate[FPEL][TPEL], data[FPEL][TPEL]);
  avg_r(intermediate[FPEL][HPEL], intermediate[HPEL][FPEL], intermediate[QPEL][TPEL], data[QPEL][TPEL]);
  avg_r(intermediate[HPEL][HPEL], intermediate[HPEL][FPEL], intermediate[HPEL][TPEL], data[HPEL][TPEL]);
  
  avg_d(intermediate[HPEL][FPEL], intermediate[FPEL][FPEL], intermediate[TPEL][FPEL], data[TPEL][FPEL]);
  avg_d(intermediate[HPEL][FPEL], intermediate[FPEL][HPEL], intermediate[TPEL][QPEL], data[TPEL][QPEL]);
  avg_d(intermediate[HPEL][HPEL], intermediate[FPEL][HPEL], intermediate[TPEL][HPEL], data[TPEL][HPEL]);
  
  for (i=ZERO; i<height-ONE; i++) 
  {
    for (j=ZERO; j<width-ONE; j++) 
    {
      a = intermediate[HPEL][FPEL][width*i+j+ONE];
      b = intermediate[FPEL][HPEL][width*i+j+width];
      intermediate[TPEL][TPEL][width*i+j] = shift_right_round(a + b, ONE);
      v = shift_right_round(a + b, (SHIFT + ONE));
      data[TPEL][TPEL][width*i+j] = clip(v, ZERO, MAX_SAMPLE_VAL);
    }
    
    intermediate[TPEL][TPEL][width*i+j] = shift_right_round(a + b, ONE);
    v = shift_right_round(a + b, (SHIFT + ONE));
    data[TPEL][TPEL][width*i+j] = clip(v, ZERO, MAX_SAMPLE_VAL);
  }
  for (j=ZERO; j<width; j++) 
  {
    intermediate[TPEL][TPEL][width*i+j] = intermediate[TPEL][TPEL][width*i+j-width];
    data[TPEL][TPEL][width*i+j] = data[TPEL][TPEL][width*i+j-width];
  }
  
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate eighth-pel interpolation (bilinear)
*
* Generates eighth-pel positions using bilinear interpolation from quarter-pel 
* positions and stores them in data (after rounding at the normal precision) 
* and in intermediate (at a higher precision)
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::generate_8p_bilinear()
{
  int i, j;
  int a, b, k, l;
  // for horizontal row 0/8,2/8,4/8,6/8 
  for (i=0; i<7; i=i+2)
  {  
    for (j=1; j<7; j=j+2)
      avg(intermediate[i][j-1], intermediate[i][j+1], intermediate[i][j], data[i][j]);
    avg_r(intermediate[i][6], intermediate[i][0], intermediate[i][7], data[i][7]);  
  }
  
  // for vertical column 0/8,2/8,4/8,6/8 
  for (j=0; j<7; j=j+2)
  {  
    for (i=1; i<7; i=i+2)  
      avg(intermediate[i-1][j], intermediate[i+1][j], intermediate[i][j], data[i][j]);
    avg_d(intermediate[6][j], intermediate[0][j], intermediate[7][j], data[7][j]);
  }
  
  // for vertical column 1/8,3/8,5/8,7/8
  for (j=1; j<=7; j=j+2)
  {  
    for (i=1; i<7; i=i+2)
    {  
      if (j==7)
        avg_r(intermediate[i+1][j-1], intermediate[i-1][0], intermediate[i][j], data[i][j]);
      else
        avg(intermediate[i-1][j+1], intermediate[i+1][j-1], intermediate[i][j], data[i][j]);
    }
    if (j==7)
    { //for (7/8,7/8)
      for (k=ZERO; k<height;k++)
      {
        for (l=ZERO; l<width; l++)
        {
          a = intermediate[TPEL][FPEL][width*k+imin(l+1,width-1)];
          b = intermediate[FPEL][TPEL][width*imin(k+1,height-1)+l];

          int val = a + b;
          if (param.bugfix87)
            val = clip(shift_right_round(val, 6), 0, MAX_SAMPLE_VAL);
          else
            val = shift_right_round(val, 1);
          data[SEPEL][SEPEL][width*k+l] = val;
        }
      }
    }
    else
      avg_d(intermediate[TPEL][j+1], intermediate[FPEL][j-1], intermediate[SEPEL][j], data[SEPEL][j]);
  }
}

/*! 
* \brief Prints the interpolated values
* 
* Prints the interpolated values specified by the input arguments
*
* \param[in] si Starting position (same for both directions)
* \param[in] ei Ending position (same for both directions)
*/
void LumaFrames8pel::print(int si, int ei)
{
  
  //si: start index for printing
  //ei: end index for printing
  int i,j,k,l;
  
  printf("\nFull pel ..........\n");
  for (i=si;i<ei;i++)
    for (j=si;j<ei;j++)
    {
      printf("%3d ",data[FPEL][FPEL][width*i+j]);
      if (j==ei-1) printf("\n");
    }
  
  printf("\nHalf pel ..................\n");
  for (i=si;i<ei;i++)
    for (k=0;k<2;k++) //number of subpel
      for (j=si;j<ei;j++)
      {
        if (k%2==0)
        {
          printf("%3d ",data[FPEL][FPEL][width*i+j]);
          printf("%3d ",data[FPEL][HPEL][width*i+j]);
        } 
        else 
        {
          printf("%3d ",data[HPEL][FPEL][width*i+j]);
          printf("%3d ",data[HPEL][HPEL][width*i+j]);
        }
        if (j==ei-1) printf("\n");
      }
  
  printf("\nQuarter pel ..................\n");
  for (i=si;i<ei;i++)
    for (k=0;k<4;k++)
      for (j=si;j<ei;j++)
      {
        if (k==0)
        {
          printf("%3d ",data[FPEL][FPEL][width*i+j]);
          printf("%3d ",data[FPEL][QPEL][width*i+j]);
          printf("%3d ",data[FPEL][HPEL][width*i+j]);
          printf("%3d ",data[FPEL][TPEL][width*i+j]);
        }
        else if (k==1)
        {
          printf("%3d ",data[QPEL][FPEL][width*i+j]);
          printf("%3d ",data[QPEL][QPEL][width*i+j]);
          printf("%3d ",data[QPEL][HPEL][width*i+j]);
          printf("%3d ",data[QPEL][TPEL][width*i+j]);
        }
        else if (k==2)
        {
          printf("%3d ",data[HPEL][FPEL][width*i+j]);
          printf("%3d ",data[HPEL][QPEL][width*i+j]);
          printf("%3d ",data[HPEL][HPEL][width*i+j]);
          printf("%3d ",data[HPEL][TPEL][width*i+j]);
        }
        else if (k==3)
        {
          printf("%3d ",data[TPEL][FPEL][width*i+j]);
          printf("%3d ",data[TPEL][QPEL][width*i+j]);
          printf("%3d ",data[TPEL][HPEL][width*i+j]);
          printf("%3d ",data[TPEL][TPEL][width*i+j]);
        }
        if (j==ei-1) printf("\n");
      }
  
  printf("\nOne Eighth pel ..................\n");
  for (i=si;i<ei;i++)
    for (k=0;k<8;k++)
      for (j=si;j<ei;j++)
      {
        for (l=0;l<8;l++)
          printf("%3d ",data[k][l][width*i+j]);
        if (j==ei-1) printf("\n");
      }
  
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate eighth-pel interpolation for AIF
*
* Generates eighth-pel positions for AIF using from quarter-pel positions and 
* stores them in data (after rounding at the normal precision) and in 
* intermediate (at a higher precision). The interpolation filter is determined
* using Param::gen8pel_filter
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::AIF_generate_8p() 
{ 
  
  switch (param.gen8pel_filter) 
  {
    case BILINEAR_FILTER:
      generate_8p_bilinear();
      break;
    case SIX_TAP_FILTER:
      generate_8p_multi_tap(SIX_TAP_FILTER);
      break;
    case EIGHT_TAP_FILTER:
      generate_8p_multi_tap(EIGHT_TAP_FILTER);
      break;
    default:
      printf("Unknown filter %d for 1/8 pel generation.",param.gen8pel_filter); 
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate quarter-pel interpolation for AIF
*
* Generates quarter-pel positions and stores them in data (after rounding at 
* the normal precision) and in intermediate (at a higher precision)
* 
* \param[in] cur Frame for determining AIF type
* \param[in] list List index (LIST_0 or LIST_1)
* \param[in] ref Reference frame index
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::AIF_generate_qp(Frame * /*cur*/, int /*list*/, int /*ref*/) 
{ 
  int m,n,k,i,j,ii,jj;

  int isum, jwidth;
  int top_ref, bottom_ref, top, bottom;
  int left, left_data, right, right_data, left_intermediate, right_intermediate; 
  int iAIF_Filter[4][4][36];
  int border = BORDER;   
  int f_bits_minus_1 = F_BITS - 1;
  int shift_f_bits_minus_1 = (1 << (f_bits_minus_1));
  int f_bits_minus_shift_minus_1 = F_BITS - SHIFT - 1;
  int shift_f_bits_minus_shift_minus_1 = (1 << (f_bits_minus_shift_minus_1));
  int f_bits_plus_shift_minus_1 = F_BITS + SHIFT - 1;
  int shift_f_bits_plus_shift_minus_1 = (1 << (f_bits_plus_shift_minus_1));
  int f_bits_plus_shift = F_BITS + SHIFT;
  int f_bits_minus_shift = F_BITS - SHIFT;
  int shift_minus_1 = SHIFT - 1; 
  int shift_minus_2 = SHIFT - 2; 

  for(k=0;k<=6;k++)
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<4;i++)
      {
        if(j||i)         
          iAIF_Filter[j][i][k]=iFilterCoef_SAIF[j*4+i-1][k];   
      }
    }
  }

  ////////////   full-pel   ////////////
  n = m = 0;
  for(j=0;j<height;j++)
  {
    jwidth = j*width;
    for(i=0;i<width;i++)
    {
      intermediate[0][0][jwidth+i]=data[0][0][jwidth+i] << SHIFT;   
    }
  }

  ////////////   b_pos   ////////////
  for(j=border;j<height-border;j++)
  {
    jwidth = j*width;
    for(i=border;i<width-border;i++)
    {
      isum = 0;
      for(ii=0;ii<6;ii++)
      {
        isum+=data[FPEL][FPEL][jwidth+i+ii-2]*iAIF_Filter[0][2][ii];
      }
      isum += iAIF_Filter[0][2][6]; 
      data[0][4][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
      intermediate[0][4][jwidth+i]=(isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 
    }
  }
   // top and bottom border padding
  top_ref = border*width;
  bottom_ref = (height-border-1)*width;
  for(j=0;j<border;j++)
  {
    top = j*width;
    bottom = (height-j-1)*width;
    for(i=border;i<width-border;i++)
    {
      data[0][4][top+i]=data[0][4][top_ref+i];
      data[0][4][bottom+i]=data[0][4][bottom_ref+i];
      intermediate[0][4][top+i]=intermediate[0][4][top_ref+i];
      intermediate[0][4][bottom+i]=intermediate[0][4][bottom_ref+i];
    }
  }
  
  // left and right border padding
  for(j=0;j<height;j++)
  {
    left = j*width;
    left_data = data[0][4][j*width+border];
    right = (j+1)*width-border;
    right_data = data[0][4][(j+1)*width-border-1];
    left_intermediate = intermediate[0][4][j*width+border];
    right_intermediate = intermediate[0][4][(j+1)*width-border-1];
  
    for(i=0;i<border;i++)
    {
      data[0][4][left+i]=left_data;
      data[0][4][right+i]=right_data;
      intermediate[0][4][left+i]=left_intermediate;
      intermediate[0][4][right+i]=right_intermediate;
    }
  }  
  ////////////   h_pos   ////////////
  for(j=border;j<height-border;j++)
  {
    jwidth = j*width;
    for(i=border;i<width-border;i++)
    {
      isum = 0;
      for(jj=0;jj<6;jj++)
      {
        isum+=data[FPEL][FPEL][jwidth+i+((jj-2)*width)]*iAIF_Filter[2][0][jj];
      }
      isum += iAIF_Filter[2][0][6]; 
      data[4][0][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
      intermediate[4][0][jwidth+i]=(isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 
    }
  }
   // top and bottom border padding
  top_ref = border*width;
  bottom_ref = (height-border-1)*width;
  for(j=0;j<border;j++)
  {
    top = j*width;
    bottom = (height-j-1)*width;
    for(i=border;i<width-border;i++)
    {
      data[4][0][top+i]=data[4][0][top_ref+i];
      data[4][0][bottom+i]=data[4][0][bottom_ref+i];
      intermediate[4][0][top+i]=intermediate[4][0][top_ref+i];
      intermediate[4][0][bottom+i]=intermediate[4][0][bottom_ref+i];
    }
  }
  
  // left and right border padding
  for(j=0;j<height;j++)
  {
    left = j*width;
    left_data = data[4][0][j*width+border];
    right = (j+1)*width-border;
    right_data = data[4][0][(j+1)*width-border-1];
    left_intermediate = intermediate[4][0][j*width+border];
    right_intermediate = intermediate[4][0][(j+1)*width-border-1];
  
    for(i=0;i<border;i++)
    {
      data[4][0][left+i]=left_data;
      data[4][0][right+i]=right_data;
      intermediate[4][0][left+i]=left_intermediate;
      intermediate[4][0][right+i]=right_intermediate;
    }
  }  

  ////////////   j_pos   ////////////
      
  for(j=border;j<height-border;j++)
  {
    jwidth = j*width;
    for(i=border;i<width-border;i++)
    {
      isum = 0;
      for(jj=0;jj<6;jj++)
      {
        isum+=intermediate[0][4][jwidth+i+((jj-2)*width)]*iAIF_Filter[2][2][jj];
      }
      isum += iAIF_Filter[2][2][6]; 
      data[4][4][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
      intermediate[4][4][jwidth+i]= (isum + shift_f_bits_minus_1) >> F_BITS; 
    }
  }
   // top and bottom border padding
  top_ref = border*width;
  bottom_ref = (height-border-1)*width;
  for(j=0;j<border;j++)
  {
    top = j*width;
    bottom = (height-j-1)*width;
    for(i=border;i<width-border;i++)
    {
      data[4][4][top+i]=data[4][4][top_ref+i];
      data[4][4][bottom+i]=data[4][4][bottom_ref+i];
      intermediate[4][4][top+i]=intermediate[4][4][top_ref+i];
      intermediate[4][4][bottom+i]=intermediate[4][4][bottom_ref+i];
    }
  }
  
  // left and right border padding
  for(j=0;j<height;j++)
  {
    left = j*width;
    left_data = data[4][4][j*width+border];
    right = (j+1)*width-border;
    right_data = data[4][4][(j+1)*width-border-1];
    left_intermediate = intermediate[4][4][j*width+border];
    right_intermediate = intermediate[4][4][(j+1)*width-border-1];
  
    for(i=0;i<border;i++)
    {
      data[4][4][left+i]=left_data;
      data[4][4][right+i]=right_data;
      intermediate[4][4][left+i]=left_intermediate;
      intermediate[4][4][right+i]=right_intermediate;
    }
  }  

  if (!AIF_mode_flag1) // AIF_MODE1
  {    
    for(j=border;j<height-border;j++)
    {
      jwidth = j*width;
      for(i=border;i<width-border;i++)
      {
        ////////////   a_pos   ////////////
        data[0][2][jwidth+i] = (data[0][0][jwidth+i] + data[0][4][jwidth+i] + 1) >> 1;
        intermediate[0][2][jwidth+i]=(data[0][0][jwidth+i] + data[0][4][jwidth+i]) << shift_minus_1; 
        ////////////   c_pos   ////////////        
        data[0][6][jwidth+i] = (data[0][0][jwidth+i+1] + data[0][4][jwidth+i] + 1) >> 1;
        intermediate[0][6][jwidth+i]=(data[0][0][jwidth+i+1] + data[0][4][jwidth+i]) << shift_minus_1; 
        ////////////   d_pos   ////////////
        data[2][0][jwidth+i] = (data[0][0][jwidth+i] + data[4][0][jwidth+i] + 1) >> 1;
        intermediate[2][0][jwidth+i]=(data[0][0][jwidth+i] + data[4][0][jwidth+i]) << shift_minus_1; 
        ////////////   l_pos   ////////////
        data[6][0][jwidth+i] = (data[0][0][jwidth+i+width] + data[4][0][jwidth+i] + 1) >> 1;
        intermediate[6][0][jwidth+i]= (data[0][0][jwidth+i+width] + data[4][0][jwidth+i]) << shift_minus_1; 
        ////////////   f_pos   ////////////
        data[2][4][jwidth+i] = (data[0][4][jwidth+i] + data[4][4][jwidth+i] + 1) >> 1;
        intermediate[2][4][jwidth+i]= (data[0][4][jwidth+i] + data[4][4][jwidth+i]) << shift_minus_1; 
        ////////////   n_pos   ////////////
        data[6][4][jwidth+i] = (data[0][4][jwidth+i+width] + data[4][4][jwidth+i] + 1) >> 1;
        intermediate[6][4][jwidth+i] = (data[0][4][jwidth+i+width] + data[4][4][jwidth+i]) << shift_minus_1; 
        ////////////   i_pos   ////////////
        data[4][2][jwidth+i] = (data[4][0][jwidth+i] + data[4][4][jwidth+i] + 1) >> 1;
        intermediate[4][2][jwidth+i] = (data[4][0][jwidth+i] + data[4][4][jwidth+i]) << shift_minus_1; 
        ////////////   e_pos   ////////////
        data[2][2][jwidth+i] = (data[0][0][jwidth+i] + data[0][4][jwidth+i] + data[4][0][jwidth+i] + data[4][4][jwidth+i] + 2) >> 2;
        intermediate[2][2][jwidth+i] = (data[0][0][jwidth+i] + data[0][4][jwidth+i] + data[4][0][jwidth+i] + data[4][4][jwidth+i]) << shift_minus_2; 
        ////////////   g_pos   ////////////
        data[2][6][jwidth+i] = (data[0][0][jwidth+i+1] + data[0][4][jwidth+i] + data[4][0][jwidth+i+1] + data[4][4][jwidth+i] + 2) >> 2;
        intermediate[2][6][jwidth+i] = (data[0][0][jwidth+i+1] + data[0][4][jwidth+i] + data[4][0][jwidth+i+1] + data[4][4][jwidth+i]) << shift_minus_2; 
        ////////////   k_pos   ////////////
        data[4][6][jwidth+i] = (data[4][4][jwidth+i] + data[4][0][jwidth+i+1] + 1) >> 1;
        intermediate[4][6][jwidth+i] = (data[4][4][jwidth+i] + data[4][0][jwidth+i+1]) << shift_minus_1; 
        ////////////   m_pos   ////////////
        data[6][2][jwidth+i] = (data[0][0][jwidth+i+width] + data[4][0][jwidth+i] + data[0][4][jwidth+i+width] + data[4][4][jwidth+i] + 2) >> 2;
        intermediate[6][2][jwidth+i] = (data[0][0][jwidth+i+width] + data[4][0][jwidth+i] + data[0][4][jwidth+i+width] + data[4][4][jwidth+i]) << shift_minus_2; 
        ////////////   o_pos   ////////////        
        data[6][6][jwidth+i] = (data[0][0][jwidth+i+width+1] + data[4][0][jwidth+i+1] + data[0][4][jwidth+i+width] + data[4][4][jwidth+i] + 2) >> 2;
        intermediate[6][6][jwidth+i] = (data[0][0][jwidth+i+width+1] + data[4][0][jwidth+i+1] + data[0][4][jwidth+i+width] + data[4][4][jwidth+i]) << shift_minus_2; 
      }
    }
  }
  else if (!AIF_mode_flag2) // AIF_MODE2
  {
    for(j=border;j<height-border;j++)
    {
      jwidth = j*width;
      for(i=border;i<width-border;i++)
      {
        ////////////   a_pos   ////////////
        isum = 0;
        for(ii=0;ii<6;ii++)
        {
          isum+=data[FPEL][FPEL][jwidth+i+ii-2]*iAIF_Filter[0][1][ii];
        }
        isum += iAIF_Filter[0][1][6]; 
        data[0][2][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        intermediate[0][2][jwidth+i]=(isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 

        ////////////   c_pos   ////////////        
        isum = 0;
        for(ii=0;ii<6;ii++)
        {
          isum+=data[FPEL][FPEL][jwidth+i+ii-2]*iAIF_Filter[0][3][ii];
        }
        isum += iAIF_Filter[0][3][6]; 
        data[0][6][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        intermediate[0][6][jwidth+i]= (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 

        ////////////   d_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=data[FPEL][FPEL][jwidth+i+((jj-2)*width)]*iAIF_Filter[1][0][jj];
        }
        isum += iAIF_Filter[1][0][6]; 
        data[2][0][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        intermediate[2][0][jwidth+i]= (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 

        ////////////   l_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=data[FPEL][FPEL][jwidth+i+((jj-2)*width)]*iAIF_Filter[3][0][jj];
        }
        isum += iAIF_Filter[3][0][6]; 
        data[6][0][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        intermediate[6][0][jwidth+i]=(isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 

        ////////////   f_pos   ////////////
        data[2][4][jwidth+i] = (data[0][4][jwidth+i] + data[4][4][jwidth+i] + 1) >> 1;
        intermediate[2][4][jwidth+i]= (data[0][4][jwidth+i] + data[4][4][jwidth+i]) << shift_minus_1; 
        ////////////   n_pos   ////////////
        data[6][4][jwidth+i] = (data[0][4][jwidth+i+width] + data[4][4][jwidth+i] + 1) >> 1;
        intermediate[6][4][jwidth+i] = (data[0][4][jwidth+i+width] + data[4][4][jwidth+i]) << shift_minus_1; 
        ////////////   i_pos   ////////////
        data[4][2][jwidth+i] = (data[4][0][jwidth+i] + data[4][4][jwidth+i] + 1) >> 1;
        intermediate[4][2][jwidth+i] = (data[4][0][jwidth+i] + data[4][4][jwidth+i]) << shift_minus_1; 
        ////////////   e_pos   ////////////
        data[2][2][jwidth+i] = (data[0][0][jwidth+i] + data[0][4][jwidth+i] + data[4][0][jwidth+i] + data[4][4][jwidth+i] + 2) >> 2;
        intermediate[2][2][jwidth+i] = (data[0][0][jwidth+i] + data[0][4][jwidth+i] + data[4][0][jwidth+i] + data[4][4][jwidth+i]) << shift_minus_2; 
        ////////////   g_pos   ////////////
        data[2][6][jwidth+i] = (data[0][0][jwidth+i+1] + data[0][4][jwidth+i] + data[4][0][jwidth+i+1] + data[4][4][jwidth+i] + 2) >> 2;
        intermediate[2][6][jwidth+i] = (data[0][0][jwidth+i+1] + data[0][4][jwidth+i] + data[4][0][jwidth+i+1] + data[4][4][jwidth+i]) << shift_minus_2; 
        ////////////   k_pos   ////////////
        data[4][6][jwidth+i] = (data[4][4][jwidth+i] + data[4][0][jwidth+i+1] + 1) >> 1;
        intermediate[4][6][jwidth+i] = (data[4][4][jwidth+i] + data[4][0][jwidth+i+1]) << shift_minus_1; 
        ////////////   m_pos   ////////////
        data[6][2][jwidth+i] = (data[0][0][jwidth+i+width] + data[4][0][jwidth+i] + data[0][4][jwidth+i+width] + data[4][4][jwidth+i] + 2) >> 2;
        intermediate[6][2][jwidth+i] = (data[0][0][jwidth+i+width] + data[4][0][jwidth+i] + data[0][4][jwidth+i+width] + data[4][4][jwidth+i]) << shift_minus_2; 
        ////////////   o_pos   ////////////        
        data[6][6][jwidth+i] = (data[0][0][jwidth+i+width+1] + data[4][0][jwidth+i+1] + data[0][4][jwidth+i+width] + data[4][4][jwidth+i] + 2) >> 2;
        intermediate[6][6][jwidth+i] = (data[0][0][jwidth+i+width+1] + data[4][0][jwidth+i+1] + data[0][4][jwidth+i+width] + data[4][4][jwidth+i]) << shift_minus_2; 
      }
    }
  }
  else // AIF_MODE3
  {
    ////////////   a_pos   ////////////
    for(j=border;j<height-border;j++)
    {
      jwidth = j*width;
      for(i=border;i<width-border;i++)
      {
        isum = 0;
        for(ii=0;ii<6;ii++)
        {
          isum+=data[FPEL][FPEL][jwidth+i+ii-2]*iAIF_Filter[0][1][ii];
        }
         isum += iAIF_Filter[0][1][6]; 
        data[0][2][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        intermediate[0][2][jwidth+i]=(isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 
      }
    }
    // top and bottom border padding
    top_ref = BORDER*width;
    bottom_ref = (height-BORDER-1)*width;
    for(j=0;j<BORDER;j++)
    {
      top = j*width;
      bottom = (height-j-1)*width;
      for(i=BORDER;i<width-BORDER;i++)
      {
        data[0][2][top+i]=data[0][2][top_ref+i];
        data[0][2][bottom+i]=data[0][2][bottom_ref+i];
        intermediate[0][2][top+i]=intermediate[0][2][top_ref+i];
        intermediate[0][2][bottom+i]=intermediate[0][2][bottom_ref+i];
      }
    }
    
    // left and right border padding
    for(j=0;j<height;j++)
    {
      left = j*width;
      left_data = data[0][2][j*width+BORDER];
      right = (j+1)*width-BORDER;
      right_data = data[0][2][(j+1)*width-BORDER-1];
      left_intermediate = intermediate[0][2][j*width+BORDER];
      right_intermediate = intermediate[0][2][(j+1)*width-BORDER-1];
    
      for(i=0;i<BORDER;i++)
      {
        data[0][2][left+i]=left_data;
        data[0][2][right+i]=right_data;
        intermediate[0][2][left+i]=left_intermediate;
        intermediate[0][2][right+i]=right_intermediate;
      }
    }  

    ////////////   c_pos   ////////////
    for(j=border;j<height-border;j++)
    {
      jwidth = j*width;
      for(i=border;i<width-border;i++)
      {
        isum = 0;
        for(ii=0;ii<6;ii++)
        {
          isum+=data[FPEL][FPEL][jwidth+i+ii-2]*iAIF_Filter[0][3][ii];
        }
        isum += iAIF_Filter[0][3][6]; 
        data[0][6][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        intermediate[0][6][jwidth+i]= (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 
      }
    }

    // top and bottom border padding
    top_ref = BORDER*width;
    bottom_ref = (height-BORDER-1)*width;
    for(j=0;j<BORDER;j++)
    {
      top = j*width;
      bottom = (height-j-1)*width;
      for(i=BORDER;i<width-BORDER;i++)
      {
        data[0][6][top+i]=data[0][6][top_ref+i];
        data[0][6][bottom+i]=data[0][6][bottom_ref+i];
        intermediate[0][6][top+i]=intermediate[0][6][top_ref+i];
        intermediate[0][6][bottom+i]=intermediate[0][6][bottom_ref+i];
      }
    }
    
    // left and right border padding
    for(j=0;j<height;j++)
    {
      left = j*width;
      left_data = data[0][6][j*width+BORDER];
      right = (j+1)*width-BORDER;
      right_data = data[0][6][(j+1)*width-BORDER-1];
      left_intermediate = intermediate[0][6][j*width+BORDER];
      right_intermediate = intermediate[0][6][(j+1)*width-BORDER-1];
    
      for(i=0;i<BORDER;i++)
      {
        data[0][6][left+i]=left_data;
        data[0][6][right+i]=right_data;
        intermediate[0][6][left+i]=left_intermediate;
        intermediate[0][6][right+i]=right_intermediate;
      }
    }  
    
    for(j=border;j<height-border;j++)
    {
      jwidth = j*width;
      for(i=border;i<width-border;i++)
      {
        ////////////   d_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=data[FPEL][FPEL][jwidth+i+((jj-2)*width)]*iAIF_Filter[1][0][jj];
        }
        isum += iAIF_Filter[1][0][6]; 
        data[2][0][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        intermediate[2][0][jwidth+i]= (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        ////////////   l_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=data[FPEL][FPEL][jwidth+i+((jj-2)*width)]*iAIF_Filter[3][0][jj];
        }
        isum += iAIF_Filter[3][0][6]; 
        data[6][0][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        intermediate[6][0][jwidth+i]=(isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift; 
        ////////////   f_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=intermediate[0][4][jwidth+i+((jj-2)*width)]*iAIF_Filter[1][2][jj];
        }
        isum += iAIF_Filter[1][2][6]; 
        data[2][4][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
        intermediate[2][4][jwidth+i]= (isum + shift_f_bits_minus_1) >> F_BITS; 
        ////////////   n_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=intermediate[0][4][jwidth+i+((jj-2)*width)]*iAIF_Filter[3][2][jj];
        }
        isum += iAIF_Filter[3][2][6]; 
        data[6][4][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
        intermediate[6][4][jwidth+i] = (isum + shift_f_bits_minus_1) >> F_BITS; 
        ////////////   i_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=intermediate[0][2][jwidth+i+((jj-2)*width)]*iAIF_Filter[2][1][jj];
        }
        isum += iAIF_Filter[2][1][6]; 
        data[4][2][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
        intermediate[4][2][jwidth+i] = (isum + shift_f_bits_minus_1) >> F_BITS; 
        ////////////   e_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=intermediate[0][2][jwidth+i+((jj-2)*width)]*iAIF_Filter[1][1][jj];
        }
        isum += iAIF_Filter[1][1][6]; 
        data[2][2][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
        intermediate[2][2][jwidth+i] = (isum + shift_f_bits_minus_1) >> F_BITS;
        ////////////   g_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=intermediate[0][6][jwidth+i+((jj-2)*width)]*iAIF_Filter[1][3][jj];
        }
        isum += iAIF_Filter[1][3][6]; 
        data[2][6][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
        intermediate[2][6][jwidth+i] = (isum + shift_f_bits_minus_1) >> F_BITS;
        ////////////   k_pos   ////////////    
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=intermediate[0][6][jwidth+i+((jj-2)*width)]*iAIF_Filter[2][3][jj];
        }
        isum += iAIF_Filter[2][3][6]; 
        data[4][6][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
        intermediate[4][6][jwidth+i] = (isum + shift_f_bits_minus_1) >> F_BITS;
        ////////////   m_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=intermediate[0][2][jwidth+i+((jj-2)*width)]*iAIF_Filter[3][1][jj];
        }
        isum += iAIF_Filter[3][1][6]; 
        data[6][2][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
        intermediate[6][2][jwidth+i] = (isum + shift_f_bits_minus_1) >> F_BITS; 
        ////////////   o_pos   ////////////
        isum = 0;
        for(jj=0;jj<6;jj++)
        {
          isum+=intermediate[0][6][jwidth+i+((jj-2)*width)]*iAIF_Filter[3][3][jj];
        }
        isum += iAIF_Filter[3][3][6]; 
        data[6][6][jwidth+i]=imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
        intermediate[6][6][jwidth+i] = (isum + shift_f_bits_minus_1) >> F_BITS; 
      }
    }
  }

  // padding
  for(n=0;n<FRACTIONAL_PEL;n+=2)    // quarter pel
  {
    for(m=0;m<FRACTIONAL_PEL;m+=2)  // quarter pel
    {
      if(!(!n && !m))
      {   
        // top and bottom border padding
        top_ref = border*width;
        bottom_ref = (height-border-1)*width;
        for(j=0;j<border;j++)
        {
          top = j*width;
          bottom = (height-j-1)*width;
          for(i=border;i<width-border;i++)
          {
            data[n][m][top+i]=data[n][m][top_ref+i];
            data[n][m][bottom+i]=data[n][m][bottom_ref+i];
            intermediate[n][m][top+i]=intermediate[n][m][top_ref+i];
            intermediate[n][m][bottom+i]=intermediate[n][m][bottom_ref+i];
          }
        }    
      
        // left and right border padding
        for(j=0;j<height;j++)
        {
          left = j*width;
          left_data = data[n][m][j*width+border];
          right = (j+1)*width-border;
          right_data = data[n][m][(j+1)*width-border-1];
          left_intermediate = intermediate[n][m][j*width+border];
          right_intermediate = intermediate[n][m][(j+1)*width-border-1];        
          for(i=0;i<border;i++)
          {
            data[n][m][left+i]=left_data;
            data[n][m][right+i]=right_data;
            intermediate[n][m][left+i]=left_intermediate;
            intermediate[n][m][right+i]=right_intermediate;
          }
        }        
      }
    }    
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate half-pel interpolation
*
* Generates half-pel positions and stores them in data (after rounding at 
* the normal precision) and in intermediate (at a higher precision)
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::generate_hp()
{
  int i,j;
  int v;
  
  for (i=ZERO; i<height; i++) 
  {
    for (j=ZERO; j<width; j++)
      buf[j] = data[FPEL][FPEL][width*i+j];
    for (j=ZERO; j<width; j++)
      intermediate[FPEL][FPEL][width*i+j] = (buf[j]<<SHIFT);   // Add 5 bits
    
    filter_6tap(buf,width);
    
    for (j=ZERO; j<width; j++) 
    {
      v = buf[j];
      intermediate[FPEL][HPEL][width*i+j] = v;                // 6 tap filter add 5 bits
      v = shift_right_round(v, SHIFT);
      v = clip(v, ZERO, MAX_SAMPLE_VAL);
      data[FPEL][HPEL][width*i+j] = v;
    }
  }
  
  for (j=ZERO; j<width; j++) 
  {
    for (i=ZERO; i<height; i++)
      buf[i] = data[FPEL][FPEL][width*i+j];
    
    filter_6tap(buf,height);
    
    for (i=ZERO; i<height; i++) 
    {
      v = buf[i];
      intermediate[HPEL][FPEL][width*i+j] = v;               // 6 tap filter add 5 bits
      v = shift_right_round(v, SHIFT);
      v = clip(v, ZERO, MAX_SAMPLE_VAL);
      data[HPEL][FPEL][width*i+j] = v;
    }
  }
  
  for (j=ZERO; j<width; j++) 
  {
    for (i=ZERO; i<height; i++)
      buf[i] = intermediate[FPEL][HPEL][width*i+j];
    
    filter_6tap(buf,height);
    
    for (i=ZERO; i<height; i++) 
    {
      v = buf[i];
      intermediate[HPEL][HPEL][width*i+j] = shift_right_round(v, SHIFT);        // 2x6 tap filter add 10 bits
      v = shift_right_round(v, TWO * SHIFT);
      v = clip(v, ZERO, MAX_SAMPLE_VAL);
      data[HPEL][HPEL][width*i+j] = v;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate eighth-pel interpolation using specified filter
*
* Generates eighth-pel positions using specified interpolation from quarter-pel 
* positions and stores them in data (after rounding at the normal precision) 
* and in intermediate (at a higher precision)
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::generate_8p_multi_tap(int filter)
{
  int i,j,k,l;
  int v;
  
  for (i=ZERO; i<height; i++) 
    for (k=ZERO; k<FRACTIONAL_PEL; k=k+2)
    {
      for (j=ZERO; j<width; j++)
        for (l=ZERO; l<FRACTIONAL_PEL; l=l+2)
          buf[j*(FRACTIONAL_PEL>>ONE)+(l>>ONE)] = intermediate[k][l][width*i+j];
      
      if (filter == SIX_TAP_FILTER)
        filter_6tap(buf,width*(FRACTIONAL_PEL>>ONE));
      else if (filter == EIGHT_TAP_FILTER)
        filter_8tap(buf,width*(FRACTIONAL_PEL>>ONE));
      else
        printf("Unknown filter type %d\n",filter);
      
      for (j=ZERO; j<width; j++) 
        for (l=ZERO; l<FRACTIONAL_PEL; l=l+2)
        {
          v = buf[j*(FRACTIONAL_PEL>>ONE)+(l>>ONE)];
          intermediate[k][l+1][width*i+j] = shift_right_round(v, SHIFT);
          v = shift_right_round(v, (2*SHIFT));
          v = clip(v, ZERO, MAX_SAMPLE_VAL);
          data[k][l+1][width*i+j] = v;
        }
    }
  
  for (j=ZERO; j<width; j++)
    for (l=ZERO; l<FRACTIONAL_PEL; l=l+2)
    {
      for (i=ZERO; i<height; i++) 
        for (k=ZERO; k<FRACTIONAL_PEL; k=k+2)
          buf[i*(FRACTIONAL_PEL>>ONE)+(k>>ONE)] = intermediate[k][l][width*i+j];
      
      // 'width' was replaced by 'height' for bug fix
      if (filter == SIX_TAP_FILTER)
        filter_6tap(buf,height*(FRACTIONAL_PEL>>ONE));
      else if (filter == EIGHT_TAP_FILTER)
        filter_8tap(buf,height*(FRACTIONAL_PEL>>ONE));
      else
        printf("Unknown filter type %d\n",filter);
      
      for (i=ZERO; i<height; i++) 
      {
        for (k=ZERO; k<FRACTIONAL_PEL; k=k+2)
        {
          v = buf[i*(FRACTIONAL_PEL>>ONE)+(k>>ONE)];
          v = shift_right_round(v, (2*SHIFT));
          v = clip(v, ZERO, MAX_SAMPLE_VAL);
          data[k+1][l][width*i+j] = v;
        }
      }
    }
  
  for (j=ZERO; j<width; j++)
  {
    for (l=ZERO; l<FRACTIONAL_PEL; l=l+2)
    {
      for (i=ZERO; i<height; i++) 
        for (k=ZERO; k<FRACTIONAL_PEL; k=k+2)
        {
          buf[i*(FRACTIONAL_PEL>>ONE)+(k>>ONE)] = intermediate[k][l+1][width*i+j];;
        }
      
      // 'width' was replaced by 'height' for bug fix
      if (filter == SIX_TAP_FILTER)
        filter_6tap(buf,height*(FRACTIONAL_PEL>>ONE));
      else if (filter == EIGHT_TAP_FILTER)
        filter_8tap(buf,height*(FRACTIONAL_PEL>>ONE));
      else
        printf("Unknown filter type %d\n",filter);
      
      
      for (i=ZERO; i<height; i++) 
        for (k=ZERO; k<FRACTIONAL_PEL; k=k+2)
        {
          v = buf[i*(FRACTIONAL_PEL>>ONE)+(k>>ONE)];
          v = shift_right_round(v, (2*SHIFT));
          v = clip(v, ZERO, MAX_SAMPLE_VAL);
          data[k+1][l+1][width*i+j] = v;
        }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Apply 6-tap filter to a line of samples
*
* \param[in,out] buf Line of samples
* \param[in] n Number of samples in line
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::filter_6tap(int *buf, int n) 
{
  int i;
  int a, b, c, d, e, f, v;
  
  b = buf[ZERO];
  c = buf[ZERO];
  d = buf[ZERO];
  e = buf[ONE];
  f = buf[TWO];
  
  for (i=ZERO; i<n; i++) 
  {
    a = b;
    b = c;
    c = d;
    d = e;
    e = f;
    if (i < n - THREE)
      f = buf[i + THREE];
    
    v = a + f;
    v -= COEFF0 * (b + e);
    v += COEFF1 * (c + d);
    buf[i] = v;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Apply 8-tap filter to a line of samples
*
* \param[in,out] buf Line of samples
* \param[in] n Number of samples in line
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::filter_8tap(int *buf, int n) 
{
  int i;
  int a, b, c, d, e, f, g, h, v;
  
  b = buf[ZERO];
  c = buf[ZERO];
  d = buf[ZERO];
  e = buf[ZERO];
  f = buf[ONE];
  g = buf[TWO];
  h = buf[THREE];
  
  for (i=ZERO; i<n; i++) 
  {
    a = b;
    b = c;
    c = d;
    d = e;
    e = f;
    f = g;
    g = h;
    if (i < n - FOUR)
      h = buf[i + FOUR];
    
    v = -a - h;
    v += COEFF0_8TAP * (b + g);
    v -= COEFF1_8TAP * (c + f);
    v += COEFF2_8TAP * (d + e);
    buf[i] = v;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Average two arrays of samples
*
* \param[in] src0 First array of samples
* \param[in] src1 Second array of samples
* \param[out] dst0 Average array (Higher Precision)
* \param[out] dst1 Average array (Lower Precision and Rounded)
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::avg(int *src0, int *src1, int *dst0, unsigned char *dst1 )
{
  int i, v, u;
  
  for (i = ZERO; i < width * height; i++) 
  {
    u = (*src0++) + (*src1++);
    v = shift_right_round(u, (SHIFT+ONE));
    *dst1++  = clip(v, ZERO, MAX_SAMPLE_VAL);
    if (dst0)
      *dst0++ = shift_right_round(u, ONE);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Average two arrays of samples (vertical shift)
*
* Second array is shifted up before averaging.
*
* \param[in] src0 First array of samples
* \param[in] src1 Second array of samples
* \param[out] dst0 Average array (Higher Precision)
* \param[out] dst1 Average array (Lower Precision and Rounded)
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::avg_d(int *src0, int *src1, int *dst0, unsigned char *dst1)
{
  int i, v, u;
  
  src1 += width;
  
  for (i = width; i < width * height; i++) 
  {
    u = (*src0++) + (*src1++);
    v = shift_right_round(u, (SHIFT+ONE));
    *dst1++  = clip(v, ZERO, MAX_SAMPLE_VAL);
    if (dst0)
      *dst0++ = shift_right_round(u, ONE);
  }
  
  src1 -= width;
  
  for (i = ZERO; i < width; i++) 
  {
    u = (*src0++) + (*src1++);
    v = shift_right_round(u, (SHIFT+ONE));
    *dst1++  = clip(v, ZERO, MAX_SAMPLE_VAL);
    if (dst0)
      *dst0++ = shift_right_round(u, ONE);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Average two arrays of samples (horizontal shift)
*
* Second array is shifted left before averaging.
*
* \param[in] src0 First array of samples
* \param[in] src1 Second array of samples
* \param[out] dst0 Average array (Higher Precision)
* \param[out] dst1 Average array (Lower Precision and Rounded)
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::avg_r(int *src0, int *src1, int *dst0, unsigned char *dst1)
{
  int i, j, v, u;
  
  for (i = ZERO; i < height; i++) 
  {
    src1++;
    
    for (j = ZERO; j < width - ONE; j++) 
    {
      u = (*src0++) + (*src1++);
      v = shift_right_round(u, (SHIFT+ONE));
      *dst1++  = clip(v, ZERO, MAX_SAMPLE_VAL);
      if (dst0)
        *dst0++ = shift_right_round(u, ONE);
    }
    
    u = (*src0++) + (src1[-1]);
    v = shift_right_round(u, (SHIFT+ONE));
    *dst1++  = clip(v, ZERO, MAX_SAMPLE_VAL);
    if (dst0)
      *dst0++ = shift_right_round(u, ONE);  
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Average two arrays of samples
*
* \param[in] src0 First array of samples
* \param[in] src1 Second array of samples
* \param[out] dst Average array
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::avg(unsigned char *src0, unsigned char *src1, unsigned char *dst)
{
  int i;
  
  for (i = ZERO; i < width * height; i++)
    *dst++ = shift_right_round((*src0++) + (*src1++), ONE);
}

//-----------------------------------------------------------------------------
/*!
* \brief Average two arrays of samples (vertical shift)
*
* Second array is shifted up before averaging.
*
* \param[in] src0 First array of samples
* \param[in] src1 Second array of samples
* \param[out] dst Average array
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::avg_d(unsigned char *src0, unsigned char *src1, unsigned char *dst)
{
  int i;
  
  src1 += width;
  
  for (i = width; i < width * height; i++)
    *dst++ = shift_right_round((*src0++) + (*src1++), ONE);
  
  src1 -= width;
  
  for (i = ZERO; i < width; i++)
    *dst++ = shift_right_round((*src0++) + (*src1++), ONE);
}

//-----------------------------------------------------------------------------
/*!
* \brief Average two arrays of samples (horizontal shift)
*
* Second array is shifted left before averaging.
*
* \param[in] src0 First array of samples
* \param[in] src1 Second array of samples
* \param[out] dst Average array
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::avg_r(unsigned char *src0, unsigned char *src1, unsigned char *dst)
{
  int i, j;
  
  for (i = ZERO; i < height; i++) 
  {
    src1++;
    
    for (j = ZERO; j < width - ONE; j++)
      *dst++ = shift_right_round((*src0++) + (*src1++), ONE);
    
    *dst++ = shift_right_round((*src0++) + (src1[-1]), ONE);
    
  }
}

//-----------------------------------------------------------------------------
// Public methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*! 
* \brief Generate interpolated data
*
* \param[in] ref Reference frame
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::makemap(Frame *ref) 
{
  
  extend(ref);

  if (!decoder_mode)
  {
  generate_hp();
  generate_qp();  
  
  if (param.B_subpel > 1 || param.P_subpel> 1) 
  {
    switch (param.gen8pel_filter) 
    { 
      case BILINEAR_FILTER:
        generate_8p_bilinear();
        break;
      case SIX_TAP_FILTER:
        generate_8p_multi_tap(SIX_TAP_FILTER);
        break;
      case EIGHT_TAP_FILTER:
        generate_8p_multi_tap(EIGHT_TAP_FILTER);
        break;
      default:
        printf("Unknown filter %d for 1/8 pel generation.",param.gen8pel_filter); 
    }
  }
    
#if 0 // Verify planes for debug purposes
    
    int step = (param.B_subpel > 1 || param.P_subpel > 1) ? 1 : 2;
    
    int i, j;
    int u, v;
    
    for (i = 0; i < 8; i += step)
    {
      for (j = 0; j < 8; j += step)
      {
        for (u = 0; u < 4; u++)
        {
          for (v = 0; v < width; v++)
          {
            if (data[i][j][width * u + v] != data[i][j][width * u + v + width])
            {
              printf("A %d %d %d %d\n", i, j, u, v);
            }
            if (data[i][j][width * (height - 1 - u) + v] != data[i][j][width * (height - 2 - u) + v])
            {
              printf("B %d %d %d %d\n", i, j, u, v);
            }
          }
        }
        for (u = 0; u < height; u++)
        {
          for (v = 0; v < 4; v++)
          {
            if (data[i][j][width * u + v] != data[i][j][width * u + v + 1])
            {
              printf("C %d %d %d %d\n", i, j, u, v);
            }
            if (data[i][j][width * u + width - 1 - v] != data[i][j][width * u + width - 2 - v])
            {
              printf("D %d %d %d %d\n", i, j, u, v);
            }
          }
        }
      }
    }
#endif
    
  }
}

//-----------------------------------------------------------------------------
/*! 
* \brief Generate interpolated data using AIF
*
* \param[in] ref Reference frame
* \param[in] cur Frame for determining AIF type
* \param[in] i List index (LIST_0 or LIST_1)
* \param[in] j Reference frame index
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::AIF_makemap(Frame *ref, Frame *cur, int i, int j, AIF_adaptive_filter *saif_half) 
{   
  extend(ref);
  setAIFcoef(saif_half);
  if (!decoder_mode)
  {
    AIF_generate_qp(cur, i, j);
    
    if (param.B_subpel > 1 || param.P_subpel> 1) 
    {
      AIF_generate_8p();
    }
  }
}

void LumaFrames8pel::interpolate_on_the_fly(int k, int l, int u, int v, unsigned char *blk_data)
{
  if ((k&1) == 0 && (l&1) == 0) // quarter-pel
  {
    getblock_QP(k, l, u, v, blk_data);
  }
  else if ((k&1) == 0 && (l&1) != 0) // horizontal eighth pel position
  {
    getblock_EP_hor(k, l, u, v, blk_data);
  }
  else if ((k&1) != 0 && (l&1) == 0) // vertical eighth pel position
  {
    getblock_EP_ver(k, l, u, v, blk_data);
  }
  else // diagonal eighth-pel position
  {
    getblock_EP_diag(k, l, u, v, blk_data);
  }
}

void LumaFrames8pel::interpolate_on_the_fly_AIF(int k, int l, int u, int v, unsigned char *blk_data)
{
  if ((k&1) == 0 && (l&1) == 0) // quarter-pel
  {
    getblock_QP_AIF(k, l, u, v, blk_data);
  }
  else if ((k&1) == 0 && (l&1) != 0) // horizontal eighth pel position
  {
    getblock_EP_hor_AIF(k, l, u, v, blk_data);
  }
  else if ((k&1) != 0 && (l&1) == 0) // vertical eighth pel position
  {
    getblock_EP_ver_AIF(k, l, u, v, blk_data);
  }
  else // diagonal eighth-pel position
  {
    getblock_EP_diag_AIF(k, l, u, v, blk_data);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Get luma block with quarter-pel precision
*
* \param[in] lf: pointer to LumaFrame
* \param[in] k: Vertical sub-pel position
* \param[in] l: Horizontal sub-pel position
* \param[in] u: Vertical position in a frame
* \param[in] v: Vertical position in a frame
* \param[in] list: LIST_0 or LIST_1
* \param[in] ref_index: reference index
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::getblock_QP(int k, int l, int u, int v, unsigned char *blk_data)
{
  int mc_blk_intermediate[64];

  QP_interpolate_fixed_on_the_fly(data[0][0] + width * u + v, blk_data, mc_blk_intermediate, k, l);
}

void LumaFrames8pel::getblock_QP_AIF(int k, int l, int u, int v, unsigned char *blk_data)
{
  int mc_blk_intermediate[64];

  QP_interpolate_adaptive_on_the_fly(data[0][0] + width * u + v, blk_data, mc_blk_intermediate, k, l);
}

//-----------------------------------------------------------------------------
/*!
* \brief Get luma block with eighth-pel precision in horizontal direction
*
* \param[in] lf: pointer to LumaFrame
* \param[in] k: Vertical sub-pel position
* \param[in] l: Horizontal sub-pel position
* \param[in] u: Vertical position in a frame
* \param[in] v: Vertical position in a frame
* \param[in] list: LIST_0 or LIST_1
* \param[in] ref_index: reference index
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::getblock_EP_hor(int k, int l, int u, int v, unsigned char *blk_data)
{
  int mc_blk_intermediate[2][64];

  QP_interpolate_fixed_on_the_fly(data[0][0] + width * u + v, blk_data, mc_blk_intermediate[0], k, l-1);
  QP_interpolate_fixed_on_the_fly(data[0][0] + width * u + v + (l == 7), blk_data, mc_blk_intermediate[1], k, (l+1)&7);

  avg(mc_blk_intermediate[0], mc_blk_intermediate[1], blk_data);
}

void LumaFrames8pel::getblock_EP_hor_AIF(int k, int l, int u, int v, unsigned char *blk_data)
{
  int mc_blk_intermediate[2][64];

  QP_interpolate_adaptive_on_the_fly(data[0][0] + width * u + v, blk_data, mc_blk_intermediate[0], k, l-1);
  QP_interpolate_adaptive_on_the_fly(data[0][0] + width * u + v + (l == 7), blk_data, mc_blk_intermediate[1], k, (l+1)&7);

  avg(mc_blk_intermediate[0], mc_blk_intermediate[1], blk_data);
}

//-----------------------------------------------------------------------------
/*!
* \brief Get luma block with eighth-pel precision in vertical direction
*
* \param[in] lf: pointer to LumaFrame
* \param[in] k: Vertical sub-pel position
* \param[in] l: Horizontal sub-pel position
* \param[in] u: Vertical position in a frame
* \param[in] v: Vertical position in a frame
* \param[in] list: LIST_0 or LIST_1
* \param[in] ref_index: reference index
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::getblock_EP_ver(int k, int l, int u, int v, unsigned char *blk_data)
{
  int mc_blk_intermediate[2][64];
  QP_interpolate_fixed_on_the_fly(data[0][0] + width * u + v, blk_data, mc_blk_intermediate[0], k-1, l);
  QP_interpolate_fixed_on_the_fly(data[0][0] + width * (u + (k==7)) + v, blk_data, mc_blk_intermediate[1], (k+1)&7, l);

  avg(mc_blk_intermediate[0], mc_blk_intermediate[1], blk_data);
}

void LumaFrames8pel::getblock_EP_ver_AIF(int k, int l, int u, int v, unsigned char *blk_data)
{
  int mc_blk_intermediate[2][64];
  QP_interpolate_adaptive_on_the_fly(data[0][0] + width * u + v, blk_data, mc_blk_intermediate[0], k-1, l);
  QP_interpolate_adaptive_on_the_fly(data[0][0] + width * (u + (k==7)) + v, blk_data, mc_blk_intermediate[1], (k+1)&7, l);

  avg(mc_blk_intermediate[0], mc_blk_intermediate[1], blk_data);
}

//-----------------------------------------------------------------------------
/*!
* \brief Get luma block with eighth-pel precision (diagonal)
*
* \param[in] lf: pointer to LumaFrame
* \param[in] k: Vertical sub-pel position
* \param[in] l: Horizontal sub-pel position
* \param[in] u: Vertical position in a frame
* \param[in] v: Vertical position in a frame
* \param[in] list: LIST_0 or LIST_1
* \param[in] ref_index: reference index
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::getblock_EP_diag(int k, int l, int u, int v, unsigned char *blk_data)
{
  int mc_blk_intermediate[2][64];

  QP_interpolate_fixed_on_the_fly(data[0][0] + width * u + v + (l==7), blk_data, mc_blk_intermediate[0], k-1, (l+1)&7);
  QP_interpolate_fixed_on_the_fly(data[0][0] + width * (u + (k==7)) + v, blk_data, mc_blk_intermediate[1], (k+1)&7, l-1);

  avg(mc_blk_intermediate[0], mc_blk_intermediate[1], blk_data, (k == 7 && l == 7 && !param.bugfix87));
}

void LumaFrames8pel::getblock_EP_diag_AIF(int k, int l, int u, int v, unsigned char *blk_data)
{
  int mc_blk_intermediate[2][64];

  QP_interpolate_adaptive_on_the_fly(data[0][0] + width * u + v + (l==7), blk_data, mc_blk_intermediate[0], k-1, (l+1)&7);
  QP_interpolate_adaptive_on_the_fly(data[0][0] + width * (u + (k==7)) + v, blk_data, mc_blk_intermediate[1], (k+1)&7, l-1);

  avg(mc_blk_intermediate[0], mc_blk_intermediate[1], blk_data, (k == 7 && l == 7 && !param.bugfix87));
}

//-----------------------------------------------------------------------------
/*!
* \brief Quarter-pel interpolation of reference blocks with fixed filter 
*
* \param[in] full_pel_data: Pointer to full-pel data
* \param[in] sub_pel_data: Pointer to sub-pel data
* \param[in] intermed_data: Pointer to intermediate data with higher precision
* \param[in] ver: Vertical sub-pel position
* \param[in] hor: Horizontal sub-pel position
* \return 
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::QP_interpolate_fixed_on_the_fly(unsigned char* full_pel_data, unsigned char* sub_pel_data, int* intermed_data, int ver, int hor)
{
  int sub_pos, x, y, fi, isum;
  int stride = width;
  int intermediate_buf[200];
  int iFilterCoef[6] = {1, -5, 20, 20, -5, 1};

  sub_pos = (ver << 1) + (hor >> 1) - 1;

  if (!ver && !hor) // full-pel position
  {
    for (y = ZERO; y < MOTION_BLK_SIZE; y++)
      for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
      {
        sub_pel_data[x + y * MOTION_BLK_SIZE] = full_pel_data[x + y * stride];
        intermed_data[x + y * MOTION_BLK_SIZE] = full_pel_data[x + y * stride] << 5;
      }
  }
  else if (!sub_pos) // a_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + y * stride] + full_pel_data[(x - fi + 3) + y * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum + (full_pel_data[x + y * stride] << 5), 1);
        isum = shift_right_round(isum + (full_pel_data[x + y * stride] << 5), 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;
      }
    }   
  }
  else if (sub_pos == 1) // b_pos
  { 
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + y * stride] + full_pel_data[(x - fi + 3) + y * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = isum;
        isum = shift_right_round(isum, 5);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;
      }
    }
  }
  else if (sub_pos == 2) // c_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + y * stride] + full_pel_data[(x - fi + 3) + y * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum + (full_pel_data[x + 1 + y * stride] << 5), 1);
        isum = shift_right_round(isum + (full_pel_data[x + 1 + y * stride] << 5), 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;
      }
    }   
  }
  else if (sub_pos == 3) // d_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[x + (y + fi - 2) * stride] + full_pel_data[x + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum + (full_pel_data[x + y * stride] << 5), 1);
        isum = shift_right_round(isum + (full_pel_data[x + y * stride] << 5), 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;        
      }
    }   
  }
  else if (sub_pos == 4) // e_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + y * stride] + full_pel_data[(x - fi + 3) + y * stride] + 
                                     full_pel_data[x + (y + fi - 2) * stride] + full_pel_data[x + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum, 1);
        isum = shift_right_round(isum, 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;  
      }
    }
  }
  else if (sub_pos == 5) // f_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE + 5; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + (y - 2) * stride] + full_pel_data[(x - fi + 3) + (y - 2) * stride]);
        intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
      }
    }
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + intermediate_buf[x + (y - fi + 5) * MOTION_BLK_SIZE]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(intermediate_buf[x + (y + 2) * MOTION_BLK_SIZE] + ((isum + 16) >> 5), 1);
        isum = shift_right_round(intermediate_buf[x + (y + 2) * MOTION_BLK_SIZE] + ((isum + 16) >> 5), 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;  
      }
    }  
  }
  else if (sub_pos == 6) // g_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + y * stride] + full_pel_data[(x - fi + 3) + y * stride] + 
                                     full_pel_data[x + 1 + (y + fi - 2) * stride] + full_pel_data[x + 1 + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum, 1);
        isum = shift_right_round(isum, 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;  
      }
    }  
  }
  else if (sub_pos == 7) // h_pos
  { 
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[x + (y + fi - 2) * stride] + full_pel_data[x + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = isum;
        isum = shift_right_round(isum, 5);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;  
      }
    }
  }
  else if (sub_pos == 8) // i_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE + 5; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + (y - 2) * stride] + full_pel_data[(x - fi + 3) + (y - 2) * stride]);
        intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
      }
    }
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + intermediate_buf[x + (y - fi + 5) * MOTION_BLK_SIZE]);
        isum = (isum + 16) >> 5;
        for(fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[x + (y + fi - 2) * stride] + full_pel_data[x + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum, 1);
        isum = shift_right_round(isum, 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum; 
      }
    }  
  }
  else if (sub_pos == 9) // j_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE + 5; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + (y - 2) * stride] + full_pel_data[(x - fi + 3) + (y - 2) * stride]);
        intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
      }
    }
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + intermediate_buf[x + (y - fi + 5) * MOTION_BLK_SIZE]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum, 5);
        isum = shift_right_round(isum, 10);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum; 
      }
    }
  }
  else if (sub_pos == 10) // k_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE + 5; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + (y - 2) * stride] + full_pel_data[(x - fi + 3) + (y - 2) * stride]);
        intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
      }
    }
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + intermediate_buf[x + (y - fi + 5) * MOTION_BLK_SIZE]);
        isum = (isum + 16) >> 5;
        for(fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[x + 1 + (y + fi - 2) * stride] + full_pel_data[x + 1 + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum, 1);        
        isum = shift_right_round(isum, 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum; 
      }
    }    
  }
  else if (sub_pos == 11) // l_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[x + (y + fi - 2) * stride] + full_pel_data[x + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum + (full_pel_data[x + (y + 1) * stride] << 5), 1);     
        isum = shift_right_round(isum + (full_pel_data[x + (y + 1) * stride] << 5), 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum; 
      }
    }   
  }
  else if (sub_pos == 12) // m_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + (y + 1) * stride] + full_pel_data[(x - fi + 3) + (y + 1) * stride] + 
                                     full_pel_data[x + (y + fi - 2) * stride] + full_pel_data[x + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum, 1);      
        isum = shift_right_round(isum, 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum; 
      }
    }  
  }
  else if (sub_pos == 13) // n_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE + 5; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + (y - 2) * stride] + full_pel_data[(x - fi + 3) + (y - 2) * stride]);
        intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
      }
    }
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + intermediate_buf[x + (y - fi + 5) * MOTION_BLK_SIZE]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(intermediate_buf[x + (y + 3) * MOTION_BLK_SIZE] + ((isum + 16) >> 5), 1);       
        isum = shift_right_round(intermediate_buf[x + (y + 3) * MOTION_BLK_SIZE] + ((isum + 16) >> 5), 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum; 
      }
    }    
  }
  else if (sub_pos == 14) // o_pos
  {
    for (y = 0; y < MOTION_BLK_SIZE; y++) 
    {
      for (x = 0; x < MOTION_BLK_SIZE; x++) 
      {
        for(isum=0, fi = 0; fi < 3; fi++)
          isum += iFilterCoef[fi] * (full_pel_data[(x + fi - 2) + (y + 1) * stride] + full_pel_data[(x - fi + 3) + (y + 1) * stride] + 
                                     full_pel_data[x + 1 + (y + fi - 2) * stride] + full_pel_data[x + 1 + (y - fi + 3) * stride]);
        intermed_data[x + y * MOTION_BLK_SIZE] = shift_right_round(isum, 1);
        isum = shift_right_round(isum, 6);
        isum = clip(isum, ZERO, MAX_SAMPLE_VAL);
        sub_pel_data[x + y * MOTION_BLK_SIZE] = isum;
      }
    }   
  }
}


//-----------------------------------------------------------------------------
/*!
* \brief Quarter-pel interpolation of reference blocks with SAIF 
*
* \param[in] full_pel_data: Pointer to full-pel data
* \param[in] sub_pel_data: Pointer to sub-pel data
* \param[in] intermed_data: Pointer to intermediate data with higher precision
* \param[in] ver: Vertical sub-pel position
* \param[in] hor: Horizontal sub-pel position
* \return 
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::QP_interpolate_adaptive_on_the_fly (unsigned char* full_pel_data, unsigned char* sub_pel_data, int* intermed_data, int ver, int hor)
{
  int sub_pos, x, y, fi, isum, isum2, isum3;
  int stride = width;
  int intermediate_buf[200];
  int shift = 5;
  int f_bits_minus_1 = F_BITS - 1;
  int shift_f_bits_minus_1 = (1 << (f_bits_minus_1));
  int f_bits_minus_shift_minus_1 = F_BITS - shift - 1;
  int shift_f_bits_minus_shift_minus_1 = (1 << (f_bits_minus_shift_minus_1));
  int f_bits_plus_shift_minus_1 = F_BITS + shift - 1;
  int shift_f_bits_plus_shift_minus_1 = (1 << (f_bits_plus_shift_minus_1));
  int f_bits_plus_shift = F_BITS + shift;
  int f_bits_minus_shift = F_BITS - shift;
  int shift_minus_1 = shift - 1; 
  int shift_minus_2 = shift - 2; 
  int offset = 0;

  sub_pos = (ver << 1) + (hor >> 1) - 1;

  //////////////////////////  full-pel position  //////////////////////////
  if (!ver && !hor) 
  {
    for (y = ZERO; y < MOTION_BLK_SIZE; y++)
      for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
      {
        sub_pel_data[x + y * MOTION_BLK_SIZE] = full_pel_data[x + y * stride];
        intermed_data[x + y * MOTION_BLK_SIZE] = full_pel_data[x + y * stride] << 5;
      }
  }
  //////////////////////////  a_pos, b_pos, c_pos  //////////////////////////
  else if (sub_pos < 3) 
  {     
    if (sub_pos == 1 || AIF_mode_flag1)
    {
      for (y = ZERO; y < MOTION_BLK_SIZE; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[sub_pos][fi] * full_pel_data[(x + fi - 2) + y * stride];
          isum += iFilterCoef_SAIF[sub_pos][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        }
      }
    }
    else
    {
      offset = (sub_pos == 2) ? 1 : 0;
      for (y = ZERO; y < MOTION_BLK_SIZE; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + y * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          isum = imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + full_pel_data[x + offset + y * stride] ) << shift_minus_1;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = ((isum + full_pel_data[x + offset + y * stride] + 1) >> 1);
        }
      }
    }
  }
  //////////////////////////  d_pos, h_pos, l_pos  //////////////////////////
  else if (sub_pos == 3 || sub_pos == 7 || sub_pos == 11) 
  {      
    if ( sub_pos == 7 || AIF_mode_flag1) // AIF_MODE2, AIF_MODE3, h_pos
    {
      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++) 
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[sub_pos][fi] * full_pel_data[x + (y - 2 + fi) * stride];
          }
          isum += iFilterCoef_SAIF[sub_pos][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
        }
      }
    }
    else // AIF_MODE1
    {
      offset = (sub_pos == 11) ? 1 : 0;
      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++) 
        {     
          isum = 0;
          {
            for(fi = 0; fi < FILTER_SIZE; fi++)
            {
              isum += iFilterCoef_SAIF[7][fi] * full_pel_data[x + (y - 2 + fi) * stride];
            }
            isum += iFilterCoef_SAIF[7][FILTER_SIZE];
          }
          isum = imin(255,imax(0,((isum + shift_f_bits_minus_1) >> F_BITS)));
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + full_pel_data[x + (y + offset) * stride]) << shift_minus_1;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = ((isum + full_pel_data[x + (y + offset) * stride] + 1) >> 1);
        }
      }
    }
  }
  //////////////////////////  e_pos  //////////////////////////
  else if (sub_pos == 4) 
  {
    if (AIF_mode_flag2) // AIF_mode3
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[0][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[0][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[4][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
          }
          isum += iFilterCoef_SAIF[4][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
        }
      }
    }
    else // AIF_mode1, AIF_mode2
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++) 
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[9][fi] * ((intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift); // j_pos
          }
          isum += iFilterCoef_SAIF[9][FILTER_SIZE]; 
          isum = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));// j_pos

          isum2 = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum2 += iFilterCoef_SAIF[7][fi] * full_pel_data[x + (y - 2 + fi) * stride]; // h_pos
          }
          isum2 += iFilterCoef_SAIF[7][FILTER_SIZE]; // h_pos
          isum2 = imin(255,imax(0,((isum2 + shift_f_bits_minus_1) >> F_BITS)));

          isum3 = imin(255,imax(0,((intermediate_buf[x + (y + 2) * MOTION_BLK_SIZE] + shift_f_bits_minus_1) >> F_BITS))); // b_pos

          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + isum3 + full_pel_data[x + y * stride]) << shift_minus_2;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + isum3 + full_pel_data[x + y * stride] + 2) >> 2; 
        }
      }
    }
  }
  //////////////////////////  f_pos  //////////////////////////
  else if (sub_pos == 5) 
  {
    if (AIF_mode_flag2) // AIF_mode3
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)  
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[5][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
          }
          isum += iFilterCoef_SAIF[5][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
        }
      }
    }
    else // AIF_mode1 , AIF_mode2 
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
        }
      }
      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[9][fi] * ((intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift);
          }
          isum += iFilterCoef_SAIF[9][FILTER_SIZE];
          isum = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
          isum += imin(255,imax(0,((intermediate_buf[x + (y + 2) * MOTION_BLK_SIZE] + shift_f_bits_minus_1) >> F_BITS)));
          intermed_data[x + y * MOTION_BLK_SIZE] = isum << shift_minus_1;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = (isum + 1) >> 1;
        }
      }
    }
  }
  //////////////////////////  g_pos  //////////////////////////
  else if (sub_pos == 6) 
  {
    if (AIF_mode_flag2) // AIF_mode3
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[2][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[2][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++) 
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[6][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
          }
          isum += iFilterCoef_SAIF[6][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
        }
      }
    }
    else // AIF_mode1, AIF_mode2
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
        }
      }
      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++) 
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[9][fi] * ((intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift); // j_pos
          }
          isum += iFilterCoef_SAIF[9][FILTER_SIZE]; 
          isum = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));// j_pos

          isum2 = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum2 += iFilterCoef_SAIF[7][fi] * full_pel_data[x + 1 + (y - 2 + fi) * stride]; // h_pos ++
          }
          isum2 += iFilterCoef_SAIF[7][FILTER_SIZE]; // h_pos ++
          isum2 = imin(255,imax(0,((isum2 + shift_f_bits_minus_1) >> F_BITS)));

          isum3 = imin(255,imax(0,((intermediate_buf[x + (y + 2) * MOTION_BLK_SIZE] + shift_f_bits_minus_1) >> F_BITS))); // b_pos
          
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + isum3 + full_pel_data[x + 1 + y * stride]) << shift_minus_2;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + isum3 + full_pel_data[x + 1 + y * stride] + 2) >> 2; 
        }
      }
    }
  }
  //////////////////////////  i_pos  //////////////////////////
  else if (sub_pos == 8) 
  {
    if (AIF_mode_flag2) // AIF_mode3
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[0][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[0][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)  
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[8][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
          }
          isum += iFilterCoef_SAIF[8][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
        }
      }
    }
    else // AIF_mode1, AIF_mode2
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)
        {     
          isum = 0;
          isum2 = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[9][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE]; // j_pos
          }
          isum += iFilterCoef_SAIF[9][FILTER_SIZE]; // j_pos
          isum = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));

          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum2 += iFilterCoef_SAIF[7][fi] * full_pel_data[x + (y - 2 + fi) * stride]; // h_pos
          }
          isum2 += iFilterCoef_SAIF[7][FILTER_SIZE]; // h_pos
          isum2 = imin(255,imax(0,((isum2 + shift_f_bits_minus_1) >> F_BITS)));
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + isum2) << shift_minus_1;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + 1) >> 1; 
        }
      }
    }
  }
  //////////////////////////  j_pos  //////////////////////////
  else if (sub_pos == 9) 
  {
    for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
    {
      for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
      {           
        isum = 0;
        for(fi = 0; fi < FILTER_SIZE; fi++)
          isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
        isum += iFilterCoef_SAIF[1][FILTER_SIZE];
        intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
      }
    }

    for (x = ZERO; x < MOTION_BLK_SIZE; x++)
    {
      for (y = ZERO; y < MOTION_BLK_SIZE; y++)  
      {     
        isum = 0;
        for(fi = 0; fi < FILTER_SIZE; fi++)
        {
          isum += iFilterCoef_SAIF[9][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
        }
        isum += iFilterCoef_SAIF[9][FILTER_SIZE];
        intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
        sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
      }
    }
  }
  //////////////////////////  k_pos  //////////////////////////
  else if (sub_pos == 10) 
  {
    if (AIF_mode_flag2) // AIF_mode3
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[2][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[2][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[10][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
          }
          isum += iFilterCoef_SAIF[10][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
        }
      }
    }
    else // AIF_mode1, AIF_mode2
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)
        {     
          isum = 0;
          isum2 = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[9][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE]; // j_pos
          }
          isum += iFilterCoef_SAIF[9][FILTER_SIZE];
          isum = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));

          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum2 += iFilterCoef_SAIF[7][fi] * full_pel_data[x + 1 + (y - 2 + fi) * stride]; // h_pos ++
          }
          isum2 += iFilterCoef_SAIF[7][FILTER_SIZE]; // h_pos ++
          isum2 = imin(255,imax(0,((isum2 + shift_f_bits_minus_1) >> F_BITS)));
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + isum2) << shift_minus_1;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + 1) >> 1; 
        }
      }
    }
  }
  //////////////////////////  m_pos  //////////////////////////
  else if (sub_pos == 12) 
  {
    if (AIF_mode_flag2) // AIF_mode3
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[0][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[0][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)  
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[12][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
          }
          isum += iFilterCoef_SAIF[12][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
        }
      }
    }
    else // AIF_mode1, AIF_mode2
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++) 
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[9][fi] * ((intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift); // j_pos
          }
          isum += iFilterCoef_SAIF[9][FILTER_SIZE];
          isum = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));// j_pos

          isum2 = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum2 += iFilterCoef_SAIF[7][fi] * full_pel_data[x + (y - 2 + fi) * stride]; // h_pos 
          }
          isum2 += iFilterCoef_SAIF[7][FILTER_SIZE]; // h_pos 
          isum2 = imin(255,imax(0,((isum2 + shift_f_bits_minus_1) >> F_BITS)));

          isum3 = imin(255,imax(0,((intermediate_buf[x + (y + 3) * MOTION_BLK_SIZE] + shift_f_bits_minus_1) >> F_BITS))); // b_pos
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + isum3 + full_pel_data[x + (y + 1) * stride]) << shift_minus_2;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + isum3 + full_pel_data[x + (y + 1) * stride] + 2) >> 2; 
        }
      }
    }
  }
  //////////////////////////  n_pos  //////////////////////////
  else if (sub_pos == 13) 
  {
    if (AIF_mode_flag2) // AIF_mode3
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++) 
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[13][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
          }
          isum += iFilterCoef_SAIF[13][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
        }
      }
    }
    else // AIF_mode1 , AIF_mode2 
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
        }
      }
      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[9][fi] * ((intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift);
          }
          isum += iFilterCoef_SAIF[9][FILTER_SIZE];
          isum = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));
          isum += imin(255,imax(0,((intermediate_buf[x + (y + 3) * MOTION_BLK_SIZE] + shift_f_bits_minus_1) >> F_BITS)));
          intermed_data[x + y * MOTION_BLK_SIZE] = isum << shift_minus_1;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = (isum + 1) >> 1;
        }
      }
    }
  }
  //////////////////////////  o_pos  //////////////////////////
  else if (sub_pos == 14) 
  {
    if (AIF_mode_flag2) // AIF_mode3
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[2][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[2][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift;
        }
      }

      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++)  
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[14][fi] * intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE];
          }
          isum += iFilterCoef_SAIF[14][FILTER_SIZE];
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + shift_f_bits_minus_1) >> F_BITS;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift))); 
        }
      }
    }
    else // AIF_mode1, AIF_mode2
    {
      for (y = ZERO; y < MOTION_BLK_SIZE + 5; y++)
      {
        for (x = ZERO; x < MOTION_BLK_SIZE; x++) 
        {           
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
            isum += iFilterCoef_SAIF[1][fi] * full_pel_data[(x + fi - 2) + (y - 2) * stride];
          isum += iFilterCoef_SAIF[1][FILTER_SIZE];
          intermediate_buf[x + y * MOTION_BLK_SIZE] = isum;
        }
      }
      for (x = ZERO; x < MOTION_BLK_SIZE; x++)
      {
        for (y = ZERO; y < MOTION_BLK_SIZE; y++) 
        {     
          isum = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum += iFilterCoef_SAIF[9][fi] * ((intermediate_buf[x + (y + fi) * MOTION_BLK_SIZE] + shift_f_bits_minus_shift_minus_1) >> f_bits_minus_shift); // j_pos
          }
          isum += iFilterCoef_SAIF[9][FILTER_SIZE];
          isum = imin(255,imax(0,((isum + shift_f_bits_plus_shift_minus_1) >> f_bits_plus_shift)));// j_pos

          isum2 = 0;
          for(fi = 0; fi < FILTER_SIZE; fi++)
          {
            isum2 += iFilterCoef_SAIF[7][fi] * full_pel_data[x + 1 + (y - 2 + fi) * stride]; // h_pos ++
          }
          isum2 += iFilterCoef_SAIF[7][FILTER_SIZE]; // h_pos ++
          isum2 = imin(255,imax(0,((isum2 + shift_f_bits_minus_1) >> F_BITS)));

          isum3 = imin(255,imax(0,((intermediate_buf[x + (y + 3) * MOTION_BLK_SIZE] + shift_f_bits_minus_1) >> F_BITS))); // b_pos
          intermed_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + isum3 + full_pel_data[x + 1 + (y + 1) * stride]) << shift_minus_2;
          sub_pel_data[x + y * MOTION_BLK_SIZE] = (isum + isum2 + isum3 + full_pel_data[x + 1 + (y + 1) * stride] + 2) >> 2; 
        }
      }
    }
  }
}



//-----------------------------------------------------------------------------
/*!
* \brief Average two arrays of samples
*
* \param[in] src0 First array of samples
* \param[in] src1 Second array of samples
* \param[out] dst0 Average array (Higher Precision)
* \param[out] dst1 Average array (Lower Precision and Rounded)
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::avg(int *src0, int *src1, unsigned char *dst1, bool bug)
{
  int i, v, u;
  int shift = (bug) ? 1 : 6;

  for (i = 0; i < MOTION_BLK_SIZE * MOTION_BLK_SIZE; i++) 
  {
    u = src0[i] + src1[i];
    v = shift_right_round(u, shift);
    if (!bug)
      v = clip(v, ZERO, MAX_SAMPLE_VAL);
    dst1[i]  = v;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Set AIF filter coefficients 
*
* \param[in] saif_half: pointer to AIF data
*/
//-----------------------------------------------------------------------------
void LumaFrames8pel::setAIFcoef(AIF_adaptive_filter *saif_half)
{
  for(int j=0;j<15;j++)
    for(int i=0;i<7;i++)
      iFilterCoef_SAIF[j][i]=saif_half->iFilterCoef_SAIF[j][i];   

  AIF_mode_flag1 = saif_half->AIF_mode_flag1;
  AIF_mode_flag2 = saif_half->AIF_mode_flag2;
}
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
