/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MCRefBlk8.cpp
*
* \brief Implementation of MCRefBlk8 class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MCRefBlk8.H"
#include "Param.H"

//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief mel_shift_sad Shifts the SAD by mel_shift
*
* Returns the SAD which is multiplied by mult and shifted by the (mel_shift - shift).
*
* \param[in] sad the raw integer sad value
* \param[in] mult the integer multiplier value
* \param[in] shift the integer shift value
* \return SAD shifted by mel_shift
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::mel_shift_sad(int sad, int mult, int shift)       
{
  sad = (int) (mult * sad);
  if (mel_shift >= shift)
    return sad << (mel_shift - shift);
  else
    return shift_right_round(sad, shift - mel_shift);
}

//-----------------------------------------------------------------------------
/*!
* \brief Transform 4x4 block and compute SAD
*
* Applies a Hadamard transform and computes the SAD in the transformed domain.
*
* \param[in] x Block array (addressed by [8*i+j])
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::transform(int *x) 
{
  int i;
  int a,b,c,d;
  int e,f,g,h;
  int r;
  int *p;

  p = x;

  for (i = ZERO; i < TWO * TX_SIZE; i++) 
  {
    a = (p[8*0] + 2) >> 2;
    b = (p[8*1] + 2) >> 2;
    c = (p[8*2] + 2) >> 2;
    d = (p[8*3] + 2) >> 2;

    e = a + d;
    f = b + c;
    g = b - c;
    h = a - d;

    p[8*0] = e + f;
    p[8*1] = h + g;
    p[8*2] = e - f;
    p[8*3] = g - h;

    p ++;
  }

  r = ZERO;
  p = x;

  for (i = ZERO; i < TWO * TX_SIZE; i++) 
  {
    a = (p[0] + 2) >> 2;
    b = (p[1] + 2) >> 2;
    c = (p[2] + 2) >> 2;
    d = (p[3] + 2) >> 2;

    e = a + d;
    f = b + c;
    g = b - c;
    h = a - d;

    r += imax(abs(e),abs(f));
    r += imax(abs(h),abs(g));

    p += 4;
  }


  if (mel_shift >= SAMPLE_ADD_PREC - ONE - FOUR)
    return r << (mel_shift + ONE - SAMPLE_ADD_PREC + FOUR);
  else
    return shift_right_round(r, SAMPLE_ADD_PREC - ONE - mel_shift - FOUR);
}

int MCRefBlk8::transformC(int *x)
{
  if (chroma_weight == 0)
    return 0;

  int r = transform(x);
  if (chroma_weight == 256)
    return r;
  else
    return (r * chroma_weight) >> 8;
}

static int m1[8][8], m2[8][8], m3[8][8];

//-----------------------------------------------------------------------------
/*!
* \brief Transform 8x8 block and compute SAD
*
* Applies a Hadamard transform and computes the SAD in the transformed domain.
*
* \param[in] diff Block array (addressed by [8*i+j])
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::transform8x8(int *diff) 
{
  int i, j, jj, satd=0;

  //horizontal
  for (j=0; j < 8; j++)
  {
    jj = j << 3;
    m2[j][0] = diff[jj  ] + diff[jj+4];
    m2[j][1] = diff[jj+1] + diff[jj+5];
    m2[j][2] = diff[jj+2] + diff[jj+6];
    m2[j][3] = diff[jj+3] + diff[jj+7];
    m2[j][4] = diff[jj  ] - diff[jj+4];
    m2[j][5] = diff[jj+1] - diff[jj+5];
    m2[j][6] = diff[jj+2] - diff[jj+6];
    m2[j][7] = diff[jj+3] - diff[jj+7];

    m1[j][0] = m2[j][0] + m2[j][2];
    m1[j][1] = m2[j][1] + m2[j][3];
    m1[j][2] = m2[j][0] - m2[j][2];
    m1[j][3] = m2[j][1] - m2[j][3];
    m1[j][4] = m2[j][4] + m2[j][6];
    m1[j][5] = m2[j][5] + m2[j][7];
    m1[j][6] = m2[j][4] - m2[j][6];
    m1[j][7] = m2[j][5] - m2[j][7];

    m2[j][0] = m1[j][0] + m1[j][1];
    m2[j][1] = m1[j][0] - m1[j][1];
    m2[j][2] = m1[j][2] + m1[j][3];
    m2[j][3] = m1[j][2] - m1[j][3];
    m2[j][4] = m1[j][4] + m1[j][5];
    m2[j][5] = m1[j][4] - m1[j][5];
    m2[j][6] = m1[j][6] + m1[j][7];
    m2[j][7] = m1[j][6] - m1[j][7];
  }

  //vertical
  for (i=0; i < 8; i++)
  {
    m3[0][i] = m2[0][i] + m2[4][i];
    m3[1][i] = m2[1][i] + m2[5][i];
    m3[2][i] = m2[2][i] + m2[6][i];
    m3[3][i] = m2[3][i] + m2[7][i];
    m3[4][i] = m2[0][i] - m2[4][i];
    m3[5][i] = m2[1][i] - m2[5][i];
    m3[6][i] = m2[2][i] - m2[6][i];
    m3[7][i] = m2[3][i] - m2[7][i];

    m1[0][i] = m3[0][i] + m3[2][i];
    m1[1][i] = m3[1][i] + m3[3][i];
    m1[2][i] = m3[0][i] - m3[2][i];
    m1[3][i] = m3[1][i] - m3[3][i];
    m1[4][i] = m3[4][i] + m3[6][i];
    m1[5][i] = m3[5][i] + m3[7][i];
    m1[6][i] = m3[4][i] - m3[6][i];
    m1[7][i] = m3[5][i] - m3[7][i];

    m2[0][i] = m1[0][i] + m1[1][i];
    m2[1][i] = m1[0][i] - m1[1][i];
    m2[2][i] = m1[2][i] + m1[3][i];
    m2[3][i] = m1[2][i] - m1[3][i];
    m2[4][i] = m1[4][i] + m1[5][i];
    m2[5][i] = m1[4][i] - m1[5][i];
    m2[6][i] = m1[6][i] + m1[7][i];
    m2[7][i] = m1[6][i] - m1[7][i];
  }
  for (j=0; j < 8; j++)
    for (i=0; i < 8; i++) 
      satd += abs(m2[j][i]);

  if (mel_shift >= SAMPLE_ADD_PREC - ONE - FOUR + 6)
    return satd << (mel_shift + ONE - SAMPLE_ADD_PREC + FOUR -6);
  else
    return shift_right_round(satd, SAMPLE_ADD_PREC - ONE - mel_shift - FOUR + 6);
}

//-----------------------------------------------------------------------------
/*!
* \brief Transform 2x8 block and compute SAD
*
* Applies a Hadamard transform and computes the SAD in the transformed domain.
*
* \param[in] diff Block array (addressed by [8*i+j])
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::transform2x8(int *diff) 
{
  int i, j, jj, satd=0;

  //horizontal
  for (j=0; j < 8; j++)
  {
    jj = j << 3;
    m2[j][0] = diff[jj  ] + diff[jj+4];
    m2[j][1] = diff[jj+1] + diff[jj+5];
    m2[j][2] = diff[jj+2] + diff[jj+6];
    m2[j][3] = diff[jj+3] + diff[jj+7];
    m2[j][4] = diff[jj  ] - diff[jj+4];
    m2[j][5] = diff[jj+1] - diff[jj+5];
    m2[j][6] = diff[jj+2] - diff[jj+6];
    m2[j][7] = diff[jj+3] - diff[jj+7];

    m1[j][0] = m2[j][0] + m2[j][2];
    m1[j][1] = m2[j][1] + m2[j][3];
    m1[j][2] = m2[j][0] - m2[j][2];
    m1[j][3] = m2[j][1] - m2[j][3];
    m1[j][4] = m2[j][4] + m2[j][6];
    m1[j][5] = m2[j][5] + m2[j][7];
    m1[j][6] = m2[j][4] - m2[j][6];
    m1[j][7] = m2[j][5] - m2[j][7];

    m2[j][0] = m1[j][0] + m1[j][1];
    m2[j][1] = m1[j][0] - m1[j][1];
    m2[j][2] = m1[j][2] + m1[j][3];
    m2[j][3] = m1[j][2] - m1[j][3];
    m2[j][4] = m1[j][4] + m1[j][5];
    m2[j][5] = m1[j][4] - m1[j][5];
    m2[j][6] = m1[j][6] + m1[j][7];
    m2[j][7] = m1[j][6] - m1[j][7];
  }

  //vertical
  for (i=0; i < 8; i++)
  {
    m3[0][i] = m2[0][i] + m2[1][i];
    m3[1][i] = m2[0][i] - m2[1][i];
    m3[2][i] = m2[2][i] + m2[3][i];
    m3[3][i] = m2[2][i] - m2[3][i];
    m3[4][i] = m2[4][i] + m2[5][i];
    m3[5][i] = m2[4][i] - m2[5][i];
    m3[6][i] = m2[6][i] + m2[7][i];
    m3[7][i] = m2[6][i] - m2[7][i];
  }
  for (j=0; j < 8; j++)
    for (i=0; i < 8; i++) 
      satd += abs(m3[j][i]);

  if (mel_shift >= SAMPLE_ADD_PREC - ONE - FOUR + 5)
    return satd << (mel_shift + ONE - SAMPLE_ADD_PREC + FOUR - 5);
  else
    return shift_right_round(satd, SAMPLE_ADD_PREC - ONE - mel_shift - FOUR + 5);
}

//-----------------------------------------------------------------------------
/*!
* \brief Transform 8x2 block and compute SAD
*
* Applies a Hadamard transform and computes the SAD in the transformed domain.
*
* \param[in] diff Block array (addressed by [8*i+j])
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::transform8x2(int *diff) 
{
  int i, j, jj, satd=0;

  //horizontal
  for (j=0; j < 8; j++)
  {
    jj = j << 3;
    m2[j][0] = diff[jj  ] + diff[jj+1];
    m2[j][1] = diff[jj  ] - diff[jj+1];
    m2[j][2] = diff[jj+2] + diff[jj+3];
    m2[j][3] = diff[jj+2] - diff[jj+3];
    m2[j][4] = diff[jj+4] + diff[jj+5];
    m2[j][5] = diff[jj+4] - diff[jj+5];
    m2[j][6] = diff[jj+6] + diff[jj+7];
    m2[j][7] = diff[jj+6] - diff[jj+7];
  }

  //vertical
  for (i=0; i < 8; i++)
  {
    m3[0][i] = m2[0][i] + m2[4][i];
    m3[1][i] = m2[1][i] + m2[5][i];
    m3[2][i] = m2[2][i] + m2[6][i];
    m3[3][i] = m2[3][i] + m2[7][i];
    m3[4][i] = m2[0][i] - m2[4][i];
    m3[5][i] = m2[1][i] - m2[5][i];
    m3[6][i] = m2[2][i] - m2[6][i];
    m3[7][i] = m2[3][i] - m2[7][i];

    m1[0][i] = m3[0][i] + m3[2][i];
    m1[1][i] = m3[1][i] + m3[3][i];
    m1[2][i] = m3[0][i] - m3[2][i];
    m1[3][i] = m3[1][i] - m3[3][i];
    m1[4][i] = m3[4][i] + m3[6][i];
    m1[5][i] = m3[5][i] + m3[7][i];
    m1[6][i] = m3[4][i] - m3[6][i];
    m1[7][i] = m3[5][i] - m3[7][i];

    m2[0][i] = m1[0][i] + m1[1][i];
    m2[1][i] = m1[0][i] - m1[1][i];
    m2[2][i] = m1[2][i] + m1[3][i];
    m2[3][i] = m1[2][i] - m1[3][i];
    m2[4][i] = m1[4][i] + m1[5][i];
    m2[5][i] = m1[4][i] - m1[5][i];
    m2[6][i] = m1[6][i] + m1[7][i];
    m2[7][i] = m1[6][i] - m1[7][i];
  }
  for (j=0; j < 8; j++)
    for (i=0; i < 8; i++) 
      satd += abs(m2[j][i]);

  if (mel_shift >= SAMPLE_ADD_PREC - ONE - FOUR + 5)
    return satd << (mel_shift + ONE - SAMPLE_ADD_PREC + FOUR - 5);
  else
    return shift_right_round(satd, SAMPLE_ADD_PREC - ONE - mel_shift - FOUR + 5);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (inter)
*
* Tests for the use of biprediction and intensity compensation and calls
* dedicated SAD function.
*
* \note Method name may be needed to change, e.g. sad_inter(), to avoid aliasing in Doxygen
*
* \param[in] i Vertical position of block (8x8 block units)
* \param[in] j Horizontal position of block (8x8 block units)
* \param[in] blk Motion data
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::sad(int i, int j, PredBlk *blk) 
{
  if (blk->list == BI_PRED) 
  {
    if (blk->BP_scale[0] != 0 || blk->BP_scale[1] != 0 || blk->BP_DC !=0)
      return bipred_sad_wp(i, j, INT_INF, blk);
    else
      return bipred_sad(i, j, INT_INF, blk);
  }
  else 
  {
    if (blk->scale[blk->list] != 0 || blk->DC[blk->list] !=0)
      return sad_wp(i, j, INT_INF, blk);
    else
      return sad(i, j, INT_INF, blk);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (bipred)
*
* Tests for the use of intensity compensation and calls
* dedicated SAD function.
*
* \note Method name may be needed to change, e.g. bipred_sad_inter(), to avoid aliasing in Doxygen
*
* \param[in] i Vertical position of block (8x8 block units)
* \param[in] j Horizontal position of block (8x8 block units)
* \param[in] blk Motion data
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::bipred_sad(int i, int j, PredBlk *blk) 
{
  if (blk->BP_scale[0] != 0 || blk->BP_scale[1] != 0 || blk->BP_DC !=0)
    return bipred_sad_wp(i, j, INT_INF, blk);
  else
    return bipred_sad(i, j, INT_INF, blk);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD with early exit for inter 8x8
*
* Tests for the use of intensity compensation and calls
* dedicated SAD function. Exits early if partial SAD exceeds maxsad.
*
* \param[in] i Vertical position of block (8x8 block units)
* \param[in] j Horizontal position of block (8x8 block units)
* \param[in] maxsad Threshold for early exit
* \param[in] blk Motion data
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::distortion(int i,int j,int maxsad,PredBlk *blk) 
{
  if (blk->list == BI_PRED) 
  {
    if (blk->BP_scale[0] != 0 || blk->BP_scale[1] != 0 || blk->BP_DC !=0)
      return bipred_sad_wp(i, j, maxsad, blk);
    else
      return bipred_sad(i, j, maxsad, blk);
  }
  else 
  {
    if (blk->scale[blk->list] != 0 || blk->DC[blk->list] !=0)
      return sad_wp(i, j, maxsad, blk);
    else
      return sad(i, j, maxsad, blk);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (general)
*
* Tests for the use of intensity compensation and calls
* dedicated SAD function. Also applies to I-coded blocks.
*
* \param[in] i Vertical position of block (8x8 block units)
* \param[in] j Horizontal position of block (8x8 block units)
* \param[in] blk Motion data
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::sad_w_intra_check(int i, int j, PredBlk *blk) 
{
  if (blk->intra != NULL)
    return blk->intra->sad;
  else
    return sad(i, j, blk);
}

//-----------------------------------------------------------------------------
/*!
* \brief Retrieve reference block
*
* Fetches reference block that will be compared during SAD computations.
*
* \param[in] ref Current frame
* \param[in] u Vertical position (8x8 block units)
* \param[in] v Horizontal position (8x8 block units)
*/
//-----------------------------------------------------------------------------
void MCRefBlk8::get_data(Frame *ref, int u, int v) 
{
  get_luma(ref, u, v);
  get_chroma(ref, u, v, 0);
  get_chroma(ref, u, v, 1);
}

//-----------------------------------------------------------------------------
/*!
* \brief Retrieve luma data of reference block
*
* Fetches luma data of reference block that will be compared during SAD computations.
*
* \param[in] ref Current frame
* \param[in] u Vertical position (8x8 block units)
* \param[in] v Horizontal position (8x8 block units)
*/
//-----------------------------------------------------------------------------
void MCRefBlk8::get_luma(Frame *ref, int u, int v) 
{
  int i,j;
  unsigned char *p;

  p = ref->luma+8*u*ref->width+8*v;

  for (i=0; i<8; i++) 
  {
    for (j=0; j<8; j++)
      val[8*i+j] = p[j];
    p += ref->width;
  }
  p = (unsigned char*)val;
  sum_luma = 0;
  for (i=0; i<64; i++)
    sum_luma += p[i];

}

//-----------------------------------------------------------------------------
/*!
* \brief Retrieve chroma data of reference block
*
* Fetches chroma data of reference block that will be compared during SAD computations.
*
* \param[in] ref Current frame
* \param[in] u Vertical position (4x4 chroma block units)
* \param[in] v Horizontal position (4x4 chroma block units)
* \param[in] c Chroma component index
*/
//-----------------------------------------------------------------------------
void MCRefBlk8::get_chroma(Frame *ref, int u, int v, int c) 
{
  int i,j;
  unsigned char *p;

  p = ref->crcb[c]+4*u*ref->width2+4*v;

  for (i=0; i<4; i++) 
  {
    for (j=0; j<4; j++)
      val[64 + 8*i + 2*j + c] = p[j];
    p += ref->width2;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Get sum of luma data in motion compensated block
*
* \param[in] i Vertical position (8x8 luma block units)
* \param[in] j Horizontal position (8x8 luma block units)
* \param[in] blk Motion data
* \param[in] list List to use for generating MC block
* \return Sum
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::get_mc_sum_luma(int i, int j, PredBlk *blk, int list) 
{
  int k, l;

  unsigned char *p = get_luma_ptr(blk, list, i, j);
  int stride = ref_mgr->get_stride();

  int sum = 0;
  for (k=0; k<8; k++) 
  {
    for (l=0; l<8; l++)
      sum += p[l];
    p += stride;
  }

  return sum;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to luma 8x8 data in reference interpolated array
*
* \param[in] blk Motion data
* \param[in] list List to use
* \param[in] i Vertical position (8x8 luma block units)
* \param[in] j Horizontal position (8x8 luma block units)
* \return Pointer to luma data
*/
//-----------------------------------------------------------------------------
unsigned char *MCRefBlk8::get_luma_ptr(PredBlk *blk, int list, int i, int j)
{
  return ref_mgr->get_luma_ptr(list, blk->ref_index[list], blk->mv[list], 8 * i, 8 * j, 8, 8);
}

//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to chroma 4x4 data in reference interpolated array
*
* \param[in] blk Motion data
* \param[in] list List to use
* \param[in] i Vertical position (8x8 luma block units)
* \param[in] j Horizontal position (8x8 luma block units)
* \return Pointer to chroma data
*/
//-----------------------------------------------------------------------------
unsigned char *MCRefBlk8::get_chroma_ptr(PredBlk *blk, int list, int i, int j, unsigned char *data)
{
  return ref_mgr->get_chroma_ptr(list, blk->ref_index[list], blk->mv[list], 8 * i, 8 * j, 8, 8, data);
}

int MCRefBlk8::distortion_luma(int *diff)
{
  int i;
  const int size = MOTION_BLK_SIZE * MOTION_BLK_SIZE;
  int result = 0;
  
  if (param.trans8x8_enable)
  {
#ifdef USE_SSD
    if (param.use_ssd)   
    {
      for (i = 0; i < size; i++)
      {
        result += param.ssd_lookup[imin(abs(diff[i] >> SAMPLE_ADD_PREC), MAX_SAMPLE_VAL)];
      }
      
      result = mel_shift_sad(result, param.ssd_mult, SAMPLE_ADD_PREC + SSD_SHIFT);
    }
    else
#endif
    if (param.hadamard == 2)
    {
      result += transform8x8(diff);    
    }
    else if (param.hadamard == 1)
    {
      result += transform(diff);
      result += transform(diff + 4 * MOTION_BLK_SIZE);
    }
    else
    {
      assert(param.hadamard == 0);
      for (i = 0; i < size; i++)
      {
        result += abs(diff[i]);
      }
      
      result = mel_shift_sad(result, param.h0_mult, SAMPLE_ADD_PREC + HADAMARD_SHIFT_0);
    }
  }
  else
  {
    result += transform(diff);
    result += transform(diff + 4 * MOTION_BLK_SIZE);
  }
  
  return result;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (single-list, no intensity compensation)
*
* \param[in] i Vertical position (8x8 block units)
* \param[in] j Horizontal position (8x8 block units)
* \param[in] maxval Threshold for early exit
* \param[in] blk Motion data
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::sad(int i, int j, int maxval, PredBlk *blk) 
{
  int k, l, n;
  int diff_luma[MOTION_BLK_SIZE * MOTION_BLK_SIZE]; 
  int diff_chroma[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2]; 
  unsigned char chroma_data[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  
  int list = blk->list;
  int w = ref_mgr->get_stride();
  unsigned char *p = get_luma_ptr(blk, list, i, j);

  for (k = 0; k < MOTION_BLK_SIZE; k++)
  {
    for (l = 0; l < MOTION_BLK_SIZE; l++)
    {
      int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
      int pred_val = p[l] << SAMPLE_ADD_PREC;
      diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
    }
    
    p += w;
  }
  
  int sad = distortion_luma(diff_luma);
  if (sad >= maxval)
    return maxval;
  
  p = get_chroma_ptr(blk, list, i, j, (param.chroma_on_fly) ? chroma_data : NULL);
  w = (param.chroma_on_fly) ? MOTION_BLK_SIZE : w;

  for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
  {
    for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
    {
      for (n = 0; n < 2; n++) 
      {
        int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
        int pred_val = p[2*l + n] << SAMPLE_ADD_PREC;
        diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
      }
    }
    
    p += w;
  }

  sad += transformC(diff_chroma);

  return sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (single-list, with intensity compensation)
*
* \param[in] i Vertical position (8x8 block units)
* \param[in] j Horizontal position (8x8 block units)
* \param[in] maxval Threshold for early exit
* \param[in] blk Motion data
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::sad_wp(int i, int j, int maxval, PredBlk *blk) 
{
  int k, l, n;
  int diff_luma[MOTION_BLK_SIZE * MOTION_BLK_SIZE]; 
  int diff_chroma[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2]; 
  unsigned char chroma_data[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  
  int list = blk->list;
  int w = ref_mgr->get_stride();
  unsigned char *p = get_luma_ptr(blk, list, i, j);
  
  int scale = blk->scale[list] + power2(SAMPLE_ADD_PREC);
  int offset = blk->DC[list] * param.DC_mult << SAMPLE_ADD_PREC;
  
  for (k = 0; k < MOTION_BLK_SIZE; k++)
  {
    for (l = 0; l < MOTION_BLK_SIZE; l++)
    {
      int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
      int pred_val = p[l] * scale + offset;
      diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
    }
    
    p += w;
  }
  
  int sad = distortion_luma(diff_luma);
  if (sad >= maxval)
    return maxval;
  
  p = get_chroma_ptr(blk, list, i, j, (param.chroma_on_fly) ? chroma_data : NULL);
  w = (param.chroma_on_fly) ? MOTION_BLK_SIZE : w;
  offset = -MID_SAMPLE_VAL * blk->scale[list];
  
  for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
  {
    for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
    {
      for (n = 0; n < 2; n++) 
      {
        int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
        int pred_val = p[2*l + n] * scale + offset;
        diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
      }
    }
    
    p += w;
  }
  
  sad += transformC(diff_chroma);
  
  return sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (bipred, no intensity compensation)
*
* \param[in] i Vertical position (8x8 block units)
* \param[in] j Horizontal position (8x8 block units)
* \param[in] maxval Threshold for early exit
* \param[in] blk Motion data
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::bipred_sad(int i, int j, int maxval, PredBlk *blk) 
{
  int k, l, n;
  int diff_luma[MOTION_BLK_SIZE * MOTION_BLK_SIZE]; 
  int diff_chroma[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2]; 
  unsigned char chroma_data[2][MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  
  int w = ref_mgr->get_stride();
  unsigned char *p_l0 = get_luma_ptr(blk, LIST_0, i, j);
  unsigned char *p_l1 = get_luma_ptr(blk, LIST_1, i, j);
  
  for (k = 0; k < MOTION_BLK_SIZE; k++)
  {
    for (l = 0; l < MOTION_BLK_SIZE; l++)
    {
      int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
      int pred_val = (p_l0[l] + p_l1[l]) << (SAMPLE_ADD_PREC - 1);
      diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
    }
    
    p_l0 += w;
    p_l1 += w;
  }
  
  int sad = distortion_luma(diff_luma);
  if (sad >= maxval)
    return maxval;
  
  p_l0 = get_chroma_ptr(blk, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
  p_l1 = get_chroma_ptr(blk, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
  w = (param.chroma_on_fly) ? MOTION_BLK_SIZE : w;
  
  for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
  {
    for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
    {
      for (n = 0; n < 2; n++) 
      {
        int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
        int pred_val = (p_l0[2*l + n] + p_l1[2*l + n]) << (SAMPLE_ADD_PREC - 1);
        diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
      }
    }
    
    p_l0 += w;
    p_l1 += w;
  }
  
  sad += transformC(diff_chroma);
  
  return sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (bipred, with intensity compensation)
*
* \param[in] i Vertical position (8x8 block units)
* \param[in] j Horizontal position (8x8 block units)
* \param[in] maxval Threshold for early exit
* \param[in] blk Motion data
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::bipred_sad_wp(int i, int j, int maxval, PredBlk *blk) 
{
  int k, l, n;
  int diff_luma[MOTION_BLK_SIZE * MOTION_BLK_SIZE]; 
  int diff_chroma[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2]; 
  unsigned char chroma_data[2][MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  
  int w = ref_mgr->get_stride();
  unsigned char *p_l0 = get_luma_ptr(blk, LIST_0, i, j);
  unsigned char *p_l1 = get_luma_ptr(blk, LIST_1, i, j);
  
  int scale_l0 = (blk->get_B_scale_sum() >> 1) + blk->get_B_scale_diff();
  int scale_l1 = blk->get_B_scale_sum() - scale_l0;
  int offset = blk->BP_DC * param.DC_mult << SAMPLE_ADD_PREC;
  
  for (k = 0; k < MOTION_BLK_SIZE; k++)
  {
    for (l = 0; l < MOTION_BLK_SIZE; l++)
    {
      int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
      int pred_val = scale_l0 * p_l0[l] + scale_l1 * p_l1[l] + offset;
      diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
    }
    
    p_l0 += w;
    p_l1 += w;
  }
  
  int sad = distortion_luma(diff_luma);
  if (sad >= maxval)
    return maxval;
  
  p_l0 = get_chroma_ptr(blk, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
  p_l1 = get_chroma_ptr(blk, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
  w = (param.chroma_on_fly) ? MOTION_BLK_SIZE : w;
  
  offset = -MID_SAMPLE_VAL * blk->BP_scale[0];
  
  for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
  {
    for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
    {
      for (n = 0; n < 2; n++) 
      {
        int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
        int pred_val = scale_l0 * p_l0[2*l + n] + scale_l1 * p_l1[2*l + n] + offset;
        diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
      }
    }
    
    p_l0 += w;
    p_l1 += w;
  }
  
  sad += transformC(diff_chroma);
  
  return sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD for inter 8x8 with boundary block shifting
*
* Tests for the use of intensity compensation and calls
* dedicated SAD function. Exits early if partial SAD exceeds maxsad.
*
* \param[in] i Vertical position of block (8x8 block units)
* \param[in] j Horizontal position of block (8x8 block units)
* \param[in] cur Motion data 
* \param[in] top Motion data from block above
* \param[in] left Motion data from block to the left
* \param[in] BiPred flag to indicate whether frame type is ft_B_frame
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::distortion_mvshare(int i,int j, PredBlk *cur, PredBlk *top, PredBlk *left, int BiPred) 
{

  if (cur->intra != NULL)
    return cur->intra->sad;
  else if (BiPred==TRUE) 
  {
    if (param.enable_intcomp)
      return bipred_sad_wp_mvshare(i, j, cur, top, left);
    else
      return bipred_sad_mvshare(i, j, cur, top, left);
  }
  else 
  {
    if (param.enable_intcomp)
      return sad_wp_mvshare(i, j, cur, top, left);
    else
      return sad_mvshare(i, j, cur, top, left);
  }

}


//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (single-list, no intensity compensation) for motion sharing
*
* \param[in] i Vertical position (8x8 block units)
* \param[in] j Horizontal position (8x8 block units)
* \param[in] blk Motion data 
* \param[in] top Motion data from block above
* \param[in] left Motion data from block to the left
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::sad_mvshare(int i, int j, PredBlk *blk, PredBlk *top, PredBlk *left) 
{
  int shift0 = blk->mvshare_size[0];
  int shift1 = blk->mvshare_size[1];
  
  if (shift0 == 0 && shift1 == 0)
  {
    return sad(i, j, INT_INF, blk);
  }
  
  int diff_luma[MOTION_BLK_SIZE * MOTION_BLK_SIZE];
  int diff_chroma[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  unsigned char chroma_data[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  int k, l, n;
  
  int w = ref_mgr->get_stride();
  unsigned char *p;
  
  if (shift0 != 0)
  {
    p = get_luma_ptr(top, top->list, i, j);
    
    for (k = 0; k < shift0; k++)
    {
      for (l = 0; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = p[l] << SAMPLE_ADD_PREC;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p += w;
    }
    
    p = get_luma_ptr(blk, blk->list, i, j) + shift0 * w;

    for (k = shift0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = 0; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = p[l] << SAMPLE_ADD_PREC;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p += w;
    }    
  }
  else
  {
    p = get_luma_ptr(left, left->list, i, j);
    
    for (k = 0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = 0; l < shift1; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = p[l] << SAMPLE_ADD_PREC;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p += w;
    }
    
    p = get_luma_ptr(blk, blk->list, i, j);
    
    for (k = 0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = shift1; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = p[l] << SAMPLE_ADD_PREC;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p += w;
    }    
  }

  int sad = 0;
  sad += transform(diff_luma);
  sad += transform(diff_luma + 4 * MOTION_BLK_SIZE);
  
  // Chroma
  shift0 >>= 1;
  shift1 >>= 1;
  w = (param.chroma_on_fly) ? MOTION_BLK_SIZE : w;
  
  if (shift0 != 0)
  {
    p = get_chroma_ptr(top, top->list, i, j, (param.chroma_on_fly) ? chroma_data : NULL);
    
    for (k = 0; k < shift0; k++) 
    {
      for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = p[2*l + n] << SAMPLE_ADD_PREC;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p += w;
    }

    p = get_chroma_ptr(blk, blk->list, i, j, (param.chroma_on_fly) ? chroma_data : NULL) + shift0 * w;
    
    for (k = shift0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = p[2*l + n] << SAMPLE_ADD_PREC;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p += w;
    }
  }
  else
  {
    p = get_chroma_ptr(left, left->list, i, j, (param.chroma_on_fly) ? chroma_data : NULL);
    
    for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = 0; l < shift1; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = p[2*l + n] << SAMPLE_ADD_PREC;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p += w;
    }
    
    p = get_chroma_ptr(blk, blk->list, i, j, (param.chroma_on_fly) ? chroma_data : NULL);
    
    for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = shift1; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = p[2*l + n] << SAMPLE_ADD_PREC;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p += w;
    }
  }

  sad += transformC(diff_chroma);

  return sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (single-list, with intensity compensation)
*
* \param[in] i Vertical position (8x8 block units)
* \param[in] j Horizontal position (8x8 block units)
* \param[in] maxval Threshold for early exit
* \param[in] blk Motion data
* \param[in] top Motion data from block above
* \param[in] left Motion data from block to the left
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::sad_wp_mvshare(int i, int j, PredBlk *blk, PredBlk *top, PredBlk *left) 
{
  int shift0 = blk->mvshare_size[0];
  int shift1 = blk->mvshare_size[1];
  
  if (shift0 == 0 && shift1 == 0)
  {
    return sad_wp(i, j, INT_INF, blk);
  }
  
  int diff_luma[MOTION_BLK_SIZE * MOTION_BLK_SIZE];
  int diff_chroma[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  unsigned char chroma_data[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  int k, l, n;
  
  int w = ref_mgr->get_stride();
  unsigned char *p;
  int scale;
  int offset;
  
  if (shift0 != 0)
  {
    p = get_luma_ptr(top, top->list, i, j);
    scale = top->scale[top->list] + power2(SAMPLE_ADD_PREC);
    offset = top->DC[top->list] * param.DC_mult << SAMPLE_ADD_PREC;
    
    for (k = 0; k < shift0; k++)
    {
      for (l = 0; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = p[l] * scale + offset;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p += w;
    }
    
    p = get_luma_ptr(blk, blk->list, i, j) + shift0 * w;
    scale = blk->scale[blk->list] + power2(SAMPLE_ADD_PREC);
    offset = blk->DC[blk->list] * param.DC_mult << SAMPLE_ADD_PREC;

    for (k = shift0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = 0; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = p[l] * scale + offset;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p += w;
    }    
  }
  else
  {
    p = get_luma_ptr(left, left->list, i, j);
    scale = left->scale[left->list] + power2(SAMPLE_ADD_PREC);
    offset = left->DC[left->list] * param.DC_mult << SAMPLE_ADD_PREC;

    for (k = 0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = 0; l < shift1; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = p[l] * scale + offset;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p += w;
    }
    
    p = get_luma_ptr(blk, blk->list, i, j);
    scale = blk->scale[blk->list] + power2(SAMPLE_ADD_PREC);
    offset = blk->DC[blk->list] * param.DC_mult << SAMPLE_ADD_PREC;

    for (k = 0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = shift1; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = p[l] * scale + offset;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p += w;
    }    
  }
  
  int sad = 0;
  sad += transform(diff_luma);
  sad += transform(diff_luma + 4 * MOTION_BLK_SIZE);
  
  // Chroma
  shift0 >>= 1;
  shift1 >>= 1;
  w = (param.chroma_on_fly) ? MOTION_BLK_SIZE : w;
  
  if (shift0 != 0)
  {
    p = get_chroma_ptr(top, top->list, i, j, (param.chroma_on_fly) ? chroma_data : NULL);
    scale = top->scale[top->list] + power2(SAMPLE_ADD_PREC);
    offset = -MID_SAMPLE_VAL * top->scale[top->list];

    for (k = 0; k < shift0; k++) 
    {
      for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = p[2*l + n] * scale + offset;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p += w;
    }
    
    p = get_chroma_ptr(blk, blk->list, i, j, (param.chroma_on_fly) ? chroma_data : NULL) + shift0 * w;
    scale = blk->scale[blk->list] + power2(SAMPLE_ADD_PREC);
    offset = -MID_SAMPLE_VAL * blk->scale[blk->list];
    
    for (k = shift0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = p[2*l + n] * scale + offset;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p += w;
    }
  }
  else
  {
    p = get_chroma_ptr(left, left->list, i, j, (param.chroma_on_fly) ? chroma_data : NULL);
    scale = left->scale[left->list] + power2(SAMPLE_ADD_PREC);
    offset = -MID_SAMPLE_VAL * left->scale[left->list];
    
    for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = 0; l < shift1; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = p[2*l + n] * scale + offset;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p += w;
    }
    
    p = get_chroma_ptr(blk, blk->list, i, j, (param.chroma_on_fly) ? chroma_data : NULL);
    scale = blk->scale[blk->list] + power2(SAMPLE_ADD_PREC);
    offset = -MID_SAMPLE_VAL * blk->scale[blk->list];
    
    for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = shift1; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = p[2*l + n] * scale + offset;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p += w;
    }
  }
  
  sad += transformC(diff_chroma);
  
  return sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (bipred, no intensity compensation)
*
* \param[in] i Vertical position (8x8 block units)
* \param[in] j Horizontal position (8x8 block units)
* \param[in] maxval Threshold for early exit
* \param[in] blk Motion data
* \param[in] top Motion data from block above
* \param[in] left Motion data from block to the left
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::bipred_sad_mvshare(int i, int j, PredBlk *blk, PredBlk *top, PredBlk *left) 
{
  int shift0 = blk->mvshare_size[0];
  int shift1 = blk->mvshare_size[1];
  
  if (shift0 == 0 && shift1 == 0)
  {
    if (blk->list == BI_PRED)
      return bipred_sad(i, j, INT_INF, blk);
    else
      return sad(i, j, INT_INF, blk);
  }
  
  int diff_luma[MOTION_BLK_SIZE * MOTION_BLK_SIZE];
  int diff_chroma[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  unsigned char chroma_data[2][MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  int k, l, n;
  
  int w = ref_mgr->get_stride();
  unsigned char *p_l0;
  unsigned char *p_l1;
  
  if (shift0 != 0)
  {
    if (top->list == BI_PRED)
    {
      p_l0 = get_luma_ptr(top, LIST_0, i, j);
      p_l1 = get_luma_ptr(top, LIST_1, i, j);
    }
    else
    {
      p_l0 = get_luma_ptr(top, top->list, i, j);
      p_l1 = p_l0;
    }

    for (k = 0; k < shift0; k++)
    {
      for (l = 0; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = (p_l0[l] + p_l1[l]) << (SAMPLE_ADD_PREC - 1);
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p_l0 += w;
      p_l1 += w;
    }
    
    if (blk->list == BI_PRED)
    {
      p_l0 = get_luma_ptr(blk, LIST_0, i, j);
      p_l1 = get_luma_ptr(blk, LIST_1, i, j);
    }
    else
    {
      p_l0 = get_luma_ptr(blk, blk->list, i, j);
      p_l1 = p_l0;
    }
    
    p_l0 += shift0 * w;
    p_l1 += shift0 * w;
    
    for (k = shift0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = 0; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = (p_l0[l] + p_l1[l]) << (SAMPLE_ADD_PREC - 1);
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p_l0 += w;
      p_l1 += w;
    }    
  }
  else
  {
    if (left->list == BI_PRED)
    {
      p_l0 = get_luma_ptr(left, LIST_0, i, j);
      p_l1 = get_luma_ptr(left, LIST_1, i, j);
    }
    else
    {
      p_l0 = get_luma_ptr(left, left->list, i, j);
      p_l1 = p_l0;
    }
    
    for (k = 0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = 0; l < shift1; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = (p_l0[l] + p_l1[l]) << (SAMPLE_ADD_PREC - 1);
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p_l0 += w;
      p_l1 += w;
    }
    
    if (blk->list == BI_PRED)
    {
      p_l0 = get_luma_ptr(blk, LIST_0, i, j);
      p_l1 = get_luma_ptr(blk, LIST_1, i, j);
    }
    else
    {
      p_l0 = get_luma_ptr(blk, blk->list, i, j);
      p_l1 = p_l0;
    }
    
    for (k = 0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = shift1; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = (p_l0[l] + p_l1[l]) << (SAMPLE_ADD_PREC - 1);
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p_l0 += w;
      p_l1 += w;
    }    
  }
  
  int sad = 0;
  sad += transform(diff_luma);
  sad += transform(diff_luma + 4 * MOTION_BLK_SIZE);
  
  // Chroma
  shift0 >>= 1;
  shift1 >>= 1;
  w = (param.chroma_on_fly) ? MOTION_BLK_SIZE : w;
  
  if (shift0 != 0)
  {
    if (top->list == BI_PRED)
    {
      p_l0 = get_chroma_ptr(top, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = get_chroma_ptr(top, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
    }
    else
    {
      p_l0 = get_chroma_ptr(top, top->list, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = p_l0;
    }

    for (k = 0; k < shift0; k++) 
    {
      for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = (p_l0[2*l + n] + p_l1[2*l + n]) << (SAMPLE_ADD_PREC - 1);
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p_l0 += w;
      p_l1 += w;
    }
    
    if (blk->list == BI_PRED)
    {
      p_l0 = get_chroma_ptr(blk, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = get_chroma_ptr(blk, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
    }
    else
    {
      p_l0 = get_chroma_ptr(blk, blk->list, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = p_l0;
    }

    p_l0 += w * shift0;
    p_l1 += w * shift0;
    
    for (k = shift0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = (p_l0[2*l + n] + p_l1[2*l + n]) << (SAMPLE_ADD_PREC - 1);
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p_l0 += w;
      p_l1 += w;
    }
  }
  else
  {
    if (left->list == BI_PRED)
    {
      p_l0 = get_chroma_ptr(left, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = get_chroma_ptr(left, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
    }
    else
    {
      p_l0 = get_chroma_ptr(left, left->list, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = p_l0;
    }
    
    for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = 0; l < shift1; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = (p_l0[2*l + n] + p_l1[2*l + n]) << (SAMPLE_ADD_PREC - 1);
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p_l0 += w;
      p_l1 += w;
    }
    
    if (blk->list == BI_PRED)
    {
      p_l0 = get_chroma_ptr(blk, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = get_chroma_ptr(blk, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
    }
    else
    {
      p_l0 = get_chroma_ptr(blk, blk->list, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = p_l0;
    }
    
    for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = shift1; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = (p_l0[2*l + n] + p_l1[2*l + n]) << (SAMPLE_ADD_PREC - 1);
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p_l0 += w;
      p_l1 += w;
    }
  }
  
  sad += transformC(diff_chroma);
  
  return sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute SAD (bipred, with intensity compensation)
*
* \param[in] i Vertical position (8x8 block units)
* \param[in] j Horizontal position (8x8 block units)
* \param[in] maxval Threshold for early exit
* \param[in] blk Motion data
* \param[in] top Motion data from block above
* \param[in] left Motion data from block to the left
* \return SAD
*/
//-----------------------------------------------------------------------------
int MCRefBlk8::bipred_sad_wp_mvshare(int i, int j, PredBlk *blk, PredBlk *top, PredBlk *left) 
{
  int shift0 = blk->mvshare_size[0];
  int shift1 = blk->mvshare_size[1];
  
  if (shift0 == 0 && shift1 == 0)
  {
    if (blk->list == BI_PRED)
      return bipred_sad_wp(i, j, INT_INF, blk);
    else
      return sad_wp(i, j, INT_INF, blk);
  }
  
  int diff_luma[MOTION_BLK_SIZE * MOTION_BLK_SIZE];
  int diff_chroma[MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  unsigned char chroma_data[2][MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  int k, l, n;
  
  int w = ref_mgr->get_stride();
  unsigned char *p_l0;
  unsigned char *p_l1;
  int scale_l0, scale_l1, offset;
  
  if (shift0 != 0)
  {
    if (top->list == BI_PRED)
    {
      p_l0 = get_luma_ptr(top, LIST_0, i, j);
      p_l1 = get_luma_ptr(top, LIST_1, i, j);
      scale_l0 = (top->get_B_scale_sum() >> 1) + top->get_B_scale_diff();
      scale_l1 = top->get_B_scale_sum() - scale_l0;
      offset = top->BP_DC * param.DC_mult << SAMPLE_ADD_PREC;
    }
    else
    {
      p_l0 = get_luma_ptr(top, top->list, i, j);
      p_l1 = p_l0;
      scale_l0 = top->scale[top->list] + power2(SAMPLE_ADD_PREC);
      scale_l1 = 0;
      offset = top->DC[top->list] * param.DC_mult << SAMPLE_ADD_PREC;
    }
    
    for (k = 0; k < shift0; k++)
    {
      for (l = 0; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = scale_l0 * p_l0[l] + scale_l1 * p_l1[l] + offset;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p_l0 += w;
      p_l1 += w;
    }
    
    if (blk->list == BI_PRED)
    {
      p_l0 = get_luma_ptr(blk, LIST_0, i, j);
      p_l1 = get_luma_ptr(blk, LIST_1, i, j);
      scale_l0 = (blk->get_B_scale_sum() >> 1) + blk->get_B_scale_diff();
      scale_l1 = blk->get_B_scale_sum() - scale_l0;
      offset = blk->BP_DC * param.DC_mult << SAMPLE_ADD_PREC;
    }
    else
    {
      p_l0 = get_luma_ptr(blk, blk->list, i, j);
      p_l1 = p_l0;
      scale_l0 = blk->scale[blk->list] + power2(SAMPLE_ADD_PREC);
      scale_l1 = 0;
      offset = blk->DC[blk->list] * param.DC_mult << SAMPLE_ADD_PREC;
    }
    
    p_l0 += shift0 * w;
    p_l1 += shift0 * w;
    
    for (k = shift0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = 0; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = scale_l0 * p_l0[l] + scale_l1 * p_l1[l] + offset;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p_l0 += w;
      p_l1 += w;
    }    
  }
  else
  {
    if (left->list == BI_PRED)
    {
      p_l0 = get_luma_ptr(left, LIST_0, i, j);
      p_l1 = get_luma_ptr(left, LIST_1, i, j);
      scale_l0 = (left->get_B_scale_sum() >> 1) + left->get_B_scale_diff();
      scale_l1 = left->get_B_scale_sum() - scale_l0;
      offset = left->BP_DC * param.DC_mult << SAMPLE_ADD_PREC;
    }
    else
    {
      p_l0 = get_luma_ptr(left, left->list, i, j);
      p_l1 = p_l0;
      scale_l0 = left->scale[left->list] + power2(SAMPLE_ADD_PREC);
      scale_l1 = 0;
      offset = left->DC[left->list] * param.DC_mult << SAMPLE_ADD_PREC;
    }
    
    for (k = 0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = 0; l < shift1; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = scale_l0 * p_l0[l] + scale_l1 * p_l1[l] + offset;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p_l0 += w;
      p_l1 += w;
    }
    
    if (blk->list == BI_PRED)
    {
      p_l0 = get_luma_ptr(blk, LIST_0, i, j);
      p_l1 = get_luma_ptr(blk, LIST_1, i, j);
      scale_l0 = (blk->get_B_scale_sum() >> 1) + blk->get_B_scale_diff();
      scale_l1 = blk->get_B_scale_sum() - scale_l0;
      offset = blk->BP_DC * param.DC_mult << SAMPLE_ADD_PREC;
    }
    else
    {
      p_l0 = get_luma_ptr(blk, blk->list, i, j);
      p_l1 = p_l0;
      scale_l0 = blk->scale[blk->list] + power2(SAMPLE_ADD_PREC);
      scale_l1 = 0;
      offset = blk->DC[blk->list] * param.DC_mult << SAMPLE_ADD_PREC;
    }
    
    for (k = 0; k < MOTION_BLK_SIZE; k++)
    {
      for (l = shift1; l < MOTION_BLK_SIZE; l++)
      {
        int target_val = val[MOTION_BLK_SIZE * k + l] << SAMPLE_ADD_PREC;
        int pred_val = scale_l0 * p_l0[l] + scale_l1 * p_l1[l] + offset;
        diff_luma[MOTION_BLK_SIZE * k + l] = pred_val - target_val;
      }
      
      p_l0 += w;
      p_l1 += w;
    }    
  }
  
  int sad = 0;
  sad += transform(diff_luma);
  sad += transform(diff_luma + 4 * MOTION_BLK_SIZE);
  
  // Chroma
  shift0 >>= 1;
  shift1 >>= 1;
  w = (param.chroma_on_fly) ? MOTION_BLK_SIZE : w;
  
  if (shift0 != 0)
  {
    if (top->list == BI_PRED)
    {
      p_l0 = get_chroma_ptr(top, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = get_chroma_ptr(top, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
      scale_l0 = (top->get_B_scale_sum() >> 1) + top->get_B_scale_diff();
      scale_l1 = top->get_B_scale_sum() - scale_l0;
      offset = -MID_SAMPLE_VAL * top->BP_scale[0];
    }
    else
    {
      p_l0 = get_chroma_ptr(top, top->list, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = p_l0;
      scale_l0 = top->scale[top->list] + power2(SAMPLE_ADD_PREC);
      scale_l1 = 0;
      offset = -MID_SAMPLE_VAL * top->scale[top->list];
    }
    
    for (k = 0; k < shift0; k++) 
    {
      for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = scale_l0 * p_l0[2*l + n] + scale_l1 * p_l1[2*l + n] + offset;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p_l0 += w;
      p_l1 += w;
    }
    
    if (blk->list == BI_PRED)
    {
      p_l0 = get_chroma_ptr(blk, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = get_chroma_ptr(blk, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
      scale_l0 = (blk->get_B_scale_sum() >> 1) + blk->get_B_scale_diff();
      scale_l1 = blk->get_B_scale_sum() - scale_l0;
      offset = -MID_SAMPLE_VAL * blk->BP_scale[0];
    }
    else
    {
      p_l0 = get_chroma_ptr(blk, blk->list, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = p_l0;
      scale_l0 = blk->scale[blk->list] + power2(SAMPLE_ADD_PREC);
      scale_l1 = 0;
      offset = -MID_SAMPLE_VAL * blk->scale[blk->list];
    }
    
    p_l0 += w * shift0;
    p_l1 += w * shift0;
    
    for (k = shift0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = 0; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = scale_l0 * p_l0[2*l + n] + scale_l1 * p_l1[2*l + n] + offset;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p_l0 += w;
      p_l1 += w;
    }
  }
  else
  {
    if (left->list == BI_PRED)
    {
      p_l0 = get_chroma_ptr(left, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = get_chroma_ptr(left, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
      scale_l0 = (left->get_B_scale_sum() >> 1) + left->get_B_scale_diff();
      scale_l1 = left->get_B_scale_sum() - scale_l0;
      offset = -MID_SAMPLE_VAL * left->BP_scale[0];
    }
    else
    {
      p_l0 = get_chroma_ptr(left, left->list, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = p_l0;
      scale_l0 = left->scale[left->list] + power2(SAMPLE_ADD_PREC);
      scale_l1 = 0;
      offset = -MID_SAMPLE_VAL * left->scale[left->list];
    }
    
    for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = 0; l < shift1; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = scale_l0 * p_l0[2*l + n] + scale_l1 * p_l1[2*l + n] + offset;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p_l0 += w;
      p_l1 += w;
    }
    
    if (blk->list == BI_PRED)
    {
      p_l0 = get_chroma_ptr(blk, LIST_0, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = get_chroma_ptr(blk, LIST_1, i, j, (param.chroma_on_fly) ? chroma_data[LIST_1] : NULL);
      scale_l0 = (blk->get_B_scale_sum() >> 1) + blk->get_B_scale_diff();
      scale_l1 = blk->get_B_scale_sum() - scale_l0;
      offset = -MID_SAMPLE_VAL * blk->BP_scale[0];
    }
    else
    {
      p_l0 = get_chroma_ptr(blk, blk->list, i, j, (param.chroma_on_fly) ? chroma_data[LIST_0] : NULL);
      p_l1 = p_l0;
      scale_l0 = blk->scale[blk->list] + power2(SAMPLE_ADD_PREC);
      scale_l1 = 0;
      offset = -MID_SAMPLE_VAL * blk->scale[blk->list];
    }
    
    for (k = 0; k < MOTION_BLK_SIZE / 2; k++) 
    {
      for (l = shift1; l < MOTION_BLK_SIZE / 2; l++) 
      {
        for (n = 0; n < 2; n++) 
        {
          int target_val = val[64 + MOTION_BLK_SIZE * k + 2*l + n] << SAMPLE_ADD_PREC;
          int pred_val = scale_l0 * p_l0[2*l + n] + scale_l1 * p_l1[2*l + n] + offset;
          diff_chroma[MOTION_BLK_SIZE * k + l + 4*n] = pred_val - target_val;
        }
      }
      
      p_l0 += w;
      p_l1 += w;
    }
  }
  
  sad += transformC(diff_chroma);
  
  return sad;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
