/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
 * \file MainEncoder.cpp
 *
 * \brief Main encoding program
 *
 */
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"
#include "Buffer.H"
#include "Param.H"
#include "Frame.H"
#include "Encoder.H"
#include "Quantizer.H"
#include "ExplicitGop.H"
#include "FastDCT.H"
#include "AIF_adaptive_filter.H"

#ifdef TRACE_DEBUG
#include "Trace.H"
#endif
//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

static int help_mode   = FALSE; //!< Flag for help mode
static int silent_mode = FALSE; //!< Flag for silent mode
static int mse_mode    = FALSE; //!< Flag for using MSE instead of PSNR
int aif_loop[LISTS][MAX_REF_FRAMES];//!< Variable used to determine whether AIF interpolated data is used for prediction

#ifdef IDR_EXTRACT
FILE *idr_extract_fptr;
int total_idr_bytes;
#endif

//-----------------------------------------------------------------------------
// Local functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
 * \brief Set the value of quantization set id based on coding structure and frame no
 */
//-----------------------------------------------------------------------------
void set_quantization_set_id(int frame_no, int intermed_idx) // cq_start, cq_end
{
  if (param.pyramid_coding == 0 || param.intermed_frame == 0 || param.enableB == 0)
  {
    if ((param.onlyI==0? frame_no : frame_no%param.onlyI)==0)
      param.quantization_set_id = 1; //!< I
    else
      param.quantization_set_id = 2; //!< P
  }
  else
  {
    if ((param.onlyI==0? frame_no : frame_no%param.onlyI)==0)
      param.quantization_set_id = 1; //!< I
    else if (intermed_idx == 0)
      param.quantization_set_id = 2; //!< P
    else
    {
      if (intermed_idx == 1)
        param.quantization_set_id = 3; //!< Br
      else
        param.quantization_set_id = 4; //!< B
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Open file
 *
 * Opens a file. If filename is empty, a NULL pointer is returned. If opening
 * the file fails, the program will exit.
 *
 * \param[in] filename File name
 * \param[in] mode Mode in which to open file (second argument to fopen)
 * \return File pointer
 */
//-----------------------------------------------------------------------------
FILE *open_file(const char *filename, const char *mode)
{
  if (filename == NULL || filename[0] == 0)
  {
    return NULL; // if empty filename, simply return NULL
  }
  
  FILE *f = fopen(filename, mode);
  if (f == NULL) 
  {
    fprintf(stderr, "Could not open file [%s].\n", filename);
    exit(EXIT_FAILURE);
  }
  
  return f;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Close file
 *
 * \param[in,out] f File pointer
 */
//-----------------------------------------------------------------------------
void close_file(FILE *f)
{
  if (f != NULL)
    fclose(f);
}

//-----------------------------------------------------------------------------
/*!
 * \brief Open file
 *
 * Opens a file. If filename is empty, a NULL pointer is returned. If opening
 * the file fails, the program will exit.
 *
 * \param[in] filename File name
 * \param[in] mode Mode in which to open file (second argument to open)
 * \return File descriptor
 */
//-----------------------------------------------------------------------------
int open_file(const char *filename, int mode)
{
  if (filename == NULL || filename[0] == 0)
    return -1; // if empty filename, simply return -1

  int fd = open(filename, mode);
  if (fd == -1) 
  {
    fprintf(stderr, "Could not open file [%s] with mode=%d.\n", filename, mode);
    exit(EXIT_FAILURE);
  }
  
  return fd;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Open file
 *
 * Opens a file. If filename is empty, a NULL pointer is returned. If opening
 * the file fails, the program will exit.
 *
 * \param[in] filename File name
 * \param[in] mode Mode in which to open file (second argument to open)
 * \param[in] permission Permission in which to open file (third argument to open)
 * \return File descriptor
 */
//-----------------------------------------------------------------------------
int open_file(const char *filename, int mode, int permission)
{
  if (filename == NULL || filename[0] == 0)
    return -1; // if empty filename, simply return -1
  
  int fd = open(filename, mode, permission);
  if (fd == -1) 
  {
    fprintf(stderr, "Could not open file [%s] with mode=%d and permission=%d.\n", filename, mode, permission);
    exit(EXIT_FAILURE);
  }
  
  return fd;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Close file
 *
 * \param[in,out] fd File pointer
 */
//-----------------------------------------------------------------------------
void close_file(int fd)
{
  if (fd != -1)
    close(fd);
}

//-----------------------------------------------------------------------------
/*!
 * \brief Write a frame to a file using lseek 
 *
 * \param[in,out] f File descriptor
 * \param[in] fr %Frame to write
 */
//-----------------------------------------------------------------------------
void write_frame(int fd, Frame *fr)
{
  if (fd != -1) 
  {
    lseek(fd, (int64) fr->frame_no * fr->size, SEEK_SET);
    write(fd, fr->data, fr->size);
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Read a frame from a file
 *
 * \param[in,out] f File pointer
 * \param[out] fr %Frame to read
 * \param[in] frame_number Position of frame in file (in frame units)
 * \param[in] f_header Size in bytes of header in file
 * \param[in] frame_skip Offset in frame numbering (in frame units)
 */
//-----------------------------------------------------------------------------
void read_one_frame(int fd, Frame *fr, int frame_number, int f_header, int frame_skip)
{
  
  if (fd != -1) 
  {
    if (lseek(fd, (int64) f_header, SEEK_SET) != f_header)
    {
      printf ("read_one_frame: cannot lseek to (Header size) in input file\n");
      exit(EXIT_FAILURE);
    }
    if (lseek(fd, (int64) fr->size * frame_skip, SEEK_CUR) == -1)
    {
      printf ("read_one_frame: cannot lseek to frame %d in input file\n",frame_skip);
      exit(EXIT_FAILURE);
    }
    
    if (lseek(fd, (int64) fr->size * frame_number, SEEK_CUR) == -1)
    {
      printf ("read_one_frame: Maximum frame number position (%d) reached. Can't advance further\n",frame_number);
      exit(EXIT_FAILURE);
    }
    
    if (read(fd, fr->data, fr->size) != fr->size)
    { 
      printf ("read_one_frame: Cannot read frame data: %d\n",frame_number);
      exit(EXIT_FAILURE);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Encode a frame
 *
 * \param[in,out] enc Pointer to encoder
 * \param[in] fr %Frame to encode
 * \param[out] buf Byte buffer containing coded frame
 * \param[in,out] f_par Pointer to file containing lambda and qp parameters for individual frames (read)
 * \param[in,out] f_dst File pointer for coded sequence - bitstream (write)
 * \param[in,out] fd_rec File descriptor for coded sequence - yuv (write)
 * \param[in,out] f_sad Pointer to file containing log of SAD values for each frame (write)
 * \param[in,out] f_log_frame Pointer to file containing log of bitrate and PSNR values for each frame (write)
 */
//-----------------------------------------------------------------------------
void encode_one_frame(Encoder  *enc,
                      Frame    *fr,
                      Frame    *fr_pred,
                      Frame    *fr_qpe_m,
                      Frame    *fr_qpe_s,
                      Frame    *fr_orig,
                      Frame    *fr_filtered,
                      Buffer   *buf,
                      FILE     *f_par,
                      FILE     *f_dst,
                      int      fd_rec,
                      FILE     *f_sad,
                      FILE     *f_log_frame)
{
  clock_t clk;
  
  if (f_par != NULL) 
  {
    fscanf(f_par, "%d %d %d", &param.qp, &param.lambdaq_base, &param.melq_base);
    param.update();
  }
  
  // encode frame and get time
  clk = clock();

  enc->encode_frame(fr, fr_pred, fr_qpe_m, fr_qpe_s, fr_orig, fr_filtered, buf);
  
  if (0!=param.wienerfilter)
  {
    enc->compute_distortion_filtered_against_orig(fr, fr_orig, fr_filtered);
  }
  else if (param.deblockingfilter)
  {
    enc->compute_distortion_filtered_against_orig(fr, fr_orig, fr);    
  }
  
  
  clk = clock() - clk;
  
  // write coded data to file
  if (f_dst != NULL)
  {
#ifdef IDR_EXTRACT
    if (fr->frame_no==0)
    {
      fprintf(idr_extract_fptr,"1st Intra Frame %d\n",total_idr_bytes+sizeof(buf->payload_type)+sizeof(buf->payload_size) +buf->p_byte + ((buf->p_bit + 7) >> 3));
    }
    else
    {
      fprintf(idr_extract_fptr,"%d %d %d %d\n",fr->frame_type,fr->intra_frame_type,fr->frame_no,total_idr_bytes);
    }
    total_idr_bytes += sizeof(buf->payload_type)+sizeof(buf->payload_size) +buf->p_byte + ((buf->p_bit + 7) >> 3);
#endif
    buf->payload_type = Buffer::PLT_FRAME;
    buf->write(f_dst);
    fflush(f_dst);
  }
  
  // write frames to files
  if (param.wienerfilter!=0)
  {
    fr_filtered->frame_no = fr->frame_no;
    write_frame(fd_rec,      fr_filtered);
  }
  else
  {
    write_frame(fd_rec,      fr);    
  }
  
  if (f_sad && fr->frame_no != 0)
    fprintf(f_sad, "%d %d\n", enc->mlc_cur.motion_bits, enc->sad);
  
  // print stats for current frame
  if (silent_mode == FALSE) 
  {
    printf("%05d", fr->frame_no);
    switch (fr->frame_type)
    {
      case ft_I_frame:
        if (fr->intra_frame_type==ft_I_IDR)
        { // IDR frame
          printf("(R)");
          break;
        }
        else if (fr->intra_frame_type==ft_I_DDR)
        { // DDR frame
          printf("(D)");
          break;
        }
        if (fr->reference_idc)
          printf("(i)");
        else
          printf("(I)");
        break;
      case ft_P_frame:
        if (fr->reference_idc)
          printf("(p)");
        else
          printf("(P)");
        break;
      case ft_B_frame:
        if (fr->reference_idc)
          printf("(b)");
        else
          printf("(B)");
        break;
      default:
        break;
    }
    
    printf("  %08d ", enc->bits);
    printf("%03d ",fr->qp);
    printf("%07.3f ", fr->snr_luma);
    printf("%07.3f ", fr->snr_chroma[0]);
    printf("%07.3f ", fr->snr_chroma[1]);
    printf("%07.3f ", 1.0 * clk / CLOCKS_PER_SEC);
    printf("%04d %04d %04d ", enc->mlc_cur.zblks, enc->mlc_cur.nzblks, enc->mlc_cur.iblks);
    printf("%06d %06d %06d ", enc->mlc_cur.motion_bits, enc->mlc_cur.luma_bits, enc->mlc_cur.chroma_bits);
    printf("%06d %06d ", enc->mlc_cur.bit_count_luma, enc->mlc_cur.bit_count_chroma);
    
    printf("%05d %05d %05d %05d ",enc->mlc_cur.intra_tma_L9, enc->mlc_cur.intra_tma_L5, enc->mlc_cur.intra_tma_L9c,
           enc->mlc_cur.intra_tma_L5c);
    
    if(param.motion_sharing_log)
    {
      if(fr->frame_type!=ft_I_frame)
        printf("%05d %05d %05d %05d %07.3f ",enc->mlc_cur.mvshare_off, enc->mlc_cur.mvshare_zero, enc->mlc_cur.mvshare_two, 
               enc->mlc_cur.mvshare_four,
               (float)(enc->mlc_cur.mvshare_two+enc->mlc_cur.mvshare_four)/(enc->mlc_cur.mvshare_off+enc->mlc_cur.mvshare_zero+enc->mlc_cur.mvshare_two+enc->mlc_cur.mvshare_four)*100);
      else
        printf("%05d %05d %05d %05d %07.3f ", 0, 0, 0, 0, 0.0);
    }
    if(param.aif)
    {
      printf("%06d ",enc->mlc_cur.aif_bits);
    }
    
    printf("%04d ",fr->melq);
    printf("%04d ",fr->lambdaq);
    printf("%03d ",fr->list_size[LIST_0]);
    printf("%03d",fr->list_size[LIST_1]);
    printf("\n");
    fflush(stdout);
  }
  
#ifdef INTRA_MODE_STATS
  if (param.intra16x16 && (fr->frame_type == ft_I_frame ))
  {
    enc->mlc_cur.print_intra_mode_stats(SIP8x8, &enc->mlc_tot, enc->coded_frames-enc->coded_inter_frames);
    enc->mlc_cur.print_intra_mode_stats(SIP16x16, &enc->mlc_tot, enc->coded_frames-enc->coded_inter_frames);
    enc->mlc_cur.print_intra_mode_stats(SIP4x4, &enc->mlc_tot, enc->coded_frames-enc->coded_inter_frames);
    enc->mlc_cur.print_intra_mode_stats(SIP2x8, &enc->mlc_tot, enc->coded_frames-enc->coded_inter_frames);
    enc->mlc_cur.print_intra_mode_stats(SIP8x2, &enc->mlc_tot, enc->coded_frames-enc->coded_inter_frames);
  }
#endif //INTRA_MODE_STATS
  fprintf(f_log_frame, "%d %d %d %d %f\n", fr->qp, fr->lambdaq, fr->melq, enc->bits, fr->snr_luma);
}

//-----------------------------------------------------------------------------
// Main function
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
 * \brief Main encoding function
 *
 * \param[in] argc Argument count
 * \param[in] argv Array of argument values
 * \return EXIT_SUCCESS or EXIT_FAILURE
 */
//-----------------------------------------------------------------------------
int main(int argc, char **argv) 
{
  Frame *fr;         // frame memory
  Frame *fr_orig;         // frame memory
  Frame *fr_filtered;     // frame memory
  Frame *fr_pred;         // frame memory
  Frame *fr_qpe_m;        // frame memory
  Frame *fr_qpe_s;        // frame memory
  Encoder *enc;      // encoder
  Buffer *buf;       // byte buffer
  int fd_src = -1;         // source file descriptor
  int fd_rec = -1;         // reconstructed file descriptor
  int fd_ref = -1;         // reference file descriptor
  FILE *f_dst = NULL;       // destination file
  FILE *f_sad = NULL;       // SAD file
  FILE *f_par = NULL;       // parameter file
  FILE *f = NULL;           // misc file
  FILE *f_log_frame = NULL; // per frame log file
  
  ExplicitGop *std_pyramid; // normal pyramid structure
  ExplicitGop *idr_pyramid; // pyramid struct used with IDR/DDR-led intermed frames.
  int frame; // frame counter
  int last_frame_no = 0;
  int last_qp = 0;
  double snr_luma, snr_chroma[2];
  double mse_luma, mse_chroma[2];
  int total_time;
  clock_t clk;
  int a;
  int current_frame_to_code;
  
  // start here
  for (a = 1; a < argc; a++) 
  {
    if (argv[a][ZERO] == '-') 
    {
      if (strcmp(argv[a], "-h") == ZERO)
        help_mode = TRUE;
      else if (strcmp(argv[a], "-s") == ZERO)
        silent_mode = TRUE;
      else if (strcmp(argv[a], "-m") == ZERO)
        mse_mode = TRUE;
      else
        help_mode = TRUE;
    }
    else
      break;
  }
  
  if (a != argc - 1)
    help_mode = TRUE;
  
  if (help_mode == TRUE)
    silent_mode = FALSE;
  
  if (silent_mode == FALSE)
    printf("CDCM video encoder\n");
  
  if (help_mode == TRUE) 
  {
    printf("Usage: %s [options] <parameter file>\n", argv[ZERO]);
    printf("Options:\n");
    printf("-h : help mode\n");
    printf("-s : silent mode\n");
    printf("-m : compute average PSNR based on average MSE\n");
    exit(EXIT_FAILURE);
  }
 
  // read configuration file
  param.read_config_file(argv[a]);
  
  param.intra16x16_org = param.intra16x16;
  
  param.ini_qmatrix(0); // cq_start
  param.ini_qmatrix(1); // cq_end
  
  param.update();
  param.read_prob_file();
  
  if (param.qmatrix_enable_flag[0]) // cq_start
    param.read_qmatrix_file(param.qmatrixfile[0], 0);
  if (param.qmatrix_enable_flag[1])
    param.read_qmatrix_file(param.qmatrixfile[1], 1); // cq_end
  
  // intermediate frames counter increase and frame skipping    
  // make sure that all frames are coded
  int max_intermed_frames = imax(param.nframes - 2, 0);
  param.intermed_frame     = imin(param.intermed_frame,     max_intermed_frames);
  param.idr_intermed_frame = imin(param.idr_intermed_frame, max_intermed_frames);
  
  int idr_refresh;
  int intermed_frame = param.intermed_frame;
  int target_last_non_idr_frame;
  int is_idr_gop = 0;
  
  if (param.onlyI && param.idr_interval)
  {
    idr_refresh = param.onlyI * param.idr_interval;
    if (idr_refresh <= param.intermed_frame)
    {
      printf("Number of frames between 2 IDR pictures (onlyI * idr_interval) should be larger than intermed_frame.\n");
      exit(EXIT_FAILURE);
    }
    param.idr_intermed_frame = (idr_refresh - 1) % (param.intermed_frame + 1);
    target_last_non_idr_frame = (idr_refresh - 1) / (param.intermed_frame + 1) * (param.intermed_frame + 1);
  }
  else
  {
    idr_refresh = 0;
    param.idr_intermed_frame = 0;
    target_last_non_idr_frame = -1;
  }
#ifdef VERBOSE_LOG
  printf (
          " idr_refreshing_frame_int=%d\n"
          " intermed_frame=%d\n"
          " idr_intermed_frame=%d\n", 
          param.onlyI*param.idr_interval, 
          param.intermed_frame, 
          param.idr_intermed_frame);
#endif
  
  std_pyramid = new ExplicitGop(param.intermed_frame);
  idr_pyramid = new ExplicitGop(param.idr_intermed_frame);
  
  if(param.q_rdoffset && param.q_adaptoffset!=4)
  {
    printf ("q_adaptoffset should be set to 4 when q_rdoffset is set to 1.\n");
    exit(EXIT_FAILURE);
  }
  
  if (param.pyramid_coding >= EGOP_EXPLICIT)
  {
    std_pyramid->interpret_gop_structure(param.expPrmdFrmt);
    idr_pyramid->interpret_gop_structure(param.expPrmdFrmtIdr);
  }
  else
  {
    std_pyramid->create_pyramid();      
    idr_pyramid->create_pyramid();
  }
  
  // open files
  fd_src = open_file(param.srcfile, OPENFLAGS_READ);
  fd_rec = open_file(param.recfile, OPENFLAGS_WRITE, OPEN_PERMISSIONS);
  fd_ref = open_file(param.reffile, OPENFLAGS_READ);
  f_dst = open_file(param.dstfile, "wb");
  f_sad = open_file(param.sadfile, "wt");
  f_par = open_file(param.parfile, "rt");
  
  // allocate memory and initialize variables
  fr = new Frame(param.frame_height, param.frame_width);
  fr_pred  = new Frame(param.frame_height, param.frame_width);
  fr_qpe_m = new Frame(param.frame_height, param.frame_width);
  fr_qpe_s = new Frame(param.frame_height, param.frame_width);
  fr_orig     = new Frame(param.frame_height, param.frame_width);
  fr_filtered = new Frame(param.frame_height, param.frame_width);
  buf = new Buffer;
  enc = new Encoder(param.frame_height, param.frame_width);
  
  snr_luma = 0.0;
  snr_chroma[0] = 0.0;
  snr_chroma[1] = 0.0;
  mse_luma = 0.0;
  mse_chroma[0] = 0.0;
  mse_chroma[1] = 0.0;
  total_time = 0;
  
  // write header
  enc->htotal_bits = ((param.write(buf) + 7)>>3)*8;
  
  if (f_dst)  
  {
#ifdef IDR_EXTRACT
    idr_extract_fptr = open_file("idr_extract.txt", "w");
    total_idr_bytes = sizeof(short) + sizeof(int) + buf->p_byte + ((buf->p_bit + 7) >> 3);
    fprintf(idr_extract_fptr, "header offset %d\n", total_idr_bytes);
#endif
    buf->payload_type = Buffer::PLT_HEADER;
    buf->write(f_dst);
    fflush(f_dst);    
  }
  
  // print column heads
  if (silent_mode == FALSE) 
  {
    printf("----------------------------------------------\n");
    printf("Frame     "); // 10
    printf("Bits     "); // 9
    printf("QP  "); // 4
    printf("PSNR Y  "); // 8
    printf("PSNR Cr "); // 8
    printf("PSNR Cb "); // 8
    printf("Time    "); // 8
    printf("Blks (Z/N/I)   "); // 15
    printf("Bits (M/L/C)         "); // 21
    printf("SIP (L/C)     "); //14
    printf("TMA (L9/L5/L9c/L5c)     "); // 24
    if(param.motion_sharing_log)
    {
      printf("MS (OFF/ZERO/TWO/FOUR/USAGE)    "); //32
    }
    if(param.aif)
    {
      printf("AIF    "); //7
    }
    printf("MELQ "); //5
    printf("LmdQ "); //5
    printf("RL0 "); //4
    printf("RL1"); //3
    printf("\n");
    printf("%05d(H)  %08d\n",0,enc->htotal_bits);
  }
  
  // main loop
  f_log_frame = open_file("log_frame.txt", "wt");
  last_qp = param.qp;
#ifdef TRACE_DEBUG
  OPEN_ETRACE;
#endif
  
  // Main coding loop
  for (frame = 0; frame < param.nframes; frame++) 
  {
    fr->frame_no = frame;
    fr->frame_order_counter = fr->frame_no - last_frame_no;
    fr->reference_idc = REFERENCE_PIC; // Edge picture is a reference picture
    fr->lq_offset = 0;
    fr->qp = param.qp;
    fr->intermed_idx = 0; // This is an edge picture
    
    // read current frame    
    current_frame_to_code = (1 + param.frame_skip) * fr->frame_no;
    read_one_frame(fd_src,fr,current_frame_to_code,param.f_header,param.frame_offset);
    fr_orig->copy(fr);
    
    fr->E();
    fr->delta_qp = fr->qp - last_qp;
    
    set_quantization_set_id(fr->frame_no, fr->intermed_idx); // cq_start
    param.already_update[0] = param.already_update[1] = 0; // cq_end
    
    encode_one_frame(enc, fr, fr_pred, fr_qpe_m, fr_qpe_s, fr_orig, fr_filtered, buf, f_par, f_dst, fd_rec,  f_sad, f_log_frame);
    
    param.qmatrix_overwrite_flag[0][param.quantization_set_id] = 0; // cq_start
    param.qmatrix_overwrite_flag[1][param.quantization_set_id] = 0; // cq_end
    
#ifdef TRACE_DEBUG
    SAVE_ETRACE;
#endif
    last_qp = fr->qp;
    last_frame_no = fr->frame_no;
    
    snr_luma += fr->snr_luma;
    snr_chroma[0] += fr->snr_chroma[0];
    snr_chroma[1] += fr->snr_chroma[1];
    mse_luma += fr->mse_luma;
    mse_chroma[0] += fr->mse_chroma[0];
    mse_chroma[1] += fr->mse_chroma[1];
    
    // Note that this implies that f_ref and f_src have the same frame organization.
    // Maybe we could also specify additional variables/counters to access f_ref.
    if (fd_ref != -1) 
    {
      read_one_frame(fd_ref,fr,current_frame_to_code,param.f_header,param.frame_offset);
      fr_orig->copy(fr);
      enc->set_ref_frame(fr);
    }
    
    if (frame != 0 || param.no_reordering)  // if there is no reordering, encoder the intermediate frames immediately after the edge picture including the first picture.
    {
      intermed_frame = (is_idr_gop) ? param.idr_intermed_frame : param.intermed_frame;
      
      if (intermed_frame > 0)
      {
        ExplicitGop *pyramid = (is_idr_gop) ? idr_pyramid : std_pyramid;
        int b_frame_to_code;

        for (b_frame_to_code = 0; b_frame_to_code < intermed_frame; b_frame_to_code++) 
        {
          fr->frame_no = frame + 1 + pyramid->gop_structure[b_frame_to_code].display_no;          
          if (!param.no_reordering)
            fr->frame_no -= intermed_frame + 1;
          
          fr->frame_order_counter = fr->frame_no - last_frame_no;
          fr->reference_idc = pyramid->gop_structure[b_frame_to_code].reference_idc; 
          fr->qp            = pyramid->gop_structure[b_frame_to_code].frame_qp; 
          fr->lq_offset = (fr->reference_idc ? param.lambdaq_offset : 0);
          fr->intermed_idx = b_frame_to_code + 1;
          
          current_frame_to_code = (1 + param.frame_skip) * fr->frame_no;
          read_one_frame(fd_src,fr,current_frame_to_code,param.f_header,param.frame_offset);        
          fr_orig->copy(fr);
          
          fr->E();
          fr->delta_qp = fr->qp - last_qp;
          
          set_quantization_set_id(fr->frame_no, fr->intermed_idx); // cq_start
          param.already_update[0] = param.already_update[1] = 0; // cq_end
          
          encode_one_frame(enc, fr, fr_pred, fr_qpe_m, fr_qpe_s, fr_orig, fr_filtered, buf, f_par, f_dst, fd_rec, f_sad, f_log_frame);
          
          param.qmatrix_overwrite_flag[0][param.quantization_set_id] = 0; // cq_start
          param.qmatrix_overwrite_flag[1][param.quantization_set_id] = 0; // cq_end
          
#ifdef TRACE_DEBUG
          SAVE_ETRACE;
#endif
          last_qp = fr->qp;
          last_frame_no = fr->frame_no;
        
          snr_luma += fr->snr_luma;
          snr_chroma[0] += fr->snr_chroma[0];
          snr_chroma[1] += fr->snr_chroma[1];
          mse_luma += fr->mse_luma;
          mse_chroma[0] += fr->mse_chroma[0];
          mse_chroma[1] += fr->mse_chroma[1];
          
          if (fd_ref != -1) 
          {
            read_one_frame(fd_ref,fr,current_frame_to_code,param.f_header,param.frame_offset);
            fr_orig->copy(fr);
            enc->set_ref_frame(fr);
          }
        }
      }
    }    
    // reallocate pyramid if needed due to frames in file.
    
    if (param.no_reordering)
    {
      frame += (is_idr_gop) ? param.idr_intermed_frame : param.intermed_frame;
    }
    
    int max_intermed_frames = imax(param.nframes - frame - 2, 0);
    param.intermed_frame     = imin(param.intermed_frame, max_intermed_frames);
    param.idr_intermed_frame = imin(param.idr_intermed_frame, max_intermed_frames);
    std_pyramid->resize(param.intermed_frame);
    idr_pyramid->resize(param.idr_intermed_frame);
    
    is_idr_gop = (idr_refresh && frame == target_last_non_idr_frame);
    if (is_idr_gop)
      target_last_non_idr_frame += idr_refresh;
    
    if (!param.no_reordering)
    {
      frame += (is_idr_gop) ? param.idr_intermed_frame : param.intermed_frame;
    }
  }
  
  if (f_dst != NULL)
  {
#ifdef IDR_EXTRACT
    close_file(idr_extract_fptr);
#endif
    buf->reset();
    buf->payload_type = Buffer::PLT_END;
    buf->write(f_dst);
    fflush(f_dst);
  }
  
  // find total elapsed time
  clk = clock();
  
  if (mse_mode == FALSE) 
  {
    snr_luma = snr_luma / enc->coded_frames;
    snr_chroma[0] = snr_chroma[0] / enc->coded_frames;
    snr_chroma[1] = snr_chroma[1] / enc->coded_frames;
  }
  else 
  {
    snr_luma = psnr(mse_luma / enc->coded_frames);
    snr_chroma[0] = psnr(mse_chroma[0] / enc->coded_frames);
    snr_chroma[1] = psnr(mse_chroma[1] / enc->coded_frames);
  }
  
  // print overall stats
  if (silent_mode == FALSE) 
  {
    printf("----------------------------------------------\n");
    printf("AVE:");
    printf("%05d ", enc->coded_frames);
    printf("%08d ", (enc->total_bits + enc->coded_frames/2)/enc->coded_frames);
    printf("    "); // Do we want average QP?
    printf("%07.3f ", snr_luma);
    printf("%07.3f ", snr_chroma[0]);
    printf("%07.3f ", snr_chroma[1]);
    printf("%07.3f ", 1.0 * clk / enc->coded_frames / CLOCKS_PER_SEC);
    
    printf("%04d %04d %04d ",
           enc->mlc_tot.zblks  / enc->coded_frames,
           enc->mlc_tot.nzblks / enc->coded_frames,
           enc->mlc_tot.iblks  / enc->coded_frames);
    printf("%06d %06d %06d ",
           enc->mlc_tot.motion_bits / enc->coded_frames,
           enc->mlc_tot.luma_bits   / enc->coded_frames,
           enc->mlc_tot.chroma_bits / enc->coded_frames);
    
    printf("%06d %06d ",
           enc->mlc_tot.bit_count_luma   / enc->coded_frames,
           enc->mlc_tot.bit_count_chroma / enc->coded_frames);
    
    printf("%05d %05d %05d %05d ",enc->mlc_tot.intra_tma_L9/ enc->coded_frames, 
           enc->mlc_tot.intra_tma_L5/ enc->coded_frames, enc->mlc_tot.intra_tma_L9c/ enc->coded_frames,
           enc->mlc_tot.intra_tma_L5c/ enc->coded_frames);

    if(enc->coded_inter_frames && param.motion_sharing_log)
    {
      printf("%05d %05d %05d %05d %07.3f ",enc->mlc_tot.mvshare_off/ enc->coded_inter_frames,
             enc->mlc_tot.mvshare_zero/ enc->coded_inter_frames, enc->mlc_tot.mvshare_two/ enc->coded_inter_frames, 
             enc->mlc_tot.mvshare_four/ enc->coded_inter_frames,  
             (float)(enc->mlc_tot.mvshare_two+enc->mlc_tot.mvshare_four)/(enc->mlc_tot.mvshare_off+enc->mlc_tot.mvshare_zero+enc->mlc_tot.mvshare_two+enc->mlc_tot.mvshare_four)*100);
    }
    else
      printf("%05d %05d %05d %05d %07.3f ", 0, 0, 0, 0, 0.0);

    if(param.aif)
    {
      printf("%06d", enc->mlc_tot.aif_bits / enc->coded_frames);
    }
  
    printf("\n");
    printf("----------------------------------------------\n");
    printf("XXXXX     %08d\n", (enc->htotal_bits + enc->coded_frames/2)/enc->coded_frames);
  }
  
#ifdef SIPNxN_OVERALL
  // print sip mode count
  int i, j;
  
  enc->getIEncoder()->print_sip_mode_count();         // tma
  for (i=0; i<NSIP; i++ )
    printf ("sip type = %d, count = %d\n", i, enc->getIEncoder()->sip_type_count[i]);
  
  
  double sum_ss_4x4 = 0.0, sum_ss_2x8 = 0.0, sum_ss_8x2 = 0.0, sum_ss_8x8 = 0.0;
  
  for (j=0; j<11; j++) 
  {
    printf ("4x4 stats, pred_mode %d\t", j);
    for (i=0; i<16; i++ ) 
    {
      printf ("\t%f", enc->getIEncoder()->ss_4x4[j][i]/enc->getIEncoder()->ss_4x4_count[j]);
      sum_ss_4x4+=enc->getIEncoder()->ss_4x4[j][i]; 
    };
    printf("\n");
  }
  for (j=0; j<11; j++) 
  {
    printf ("2x8 stats, pred_mode %d\t", j);
    for (i=0; i<16; i++ ) 
    {
      printf ("\t%f", enc->getIEncoder()->ss_2x8[j][i]/enc->getIEncoder()->ss_2x8_count[j]);
      sum_ss_2x8+=enc->getIEncoder()->ss_2x8[j][i]; 
    };
    printf("\n");
  }
  for (j=0; j<11; j++) 
  {
    printf ("8x2 stats, pred_mode %d\t", j);
    for (i=0; i<16; i++ ) 
    {
      printf ("\t%f", enc->getIEncoder()->ss_8x2[j][i]/enc->getIEncoder()->ss_8x2_count[j]);
      sum_ss_8x2+=enc->getIEncoder()->ss_8x2[j][i]; 
    };
    printf("\n");
  }
  for (j=0; j<11; j++) 
  {
    printf ("8x8 stats, pred_mode %d\t", j);
    for (i=0; i<64; i++ ) 
    {
      printf ("\t%f", enc->getIEncoder()->ss_8x8[j][i]/enc->getIEncoder()->ss_8x8_count[j]);
      sum_ss_8x8+=enc->getIEncoder()->ss_8x8[j][i]; 
    };
    printf("\n");
  }
  
  printf ("Sum Square Coeff 4x4 = %f\n", sum_ss_4x4); 
  printf ("Sum Square Coeff 2x8 = %f\n", sum_ss_2x8); 
  printf ("Sum Square Coeff 8x2 = %f\n", sum_ss_8x2); 
  printf ("Sum Square Coeff 8x8 = %f\n", sum_ss_8x8); 
  printf ("Sum Square Coeff All = %f\n", sum_ss_4x4 + sum_ss_2x8 + sum_ss_8x2 + sum_ss_8x8); 
  
#endif
  
  // write log file
  f = open_file(param.logfile, "at");
  if (f != NULL)
  {
    fprintf(f, "%s ", param.srcfile);
    fprintf(f, "%d ", param.qp);
    fprintf(f, "%d ", param.lambdaq_base);
    fprintf(f, "%d ", param.nframes);
    fprintf(f, "%d ", param.melq_base);
    fprintf(f, "%f ", 1.0*enc->total_bits/enc->coded_frames);
    fprintf(f, "%f ", snr_luma);
    fprintf(f, "%f ", snr_chroma[0]);
    fprintf(f, "%f ", snr_chroma[1]);
    fprintf(f, "%f\n", 1.0 * clk / enc->coded_frames / CLOCKS_PER_SEC);
    close_file(f);
  }
    
  f = open_file(param.sadlogfile, "at");
  if (f != NULL)
  {
    fprintf(f, "%s ", param.srcfile);
    fprintf(f, "%d ", param.nframes);
    fprintf(f, "%d ", param.melq_base);
    fprintf(f, "%f ", 1.0*enc->mlc_tot.motion_bits/(enc->coded_frames-1));
    if(param.aif)
    {
      fprintf(f, "AIF %f ", 1.0*enc->mlc_tot.aif_bits/(enc->coded_frames-1));
    }
    fprintf(f, "%f\n", 1.0*enc->total_sad/(enc->coded_frames-1)/enc->frame_height/enc->frame_width);
    close_file(f);
  }
  
  // close files
  close_file(f_log_frame);
  close_file(fd_src);
  close_file(fd_rec);
  close_file(f_dst);
  close_file(f_sad);
  close_file(f_par);
  close_file(fd_ref);
#ifdef TRACE_DEBUG
  CLOSE_ETRACE;
#endif
  // clear memory
  delete std_pyramid;  
  delete idr_pyramid;  
  delete buf;
  delete enc;
  delete fr;
  delete fr_orig;
  delete fr_filtered;
  delete fr_pred;
  delete fr_qpe_m;
  delete fr_qpe_s;

  return EXIT_SUCCESS;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
