/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEst.cpp
*
* \brief Implementation of MotionEst class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MotionEst.H"
#include "Param.H"
#include "Quantizer.H"
#include "MCRefBlk8.H"
#include "EntropyEncoder.H"
#include "IEncoder6.H"

//-----------------------------------------------------------------------------
// Define macros
//-----------------------------------------------------------------------------

#define MEL_PRECISION 8

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] mf0 Pointer to motion field
* \param[in] cur Pointer to current frame
* \param[in] ref_mgr Pointer to reference frame manager
*/
//-----------------------------------------------------------------------------
MotionEst::MotionEst(MotionField *mf, RefFrameManager *ref_mgr)
{
  this->mf = mf;
  this->ref_mgr = ref_mgr;

  int i;

  for (i = 0; i < 2; i++)
  {
    c_mv_lvl[i] = NULL;
    c_ofs[i] = new int[FOUR * (MAX_SAMPLE_VAL + ONE)];
    c_offset[i] = c_ofs[i] + TWO * (MAX_SAMPLE_VAL + ONE);
  }

  for (i = 0; i < NCTX_SCALE; i++)
  {
    c_scl[i] = new int[FOUR * (power2(SAMPLE_ADD_PREC) + ONE)];
    c_scale[i] = c_scl[i] + TWO * (power2(SAMPLE_ADD_PREC) + ONE);
  }
}

void MotionEst::reset(Frame *cur)
{
  int i;

  int j, max_inter_pel=1;

  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      if(ref_mgr->ref_process->inter_pel[i][j] > max_inter_pel)
      {
        max_inter_pel=ref_mgr->ref_process->inter_pel[i][j];
      }
    }
  }

  me_range = max_inter_pel * param.me_range;

  for (i = ZERO; i < TWO; i++)
  {
    if (c_mv_lvl[i] != NULL)
      delete[] c_mv_lvl[i];
    c_mv_lvl[i] = new int[FOUR * me_range];
    c_mv_level[i] = c_mv_lvl[i] + TWO * me_range;
  }

  mel = (int)(pow(2.0, MEL_PRECISION + (cur->melq & 15) / 16.0) + 0.5);
  mel_shift = RATE_BITS - (cur->melq >> 4);

  if (param.use_chroma_me == 0)
    chroma_weight = ZERO;
  else if (param.use_chroma_me == 1)
    chroma_weight = 256;
  else
    chroma_weight = int(pow(2.0, 8 + (cur->qp - cur->qpC) / 16.0) + 0.5);  
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
MotionEst::~MotionEst()
{
  int i;

  for (i = ZERO; i < TWO; i++)
    delete[] c_mv_lvl[i];

  for (i=0; i<NCTX_SCALE; i++)
    delete[] c_scl[i];

  for (i=0; i<2; i++)
    delete[] c_ofs[i];

}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Multiply rate by lambda parameter
*
* \param[in] x Rate
*
* \return Rate multiplied by lambda
*/
//-----------------------------------------------------------------------------
int MotionEst::mult_mel(int x)
{
  return shift_right_round(x * mel, MEL_PRECISION);
}

//-----------------------------------------------------------------------------
/*!
* \brief Multiply rate by lambda parameter in double precision
*
* \param[in] x Rate
*
* \return Rate multiplied by lambda
*/
//-----------------------------------------------------------------------------
int MotionEst::mult_mel_double(int x)
{
  double b = (1<<MEL_PRECISION);
  double b_offset = (1 << (MEL_PRECISION-1));

  return (int)(((double)x *mel + b_offset) / b);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute all rate costs used in R-D optimization
*
* \param[in] eenc Pointer to entropy encoder
*/
//-----------------------------------------------------------------------------
void MotionEst::reset_cost(EntropyEncoder *eenc)
{
  int i, j, x0, x1;
  int low;
  Block blk;
  ZeroTree *ztree;
  int c_skip, c_noskip;
  int check_mv_level[2];
  int keep_max_mv_level[2];

  c_skip   = param.bitsM[1][eenc->output_mv_zero(0)];
  c_noskip = param.bitsM[0][eenc->output_mv_zero(0)];

  ztree = param.ztree[4];
  for (i = ZERO; i < power2(ztree->nleaves); i++)
  {
    for (j = ZERO; j < ztree->nleaves; j++)
      blk.val[j] = (i >> j) & ONE;

    ztree->update_values(&blk);
    c_mv_zero[i] = mult_mel(ztree->get_ztcost(ztree->root) + ((i == 0) ? c_skip : c_noskip));
  }

  ztree = param.ztree[5];
  for (i = ZERO; i < power2(ztree->nleaves); i++)
  {
    for (j = ZERO; j < ztree->nleaves; j++)
      blk.val[j] = (i >> j) & ONE;

    ztree->update_values(&blk);
    c_mv_zero_bi[i] = mult_mel(ztree->get_ztcost(ztree->root) + ((i == 0) ? c_skip : c_noskip));
  }

  check_mv_level[0]= check_mv_level[1]= ZERO;
  keep_max_mv_level[0]= keep_max_mv_level[1]= ZERO;

  for (i = ONE; i < TWO * me_range; i++)
  {
    c_mv_level[0][ i] = mult_mel_double(eenc->cost_mv(i, 0));
    c_mv_level[1][ i] = mult_mel_double(eenc->cost_mv(i, 1));
    c_mv_level[0][-i] = c_mv_level[0][i];
    c_mv_level[1][-i] = c_mv_level[1][i];
    if(c_mv_level[0][i] > keep_max_mv_level[0]) keep_max_mv_level[0] = c_mv_level[0][i];
    if(c_mv_level[1][i] > keep_max_mv_level[1]) keep_max_mv_level[1] = c_mv_level[1][i];
    if(check_mv_level[0]==0 && c_mv_level[0][i] < 0) check_mv_level[0] = i;
    if(check_mv_level[1]==0 && c_mv_level[1][i] < 0) check_mv_level[1] = i;
  }
  if(check_mv_level[0])
  {
    for (i = check_mv_level[0]; i < TWO * me_range; i++)
    {
      c_mv_level[0][ i] = c_mv_level[0][-i] = keep_max_mv_level[0];
    }
  }
  if(check_mv_level[1])
  {
    for (i = check_mv_level[1]; i < TWO * me_range; i++)
    {
      c_mv_level[1][ i] = c_mv_level[1][-i] = keep_max_mv_level[1];
    }
  }

#ifndef MVCOMP_ZERORES
  for (i = ZERO; i < TWO; i++)
#else
  for (i = ZERO; i < THREE; i++)
#endif
  {
    c_mvp[i][0] = mult_mel(eenc->cost_mvp(0, i));
    c_mvp[i][1] = mult_mel(eenc->cost_mvp(1, i));
  }

#ifdef MVCOMP_ZERORES
  int proba = eenc->proba_mvp(2);
  int cnt0 = eenc->count_mvp(2, 0);
  int cnt1 = eenc->count_mvp(2, 1);

  // Global probability of signalling T is cnt1/(cnt0+cnt1) -- we signal T using '1'.
  double p = cnt0+cnt1 == 0 || cnt1 == 0 ? ((double)0x10000 - proba)/0x10000: (double)cnt1/(cnt0+cnt1);
  double thresh = sqrt((1.0-p)/p);
  int ceiling = (int)thresh;
  mv_temporal_thresh = -ceiling-1; // mv_temporal_thresh -ve => toss close values.
#endif

  c_pmode[0] = mult_mel(param.bitsM[0][eenc->output_prev_intra_pred_mode_flag()]);
  c_pmode[1] = mult_mel(param.bitsM[1][eenc->output_prev_intra_pred_mode_flag()]);
  for (i = ZERO; i < 5; i++)
  {
    c_rmode[i][0] = mult_mel(param.bitsM[0][eenc->output_rem_intra_pred_mode_flag(i)]);
    c_rmode[i][1] = mult_mel(param.bitsM[1][eenc->output_rem_intra_pred_mode_flag(i)]);
  }
  for (i = ZERO; i < THREE; i++)
  {
    c_cmode[i][0] = mult_mel(param.bitsM[0][eenc->output_intra_chroma_pred_mode(i)]);
    c_cmode[i][1] = mult_mel(param.bitsM[1][eenc->output_intra_chroma_pred_mode(i)]);
  }
  c_cmode_1[0] = mult_mel(param.bitsM[0][eenc->output_intra_chroma_pred_mode_1()]);
  c_cmode_1[1] = mult_mel(param.bitsM[1][eenc->output_intra_chroma_pred_mode_1()]);
  c_cmode_2[0] = mult_mel(param.bitsM[0][eenc->output_intra_chroma_pred_mode_2()]);
  c_cmode_2[1] = mult_mel(param.bitsM[1][eenc->output_intra_chroma_pred_mode_2()]); 

  low = INT_INF;
  for (x0=0; x0<4; x0++)
  {
    for (x1=0; x1<4; x1++)
    {
      for (i=0; i<4; i++)
      {
        c_listidx[x0][x1][i] = mult_mel(eenc->cost_mode(i, x0, x1, mf->frame_type == ft_B_frame));
        low = imin(low, c_listidx[x0][x1][i]);
      }
    }
  }

  for (x0=0; x0<4; x0++)
  {
    for (x1=0; x1<4; x1++)
    {
      for (i=0; i<4; i++)
      {
        c_listidx[x0][x1][i] -= low;
      }
    }
  }

  for (x0=0; x0<5; x0++)
  {
    if(x0==3 || x0==1) continue;
    for (x1=0; x1<5; x1++)
    {
      if(x1==3 || x1==1) continue;
      for (j=0; j<3; j++) 
      {
        for (i=0; i<7; i++) 
        {
          if(i==4 || i==3 || i==0) continue;
          c_mvshare[x0][x1][j][i] = mult_mel(eenc->cost_mvshare_size(x0, x1, j, i));
        }
      }
    }
  }

  low = INT_INF;
  for (x0=0; x0<2; x0++)
  {
    for (x1=0; x1<2; x1++)
    {
      for (i=0; i<2; i++)
      {
        c_listidx_P[x0][x1][i] = mult_mel(eenc->cost_mode(i, x0, x1, mf->frame_type == ft_B_frame));
        low = imin(low, c_listidx_P[x0][x1][i]);
      }
    }
  }
  for (x0=0; x0<2; x0++)
  {
    for (x1=0; x1<2; x1++)
    {
      for (i=0; i<2; i++)
      {
        c_listidx_P[x0][x1][i] -= low;
      }
    }
  }

  for (i=0; i<LISTS; i++)
  {
    for (x0=0; x0<ref_mgr->lsize[i]; x0++)
    {
      for (x1=0; x1<ref_mgr->lsize[i]; x1++)
      {
        for (j=0; j<ref_mgr->lsize[i]; j++)
        {
          c_refidx[i][x0][x1][j] = mult_mel(eenc->cost_refindex(j, ref_mgr->lsize[i], x0, x1));
        }
      }
    }
  }

  for (j = ZERO; j < NSIP; j++)
    for (i = ONE; i < NCTX_SIP_TYPE; i++)
    {
      c_siptype[j][i] = mult_mel(eenc->cost_sip_type(j, i));   
    }

  for (j = ZERO; j < NCTX_SCALE; j++)
  {
    for (i = ONE; i <= TWO * power2(SAMPLE_ADD_PREC); i++)
    {
      c_scale[j][ i] = mult_mel(eenc->cost_scale(i, j));
      c_scale[j][-i] = c_scale[j][i];
    }
  }

  for (j = ZERO; j < TWO; j++)
  {
    for (i = ONE; i <= TWO * MAX_SAMPLE_VAL; i++)
    {
      c_offset[j][ i] = mult_mel(eenc->cost_offset(i, j));
      c_offset[j][-i] = c_offset[j][i];
    }
  }

}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding MVD (single-list)
*
* \param[in] d0 Vertical MV difference
* \param[in] d1 Horizontal MV difference
* \param[in] d_scale Scale parameter difference
* \param[in] d_offset Offset parameter difference
*
* \return Cost of coding MVD
*/
//-----------------------------------------------------------------------------
int MotionEst::mv_cost_d(int d0, int d1, int d_scale, int d_offset)
{
  int result, index;
  
  index = 0;
  result = 0;
  
  if (d0 != 0)
  {
    int d0x = abs(d0);
    int max_val = TWO * me_range - 1;
    while (d0x >= max_val)
    {
      result += c_mv_level[vertical][max_val];
      d0x -= max_val;
    }
    if (d0x != 0)
    {
      result += c_mv_level[vertical][d0x];
    }
    index += 1;
    
    if (result < 0)
      goto bail;
  }
  
  if (d1 != 0)
  {
    int d1x = abs(d1);
    int max_val = TWO * me_range - 1;
    while (d1x >= max_val)
    {
      result += c_mv_level[horizontal][max_val];
      d1x -= max_val;
    }
    if (d1x != 0)
    {
      result += c_mv_level[horizontal][d1x];
    }
    index += 2;
    
    if (result < 0)
      goto bail;
  }
    
  if (d_scale != 0)
  {
    result += c_scale[0][d_scale];
    index += 4;
    
    if (result < 0)
      goto bail;
  }
  
  if (d_offset != 0)
  {
    result += c_offset[0][d_offset];
    index += 8;
    
    if (result < 0)
      goto bail;
  }
  
  result += c_mv_zero[index];
  
  if (result < 0)
    goto bail;
  
  return result;

bail:
  printf("Warning: negative result = %d (%d,%d,%d,%d)\n", result ,d0, d1, d_scale, d_offset);
  return INT_INF / 2;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding MVD (bipred)
*
* \param[in] d0_0 Vertical MV difference for list 0
* \param[in] d1_0 Horizontal MV difference for list 0
* \param[in] d0_1 Vertical MV difference for list 1
* \param[in] d1_1 Horizontal MV difference for list 1
* \param[in] d_scale Scale parameter difference
* \param[in] d_soffset Scale difference parameter difference
* \param[in] d_offset Offset parameter difference
*
* \return Cost of coding MVD
*/
//-----------------------------------------------------------------------------
int MotionEst::mv_cost_d_bipred(int d0_0, int d0_1, int d1_0, int d1_1, int d_scale, int d_soffset, int d_offset)
{
  int result, index;
  
  index = 0;
  result = 0;
  
  if (d0_0 != 0)
  {
    int d0x = abs(d0_0);
    int max_val = TWO * me_range - 1;
    while (d0x >= max_val)
    {
      result += c_mv_level[vertical][max_val];
      d0x -= max_val;
    }
    if (d0x != 0)
    {
      result += c_mv_level[vertical][d0x];
    }
    index += 1;
  }
  
  if (d0_1 != 0)
  {
    int d1x = abs(d0_1);
    int max_val = TWO * me_range - 1;
    while (d1x >= max_val)
    {
      result += c_mv_level[horizontal][max_val];
      d1x -= max_val;
    }
    if (d1x != 0)
    {
      result += c_mv_level[horizontal][d1x];
    }
    index += 2;
  }

  if (d1_0 != 0)
  {
    int d0x = abs(d1_0);
    int max_val = TWO * me_range - 1;
    while (d0x >= max_val)
    {
      result += c_mv_level[vertical][max_val];
      d0x -= max_val;
    }
    if (d0x != 0)
    {
      result += c_mv_level[vertical][d0x];
    }
    index += 4;
  }
  
  if (d1_1 != 0)
  {
    int d1x = abs(d1_1);
    int max_val = TWO * me_range - 1;
    while (d1x >= max_val)
    {
      result += c_mv_level[horizontal][max_val];
      d1x -= max_val;
    }
    if (d1x != 0)
    {
      result += c_mv_level[horizontal][d1x];
    }
    index += 8;
  }
  
  if (d_scale != 0)
  {
    result += c_scale[1][d_scale];
    index += 16;
  }
  
  if (d_soffset != 0)
  {
    result += c_scale[2][d_soffset];
    index += 32;
  }
  
  if (d_offset != 0)
  {
    result += c_offset[1][d_offset];
    index += 64;
  }
  
  result += c_mv_zero_bi[index];
  
  return result;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding block in intra mode
*
* Computes the cost of coding a block in intra mode. The cost may depend on
* whether neighbouring blocks are coded using intra mode as it affects DC
* prediction. The result is cached in PredBlkI::cost to avoid additional
* computation on succeeding calls.
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
* \param[in] topleft Motion data of block to the upper-left of current block
*
* \return Cost of coding block
*/
//-----------------------------------------------------------------------------
int MotionEst::bit_cost_intra(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft)
{
  int ctx;
  int r, k, l;
  int min_mode = -1;
  int mode;

  ctx = 0;
  if (left != NULL && left->intra != NULL)
    ctx += 1;
  if (top != NULL && top->intra != NULL)
    ctx += 2;
  if (topleft != NULL && topleft->intra != NULL)
    ctx += 4;

  if (cur->intra->cost[ctx] == -1)
  {

    if (mf->frame_type == ft_B_frame)
    {
      r = bit_cost_listindex(cur, left, top); 
    }
    else
    {
      r = bit_cost_listindex_P(cur, left, top);
    }

    r += c_siptype[cur->intra->sip_type][1];     // todo ctx for sip_type in Inter picture is not known since NxN SIP is not created or stored.;

    for (k=0; k<2; k++)
    {
      for (l=0; l<2; l++)
      {
        if((cur->intra->sip_type == SIP8x8)&& (k||l))
          continue;

        min_mode = cur->intra->lmode_luma[k][l];

        mode = (int)cur->intra->mode_luma[k][l];
        if (mode == min_mode)    
        {
          r += c_pmode[1];
        }
        else 
        {
          r += c_pmode[0];

          if (min_mode < mode)  
          {
            mode--;
          }

          if (mode > 9)
          {
            fprintf(stderr, "MotionEst::init_intra No such SIP mode: %d\n", mode);
          }

          int mode2 = imin(mode, 7);
          r += c_rmode[0][(mode2 >> 2) & 1];
          r += c_rmode[1][(mode2 >> 1) & 1];
          r += c_rmode[2][mode2 & 1];

          if (mode >= 7)
          {
            r += c_rmode[3][mode > 7];
            if (mode > 7)
            {
              r += c_rmode[4][mode > 8];
            }
          }
        }
      }
    }

    int flag_count = 0;

    if (top != NULL && top->intra != NULL && top->intra->mode_chroma != 0)
      flag_count++;

    if (left != NULL && left->intra != NULL && left->intra->mode_chroma != 0)
      flag_count++;

    mode = cur->intra->mode_chroma;

    if (mode > 5)
    {
      printf("ERROR:at IEncoder.cpp#cost_bits()\n");
    }

    r += c_cmode[flag_count][mode != 0];

    if (mode > 0)
    {
      r += c_cmode_1[mode > 1];
      if (mode > 1)
      {
        r += c_cmode_2[mode > 2];
        if (mode > 2)
        {
          r += c_cmode_2[mode > 3];
          if (mode > 3)
          {
            r += c_cmode_2[mode > 4];
          }
        }
      }
    }

    cur->intra->cost[ctx] = r;
  }
  else
    r = cur->intra->cost[ctx];

  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding MVD
*
* This is used when a single MV predictor is available.
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of neighbouring block
*
* \return Cost of coding MVD
*/
//-----------------------------------------------------------------------------
int MotionEst::bit_cost_inter1(PredBlk *cur, PredBlk *left)
{
  int list;
#ifdef MV_CODING_ENHANCE
  int first_list, rscale[2], delta_mv[2][2];
  int flag_xpred = 0;
#endif

  list = cur->list;

  if (list == BI_PRED)
#ifdef MV_CODING_ENHANCE
  {
    delta_mv[LIST_0][vertical  ] = cur->mv[LIST_0][vertical  ] - left->mv[LIST_0][vertical  ];
    delta_mv[LIST_0][horizontal] = cur->mv[LIST_0][horizontal] - left->mv[LIST_0][horizontal];
    delta_mv[LIST_1][vertical  ] = cur->mv[LIST_1][vertical  ] - left->mv[LIST_1][vertical  ];
    delta_mv[LIST_1][horizontal] = cur->mv[LIST_1][horizontal] - left->mv[LIST_1][horizontal];

    rscale[LIST_0] = ref_mgr->dist[LIST_0][cur->ref_index[LIST_0]];
    rscale[LIST_1] = ref_mgr->dist[LIST_1][cur->ref_index[LIST_1]];
    first_list = abs(rscale[LIST_0]) < abs(rscale[LIST_1]);
    flag_xpred = (sign(left->mv[LIST_0][vertical]*rscale[LIST_0])==sign(left->mv[LIST_1][vertical]*rscale[LIST_1]));
    flag_xpred = (flag_xpred && (sign(left->mv[LIST_0][horizontal]*rscale[LIST_0])==sign(left->mv[LIST_1][horizontal]*rscale[LIST_1])));
    if (flag_xpred)
    {
      rscale[first_list] = ref_mgr->inv_dist[first_list][cur->ref_index[first_list]];
      delta_mv[first_list^1][vertical  ] -= shift_right_round(delta_mv[first_list][vertical  ] * rscale[0] * rscale[1], 12);
      delta_mv[first_list^1][horizontal] -= shift_right_round(delta_mv[first_list][horizontal] * rscale[0] * rscale[1], 12);
    }

    return mv_cost_d_bipred(delta_mv[LIST_0][0], delta_mv[LIST_0][1], delta_mv[LIST_1][0], delta_mv[LIST_1][1], 
      cur->BP_scale[0]   - left->BP_scale[0], 
      cur->BP_scale[1]   - left->BP_scale[1], 
      cur->BP_DC         - left->BP_DC);
  }
#else
    return mv_cost_d_bipred(cur->mv[LIST_0][0] - left->mv[LIST_0][0], 
    cur->mv[LIST_0][1] - left->mv[LIST_0][1],
    cur->mv[LIST_1][0] - left->mv[LIST_1][0],
    cur->mv[LIST_1][1] - left->mv[LIST_1][1],
    cur->BP_scale[0]   - left->BP_scale[0], 
    cur->BP_scale[1]   - left->BP_scale[1], 
    cur->BP_DC         - left->BP_DC);
#endif
  else
    return mv_cost_d(cur->mv[list][0] - left->mv[list][0],
    cur->mv[list][1] - left->mv[list][1],
    cur->scale[list] - left->scale[list],
    cur->DC[list]    - left->DC[list]);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding MVD
*
* This is used when a two MV predictors are available.
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
* \param[in] topleft Motion data of block to the upper-left of current block
* \param[in] extra_pred indicates we're signalling with a dedicated predictor context
*
* \return Cost of coding MVD
*/
//-----------------------------------------------------------------------------
#ifndef MVCOMP_ZERORES
int MotionEst::bit_cost_inter2(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft)
#else
// left and top are actually predictor 1 and predictor 2.
int MotionEst::bit_cost_inter2(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft, int extra_pred)
#endif
{
  int c0, c1, ctx;

#ifndef MVCOMP_ZERORES
  ctx = topleft->get_pred_context(left, top, cur->list);
#else
  ctx = topleft->get_pred_context(left, top, cur->list, extra_pred);
#endif

  c0 = bit_cost_inter1(cur, left) + c_mvp[ctx][0];
  c1 = bit_cost_inter1(cur, top ) + c_mvp[ctx][1];

  return imin(c0, c1);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding reference index
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
* \param[in] list List index (LIST_0 or LIST_1)
*
* \return Cost of coding reference index
*/
//-----------------------------------------------------------------------------
int MotionEst::bit_cost_refindex(PredBlk *cur, PredBlk *left, PredBlk *top, int list)
{
  int ref_left, ref_top;

  ref_left = (left == NULL || left->intra != NULL) ? 4 : left->ref_index[list];
  ref_top  = (top  == NULL ||  top->intra != NULL) ? 4 :  top->ref_index[list];

  if (ref_left == 4)
    ref_left = ref_top & 3;
  if (ref_top == 4)
    ref_top = ref_left & 3;

  return c_refidx[list][ref_left][ref_top][cur->ref_index[list]];
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding prediction mode
*
* This considers the cost of signalling whether a block uses LIST_0, LIST_1, or
* BI_PRED prediction.
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
*
* \return Cost of coding prediction mode.
*/
//-----------------------------------------------------------------------------
int MotionEst::bit_cost_listindex(PredBlk *cur, PredBlk *left, PredBlk *top)
{
  int x0, x1;
  int x2;

  x0 = left ?  left->mode(): BI_PRED;
  x1 = top ?  top->mode(): BI_PRED;
  x2 = cur ?  cur->mode(): BI_PRED;

  return c_listidx[x0][x1][x2];
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding prediction mode for P-picï¿½@
*
* This considers the cost of signalling whether a block uses LIST_0, LIST_1, or
* BI_PRED prediction.
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
*
* \return Cost of coding prediction mode.
*/
//-----------------------------------------------------------------------------
int MotionEst::bit_cost_listindex_P(PredBlk *cur, PredBlk *left, PredBlk *top)
{
  int x0, x1;
  int x2;

  x0 = left ?  (left->mode()) : MODE_P_DEF;
  x1 = top ?  (top->mode()) : MODE_P_DEF;
  x2 = cur ?  (cur->mode()) : MODE_P_DEF;

  return c_listidx_P[x0][x1][x2];
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding shift value for boundary block shifting
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
*
* \return Cost of coding shift value.
*/
//-----------------------------------------------------------------------------
int MotionEst::bit_cost_mvshare(PredBlk *cur, PredBlk *top, PredBlk *left)
{
  int mvshare_type;
  int mvshare_size_ctx;
  int r = 0;
  int prev_val;

  mvshare_type = mf->mvshare_type_simple(cur, top, left); // 0, 1, 2 or 4
  
  if (mvshare_type)
  {
    mvshare_size_ctx = mf->get_mvshare_size_ctx(mvshare_type, top, left);
    
    if (mvshare_type == 1)
    {
      prev_val = (left == NULL) ? 0 : left->mvshare_size[0];
      r += c_mvshare[cur->mvshare_size[0]][prev_val][mvshare_size_ctx][1];
    }
    else if (mvshare_type == 2)
    {
      prev_val = (top == NULL) ? 0 : top->mvshare_size[1];
      r += c_mvshare[cur->mvshare_size[1]][prev_val][mvshare_size_ctx][2];
    }
    else
    {
      if (cur->mvshare_size[0])
      {
        r+= c_mvshare[cur->mvshare_size[0]][left->mvshare_size[0]][mvshare_size_ctx][5];
      }
      else
      {
        r+= c_mvshare[cur->mvshare_size[1]][top->mvshare_size[1]][mvshare_size_ctx][6];
      }
    }
  }

  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding block in inter mode
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
* \param[in] topleft Motion data of block to the upper-left of current block
*
* \return Cost of coding block
*/
//-----------------------------------------------------------------------------
#ifndef MVCOMP_ZERORES
int MotionEst::bit_cost_inter(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft)
#else
int MotionEst::bit_cost_inter(PredBlk *cur, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft)
#endif
{
  int r;
  int list;
  PredBlk *pred_left, *pred_top, *pred_tleft;
#ifdef MVCOMP_ZERORES
  PredBlkSet predblkset;
  PredBlk **ptmp_preds, **ptmp_diff_preds;
  int n_diff_preds;
#endif

  list = cur->list;

  if (mf->frame_type == ft_B_frame)
  {
    r = bit_cost_listindex(cur, left, top);
  }
  else
  {
    r = bit_cost_listindex_P(cur, left, top);
  }

  if (cur->list != BI_PRED)
  {
    r += bit_cost_refindex(cur, left, top, list);
  }
  else 
  {
    r += bit_cost_refindex(cur, left, top, LIST_0);
    r += bit_cost_refindex(cur, left, top, LIST_1);
  }

#ifndef MVCOMP_ZERORES
#ifdef MV_CODING_ENHANCE

#ifdef PMV0_AT_OUT_OF_BOUNDARY
  pred_left = pleft.get_prediction(i,j,left, cur, ref_mgr);
  pred_top = ptop.get_prediction(i,j,top,  cur, ref_mgr);
#else
  pred_left = pleft.get_prediction(left, cur, ref_mgr);
  pred_top = ptop.get_prediction(top,  cur, ref_mgr);
#endif

#else
  pred_left = pleft.get_prediction(left, cur);
  pred_top = ptop.get_prediction(top,  cur);
#endif
#else
  ptmp_preds = predblkset.get_predictor_set_complete(ref_mgr, i, j, left, top, topleft, cur);
  ptmp_diff_preds = predblkset.collapse_predictor_set(list, n_diff_preds);
  int extra_pred;
  predblkset.signal_treatment(n_diff_preds, extra_pred, list,
                              0, 0, 0, mv_temporal_thresh);
  pred_left = n_diff_preds > 0 ? ptmp_diff_preds[0] : NULL;
  pred_top = n_diff_preds > 1 ? ptmp_diff_preds[1] : NULL;
#endif

  if (pred_left == NULL)
  {
    if (pred_top == NULL)
      r += bit_cost_inter1(cur, &zero_mv);
    else
      r += bit_cost_inter1(cur, pred_top);
  }
  else if (pred_top == NULL || pred_left->equal_pred(pred_top, list))
    r += bit_cost_inter1(cur, pred_left);
  else 
  {
#ifdef MV_CODING_ENHANCE

#ifdef PMV0_AT_OUT_OF_BOUNDARY
    pred_tleft = ptleft.get_prediction(i,j,topleft, cur, ref_mgr);
#else
    pred_tleft = ptleft.get_prediction(topleft, cur, ref_mgr);
#endif


#else
    pred_tleft = ptleft.get_prediction(topleft, cur);
#endif
    if (pred_tleft == NULL)
      pred_tleft = &zero_mv;
#ifndef MVCOMP_ZERORES
    r += bit_cost_inter2(cur, pred_left, pred_top, pred_tleft);
#else
    r += bit_cost_inter2(cur, pred_left, pred_top, pred_tleft, extra_pred);
#endif
  }

  if(param.motion_sharing_enable)
  {
    r += bit_cost_mvshare(cur, top, left);
  }

  return r;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute cost of coding block in any mode 
*
* \param[in] cur Motion data of current block
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
* \param[in] topleft Motion data of block to the upper-left of current block
*
* \return Cost of coding block
*/
//-----------------------------------------------------------------------------
#ifndef MVCOMP_ZERORES
int MotionEst::bit_cost(PredBlk *cur, PredBlk *left, PredBlk *top, PredBlk *topleft)
#else
int MotionEst::bit_cost(PredBlk *cur, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft)
#endif
{
  if (cur == NULL)
    return ZERO;
  else if (cur->intra != NULL)
    return bit_cost_intra(cur, left, top, topleft);
  else
#ifndef MVCOMP_ZERORES
    return bit_cost_inter(cur, left, top, topleft);
#else
    return bit_cost_inter(cur, i, j, left, top, topleft);
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Initialize costs for intra coding
*
* Stores DC values from ienc for estimating cost of coding it. Computes SAD
* values for intra blocks.
*
* \param[in] ienc Pointer to I-block encoder
* \param[in] cur2 Current frame
* \param[in] eenc Pointer to entropy encoder
*/
//-----------------------------------------------------------------------------
void MotionEst::intra_init(IEncoder6 *ienc, EntropyEncoder *eenc)
{
  int i,j,c;
  Block blk_pred;
  MCRefBlk8 blkr;

  blkr.ref_mgr = ref_mgr;
  blkr.mel_shift = mel_shift;
  blkr.chroma_weight = chroma_weight;

  reset_cost(eenc);


  if(intra_cost_update == ICOST_COPY)  // copy of stored data
  {
    for (i = 0; i < mf->width * mf->height; i++)
    {
      mf->blk_array_I[i] = ienc->ipic_store[i];
      mf->blk_array_I[i].reset_cost();
    }
  }
  else if(intra_cost_update == ICOST_UPDATE || intra_cost_update == ICOST_STORED) // update intra cost or store cost
  {

  for (i = 0; i < mf->width * mf->height; i++)
  {
    mf->blk_array_I[i] = ienc->store[i];
    mf->blk_array_I[i].reset_cost();
  }

  for (i=0; i<mf->height; i++)
  {
    for (j=0; j<mf->width; j++)
    {
      int sip_type = mf->blk_array_I[mf->width*i+j].sip_type;
      int sad = 0;

      blk_pred.set_size(8, 8);
      ienc->store_ipred_img->get_blk_luma(&blk_pred, i, j);

      if (sip_type == SIP2x8)
      {
        sad += blkr.transform2x8(blk_pred.val);
      }
      else if (sip_type == SIP8x2)
      {
        sad += blkr.transform8x2(blk_pred.val);
      }
      else if (sip_type == SIP8x8)
      {
        sad += blkr.transform8x8(blk_pred.val);
      }
      else
      {
        sad += blkr.transform(blk_pred.val);
        sad += blkr.transform(blk_pred.val+32);
      }

      blk_pred.set_size(4, 8);
      for (c=0; c<2; c++)
      {
        Block blk_pred_tmp(4, 4);
        ienc->store_ipred_img->get_blk_chroma(&blk_pred_tmp, i, j, c);
        blk_pred.insert(blk_pred_tmp, 0, 4 * c);
      }
      sad += blkr.transformC(blk_pred.val);

      mf->blk_array_I[mf->width*i+j].sad = sad;
    }
  }

  }

  if(intra_cost_update == ICOST_STORED)
  {
    for (i = 0; i < mf->width * mf->height; i++)
    {
      ienc->ipic_store[i] = mf->blk_array_I[i];
    }
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
