/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEstRefine.cpp
*
* \brief Implementation of MotionEstRefine class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MotionEstRefine.H"
#include "Param.H"
#include "MCRefBlk8.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] mf0 Pointer to motion field
* \param[in] cur Pointer to current frame
* \param[in] ref_mgr Pointer to reference frame manager
*/
//-----------------------------------------------------------------------------
MotionEstRefine::MotionEstRefine(MotionField *mf, RefFrameManager *ref_mgr) : MotionEst(mf, ref_mgr)
{
  flag = new int[mf->width * mf->height];
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
MotionEstRefine::~MotionEstRefine()
{
  delete[] flag;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Insert a candidate in candidate array
*
* If candidate already in list, it is not added a second time.
*
* \param[in] val Value to be inserted
* \param[in,out] lcand Array of candidates
* \param[in,out] ncand Number of candidates in array
*/
//----------------------------------------------------------------------------
void MotionEstRefine::insert_val(int val, int *lcand, int *ncand)
{
  int i;
  int n;
  
  n = *ncand;
  
  for (i = 0; i < n; i++)
    if (val == lcand[i])
      return;
  
  lcand[n++] = val;
  *ncand = n;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute rate-distortion cost of all blocks affected by changing current block
*
* Terminates early if partial cost exceed bsad.
*
* \param[in] i Vertical 8x8 block position in a frame
* \param[in] j Horizontal 8x8 block position in a frame
* \param[in] sadin Threshold for early exit
* \param[in] curblk Motion data of current block
* \param[in] mf Pointer to motion field
* \param[in] blk Pointer to MC reference block
* \param[in] BiPred flag to indicate whether frame type is ft_B_frame
*
* \return Rate-Distortion cost
*/
//-----------------------------------------------------------------------------
int MotionEstRefine::total_cost_mvshare(int i, int j, int sadin, PredBlk *curblk, MotionField *mf, MCRefBlk8 *blk, int Bframe)
{
  PredBlk *top;
  PredBlk *left;
  int sad;
  int min_sad=sadin;
  int best_mvshare_size[2];
  int mvshare_type;
   
  left = mf->get_pblk(i,j-1);
  top = mf->get_pblk(i-1,j);
   

  mvshare_type = mf->mvshare_type_simple(curblk, top, left);
   
   
  best_mvshare_size[0]=curblk->mvshare_size[0];
  best_mvshare_size[1]=curblk->mvshare_size[1];
   
  curblk->mvshare_size[0]=curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
  sad = bit_cost_all(curblk, sadin);
#else
  sad = bit_cost_all(curblk, i, j, sadin);
#endif
  if(sad < min_sad)
  {
    sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
    if(sad < min_sad)
    {
      min_sad =sad;
      best_mvshare_size[0]=best_mvshare_size[1]=0;
    }
  }
   
  if(mvshare_type)
  {  
    if(mvshare_type==1 || mvshare_type==4)
    {
      if(param.mvshare_two_enable_me)
      {
        curblk->mvshare_size[0]=2; 
        curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
        sad = bit_cost_all(curblk, sadin);
#else
        sad = bit_cost_all(curblk, i, j, sadin);
#endif
        if(sad < min_sad)
        {
          sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
          if(sad < min_sad)
          {
            min_sad = sad;
            best_mvshare_size[0]=2;
            best_mvshare_size[1]=0;
          }
        }
      }
     
      curblk->mvshare_size[0]=4; 
      curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
      sad = bit_cost_all(curblk, sadin);
#else
      sad = bit_cost_all(curblk, i, j, sadin);
#endif
      if(sad < min_sad)
      {
        sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
        if(sad < min_sad)
        {
          min_sad = sad;
          best_mvshare_size[0]=4;
          best_mvshare_size[1]=0;
        }
      }
       
    }
     
    if(mvshare_type==2 || mvshare_type==4)
    {
      if(param.mvshare_two_enable_me)
      {
        curblk->mvshare_size[0]=0; 
        curblk->mvshare_size[1]=2;
#ifndef MVCOMP_ZERORES
        sad = bit_cost_all(curblk, sadin);
#else
        sad = bit_cost_all(curblk, i, j, sadin);
#endif
        if(sad < min_sad)
        {
          sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
          if(sad < min_sad)
          {
            min_sad = sad;
            best_mvshare_size[0]=0;
            best_mvshare_size[1]=2;
          }
        }
      }
       
      curblk->mvshare_size[0]=0; 
      curblk->mvshare_size[1]=4;
#ifndef MVCOMP_ZERORES
      sad = bit_cost_all(curblk, sadin);
#else
      sad = bit_cost_all(curblk, i, j, sadin);
#endif
      if(sad < min_sad)
      {
        sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
        if(sad < min_sad)
        {
          min_sad = sad;
          best_mvshare_size[0]=0;
          best_mvshare_size[1]=4;
        }
      }
       
    }
     
  }
  
  curblk->mvshare_size[0] = best_mvshare_size[0];
  curblk->mvshare_size[1] = best_mvshare_size[1];
   
  return min_sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute rate cost of all blocks affected by changing current block
*
* Terminates early if partial cost exceed bsad.
*
* \param[in] cand Motion data of current block
* \param[in] bsad Best cost observed so far
*
* \return Rate cost
*/
//-----------------------------------------------------------------------------
#ifndef MVCOMP_ZERORES
int MotionEstRefine::bit_cost_all(PredBlk *cand, int bsad)
#else
int MotionEstRefine::bit_cost_all(PredBlk *cand, int i, int j, int bsad)
#endif
{
  int cost;

#ifndef MVCOMP_ZERORES
  cost = bit_cost(cand, neighbour[ONE][ZERO], neighbour[ZERO][ONE], neighbour[ZERO][ZERO]);
#else
  cost = bit_cost(cand, i, j, neighbour[ONE][ZERO], neighbour[ZERO][ONE], neighbour[ZERO][ZERO]);
#endif
  if (cost >= bsad)
    return cost;
  
  if (neighbour[TWO][ONE]) 
  { // bottom
#ifndef MVCOMP_ZERORES
    cost += bit_cost(neighbour[TWO][ONE], neighbour[TWO][ZERO], cand, neighbour[ONE][ZERO]);
#else
    cost += bit_cost(neighbour[TWO][ONE], i+ONE, j, neighbour[TWO][ZERO], cand, neighbour[ONE][ZERO]);
#endif
    if (cost >= bsad)
      return cost;
  }
  
  if (neighbour[ONE][TWO]) 
  { // right
#ifndef MVCOMP_ZERORES
    cost += bit_cost(neighbour[ONE][TWO], cand, neighbour[ZERO][TWO], neighbour[ZERO][ONE]);
#else
    cost += bit_cost(neighbour[ONE][TWO], i, j+ONE, cand, neighbour[ZERO][TWO], neighbour[ZERO][ONE]);
#endif
    if (cost >= bsad)
      return cost;
  }
  
  if (neighbour[TWO][TWO]) // bottom right
#ifndef MVCOMP_ZERORES
    cost += bit_cost(neighbour[TWO][TWO], neighbour[TWO][ONE], neighbour[ONE][TWO], cand);
#else
    cost += bit_cost(neighbour[TWO][TWO], i+ONE, j+ONE, neighbour[TWO][ONE], neighbour[ONE][TWO], cand);
#endif
  
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Estimate motion field
*
* Iterates through every 8x8 block in frame and tries to find a set of motion
* data that lowers the R-D cost.
*
* \param[in] cur Current frame
* \param[in] list Estimation mode (BI_PRED or not)
*/
//-----------------------------------------------------------------------------
void MotionEstRefine::estimate(Frame *cur, int list)
{
  int i, j, k, l, m, n, scale, sv;
  MCRefBlk8 blk;
  PredBlk best_mv, ref_mv, cand_mv;
  int bsad, sad;
  int changed;
  int cur_list, i_list;
  int lmin, lmax;
  
  int n_size;
  
  blk.ref_mgr = ref_mgr;
  blk.mel_shift = mel_shift;
  blk.chroma_weight = chroma_weight;
  
  n_size = THREE;
  
  for (i = ZERO; i < mf->height * mf->width; i++)
    flag[i] = TRUE; 
  
  for (n = ZERO; n < nloops; n++)
  {  
    changed = ZERO;
    
    for (i = ZERO; i < mf->height; i++)
    {
      for (j = ZERO; j < mf->width; j++) 
      {
        if (flag[mf->width*i+j] == FALSE)
          continue; 
        
        flag[mf->width*i+j] = FALSE;
        blk.get_data(cur, i, j);
        
        for (k=0; k<3; k++) 
        {
          for (l=0; l<3; l++) 
          {
            neighbour[k][l] = mf->get_pblk(i + k - ONE, j + l - ONE);
          }
        }

          if (neighbour[ONE][ONE]->intra) 
          {
            best_mv.clear();
            if (list == BI_PRED)
              best_mv.list = BI_PRED;
          }
          else
            best_mv = *neighbour[ONE][ONE];   
          
          cur_list = best_mv.list;  

          if(param.motion_sharing_me_refine)
          {
            bsad = total_cost_mvshare(i, j, INT_INF, &best_mv, mf, &blk, cur->frame_type==ft_B_frame);
          }
          else
          {
            bsad = blk.distortion(i, j, INT_INF, &best_mv);
#ifndef MVCOMP_ZERORES
            bsad += bit_cost_all(&best_mv, INT_INF);
#else
            bsad += bit_cost_all(&best_mv, i, j, INT_INF);
#endif
          }

          for (k = ZERO; k < n_size; k++) 
          {
            for (l = ZERO; l < n_size; l++) 
            {
              if (neighbour[k][l] == NULL || neighbour[k][l]->intra != NULL)
                continue;
              
              if (best_mv.equal(neighbour[k][l]))
                continue;
            
              cand_mv = *neighbour[k][l];

              if (param.motion_sharing_me_refine_full)
              {
                sad = total_cost_mvshare(i, j, bsad, &cand_mv, mf, &blk, cur->frame_type==ft_B_frame);
             
                if (sad >= bsad)
                  continue;
              }
              else
              {
                sad = blk.distortion(i, j, bsad, &cand_mv);
                if (sad >= bsad)
                  continue;
            
#ifndef MVCOMP_ZERORES
                sad += bit_cost_all(&cand_mv, bsad - sad);
#else
                sad += bit_cost_all(&cand_mv, i, j, bsad - sad);
#endif
                if (sad >= bsad)
                  continue;
              }

              bsad = sad;
              best_mv = cand_mv;
            }
          }
          
          cur_list = best_mv.list;
          
          if (param.enable_intcomp)
          {
            cand_mv = best_mv;
            cur_list = cand_mv.list;
            
            for (k=0; k<3; k++)
            {
              for (l=0; l<3; l++)
              {
                if (neighbour[k][l] == NULL || neighbour[k][l]->intra != NULL)
                  continue;
                
                if (cur_list != neighbour[k][l]->list)
                  continue;
                
                if (cur_list == BI_PRED &&
                  cand_mv.ref_index[0] == neighbour[k][l]->ref_index[0] &&
                  cand_mv.ref_index[1] == neighbour[k][l]->ref_index[1]) 
                {
                  cand_mv.BP_DC       = neighbour[k][l]->BP_DC;
                  cand_mv.BP_scale[0] = neighbour[k][l]->BP_scale[0];
                  cand_mv.BP_scale[1] = neighbour[k][l]->BP_scale[1];
                }
                else if (cur_list != BI_PRED && cand_mv.ref_index[cur_list] == neighbour[k][l]->ref_index[cur_list]) 
                {
                  cand_mv.scale[cur_list] = neighbour[k][l]->scale[cur_list];
                  cand_mv.DC[cur_list]    = neighbour[k][l]->DC[cur_list];
                }
                else
                  continue;
                
                if (best_mv.equal(&cand_mv))
                  continue;
                
                if(param.motion_sharing_me_refine_full)
                {
                  sad = total_cost_mvshare(i, j, bsad, &cand_mv, mf, &blk, cur->frame_type==ft_B_frame);
               
                  if (sad >= bsad)
                    continue;
                }
                else
                {
                  sad = blk.distortion(i, j, bsad, &cand_mv);
                  if (sad >= bsad)
                    continue;
                
#ifndef MVCOMP_ZERORES
                  sad += bit_cost_all(&cand_mv, bsad - sad);
#else
                  sad += bit_cost_all(&cand_mv, i, j, bsad - sad);
#endif
                  if (sad >= bsad)
                    continue;
                }
                
                bsad = sad;
                best_mv = cand_mv;
                
              }
            }
          }
          
          cur_list = best_mv.list;
          
          if (cur_list == BI_PRED) 
          {
            lmin = LIST_0;
            lmax = LIST_1;
          }
          else 
          {
            lmin = cur_list;
            lmax = cur_list;
          }
          
          for (i_list = lmin; i_list <= lmax; i_list++)
          {
            for (m=0; m < ref_mgr->lsize[i_list]; m++)
            {
              for (scale=3; scale>=0; scale--) 
              {
                cand_mv = best_mv;
                cand_mv.ref_index[i_list] = m;
                ref_mv = best_mv;
                
                for (k=-1; k<=1; k++) 
                {
                  for (l=-1; l<=1; l++) 
                  {
                    sv = ref_mgr->mv_scale[i_list][m][ref_mv.ref_index[i_list]];
                    
                    cand_mv.mv[i_list][0] = shift_right_round(sv * ref_mv.mv[i_list][0], RESCALE_BITS) + (k << scale);
                    cand_mv.mv[i_list][1] = shift_right_round(sv * ref_mv.mv[i_list][1], RESCALE_BITS) + (l << scale);
                    
                    if(param.motion_sharing_me_refine_full)
                    {
                      sad = total_cost_mvshare(i, j, bsad, &cand_mv, mf, &blk, cur->frame_type==ft_B_frame);
                     
                      if (sad >= bsad)
                        continue;
                    }
                    else
                    {
                      sad = blk.distortion(i, j, bsad, &cand_mv);
                      if (sad >= bsad)
                        continue;
                    
#ifndef MVCOMP_ZERORES
                      sad += bit_cost_all(&cand_mv, bsad - sad);
#else
                      sad += bit_cost_all(&cand_mv, i, j, bsad - sad);
#endif
                      if (sad >= bsad)
                        continue;
                    }
                    
                    bsad = sad;
                    best_mv = cand_mv;
                    
                  }
                }
              }
            }
          }
          
          cur_list = best_mv.list;
          
          if (param.enable_intcomp) 
          {
            int sum1, sum2, sum3;
            int lcand[2][7];
            int offset_value;
            int ncand;
            int scale[2];
            int s0, s1;
            
            cand_mv = best_mv;
            ref_mv = best_mv;
            
            if (cur_list == BI_PRED) 
            {
              for (m = -param.me_refine_s1; m <= param.me_refine_s1; m++) 
              {
                for (k = -param.me_refine_s2; k <= param.me_refine_s2; k++) 
                {
                  sum1 = blk.sum_luma;
                  sum2 = blk.get_mc_sum_luma(i, j, &cand_mv, LIST_0);
                  sum3 = blk.get_mc_sum_luma(i, j, &cand_mv, LIST_1);
                  
                  ncand = ZERO;          
                  
                  if (neighbour[0][1] != NULL) 
                    insert_val(neighbour[0][1]->BP_DC, lcand[LIST_0], &ncand);
                  if (neighbour[1][0] != NULL) 
                    insert_val(neighbour[1][0]->BP_DC, lcand[LIST_0], &ncand);
                  if (neighbour[2][1] != NULL) 
                    insert_val(neighbour[2][1]->BP_DC, lcand[LIST_0], &ncand);
                  if (neighbour[1][2] != NULL) 
                    insert_val(neighbour[1][2]->BP_DC, lcand[LIST_0], &ncand);
                  
                  cand_mv.BP_scale[0] = ref_mv.BP_scale[0] + m;
                  cand_mv.BP_scale[1] = ref_mv.BP_scale[1] + k;
                  
                  s0 = (cand_mv.get_B_scale_sum() >> ONE) + cand_mv.get_B_scale_diff();
                  s1 = cand_mv.get_B_scale_sum() - s0;
                  
                  offset_value = (power2(SAMPLE_ADD_PREC) * sum1 - sum2 * s0 - sum3 * s1) / (param.DC_mult << (SAMPLE_ADD_PREC + 6));
                  insert_val(offset_value    , lcand[LIST_0], &ncand);
                  insert_val(offset_value + 1, lcand[LIST_0], &ncand);
                  insert_val(offset_value - 1, lcand[LIST_0], &ncand);                
                  
                  for (l = ZERO; l < ncand; l++) 
                  {
                    cand_mv.BP_DC  = lcand[LIST_0][l];
                    
                    if (cand_mv.inbounds_wp() == FALSE) 
                      continue;
                    
                    if(param.motion_sharing_me_refine_full)
                    {
                      sad = total_cost_mvshare(i, j, bsad, &cand_mv, mf, &blk, cur->frame_type==ft_B_frame);
                     
                      if (sad >= bsad)
                        continue;
                    }
                    else
                    {
                      sad = blk.distortion(i, j, bsad, &cand_mv);
                      if (sad >= bsad)
                        continue;
                    
#ifndef MVCOMP_ZERORES
                      sad += bit_cost_all(&cand_mv, bsad - sad);
#else
                      sad += bit_cost_all(&cand_mv, i, j, bsad - sad);
#endif
                      if (sad >= bsad)
                        continue;
                    }
                    
                    bsad = sad;
                    best_mv = cand_mv;
                  }
                }
              }
            }
            else
            {            
              for (k = -param.me_refine_s0; k <= param.me_refine_s0; k++) 
              {
                sum1 = blk.sum_luma;  
                cand_mv.scale[cur_list] = ZERO;
                cand_mv.DC[cur_list] = ZERO;
                sum2 = blk.get_mc_sum_luma(i, j, &cand_mv, cur_list);
                ncand = ZERO;
                
                if (neighbour[0][1] != NULL) 
                  insert_val(neighbour[0][1]->DC[cur_list], lcand[cur_list], &ncand);
                if (neighbour[1][0] != NULL) 
                  insert_val(neighbour[1][0]->DC[cur_list], lcand[cur_list], &ncand);
                if (neighbour[1][2] != NULL)
                  insert_val(neighbour[1][2]->DC[cur_list], lcand[cur_list], &ncand);
                if (neighbour[2][1] != NULL)
                  insert_val(neighbour[2][1]->DC[cur_list], lcand[cur_list], &ncand);
                
                scale[cur_list] = ref_mv.scale[cur_list] + k + power2(SAMPLE_ADD_PREC);
                
                offset_value = (power2(SAMPLE_ADD_PREC) * sum1 - sum2 * scale[cur_list]) / (param.DC_mult << (SAMPLE_ADD_PREC + 6));
                insert_val(offset_value,       lcand[cur_list], &ncand);
                insert_val(offset_value + ONE, lcand[cur_list], &ncand);
                insert_val(offset_value - ONE, lcand[cur_list], &ncand);              
                
                for (l = ZERO; l < ncand; l++)
                {
                  cand_mv.scale[cur_list] = ref_mv.scale[cur_list] + k;
                  cand_mv.DC[cur_list]    = lcand[cur_list][l];
                  
                  if (cand_mv.inbounds_wp() == FALSE) 
                    continue;
                  
                  if(param.motion_sharing_me_refine_full)
                  {
                    sad = total_cost_mvshare(i, j, bsad, &cand_mv, mf, &blk, cur->frame_type==ft_B_frame);
                   
                    if (sad >= bsad)
                      continue;
                  }
                  else
                  {
                    sad = blk.distortion(i, j, bsad, &cand_mv);
                    if (sad >= bsad)
                      continue;
                  
#ifndef MVCOMP_ZERORES
                    sad += bit_cost_all(&cand_mv, bsad - sad);
#else
                    sad += bit_cost_all(&cand_mv, i, j, bsad - sad);
#endif
                    if (sad >= bsad)
                      continue;
                  }
                  
                  bsad = sad;
                  best_mv = cand_mv;
                  
                }
              }
            }
          }
          
          cand_mv.clear();
          cand_mv.intra = &mf->blk_array_I[mf->width*i+j];
          sad = cand_mv.intra->sad;
#ifndef MVCOMP_ZERORES
          sad += bit_cost_all(&cand_mv, INT_INF);
#else
          sad += bit_cost_all(&cand_mv, i, j, INT_INF);
#endif
          
          if (sad < bsad) 
          {         
            bsad = sad;
            best_mv.intra = cand_mv.intra;
          }
          
          if ( best_mv.equal(neighbour[1][1]) == FALSE || best_mv.intra!=NULL) 
          {
            *mf->get_pblk(i,j) = best_mv;
            if(best_mv.intra)
            {
              mf->get_pblk(i,j)->intra=best_mv.intra;
            }
            changed++;
            
            for (k=i-1; k<=i+1; k++) 
            {
              for (l=j-1; l<=j+1; l++) 
              {
                if (k >= 0 && k < mf->height && l >= 0 && l < mf->width)
                  flag[mf->width*k+l] = TRUE;
              }
            }
          }

      }
    }
    
    if (changed == ZERO)
      break;
  }

  if(param.motion_sharing_me_refine)
  {
    for (i = ZERO; i < mf->height; i++)
    {
      for (j = ZERO; j < mf->width; j++) 
      {
        mf->get_pblk(i, j)->mvshare_size[0]= mf->get_pblk(i, j)->mvshare_size[1]=0;
      }
    }
  }

}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
