/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEstRegion.cpp
*
* \brief Implementation of MotionEstRegion class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MotionEstRegion.H"
#include "Param.H"
#include "MCRefBlk8.H"


//-----------------------------------------------------------------------------
// Define macros
//-----------------------------------------------------------------------------

#define WINDOW 2
#define BIPRED_WINDOW 1

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] mf0 Pointer to motion field
* \param[in] cur Pointer to current frame
* \param[in] ref_mgr Pointer to reference frame manager
*/
//-----------------------------------------------------------------------------
MotionEstRegion::MotionEstRegion(MotionField *mf, RefFrameManager *ref_mgr) : MotionEst(mf, ref_mgr)
{
  max_ncand = param.region_ncand;
  
  region = new int[mf->width * mf->height];
  cand = new PredBlk[max_ncand];
  cand_sl = new PredBlk[mf->width * mf->height];
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
MotionEstRegion::~MotionEstRegion()
{
  delete[] region;
  delete[] cand;
  delete[] cand_sl;
}

//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Get region index of block at position (i,j)
*
* \param[in] i Vertical position (8x8 block unit)
* \param[in] j Horizontal position (8x8 block unit)
*
* \return Region index
*/
//-----------------------------------------------------------------------------
int MotionEstRegion::get_region(int i, int j) 
{
  if (i < ZERO || i >= mf->height || j < ZERO || j >= mf->width)
    return UNDEFINED;
  else
    return region[mf->width*i+j];
}

//-----------------------------------------------------------------------------
/*!
* \brief Test block
*
* Tests whether the block at (i,j) has not yet an assigned region index and whether
* it shares the same motion data as block blk. 
*
* \param[in] blk Motion data of reference block
* \param[in] i Vertical position (8x8 block unit)
* \param[in] j Horizontal position (8x8 block unit)
*
* \return 0:different region 1:same region 
*/
//-----------------------------------------------------------------------------
int MotionEstRegion::test_and_fill(PredBlk *blk, int i, int j)
{
  if (mf->get_pblk_unsafe(i,j)->equal(blk) && region[mf->width*i+j] == UNDEFINED)
    return 1;
  else
    return 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief Filling
*
* Assigns current region number to block at (i,j) and recursively test and fill
* neighbouring blocks.
*
* \param[in] blk Motion data of reference block
* \param[in] same_region Vertical and Horizontal position (8x8 block unit)
* \param[in] nstored Number of stored 8x8 blocks to be unified to same region 
*/
//-----------------------------------------------------------------------------
void MotionEstRegion::fill(PredBlk *blk,int *same_region, int *nstored)
{  
  int flag;
  int i,j;
  i=same_region[(nstored[0]-1)*2+0];
  j=same_region[(nstored[0]-1)*2+1];
  
  region[mf->width*i+j] = nregions;
  
  flag=0; 
  if (i > ZERO)
  {
    if(test_and_fill(blk, i-ONE, j)==1)
    {
      same_region[nstored[0]*2+0]=i-ONE;
      same_region[nstored[0]*2+1]=j;
     flag=1;
      nstored[0]++;
    }
  }
  if (flag==0 && i < mf->height-ONE)
  {
    if(test_and_fill(blk, i+ONE, j)==1)
    {
      same_region[nstored[0]*2+0]=i+ONE;
      same_region[nstored[0]*2+1]=j;
      flag=1;
      nstored[0]++;
    }
  }
  if (flag==0 && j > ZERO)
  {
    if(test_and_fill(blk, i, j-ONE)==1)
    {
      same_region[nstored[0]*2+0]=i;
      same_region[nstored[0]*2+1]=j-ONE;
      flag=1;
      nstored[0]++;
    }
  }
  if (flag==0 && j < mf->width-ONE)
  {
    if(test_and_fill(blk, i, j+ONE)==1)
    {
      same_region[nstored[0]*2+0]=i;
      same_region[nstored[0]*2+1]=j+ONE;
      flag=1;
      nstored[0]++;
    }
  }
  if(flag==0)
  {
    nstored[0]=nstored[0]-1;
  }

}

//-----------------------------------------------------------------------------
/*!
* \brief Find region bounding box
*
* Finds rectangular bounding box of a region
*
* \note This is used to speed up cost computation. This should thus
* be invoked before calling MotionEstRegion::cost
*
* \param[in] r Region index
*/
//-----------------------------------------------------------------------------
void MotionEstRegion::find_region_bbox(int r)
{
  int i,j;
  
  bbox_left = mf->width - ONE;
  bbox_right = ZERO;
  bbox_top = mf->height - ONE;
  bbox_bottom = ZERO;
  
  for (i = ZERO; i < mf->height; i++)
  {
    for (j = ZERO; j < mf->width; j++) 
    {
      if (region[i * mf->width + j] == r)
      {
        bbox_top = imin(bbox_top, i);
        bbox_bottom = imax(bbox_bottom, i);
        bbox_left = imin(bbox_left, j);
        bbox_right = imax(bbox_right, j);
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute rate-distortion cost of all blocks affected by changing current block
*
*
* \param[in] i Vertical 8x8 block position in a frame
* \param[in] j Horizontal 8x8 block position in a frame
* \param[in] curblk Motion data of current block
* \param[in] mf Pointer to motion field
* \param[in] blk Pointer to MC reference block
* \param[in] BiPred flag to indicate whether frame type is ft_B_frame
*
* \return Rate-Distortion cost
*/
//-----------------------------------------------------------------------------
int MotionEstRegion::total_cost_mvshare(int i, int j, PredBlk *curblk, MotionField *mf, MCRefBlk8 *blk, int Bframe)
{
  PredBlk *top;
  PredBlk *left;
  PredBlk *tleft;
  int sad;
  int min_sad;
  int best_mvshare_size[2];
  int mvshare_type;
   
  left = mf->get_pblk(i,j-1);
  top = mf->get_pblk(i-1,j);
  tleft = mf->get_pblk(i-1, j-1);
   
  mvshare_type = mf->mvshare_type_simple(curblk, top, left);
   
  curblk->mvshare_size[0]=curblk->mvshare_size[1]=0;
  best_mvshare_size[0]=best_mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
  min_sad = bit_cost(curblk, left, top, tleft);
#else
  min_sad = bit_cost(curblk, i, j, left, top, tleft);
#endif
  min_sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
   
  if(mvshare_type)
  {
     
    if(mvshare_type==1 || mvshare_type==4)
    {

      if(param.mvshare_two_enable_me)
      {
        curblk->mvshare_size[0]=2; 
        curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
        sad = bit_cost(curblk, left, top, tleft);
#else
        sad = bit_cost(curblk, i, j, left, top, tleft);
#endif
       
        sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
        if(sad < min_sad)
        {
          min_sad = sad;
          best_mvshare_size[0]=2;
          best_mvshare_size[1]=0;
        }
      }
     
      curblk->mvshare_size[0]=4; 
      curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
      sad = bit_cost(curblk, left, top, tleft);
#else
      sad = bit_cost(curblk, i, j, left, top, tleft);
#endif
       
      sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
      if(sad < min_sad)
      {
        min_sad = sad;
        best_mvshare_size[0]=4;
        best_mvshare_size[1]=0;
      }
       
    }
    if(mvshare_type==2 || mvshare_type==4)
    {
      if(param.mvshare_two_enable_me)
      {
        curblk->mvshare_size[0]=0; 
        curblk->mvshare_size[1]=2;
#ifndef MVCOMP_ZERORES
        sad = bit_cost(curblk, left, top, tleft);
#else
        sad = bit_cost(curblk, i, j, left, top, tleft);
#endif
       
        sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
        if(sad < min_sad)
        {
          min_sad = sad;
          best_mvshare_size[0]=0;
          best_mvshare_size[1]=2;
        }
      }
       
      curblk->mvshare_size[0]=0; 
      curblk->mvshare_size[1]=4;
#ifndef MVCOMP_ZERORES
      sad = bit_cost(curblk, left, top, tleft);
#else
      sad = bit_cost(curblk, i, j, left, top, tleft);
#endif
       
      sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
      if(sad < min_sad)
      {
        min_sad = sad;
        best_mvshare_size[0]=0;
        best_mvshare_size[1]=4;
      }
       
    }
     
  }

  curblk->mvshare_size[0] = best_mvshare_size[0];
  curblk->mvshare_size[1] = best_mvshare_size[1];
   
   
  return min_sad;
}

//-----------------------------------------------------------------------------
/*!
* \brief Cost of coding region
*
* Computes R-D cost of coding a region. It includes the coding cost of blocks
* that are affected by the motion data of the region. Exits early if partial
* cost exceeds mincost.
*
* \note This method relies on the bounding box parameters.
*
* \param[in] cur Current frame
* \param[in] r Region index
* \param[in] mincost Best cost so far
*
* \return R-D cost
*/
//-----------------------------------------------------------------------------
double MotionEstRegion::cost(Frame *cur, int r, double mincost)
{
  double cost;
  MCRefBlk8 blk;
  PredBlk *top, *topleft, *left, *curblk;
  
  int left_i, left_j;
  int topleft_i, topleft_j;
  int top_i, top_j;
  int cur_i, cur_j;
  int i,j;
  int maxi, maxj;
  int rcost, dcost;
  
  blk.ref_mgr = ref_mgr;
  blk.mel_shift = mel_shift;
  blk.chroma_weight = chroma_weight;
  
  cost = 0.0;

  maxi = imin(bbox_bottom + ONE, mf->height - ONE);
  maxj = imin(bbox_right  + ONE, mf->width  - ONE);
  
  for (i=bbox_top; i<=maxi; i++)
  {
    curblk = mf->get_pblk(i,       bbox_left - ONE);
    top    = mf->get_pblk(i - ONE, bbox_left - ONE);
    
    cur_i = i;
    cur_j = bbox_left - ONE;
    top_i = i - ONE;
    top_j = bbox_left - ONE;
    
    for (j=bbox_left; j<=maxj; j++)
    {
      left = curblk;
      topleft = top;
      
      curblk = mf->get_pblk_unsafe(i,j);
      if (i != ZERO)
        top = mf->get_pblk_unsafe(i-ONE, j);
      
      left_i = cur_i;
      left_j = cur_j;
      topleft_i = top_i;
      topleft_j = top_j;
      cur_i = i;
      cur_j = j;
      if(i != ZERO)
      {
        top_i = i-ONE;
        top_j = j;
      }
      
      if (region[mf->width*i+j] == r)
      {  
        if(param.motion_sharing_me_region)
        {
          blk.get_data(cur, i, j);
          dcost = total_cost_mvshare(i, j, curblk, mf, &blk, cur->frame_type==ft_B_frame);
          rcost = 0;
        }
        else
        {
          if (curblk->intra)
          {
            dcost = curblk->intra->sad;
          }
          else 
          {
            blk.get_data(cur, i, j);
            dcost = blk.sad(i,j,curblk);
          }
  
#ifndef MVCOMP_ZERORES
          if(left==NULL && top==NULL && topleft==NULL)
            rcost = bit_cost(curblk, NULL, NULL, NULL);
          else if(left==NULL && top==NULL)
            rcost = bit_cost(curblk, NULL, NULL, topleft);
          else if(top==NULL && topleft==NULL)
            rcost = bit_cost(curblk, left, NULL, NULL);
          else if(left==NULL && topleft==NULL)
            rcost = bit_cost(curblk, NULL, top, NULL);
          else if(left==NULL)
            rcost = bit_cost(curblk, NULL, top, topleft);
          else if(top==NULL)
            rcost = bit_cost(curblk, left, NULL, topleft);
          else if(topleft==NULL)
            rcost = bit_cost(curblk, left, top, NULL);
          else
            rcost = bit_cost(curblk, left, top, topleft);
#else
          rcost = bit_cost(curblk,
                           i, j, left, top, topleft);
#endif
        }
        
      }
      
      else if (get_region(i-ONE, j) == r || get_region(i-ONE, j-ONE) == r || get_region(i, j-ONE) == r) 
      {

#ifndef MVCOMP_ZERORES
          rcost = bit_cost(curblk, left, top, topleft);
#else
          rcost = bit_cost(curblk, i, j, left, top, topleft);
#endif
        
        dcost = ZERO;
      }
      else 
      {
        rcost = ZERO;
        dcost = ZERO;
      }
      
      cost += rcost + dcost;

      if (cost >= mincost)
        return cost;
    }
  }
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Insert candidate into short list
*
* Inserts candidate only if not already in list.
*
* \param[in] blk Motion data candidate to insert
*
* \return TRUE if candidate was inserted, FALSE otherwise.
*/
//-----------------------------------------------------------------------------
int MotionEstRegion::insert_cand_short_list(PredBlk *blk)
{
  int i;
  
  if (blk == NULL)
    return FALSE;
  
  for (i=ZERO; i<ncand_sl; i++)
    if (cand_sl[i].equal(blk))
      return FALSE;
  
  cand_sl[ncand_sl] = *blk;
  
  ncand_sl++;
  
  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Insert candidate into short list
*
* Inserts candidate only if not already in list.
*
* \note Unused
*
* \param[in] blk Motion data candidate to insert
*
* \return TRUE if candidate was inserted, FALSE otherwise.
*/
//-----------------------------------------------------------------------------
void MotionEstRegion::insert_cand_vs_slist(PredBlk *blk)
{
  int i;
  
  if (blk == NULL || ncand == max_ncand)
    return;
  
  for (i=ZERO; i<ncand_sl; i++)
    if (cand_sl[i].equal(blk))
      return;
  
  cand[ncand] = *blk;
  ncand++;    
}

//-----------------------------------------------------------------------------
/*!
* \brief Insert candidate into candidate list
*
* Inserts candidate only if not already in list.
*
* \param[in] blk Motion data candidate to insert
*/
//-----------------------------------------------------------------------------
void MotionEstRegion::insert_cand(PredBlk *blk)
{
  int i;
  
  if (blk == NULL || ncand == max_ncand)
    return;
  
  for (i=ZERO; i<ncand; i++)
    if (cand[i].equal(blk))
      return;
  
  cand[ncand] = *blk;
  
  ncand++;
}

//-----------------------------------------------------------------------------
/*!
* \brief Insert candidates into candidate list
*
* Inserts provided candidate and variations thereof into the candidate list
*
* \param[in] blk Motion data candidate to insert
* \param[in] extend Flag indicating whether variations should be inserted
*/
//-----------------------------------------------------------------------------
void MotionEstRegion::insert_cands(PredBlk *blk, int extend) 
{
  PredBlk mycand;
  int k,l, m,n;
  int cur_list, ilist;
  int scale;

#ifdef MV_CODING_ENHANCE
  #ifdef SCALED_CANDIDATES_IN_REGION_ME
  int rscale[2];
  #endif
#endif 
  
  if (blk == NULL)
    return;
  
  if (insert_cand_short_list(blk) == FALSE)
    return;
  
  insert_cand(blk);
  
  if (extend == FALSE)
    return;
  
  if (blk->intra == NULL) 
  {
    cur_list = blk->list;
    if (cur_list == BI_PRED) 
    {
      
      for (m=-BIPRED_WINDOW; m<=BIPRED_WINDOW; m++) 
      {
        for (n=-BIPRED_WINDOW; n<=BIPRED_WINDOW; n++) 
        {          
          for (k=-BIPRED_WINDOW; k<=BIPRED_WINDOW; k++) 
          {
            for (l=-BIPRED_WINDOW; l<=BIPRED_WINDOW; l++) 
            {
              if (k != ZERO || l != ZERO || m != ZERO || n != ZERO) 
              {
                mycand = *blk;
                mycand.mv[LIST_0][vertical] += k;
                mycand.mv[LIST_0][horizontal] += l;
                mycand.mv[LIST_1][vertical] += m;
                mycand.mv[LIST_1][horizontal] += n;
                
                insert_cand(&mycand);
              }
            }
          }
        }
      }
    #ifdef MV_CODING_ENHANCE
     #ifdef SCALED_CANDIDATES_IN_REGION_ME
      //candidate0
      mycand = *blk;
      mycand.mv[LIST_0][vertical]=0;
      mycand.mv[LIST_0][horizontal]=0;
      mycand.mv[LIST_1][vertical]=0;
      mycand.mv[LIST_1][horizontal]=0;
      insert_cand(&mycand);

      //candidate1
      rscale[LIST_0]= ref_mgr->inv_dist[LIST_0][blk->ref_index[LIST_0]];
      rscale[LIST_1]= ref_mgr->dist[LIST_1][blk->ref_index[LIST_1]];
     
      mycand = *blk;
      mycand.mv[LIST_1][vertical]=shift_right_round(mycand.mv[LIST_0][vertical]*rscale[LIST_0]*rscale[LIST_1],12);
      mycand.mv[LIST_1][horizontal]=shift_right_round(mycand.mv[LIST_0][horizontal]*rscale[LIST_0]*rscale[LIST_1],12);
      insert_cand(&mycand);
      
      //candidate2 
      rscale[LIST_0]= ref_mgr->dist[LIST_0][blk->ref_index[LIST_0]];
      rscale[LIST_1]= ref_mgr->inv_dist[LIST_1][blk->ref_index[LIST_1]];
     
      mycand = *blk;
      mycand.mv[LIST_0][vertical]=shift_right_round(mycand.mv[LIST_1][vertical]*rscale[LIST_0]*rscale[LIST_1],12);
      mycand.mv[LIST_0][horizontal]=shift_right_round(mycand.mv[LIST_1][horizontal]*rscale[LIST_0]*rscale[LIST_1],12);

      insert_cand(&mycand);

     #endif
    #endif
    }
    else 
    {
      for (k=-WINDOW; k<=WINDOW; k++) 
      {
        for (l=-WINDOW; l<=WINDOW; l++) 
        {
          
          if (k != ZERO || l != ZERO) 
          {
            mycand = *blk;
            mycand.mv[cur_list][vertical] += k;
            mycand.mv[cur_list][horizontal] += l;
            
            insert_cand(&mycand);
          }
        }
      }
    }
    if (param.enable_intcomp) 
    {
      
      if (cur_list == BI_PRED) 
      {
        for (k = -param.me_region_s1; k <= param.me_region_s1; k++) 
        {
          for (l = -param.me_region_o1; l <= param.me_region_o1; l++) 
          {
            mycand = *blk;
            if (param.bugfix80)
            {
              mycand.BP_scale[0] += k;
              mycand.BP_DC       += l - shift_right_round(MID_SAMPLE_VAL * k, SAMPLE_ADD_PREC);
            } else
            {
              if (k == ZERO && l == ZERO) 
              {
                mycand.BP_scale[0] = ZERO;
                mycand.BP_DC       = ZERO;
              }            
              else 
              {
                mycand.BP_scale[0] += k;
                mycand.BP_DC       += l - shift_right_round(MID_SAMPLE_VAL * k, SAMPLE_ADD_PREC);
              }
            }
            if (mycand.inbounds_wp() == TRUE)
              insert_cand(&mycand);
          }
        }
        
        for (k= -param.me_region_s2; k <= param.me_region_s2; k++) 
        {
          for (l = -param.me_region_o1; l <= param.me_region_o1; l++) 
          {
            mycand = *blk;
            if (param.bugfix80)
            {
                mycand.BP_scale[1] += k;
                mycand.BP_DC       += l;
            } else
            {
              if (k == ZERO && l == ZERO) 
              {
                mycand.BP_scale[1] = ZERO;
                mycand.BP_DC       = ZERO;
              }
              else 
              {
                mycand.BP_scale[1] += k;
                mycand.BP_DC       += l;
              }
            }
            if (mycand.inbounds_wp() == TRUE)
              insert_cand(&mycand);
          }
        }
        if (param.bugfix80)
        {
          mycand = *blk;
          mycand.BP_scale[0] = ZERO;
          mycand.BP_scale[1] = ZERO;
          mycand.BP_DC       = ZERO;
          insert_cand(&mycand);
        }
      }
      else
      {
        for (k = -param.me_region_s0; k <= param.me_region_s0; k++) 
        {
          for (l = -param.me_region_o0; l <= param.me_region_o0; l++) 
          {
            mycand = *blk;
            
            if (k == ZERO && l == ZERO) 
            {
              mycand.scale[cur_list] = ZERO;
              mycand.DC[cur_list] = ZERO;
            }
            else 
            {              
              mycand.scale[cur_list] += k;
              mycand.DC[cur_list] += l - shift_right_round(MID_SAMPLE_VAL * k, SAMPLE_ADD_PREC);
            }
            
            if (mycand.inbounds_wp() == TRUE)
              insert_cand(&mycand);
          }
        }
      }
    }   
    if (cur_list == BI_PRED) 
    {
      for (k=ZERO; k<ref_mgr->lsize[LIST_0]; k++) 
      {      
        for (l=ZERO; l<ref_mgr->lsize[LIST_1]; l++) 
        {      
          mycand = *blk;
          mycand.ref_index[LIST_0] = k;
          mycand.ref_index[LIST_1] = l;
          
          for (ilist = LIST_0; ilist < LISTS; ilist ++) 
          {
            scale = ref_mgr->mv_scale[ilist][mycand.ref_index[ilist]][blk->ref_index[ilist]];
            mycand.mv[ilist][vertical  ] = shift_right_round(scale * blk->mv[ilist][vertical  ], RESCALE_BITS);
            mycand.mv[ilist][horizontal] = shift_right_round(scale * blk->mv[ilist][horizontal], RESCALE_BITS);
          }
          mycand.BP_scale[0] = ZERO;
          mycand.BP_scale[1] = ZERO;
          mycand.BP_DC = ZERO;
          
          insert_cand(&mycand);
        }
      }
    }
    else 
    {
      for (k=ZERO; k<ref_mgr->lsize[cur_list] ; k++) 
      {
        mycand = *blk;
        mycand.ref_index[cur_list] = k;
        scale = ref_mgr->mv_scale[cur_list][mycand.ref_index[cur_list]][blk->ref_index[cur_list]];
        mycand.mv[cur_list][vertical  ] = shift_right_round(scale * blk->mv[cur_list][vertical  ], RESCALE_BITS);
        mycand.mv[cur_list][horizontal] = shift_right_round(scale * blk->mv[cur_list][horizontal], RESCALE_BITS);
        mycand.scale[cur_list] = ZERO;
        mycand.DC[cur_list] = ZERO;
        
        insert_cand(&mycand);
      }
    }
  } 
}

//-----------------------------------------------------------------------------
// Public methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Estimate motion field
*
* Iterates through every homogeneous motion region in a frame and tries 
* to find a set of motion data that lowers the R-D cost.
*
* \param[in] cur Current frame
*
*/
//-----------------------------------------------------------------------------
void MotionEstRegion::estimate(Frame *cur, int) 
{
  int i,j,k,l;
  int r,c;
  int bestc;
  double mincost;
  double sad;
  int loop;
  int first;
  int changed;
  int nchanged;
  
  int *same_region;
  same_region = new int[mf->width*mf->height*2];
  int nstored[1];
 
  for (loop=ZERO; loop<nloops; loop++) 
  {
    changed = FALSE;
    nchanged = ZERO;
    
    for (i=0; i<mf->width*mf->height; i++)
      region[i] = UNDEFINED;
    
    nregions = ZERO;
    
    for (i=ZERO; i<mf->height; i++) 
    {
      for (j=ZERO; j<mf->width; j++) 
      {
        if (region[mf->width*i+j] == UNDEFINED) 
        {
          nstored[0] = ONE;    
          same_region[0]=i;    
          same_region[1]=j;    
          while(nstored[0]>=ONE)
          {
            fill(mf->get_pblk_unsafe(same_region[(nstored[0]-1)*2+0],same_region[(nstored[0]-1)*2+1]),same_region, nstored); 
          }  
          nregions++;
        }
      }
    }
    for (r=ZERO; r<nregions; r++) 
    {
      ncand = ZERO;
      ncand_sl = ZERO;
      first = TRUE;
      
      find_region_bbox(r);
      
      for (i = bbox_top; i <= bbox_bottom; i++) 
      {
        for (j = bbox_left; j <= bbox_right; j++) 
        {
          if (region[mf->width*i+j] == r) 
          {
              
              if (first) 
              {
                first = FALSE;
                insert_cands(mf->get_pblk_unsafe(i,j), TRUE);
              }
              
              for (k=-ONE; k<=ONE; k++)
              {
                for (l=-ONE; l<=ONE; l++)
                {
                  if (k != ZERO || l != ZERO)
                  {
                    
                    if(mf->get_pblk(i+k,j+l))
                    {
                      insert_cands(mf->get_pblk(i+k,j+l), FALSE);
                      
                    }
                  }
                }
              }
          }
        }
      }
      bestc = ZERO;
      mincost = FLT_INF;
      
      for (c=ZERO; c<ncand; c++)
      {
        for (i=bbox_top; i<=bbox_bottom; i++) 
        {
          for (j=bbox_left; j<=bbox_right; j++) 
          {
            if (region[mf->width*i+j] == r) 
            {                   
                *mf->get_pblk_unsafe(i,j) = cand[c];
                
                if (cand[c].intra)
                  mf->get_pblk_unsafe(i,j)->intra = &mf->blk_array_I[i*mf->width+j];
            }
          } 
        } 
        
        sad = cost(cur, r, mincost);
        
        if (sad < mincost) 
        {
          mincost = sad;
          bestc = c;
        }
      }
      
      for (i=bbox_top; i<=bbox_bottom; i++) 
      {
        for (j=bbox_left; j<=bbox_right; j++) 
        {
          if (region[mf->width*i+j] == r) 
          {
                          
              *mf->get_pblk_unsafe(i,j) = cand[bestc];
              if (cand[bestc].intra)
                mf->get_pblk_unsafe(i,j)->intra = &mf->blk_array_I[i*mf->width+j];
              
          }
        } 
      }
      if (bestc != ZERO) 
      {
        changed = TRUE;
        nchanged++;
      }
    }
    
    if (!changed)
      break;
  }

/*
  if(param.motion_sharing_me_region)
  {
    for (i = ZERO; i < mf->height; i++)
    {
      for (j = ZERO; j < mf->width; j++) 
      {
        mf->get_pblk(i, j)->mvshare_size[0]= mf->get_pblk(i, j)->mvshare_size[1]=0;
      }
    }
  }
*/

  delete[] same_region;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
