/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEstTrellis.cpp
*
* \brief Implementation of MotionEstTrellis class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MotionEstTrellis.H"
#include "Param.H"
#include "MCRefBlk8.H"


//-----------------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------------

#define WIN 1
#define TWO_WIN (TWO * WIN + ONE)

#define OFFWIN 2
#define TWO_OFFWIN (TWO * OFFWIN + ONE)

#define OFFWIN_BI 2 
#define TWO_OFFWIN_BI (TWO * OFFWIN_BI + ONE)

#define MAX_NEIGHBOURS 2

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] mf Pointer to motion field
* \param[in] cur Pointer to current frame
* \param[in] ref_mgr Pointer to reference frame manager
*/
//-----------------------------------------------------------------------------
MotionEstTrellis::MotionEstTrellis(MotionField *mf, RefFrameManager *ref_mgr) : MotionEst(mf, ref_mgr) 
{
  int i;
  
  size = imax(mf->width, mf->height);
  
  mvcand = new PredBlk *[size];
  for (i=ZERO; i<size; i++)
    mvcand[i] = new PredBlk[MAXNCAND];
  ncand = new int[size];
  backtrack = new candarray[size];
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
MotionEstTrellis::~MotionEstTrellis() 
{
  int i;
  
  for (i=ZERO; i<size; i++)
    delete[] mvcand[i];
  delete[] mvcand;
  delete[] ncand;
  delete[] backtrack;
}

//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Reorder candidates based on cost
*
* \note Uses bubble sort.
*
* \param[in] z Block index for which to do reordering
* \param[in] cost Array of cost values for each candidate
*/
//-----------------------------------------------------------------------------
void MotionEstTrellis::reorder(int z, int *cost) 
{
  int k,l;
  int tcost;
  int back;
  PredBlk tmv;
  
  for (k=ONE; k<ncand[z]; k++) 
  {
    tcost = cost[k];
    if (tcost < cost[k-1]) 
    {
      l = k;
      tmv = mvcand[z][l];
      back = backtrack[z][l];
      
      do 
      {
        cost[l] = cost[l-1];
        mvcand[z][l] = mvcand[z][l-1];
        backtrack[z][l] = backtrack[z][l-1];
        l--;
      } while (l > 0 && tcost < cost[l-1]);
      
      cost[l] = tcost;
      mvcand[z][l] = tmv;
      backtrack[z][l] = back;
    }
  }
  
  cost[ncand[z]] = INT_INF;
}

//-----------------------------------------------------------------------------
/*!
* \brief Insert default candidates in candidates list
*
* \param[out] lmvcand Candidate list
* \param[in] i Vertical position of block for which to generate candidates
* \param[in] j Horizontal position of block for which to generate candidates
* \param[in] list Set to BI_PRED if should also generate candidates using BI_PRED mode
*
* \return Number of inserted candidates
*/
//-----------------------------------------------------------------------------
int MotionEstTrellis::insert_default_cands(PredBlk *lmvcand, int i, int j, int list) 
{
  int n, ilist;
  int k,l,m, o,p;
  int cur_list;
  n = ZERO;
  
  lmvcand[n] = *mf->get_pblk_unsafe(i,j); 
  lmvcand[n].intra = &mf->blk_array_I[mf->width*i+j];
  n++;
  
  if (mf->get_pblk_unsafe(i,j)->intra == NULL)
  {    
    cur_list = mf->get_pblk_unsafe(i,j)->list;
    
    if (cur_list == BI_PRED) 
    {
      for (o=ZERO; o<TWO_WIN; o++) 
      {
        for (p=ZERO; p<TWO_WIN; p++) 
        {
          for (k=ZERO; k<TWO_WIN; k++) 
          {
            for (l=ZERO; l<TWO_WIN; l++) 
            {
              if ((o != ZERO && p != ZERO) || (k != ZERO && l != ZERO))
                continue;
              lmvcand[n] = *mf->get_pblk_unsafe(i,j);                
              lmvcand[n].mv[LIST_0][vertical] += (k > WIN) ? k - TWO_WIN : k;
              lmvcand[n].mv[LIST_0][horizontal] += (l > WIN) ? l - TWO_WIN : l;
              lmvcand[n].mv[LIST_1][vertical] += (o > WIN) ? o - TWO_WIN : o;
              lmvcand[n].mv[LIST_1][horizontal] += (p > WIN) ? p - TWO_WIN : p;
              
              n++;
            }
          }
        }
      }
      if (param.enable_intcomp) 
      {
        for (k = -param.me_trellis_s1; k <= param.me_trellis_s1; k++)
        {
          for (m = -param.me_trellis_o1; m <= param.me_trellis_o1; m++) 
          {
            lmvcand[n] = *mf->get_pblk_unsafe(i,j);
            
            if (param.bugfix80)
            {
              if (k == ZERO && m == ZERO)
                continue;
              lmvcand[n].BP_scale[0] += k;
              lmvcand[n].BP_DC += m - shift_right_round(MID_SAMPLE_VAL * k, SAMPLE_ADD_PREC);
              if (lmvcand[n].inbounds_wp() == FALSE)
                continue;
            } else
            {
              if (k == ZERO && m == ZERO) 
              {
                if (lmvcand[n].BP_scale[0] == ZERO && lmvcand[n].BP_DC == ZERO)
                  continue;              
                lmvcand[n].BP_scale[0] = ZERO;
                lmvcand[n].BP_DC = ZERO;
              }
              else 
              {
                lmvcand[n].BP_scale[0] += k;
                lmvcand[n].BP_DC += m - shift_right_round(MID_SAMPLE_VAL * k, SAMPLE_ADD_PREC);
                
                if (lmvcand[n].inbounds_wp() == FALSE)
                  continue;
              }
            }
            n++;
          }
        }
        
        for (l = -param.me_trellis_s2; l <= param.me_trellis_s2; l++) 
        {
          if (param.bugfix80)
          {
            lmvcand[n] = *mf->get_pblk_unsafe(i,j);
            if (l == ZERO)
              continue;
            lmvcand[n].BP_scale[1] += l;
            if (lmvcand[n].inbounds_wp() == FALSE)
              continue;
            n++;
          } else
          {
            for (m = -param.me_trellis_o1; m <= param.me_trellis_o1; m++) 
            {
              lmvcand[n] = *mf->get_pblk_unsafe(i,j);
              
              if (l == ZERO && m == ZERO) 
              {
                if (lmvcand[n].BP_scale[1] == ZERO && lmvcand[n].BP_DC == ZERO)
                  continue;
                
                lmvcand[n].BP_scale[1] = ZERO;
                lmvcand[n].BP_DC       = ZERO;
              }
              else 
              {
                lmvcand[n].BP_scale[1] += l;
                lmvcand[n].BP_DC += m;
                
                if (lmvcand[n].inbounds_wp() == FALSE)
                  continue;
              }
            }
            n++;
          }
        }
        if (param.bugfix80)
        {
          lmvcand[n] = *mf->get_pblk_unsafe(i,j);
          if (lmvcand[n].BP_scale[0] != ZERO || lmvcand[n].BP_scale[1] != ZERO || lmvcand[n].BP_DC != ZERO)
          {
            lmvcand[n].BP_scale[0] = ZERO;
            lmvcand[n].BP_scale[1] = ZERO;
            lmvcand[n].BP_DC = ZERO;
            n++;
          }
        }
      }
      
      for (ilist = LIST_0; ilist <= LIST_1; ilist++) 
      {
        for (k = ZERO; k < ref_mgr->lsize[ilist]; k++) 
        {
          int sv;
          
          lmvcand[n] = *mf->get_pblk_unsafe(i,j);
          
          if (k == lmvcand[n].ref_index[ilist])
            continue;
          
          if (param.bugfix80)
          {
            sv = ref_mgr->mv_scale[ilist][k][lmvcand[n].ref_index[ilist]];
            lmvcand[n].ref_index[ilist] = k;
          } else
          {
            lmvcand[n].ref_index[ilist] = k;
            sv = ref_mgr->mv_scale[ilist][k][lmvcand[n].ref_index[ilist]];
          }
          
          lmvcand[n].mv[ilist][0] = shift_right_round(sv * lmvcand[n].mv[ilist][0], RESCALE_BITS);
          lmvcand[n].mv[ilist][1] = shift_right_round(sv * lmvcand[n].mv[ilist][1], RESCALE_BITS);
          
          n++;
        }                  
      }
      
      for (ilist = LIST_0; ilist < LISTS; ilist++) 
      {
        lmvcand[n] = *mf->get_pblk_unsafe(i,j);
        lmvcand[n].list = ilist;
        lmvcand[n].BP_scale[0] = ZERO;
        lmvcand[n].BP_scale[1] = ZERO;
        lmvcand[n].BP_DC = ZERO;
        lmvcand[n].mv[ilist^1][0] = ZERO;
        lmvcand[n].mv[ilist^1][1] = ZERO;
        lmvcand[n].ref_index[ilist^1] = ZERO;
        n++;
      }
    }
    else 
    { 
      
      
      for (k=ZERO; k<TWO_WIN; k++) 
      {
        for (l=ZERO; l<TWO_WIN; l++) 
        {
          lmvcand[n] = *mf->get_pblk_unsafe(i,j);                
          cur_list = lmvcand[n].list;          
          lmvcand[n].mv[cur_list][vertical] += (k > WIN) ? k - TWO_WIN : k;
          lmvcand[n].mv[cur_list][horizontal] += (l > WIN) ? l - TWO_WIN : l;
          
          n++;
        }
      }
      
      
      if (param.enable_intcomp == TRUE) 
      {
        for (k = -param.me_trellis_s0; k <= param.me_trellis_s0; k++)
        {
          for (l = -param.me_trellis_o0; l <= param.me_trellis_o0; l++) 
          {
            lmvcand[n] = *mf->get_pblk_unsafe(i,j);
            cur_list = lmvcand[n].list;
            
            if (k == ZERO && l == ZERO) 
            {
              if (lmvcand[n].scale[cur_list] == ZERO && lmvcand[n].DC[cur_list] == ZERO)
                continue; 
              
              lmvcand[n].scale[cur_list] = ZERO;
              lmvcand[n].DC[cur_list] = ZERO;
            }
            else 
            {
              lmvcand[n].scale[cur_list] += k;
              lmvcand[n].DC[cur_list] += l  - shift_right_round(MID_SAMPLE_VAL * k, SAMPLE_ADD_PREC);
              
              if (lmvcand[n].inbounds_wp() == FALSE)
                continue;
            }           
            n++;
          }
        }
      }
      
      for (k = ZERO; k < ref_mgr->lsize[cur_list]; k++) 
      {
        int sv;
        
        lmvcand[n] = *mf->get_pblk_unsafe(i,j);
        
        if (k == lmvcand[n].ref_index[cur_list])
          continue;
        
        if (param.bugfix80)
        {
          sv = ref_mgr->mv_scale[cur_list][k][lmvcand[n].ref_index[cur_list]];
          lmvcand[n].ref_index[cur_list] = k;
        } else
        {
          lmvcand[n].ref_index[cur_list] = k;
          sv = ref_mgr->mv_scale[cur_list][k][lmvcand[n].ref_index[cur_list]];
        }
        
        lmvcand[n].mv[cur_list][0] = shift_right_round(sv * lmvcand[n].mv[cur_list][0], RESCALE_BITS);
        lmvcand[n].mv[cur_list][1] = shift_right_round(sv * lmvcand[n].mv[cur_list][1], RESCALE_BITS);
        
        
        n++;
      }                  
      
      if (list == BI_PRED) 
      {
        lmvcand[n] = *mf->get_pblk_unsafe(i,j);
        lmvcand[n].list = BI_PRED;
        lmvcand[n].scale[LIST_0] = lmvcand[n].scale[LIST_1] = ZERO;
        lmvcand[n].DC[LIST_0] = lmvcand[n].DC[LIST_1] = 0;
        n++;
      }
    }
  }
  return n;
}

//-----------------------------------------------------------------------------
/*!
* \brief Insert motion data from neighbouring blocks into candidate list
*
* \param[in,out] mvcand Candidate list
* \param[in,out] n Number of candidates in list
* \param[in] i Vertical position of block for which to generate candidates
* \param[in] j Horizontal position of block for which to generate candidates
* \param[in] di Vertical offset
* \param[in] dj Horizontal offset
* \param[in] remaining Number of remaining slots in candidate list before reaching maximum size
*/
//-----------------------------------------------------------------------------
void MotionEstTrellis::insert_neighbour_cands(PredBlk *mvcand, int *n, int i, int j, int di, int dj, int remaining)
{
  PredBlk *last, *current;
  
  last = mf->get_pblk(i,j);
  
  while (last != NULL && remaining >= ZERO) 
  {
    i += di;
    j += dj;
    
    current = mf->get_pblk(i,j);
    if (current == NULL)
      break;
        
    if (last->equal(current) == FALSE) 
    {
      insert_cand(mvcand, n, current);
      last = current;
      remaining--;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Insert a candidate into list
*
* \param[in,out] mvcand Candidate list
* \param[in,out] n Number of candidates in list
* \param[in] cand Candidate to insert
*/
//-----------------------------------------------------------------------------
void MotionEstTrellis::insert_cand(PredBlk *mvcand, int *n, PredBlk *cand) 
{
  int i;
  
  if (cand == NULL || (*n) >= param.treillis_ncand)
    return; 
  
  for (i=ZERO; i<(*n); i++)
  if (mvcand[i].equal(cand))
    return;
  
  mvcand[(*n)] = *cand;
  
  (*n)++;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute rate-distortion cost of all blocks affected by changing current block
*
* Terminates early if partial cost exceed bsad.
*
* \param[in] type Trellis direction indicator
* \param[in] i Vertical 8x8 block position in a frame
* \param[in] j Horizontal 8x8 block position in a frame
* \param[in] curblk Motion data of current block
* \param[in] top_left_blk Motion data of top or left block
* \param[in] mf Pointer to motion field
* \param[in] blk Pointer to MC reference block
* \param[in] BiPred flag to indicate whether frame type is ft_B_frame
*
* \return Rate-Distortion cost
*/
//-----------------------------------------------------------------------------
int MotionEstTrellis::total_cost_mvshare(int type, int i, int j, PredBlk *curblk, PredBlk *top_left_blk, MotionField *mf, MCRefBlk8 *blk, int Bframe)
{
  PredBlk *top;
  PredBlk *left;
  PredBlk *tleft;
  PredBlk *block_a;
  PredBlk *block_b;
  PredBlk *block_c;
  PredBlk *block_d;
  int sad;
  int min_sad;
  int best_mvshare_size[2];
  int mvshare_type;
       
  if(type==1)
  {
    left = top_left_blk;
    top = mf->get_pblk(i-1,j);
  }
  else
  {
    left = mf->get_pblk(i,j-1);
    top = top_left_blk;
  }
  tleft = mf->get_pblk(i-1, j-1);
   
  if(type==1)
  {
    block_a = mf->get_pblk(i+1,j);    // bottom
    block_b = mf->get_pblk(i+1, j-1); // bottomleft
    block_c = curblk;
    block_d = left;
  }
  else
  {
    block_a = mf->get_pblk(i,j+1);     // right
    block_b = curblk;
    block_c = mf->get_pblk(i-1,j+1);  // topright
    block_d = top;
  }
       
  mvshare_type = mf->mvshare_type_simple(curblk, top, left);
       
  curblk->mvshare_size[0]=curblk->mvshare_size[1]=0;
  best_mvshare_size[0]=best_mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
  min_sad = bit_cost(curblk, left, top, tleft);
  if(block_a) min_sad += bit_cost(block_a, block_b, block_c, block_d);
#else
  min_sad = bit_cost(curblk, i, j, left, top, tleft);
  if(block_a) min_sad += bit_cost(block_a, i+(type == 1), j+(type == 0), block_b, block_c, block_d);
#endif
  min_sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
       
  if(mvshare_type)
   {
         
    if(mvshare_type==1 || mvshare_type==4)
    {
      if(param.mvshare_two_enable_me)
      {
        curblk->mvshare_size[0]=2; 
        curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
        sad = bit_cost(curblk, left, top, tleft);
        if(block_a) sad += bit_cost(block_a, block_b, block_c, block_d);
#else
        sad = bit_cost(curblk, i, j, left, top, tleft);
        if(block_a) sad += bit_cost(block_a, i+(type == 1), j+(type == 0), block_b, block_c, block_d);
#endif
        if(sad < min_sad)
        {
          sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
          if(sad < min_sad)
          {
            min_sad = sad;
            best_mvshare_size[0]=2;
            best_mvshare_size[1]=0;
          }
        }
      }
     
      curblk->mvshare_size[0]=4; 
      curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
      sad = bit_cost(curblk, left, top, tleft);
      if(block_a) sad += bit_cost(block_a, block_b, block_c, block_d);
#else
      sad = bit_cost(curblk, i, j, left, top, tleft);
      if(block_a) sad += bit_cost(block_a, i+(type == 1), j+(type == 0), block_b, block_c, block_d);
#endif
      if(sad < min_sad)
      {
        sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
        if(sad < min_sad)
        {
          min_sad = sad;
          best_mvshare_size[0]=4;
          best_mvshare_size[1]=0;
        }
      }
    }
    if(mvshare_type==2 || mvshare_type==4)
    {
      if(param.mvshare_two_enable_me)
      {
        curblk->mvshare_size[0]=0; 
        curblk->mvshare_size[1]=2;
#ifndef MVCOMP_ZERORES
        sad = bit_cost(curblk, left, top, tleft);
        if(block_a) sad += bit_cost(block_a, block_b, block_c, block_d);
#else
        sad = bit_cost(curblk, i, j, left, top, tleft);
        if(block_a) sad += bit_cost(block_a, i+(type == 1), j+(type == 0), block_b, block_c, block_d);
#endif
        if(sad < min_sad)
        {
          sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
          if(sad < min_sad)
          {
            min_sad = sad;
            best_mvshare_size[0]=0;
            best_mvshare_size[1]=2;
          }
        }
      }
      
      curblk->mvshare_size[0]=0; 
      curblk->mvshare_size[1]=4;
#ifndef MVCOMP_ZERORES
      sad = bit_cost(curblk, left, top, tleft);
      if(block_a) sad += bit_cost(block_a, block_b, block_c, block_d);
#else
      sad = bit_cost(curblk, i, j, left, top, tleft);
      if(block_a) sad += bit_cost(block_a, i+(type == 1), j+(type == 0), block_b, block_c, block_d);
#endif
      if(sad < min_sad)
      {
        sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
        if(sad < min_sad)
        {
          min_sad = sad;
          best_mvshare_size[0]=0;
          best_mvshare_size[1]=4;
        }
      }
           
    }
         
  }

  best_mvshare_size[0]= best_mvshare_size[1]=0;

  curblk->mvshare_size[0] = best_mvshare_size[0];
  curblk->mvshare_size[1] = best_mvshare_size[1];
       
       
  return min_sad;
}


//-----------------------------------------------------------------------------
/*!
* \brief Row optimization of motion data
*
* Optimizes multiples rows of motion data.
* If all rows should be optimized, use first=0 and step=1.
* To optimize only 'odd' rows, use first=1 and step=2.
*
* \param[in] cur Current frame
* \param[in] first Index of first row to optimize
* \param[in] step Delta index of rows to optimize (should be positive)
* \param[in] list Set to BI_PRED if BI_PRED mode should be used
*
* \return TRUE if one of more blocks are changed, FALSE otherwise
*/
//-----------------------------------------------------------------------------
int MotionEstTrellis::treillisH(Frame *cur, int first, int step, int list) 
{
  int i,j,k,l;
  int cost_array[2][MAXNCAND];
  int *pcost, *cost;
  MCRefBlk8 blk;
  int tcost, bcost;
  int n;
  int back;
  PredBlk *top, *topleft;
  PredBlk *bottom, *bottomleft;
  PredBlk *curblk;
  int changed;
  
  PredBlk *left;
  
  cost = NULL;
  changed = FALSE;
  blk.ref_mgr       = ref_mgr;
  blk.mel_shift     = mel_shift;
  blk.chroma_weight = chroma_weight;
  
  for (i = first; i < mf->height; i += step) 
  {
    for (j = ZERO; j < mf->width; j++) 
    {
             
      n = insert_default_cands(mvcand[j], i, j, list);
        
      insert_neighbour_cands(mvcand[j], &n, i, j, 0, -1, MAX_NEIGHBOURS);
      insert_neighbour_cands(mvcand[j], &n, i, j, 0,  1, MAX_NEIGHBOURS);
        
      top = mf->get_pblk(i-1,j);
      if (top != NULL) 
      {
        insert_cand(mvcand[j], &n, top);
                                     
        insert_neighbour_cands(mvcand[j], &n, i-1, j, 0, -1, MAX_NEIGHBOURS);
        insert_neighbour_cands(mvcand[j], &n, i-1, j, 0,  1, MAX_NEIGHBOURS);
      }   
      bottom = mf->get_pblk(i+1,j);
      if (bottom != NULL) 
      {
        insert_cand(mvcand[j], &n, bottom);
          
        insert_neighbour_cands(mvcand[j], &n, i+1, j, 0, -1, MAX_NEIGHBOURS);
        insert_neighbour_cands(mvcand[j], &n, i+1, j, 0,  1, MAX_NEIGHBOURS);
      }
        
      ncand[j] = n;
        
    } 
    
    
    for (j = ZERO; j < mf->width; j++) 
    {
      pcost = cost;            
      cost  = cost_array[j&1]; 
      
      top        = mf->get_pblk(i-1,j  );
      topleft    = mf->get_pblk(i-1,j-1);
      bottom     = mf->get_pblk(i+1,j  );
      bottomleft = mf->get_pblk(i+1,j-1);
      
      left = mf->get_pblk(i,j-1);
      
      if (j == ZERO) 
      {
                  
          for (l = ZERO; l < ncand[j]; l++) 
          {
            
#ifndef MVCOMP_ZERORES
            cost[l] = bit_cost(&mvcand[ZERO][l], NULL, top, NULL);
            if (bottom != NULL)
              cost[l] += bit_cost(bottom, NULL, &mvcand[ZERO][l], NULL);
#else
            cost[l] = bit_cost(&mvcand[ZERO][l], i, j, NULL, top, NULL);
            if (bottom != NULL)
              cost[l] += bit_cost(bottom, i+ONE, j, NULL, &mvcand[ZERO][l], NULL);
#endif

            backtrack[ZERO][l] = ZERO;
          }
      }
      else
      { 

          for (l=0; l<ncand[j]; l++) 
          {
            curblk = &mvcand[j][l];
            bcost = pcost[0];
            
#ifndef MVCOMP_ZERORES
            bcost += bit_cost(curblk, &mvcand[j-1][0], top, topleft);
#else
            bcost += bit_cost(curblk, i, j, &mvcand[j-1][0], top, topleft);
#endif
            
            if (bottom)
#ifndef MVCOMP_ZERORES
              bcost += bit_cost(bottom, bottomleft, curblk, &mvcand[j-1][0]);
#else
              bcost += bit_cost(bottom, i+ONE, j, bottomleft, curblk, &mvcand[j-1][0]);
#endif

            back = 0;
            k = 0;
            
            while (TRUE) 
            {
              k++;
              tcost = pcost[k];
              
              if (tcost >= bcost)
                break;
              
#ifndef MVCOMP_ZERORES
              tcost += bit_cost(curblk, &mvcand[j-1][k], top, topleft);
#else
              tcost += bit_cost(curblk, i, j, &mvcand[j-1][k], top, topleft);
#endif
              
              if (tcost >= bcost)
                continue;
              
              if (bottom) 
              {
#ifndef MVCOMP_ZERORES
                tcost += bit_cost(bottom, bottomleft, curblk, &mvcand[j-1][k]);
#else
                tcost += bit_cost(bottom, i+ONE, j, bottomleft, curblk, &mvcand[j-1][k]);
#endif

                if (tcost >= bcost)
                  continue;
              }
              
              bcost = tcost;        
              back = k;
            } 
            
            backtrack[j][l] = back;
            cost[l] = bcost;
          }
      }
      
      blk.get_data(cur, i, j);
      
      for (k = ZERO; k < ncand[j]; k++)
      {
        if(param.motion_sharing_me_trellis)
        {
          if(param.motion_sharing_me_trellis_full)
          {
            if(j==0)
            {
#ifndef MVCOMP_ZERORES
              cost[k] -= bit_cost(&mvcand[j][k], left, top, topleft);
              if(bottom) cost[k] -= bit_cost(bottom, bottomleft, &mvcand[j][k], left);          
#else
              cost[k] -= bit_cost(&mvcand[j][k], i, j, left, top, topleft);
              if(bottom) cost[k] -= bit_cost(bottom, i+ONE, j, bottomleft, &mvcand[j][k], left);          
#endif
              cost[k] += total_cost_mvshare(1, i, j, &mvcand[j][k], NULL, mf, &blk, cur->frame_type==ft_B_frame);
            }
            else
            {
#ifndef MVCOMP_ZERORES
              cost[k] -= bit_cost(&mvcand[j][k], &mvcand[j-1][backtrack[j][k]], top, topleft);
              if(bottom) cost[k] -= bit_cost(bottom, bottomleft, &mvcand[j][k], &mvcand[j-1][backtrack[j][k]]);          
#else
              cost[k] -= bit_cost(&mvcand[j][k], i, j, &mvcand[j-1][backtrack[j][k]], top, topleft);
              if(bottom) cost[k] -= bit_cost(bottom, i+ONE, j, bottomleft, &mvcand[j][k], &mvcand[j-1][backtrack[j][k]]);          
#endif
              cost[k] += total_cost_mvshare(1, i, j, &mvcand[j][k], &mvcand[j-1][backtrack[j][k]], mf, &blk, cur->frame_type==ft_B_frame);
            }
          }
          else
          {
            if(k<BLIM)
            {
            if(j==0)
            {
#ifndef MVCOMP_ZERORES
              cost[k] -= bit_cost(&mvcand[j][k], left, top, topleft);
              if(bottom) cost[k] -= bit_cost(bottom, bottomleft, &mvcand[j][k], left);          
#else
              cost[k] -= bit_cost(&mvcand[j][k], i, j, left, top, topleft);
              if(bottom) cost[k] -= bit_cost(bottom, i+ONE, j, bottomleft, &mvcand[j][k], left);          
#endif
              cost[k] += total_cost_mvshare(1, i, j, &mvcand[j][k], NULL, mf, &blk, cur->frame_type==ft_B_frame);
            }
            else
            {
#ifndef MVCOMP_ZERORES
              cost[k] -= bit_cost(&mvcand[j][k], &mvcand[j-1][backtrack[j][k]], top, topleft);
              if(bottom) cost[k] -= bit_cost(bottom, bottomleft, &mvcand[j][k], &mvcand[j-1][backtrack[j][k]]);          
#else
              cost[k] -= bit_cost(&mvcand[j][k], i, j, &mvcand[j-1][backtrack[j][k]], top, topleft);
              if(bottom) cost[k] -= bit_cost(bottom, i+ONE, j, bottomleft, &mvcand[j][k], &mvcand[j-1][backtrack[j][k]]);          
#endif
              cost[k] += total_cost_mvshare(1, i, j, &mvcand[j][k], &mvcand[j-1][backtrack[j][k]], mf, &blk, cur->frame_type==ft_B_frame);
            }
            }
            else
            {
              cost[k] += blk.sad_w_intra_check(i, j, &mvcand[j][k]);
            }
          }
        }
        else
        {
          cost[k] += blk.sad_w_intra_check(i, j, &mvcand[j][k]);
        }
      }

      reorder(j, cost);
    }
    
    l = ZERO;
    j = mf->width - ONE;
    
    for (k=ONE; k<ncand[j]; k++)
    {
      if (cost[k] < cost[l]) l = k;
    }
    
    while (j >= ZERO) 
    {
      if (mvcand[j][l].equal(mf->get_pblk_unsafe(i,j)) == FALSE ) 
      {
        *mf->get_pblk_unsafe(i,j) = mvcand[j][l];
        if (mvcand[j][l].intra)
        {
          mf->get_pblk_unsafe(i,j)->intra = &mf->blk_array_I[i*mf->width+j];
        }

        changed = TRUE;
      }

      l = backtrack[j][l];
      j--;
    }
  }  
  return changed;
}

//-----------------------------------------------------------------------------
/*!
* \brief Column optimization of motion data
*
* Optimizes multiples columns of motion data.
* If all columns should be optimized, use first=0 and step=1.
* To optimize only 'odd' columns, use first=1 and step=2.
*
* \param[in] cur Current frame
* \param[in] first Index of first column to optimize
* \param[in] step Delta index of columns to optimize (should be positive)
* \param[in] list Set to BI_PRED if BI_PRED mode should be used
*
* \return TRUE if one of more blocks are changed, FALSE otherwise
*/
//-----------------------------------------------------------------------------
int MotionEstTrellis::treillisV(Frame *cur, int first, int step, int list)
{
  int i,j,k,l;
  int cost_array[TWO][MAXNCAND];
  int n = ZERO;
  MCRefBlk8 blk;
  int *pcost, *cost;
  int bcost,tcost;
  int back;
  PredBlk *left, *topleft;
  PredBlk *right, *topright;
  int changed;
  
  PredBlk *curblk;
  PredBlk *top;
  
  cost = NULL;
  changed = FALSE;
  blk.ref_mgr = ref_mgr;
  blk.mel_shift = mel_shift;
  blk.chroma_weight = chroma_weight;
  
  for (j=first; j<mf->width; j+=step) 
  {
    for (i=0; i<mf->height; i++) 
    {
              
      n = insert_default_cands(mvcand[i], i, j, list);
        
      insert_neighbour_cands(mvcand[i], &n, i, j, -1, 0, MAX_NEIGHBOURS);
      insert_neighbour_cands(mvcand[i], &n, i, j,  1, 0, MAX_NEIGHBOURS);
        
      left = mf->get_pblk(i,j-1);
      if (left != NULL) 
      {
        insert_cand(mvcand[i], &n, left);

        insert_neighbour_cands(mvcand[i], &n, i, j-1, -1, 0, MAX_NEIGHBOURS);
        insert_neighbour_cands(mvcand[i], &n, i, j-1,  1, 0, MAX_NEIGHBOURS);
      }
        
      right = mf->get_pblk(i,j+1);
      if (right != NULL) 
      {
        insert_cand(mvcand[i], &n, right);
          
        insert_neighbour_cands(mvcand[i], &n, i, j+1, -1, 0, MAX_NEIGHBOURS);
        insert_neighbour_cands(mvcand[i], &n, i, j+1,  1, 0, MAX_NEIGHBOURS);
      }
        
        ncand[i] = n;
        
    } 
    
    for (i=ZERO; i<mf->height; i++) 
    {
      pcost = cost;
      cost = cost_array[i&1];
      
      left = mf->get_pblk(i,j-1);
      topleft = mf->get_pblk(i-1,j-1);
      right = mf->get_pblk(i,j+1);
      topright = mf->get_pblk(i-1,j+1);
      
      top = mf->get_pblk(i-1,j);
      
      if (i == ZERO) 
      {
                  
        for (l=ZERO; l<ncand[ZERO]; l++) 
        {
#ifndef MVCOMP_ZERORES
          cost[l] = bit_cost(&mvcand[ZERO][l], left, NULL, NULL);
          if (right)
            cost[l] += bit_cost(right, &mvcand[ZERO][l], NULL, NULL);
#else
          cost[l] = bit_cost(&mvcand[ZERO][l], i, j, left, NULL, NULL);
          if (right)
            cost[l] += bit_cost(right, i, j+ONE, &mvcand[ZERO][l], NULL, NULL);
#endif

          backtrack[ZERO][l] = ZERO;
        }
        
      }
      else 
      {
                 
          for (l=0; l<ncand[i]; l++) 
          {
            
            curblk = &mvcand[i][l];
            bcost = pcost[0];
            
#ifndef MVCOMP_ZERORES
            bcost += bit_cost(&mvcand[i][l], left, &mvcand[i-1][0], topleft);
#else
            bcost += bit_cost(&mvcand[i][l], i, j, left, &mvcand[i-1][0], topleft);
#endif
            
            if (right)
#ifndef MVCOMP_ZERORES
              bcost += bit_cost(right, &mvcand[i][l], topright, &mvcand[i-1][0]);
#else
              bcost += bit_cost(right, i, j+ONE, &mvcand[i][l], topright, &mvcand[i-1][0]);
#endif

            back = 0;
            
            for (k=1; pcost[k]<bcost; k++)
            {
              tcost = pcost[k];
              
#ifndef MVCOMP_ZERORES
              tcost += bit_cost(&mvcand[i][l], left, &mvcand[i-1][k], topleft);
#else
              tcost += bit_cost(&mvcand[i][l], i, j, left, &mvcand[i-1][k], topleft);
#endif

              if (tcost >= bcost)
                continue;
              
              if (right)
#ifndef MVCOMP_ZERORES
                tcost += bit_cost(right, &mvcand[i][l], topright, &mvcand[i-1][k]);
#else
                tcost += bit_cost(right, i, j+ONE, &mvcand[i][l], topright, &mvcand[i-1][k]);
#endif
             
              if (tcost < bcost) 
              {
                bcost = tcost;
                back = k;
              }
            }
            
            backtrack[i][l] = back;
            cost[l] = bcost;
          }
          
      } 
      
      blk.get_data(cur, i, j);

      for (k = ZERO; k < ncand[i]; k++)
      {
        if(param.motion_sharing_me_trellis)
        {
          if(param.motion_sharing_me_trellis_full)
          {
            if(i==ZERO)
            {
#ifndef MVCOMP_ZERORES
              cost[k] -= bit_cost(&mvcand[i][k], left, top, topleft);
              if(right) cost[k] -= bit_cost(right, &mvcand[i][k], topright, top);
#else
              cost[k] -= bit_cost(&mvcand[i][k], i, j, left, top, topleft);
              if(right) cost[k] -= bit_cost(right, i, j+ONE, &mvcand[i][k], topright, top);
#endif
              cost[k] += total_cost_mvshare(2, i, j, &mvcand[i][k], NULL, mf, &blk, cur->frame_type==ft_B_frame);
            }
            else
            {
#ifndef MVCOMP_ZERORES
              cost[k] -= bit_cost(&mvcand[i][k], left, &mvcand[i-1][backtrack[i][k]], topleft);
              if(right) cost[k] -= bit_cost(right, &mvcand[i][k], topright, &mvcand[i-1][backtrack[i][k]]);
#else
              cost[k] -= bit_cost(&mvcand[i][k], i, j, left, &mvcand[i-1][backtrack[i][k]], topleft);
              if(right) cost[k] -= bit_cost(right, i, j+ONE, &mvcand[i][k], topright, &mvcand[i-1][backtrack[i][k]]);
#endif
              cost[k] += total_cost_mvshare(2, i, j, &mvcand[i][k], &mvcand[i-1][backtrack[i][k]], mf, &blk, cur->frame_type==ft_B_frame);
            }
          }
          else
          {
            if(k<BLIM)
            {
            if(i==ZERO)
            {
#ifndef MVCOMP_ZERORES
              cost[k] -= bit_cost(&mvcand[i][k], left, top, topleft);
              if(right) cost[k] -= bit_cost(right, &mvcand[i][k], topright, top);
#else
              cost[k] -= bit_cost(&mvcand[i][k], i, j, left, top, topleft);
              if(right) cost[k] -= bit_cost(right, i, j+ONE, &mvcand[i][k], topright, top);
#endif
              cost[k] += total_cost_mvshare(2, i, j, &mvcand[i][k], NULL, mf, &blk, cur->frame_type==ft_B_frame);
            }
            else
            {
#ifndef MVCOMP_ZERORES
              cost[k] -= bit_cost(&mvcand[i][k], left, &mvcand[i-1][backtrack[i][k]], topleft);
              if(right) cost[k] -= bit_cost(right, &mvcand[i][k], topright, &mvcand[i-1][backtrack[i][k]]);
#else
              cost[k] -= bit_cost(&mvcand[i][k], i, j, left, &mvcand[i-1][backtrack[i][k]], topleft);
              if(right) cost[k] -= bit_cost(right, i, j+ONE, &mvcand[i][k], topright, &mvcand[i-1][backtrack[i][k]]);
#endif
              cost[k] += total_cost_mvshare(2, i, j, &mvcand[i][k], &mvcand[i-1][backtrack[i][k]], mf, &blk, cur->frame_type==ft_B_frame);
            }
            }
            else
            {
              cost[k] += blk.sad_w_intra_check(i, j, &mvcand[i][k]);
            }
          }
        }
        else
        {
          cost[k] += blk.sad_w_intra_check(i, j, &mvcand[i][k]);
        }
      }

      reorder(i, cost);
    }
    
    l = ZERO;
    i = mf->height - ONE;
    
    for (k=ONE; k<ncand[i]; k++)
    {
      if (cost[k] < cost[l]) l = k;
    }
    
    while (i >= ZERO) 
    {        
      if (mvcand[i][l].equal(mf->get_pblk_unsafe(i,j)) == FALSE ) 
      {
        *mf->get_pblk_unsafe(i,j) = mvcand[i][l];
        changed = TRUE;
        if (mvcand[i][l].intra)
        {
            mf->get_pblk_unsafe(i,j)->intra = &mf->blk_array_I[i*mf->width+j];
        }
      }

      l = backtrack[i][l];
      i--;
    }   
  }
  return changed;
}

//-----------------------------------------------------------------------------
// Public methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Main estimation loop
*
* Several modes are supported as a function of the trellis_mode value from
* the configuration file:
* - Mode 0 optimizes all rows, then all columns.
* - Mode 1 optimizes all even rows, all even columns, all odd rows, and finally all odd columns.
* - Mode 2 optimizes all even rows, all odd rows, all even columns, and finally all odd columns.
*
* Multiple loop of optimization may be performed as determined by nloops. Early exit is
* provided when no change is made in one iteration.
*
* \param[in] cur Current frame
* \param[in] list Set to BI_PRED if BI_PRED mode should be used
*/
//-----------------------------------------------------------------------------
void MotionEstTrellis::estimate(Frame *cur, int list) 
{
  int i;
  int changed;
  
  for (i=ZERO; i<nloops; i++) 
  {
    switch (param.trellis_mode) 
    {
      case 0:
        changed  = treillisH(cur, ZERO, ONE, list);
        changed += treillisV(cur, ZERO, ONE, list);
        break;
        
      case 1:
        changed  = treillisH(cur, ZERO, TWO, list);
        changed += treillisV(cur, ZERO, TWO, list);
        changed += treillisH(cur,  ONE, TWO, list);
        changed += treillisV(cur,  ONE, TWO, list);
        break;
        
      case 2:
        changed  = treillisH(cur, ZERO, TWO, list);
        changed += treillisH(cur,  ONE, TWO, list);
        changed += treillisV(cur, ZERO, TWO, list);
        changed += treillisV(cur,  ONE, TWO, list);  
        break;
        
      default:
        changed = 0;
        break;
    }
    
    if (changed == FALSE)
      break;
  }

/*
  if(param.motion_sharing_me_trellis)
  {
    for (i = ZERO; i < mf->height; i++)
    {
      for (j = ZERO; j < mf->width; j++) 
      {
        mf->get_pblk(i, j)->mvshare_size[0]= mf->get_pblk(i, j)->mvshare_size[1]=0;
          } 
        } 
   }
*/

}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
