/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionEstUZSearchBiPred.cpp
*
* \brief Implementation of MotionEstUZSearchBiPred class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MotionEstUZSearchBiPred.H"
#include "Param.H"

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

#define I16_SQRT_HALF 46341 // 655536 * sqrt(0.5)
#define SDIAMOND_SP 4
#define SQUARE_SP 8
#define EDIAMOND_SP 12
#define LDIAMOND_SP 8
#define SBDIAMOND_SP 12

static int sdiamond_data[SDIAMOND_SP][4] = 
{
  {  0,  4, 3, 3 
  },
  {  4,  0, 0, 3 
  },
  {  0, -4, 1, 3 
  },
  { -4,  0, 2, 3 
  }
};

static int square_data[SQUARE_SP][4] = 
{
  {  0,  4, 7, 3 
  },
  {  4,  4, 7, 5 
  },
  {  4,  0, 1, 3 
  },
  {  4, -4, 1, 5 
  },
  {  0, -4, 3, 3 
  },
  { -4, -4, 3, 5 
  },
  { -4,  0, 5, 3 
  },
  { -4,  4, 5, 5 
  }
};

static int ediamond_data[EDIAMOND_SP][4] = 
{
  { -4,  4, 10, 5 
  },
  {  0,  8, 10, 8 
  },
  {  0,  4, 10, 7 
  },
  {  4,  4, 1, 5 
  },
  {  8,  0, 1, 8 
  },
  {  4,  0, 1, 7 
  },
  {  4, -4, 4, 5 
  },
  {  0, -8, 4, 8 
  },
  {  0, -4, 4, 7 
  },
  { -4, -4, 7, 5 
  },
  { -8,  0, 7, 8 
  },
  { -4,  0, 7, 7 
  }
};

static int ldiamond_data[LDIAMOND_SP][4] = 
{
  {  0,  8, 6, 5 
  },
  {  4,  4, 0, 3 
  },
  {  8,  0, 0, 5 
  },
  {  4, -4, 2, 3 
  },
  {  0, -8, 2, 5 
  },
  { -4, -4, 4, 3 
  },
  { -8,  0, 4, 5 
  },
  { -4,  4, 6, 3 
  }
};

static int sbdiamond_data[SBDIAMOND_SP][4] = 
{
  {  0,  8, 6, 12 
  },
  {  4,  4, 0, 12 
  },
  {  8,  0, 0, 12 
  },
  {  4, -4, 2, 12 
  },
  {  0, -8, 2, 12 
  },
  { -4, -4, 4, 12 
  },
  { -8,  0, 4, 12 
  },
  { -4,  4, 6, 12 
  },
  {  0,  2, 6, 12 
  },
  {  2,  0, 0, 12 
  },
  {  0, -2, 2, 12 
  },
  { -2,  0, 4, 12 
  }
};

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] mf0 Pointer to motion field
* \param[in] cur Pointer to current frame
* \param[in] ref_mgr Pointer to reference frame manager
*/
//-----------------------------------------------------------------------------
MotionEstUZSearchBiPred::MotionEstUZSearchBiPred(MotionField *mf, RefFrameManager *ref_mgr) : MotionEst(mf, ref_mgr)
{
  motionMap = NULL;
  row = new PredBlk[mf->width];  
}
 
void MotionEstUZSearchBiPred::reset(Frame *cur)
{
  int i;
  
  MotionEst::reset(cur);
  
  mvset_size = (TWO * me_range) * (TWO * me_range);  

  if (motionMap)
    delete[] motionMap;
  
  motionMap = new int[mvset_size];
  motionMapWidth = 2 * me_range;
  motionMapMinValue = -me_range;
  motionMapMaxValue = me_range - 1;

  for (i = ZERO; i < mvset_size; i++)
    motionMap[i] = ZERO;
  
  UnifiedZonalSearchInit(); 
}

bool MotionEstUZSearchBiPred::isInBounds(int val[2])
{
  return val[0] >= motionMapMinValue && val[0] <= motionMapMaxValue && val[1] >= motionMapMinValue && val[1] <= motionMapMaxValue;
}

bool MotionEstUZSearchBiPred::testAndMark(int val[2])
{
  int idx = (val[1] - motionMapMinValue) * motionMapWidth + (val[0] - motionMapMinValue);

  if (motionMap[idx] == uZonalSearchBlockIndex)
    return false;

  motionMap[idx] = uZonalSearchBlockIndex;

  return true;
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
MotionEstUZSearchBiPred::~MotionEstUZSearchBiPred()
{ 
  if (motionMap)
    delete[] motionMap;
  delete[] row;
  
}

//-----------------------------------------------------------------------------
// Private methods
//-----------------------------------------------------------------------------

void MotionEstUZSearchBiPred::forceIntegerPel(PredBlk *pblk)
{
  //assert(pblk != NULL && (pblk->list == LIST_0 || pblk->list == LIST_1));
  int logPel = ref_mgr->ref_process->inter_type[pblk->list][pblk->ref_index[pblk->list]];
  //assert(logPel >= 0 && logPel <= 3);
  pblk->mv[pblk->list][0] = shift_right_round(pblk->mv[pblk->list][0], logPel) << logPel;
  pblk->mv[pblk->list][1] = shift_right_round(pblk->mv[pblk->list][1], logPel) << logPel;
}

//-----------------------------------------------------------------------------
/*!
* \brief Insert a value into a list
*
* \note The value is inserted only if not already present in the list.
*
* \param[in] val Value to insert
* \param[in,out] list Array containing list
* \param[in,out] list_size Size of list
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::add_to_list(int val, int *list, int *list_size)
{
  int i;
  int n;
  
  n = *list_size;
  
  for (i = ZERO; i < n; i++)
    if (val == list[i])
      return;
  
  list[n++] = val;
  *list_size = n;
}

//-----------------------------------------------------------------------------
/*!
* \brief Initializes list of fixed predictors
*
* The fixed predictors lie on concentric octogons. The radii of the octogons
* increase logarithmically up to the search range.
*
* \param[in] search_range Search range
* \param[out] predictor List containing fixed predictors
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::UnifiedZonalSearchFixedPredictorInit(int search_range, FastMEPattern *predictor)
{ 
  int pos, i, j;
  int searchpos, sqsearchpos;
  int prednum;
  int ipos;
  int max_inter_pel;

  max_inter_pel = 1;

  for (i = 0; i < LISTS; i++)
  {
    for (j = 0; j < ref_mgr->lsize[i]; j++)
    {
      max_inter_pel = imax(max_inter_pel, ref_mgr->ref_process->inter_pel[i][j]);
    }
  }
  
  ipos = 0;
  while (power2(ipos + 1) <= search_range * search_range)
    ipos++;
  ipos = (ipos + 1) >> 1;
  
  prednum = ZERO;
  for (pos = ipos - TWO; pos >= ZERO; pos--)
  {
    searchpos   = max_inter_pel * (search_range >> pos);
    sqsearchpos = shift_right_round(searchpos * I16_SQRT_HALF, 16);
    
    for (i=-1; i<=1; i++)
      for (j=-1; j<=1; j++) 
      {
        switch(iabs(i) + iabs(j)) 
        {
          case ONE:
            predictor->point[prednum  ].x  = i * searchpos;
            predictor->point[prednum++].y  = j * searchpos;
            break;
          case TWO:
            predictor->point[prednum  ].x  = i * sqsearchpos;
            predictor->point[prednum++].y  = j * sqsearchpos;
            break;
        }
      }
  } 
  predictor->searchPoints = prednum;
}

//-----------------------------------------------------------------------------
/*!
* \brief Load a search pattern
*
* \param[out] pattern Search pattern
* \param[in] data Array describing the search pattern
* \param[in] searchPoints Number of points in pattern
* \param[in] stopSearch Flag indicating to stop search after current pattern
* \param[in] nextLast ??
* \param[in] nextpattern Pointer to next search pattern
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::assign_pattern(FastMEPattern *pattern, int data[][4], int searchPoints, int stopSearch, int nextLast, FastMEPattern *nextpattern)
{
  int i;
  
  for (i = ZERO; i < searchPoints; i++)
  {
    pattern->point[i].x           = data[i][0];
    pattern->point[i].y           = data[i][1];
    pattern->point[i].start_nmbr  = data[i][2];
    pattern->point[i].next_points = data[i][3];
  }
  pattern->searchPoints = searchPoints;
  pattern->stopSearch   = stopSearch;
  pattern->nextLast     = nextLast;
  pattern->nextpattern  = nextpattern;
}

//-----------------------------------------------------------------------------
/*!
* \brief Set search pattern
*
* \param[out] pattern Search pattern
* \param[in] shape Search pattern index
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::set_pattern(FastMEPattern *pattern, int shape)
{
  switch (shape) 
  {
    case 5:
      *pattern = sbdiamond;      
      break;
    case 4:
      *pattern = ldiamond;      
      break;
    case 3:
      *pattern = ediamond;      
      break;
    case 2:
      *pattern = square;      
      break;
    default:
      *pattern = sdiamond;      
      break;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Initialize search patterns
*
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::UnifiedZonalSearchInit()
{
  int i,j,max_inter_pel=1,max_log_inter_pel=0;
  
  uZonalSearchBlockIndex = ZERO;
  
  medianthres = 0;

  assign_pattern(&sdiamond,  sdiamond_data,  SDIAMOND_SP,  TRUE,  TRUE, &sdiamond);
  assign_pattern(&square,    square_data,    SQUARE_SP,    TRUE,  TRUE, &square  );
  assign_pattern(&ediamond,  ediamond_data,  EDIAMOND_SP,  TRUE,  TRUE, &ediamond);
  assign_pattern(&ldiamond,  ldiamond_data,  LDIAMOND_SP,  FALSE, TRUE, &sdiamond);
  assign_pattern(&sbdiamond, sbdiamond_data, SBDIAMOND_SP, FALSE, TRUE, &sdiamond);
  
  set_pattern(&searchPattern,  param.fastME    );
  set_pattern(&searchPatternD, param.fastMEDual);
  
  if (param.fastMEFixed)
  {
    for(i=0;i<LISTS;i++)
    {
      for (j = 0; j < ref_mgr->lsize[i]; j++)
      {
        max_inter_pel = max(max_inter_pel, ref_mgr->ref_process->inter_pel[i][j]);
      }
    }
    
    if (max_inter_pel== PEL_ONE_QUAD)
    {
      max_log_inter_pel = LOG_PEL_ONE_QUAD;
    }
    else if (max_inter_pel== PEL_ONE_EIGHTH)
    {
      max_log_inter_pel = LOG_PEL_ONE_EIGHTH;
    }
    else if(max_inter_pel == PEL_ONE_HALF)
    {
      max_log_inter_pel = LOG_PEL_ONE_HALF; 
    }
    else if(max_inter_pel == PEL_ONE_FULL)
    {
      max_log_inter_pel = LOG_PEL_ONE_FULL; 
    }
    
    UnifiedZonalSearchFixedPredictorInit(me_range >> max_log_inter_pel, &fixed_predictor);  
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Construct set of predictors
*
* Generates a set of predictors by adding the search pattern contained in
* fixed_predictor to a reference motion vector.
*
* \param[in] medianMV Reference motion data
* \param[in,out] predictor Set of predictors
* \param[in,out] prednum Number of predictors in set
* \param[in] list List for which to construct predictors (LIST_0 or LIST_1)
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::UnifiedZonalSearchFixedPredictors(PredBlk medianMV,FastMEPattern *predictor, int *prednum, int list) 
{ 
  int pos;
  
  for (pos = 0; pos < fixed_predictor.searchPoints; pos++) 
  {
    predictor->point[(*prednum)  ].x  = medianMV.mv[list][horizontal] + fixed_predictor.point[pos].x;
    predictor->point[(*prednum)++].y  = medianMV.mv[list][vertical  ] + fixed_predictor.point[pos].y;  
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Construct set of spatial predictors
*
* Scales motion vectors of the 4 provide motion data to match the provided
* reference frame index. The scaled motion vectors are then inserted into
* the set of predictors.
*
* \param[in] ref Reference frame index
* \param[in] block_a First motion data
* \param[in] block_b Second motion data
* \param[in] block_c Third motion data
* \param[in] block_d Fourth motion data
* \param[out] predictor Set of predictors
* \param[in] list List for which to construct predictors (LIST_0 or LIST_1)
*/
//-----------------------------------------------------------------------------
int MotionEstUZSearchBiPred::spatial_predictors(int ref, PredBlk  *block_a,PredBlk  *block_b, PredBlk  *block_c, PredBlk  *block_d, FastMEPattern *predictor, int list) 
{
  PredBlk  sblock_a,  sblock_b,  sblock_c,  sblock_d;
  
  sblock_a.ref_index[list] = ref;
  sblock_b.ref_index[list] = ref;
  sblock_c.ref_index[list] = ref;
  sblock_d.ref_index[list] = ref;
  
  block_a = sblock_a.rescale(ref_mgr, block_a, list);
  block_b = sblock_b.rescale(ref_mgr, block_b, list);
  block_c = sblock_c.rescale(ref_mgr, block_c, list);
  block_d = sblock_d.rescale(ref_mgr, block_d, list);  
  
  predictor->point[0].x  = 0;
  predictor->point[0].y  = 0;
  
  predictor->point[1].x  = (block_a != NULL) ? block_a->mv[list][horizontal] :  8;
  predictor->point[1].y  = (block_a != NULL) ? block_a->mv[list][vertical  ] :  0;    
  
  predictor->point[2].x  = (block_b != NULL) ? block_b->mv[list][horizontal] :  0;
  predictor->point[2].y  = (block_b != NULL) ? block_b->mv[list][vertical  ] :  8;
  
  predictor->point[3].x  = (block_c != NULL) ? block_c->mv[list][horizontal] : -8;
  predictor->point[3].y  = (block_c != NULL) ? block_c->mv[list][vertical  ] :  0;
  
  predictor->point[4].x  = (block_d != NULL) ? block_d->mv[list][horizontal] :  0;
  predictor->point[4].y  = (block_d != NULL) ? block_d->mv[list][vertical  ] : -8;    

  return 5;
}


//-----------------------------------------------------------------------------
/*!
* \brief Construct set of predictors from previous row
*
* \param[in] ref Reference frame index
* \param[in] i Vertical position of current 8x8 block
* \param[in] j Horizontal position of current 8x8 block
* \param[in,out] prednum Number of predictors in set
* \param[in,out] predictor Set of predictors
* \param[in] list List for which to construct predictors (LIST_0 or LIST_1)
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::row_predictors(int ref, int i,int j, int *prednum, FastMEPattern *predictor, int list) 
{
  
  if (i > 0) 
  {
    predictor->point[*prednum].x = clip(shift_right_round(ref_mgr->mv_scale[list][ref][row[j].ref_index[list]] * row[j].mv[list][vertical  ], RESCALE_BITS), -me_range, me_range-1);
    predictor->point[*prednum].y = clip(shift_right_round(ref_mgr->mv_scale[list][ref][row[j].ref_index[list]] * row[j].mv[list][horizontal], RESCALE_BITS), -me_range, me_range-1);
    *prednum += (predictor->point[(*prednum)].x != 0 || predictor->point[(*prednum)].y != 0);
    
    if (j < mf->width - 1) 
    {
      predictor->point[*prednum].x = clip(shift_right_round(ref_mgr->mv_scale[list][ref][row[j+1].ref_index[list]] * row[j+1].mv[list][vertical  ], RESCALE_BITS), -me_range, me_range-1);
      predictor->point[*prednum].y = clip(shift_right_round(ref_mgr->mv_scale[list][ref][row[j+1].ref_index[list]] * row[j+1].mv[list][horizontal], RESCALE_BITS), -me_range, me_range-1);
      *prednum += (predictor->point[*prednum].x != 0 || predictor->point[*prednum].y != 0);
    }
  }
  if (j>0)
  {
    predictor->point[*prednum].x = clip(shift_right_round(ref_mgr->mv_scale[list][ref][row[j-1].ref_index[list]] * row[j-1].mv[list][vertical  ], RESCALE_BITS), -me_range, me_range-1);
    predictor->point[*prednum].y = clip(shift_right_round(ref_mgr->mv_scale[list][ref][row[j-1].ref_index[list]] * row[j-1].mv[list][horizontal], RESCALE_BITS), -me_range, me_range-1);
    *prednum += (predictor->point[(*prednum)].x != 0 || predictor->point[(*prednum)].y != 0);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Search motion vector for single-list prediction
*
* \param[in] list List for which to search (LIST_0 or LIST_1)
* \param[in] block_a First motion data
* \param[in] block_b Second motion data
* \param[in] block_c Third motion data
* \param[in] block_d Fourth motion data
* \param[in] i Vertical position of current 8x8 block
* \param[in] j Horizontal position of current 8x8 block
* \param[in] blk Reference block to be matched
*
* \return Estimated motion data
*/
//-----------------------------------------------------------------------------
PredBlk MotionEstUZSearchBiPred::estimate_one_list(int list, PredBlk *block_a, PredBlk *block_b, PredBlk *block_c, PredBlk *block_d, int i, int j, MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], int u, int v, int n)
{
  int gsad, bsad;
  PredBlk gcand, cand, imv;
  int r;
  int active_ref_frames =  ref_mgr->lsize[list];
  int bk, bl;
  int temp_inter_pel;
  
  gsad = INT_INF;    
  cand.clear();
  cand.list = list;
  
  for (r = 0; r < active_ref_frames; r++) 
  {  
    uZonalSearchBlockIndex ++;
    imv = estimate(r, i, j, block_a, block_b, block_c, block_d, blk, &bsad, list, u, v, n);
    bk = 0;
    bl = 0;
    cand.ref_index[list] = imv.ref_index[list];        
    
    temp_inter_pel = ref_mgr->ref_process->inter_pel[list][r];

    switch (param.me_subpel) 
    {
      case 0:
        break;
      case 1: 
        if ((r == 0 || bsad < 2 * gsad))           
          search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>1, temp_inter_pel>>1, i, j, u, v, n);
        break;
      case 2: 
        if ((r == 0 || 2 * bsad < 3 * gsad))           
          search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>2, temp_inter_pel>>2, i, j, u, v, n);
        break;
      case 3: 
        if ((r == 0 || bsad < 2 * gsad)) 
        {          
          search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>1, temp_inter_pel>>1, i, j, u, v, n);
          if ((r == 0 || 2 * bsad < 3 * gsad))           
            search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>2, temp_inter_pel>>2, i, j, u, v, n);
        }
        break;
      case 4: 
        if ((r == 0 || bsad < gsad))          
          search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>3, temp_inter_pel>>3, i, j, u, v, n);  
        break;
      case 5: 
        if ((r == 0 || bsad < 2 * gsad)) 
        {      
          search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>1, temp_inter_pel>>1, i, j, u, v, n);
          if ((r == 0 || bsad < gsad))          
            search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>3, temp_inter_pel>>3, i, j, u, v, n);  
        }
        break;
      case 6: 
        if ((r == 0 || 2 * bsad < 3 * gsad)) 
        {           
          search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>2, temp_inter_pel>>2, i, j, u, v, n);
          if ((r == 0 || bsad < gsad))         
            search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>3, temp_inter_pel>>3, i, j, u, v, n);  
        }
        break;
      case 7: 
        if ((r == 0 || bsad < 2 * gsad)) 
        {
          search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>1, temp_inter_pel>>1, i, j, u, v, n);
          if ((r == 0 || 2 * bsad < 3 * gsad)) 
          {          
            search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>2, temp_inter_pel>>2, i, j, u, v, n);
            if ((r == 0 || bsad < gsad))          
              search_subpel(list, &imv, &cand, block_a, block_b, block_c, blk, &bsad, temp_inter_pel>>3, temp_inter_pel>>3, i, j, u, v, n);  
          }
        }
        break;
      default:
        printf("Error: unknown subpel search type me_subpel in MotionEstUZSearchBiPred::estimate_one_list : %d\n",param.me_subpel);
        break;
    }
    
    cand.mv[list][0] = imv.mv[list][0];
    cand.mv[list][1] = imv.mv[list][1];
    
    if (bsad < gsad) 
    {
      gsad = bsad;
      gcand = cand;
      gcand.gsad = gsad;
    }        
  }
  return gcand;
}

//-----------------------------------------------------------------------------
/*!
* \brief Search motion vector for subpel position 
*
* \param[in] list List for which to search (LIST_0 or LIST_1)
* \param[in,out] imv Estimated motion vector?
* \param[in] cand Motion data of Candidate block
* \param[in] block_a First motion data
* \param[in] block_b Second motion data
* \param[in] block_c Third motion data
* \param[in] blk Reference block to be matched
* \param[in] Best cost observed so far
* \param[in] k_range Vertical size of inter pel
* \param[in] l_range Horizontal size of inter pel
* \param[in] i Vertical position of current 8x8 block
* \param[in] j Horizontal position of current 8x8 block
*
* \return Estimated motion data for subpel position
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::search_subpel(int list, PredBlk *imv, PredBlk *cand, PredBlk *block_a, PredBlk *block_b, PredBlk * /*block_c*/, MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], int *bsad, int k_range, int l_range, int i, int j, int u, int v, int n)
{
  int bk=0, bl=0;
  int k,l;
  int sad;
  int Bframe;

  Bframe = ((list==BI_PRED) || (block_a && block_a->list==BI_PRED) || (block_b && block_b->list==BI_PRED));

  if (k_range > 0 || l_range > 0) 
  {
    for (k = -k_range; k <= k_range; k += k_range) 
    {
      for (l = -l_range; l <= l_range; l += l_range) 
      {  
        if (k == ZERO && l == ZERO)
          continue;
        
        cand->mv[list][0] = imv->mv[list][0] + k;
        cand->mv[list][1] = imv->mv[list][1] + l;
        
        if (isInBounds(cand->mv[list]) == false)
          continue;
        if (testAndMark(cand->mv[list]) == false)
          continue;
        sad = bitcostN(cand, i, j, u, v, n);
        sad += sadN(blk, cand, i, j, u, v, n);
        
        if (sad < *bsad)
        {
          *bsad = sad;
          bk = k;
          bl = l;
        }
      }
    }

    imv->mv[list][0] = imv->mv[list][0] + bk;
    imv->mv[list][1] = imv->mv[list][1] + bl;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Search motion vector for single-list prediction and a single reference frame
*
* \note This method name was changed from estimate to avoid name aliasing.
*
* \param[in] ref Reference frame index
* \param[in] i Vertical position of current 8x8 block
* \param[in] j Horizontal position of current 8x8 block
* \param[in] block_a First motion data
* \param[in] block_b Second motion data
* \param[in] block_c Third motion data
* \param[in] block_d Fourth motion data
* \param[in] blk Reference block to be matched
* \param[out] bs SAD of estimated motion data
* \param[in] list List for which to search (LIST_0 or LIST_1)
*
* \return Estimated motion data
*/
//-----------------------------------------------------------------------------
PredBlk MotionEstUZSearchBiPred::estimate(int ref, int i, int j, PredBlk *block_a, PredBlk *block_b, PredBlk *block_c, PredBlk *block_d, MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], int *bs, int list, int u, int v, int n) 
{
  int sad, bsad = INT_INF, sbsad = INT_INF;
  PredBlk medianMV, curr_mv, center_mv, best_mv, sbest_mv;
  int noblock_a, noblock_b, noblock_c, noblock_d;
  int pos = 0, prednum = 0;

  FastMEPattern *searchPatternF; 
  int threshold = 4*medianthres;
  
  int patternStop = 0, pointNumber = 0, checkPts, nextLast = 0;
  int totalCheckPts = 0, motionDirection = 0;

  int Bframe;

  Bframe = (list==BI_PRED) || (block_a && block_a->list==BI_PRED) || (block_b && block_b->list==BI_PRED);
  
  noblock_a = (block_a == NULL);
  noblock_b = (block_b == NULL);
  noblock_c = (block_c == NULL);
  noblock_d = (block_d == NULL);
  
  medianMV.ref_index[list] = ref;
  
  if (noblock_a || (noblock_b == 0 && (block_a->ref_index[list] != ref && block_b->ref_index[list] == ref))) 
  {
    if (noblock_b) 
    {
      if (noblock_d) 
      {
        medianMV.mv[list][0] = 0;
        medianMV.mv[list][1] = 0;
      }
      else 
      {
        medianMV.mv[list][0] = block_d->mv[list][0];
        medianMV.mv[list][1] = block_d->mv[list][1];
      }
    }
    else 
    {
      medianMV.mv[list][0] = block_b->mv[list][0];
      medianMV.mv[list][1] = block_b->mv[list][1];
    }
  }
  else if (noblock_b || block_a->equal(block_b, list) || (block_b->ref_index[list] != ref && block_a->ref_index[list] == ref)) 
  {
    medianMV.mv[list][0] = block_a->mv[list][0];
    medianMV.mv[list][1] = block_a->mv[list][1];
  }
  else 
  {
    if (noblock_d && noblock_c) 
    {
      medianMV.mv[list][0] = block_a->mv[list][0];
      medianMV.mv[list][1] = block_a->mv[list][1];
    }
    else if (noblock_d) 
    {
      if (block_c->ref_index[list] == ref && block_b->ref_index[list] != ref && block_a->ref_index[list] != ref) 
      {
        medianMV.mv[list][0] = block_c->mv[list][0];
        medianMV.mv[list][1] = block_c->mv[list][1];
      }
      else 
      {
        medianMV.mv[list][0] = median(block_a->mv[list][0], block_b->mv[list][0], block_c->mv[list][0]);
        medianMV.mv[list][1] = median(block_a->mv[list][1], block_b->mv[list][1], block_c->mv[list][1]);
      }
    }
    else 
    {
      if (block_d->ref_index[list] == ref && block_b->ref_index[list] != ref && block_a->ref_index[list] != ref) 
      {
        medianMV.mv[list][0] = block_d->mv[list][0];
        medianMV.mv[list][1] = block_d->mv[list][1];
      }
      else 
      {
        medianMV.mv[list][0] = median(block_a->mv[list][0], block_b->mv[list][0], block_d->mv[list][0]);
        medianMV.mv[list][1] = median(block_a->mv[list][1], block_b->mv[list][1], block_d->mv[list][1]);
      }
    }
  }
  
  forceIntegerPel(&medianMV);
  if (isInBounds(medianMV.mv[list]) == false)
  {
    medianMV.mv[list][0] = 0;
    medianMV.mv[list][1] = 0;
  }

  curr_mv.mv[list][0] = medianMV.mv[list][0];
  curr_mv.mv[list][1] = medianMV.mv[list][1];
  curr_mv.ref_index[list] = ref;
  curr_mv.list = medianMV.list = list;
  
  sad = bitcostN(&curr_mv, i, j, u, v, n);
  sad += sadN(blk, &curr_mv, i, j, u, v, n);

  best_mv.list = sbest_mv.list = list;
  best_mv.ref_index[list] = sbest_mv.ref_index[list] = ref;
  best_mv.mv[list][0]  = curr_mv.mv[list][0];
  best_mv.mv[list][1]  = curr_mv.mv[list][1];
  sbest_mv.mv[list][0] = 0;
  sbest_mv.mv[list][1] = 0;   
  bsad = sad;
  testAndMark(curr_mv.mv[list]);

  if (bsad > medianthres) // Do more work if medianMV above is not good enough
  {
    // Generate a set of candidates from neighbouring blocks
    prednum = spatial_predictors(ref, block_a, block_b, block_c, block_d, &predictor, list);        
    
    // Add a set of candidates from neighouring blocks using opposite list
    if (mf->frame_type == ft_B_frame)
      row_predictors(ref, i, j, &prednum, &predictor, list);

    {
      int k, l;

      for (k = -1; k <= 1; k++)
      {
        for (l = -1; l <= 1; l++)
        {
          PredBlk *a = prevMf->get_pblk(i + k, j + l);
          if (a)
          {
            predictor.point[prednum].x = shift_right_round(a->mv[LIST_0][1] * ref_mgr->dist[list][ref], 12);
            predictor.point[prednum].y = shift_right_round(a->mv[LIST_0][0] * ref_mgr->dist[list][ref], 12);
            prednum++;
            predictor.point[prednum].x = shift_right_round(a->mv[LIST_1][1] * ref_mgr->dist[list][ref], 12);
            predictor.point[prednum].y = shift_right_round(a->mv[LIST_1][0] * ref_mgr->dist[list][ref], 12);
            prednum++;
          }
        }
      }
    }

    if (!noblock_a)
      threshold = imin(threshold, block_a->gsad);
    if (!noblock_b)
      threshold = imin(threshold, block_b->gsad);
    if (!noblock_c)
      threshold = imin(threshold, block_c->gsad);
    if (!noblock_d)
      threshold = imin(threshold, block_d->gsad);
    threshold = imax(threshold, medianthres);
    
    // Add a set candidates using a fixed pattern of growing diamonds around medianMV
    if (param.fastMEFixed)
      UnifiedZonalSearchFixedPredictors(medianMV, &predictor, &prednum, list);
    
    // Test all candidates derived above, and keep best and second best
    for (pos = 0; pos < prednum; pos++) 
    {
      curr_mv.mv[list][horizontal] = predictor.point[pos].x;
      curr_mv.mv[list][vertical  ] = predictor.point[pos].y;
      
      forceIntegerPel(&curr_mv);
      if (isInBounds(curr_mv.mv[list]) == false)
        continue;
      if (testAndMark(curr_mv.mv[list]) == false)
        continue;

      sad = bitcostN(&curr_mv, i, j, u, v, n);
      
      if (sad < sbsad) 
      {
        sad += sadN(blk, &curr_mv, i, j, u, v, n);

        if (sad < bsad) 
        {
          // Best becomes second-best
          sbsad = bsad;
          sbest_mv.mv[list][0] = best_mv.mv[list][0];
          sbest_mv.mv[list][1] = best_mv.mv[list][1];
          
          // Set new best
          bsad = sad;
          best_mv.mv[list][0]  = curr_mv.mv[list][0];
          best_mv.mv[list][1]  = curr_mv.mv[list][1];
        }
        else if (sad < sbsad && (best_mv.mv[list][0] != curr_mv.mv[list][0] || best_mv.mv[list][1] != curr_mv.mv[list][1])) 
        {
          // Set second-best
          sbsad = sad;
          sbest_mv.mv[list][0] = curr_mv.mv[list][0];
          sbest_mv.mv[list][1] = curr_mv.mv[list][1];
        }          
      }
    }
  
    if (bsad > threshold)
    {
      if ((param.fastME!=1) && (2 * bsad< 5 * medianthres)) 
      {
        if ((best_mv.mv[list][0] == 0 && best_mv.mv[list][1] == 0))
          searchPatternF=&sdiamond;
        else
          searchPatternF=&square;      
      } 
      else
        searchPatternF=&searchPattern;      
      
      totalCheckPts = searchPatternF->searchPoints;                 
      
      center_mv.mv[list][0] = best_mv.mv[list][0];    
      center_mv.mv[list][1] = best_mv.mv[list][1];    
      
      do 
      {
        checkPts = totalCheckPts;
        bool foundBetterMv = false;
        do 
        {            
          curr_mv.mv[list][horizontal] = center_mv.mv[list][horizontal] + searchPatternF->point[pointNumber].x;
          curr_mv.mv[list][vertical  ] = center_mv.mv[list][vertical  ] + searchPatternF->point[pointNumber].y;

          forceIntegerPel(&curr_mv); // Shouldn't be needed but one never knows...
          if (isInBounds(curr_mv.mv[list]))
          {
            if (testAndMark(curr_mv.mv[list]) == false)
            {
              pointNumber += 1;
              if (pointNumber >= searchPatternF->searchPoints)
                pointNumber -= searchPatternF->searchPoints;
              checkPts -= 1;
              continue;
            }
           
            sad = bitcostN(&curr_mv, i, j, u, v, n);
            
            if (sad < bsad) 
            {              
              sad += sadN(blk, &curr_mv, i, j, u, v, n);

              if (sad < bsad) 
              {
                bsad = sad;
                best_mv.mv[list][0] = curr_mv.mv[list][0];
                best_mv.mv[list][1] = curr_mv.mv[list][1];
                motionDirection = pointNumber;
                foundBetterMv = true;
              }
            }            
          }
          pointNumber += 1;
          if (pointNumber >= searchPatternF->searchPoints)
            pointNumber -= searchPatternF->searchPoints;
          checkPts -= 1;
        }
        while (checkPts > 0);
      
        if (nextLast || !foundBetterMv)
        {
          patternStop     = searchPatternF->stopSearch;
          searchPatternF  = searchPatternF->nextpattern; 
          totalCheckPts   = searchPatternF->searchPoints;
          nextLast        = searchPatternF->nextLast;
          motionDirection = 0;
          pointNumber     = 0;
        }
        else 
        {
          totalCheckPts = searchPatternF->point[motionDirection].next_points;
          pointNumber =searchPatternF->point[motionDirection].start_nmbr;
          center_mv.mv[list][0] = best_mv.mv[list][0];
          center_mv.mv[list][1] = best_mv.mv[list][1];         
        }
      }
      while (patternStop != 1);
  
      // Do refinement around second-best MV using dual pattern
      if (param.fastMEDual && bsad > threshold) 
      {
        pointNumber     = 0; 
        motionDirection = 0;
        patternStop     = 0;
        nextLast        = 0;
        
        if ((sbest_mv.mv[list][0] == 0 && sbest_mv.mv[list][1] == 0))
          searchPatternF=&sdiamond;      
        else
          searchPatternF=&searchPatternD;      
        
        totalCheckPts=searchPatternF->searchPoints;
        
        center_mv.mv[list][0] = sbest_mv.mv[list][0];
        center_mv.mv[list][1] = sbest_mv.mv[list][1];
        
        do 
        {
          checkPts = totalCheckPts;
          bool foundBetterMv = false;
          do 
          {            
            curr_mv.mv[list][horizontal] = center_mv.mv[list][horizontal] + searchPatternF->point[pointNumber].x;
            curr_mv.mv[list][vertical  ] = center_mv.mv[list][vertical  ] + searchPatternF->point[pointNumber].y;
            
            forceIntegerPel(&curr_mv);
            if (isInBounds(curr_mv.mv[list]))
            {
              if (testAndMark(curr_mv.mv[list]) == false)
              {                
                pointNumber += 1;
                if (pointNumber >= searchPatternF->searchPoints)
                  pointNumber -= searchPatternF->searchPoints;
                checkPts -= 1;
                continue;
              }
              
              sad = bitcostN(&curr_mv, i, j, u, v, n);
              
              if (sad < sbsad) 
              {             
                sad += sadN(blk, &curr_mv, i, j, u, v, n);
                if (sad < sbsad) 
                {
                  sbsad = sad;
                  sbest_mv.mv[list][0] = curr_mv.mv[list][0];
                  sbest_mv.mv[list][1] = curr_mv.mv[list][1];
                  foundBetterMv = true;
                  motionDirection = pointNumber;
                }
              }              
            }
            pointNumber += 1;
            if (pointNumber >= searchPatternF->searchPoints)
              pointNumber -= searchPatternF->searchPoints;
            checkPts -= 1;
          }
          while (checkPts > 0);
          
          if (nextLast || !foundBetterMv)
          {
            patternStop     = searchPatternF->stopSearch;
            searchPatternF  = searchPatternF->nextpattern; 
            totalCheckPts   = searchPatternF->searchPoints;
            nextLast        = searchPatternF->nextLast;
            motionDirection = 0;
            pointNumber     = 0;
          }
          else 
          {
            totalCheckPts = searchPatternF->point[motionDirection].next_points;
            pointNumber =searchPatternF->point[motionDirection].start_nmbr;

            center_mv.mv[list][0] = sbest_mv.mv[list][0];
            center_mv.mv[list][1] = sbest_mv.mv[list][1];
          }
        }
        while (patternStop != 1);                
        if (sbsad < bsad)
        {
          bsad = sbsad;
          best_mv.mv[list][0] = sbest_mv.mv[list][0];
          best_mv.mv[list][1] = sbest_mv.mv[list][1];
        }
      }      
    }
  }
  
  *bs = bsad;

  return best_mv;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute rate-distortion cost of all blocks affected by changing current block
*
* Terminates early if partial cost exceed bsad.
*
* \param[in] i Vertical 8x8 block position in a frame
* \param[in] j Horizontal 8x8 block position in a frame
* \param[in] sadin Threshold for early exit
* \param[in] curblk Motion data of current block
* \param[in] mf Pointer to motion field
* \param[in] blk Pointer to MC reference block
* \param[in] BiPred flag to indicate whether frame type is ft_B_frame
*
* \return Rate-Distortion cost
*/
//-----------------------------------------------------------------------------
int MotionEstUZSearchBiPred::total_cost_mvshare(int i, int j, int sadin, PredBlk *curblk, MotionField *mf, MCRefBlk8 *blk, int Bframe)
{
  PredBlk *top;
  PredBlk *left;
  PredBlk *tleft;
  int sad;
  int min_sad=sadin;
  int best_mvshare_size[2];
  int mvshare_type;
   
  left = mf->get_pblk(i,j-1);
  top = mf->get_pblk(i-1,j);
  tleft = mf->get_pblk(i-1, j-1);
   
  mvshare_type = mf->mvshare_type_simple(curblk, top, left);
   
  curblk->mvshare_size[0]=curblk->mvshare_size[1]=0;
  best_mvshare_size[0]=best_mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
  if (param.trans8x8_enable) sad = bit_cost(curblk, left, top, tleft);
  else sad = bit_cost(curblk, left, top, tleft);
#else
  if (param.trans8x8_enable) sad = bit_cost(curblk, i, j, left, top, tleft);
  else sad = bit_cost(curblk, i, j, left, top, tleft);
#endif
  if(sad < min_sad)
  {
     sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
     if(sad < min_sad) min_sad = sad;
  }
   
  if(mvshare_type)
  {
     
    if(mvshare_type==1 || mvshare_type==4)
    {
      if(param.mvshare_two_enable_me)
      {
        curblk->mvshare_size[0]=2; 
        curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
        sad = bit_cost(curblk, left, top, tleft);
#else
        sad = bit_cost(curblk, i, j, left, top, tleft);
#endif

        if(sad < min_sad)
        {
          sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
          if(sad < min_sad)
          {
            min_sad = sad;
            best_mvshare_size[0]=2;
            best_mvshare_size[1]=0;
          }
        }
      }
       
      curblk->mvshare_size[0]=4; 
      curblk->mvshare_size[1]=0;
#ifndef MVCOMP_ZERORES
      sad = bit_cost(curblk, left, top, tleft);
#else
      sad = bit_cost(curblk, i, j, left, top, tleft);
#endif

      if(sad < min_sad)
      {
        sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
        if(sad < min_sad)
        {
          min_sad = sad;
          best_mvshare_size[0]=4;
          best_mvshare_size[1]=0;
        }
      }

    }
    if(mvshare_type==2 || mvshare_type==4)
    {
      if(param.mvshare_two_enable_me)
      {
        curblk->mvshare_size[0]=0; 
        curblk->mvshare_size[1]=2;
#ifndef MVCOMP_ZERORES
        sad = bit_cost(curblk, left, top, tleft);
#else
        sad = bit_cost(curblk, i, j, left, top, tleft);
#endif

        if(sad < min_sad)
        {
          sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
          if(sad < min_sad)
          {
            min_sad = sad;
            best_mvshare_size[0]=0;
            best_mvshare_size[1]=2;
          }
        }
      }
       
      curblk->mvshare_size[0]=0; 
      curblk->mvshare_size[1]=4;
#ifndef MVCOMP_ZERORES
      sad = bit_cost(curblk, left, top, tleft);
#else
      sad = bit_cost(curblk, i, j, left, top, tleft);
#endif

      if(sad < min_sad)
      {
        sad += blk->distortion_mvshare(i, j, curblk, top, left, Bframe);
        if(sad < min_sad)
        {
          min_sad = sad;
          best_mvshare_size[0]=0;
          best_mvshare_size[1]=4;
        }
      }

    }
     
  }
  best_mvshare_size[0]= best_mvshare_size[1]=0;

  curblk->mvshare_size[0] = best_mvshare_size[0];
  curblk->mvshare_size[1] = best_mvshare_size[1];
   
   
  return min_sad;
}


//-----------------------------------------------------------------------------
// Public methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Estimate motion field
*
* Iterates through every homogeneous motion region in a frame and tries to find a set of motion
* data that lowers the R-D cost.
*
* \param[in] cur Current frame
* \param[in] list Estimation mode (BI_PRED or not)
*/
//-----------------------------------------------------------------------------
void MotionEstUZSearchBiPred::estimate(Frame *cur, int list)
{
  MCRefBlk8 blk[MAXMEBLK][MAXMEBLK];

  int i,j;
  int k,l,m;
  int gsad;
  int sad;
  int bk, bl;
  int bk0, bk1, bl0;
  PredBlk *block_a, *block_b, *block_c;
  PredBlk cand;
  int enable_intcomp;
  int u, v;

  MCRefBlk8 blk2;
  
  PredBlk tmp_pblk;
  
  for (u = 0; u < MAXMEBLK; u++)
  {
    for (v = 0; v < MAXMEBLK; v++)
    {
      blk[u][v].ref_mgr = ref_mgr;
      blk[u][v].mel_shift = mel_shift;
      blk[u][v].chroma_weight = chroma_weight;
    }
  }
  
  blk2.ref_mgr = ref_mgr;
  blk2.mel_shift = mel_shift;
  blk2.chroma_weight = chroma_weight;
  
  enable_intcomp = param.enable_intcomp;
  param.enable_intcomp = FALSE;
  
  for (i=0; i<mf->height; i += param.lime_size) 
  {
    for (j=0; j<mf->width; j += param.lime_size) 
    {
      for (u = 0; u < imin(mf->height - i, param.lime_size); u++)
      {
        for (v = 0; v < imin(mf->width - j, param.lime_size); v++)
        {
          blk[u][v].get_data(cur, i + u, j + v);
        }
      }
      
      estimateR(i, j, blk, 0, 0, param.lime_size, list, cur);
    }
  }

#if 1
    param.enable_intcomp = enable_intcomp;

    if (param.enable_intcomp) 
    {
      MCRefBlk8 blk;
  blk.ref_mgr = ref_mgr;
  blk.mel_shift = mel_shift;
  blk.chroma_weight = chroma_weight;

      for (i=0; i<mf->height; i++) 
      {
        for (j=0; j<mf->width; j++) 
        {
          int sum1, sum2, sum3;
          int lcand[8];
          int ncand;
          int s0,s1;
          int val;
          int cur_list;     
        
          blk.get_data(cur, i, j);
       

          block_b = mf->get_pblk(i-1,j);
          block_a = mf->get_pblk(i,j-1);
          block_c = mf->get_pblk(i-1,j-1);

          cand = *mf->get_pblk(i,j);
          int minDist = blk.distortion(i, j, INT_INF, &cand);
        
          cur_list = cand.list; 
       
          if (cur_list == BI_PRED) 
          {
          
            bk0 = ZERO;
            bl0 = ZERO;
            bk1 = ZERO;
 
            if(param.motion_sharing_me_init_offset)
            {
              gsad = total_cost_mvshare(i, j, INT_INF, &cand, mf, &blk, cur->frame_type==ft_B_frame);
            }
            else
            {         
#ifndef MVCOMP_ZERORES
              gsad = bit_cost(&cand, block_a, block_b, block_c);
#else
              gsad = bit_cost(&cand, i, j, block_a, block_b, block_c);
#endif
              gsad += blk.distortion(i, j, INT_INF, &cand);
            }
          
            sum1 = blk.sum_luma;
            sum2 = blk.get_mc_sum_luma(i, j, &cand, LIST_0);
            sum3 = blk.get_mc_sum_luma(i, j, &cand, LIST_1);
          
            for (k=-power2(SAMPLE_ADD_PREC); k <= power2(SAMPLE_ADD_PREC); k++) 
            {
              ncand = ZERO;
            
              add_to_list(ZERO, lcand, &ncand);
            
              if (block_b != NULL)
                add_to_list(block_b->BP_DC, lcand, &ncand);
            
              if (block_a != NULL)
                add_to_list(block_a->BP_DC, lcand, &ncand);
            
              s0 = (k + power2(SAMPLE_ADD_PREC)) >> 1;
              s1 = (k + power2(SAMPLE_ADD_PREC) + 1) >> 1;
            
              val = (power2(SAMPLE_ADD_PREC) * sum1 - sum2 * s0 - sum3 * s1) >> (SAMPLE_ADD_PREC + 6);
              add_to_list(val, lcand, &ncand);
              add_to_list(val + ONE, lcand, &ncand);
            
              cand.BP_scale[0] = k;
              cand.BP_scale[1] = ZERO;
            
              for (l=0; l<ncand; l++) 
              {
                cand.BP_DC = lcand[l];
              
                if (cand.inbounds_wp() == FALSE)
                  continue;

                if(param.motion_sharing_me_init_offset)
                {
                  sad = total_cost_mvshare(i, j, gsad, &cand, mf, &blk, cur->frame_type==ft_B_frame);
                }
                else
                {              
                  sad = blk.distortion(i, j, gsad, &cand);
                  if (sad >= minDist)
                    continue;
              
                  if (sad >= gsad) 
                    continue;
#ifndef MVCOMP_ZERORES
                  sad += bit_cost(&cand, block_a, block_b, block_c);
#else
                  sad += bit_cost(&cand, i, j, block_a, block_b, block_c);
#endif
                } 
              
                if (sad < gsad)
                {
                  gsad = sad;
                  bk0 = k;
                  bk1 = ZERO;
                  bl0 = cand.BP_DC;
                }
              }
            } 
          
            for (m=-power2(SAMPLE_ADD_PREC); m<=power2(SAMPLE_ADD_PREC); m++) 
            {
              ncand = 0;
            
              add_to_list(ZERO, lcand, &ncand);
            
              if (block_b != NULL)
                add_to_list(block_b->BP_DC, lcand, &ncand);
            
              if (block_a != NULL)
                add_to_list(block_a->BP_DC, lcand, &ncand);
            
              cand.BP_scale[0] = ZERO;
              cand.BP_scale[1] = m;
              s0 = (cand.get_B_scale_sum() >> ONE) + cand.get_B_scale_diff();
              s1 = cand.get_B_scale_sum() - s0;
            
              val = (power2(SAMPLE_ADD_PREC) * sum1 - sum2 * s0 - sum3 * s1) >> (SAMPLE_ADD_PREC + 6);
              add_to_list(val, lcand, &ncand);
              add_to_list(val + ONE, lcand, &ncand);
            
              for (l=0; l<ncand; l++) 
              {
                cand.BP_DC = lcand[l]; 
              
                if (cand.inbounds_wp() == FALSE)
                  continue;

                if(param.motion_sharing_me_init_offset)
                {
                  sad = total_cost_mvshare(i, j, gsad, &cand, mf, &blk, cur->frame_type==ft_B_frame);
                }
                else
                {              
                  sad = blk.distortion(i, j, gsad, &cand);

                  if (sad >= minDist)
                    continue;

#ifndef MVCOMP_ZERORES
                  sad += bit_cost(&cand, block_a, block_b, block_c);
#else
                  sad += bit_cost(&cand, i, j, block_a, block_b, block_c);
#endif
              
                  if (sad >= gsad) 
                    continue;
              
                }
              
                if (sad < gsad) 
                {
                  gsad = sad;
                  bk0 = ZERO;
                  bk1 = m;
                  bl0 = cand.BP_DC;
                }
              }
            }
          
            mf->get_pblk(i,j)->BP_scale[0] = bk0;
            mf->get_pblk(i,j)->BP_scale[1] = bk1;
            mf->get_pblk(i,j)->BP_DC = bl0;
          
            mf->get_pblk(i,j)->list = cur_list;
          
            mf->get_pblk(i,j)->gsad = gsad;
          }
          else 
          { 
            bk = ZERO;
            bl = ZERO;
          
            if(param.motion_sharing_me_init_offset)
            {
              gsad = total_cost_mvshare(i, j, INT_INF, &cand, mf, &blk, cur->frame_type==ft_B_frame);
            }
            else
            {
#ifndef MVCOMP_ZERORES
              gsad = bit_cost(&cand, block_a, block_b, block_c);
#else
              gsad = bit_cost(&cand, i, j, block_a, block_b, block_c);
#endif
              gsad += blk.distortion(i, j, INT_INF, &cand);
            }
          
            sum1 = blk.sum_luma;
          
            sum2 = blk.get_mc_sum_luma(i, j, &cand, cur_list);
          
            for (k=-power2(SAMPLE_ADD_PREC); k<=power2(SAMPLE_ADD_PREC); k++) 
            {
              ncand = ZERO;
            
              add_to_list(ZERO, lcand, &ncand);
            
              if (block_b != NULL)
                add_to_list(block_b->DC[cur_list], lcand, &ncand);
            
              if (block_a != NULL)
                add_to_list(block_a->DC[cur_list], lcand, &ncand);
            
              val = (power2(SAMPLE_ADD_PREC) * sum1 - sum2 * (k + power2(SAMPLE_ADD_PREC))) >> (SAMPLE_ADD_PREC + 6);
              val = clip(val, -MAX_SAMPLE_VAL, MAX_SAMPLE_VAL - ONE);
              add_to_list(val, lcand, &ncand);
              add_to_list(val + ONE, lcand, &ncand);
            
              for (l = ZERO; l < ncand; l++) 
              {
                cand.scale[cur_list] = k; 
                cand.DC[cur_list] = lcand[l];         
              
                if (cand.inbounds_wp() == FALSE)
                  continue;

                if(param.motion_sharing_me_init_offset)
                {
                  sad = total_cost_mvshare(i, j, gsad, &cand, mf, &blk, cur->frame_type==ft_B_frame);
                }
                else
                {            
                  sad = blk.distortion(i, j, gsad, &cand);

                  if (sad >= minDist)
                    continue;

#ifndef MVCOMP_ZERORES
                  sad += bit_cost(&cand, block_a, block_b, block_c);
#else
                  sad += bit_cost(&cand, i, j, block_a, block_b, block_c);
#endif
              
                  if (sad >= gsad) 
                    continue;
              
                }
              
                if (sad < gsad) 
                {
                  gsad = sad;
                  bk = k;
                  bl = lcand[l];              
                }
              }
            }
            mf->get_pblk(i,j)->scale[cur_list] = bk; 
            mf->get_pblk(i,j)->DC[cur_list] = bl;
            mf->get_pblk(i,j)->list = cur_list;
          
            mf->get_pblk(i,j)->gsad = gsad;
          }

        } // j
      }  // i
    }  //(param.enable_intcomp)
#endif

  for (i=0; i<mf->height; i++) 
  {
    for (j=0; j<mf->width; j++) 
    {
      PredBlk *a = mf->get_pblk(i, j);
      PredBlk *b = prevMf->get_pblk(i, j);
      int scale;

      if (a->list != LIST_1)
      {
        scale = ref_mgr->inv_dist[LIST_0][a->ref_index[LIST_0]];
        b->mv[LIST_0][0] = a->mv[LIST_0][0] * scale;
        b->mv[LIST_0][1] = a->mv[LIST_0][1] * scale;
      }
      if (a->list != LIST_0)
      {
        scale = ref_mgr->inv_dist[LIST_1][a->ref_index[LIST_1]];
        b->mv[LIST_1][0] = a->mv[LIST_1][0] * scale;
        b->mv[LIST_1][1] = a->mv[LIST_1][1] * scale;
      }
    }
  }
}

MotionField *MotionEstUZSearchBiPred::prevMf;
void MotionEstUZSearchBiPred::initPrevMf(int height, int width)
{
  prevMf = new MotionField(height, width);
}
void MotionEstUZSearchBiPred::freePrevMf()
{
  delete prevMf;
}

void MotionEstUZSearchBiPred::setMv(PredBlk *pblk, int i, int j, int u, int v, int n)
{
  int k, l;

  for (k = 0; k < n; k++)
  {
    for (l = 0; l < n; l++)
    {
      *mf->get_pblk(i + u + k, j + v + l) = *pblk;
    }
  }
}

int MotionEstUZSearchBiPred::bitcostN(PredBlk *pblk, int i, int j, int u, int v, int n)
{
  int k, l;
  PredBlk *a, *b, *c;
  int sum = 0;

  for (k = 0; k < n; k++)
  {
    for (l = 0; l < n; l++)
    {
      if (k != 0)
        b = pblk;
      else
        b = mf->get_pblk(i+u-1,j+v+l);
      if (l != 0)
        a = pblk;
      else
        a = mf->get_pblk(i+u+k,j+v-1);
      if (k != 0 && l != 0)
        c = pblk;
      else
        c = mf->get_pblk(i+u+k-1,j+v+l-1);

#ifndef MVCOMP_ZERORES
      sum += bit_cost(pblk, a, b, c);
#else
      sum += bit_cost(pblk, i+u+k, j+v+l, a, b, c);
#endif
    }
  }

  return sum;
}

int MotionEstUZSearchBiPred::sadN(MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], PredBlk *pblk, int i, int j, int u, int v, int n)
{
  int k, l;
  int sum = 0;

  for (k = 0; k < n; k++)
  {
    for (l = 0; l < n; l++)
    {
      sum += blk[u+k][v+l].sad(i+u+k,j+v+l, pblk);
    }
  }

  return sum;
}

int MotionEstUZSearchBiPred::estimateR(int i, int j, MCRefBlk8 blk[MAXMEBLK][MAXMEBLK], int u, int v, int n, int list, Frame *cur)
{
  int k, l;
  int n2 = n >> 1;
  int cost_split;
  int bsad;

  if (n2 != 0)
  {
    cost_split = 0;

    for (k = 0; k < 2; k++)
    {
      for (l = 0; l < 2; l++)
      {
        cost_split += estimateR(i, j, blk, u+n2*k, v+n2*l, n2, list, cur);
      }
    }
  }
  else
  {
    cost_split = INT_INF;
  }

  if (i + u >= mf->height || j + v >= mf->width)
    return 0;

  if (i + u + n > mf->height || j + v + n > mf->width)
    return cost_split;

  bsad = cost_split;
  PredBlk gcand[2];
  PredBlk *block_a, *block_b, *block_c, *block_d;

  block_a = mf->get_pblk(i+u, j+v-1);
  block_b = mf->get_pblk(i+u-1, j+v);
  block_c = mf->get_pblk(i+u-1, j+v-1);
  block_d = mf->get_pblk(i+u-1, j+v+n);

  if (list == LIST_0 || list == BI_PRED)
    gcand[LIST_0] = estimate_one_list(LIST_0, block_a, block_b, block_c, block_d, i, j, blk, u, v, n);
  
  if (list == LIST_1 || list == BI_PRED)
    gcand[LIST_1] = estimate_one_list(LIST_1, block_a, block_b, block_c, block_d, i, j, blk, u, v, n);

  int bi_cases = 1;
  PredBlk bipred[3];

  if (list == BI_PRED)  
  {
    for (k = 0; k < bi_cases; k++) 
    {
      bipred[k].list = BI_PRED;
      
      if ((k & 2) == 0) 
      {
        bipred[k].mv[LIST_0][0]     = gcand[LIST_0].mv[LIST_0][0];
        bipred[k].mv[LIST_0][1]     = gcand[LIST_0].mv[LIST_0][1];
        bipred[k].ref_index[LIST_0] = gcand[LIST_0].ref_index[LIST_0];
      }
      
      if ((k & 1) == 0) 
      {
        bipred[k].mv[LIST_1][0]     = gcand[LIST_1].mv[LIST_1][0];
        bipred[k].mv[LIST_1][1]     = gcand[LIST_1].mv[LIST_1][1];
        bipred[k].ref_index[LIST_1] = gcand[LIST_1].ref_index[LIST_1];
      }

      int sad = bitcostN(&bipred[k], i, j, u, v, n);
      sad += sadN(blk, &bipred[k], i, j, u, v, n);

      bipred[k].gsad = sad;
    }
  }

  if (list == LIST_0)
  {
    if (gcand[LIST_0].gsad < bsad)
    {
      bsad = gcand[LIST_0].gsad;
      setMv(&gcand[LIST_0], i, j, u, v, n);
    }
  }
  else if (list == LIST_1)
  {
    if (gcand[LIST_1].gsad < bsad)
    {
      bsad = gcand[LIST_1].gsad;
      setMv(&gcand[LIST_1], i, j, u, v, n);
    }
  }
  else
  {
    if (cur->me_mode == BI_PRED_ONLY)     
    {
      int z = 0;
      for (k=1; k<bi_cases; k++)
        if (bipred[k].gsad < bipred[z].gsad)
          z = k;
      if (bipred[z].gsad < bsad)
      {
        bsad = bipred[z].gsad;
        setMv(&bipred[z], i, j, u, v, n);
      }
    }
    else
    {
      if (n == 1)
      {
        row[j+u] = gcand[LIST_0];
        row[j+u].ref_index[LIST_1]=gcand[LIST_1].ref_index[LIST_1];
        row[j+u].mv[LIST_1][vertical]=gcand[LIST_1].mv[LIST_1][vertical];
        row[j+u].mv[LIST_1][horizontal]=gcand[LIST_1].mv[LIST_1][horizontal];
      }
      
      int bsad = cost_split;
      if (gcand[LIST_0].gsad < bsad)
      {
        bsad = gcand[LIST_0].gsad;
        setMv(&gcand[LIST_0], i, j, u, v, n);
      }
      if (gcand[LIST_1].gsad < bsad)
      {
        bsad = gcand[LIST_1].gsad;
        setMv(&gcand[LIST_1], i, j, u, v, n);
      }

      int z = 0;

      for (k = 1; k < bi_cases; k++)
        if (bipred[k].gsad < bipred[z].gsad)
          z = k;

      if (bipred[z].gsad < bsad)
      {
        bsad = bipred[z].gsad;
        setMv(&bipred[z], i, j, u, v, n);
      }
    }
  }

  return bsad;
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
