/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file MotionField.cpp
*
* \brief Implementation of MotionField class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "MotionField.H"
#include "Global.H"
#include "Param.H"
#include "LumaFrames.H"
#include "ChromaFrames.H"
#include "MCRefBlk8.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in sample units
* \param[in] w width of frame in sample units
*/
//-----------------------------------------------------------------------------
MotionField::MotionField(int h, int w)
{
  width = w >> LOG_MOTION_BLK_SIZE;
  height = h >> LOG_MOTION_BLK_SIZE;

  blk_array = new PredBlk[height * width];
  blk_array_I = new PredBlkI[height * width];
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*/
//-----------------------------------------------------------------------------
MotionField::~MotionField()
{
  delete[] blk_array;
  delete[] blk_array_I;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

#ifdef MVCOMP_ZERORES
//-----------------------------------------------------------------------------
/*!
* \brief Copy content
*
* \param[in] mf_src
*
* \return void
*/
//-----------------------------------------------------------------------------
void MotionField::copy_field_from(MotionField *mf_src)
{
  memcpy(blk_array, mf_src->blk_array, width * height * sizeof(blk_array[0]));
  memcpy(blk_array_I, mf_src->blk_array_I, width * height * sizeof(blk_array_I[0]));

  int i;
  for (i = 0; i < width * height; i++)
  {
    if (blk_array[i].intra != NULL)
      blk_array[i].intra = &blk_array_I[i];
  }
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to luma data in interpolated reference frame
*
* \param[in] ref_mgr Pointer to reference frame manager
* \param[in] blk Pointer to motion data of current block
* \param[in] list Prediction list (LIST_0 or LIST_1)
* \param[in] i Vertical position of current 4x4 block (in 4-luma-sample units)
* \param[in] j Horizontal position of current 4x4 block (in 4-luma-sample units)
*
* \return Pointer to luma data
*/
//-----------------------------------------------------------------------------
unsigned char *MotionField::get_luma_ptr(RefFrameManager *ref_mgr, PredBlk *blk, int list, int i, int j, unsigned char *data)
{
  return ref_mgr->get_luma_ptr(list, blk->ref_index[list], blk->mv[list], 8 * i, 8 * j, 8, 8, data);
}

//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to chroma data in interpolated reference frame
*
* \param[in] ref_mgr Pointer to reference frame manager
* \param[in] blk Pointer to motion data of current block
* \param[in] list Prediction list (LIST_0 or LIST_1)
* \param[in] i Vertical position of current 4x4 block (in 4-chroma-sample units)
* \param[in] j Horizontal position of current 4x4 block (in 4-chroma-sample units)
* \param[in] c Chroma component index (0 or 1)
*
* \return Pointer to chroma data
*/
//-----------------------------------------------------------------------------
unsigned char *MotionField::get_chroma_ptr(RefFrameManager *ref_mgr, PredBlk *blk, int list, int i, int j, unsigned char *data)
{
  return ref_mgr->get_chroma_ptr(list, blk->ref_index[list], blk->mv[list], 8 * i, 8 * j, 8, 8, data);
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate motion-predicted frame
*
* Generates a motion-predicted frame based on motion field. For blocks that are marked intra, the prediction is set to zero.
*
* \param[out] cur Pointer to motion-predicted frame
* \param[in] ref_mgr Pointer to reference frame manager
*/
//-----------------------------------------------------------------------------
void MotionField::predict(WFrame *cur, RefFrameManager *ref_mgr)
{
  int i, j;

  for (i = 0; i < height; i++) 
  {
    for (j = 0; j < width; j++) 
    {
      predict_blk(cur, ref_mgr, i, j);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate predicted inter 8x8 block
*
* \param[out] cur Pointer to motion-predicted frame
* \param[in] ref_mgr Pointer to reference frame manager
* \param[in] i Vertical position of current 8x8 block
* \param[in] j Horizontal position of current 8x8 block
*/
//-----------------------------------------------------------------------------
void MotionField::predict_blk(WFrame *cur, RefFrameManager *ref_mgr, int i, int j)
{
  int k,l;
  unsigned char *p_l0=NULL, *p_l1=NULL;      // initialize to avoid compiler warning.
  PredBlk *pblk;
  int x;
  int list;  
  int scale_l0, scale_l1, offset;
  int stride = 0;
  short luma[64], chroma[2][16];
  unsigned char luma_data[2][MOTION_BLK_SIZE * MOTION_BLK_SIZE];
  unsigned char chroma_data[2][MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  
  pblk = get_pblk_unsafe(i, j);
  list = pblk->list;

  if (pblk->intra != NULL) 
  {
    for (k=0; k<64; k++)
      luma[k] = ZERO;

    for (k=0; k<16; k++) 
    {
      chroma[0][k] = ZERO;
      chroma[1][k] = ZERO;
    }
  }
  else if (list == BI_PRED) 
  {
    // do luma
    if (ref_mgr->decoder_mode)
    {
      p_l0 = get_luma_ptr(ref_mgr, pblk, LIST_0, i, j, luma_data[0]);
      p_l1 = get_luma_ptr(ref_mgr, pblk, LIST_1, i, j, luma_data[1]);
      stride = MOTION_BLK_SIZE;
    }
    else
    {
      p_l0 = get_luma_ptr(ref_mgr, pblk, LIST_0, i, j);
      p_l1 = get_luma_ptr(ref_mgr, pblk, LIST_1, i, j);
      stride = ref_mgr->get_stride();
    }

    scale_l0 = (pblk->get_B_scale_sum() >> ONE) + pblk->get_B_scale_diff();
    scale_l1 = pblk->get_B_scale_sum() - scale_l0;
    offset = pblk->BP_DC * param.DC_mult;

    for (k = ZERO; k < MOTION_BLK_SIZE; k++) 
    {
      for (l = ZERO; l < MOTION_BLK_SIZE; l++) 
      {
        x  = p_l0[stride*k+l] * scale_l0;
        x += p_l1[stride*k+l] * scale_l1;
        x += (offset - MID_SAMPLE_VAL) << SAMPLE_ADD_PREC;
        luma[MOTION_BLK_SIZE * k + l] = x;
      }
    }

    // do chroma
    int chroma_stride = stride;
    if (param.chroma_on_fly)
    {
      p_l0 = get_chroma_ptr(ref_mgr, pblk, LIST_0, i, j, chroma_data[0]);
      p_l1 = get_chroma_ptr(ref_mgr, pblk, LIST_1, i, j, chroma_data[1]);
      chroma_stride = MOTION_BLK_SIZE;
    }
    else
    {
      p_l0 = get_chroma_ptr(ref_mgr, pblk, LIST_0, i, j);
      p_l1 = get_chroma_ptr(ref_mgr, pblk, LIST_1, i, j);
    }
    
    offset = MID_SAMPLE_VAL * (- scale_l0 - scale_l1);

    for (k = ZERO; k < TX_SIZE; k++) 
    {
      for (l = ZERO; l < TX_SIZE; l++) 
      {
        x  = p_l0[TWO * l] * scale_l0;
        x += p_l1[TWO * l] * scale_l1;          
        x += offset;
        chroma[0][TX_SIZE * k + l] = x;

        x  = p_l0[TWO * l + ONE] * scale_l0;
        x += p_l1[TWO * l + ONE] * scale_l1;          
        x += offset;
        chroma[1][TX_SIZE * k + l] = x;
      }
      p_l0 += chroma_stride;
      p_l1 += chroma_stride;
    }
  }
  else 
  {
    if (ref_mgr->decoder_mode)
    {
      p_l0 = get_luma_ptr(ref_mgr, pblk, list, i, j, luma_data[0]);
      stride = MOTION_BLK_SIZE;
    }
    else
    {
      p_l0 = get_luma_ptr(ref_mgr, pblk, list, i, j);
      stride = ref_mgr->get_stride();
    }

    scale_l0 = pblk->scale[list] + power2(SAMPLE_ADD_PREC);
    offset = pblk->DC[list] * param.DC_mult;

    for (k = ZERO; k < MOTION_BLK_SIZE; k++) 
    {
      for (l = ZERO; l < MOTION_BLK_SIZE; l++) 
      {
        x = p_l0[stride*k+l] * scale_l0;
        x += (offset - MID_SAMPLE_VAL) << SAMPLE_ADD_PREC;
        luma[MOTION_BLK_SIZE * k + l] = x;
      }
    }

    int chroma_stride = stride;
    if (param.chroma_on_fly)
    {
      p_l0 = get_chroma_ptr(ref_mgr, pblk, list, i, j, chroma_data[0]);
      chroma_stride = MOTION_BLK_SIZE;
    }
    else
    {
      p_l0 = get_chroma_ptr(ref_mgr, pblk, list, i, j);
    }

    offset = MID_SAMPLE_VAL * (- scale_l0);

    for (k = ZERO; k < TX_SIZE; k++) 
    {
      for (l = ZERO; l < TX_SIZE; l++) 
      {

        x = p_l0[TWO * l] * scale_l0;
        x += offset;
        chroma[0][TX_SIZE * k + l] = x;

        x = p_l0[TWO * l + ONE] * scale_l0;
        x += offset;
        chroma[1][TX_SIZE * k + l] = x;

      }
      p_l0 += chroma_stride;
    }

  }

  cur->put_mc_block(i, j, luma, chroma);
}

//-----------------------------------------------------------------------------
/*!
* \brief Get motion data for a given 4x4 block
*
* This method correct for a mvshare of 4. A mvshare of 2 is ignored.
* This method checks whether block position is within frame boundaries. If not, a NULL pointer is returned.
*
* \param[in] i Vertical position of 4x4 block (in 4x4 block units)
* \param[in] j Horizontal position of 4x4 block (in 4x4 block units)
*
* \return Pointer to motion data of 4x4 block
*/
//-----------------------------------------------------------------------------
PredBlk *MotionField::get_pblk_4x4(int i2, int j2)
{
  int i = i2 >> 1;
  int j = j2 >> 1;
  PredBlk *tmp_blk = get_pblk(i, j);
  
  if (tmp_blk != NULL)
  {
    if ((i2 & 1) == 0 && tmp_blk->mvshare_size[0] == 4 && i > 0) //Condition i > 0 is actually redundant if mvshare variable is set properly
      tmp_blk = get_pblk(i - 1, j);
    else if ((j2 & 1) == 0 && tmp_blk->mvshare_size[1] == 4 && j > 0) //Condition j > 0 is actually redundant if mvshare variable is set properly
      tmp_blk = get_pblk(i, j - 1);
  }

  return tmp_blk;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get motion data for a given 8x8 block
*
* This methods checks whether block position is within frame boundaries. If not, a NULL pointer is returned.
*
* \param[in] i Vertical position of 8x8 block (in 8x8 block units)
* \param[in] j Horizontal position of 8x8 block (in 8x8 block units)
*
* \return Pointer to motion data of 8x8 block
*/
//-----------------------------------------------------------------------------
PredBlk *MotionField::get_pblk(int i, int j) 
{
  if (i >= ZERO && i < height && j >= ZERO && j < width)
    return &blk_array[width*i+j];
  else
    return NULL;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get motion data for a given 8x8 block
*
* \note This methods DOES NOT check whether block position is within frame
* boundaries. If not in bounds, this may result in a crash :-)
* This method is much faster than get_pblk and should only be used when the
* block position is known to be within frame boundaries.
*
* \param[in] i Vertical position of 8x8 block (in 8x8 block units)
* \param[in] j Horizontal position of 8x8 block (in 8x8 block units)
* \return Pointer to motion data of 8x8 block
*
*/
//-----------------------------------------------------------------------------
PredBlk *MotionField::get_pblk_unsafe(int i, int j) 
{
  return &blk_array[width*i+j];
}

//-----------------------------------------------------------------------------
/*!
* \brief Set all blocks to intra
*
* Sets all the motion data in a motion field to intra mode.
*/
//-----------------------------------------------------------------------------
void MotionField::set_all_intra()
{
  int i;

  for (i = 0; i < height * width; i++)
  {
    blk_array[i].clear();
    blk_array[i].intra = &blk_array_I[i];
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Reset motion field
*
*/
//-----------------------------------------------------------------------------
void MotionField::reset_motion_field()
{
  int i;

  for (i = 0; i < height * width; i++)
  {
    blk_array[i].clear();
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Change resolution of motion field
*
* \param[in] ref_mgr Pointer to reference frame manager
* \param[in] frame_type indication for type of frame (ft_P_frame or ft_B_frame)
*/
//-----------------------------------------------------------------------------
void MotionField::change_resolution_motion_field(RefFrameManager *ref_mgr, int /*frame_type*/)
{
  int i,j,k;
  int temp_inter_pel, temp_mv;

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      PredBlk &blk = blk_array[width * i + j];

      for (k = 0; k < 2; k++) 
      {
        int refIndex = blk.ref_index[k];
        int prev_inter_pel = 1 << ref_mgr->inter_type_set[ref_mgr->prev_rpps_id][k][refIndex];

        temp_inter_pel = ref_mgr->ref_process->inter_pel[k][refIndex];

        if (temp_inter_pel != prev_inter_pel)
        {
          temp_mv = blk.mv[k][0] * temp_inter_pel / prev_inter_pel;
          blk.mv[k][0] = clip(temp_mv, -temp_inter_pel * param.me_range, temp_inter_pel * param.me_range - 1);

          temp_mv = blk.mv[k][1] * temp_inter_pel / prev_inter_pel;   
          blk.mv[k][1] = clip(temp_mv, -temp_inter_pel * param.me_range, temp_inter_pel * param.me_range - 1);          
        }
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Set current block intra mode
*
* \param[in] i Vertical position of current 8x8 block
* \param[in] j Horizontal position of current 8x8 block
*/
//-----------------------------------------------------------------------------
void MotionField::set_block_intra(int i, int j)
{  
  blk_array[width*i+j].clear();
  blk_array[width*i+j].intra = &blk_array_I[width*i+j];  
}

//-----------------------------------------------------------------------------
/*!
* \brief Get motion sharing type for a target block
*
* \param[in] i Vertical position of current 8x8 block (in 8-luma-sample units)
* \param[in] j Horizontal position of current 8x8 block (in 8-luma-sample units)
*
* \return motion sharing type
*/
//-----------------------------------------------------------------------------                        
int MotionField::mvshare_type(int i, int j)
{
  return mvshare_type_simple(get_pblk_unsafe(i, j), get_pblk(i - 1, j), get_pblk(i, j - 1));
}

//-----------------------------------------------------------------------------
/*!
* \brief Get motion sharing type for a target block for ME
*
* \param[in] blk Motion data
* \param[in] top Motion data from block above
* \param[in] left Motion data from block to the left*
*
* \return motion sharing type
*/
//-----------------------------------------------------------------------------                        
int MotionField::mvshare_type_simple(PredBlk *cur, PredBlk *top, PredBlk *left)
{
  if (cur->intra)
    return 0;

  int leftIsUsable = (left != NULL) && !left->intra && !cur->equal(left);
  int topIsUsable  = (top  != NULL) && !top ->intra && !cur->equal(top);

  // return 1 if top only, 2 if left only, 4 if both
  return topIsUsable + 2 * leftIsUsable + (topIsUsable & leftIsUsable);
}

int MotionField::get_mvshare_size_ctx(int mvshare_type, PredBlk *top, PredBlk *left)
{
  int mvshare_type_top = 0;
  if (top != NULL)
  {
    if (top->intra != NULL) 
    {
      mvshare_type_top = 1;
    }
    else
    {
      if (mvshare_type != 2)
        mvshare_type_top += top->mvshare_size[0];
      if (mvshare_type != 1)
        mvshare_type_top += top->mvshare_size[1];
    }
  }
  
  int mvshare_type_left = 0;
  if (left != NULL)
  {
    if (left->intra != NULL) 
    {
      mvshare_type_left = 1;
    }
    else
    {
      if (mvshare_type != 2)
        mvshare_type_left += left->mvshare_size[0];
      if (mvshare_type != 1)
        mvshare_type_left += left->mvshare_size[1];
    }
  }
  
  return (mvshare_type_top != 0) + (mvshare_type_left != 0); 
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate motion-predicted frame with motion sharing
*
* \param[out] cur Pointer to motion-predicted frame
* \param[in] ref_mgr Pointer to reference frame manager
* \param[in] i Vertical position of 8x8 block 
* \param[in] j Horizontal position of 8x8 block 
*/
//-----------------------------------------------------------------------------
void MotionField::predict_blk_mvshare(WFrame *pred, RefFrameManager *ref_mgr, int i, int j)
{
  short luma_mvshare[64];
  short chroma_mvshare[2][16];
  PredBlk *pblk = get_pblk_unsafe(i, j);
  int ver_mvshare = pblk->mvshare_size[0] != 0;
  int hor_mvshare = pblk->mvshare_size[1] != 0;

  if (ver_mvshare || hor_mvshare)
  {
    predict_blk_part_mvshare(luma_mvshare, chroma_mvshare, ref_mgr, i-ver_mvshare, j-hor_mvshare, i, j);
    pred->put_mc_block_mvshare(i, j, luma_mvshare, chroma_mvshare, pblk->mvshare_size[0], pblk->mvshare_size[1]);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Get prediction signalin interpolated reference frame
*
* \param[out] luma Pointer to luma predicted block
* \param[out] chroma Pointer to chroma predicted block
* \param[in] ref_mgr Pointer to reference frame manager
* \param[in] ii_t Vertical 8x8 block position including mv, scale and offset
* \param[in] jj_t Horizontal 8x8 block position including mv, scale and offset
* \param[in] ii_b Vertical 8x8 block position to be predicted
* \param[in] jj_b Horizontal 8x8 block position to be predicted
*/
//-----------------------------------------------------------------------------                        
void MotionField::predict_blk_part_mvshare(short luma[64], short chroma[2][16], RefFrameManager *ref_mgr, int ii_t, int jj_t, int ii_b, int jj_b)
{
  int k,l;
  unsigned char *p_l0, *p_l1;
  PredBlk *pblk;
  int x;
  int list;  
  int scale_l0, scale_l1, offset;
  int stride = 0;
  unsigned char luma_data[2][MOTION_BLK_SIZE * MOTION_BLK_SIZE];
  unsigned char chroma_data[2][MOTION_BLK_SIZE * MOTION_BLK_SIZE / 2];
  
  pblk = get_pblk_unsafe(ii_t, jj_t);
  list = pblk->list;

  if (list == BI_PRED) 
  {
    if (ref_mgr->decoder_mode)
    {
      p_l0 = get_luma_ptr(ref_mgr, pblk, LIST_0, ii_b, jj_b, luma_data[0]);
      p_l1 = get_luma_ptr(ref_mgr, pblk, LIST_1, ii_b, jj_b, luma_data[1]);
      stride = MOTION_BLK_SIZE;
    }
    else
    {
      p_l0 = get_luma_ptr(ref_mgr, pblk, LIST_0, ii_b, jj_b);
      p_l1 = get_luma_ptr(ref_mgr, pblk, LIST_1, ii_b, jj_b);
      stride = ref_mgr->get_stride();
    }

    scale_l0 = (pblk->get_B_scale_sum() >> ONE) + pblk->get_B_scale_diff();
    scale_l1 = pblk->get_B_scale_sum() - scale_l0;
    offset = pblk->BP_DC * param.DC_mult;

    for (k = ZERO; k < MOTION_BLK_SIZE; k++) 
    {
      for (l = ZERO; l < MOTION_BLK_SIZE; l++) 
      {
        x  = p_l0[stride*k+l] * scale_l0;
        x += p_l1[stride*k+l] * scale_l1;
        x += (offset - MID_SAMPLE_VAL) << SAMPLE_ADD_PREC;
        luma[MOTION_BLK_SIZE * k + l] = x;
      }
    }     

    int chroma_stride = stride;
    if (param.chroma_on_fly)
    {
      p_l0 = get_chroma_ptr(ref_mgr, pblk, LIST_0, ii_b, jj_b, chroma_data[0]);
      p_l1 = get_chroma_ptr(ref_mgr, pblk, LIST_1, ii_b, jj_b, chroma_data[1]);
      chroma_stride = MOTION_BLK_SIZE;
    }
    else
    {
      p_l0 = get_chroma_ptr(ref_mgr, pblk, LIST_0, ii_b, jj_b);
      p_l1 = get_chroma_ptr(ref_mgr, pblk, LIST_1, ii_b, jj_b);
    }
    
    offset = MID_SAMPLE_VAL * (- scale_l0 - scale_l1);


    for (k = ZERO; k < TX_SIZE; k++) 
    {
      for (l = ZERO; l < TX_SIZE; l++) 
      {
        x  = p_l0[TWO * l] * scale_l0;
        x += p_l1[TWO * l] * scale_l1;          
        x += offset;
        chroma[0][TX_SIZE * k + l] = x;

        x  = p_l0[TWO * l + ONE] * scale_l0;
        x += p_l1[TWO * l + ONE] * scale_l1;          
        x += offset;
        chroma[1][TX_SIZE * k + l] = x;
      }
      p_l0 += chroma_stride;
      p_l1 += chroma_stride;
    }
  }
  else 
  {
    if (ref_mgr->decoder_mode)
    {
      p_l0 = get_luma_ptr(ref_mgr, pblk, list, ii_b, jj_b, luma_data[0]);
      stride = MOTION_BLK_SIZE;
    }
    else
    {
      p_l0 = get_luma_ptr(ref_mgr, pblk, list, ii_b, jj_b);
      stride = ref_mgr->get_stride();
    }

    scale_l0 = pblk->scale[list] + power2(SAMPLE_ADD_PREC);
    offset = pblk->DC[list] * param.DC_mult;

    for (k = ZERO; k < MOTION_BLK_SIZE; k++) 
    {
      for (l = ZERO; l < MOTION_BLK_SIZE; l++) 
      {
        x = p_l0[stride*k+l] * scale_l0;
        x += (offset - MID_SAMPLE_VAL) << SAMPLE_ADD_PREC;
        luma[MOTION_BLK_SIZE * k + l] = x;
      }
    }

    int chroma_stride = stride;
    if (param.chroma_on_fly)
    {
      p_l0 = get_chroma_ptr(ref_mgr, pblk, list, ii_b, jj_b, chroma_data[0]);
      chroma_stride = MOTION_BLK_SIZE;
    }
    else
    {
      p_l0 = get_chroma_ptr(ref_mgr, pblk, list, ii_b, jj_b);
    }

    offset = MID_SAMPLE_VAL * (- scale_l0);

    for (k = ZERO; k < TX_SIZE; k++) 
    {
      for (l = ZERO; l < TX_SIZE; l++) 
      {
        x = p_l0[TWO * l] * scale_l0;
        x += offset;
        chroma[0][TX_SIZE * k + l] = x;

        x = p_l0[TWO * l + ONE] * scale_l0;
        x += offset;
        chroma[1][TX_SIZE * k + l] = x;
      }
      p_l0 += chroma_stride;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Get block size of transform for inter block
*
* Transform size is 4x4 for blocks that use motion sharing, and 8x8 othewise
*
* \param[in] i Vertical 8x8 block position 
* \param[in] j Horizontal 8x8 block position 
* \return indication of block size (1: 8x8, 0: 4x4, -1: invalid block)
*/
//-----------------------------------------------------------------------------                        
int MotionField::get_transform_blocksize(int i, int j)
{
  PredBlk *cur = get_pblk(i, j);

  if (cur == NULL || cur->intra != NULL)
    return -1;

  return (param.trans8x8_enable && cur->mvshare_size[0] == 0 && cur->mvshare_size[1] == 0);
}

//-----------------------------------------------------------------------------
/*!
* \brief Partial reset of motion field
*
*/
//-----------------------------------------------------------------------------                        
void MotionField::fix_lists()
{
  int i, j;

  for (i=0; i<height; i++)
  {
    for (j=0; j<width; j++)
    {
      PredBlk *cur_blk = get_pblk_unsafe(i,j);
      if (cur_blk->list == -1)
        cur_blk->list = LIST_0;

      if (cur_blk->list != BI_PRED)
      {
        cur_blk->ref_index[cur_blk->list ^ 1] = 0;
      }

      cur_blk->mvshare_size[0] = 0;
      cur_blk->mvshare_size[1] = 0;
#ifdef MVCOMP_ZERORES
      cur_blk->mv_zero_res = 0;
#endif
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Print motion field info
*
* \param[in] f File to which to print
* \param[in] frame_no Frame number
*/
//-----------------------------------------------------------------------------                        
void MotionField::print_info(FILE *f, int frame_no)
{
  int i, j;
  PredBlk* cur_blk;

  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      cur_blk = get_pblk(i,j);
      fprintf(f,"%d %d %d %d", frame_no, i, j, (cur_blk->intra==NULL));
      if (cur_blk->intra==NULL)
      {
        fprintf(f," %d", cur_blk->list);
        fprintf(f," L0(%d,%d,%d) L1(%d,%d,%d)", cur_blk->ref_index[0], cur_blk->mv[0][0], cur_blk->mv[0][1], cur_blk->ref_index[1], cur_blk->mv[1][0], cur_blk->mv[1][1]);
        fprintf(f," MVSHARE(%d,%d)", cur_blk->mvshare_size[0], cur_blk->mvshare_size[1]);
      }
      fprintf(f,"\n");
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Write motion field info to stat files
 *
 * \param[in] cur Pointer to current frame
 */
//-----------------------------------------------------------------------------                        
void MotionField::write_stat_files(Frame *cur)
{
  // Warning: this will only work on little endian architectures (i386)
  int i, j;
  PredBlk *curblk;
  
  if (param.mode_stats_file)
  {
    // Go to appropriate position in mode_stats_file. 7 bytes for header.    
    fseek(param.mode_stats_file, (7 + (width * height)) * cur->frame_no, SEEK_SET);
    fwrite(&cur->frame_type,    1, 1, param.mode_stats_file);
    fwrite(&cur->frame_no,      4, 1, param.mode_stats_file);
    fwrite(&cur->qp,            1, 1, param.mode_stats_file);
    fwrite(&cur->reference_idc, 1, 1, param.mode_stats_file);
  }
  if (param.mv_stats_file)
  {
    // Go to appropriate position in mv_stats_file. 2 Bytes for reference index, 8 * 2 for motion info.    
    fseek(param.mv_stats_file, (18 * (width * height)) * cur->frame_no, SEEK_SET);
  }
  
  for (i = 0; i < height; i++)
  {
    for (j = 0; j < width; j++)
    {
      curblk = get_pblk(i,j);
      if (curblk->intra != NULL)
      {
        char modes = -1;
        if (param.mode_stats_file)
        {
          fwrite(&modes, 1, 1, param.mode_stats_file);
        }
        
        if (param.mv_stats_file)
        {
          modes = 0;
          for (int m= 0; m<18; m++)
            fwrite(&modes, 1, 1, param.mv_stats_file);
        }
      }
      else
      {
        if (param.mode_stats_file)
        {
          fwrite(&curblk->list, 1, 1,param.mode_stats_file);
        }
        
        if (param.mv_stats_file)
        {
          fwrite(&curblk->ref_index[0], 1, 1, param.mv_stats_file);
          fwrite(&curblk->ref_index[1], 1, 1, param.mv_stats_file);
          fwrite(&curblk->mv[LIST_0][0], 2, 1, param.mv_stats_file);
          fwrite(&curblk->mv[LIST_0][1], 2, 1, param.mv_stats_file);
          fwrite(&curblk->mv[LIST_1][0], 2, 1, param.mv_stats_file);
          fwrite(&curblk->mv[LIST_1][1], 2, 1, param.mv_stats_file);
          if (curblk->list == BI_PRED) 
          {
            fwrite(&curblk->BP_scale[0], 2, 1, param.mv_stats_file);
            fwrite(&curblk->BP_scale[1], 2, 1, param.mv_stats_file);
            fwrite(&curblk->BP_DC, 2, 1, param.mv_stats_file);
            fwrite(&curblk->BP_DC, 2, 1, param.mv_stats_file);
          }
          else 
          {
            fwrite(&curblk->scale[0], 2, 1, param.mv_stats_file);
            fwrite(&curblk->scale[1], 2, 1, param.mv_stats_file);
            fwrite(&curblk->DC[0], 2, 1, param.mv_stats_file);
            fwrite(&curblk->DC[1], 2, 1, param.mv_stats_file);
          }
        }
      }
    }
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
