/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file Plane.cpp
*
* \brief Implementation of Plane class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Plane.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
 * \brief Default constructor
 */
//-----------------------------------------------------------------------------
template <typename T>
Plane<T>::Plane()
{
  m_width = 0;
  m_height = 0;
  m_pitch = 0;
  m_border_size =0;
  m_data = NULL;
  m_origin = NULL;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Constructor
 *
 * \param[in] width Width of plane
 * \param[in] height Height of plane
 * \param[in] border_size Number of samples added at each border
 */
//-----------------------------------------------------------------------------
template <typename T>
Plane<T>::Plane(int width, int height, int border_size)
{
  m_data = NULL;
  set_size(width, height, border_size);
}

//-----------------------------------------------------------------------------
/*!
 * \brief Destructor
 */
//-----------------------------------------------------------------------------
template <typename T>
Plane<T>::~Plane()
{
  if (m_data != NULL)
    delete[] m_data;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Set size
 *
 * \param[in] width Width of plane
 * \param[in] height Height of plane
 * \param[in] border_size Number of samples added at each border
 */
//-----------------------------------------------------------------------------
template <typename T>
void Plane<T>::set_size(int width, int height, int border_size)
{
  if (m_data != NULL)
    delete[] m_data;

  m_width = width;
  m_height = height;
  m_pitch = width + 2 * border_size;
  m_border_size = border_size;

  m_data = new T[m_pitch * (height + 2 * border_size)];
  m_origin = m_data + border_size * m_pitch + border_size;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Fill a rectangular window with a given value
 *
 * \param[in] row Top row of window
 * \param[in] col Leftmost column of window
 * \param[in] val Value to be assigned
 * \param[in] numRows Number of rows to fill
 * \param[in] numCols Number of columns to fill
 */
//-----------------------------------------------------------------------------
template <typename T>
void Plane<T>::set(int row, int col, T val, int numRows, int numCols)
{
  int i, j;

  for (i = 0; i < numRows; i++)
  {
    for (j = 0; j < numCols; j++)
    {
      (*this)[row+i][col+j] = val;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Fill border samples by mirroring
 */
//-----------------------------------------------------------------------------
template <typename T>
void Plane<T>::mirror()
{
  int i, j;

  for (i = 0; i < m_height; i++)
  {
    for (j = 0; j < m_border_size; j++)
    {
      (*this)[i][-(1 + j)] = (*this)[i][1 + j];
      (*this)[i][m_width + j] = (*this)[i][m_width - 2 - j];
    }
  }

  for (i = 0; i < m_border_size; i++)
  {
    for (j = 0; j < m_pitch; j++)
    {
      (*this)[-(1 + i)][j - m_border_size] = (*this)[1 + i][j - m_border_size];
      (*this)[m_height + i][j - m_border_size] = (*this)[m_height - 2 - i][j - m_border_size];
    }
  }
}

// Explicit instantiations
template class Plane<unsigned char>;
template class Plane<short>;
template class Plane<int>;

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
