/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file PredBlk.cpp
* \brief Implementation of PredBlk and PredBlkI classes
*/
//-----------------------------------------------------------------------------

#include "PredBlk.H"
#include "Global.H"
#include "Param.H"

#ifdef MVCOMP_ZERORES
#include "MotionField.H"
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Clears all data fields.
*/
//-----------------------------------------------------------------------------
PredBlk::PredBlk() 
{
  clear();
}

//-----------------------------------------------------------------------------
/*!
* \brief Clear all data fields
*/
//-----------------------------------------------------------------------------
void PredBlkI::reset_cost()
{
  int i;

  for (i=0; i<8; i++)
  {
    cost[i] = -1;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Clear all data fields
*
*/
//-----------------------------------------------------------------------------
void PredBlk::clear() 
{
  intra = NULL;
  mv[LIST_0][vertical  ] = ZERO;
  mv[LIST_0][horizontal] = ZERO;
  mv[LIST_1][vertical  ] = ZERO;
  mv[LIST_1][horizontal] = ZERO;

  ref_index[LIST_0] = ZERO;
  ref_index[LIST_1] = ZERO;
  gsad = INT_INF;
  list = LIST_0;
  scale[LIST_0] = ZERO;
  scale[LIST_1] = ZERO;
  DC[LIST_0] = ZERO;
  DC[LIST_1] = ZERO;
  BP_DC = ZERO;
  BP_scale[0] = ZERO;
  BP_scale[1] = ZERO;

  mvshare_size[0] = 0;
  mvshare_size[1] = 0;

#ifdef MVCOMP_ZERORES
  mv_zero_res = 0;
#endif
}


//-----------------------------------------------------------------------------
/*!
* \brief Copy all data fields
*
* \param[in] blk Motion data to copy from
*/
//-----------------------------------------------------------------------------
void PredBlk::copy(PredBlk *blk) 
{
  intra = blk->intra;

  mv[LIST_0][vertical  ] = blk->mv[LIST_0][vertical  ];
  mv[LIST_0][horizontal] = blk->mv[LIST_0][horizontal];
  mv[LIST_1][vertical  ] = blk->mv[LIST_1][vertical  ];
  mv[LIST_1][horizontal] = blk->mv[LIST_1][horizontal];

  ref_index[LIST_0] = blk->ref_index[LIST_0];
  ref_index[LIST_1] = blk->ref_index[LIST_1];
  list = blk->list;
  scale[LIST_0] = blk->scale[LIST_0];
  scale[LIST_1] = blk->scale[LIST_1];
  DC[LIST_0] = blk->DC[LIST_0];
  DC[LIST_1] = blk->DC[LIST_1];
  BP_DC = blk->BP_DC;
  BP_scale[0] = blk->BP_scale[0];
  BP_scale[1] = blk->BP_scale[1];

#ifdef MVCOMP_ZERORES
  mv_zero_res = blk->mv_zero_res;
#endif

  mvshare_size[0] = blk->mvshare_size[0];
  mvshare_size[1] = blk->mvshare_size[1];
}


#ifdef MVCOMP_ZERORES
// Return the |v1-v2| value.
int PredBlk::mv_vector_diff(PredBlk *blk, int list_idx)
{
    int dx;
    int dy;
    dx = abs(blk->mv[list_idx][0] - mv[list_idx][0]);
    dy = abs(blk->mv[list_idx][1] - mv[list_idx][1]);

    return dx+dy;
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Compare motion data
*
* Compares two sets of motion data relevant to prediction mode cur_list.
*
* \param[in] blk Motion data to compare with
* \param[in] cur_list Prediction mode (LIST_0, LIST_1, BI_PRED)
* \return TRUE if equal, FALSE if not
*/
//-----------------------------------------------------------------------------
int PredBlk::equal(PredBlk *blk, int cur_list) 
{
  if (intra != NULL)
    return blk->intra != NULL;
  if (blk->intra != NULL)
    return FALSE;

  if (cur_list != BI_PRED) 
  {
    if (mv[cur_list][0]     != blk->mv[cur_list][0] ||
      mv[cur_list][1]     != blk->mv[cur_list][1] ||
      ref_index[cur_list] != blk->ref_index[cur_list] ||
      DC[cur_list]        != blk->DC[cur_list]    ||
      scale[cur_list]     != blk->scale[cur_list])
      return FALSE;
  }
  else 
  {
    if (mv[LIST_0][0]     != blk->mv[LIST_0][0] ||
      mv[LIST_0][1]     != blk->mv[LIST_0][1] ||
      mv[LIST_1][0]     != blk->mv[LIST_1][0] ||
      mv[LIST_1][1]     != blk->mv[LIST_1][1] ||
      ref_index[LIST_0] != blk->ref_index[LIST_0] ||
      ref_index[LIST_1] != blk->ref_index[LIST_1] ||
      BP_scale[0]       != blk->BP_scale[0]   ||
      BP_scale[1]       != blk->BP_scale[1]   ||
      BP_DC             != blk->BP_DC)
      return FALSE;
  }

  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compare motion data
*
* Compares two sets of motion data. Prediction mode must match as well.
*
* \param[in] blk Motion data to compare with
* \return TRUE if equal, FALSE if not
*/
//-----------------------------------------------------------------------------
int PredBlk::equal(PredBlk *blk) 
{
  if (intra != NULL)
    return blk->intra != NULL;

  if (blk->intra != NULL)
    return FALSE;

  if (list != blk->list)
    return FALSE;

  if (list != BI_PRED) 
  {
    if (mv[list][0] != blk->mv[list][0])
      return FALSE;    
    if (mv[list][1] != blk->mv[list][1])
      return FALSE;    
    if (ref_index[list] != blk->ref_index[list])
      return FALSE;
    if (DC[list] != blk->DC[list])
      return FALSE;        
    if (scale[list] != blk->scale[list])
      return FALSE;
  }
  else 
  {
    // LIST 0
    if (mv[LIST_0][0] != blk->mv[LIST_0][0])
      return FALSE;   
    if (mv[LIST_0][1] != blk->mv[LIST_0][1])
      return FALSE;
    if (ref_index[LIST_0] != blk->ref_index[LIST_0])
      return FALSE;
    // LIST 1
    if (mv[LIST_1][0] != blk->mv[LIST_1][0])
      return FALSE;    
    if (mv[LIST_1][1] != blk->mv[LIST_1][1])
      return FALSE;       
    if (ref_index[LIST_1] != blk->ref_index[LIST_1])
      return FALSE;
    //WP params
    if (BP_DC != blk->BP_DC)
      return FALSE;    
    if (BP_scale[0] != blk->BP_scale[0])
      return FALSE;    
    if (BP_scale[1] != blk->BP_scale[1])
      return FALSE;
  }  

  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compare motion data
*
* Compares two sets of motion data. Motion vectors are considered to match if
* differences of individual components are equal to or less than th.
*
* \param[in] blk Motion data to compare with
* \param[in] th Threshold
* \return TRUE if equal, FALSE if not
*/
//-----------------------------------------------------------------------------
int PredBlk::soft_equal(PredBlk *blk, int th)
{
  if (intra != NULL)
    return blk->intra != NULL;

  if (blk->intra != NULL)
    return FALSE;

  if (list != blk->list)
    return FALSE;

  if (list != BI_PRED) 
  {
    if (abs(mv[list][0] - blk->mv[list][0]) > th)
      return FALSE;   
    if (abs(mv[list][1] - blk->mv[list][1]) > th)
      return FALSE;    
    if (ref_index[list] != blk->ref_index[list])
      return FALSE;
    if (DC[list] != blk->DC[list])
      return FALSE;        
    if (scale[list] != blk->scale[list])
      return FALSE;
  }
  else  
  {
    // LIST 0
    if (abs(mv[LIST_0][0] - blk->mv[LIST_0][0]) > th)
      return FALSE;   
    if (abs(mv[LIST_0][1] - blk->mv[LIST_0][1]) > th)
      return FALSE;
    if (ref_index[LIST_0] != blk->ref_index[LIST_0])
      return FALSE;
    // LIST 1
    if (abs(mv[LIST_1][0] - blk->mv[LIST_1][0]) > th)
      return FALSE;   
    if (abs(mv[LIST_1][1] - blk->mv[LIST_1][1]) > th)
      return FALSE;       
    if (ref_index[LIST_1] != blk->ref_index[LIST_1])
      return FALSE;
    //WP params
    if (BP_DC != blk->BP_DC)
      return FALSE;    
    if (BP_scale[0] != blk->BP_scale[0])
      return FALSE;    
    if (BP_scale[1] != blk->BP_scale[1])
      return FALSE;
  }

  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compare motion data
*
* Compares two sets of motion data relevant to prediction mode cur_list.
* Presence of intra mode is not considered.
*
* \param[in] blk Motion data to compare with
* \param[in] cur_list Prediction mode (LIST_0, LIST_1, BI_PRED)
* \return TRUE if equal, FALSE if not
*/
//-----------------------------------------------------------------------------
int PredBlk::equal_pred(PredBlk *blk, int cur_list) 
{
  if (cur_list != BI_PRED) 
  {
    if (mv[cur_list][0]     != blk->mv[cur_list][0] ||
      mv[cur_list][1]     != blk->mv[cur_list][1] ||
      DC[cur_list]        != blk->DC[cur_list]    ||
      scale[cur_list]     != blk->scale[cur_list])
      return FALSE;
  }
  else 
  {
    if (mv[LIST_0][0]     != blk->mv[LIST_0][0] ||
      mv[LIST_0][1]     != blk->mv[LIST_0][1] ||
      mv[LIST_1][0]     != blk->mv[LIST_1][0] ||
      mv[LIST_1][1]     != blk->mv[LIST_1][1] ||
      BP_scale[0]       != blk->BP_scale[0]   ||
      BP_scale[1]       != blk->BP_scale[1]   ||
      BP_DC             != blk->BP_DC)
      return FALSE;
  }

  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compare motion data
*
* Compares two sets of motion data 
*
* \param[in] blk Motion data to compare with
* \return TRUE true if two blocks have the same motion parameters, FALSE if not
*/
//-----------------------------------------------------------------------------
int PredBlk::equal_mot(PredBlk *blk) const 
{
  if (intra != NULL)
    return blk->intra != NULL;

  if (blk->intra != NULL)
    return FALSE;

  if (list != blk->list)
    return FALSE;

  if (list != BI_PRED) 
  {
    if (mv[list][0]     != blk->mv[list][0] ||
      mv[list][1]     != blk->mv[list][1] ||
      ref_index[list] != blk->ref_index[list])
      return FALSE;
  }
  else 
  {
    if (mv[LIST_0][0]     != blk->mv[LIST_0][0] ||
      mv[LIST_0][1]     != blk->mv[LIST_0][1] ||
      ref_index[LIST_0] != blk->ref_index[LIST_0] ||
      mv[LIST_1][0]     != blk->mv[LIST_1][0] ||
      mv[LIST_1][1]     != blk->mv[LIST_1][1] ||
      ref_index[LIST_1] != blk->ref_index[LIST_1])
      return FALSE;
  }

  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compare IC offset
*
* Compares two sets of IC offset 
*
* \param[in] blk Motion data to compare with
* \return TRUE true if two blocks have the same IC DC parameters, FALSE if not
*/
//-----------------------------------------------------------------------------
int PredBlk::equal_ic_DC(PredBlk *blk) const 
{

  if (list != BI_PRED) 
  {
    if (DC[list] != blk->DC[blk->list])
      return FALSE;        
  }
  else  
  {
    //WP params
    if (BP_DC != blk->BP_DC)
      return FALSE;    
  }

  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Compare IC scale
*
* Compares two sets of IC scale 
*
* \param[in] blk Motion data to compare with
* \return TRUE true if two blocks have the same IC scale parameters, FALSE if not
*/
//-----------------------------------------------------------------------------
int PredBlk::equal_ic_scale(PredBlk *blk) const 
{
  if (list != BI_PRED) 
  {
    if (scale[list] != blk->scale[blk->list])
      return FALSE;        
  }
  else  
  {
    if (BP_scale[0] != blk->BP_scale[0])
      return FALSE;    
    if (BP_scale[1] != blk->BP_scale[1])
      return FALSE;    
  }

  return TRUE;
}

#ifdef MV_CODING_ENHANCE
//-----------------------------------------------------------------------------
/*!
* \brief Computes normalized motion vectors
*
* Computes normalized motion vectors for each list. If a list is not valid, the 
* normalized motion vectors of the other list are used to set the normalized
* motion vectors for the current list.
* 
* \param[in] ref_mgr Reference Frame Manager
*/
//-----------------------------------------------------------------------------
void PredBlk::compute_mv_norm(RefFrameManager *ref_mgr)
{
  int listIter, inv_dist;

  for (listIter = LIST_0; listIter <= LIST_1; listIter++)
  {
    if (list == listIter || list == BI_PRED)
    {
      inv_dist = ref_mgr->inv_dist[listIter][ref_index[listIter]];
      mv_norm[listIter][0] = mv[listIter][0] * inv_dist;
      mv_norm[listIter][1] = mv[listIter][1] * inv_dist;
    }
    else
    {
      mv_norm[listIter][0] = 0;
      mv_norm[listIter][1] = 0;
    }
  }
  for (listIter = LIST_0; listIter <= LIST_1; listIter++)
  {
    if (list != listIter && list != BI_PRED)
    {
      mv_norm[listIter][0] = mv_norm[listIter ^ 1][0];
      mv_norm[listIter][1] = mv_norm[listIter ^ 1][1];
    }
  }
}
#endif //MV_CODING_ENHANCE

#ifdef MV_CODING_ENHANCE
//-----------------------------------------------------------------------------
/*!
* \brief Adjust motion data predictor
*
* Adjusts the motion data predictor used for differential coding of motion
* data. The adjustement takes the reference frame indices of the current
* and predicting blocks into account.
*
* \param[in] blk Motion data of predicting block
* \param[in] target Motion data of current block
* \param[in] ref_mgr Reference Frame Manager
* \return Adjusted motion data predictor
*/
//-----------------------------------------------------------------------------
#ifdef PMV0_AT_OUT_OF_BOUNDARY
PredBlk *PredBlk::get_prediction(int i, int j, PredBlk *blk, PredBlk *target, RefFrameManager *ref_mgr)
#else
PredBlk *PredBlk::get_prediction(PredBlk *blk, PredBlk *target, RefFrameManager *ref_mgr)
#endif

#else
//-----------------------------------------------------------------------------
/*!
* \brief Adjust motion data predictor
*
* Adjusts the motion data predictor used for differential coding of motion
* data. The adjustement takes the reference frame indices of the current
* and predicting blocks into account.
*
* \param[in] blk Motion data of predicting block
* \param[in] target Motion data of current block
* \return Adjusted motion data predictor
*/
//-----------------------------------------------------------------------------
PredBlk *PredBlk::get_prediction(PredBlk *blk, PredBlk *target)
#endif //MV_CODING_ENHANCE
{
  int l;
  PredBlk *defaultp;
#ifdef MV_CODING_ENHANCE
  int rscale[2];
#endif

#ifdef PMV0_AT_OUT_OF_BOUNDARY
  PredBlk *zero_pmv;
#endif 

  defaultp = NULL;
  
#ifdef PMV0_AT_OUT_OF_BOUNDARY
  if (blk == NULL || blk->intra != NULL){
  if((blk == NULL) && (i==0 || j==0)){
      zero_pmv = & zero_mv;      
    return zero_pmv;
  }
  else{
      return defaultp;
  }
  }

#else

  if (blk == NULL || blk->intra != NULL)
    return defaultp;

#endif

  l = target->list;

  if (l == BI_PRED) 
  {
    if (blk->list == BI_PRED) 
    {
      if (blk->ref_index[LIST_0] == target->ref_index[LIST_0] && blk->ref_index[LIST_1] == target->ref_index[LIST_1])
        return blk;
      else if (blk->ref_index[LIST_0] == target->ref_index[LIST_0]) 
      {
        mv[LIST_0][0] = blk->mv[LIST_0][0];
        mv[LIST_0][1] = blk->mv[LIST_0][1];
        mv[LIST_1][0] = 0;
        mv[LIST_1][1] = 0;
#ifdef MV_CODING_ENHANCE
        ref_index[LIST_0] = blk->ref_index[LIST_0];
        ref_index[LIST_1] = 0;
        list = LIST_0;
#else
        return this;
#endif
      }
      else if (blk->ref_index[LIST_1] == target->ref_index[LIST_1]) 
      {
        mv[LIST_1][0] = blk->mv[LIST_1][0];
        mv[LIST_1][1] = blk->mv[LIST_1][1];
        mv[LIST_0][0] = 0;
        mv[LIST_0][1] = 0;
#ifdef MV_CODING_ENHANCE
        ref_index[LIST_1] = blk->ref_index[LIST_1];
        ref_index[LIST_0] = 0;
        list = LIST_1;
#else
        return this;
#endif
      }
      else
        return defaultp;
    }
    else if (blk->ref_index[blk->list] == target->ref_index[blk->list]) 
    {
      mv[blk->list][0] = blk->mv[blk->list][0];
      mv[blk->list][1] = blk->mv[blk->list][1];
      mv[blk->list^1][0] = 0;
      mv[blk->list^1][1] = 0;
#ifdef MV_CODING_ENHANCE
      ref_index[blk->list] = blk->ref_index[blk->list];
      ref_index[blk->list^1] = 0;
      list = blk->list;
#else
      return this;
#endif
    }
    else
      return defaultp;
  }
  else 
  {
    if (blk->list == BI_PRED && blk->ref_index[l] == target->ref_index[l])
      return blk;
    else if (blk->list != l || blk->ref_index[blk->list] != target->ref_index[l])
      return defaultp;
    else
      return blk;
  }
#ifdef MV_CODING_ENHANCE
  compute_mv_norm(ref_mgr);

  if (l == BI_PRED)
  {
    rscale[LIST_0] = ref_mgr->dist[LIST_0][target->ref_index[LIST_0]];
    rscale[LIST_1] = ref_mgr->dist[LIST_1][target->ref_index[LIST_1]];
    mv[LIST_0][0] = shift_right_round(mv_norm[LIST_0][0] * rscale[LIST_0], 12);
    mv[LIST_0][1] = shift_right_round(mv_norm[LIST_0][1] * rscale[LIST_0], 12);
    mv[LIST_1][0] = shift_right_round(mv_norm[LIST_1][0] * rscale[LIST_1], 12);
    mv[LIST_1][1] = shift_right_round(mv_norm[LIST_1][1] * rscale[LIST_1], 12);
  }
  else
  {
    rscale[l] = ref_mgr->dist[l][target->ref_index[l]];
    mv[l][0] = shift_right_round(mv_norm[l][0] * rscale[l], 12);
    mv[l][1] = shift_right_round(mv_norm[l][1] * rscale[l], 12);
  }

  return this;
#endif
}

#ifdef MVCOMP_ZERORES
// Copy some motion info from a neighbour, ref, mv, dc, scale.
// It's all copied to the indicated list.
void PredBlk::copy_mvinfo_from(PredBlk *neighbour, int list_idx)
{
    int ref_diff = ref_index[list_idx] != neighbour->ref_index[list_idx];
    ref_index[list_idx] = neighbour->ref_index[list_idx];
    mv[list_idx][0] = neighbour->mv[list_idx][0];
    mv[list_idx][1] = neighbour->mv[list_idx][1];
    // Zero IC if coming from a different reference, or different
    // type of mb.
    if (ref_diff || ((neighbour->list == BI_PRED) != (list == BI_PRED)))
        {
            DC[list_idx] = 0;
            scale[list_idx] = 0;
            return;
        }
    DC[list_idx] = neighbour->list == BI_PRED ? neighbour->BP_DC : neighbour->DC[list_idx];
    scale[list_idx] = neighbour->list == BI_PRED ? neighbour->BP_scale[list_idx] : neighbour->scale[list_idx];
}
#endif

#ifdef MVCOMP_ZERORES
  // If the neighbour exists as INTER using the list we're interested in, we use it, otherwise
  // we use 'unavail'.
#define CHECK_EXISTENCE(neighbour) \
  ((neighbour) != NULL && (neighbour)->intra == NULL && ((neighbour)->list == list_idx || (neighbour)->list == BI_PRED)) \
      ? neighbour \
      : &unavail;

  // Does the neighbour use our ref for the list we're interested in?
#define REFOK(neighbour) ((neighbour)->ref_index[list_idx] == curblk->ref_index[list_idx])

// Examine neighbours to determine the references to use prior to a call
// of get_*_prediction.  The references will be stored in curblk
// along with a correct list setting (LIST_O, LIST_1, BI_PRED).
// We use the minimum valid reference found in a neighbour for each list.
int PredBlk::get_best_median_refs(PredBlk *left, PredBlk *top, PredBlk *topleft)
{
    PredBlk unavail;

    unavail.ref_index[LIST_0] = -1;
    unavail.ref_index[LIST_1] = -1;

    ref_index[LIST_0] = -1;
    ref_index[LIST_1] = -1;

    for (int list_idx = 0; list_idx < 2; list_idx++)
    {
        PredBlk *pred;
        int pred_ref_index;
        pred = CHECK_EXISTENCE(left);
        if (pred->ref_index[list_idx] >= 0)
        {
            pred_ref_index = pred->ref_index[list_idx];
            if (ref_index[list_idx] < 0 || pred_ref_index < ref_index[list_idx])
                ref_index[list_idx] = pred_ref_index;
        }
        pred = CHECK_EXISTENCE(top);
        if (pred->ref_index[list_idx] >= 0)
        {
            pred_ref_index = pred->ref_index[list_idx];
            if (ref_index[list_idx] < 0 || pred_ref_index < ref_index[list_idx])
                ref_index[list_idx] = pred_ref_index;
        }
        pred = CHECK_EXISTENCE(topleft);
        if (pred->ref_index[list_idx] >= 0)
        {
            pred_ref_index = pred->ref_index[list_idx];
            if (ref_index[list_idx] < 0 || pred_ref_index < ref_index[list_idx])
                ref_index[list_idx] = pred_ref_index;
        }
    }

    if (ref_index[LIST_0] >= 0 && ref_index[LIST_1] >= 0)
        list = BI_PRED;
    else if (ref_index[LIST_0] >= 0)
    {
        list = LIST_0;
        ref_index[LIST_1] = 0;
    }
    else if (ref_index[LIST_1] >= 0)
    {
        list = LIST_1;
        ref_index[LIST_0] = 0;
    }
    else
    {
        ref_index[LIST_0] = 0;
        ref_index[LIST_1] = 0;
        list = LIST_0;
        return FALSE; // No ref from neighbour.
    }

    return TRUE;
}

int PredBlk::get_best_temporal_refs(RefFrameManager * /*ref_mgr*/, int /*i*/, int /*j*/, PredBlk *left, PredBlk *top, PredBlk *topleft)
{
    return get_best_median_refs(left, top, topleft);
}
#endif

#ifdef MVCOMP_ZERORES
/*
 * We're the neighbour -- we scale our mvs to account for any ref differences
 * between curblk and ourselves.
 */
void PredBlk::ref_scale_mvs(RefFrameManager *ref_mgr, PredBlk *curblk, int list_idx)
{

    int ref_neighbour = ref_index[list_idx];
    int ref_current   = curblk->ref_index[list_idx];

    if (ref_neighbour < 0 || ref_neighbour == ref_current)
        return;

    RefFrame *ref_frame_neighbour = ref_mgr->ref[ref_mgr->list[list_idx][ref_neighbour]];
    RefFrame *ref_frame_current   = ref_mgr->ref[ref_mgr->list[list_idx][ref_current  ]];
    int inter_pel_neighbour = ref_frame_current->saved_inter_pel[list_idx][ref_neighbour];
    int inter_pel_current   = ref_frame_current->saved_inter_pel[list_idx][ref_current  ];
    int ref_fno_neighbour = ref_frame_neighbour->frame_no;
    int ref_fno_current   = ref_frame_current  ->frame_no;

    int d_neighbour = ref_mgr->cur_frame_no - ref_fno_neighbour; // left mb's reference distance, in frames.
    int d_current   = ref_mgr->cur_frame_no - ref_fno_current  ; // Our reference distance, in frames.
    mv[list_idx][0] = shift_right_round(mv[list_idx][0]*power2(RESCALE_BITS)*d_current*inter_pel_current/d_neighbour/inter_pel_neighbour, RESCALE_BITS);
    mv[list_idx][1] = shift_right_round(mv[list_idx][1]*power2(RESCALE_BITS)*d_current*inter_pel_current/d_neighbour/inter_pel_neighbour, RESCALE_BITS);
}

int PredBlk::get_temporal_prediction(RefFrameManager *ref_mgr, int i, int j, PredBlk * /*left*/, PredBlk * /*top*/, PredBlk * /*topleft*/, PredBlk *curblk, int list_idx)
{
  RefFrame *frame_ref; // Our reference frame.
  RefFrame *frame_colocated; // Co-located frame.  The closest in the list.
  PredBlk *mb_colocated; // Co-located mb.
  PredBlk *p_mb_colocated;
  int ref_colocated; // The ref used by the co-located mb.
  int fno_colocated; // Frame no of co-located frame.
  int fno_at_ref_colocated; // Frame no of co-located ref frame.
  int inter_pel_colocated; // Interpolation accuracy at co-located mb with its reference.
  int log_inter_pel_colocated; // Log of inter_pel_colocated;

  int fno_cur; // Our frame number.
  int fno_at_ref_cur; // Our ref's frame number.
  int inter_pel_cur; // Interpolation accuracy of current mb and reference.
  PredBlk unavail;

  // Look at the co-located mb, and use its vector, scaling
  // where necessary to account for different ref distances
  // and different ref pelness.
  frame_colocated = ref_mgr->ref[ref_mgr->list[list_idx][0]];
  mb_colocated = frame_colocated->saved_mf->get_pblk(i, j);

  // For now, restrict to non-B.
  if (frame_colocated->frame->frame_type == ft_B_frame)
      return FALSE; // No predictor.

  p_mb_colocated = CHECK_EXISTENCE(mb_colocated);
  if (p_mb_colocated == &unavail)
      return FALSE;

  ref_colocated = mb_colocated->ref_index[list_idx];
  inter_pel_colocated = frame_colocated->saved_inter_pel[list_idx][ref_colocated];
  log_inter_pel_colocated = frame_colocated->saved_log_inter_pel[list_idx][ref_colocated];
  fno_colocated = frame_colocated->frame_no;
  fno_at_ref_colocated = frame_colocated->saved_ref_framenos[frame_colocated->saved_ref_list[list_idx][ref_colocated]];

  frame_ref = ref_mgr->ref[ref_mgr->list[list_idx][curblk->ref_index[list_idx]]];
  inter_pel_cur = ref_mgr->ref_process->inter_pel[list_idx][curblk->ref_index[list_idx]];
  fno_cur = ref_mgr->cur_frame_no;
  fno_at_ref_cur = frame_ref->frame_no;

  // Set unavail ref to illegal value and mv info to zero.
  unavail.ref_index[list_idx] = -1;
  unavail.mv[list_idx][0] = 0;
  unavail.mv[list_idx][1] = 0;
  unavail.BP_DC = 0;
  unavail.BP_scale[list_idx] = 0;
  unavail.DC[list_idx] = 0;
  unavail.scale[list_idx] = 0;

  // Apply scale to compensate for any change in reference image distance.
  int d_cur = fno_cur-fno_at_ref_cur;  // Our reference distance, in frames.
  int d_col = fno_colocated-fno_at_ref_colocated; // Co-located's reference distance, in frames.

  copy_mvinfo_from(p_mb_colocated, list_idx);
  mv[list_idx][0] = shift_right_round(mv[list_idx][0]*power2(RESCALE_BITS)*d_cur*inter_pel_cur/d_col/inter_pel_colocated, RESCALE_BITS);
  mv[list_idx][1] = shift_right_round(mv[list_idx][1]*power2(RESCALE_BITS)*d_cur*inter_pel_cur/d_col/inter_pel_colocated, RESCALE_BITS);
  // Zero the IC all the time; either the ref is different or the ref
  // distance is different.
  scale[list_idx] = 0;
  DC[list_idx] = 0;

  ref_index[list_idx] = curblk->ref_index[list_idx];

  return TRUE;
}

// Calls the above get_median_prediction twice, once for each list,
// then determine single- or bi-pred mode.
PredBlk *PredBlk::get_temporal_prediction(RefFrameManager *ref_mgr, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft, PredBlk *curblk)
{
  // for now, restrict to non-B.
  if (ref_mgr->cur_frame_type == ft_B_frame)
      return NULL;

  int pred_median0 = curblk->list != LIST_1 ? get_temporal_prediction(ref_mgr, i, j, left, top, topleft, curblk, LIST_0) : 0;
  int pred_median1 = curblk->list != LIST_0 ? get_temporal_prediction(ref_mgr, i, j, left, top, topleft, curblk, LIST_1) : 0;

  if (pred_median0 && pred_median1)
  {
      list = BI_PRED;
      // Transfer info to bipred fields.
      BP_DC = 0;
      BP_scale[LIST_0] = 0;
      BP_scale[LIST_1] = 0;
  }
  else if (pred_median0)
  {
      list = LIST_0;
  }
  else if (pred_median1)
  {
      list = LIST_1;
  }
  else
      return NULL;

  return this;
}

#endif

//-----------------------------------------------------------------------------
/*!
* \brief Get context for coding motion data prediction direction
*
* \note The object to which this method is applied should contain the motion
* data of the block to the upper-left of the current block.
*
* \param[in] left Motion data of block to the left of current block
* \param[in] top Motion data of block above current block
* \param[in] target_list Prediction list used by current block (LIST_0, LIST_1, BI_PRED)
* \return Context index (0 or 1)
*/
//-----------------------------------------------------------------------------
#ifndef MVCOMP_ZERORES
int PredBlk::get_pred_context(PredBlk *left, PredBlk *top, int target_list)
#else
int PredBlk::get_pred_context(PredBlk *left, PredBlk *top, int target_list, int extra_pred)
#endif
{
  int r;

#ifdef MVCOMP_ZERORES
  if (extra_pred)
      return 2;
#endif

  if (target_list != BI_PRED) 
  {
    r  = iabs(mv[target_list][0] - top->mv[target_list][0]) - iabs(mv[target_list][0] - left->mv[target_list][0]);
    r += iabs(mv[target_list][1] - top->mv[target_list][1]) - iabs(mv[target_list][1] - left->mv[target_list][1]);
    r += iabs(scale[target_list] - top->scale[target_list]) - iabs(scale[target_list] - left->scale[target_list]);
    r += iabs(DC   [target_list] - top->DC   [target_list]) - iabs(DC   [target_list] - left->DC   [target_list]);
  }
  else 
  {
    r  = iabs(mv[LIST_0][0] - top->mv[LIST_0][0]) - iabs(mv[LIST_0][0] - left->mv[LIST_0][0]);
    r += iabs(mv[LIST_0][1] - top->mv[LIST_0][1]) - iabs(mv[LIST_0][1] - left->mv[LIST_0][1]);
    r += iabs(mv[LIST_1][0] - top->mv[LIST_1][0]) - iabs(mv[LIST_1][0] - left->mv[LIST_1][0]);
    r += iabs(mv[LIST_1][1] - top->mv[LIST_1][1]) - iabs(mv[LIST_1][1] - left->mv[LIST_1][1]);
    r += iabs(BP_scale[0]   - top->BP_scale[0]  ) - iabs(BP_scale[0]   - left->BP_scale[0]  );
    r += iabs(BP_scale[1]   - top->BP_scale[1]  ) - iabs(BP_scale[1]   - left->BP_scale[1]  );
    r += iabs(BP_DC         - top->BP_DC        ) - iabs(BP_DC         - left->BP_DC        );
  }

  return r >= ZERO;
}

#ifdef MV_CODING_ENHANCE
//-----------------------------------------------------------------------------
/*!
* \brief Subtract motion data
*
* \param[in] blk Motion data to be subtracted
* \param[in] ref_mgr Reference Frame Manager
*/
//-----------------------------------------------------------------------------
void PredBlk::sub(PredBlk *blk, RefFrameManager *ref_mgr)
#else
//-----------------------------------------------------------------------------
/*!
* \brief Subtract motion data
*
* \param[in] blk Motion data to be subtracted
*/
//-----------------------------------------------------------------------------
void PredBlk::sub(PredBlk *blk)
#endif
{
#ifdef MV_CODING_ENHANCE
  int first_list, rscale[2];
  int flag_xpred = 0;
#endif
  if (list != BI_PRED)  
  {
    mv[list][vertical  ] -= blk->mv[list][vertical  ];
    mv[list][horizontal] -= blk->mv[list][horizontal];    
    scale[list] -= blk->scale[list];
    DC[list] -= blk->DC[list];

    mv[list^1][vertical  ] = 0;
    mv[list^1][horizontal] = 0;    
    scale[list^1] = 0;
    DC[list^1] = 0;
    BP_scale[0] = 0;
    BP_scale[1] = 0;
    BP_DC = 0;
  }
  else  
  {
    mv[LIST_0][vertical  ] -= blk->mv[LIST_0][vertical  ];
    mv[LIST_0][horizontal] -= blk->mv[LIST_0][horizontal];
    mv[LIST_1][vertical  ] -= blk->mv[LIST_1][vertical  ];
    mv[LIST_1][horizontal] -= blk->mv[LIST_1][horizontal];
#ifdef MV_CODING_ENHANCE
    rscale[LIST_0] = ref_mgr->dist[LIST_0][ref_index[LIST_0]];
    rscale[LIST_1] = ref_mgr->dist[LIST_1][ref_index[LIST_1]];
    first_list = abs(rscale[LIST_0]) < abs(rscale[LIST_1]);
    flag_xpred = (sign(blk->mv[LIST_0][vertical]*rscale[LIST_0])==sign(blk->mv[LIST_1][vertical]*rscale[LIST_1]));
    flag_xpred = (flag_xpred && (sign(blk->mv[LIST_0][horizontal]*rscale[LIST_0])==sign(blk->mv[LIST_1][horizontal]*rscale[LIST_1])));
    if (flag_xpred)
    {
      rscale[first_list] = ref_mgr->inv_dist[first_list][ref_index[first_list]];
      mv[first_list^1][vertical  ] -= shift_right_round(mv[first_list][vertical  ] * rscale[0] * rscale[1], 12);
      mv[first_list^1][horizontal] -= shift_right_round(mv[first_list][horizontal] * rscale[0] * rscale[1], 12);
    }
#endif

    BP_scale[0] -= blk->BP_scale[0];
    BP_scale[1] -= blk->BP_scale[1];
    BP_DC -= blk->BP_DC;
    scale[LIST_0] = scale[LIST_1] = 0;
    DC[LIST_0] = DC[LIST_1] = 0;
  }
}

#ifdef MV_CODING_ENHANCE
//-----------------------------------------------------------------------------
/*!
* \brief Add motion data
*
* \param[in] blk Motion data to be added
* \param[in] ref_mgr Reference Frame Manager
*/
//-----------------------------------------------------------------------------
void PredBlk::add(PredBlk *blk, RefFrameManager *ref_mgr)
#else
//-----------------------------------------------------------------------------
/*!
* \brief Add motion data
*
* \param[in] blk Motion data to be added
*/
//-----------------------------------------------------------------------------
void PredBlk::add(PredBlk *blk)
#endif
{
#ifdef MV_CODING_ENHANCE
  int first_list, rscale[2];
  int flag_xpred = 0;
#endif
  if (list != BI_PRED)  
  {
    mv[list][vertical  ] += blk->mv[list][vertical  ];
    mv[list][horizontal] += blk->mv[list][horizontal];    
    scale[list] += blk->scale[list];
    DC[list] += blk->DC[list];

    mv[list^1][vertical  ] = 0;
    mv[list^1][horizontal] = 0;    
    scale[list^1] = 0;
    DC[list^1] = 0;
    BP_scale[0] = 0;
    BP_scale[1] = 0;
    BP_DC = 0;
  }
  else  
  {
#ifdef MV_CODING_ENHANCE
    rscale[LIST_0] = ref_mgr->dist[LIST_0][ref_index[LIST_0]];
    rscale[LIST_1] = ref_mgr->dist[LIST_1][ref_index[LIST_1]];
    first_list = abs(rscale[LIST_0]) < abs(rscale[LIST_1]);
    flag_xpred = (sign(blk->mv[LIST_0][vertical]*rscale[LIST_0])==sign(blk->mv[LIST_1][vertical]*rscale[LIST_1]));
    flag_xpred = (flag_xpred && (sign(blk->mv[LIST_0][horizontal]*rscale[LIST_0])==sign(blk->mv[LIST_1][horizontal]*rscale[LIST_1])));
    if (flag_xpred)
    {
      rscale[first_list] = ref_mgr->inv_dist[first_list][ref_index[first_list]];
      mv[first_list^1][vertical  ] += shift_right_round(mv[first_list][vertical  ] * rscale[0] * rscale[1], 12);
      mv[first_list^1][horizontal] += shift_right_round(mv[first_list][horizontal] * rscale[0] * rscale[1], 12);
    }
#endif
    mv[LIST_0][vertical  ] += blk->mv[LIST_0][vertical  ];
    mv[LIST_0][horizontal] += blk->mv[LIST_0][horizontal];
    mv[LIST_1][vertical  ] += blk->mv[LIST_1][vertical  ];
    mv[LIST_1][horizontal] += blk->mv[LIST_1][horizontal];

    BP_scale[0] += blk->BP_scale[0];
    BP_scale[1] += blk->BP_scale[1];
    BP_DC += blk->BP_DC;
    scale[LIST_0] = scale[LIST_1] = 0;
    DC[LIST_0] = DC[LIST_1] = 0;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Scale motion vector
*
* Scales the motion vector of list cur_list in motion data blk to match the
* temporal scaling resulting from changing the reference index to that of the
* current motion data.
*
* \param[in] ref_mgr Pointer to reference frame manager
* \param[in] blk Motion data to be scaled
* \param[in] cur_list List to be scaled (LIST_0 or LIST_1)
* \return Motion data with scaled motion vector
*/
//-----------------------------------------------------------------------------
PredBlk *PredBlk::rescale(RefFrameManager *ref_mgr, PredBlk *blk, int cur_list) 
{
  int temp_inter_pel;

  if (blk == NULL || blk->intra != NULL)
    return NULL;
  else if (blk->ref_index[cur_list] == ref_index[cur_list]) 
    return blk;
  else 
  {    
    temp_inter_pel = ref_mgr->ref_process->inter_pel[cur_list][ref_index[cur_list]];
    mv[cur_list][vertical  ] = clip(shift_right_round(ref_mgr->mv_scale[cur_list][ref_index[cur_list]][blk->ref_index[cur_list]] * blk->mv[cur_list][vertical  ], RESCALE_BITS), -temp_inter_pel * param.me_range,temp_inter_pel * param.me_range-1);
    mv[cur_list][horizontal] = clip(shift_right_round(ref_mgr->mv_scale[cur_list][ref_index[cur_list]][blk->ref_index[cur_list]] * blk->mv[cur_list][horizontal], RESCALE_BITS), -temp_inter_pel * param.me_range,temp_inter_pel * param.me_range-1);

    scale[cur_list] = blk->scale[cur_list];
    DC[cur_list] = blk->DC[cur_list];
    ref_index[cur_list] = blk->ref_index[cur_list];
    return this;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Check whether motion vector is in bounds
*
* Checks if the motion vector of list curr_list in within the search range
* bounds defined by Param::me_range.
*
* \param[in] ref_mgr RefFrameManager to be considered
* \param[in] curr_list List to be considered
* \param[in] ref Reference frame index
* \return TRUE if in bounds, FALSE otherwise
*/
//-----------------------------------------------------------------------------
int PredBlk::inbounds(RefFrameManager *ref_mgr, int curr_list, int ref)
{

  int temp_inter_pel;

  temp_inter_pel = ref_mgr->ref_process->inter_pel[curr_list][ref];

  if (mv[curr_list][vertical  ] <  -temp_inter_pel * param.me_range ||
    mv[curr_list][vertical  ] >=  temp_inter_pel * param.me_range ||
    mv[curr_list][horizontal] <  -temp_inter_pel * param.me_range ||
    mv[curr_list][horizontal] >=  temp_inter_pel * param.me_range)
    return FALSE;

  return TRUE;
}

//-----------------------------------------------------------------------------
/*!
* \brief Check whether motion vector is in bounds from search center
*
* Checks if the motion vector of list curr_list in within the search range
* bounds defined by Param::me_range.
*
* \param[in] ref_mgr RefFrameManager to be considered
* \param[in] curr_list List to be considered
* \param[in] ref Reference frame index
* \param[in] mv0 Vertical mv tollerance
* \param[in] mv1 Horizontal mv tollerance
* \return TRUE if in bounds, FALSE otherwise
*/
//-----------------------------------------------------------------------------
int PredBlk::offset_inbounds(RefFrameManager *ref_mgr, int curr_list, int ref, int mv0, int mv1)
{

  int temp_inter_pel;

  temp_inter_pel = ref_mgr->ref_process->inter_pel[curr_list][ref];

  if (mv[curr_list][vertical  ] <  -temp_inter_pel * param.me_range +mv0||
    mv[curr_list][vertical  ] >=  temp_inter_pel * param.me_range +mv0||
    mv[curr_list][horizontal] <  -temp_inter_pel * param.me_range +mv1||
    mv[curr_list][horizontal] >=  temp_inter_pel * param.me_range +mv1)
    return FALSE;

  return TRUE;
}


//-----------------------------------------------------------------------------
/*!
* \brief Check whether intensity compensation parameters are in bounds
*
* \return TRUE if in bounds, FALSE otherwise
*/
//-----------------------------------------------------------------------------
int PredBlk::inbounds_wp()
{
  if (list == BI_PRED)
    return abs(BP_scale[0]) <= 16 && abs(BP_scale[1]) <= (param.weight_diff_max >> param.weight_diff_scale) && abs(BP_DC) <= 16;
  else
    return abs(scale[list]) <= 16 && abs(DC[list]) <= 16;
}


//-----------------------------------------------------------------------------
/*!
* \brief Get coding mode associated with motion data
*
* \param[in] mode4    flag to indicate if mode if for 4x4 or for 8x8
* \return 0 for intra, 1: 8x8 LIST_0, 2: 8x8 LIST_1  3: BiPred  4: 4x4 LIST_0, 5: 4x4 LIST_1
*/
//-----------------------------------------------------------------------------
int PredBlk::mode()
{
  if (intra != NULL)// 0: INTRA
    return ZERO;
  else 
  {
    return imin(list + 1, BI_PRED);// 1: 8x8 LIST_0, 2: 8x8 LIST_1  3: BiPred
  }

}


//-----------------------------------------------------------------------------
/*!
* \brief Get coding mode associated with motion data
*
* \param[in] mode4    flag to indicate if mode if for 4x4 or for 8x8
* \return 0 for intra, 1: 8x8 LIST_0, 3: 4x4 LIST_0
*/
//-----------------------------------------------------------------------------
int PredBlk::mode_P()
{
  return (intra != NULL) ? MODE_P_INTRA : MODE_P_LIST_0_8x8;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get coding mode associated with motion data
*
* \return 0 for intra, 1: 8x8 LIST_0, 2: 8x8 LIST_1, 3: 8x8 BiPred
*/
//-----------------------------------------------------------------------------
int PredBlk::mode_B()
{
  if (intra != NULL)// 0: INTRA
    return MODE_INTRA;
  else 
  {
    if (list ==0) 
    {
      return MODE_LIST_0_8x8;  // 1: 8x8 LIST_0
    }
    else if(list ==1) 
    {
      return MODE_LIST_1_8x8;  // 2: 8x8 LIST_1
    }
    else 
    {
      return MODE_BI_PRED;  // 3: 8x8 BiPred
    }
  }
}


//-----------------------------------------------------------------------------
/*!
* \brief Print motion data
*
*/
//-----------------------------------------------------------------------------
void PredBlk::print()
{
  if (intra != NULL)
    printf("INTRA\n");
  else if (list == BI_PRED)
    printf("BI_PRED %d %d %d %d %d %d %d\n", mv[0][0], mv[0][1], mv[1][0], mv[1][1], BP_scale[0], BP_scale[1], BP_DC);
  else
    printf("LIST%d %d %d %d %d\n", list, mv[list][0], mv[list][1], scale[list], DC[list]);
}

#ifdef MVCOMP_ZERORES
PredBlk **PredBlkSet::get_predictor_set_complete(RefFrameManager *ref_mgr, int i, int j, PredBlk *left, PredBlk *top, PredBlk *topleft, PredBlk *curblk)
{
  MVCompConfig *pconfig = param.mv_cur_comp_config();
  predictor_set[MV_PRED_LEFT] = pconfig->mv_predictor_used[MV_PRED_LEFT]
#ifdef MV_CODING_ENHANCE

  #ifdef PMV0_AT_OUT_OF_BOUNDARY
                                    ? predictor[MV_PRED_LEFT].get_prediction(i,j,left, curblk, ref_mgr)
  #else
                                    ? predictor[MV_PRED_LEFT].get_prediction(left, curblk, ref_mgr)
  #endif

#else
                                    ? predictor[MV_PRED_LEFT].get_prediction(left, curblk)
#endif
                                    : NULL;
  predictor_set[MV_PRED_TOP] = pconfig->mv_predictor_used[MV_PRED_TOP]
#ifdef MV_CODING_ENHANCE
  #ifdef PMV0_AT_OUT_OF_BOUNDARY
                                    ? predictor[MV_PRED_TOP].get_prediction(i,j,top,  curblk, ref_mgr)
  #else
                                    ? predictor[MV_PRED_TOP].get_prediction(top,  curblk, ref_mgr)
  #endif
#else
                                    ? predictor[MV_PRED_TOP].get_prediction(top,  curblk)
#endif
                                    : NULL;
  predictor_set[MV_PRED_TEMPORAL] = pconfig->mv_predictor_used[MV_PRED_TEMPORAL]
                                      ? predictor[MV_PRED_TEMPORAL].get_temporal_prediction(ref_mgr, i, j, left, top, topleft, curblk)
                                      : NULL;
  predictor_set[MV_PRED_ZERO] = &zero_mv;
  predictor_set[MV_PRED_NONE] = NULL;

  return predictor_set;
}

PredBlk **PredBlkSet::collapse_predictor_set(int list, int &n_diff_preds)
{
  MVCompConfig *pconfig = param.mv_cur_comp_config();
  n_diff_preds = 0;

  for (int p = 0; p < pconfig->mv_npredictors; p++)
  {
      int pdst;
      PredBlk *pred = predictor_set[pconfig->mv_predictors[p]];
      if (pred == NULL)
          continue;
      for (pdst = 0; pdst < n_diff_preds; pdst++)
      {
          if (collapsed_predictor_set[pdst]->equal_pred(pred, list))
              break; // same.
      }
      if (pdst == n_diff_preds)
      {
          // store different predictor.
          collapsed_predictor_set[n_diff_preds++] = pred;
      }
  }

  return collapsed_predictor_set;
}

// Look at how we signal our predictors.
// We can further reduce our # predictors, and we can indicate a special
// context for a third predictor.
// Also, for dyanamic probability calculations, we pass in the T context proba and symbol
// counts.  If the proba is given as zero, we use the given 'thresh' directly -- this happens
// during ME where a single thresh is calculated for use during the entire pass.
void PredBlkSet::signal_treatment(int &n_diff_preds, int &extra_pred, int list, int t_proba, int /*t_cnt0*/, int /*t_cnt1*/, int thresh)
{
  MVCompConfig *pconfig = param.mv_cur_comp_config();

  extra_pred = 0; // 1 indicates it's possible we want to signal 'temporal' (our third choice).
  if (n_diff_preds > 2)
    n_diff_preds = 2;

  // If we're looking at x&T, and T is too close to or too far from x, ignore T.
  if (n_diff_preds == 2
      && predictor_set[pconfig->mv_predictors[2]] == collapsed_predictor_set[1]
      && (pconfig->mv_auto_max_temporal_diff || pconfig->mv_max_temporal_diff != 0))
  {
      int ceiling = pconfig->mv_max_temporal_diff;
      if (pconfig->mv_auto_max_temporal_diff)
      {
          // calculate the "dynamic" threshold sqrt((1-p)/p), p is P(t/A,t).
          // P is P(t/A,t) ie, the probability of choosing T when T's in competition.
          // We use a dedicated context for signalling T in this case, so we get its
          // stats.

          //int proba = edec->proba_mvp(2);
          //int cnt0 = edec->count_mvp(2, 0);
          //int cnt1 = edec->count_mvp(2, 1);
          // Probability of signalling T is cnt1/(cnt0+cnt1) -- we signal T using '1'.
          // Actually, we use the instantaneous probability, 'p', as *that* defines
          // the cost of signalling.  Not the complete image counts.
          //double p = cnt0+cnt1 == 0 || cnt1 == 0 ? ((double)0x10000 - proba)/0x10000: (double)cnt1/(cnt0+cnt1);
          if (t_proba != 0)
          {
              // dynamic -- using last known probability of T.
              double p = ((double)0x10000-t_proba)/0x10000;

              double dynamic_thresh = sqrt((1.0-p)/p);
              ceiling = (int)dynamic_thresh;
              ceiling = -ceiling-1; // negative => toss close values, keep far values.
          }
          else
          {
              // ME: pre-calculated thresh given directly.
              ceiling = thresh;
          }
      }
      if (ceiling > 0
        ? collapsed_predictor_set[0]->mv_vector_diff(collapsed_predictor_set[1], list) > ceiling
        : collapsed_predictor_set[0]->mv_vector_diff(collapsed_predictor_set[1], list) < -ceiling)
      {
        n_diff_preds = 1;
        //printf("(t-rejected (dist))\n");
      }
  }

  // If A&B were equal, and we've got an available 3rd (now 2nd) predictor,
  // we want to signal x/3rd using a dedicated context.
  if (n_diff_preds == 2
      && predictor_set[pconfig->mv_predictors[2]] == collapsed_predictor_set[1])
  {
      extra_pred = 1;
  }
}

void PredBlkSet::print(int n_diff_preds, int i, int j)
{
  // print final predictors that are in competition.
  MVCompConfig *pconfig = param.mv_cur_comp_config();
  char names[1000];
  names[0] = 0;
  int p, n;

  for (p = 0; p < n_diff_preds; p++)
      for (n = 0; n < pconfig->mv_npredictors; n++)
          if (predictor_set[pconfig->mv_predictors[n]] == collapsed_predictor_set[p])
          {
              if (names[0] != 0)
                  strcat(names, " ");
              strcat(names, param.mv_pred_name[pconfig->mv_predictors[n]]);
              break;
          }
  printf("MV Y=%d X=%d: %d different predictors (%s)\n", i, j, n_diff_preds, names);
  for (p = 0; p < n_diff_preds; p++)
  {
      printf("MV Y=%d X=%d: predgroup %d: ", i, j, p);
      collapsed_predictor_set[p]->print();
  }
}
#endif

PredBlk zero_mv;
#ifndef SUBSTREAM_MT // remove statics!
#ifdef FASTER_MVC
PredBlk PredBlkSet::predictor[MV_MAX_PRED_INDEX+1]; // working area.
#endif
#endif

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
