/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ProbEst.cpp
* \brief Implementation of ProbEst class
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"
#include "ProbEst.H"

//-----------------------------------------------------------------------------
// Public methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Reset counts
*
*/
//-----------------------------------------------------------------------------
void ProbEst::reset_counts() 
{
  symcount[ZERO] = ZERO;
  symcount[ONE] = ZERO;
  
  bitcount[ZERO] = ZERO;
  bitcount[ONE] = ZERO;
  
  fbitcount = 0.0;
}

//-----------------------------------------------------------------------------
/*!
* \brief Prints statistics to file
*
* Prints symbol counts, renormalization counts, and fractional bit count.
*
* \param[in] f File pointer
*/
//-----------------------------------------------------------------------------
void ProbEst::print(FILE *f) 
{
  fprintf(f, "%03d %-40s ", idCtx, idName);
  fprintf(f, "%8d %8d ", symcount[ZERO], symcount[ONE]);
  fprintf(f, "%8d %8d ", bitcount[ZERO], bitcount[ONE]);
  fprintf(f, "%10.2f\n", fbitcount);
}

//-----------------------------------------------------------------------------
/*!
* \brief Compute entropy of coded sequence
*
* Computes the entropy of the coded sequence based on symbol counts.
*
* \return Entropy
*/
//-----------------------------------------------------------------------------
double ProbEst::entropy() 
{
  int n;
  double r;
  int i;
  
  n = symcount[ZERO] + symcount[ONE];
  if (n == ZERO)
    return 0.0;
  
  r = n * log((double)n);
  
  for (i=ZERO; i<=ONE; i++)
    if (symcount[i] != ZERO)
      r = r - symcount[i] * log((double)symcount[i]);
  
  r = r / log(2.0);
  
  return r;
}

void ProbEst::copyInitialProba(ProbEst *src)
{
  // The copies should be between identical contexts.
  // Can compare name pointers as they're original literal strings.
  if (idCtx != src->idCtx || idName != src->idName)
  {
    printf("copyInitialProba error: src ctxt %d (%s) to ctxt %d (%s)\n",
           src->idCtx, src->idName, idCtx, idName);
    abort();
  }
  i_p = src->i_p;
  p = src->p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Store probability for further reset
*
* Store given probabiliy as default value
*
*/
//-----------------------------------------------------------------------------
void ProbEst::set_default_prob(unsigned int p)
{
  i_p=p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Restore current probability from stored value
*
* Restore current Probabiliy status from default value
*
*/
//-----------------------------------------------------------------------------
void ProbEst::restoreDefaultState()
{
  p=i_p;
}

#ifdef CABAC_SYNC_POINTS
//-----------------------------------------------------------------------------
/*!
* \brief Set external data pointers.
*
* This function is a no-op at this level.  Derived classes do appropriate work.
*/
//-----------------------------------------------------------------------------
void ProbEst::setExternalData(unsigned int *p2d, unsigned int *mnb)
{
  (void)p2d; // unused parameter.
  (void)mnb; // unused parameter.
}
#endif // CABAC_SYNC_POINTS

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
