/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file Quantizer.cpp
* \brief Implementation of Quantizer class
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Quantizer.H"
#include "Param.H"
#include "Global.H"
#include "WFrame.H"

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Allocates memory and initializes offset tables.
*
* \note The entropy model is used to determine the intra field which affects
* the quantization of the DC coefficient.
*
* \param[in] emodel Entropy model (EntropyModel::I_luma, etc.)
*/
//-----------------------------------------------------------------------------
Quantizer::Quantizer(int emodel)
{
  int i, j, k;

  intra = emodel < TWO;

  if (param.trans8x8_enable) 
  {
    for (j = ZERO; j < TWO; j++) 
    { 
      if (param.intra16x16)
      {
        for (i = ZERO; i < TX_NCOEFF16; i++) 
          offset16x16[j][i] = param.q_offset[emodel];              
      }
      for (i = ZERO; i < TX_NCOEFF8; i++) 
      {
        offset8x8[j][i] = param.q_offset[emodel];              
      }
      for (i = ZERO; i < TX_NCOEFF; i++) 
      {
        offset[j][i] = param.q_offset[emodel];              
      }
    }

    if (emodel==0) 
    {
      if (param.intra16x16)
      {
        for (i = ZERO; i < TX_NCOEFF16; i++) 
        {
          offset16x16[2][i] = param.q_offset[4];
          offset16x16[3][i] = param.q_offset[4];
        }
      }
      for (i = ZERO; i < TX_NCOEFF8; i++) 
      {
        offset8x8[2][i] = param.q_offset[4];
        offset8x8[3][i] = param.q_offset[4];
      }
      for (i = ZERO; i < TX_NCOEFF; i++) 
      {
        offset[2][i] = param.q_offset[4];
        offset[3][i] = param.q_offset[4];
      }
    }
    else if(emodel==1) 
    {
      if (param.intra16x16)
      {
        for (i = ZERO; i < TX_NCOEFF16; i++) 
        {
          offset16x16[2][i] = param.q_offset[5];
          offset16x16[3][i] = param.q_offset[5];
        }
      }
      for (i = ZERO; i < TX_NCOEFF8; i++) 
      {
        offset8x8[2][i] = param.q_offset[5];
        offset8x8[3][i] = param.q_offset[5];
      }
      for (i = ZERO; i < TX_NCOEFF; i++) 
      {
        offset[2][i] = param.q_offset[5];
        offset[3][i] = param.q_offset[5];
      }
    }

    inv_scale8x8 = new int[NQPS * 64];
    fwd_scale8x8 = new int[NQPS * 64];
    iquant_scale8x8 = new int[64];
    quant_scale8x8 = new int[64];
    coeff_norm_rdoq_qp8x8 = new double[NQPS * ZTsize64];
    coeff_norm_rdoq8x8 = new double[ZTsize64];
  }  
  else
  {
    for (j = ZERO; j < TWO; j++)
    {
      for (i = ZERO; i < TX_NCOEFF; i++)
      {
        offset[j][i] = param.q_offset[emodel];
      }
    }

    if(emodel==0) 
    {
      for (i = ZERO; i < TX_NCOEFF; i++)
      {
        offset[2][i] = param.q_offset[4];
        offset[3][i] = param.q_offset[4];
      }
    }
    else if(emodel==1) 
    {
      for (i = ZERO; i < TX_NCOEFF; i++)
      {
        offset[2][i] = param.q_offset[5];
        offset[3][i] = param.q_offset[5];
      }
    }
  }

  if (param.intra16x16)
    qp16x16 = new int[TX_NCOEFF16];

  for (k=0; k<NSIP; k++) 
  {
    inv_scale[k] = new int[NQPS * TX_NCOEFF];
    fwd_scale[k] = new int[NQPS * TX_NCOEFF];
    iquant_scale[k] = new int [TX_NCOEFF];
    quant_scale[k] = new int[TX_NCOEFF];
  }
  for (k=0; k<NSIP; k++) 
  {
    coeff_norm_rdoq_qp[k] = new double[NQPS * TX_NCOEFF];
    coeff_norm_rdoq[k] = new double[TX_NCOEFF];
  }

#ifdef Q_OFFSET
  for (k=0; k<2; k++) 
 {
    iTheta[k] = param.iTheta[k];
    iTheta8x8[k] = param.iTheta8x8[k];
  }
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory.
*/
//-----------------------------------------------------------------------------
Quantizer::~Quantizer()
{
  int i; 

  if (param.trans8x8_enable) 
  {
    delete[] inv_scale8x8;
    delete[] fwd_scale8x8;
    delete iquant_scale8x8;
    delete[] quant_scale8x8;
    delete[] coeff_norm_rdoq_qp8x8;
    delete[] coeff_norm_rdoq8x8;
  }

  if (param.intra16x16 && (qp16x16 != NULL))
  {
    delete [] qp16x16;
    qp16x16 = NULL;
  }

  for (i=0; i<NSIP; i++) 
  {
    delete[] inv_scale[i];
    delete[] fwd_scale[i];
    delete iquant_scale[i];
    delete[] quant_scale[i];
  }

  for (i=0; i<NSIP; i++) 
  {
    delete[] coeff_norm_rdoq_qp[i];
    delete[] coeff_norm_rdoq[i];
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Set quantizer
*
* \param[in] qp Array of quantization parameter values (one for each coefficient)
*/
//-----------------------------------------------------------------------------
void Quantizer::set_qp(int *qp) 
{
  int i, k;

  for (k = ZERO; k < TX_NCOEFF; k++) 
  {
    for (i=0; i<NSIP; i++) 
    {
      iquant_scale[i][k] = inv_scale[i][TX_NCOEFF * clip(qp[k], ZERO, NQPS - ONE) + k];
      quant_scale[i][k] = fwd_scale[i][TX_NCOEFF * clip(qp[k], ZERO, NQPS - ONE) + k];
    }
    for (i=0; i<NSIP; i++) 
    {
      coeff_norm_rdoq[i][k] = 1.0 / coeff_norm_rdoq_qp[i][TX_NCOEFF * clip(qp[k], ZERO, NQPS - ONE) + k];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Rescan a block of double
*
* \param[in] val Pointer to Array to be rescan
* \param[in] type1 type of current scan
* \param[in] type2 type of scan to scan to
*/
//-----------------------------------------------------------------------------
void Quantizer::rescan_d(double *val, int type1, int type2)
{
  double tmp[TX_NCOEFF];
  int i;

  if (type1 == type2) return; // do nothing

  for (i=0; i<16; i++) 
  {
    tmp[i] = val[Block::scan[type1][i]];    // tem.val is sorted in increasing order now.
  }

  for (i=0; i<16; i++) 
  {
    val[Block::scan[type2][i]] = tmp[i];
  }

}

//-----------------------------------------------------------------------------
/*!
* \brief Set quantizer for 8x8
*
* \param[in] qp Array of quantization parameter values (one for each coefficient)
*/
//-----------------------------------------------------------------------------
void Quantizer::set_qp8x8(int *qp) 
{
  int k;

  for (k = ZERO; k < 64; k++) 
  {
    iquant_scale8x8[k] = inv_scale8x8[64 * clip(qp[k], ZERO, NQPS - ONE) + k];
    quant_scale8x8[k] = fwd_scale8x8[64 * clip(qp[k], ZERO, NQPS - ONE) + k];

    coeff_norm_rdoq8x8[k] = coeff_norm_rdoq_qp8x8[ZTsize64 * clip(qp[k], ZERO, NQPS - ONE) + k];
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Quantize 4x4 block with rdoq
*
* Quantizes all coefficients in a block and adapts the offsets used in quantization.
*
* \param[in,out] blk Block to quantize
* \param[in] eenc EntropyEncoder
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] lambda_mode 
* \param[in] context 
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
static inline int get_sign(int x) { return x >> 31; }
static inline int mul_sign(int x, int s) { return (x ^ s) - s; }

void Quantizer::quant_rdoq(Block *blk, EntropyEncoder *eenc, int type, double lambda_mode, int context, ZeroTree *ztree, int sip_type, int parity) 
{
  if (blk->get_num_elements() == 16)
  {
    quant_rdoq4x4(blk, eenc, type, lambda_mode, context, ztree, sip_type, parity);
  }
  else
  {
    quant_rdoq8x8(blk, eenc, type, lambda_mode, context, ztree, parity);
  }
}

void Quantizer::quant_rdoq4x4(Block *blk, EntropyEncoder *eenc, int type, double lambda_mode, int context, ZeroTree *ztree, int sip_type, int parity) 
{
  int k, ztk;
  int i;
  int sgn; // sign
  int v;   // value of coefficient
  int vq;  // quantized value
  int zti;
  
  double cij;
  double dfq;
  int Lfloor, Lceil;
  double cost_floor, cost_ceil, cost_zero;
  double bit_1;
  Block check_one(4, 4);
  
  int min_one=ZERO;
  int max_one=ZERO;
  double jsum=0.0;
  int second_loop_end=TX_NCOEFF;
  double min_cost=FLT_INF;
  
  int total_val=0;
  
  int total_ar_val=0;
  
  Block ar_blk(4, 4);
  double ar_cost=0;
  double ar_distortion=0;
  double rdoq_distortion=0;
  double rdoq_cost=0;
  int ar_count=0;
  int rdoq_count=0;
  double c_ar, c_rd;
  
  // Hidden Flag
  int qblk_sum = 0;
  int index_max = -1;
  int vq_max = 0;
  int vqadd = 0;
  int mod_off_max = 0;
  int err_max = -1;
  int second_coeff = 0, second_coeff_max = 0, second_sgn_max = 0, second_index_max = -1;
  int second_qblk_parity;
  unsigned int second_qblk_sum = 0;  
  double second_err_max = -1, second_v_err_d;
  if (!param.transform_selection)
    parity = -1;
  
#ifdef TS_RDOQ_4X4_ENHANCE
  int mod_coeff = -1;
  int ori_coeff, new_coeff = 0;
  double rdoq_cost_2 = 0;
  double rdoq_distortion_2 = 0;
  int rdoq_count_2 = 0;
  int total_val_2 = 0;
  double c_rd_2;
  int save_blk_val;
  int save_ztk = 0;
#endif //TS_RDOQ_4X4_ENHANCE
  
  Block abs_scaled_blk(4, 4);
  const double *norm = coeff_norm_rdoq[sip_type];
  
  for (k = 0; k < TX_NCOEFF; k++) 
  {
    int v = blk->val[k];    
    int sgn = get_sign(v);
    v = mul_sign(v, sgn);
    v = v * quant_scale[sip_type][k]*16/active_qmatrix_4x4[0][k]; // cq_start, cq_end
    abs_scaled_blk.val[k] = v;
    
    int offset_index = (intra && !intra_frame) ? 3 : 1;
    if (!intra && !low_zeromv)
      offset_index = (v >= 0x10000);
    
    int vq = (v + offset[offset_index][k]) >> 16;
    int delta = v - (vq << 16);
    
    int mod_off = 0;
    
    if (vq != 0 && (low_zeromv || v >= 0x18000))
    {
      mod_off = shift_right_round(-delta, 10); // for compatibility with previous code
      offset[offset_index][k] += shift_right_round(delta, 10); // add 1/1024 of difference
      offset[offset_index][k] = clip(offset[offset_index][k], 0, 32768);
    }
    
    ar_blk.val[k] = mul_sign(vq, sgn);
    
    if (abs(delta) >= err_max)
    {
      index_max = k;
      err_max = abs(delta);
      vqadd = (delta >> 31) | 1; // -1 if delta is negative, +1 otherwise
      vq_max = vq + vqadd;
      vq_max = mul_sign(vq_max, sgn);
      mod_off_max = mod_off;
    }
    qblk_sum += vq; // Using absolute value such that qblk_sum can be used to check parity AND whether all coeffs are 0
  }
  
  if (parity != -1 && qblk_sum > 0 && (qblk_sum & 1) != parity)
  {
    int k = index_max;
    ar_blk.val[k] = vq_max;
    
    // Is the stuff below really needed?
    int v = abs_scaled_blk.val[k];
    int offset_index = (intra && !intra_frame) ? 3 : 1;      
    offset[offset_index][k] += mod_off_max;
    
    int vq = abs(vq_max);
    if (vq != 0 && (low_zeromv || v >= 0x18000))
    {
      offset[offset_index][k] = update_offset(offset[offset_index][k], v, vq, 32768);
    }
  }
  
  // grouping
  max_one = -1;
  for (k = ZERO; k < TX_NCOEFF; k++) 
  {    
    ztk = ztree->leaf_index[k];
    
    v = abs_scaled_blk.val[ztk];
    int vs = v >> 15;
    int large = (vs > 2);
    int vs2 = (large) ? 2 : vs;
    
    if (vs == 0) // v/65536 < 0.5
      blk->val[ztk] = 0;
    
    if (vs != 0)
      max_one = k + large;
    
    if (vs2 == 2)
      min_one = max_one;
    
    check_one.val[ztk] = vs2;
  }
  
  if (max_one < 0)
    return;
  
  if(intra)
  {
    ar_blk.rescan(sip_type, 0);
    abs_scaled_blk.rescan(sip_type, 0);
    blk->rescan(sip_type, 0);
    check_one.rescan(sip_type, 0);
  }
  
  // first loop
  if (min_one < TX_NCOEFF)
  {
    Block check_two(4, 4);
    check_two.set(0);
    
    for (k = 0; k < min_one; k++)
    {
      ztk = ztree->leaf_index[k];
      check_two.val[ztk] = check_one.val[ztk];
    }
    
    max_one = imin(max_one, TX_NCOEFF - 1);
    
    jsum = 0.0;
    second_loop_end = min_one;    
    min_cost = lambda_mode * eenc->cost_ztree_rdo(&check_two, type, context, ztree);
    
    for (k = min_one; k <= max_one; k++)
    {      
      ztk = ztree->leaf_index[k];
      if (check_one.val[ztk])
      {        
        check_two.val[ztk] = check_one.val[ztk];
        v = abs_scaled_blk.val[ztk];
        jsum += (65536.0 * 65536.0 - 2 * 65536.0 * v) * norm[ztk]; // (65536 - v)^2 - v^2
        jsum += lambda_mode * eenc->cost_one_coeff(1, type, TX_NCOEFF-1);
        bit_1 = lambda_mode * eenc->cost_ztree_rdo(&check_two, type, context, ztree);
        
        if (jsum + bit_1 < min_cost)
        {
          second_loop_end = k + 1;
          min_cost = jsum + bit_1;
        }
        else if (!low_zeromv && !intra)
        {
          check_two.val[ztk] = 0;
        }
      }      
    }
    
    // second_loop
    for (k = second_loop_end; k < TX_NCOEFF; k++) 
    {
      ztk = ztree->leaf_index[k];
      blk->val[ztk] = 0;
      check_one.val[ztk] = 0;
    }
  }
  
  for (k = TX_NCOEFF - 1, i = eenc->init_coeff_ctx_4x4(); k >=0 ; k--) 
  {    
    ztk = (intra) ? Block::ZigZag_Table_4x4[k] : k;
    zti = (intra) ? Block::ZigZag_Table_4x4[i] : i;
    
    v = abs_scaled_blk.val[ztk];    
    cij = (double)v ;
    
    sgn = (blk->val[ztk] < 0) ? -1 : 1;
    
    if (check_one.val[ztk] == 0)
    {      
      rdoq_distortion += cij * cij * norm[ztk];
      
      second_v_err_d = -1; //cij;
      second_coeff = 1;
      second_v_err_d = second_v_err_d * second_v_err_d * norm[ztk];
      if (second_v_err_d >= second_err_max)
      {
        second_index_max = ztk;
        second_err_max  = second_v_err_d;
        second_sgn_max = sgn;
        second_coeff_max = second_coeff;
      }
      
      continue;
    }
    else if (check_one.val[ztk] == 1)
    {
      double delta_cost = (65536.0 * 65536.0 - 2 * 65536.0 * cij) * norm[ztk];
      
      delta_cost += lambda_mode * (eenc->cost_ztree_rdo(&check_one, type, context, ztree) + eenc->cost_one_coeff(1, type, zti));
      check_one.val[ztk] = 0;
      delta_cost -= lambda_mode * eenc->cost_ztree_rdo(&check_one, type, context, ztree);
      
      vq = delta_cost < 0.0;
      check_one.val[ztk] = vq;
      
      second_coeff = !vq;
      second_v_err_d = (vq) ? (65536.0 - cij) : cij;
#ifdef TS_RDOQ_4X4_ENHANCE
      if (!vq)
      {
        mod_coeff = k;
        new_coeff = sgn;
      }
#endif //TS_RDOQ_4X4_ENHANCE
    }
    else
    {
      Lfloor = v >> 16;
      Lceil = Lfloor + 1;
      
      if ((v & 0x8000) == 0)
      { 
        vq = Lfloor;
        second_v_err_d = (double)(v & 0xffff);
        second_coeff = Lceil;
      }
      else
      {        
        double delta = (double)(v & 0xffff); // (cij - 65536 * Lfloor)
        cost_floor = delta * delta * norm[ztk] + lambda_mode*(eenc->cost_one_coeff(Lfloor, type, zti));
        double delta2 = 65536.0 - delta; // (65536 * Lceil - cij)
        cost_ceil = delta2 * delta2 * norm[ztk] + lambda_mode*(eenc->cost_one_coeff(Lceil, type, zti));
        cost_zero = cij * cij * norm[ztk];
        
        if (cost_ceil < cost_floor && cost_ceil < cost_zero)
        {
          vq = Lceil;
          second_v_err_d = delta2;
          second_coeff = Lfloor;
        }
        else if (cost_floor < cost_zero)
        {
          vq = Lfloor;
          second_v_err_d = delta;
          second_coeff = Lceil;
        }
        else
        {
          vq = 0;
          check_one.val[ztk] = 0;
          second_v_err_d = cij;
          second_coeff = 1;
        }
      }      
    }        
    
    dfq = (double)(vq<<16) ;
    rdoq_distortion += (cij-dfq)*(cij-dfq)*norm[ztk];
    rdoq_cost += eenc->cost_one_coeff(vq, type, zti);
    
    total_val+=vq;
    
    if(vq) rdoq_count++;
    
    if(second_v_err_d > 0 )
      second_v_err_d = second_v_err_d*second_v_err_d*norm[ztk];
    if(second_v_err_d >= second_err_max)
    {
      second_index_max = ztk;
      second_err_max  = second_v_err_d;
      second_sgn_max = sgn;
      second_coeff_max = second_coeff;
    }
    second_qblk_sum += vq;
    
    i = eenc->update_coeff_ctx_4x4(i, vq, intra);
    
    blk->val[ztk] = vq * sgn;    
  }
  
  if (param.transform_selection && parity != -1 )
  {
    // get the parity of the sum of coefficients
    second_qblk_parity = (second_qblk_sum & 01);
    
    if (second_qblk_parity != parity && second_qblk_sum>0 )
    {
#ifdef TS_RDOQ_4X4_ENHANCE
      if(mod_coeff>=0)
      {
        ori_coeff = 0;
        save_ztk = (intra) ? Block::ZigZag_Table_4x4[mod_coeff] : mod_coeff;
        check_one.val[save_ztk] = 1;
        blk->val[save_ztk] = new_coeff;
        rdoq_cost_2 = eenc->cost_ztree_rdo(&check_one, type, context, ztree);
        
        for (k = TX_NCOEFF-1, i = eenc->init_coeff_ctx_4x4(); k >=0 ; k--) 
        {
          ztk = (intra) ? Block::ZigZag_Table_4x4[k] : k;
          zti = (intra) ? Block::ZigZag_Table_4x4[i] : i;
          
          v = abs_scaled_blk.val[ztk];
          vq = abs(blk->val[ztk]);
          double delta = (double)(v - (vq << 16));
          rdoq_distortion_2 += delta * delta * norm[ztk];
          i = eenc->update_coeff_ctx_4x4(i, vq, intra);
          total_val_2 += vq;          
          rdoq_count_2 += (vq != 0);          
          rdoq_cost_2 += eenc->cost_one_coeff(vq, type, zti);
        }
        check_one.val[save_ztk] = abs(ori_coeff);
        blk->val[save_ztk] = ori_coeff;
      }
      save_blk_val = blk->val[second_index_max];
#endif //TS_RDOQ_4X4_ENHANCE
      
      blk->val[second_index_max] = second_coeff_max * second_sgn_max;
      check_one.val[second_index_max] = second_coeff_max != 0;
      
      rdoq_cost = eenc->cost_ztree_rdo(&check_one, type, context, ztree);
      rdoq_distortion = 0;
      rdoq_count = 0;
      total_val=0;
      for (k = TX_NCOEFF - 1, i = eenc->init_coeff_ctx_4x4(); k >=0 ; k--) 
      {    
        ztk = (intra) ? Block::ZigZag_Table_4x4[k] : k;
        zti = (intra) ? Block::ZigZag_Table_4x4[i] : i;
        
        v = abs_scaled_blk.val[ztk];
        vq = abs(blk->val[ztk]);
        double delta = (double)(v - (vq << 16));
        rdoq_distortion += delta * delta * norm[ztk];
        i = eenc->update_coeff_ctx_4x4(i, vq, intra);
        total_val += vq;          
        rdoq_count += (vq != 0);          
        rdoq_cost += eenc->cost_one_coeff(vq, type, zti);
      }
#ifdef TS_RDOQ_4X4_ENHANCE
      if ( mod_coeff>=0 ) 
      {
        c_rd = rdoq_cost*lambda_mode+rdoq_distortion;
        c_rd_2 = rdoq_cost_2*lambda_mode+rdoq_distortion_2;
        if ( c_rd > c_rd_2 )    
        {
          blk->val[second_index_max] = save_blk_val;
          blk->val[save_ztk] = new_coeff;
          rdoq_distortion = rdoq_distortion_2;
          rdoq_cost = rdoq_cost_2;
          total_val = total_val_2;    
          rdoq_count = rdoq_count_2;
        }
      }
#endif // TS_RDOQ_4X4_ENHANCE
    }
    else
    {
      rdoq_cost += eenc->cost_ztree_rdo(&check_one, type, context, ztree);
    }
  }
  else
  {
    rdoq_cost += eenc->cost_ztree_rdo(&check_one, type, context, ztree);
  }
  
  ar_cost = eenc->cost_ztree_rdo(&ar_blk, type, context, ztree); // Note: calls ZeroTree::update_values
  for (k = TX_NCOEFF - 1, i = eenc->init_coeff_ctx_4x4(); k >= 0 ; k--) 
  {    
    int ztk = (intra) ? Block::ZigZag_Table_4x4[k] : k;    
    int abs_v = abs_scaled_blk.val[ztk];
    int abs_vq = abs(ar_blk.val[ztk]);    
    double delta = (double)(abs_v - (abs_vq << 16));
    
    total_ar_val += abs_vq;
    ar_count += abs_vq != 0;  
    int zti = (intra) ? Block::ZigZag_Table_4x4[i] : i;
    i = eenc->update_coeff_ctx_4x4(i, abs_vq, intra);
    
    ar_distortion += delta * delta * norm[ztk];
    ar_cost += eenc->cost_one_coeff(abs_vq, type, zti);
  }
  
  c_ar = ar_cost*lambda_mode+ar_distortion;
  c_rd = rdoq_cost*lambda_mode+rdoq_distortion; 
  
  int revert = FALSE;
  if (c_ar < c_rd && ((!low_zeromv && !intra) || (ar_distortion < rdoq_distortion)))
  {    
    revert = TRUE;
  }
  else if (reference_idc == REFERENCE_PIC)
  {
    if ((low_zeromv || intra) && total_val == 0 ||
        frame_type == ft_P_frame && low_zeromv && (ar_distortion<rdoq_distortion) && (ar_blk.val[0]==blk->val[0]) && (rdoq_count < ar_count) && (intra || ar_blk.val[0] == 0))
    {
      revert = ar_distortion * (ar_cost * lambda_mode + rdoq_distortion) < c_rd * rdoq_distortion;
    }
  }
  
  if (revert)
  {
    for (k = 0; k < TX_NCOEFF; k++)
      blk->val[k] = ar_blk.val[k];
  }
  
  if (intra)
  {
    blk->rescan(0, sip_type);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Quantize 4x4 block
*
* Quantizes all coefficients in a block and adapts the offsets used in quantization.
*
* \param[in,out] blk Block to quantize
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
void Quantizer::quant(Block *blk, int sip_type, int parity)
{
  if (blk->get_num_elements() == 64)
    quant8x8(blk, 0, parity);
  else
    quant4x4(blk, sip_type, parity);
}

void Quantizer::quant4x4(Block *blk, int sip_type, int parity) 
{
  int k;
  int sgn; // sign
  int v;   // value of coefficient
  int vq;  // quantized value
  int offset_limit = get_offset_limit();

  int qblk_sum = 0;
  int vqadd = 0;
  int v_max = 0;
  int vq_max = 0, sgn_max = 0, index_max = -1;
  int err_max = -1;
  int mod_off_max = 0;

  int offsetValue; // cq_start, cq_end

  for (k=0; k<16; k++) 
  {
    v = blk->val[k];    
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * quant_scale[sip_type][k]*16/active_qmatrix_4x4[0][k]; // cq_start

    int offset_index = get_offset_index(intra, intra_frame, v, low_zeromv);

    vq = (v + offset[offset_index][k]) >> 16;
    offsetValue = offset[offset_index][k]; // cq_start
    bool update_offset_flag = get_update_offset(v, vq, low_zeromv);
    if (update_offset_flag)
    {
      offset[offset_index][k] = update_offset(offset[offset_index][k], v, vq, offset_limit);
    }

    if (param.qmatrix_enable_flag[1]) // cq_start
      vq = max(0, (v + (offsetValue*(128-active_qmatrix_4x4[1][k])/128) + ((1<<16)*active_qmatrix_4x4[1][k]/128))) >> 16; // cq_end

    if (param.transform_selection)
    {    
      int vqshift = vq << 16;
      int v_err_d = abs(v - vqshift);

      if (v_err_d >= err_max)
      {
        index_max = k;
        err_max  = v_err_d;
        vq_max  = vq;
        sgn_max = sgn;
        v_max = v;
        mod_off_max = update_offset_flag;
        vqadd = (v >= vqshift) ? 1 : -1;
      }
      qblk_sum += vq;
    }
    blk->val[k] = vq * sgn;
  }

  if (param.transform_selection && parity != -1 && (qblk_sum & 1) != parity && qblk_sum > 0)
  {
    v = v_max;
    vq = vq_max+vqadd;
    k = index_max;
    blk->val[k] = vq * sgn_max;

    int offset_index = get_offset_index(intra, intra_frame, v, low_zeromv);
    
    if(mod_off_max)
    {
      offset[offset_index][k] += shift_right_round( - v + (vq_max<<16), 10); // add 1/1024 of difference          
    }
    
    bool update_offset_flag = get_update_offset(v, vq, low_zeromv);
    if (update_offset_flag)
    {
      offset[offset_index][k] = update_offset(offset[offset_index][k], v, vq, offset_limit);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Quantize 4x4 block without actually affecting the offsets
*
* Quantizes all coefficients in a block.
*
* \param[in,out] blk Block to quantize
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
void Quantizer::quant_loop(Block *blk, int sip_type, int parity)
{
  if (blk->get_num_elements() == 16)
    quant_loop4x4(blk, sip_type, parity);
  else
    quant8x8(blk, 0, parity);
}

void Quantizer::quant_loop4x4(Block *blk, int sip_type, int parity)
{
  int k;
  int sgn; // sign
  int v;   // value of coefficient
  int vq;  // quantized value

  int qblk_sum = 0;
  int vqadd = 0;
  int vq_max = 0, sgn_max = 0, index_max = -1;
  int err_max = -1;

  int offsetValue; // cq_start, cq_end

  for (k = 0; k < 16; k++) 
  {
    v = blk->val[k];    
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * quant_scale[sip_type][k]*16/active_qmatrix_4x4[0][k]; // cq_start

    int offset_index = get_offset_index(intra, intra_frame, v, low_zeromv);

    vq = (v + offset[offset_index][k]) >> 16;

    offsetValue = offset[offset_index][k] ; // cq_start
    if (param.qmatrix_enable_flag[1])
      vq = max(0, (v + (offsetValue*(128-active_qmatrix_4x4[1][k])/128) + ((1<<16)*active_qmatrix_4x4[1][k]/128))) >> 16; // cq_end

    if (param.transform_selection)
    {   
      int vqshift = vq << 16;
      int v_err_d = abs(v - vqshift);

      if (v_err_d >= err_max )
      {
        index_max = k;
        err_max  = v_err_d;
        vq_max  = vq;
        sgn_max = sgn;
        vqadd = (v >= vqshift) ? 1 : -1;
      }
      qblk_sum += vq;
    }

    blk->val[k] = vq * sgn;
  }

  if (param.transform_selection && parity != -1 && (qblk_sum & 1) != parity && qblk_sum > 0)
  {
     blk->val[index_max] = (vq_max + vqadd) * sgn_max;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Inverse quantize 4x4 block
*
* \param[in,out] blk Block to inverse quantize
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
void Quantizer::iquant(Block *blk, int sip_type) 
{
  if (blk->get_width() == 8 && blk->get_height() == 8)
    iquant8x8(blk);
  else
    iquant4x4(blk, sip_type);
}

void Quantizer::iquant4x4(Block *blk, int sip_type) 
{
  int k;  

  for (k = ZERO; k < TX_NCOEFF; k++)
  {
#ifdef Q_OFFSET // cq_start
    if (blk->val[k])
      blk->val[k] = ((((abs(blk->val[k]) << 7) -  active_qmatrix_4x4[1][k]) * iquant_scale[sip_type][k] * active_qmatrix_4x4[0][k]+ 1024) >> 11) * sign(blk->val[k]);
    else
      blk->val[k] = (blk->val[k] * iquant_scale[sip_type][k] * active_qmatrix_4x4[0][k]+8)>>4;
#else
    blk->val[k] = (blk->val[k] * iquant_scale[sip_type][k] * active_qmatrix_4x4[0][k]+8)>>4;
#endif // cq_end
  }
}



#ifdef Q_OFFSET
//-----------------------------------------------------------------------------
/*!
* \brief Set offset values
*/
//-----------------------------------------------------------------------------
void Quantizer::setOffset(int iTheta[2], int iTheta8x8[2]) 
{
  this->iTheta[0] = iTheta[0];
  this->iTheta[1] = iTheta[1];
  this->iTheta8x8[0] = iTheta8x8[0]; 
  this->iTheta8x8[1] = iTheta8x8[1];
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Load q matrix to active q matrixe
*/
//-----------------------------------------------------------------------------
void Quantizer::loadmatrix(int emodel, int CbCrFlag, int sip_type, int matrix_type) // cq_start, cq_end
{
  switch (emodel)
  {
  case 0: //!< Intra_Luma
    switch (sip_type)
    {
    case 0:
      active_qmatrix_4x4[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][0][4]);
      break;
    case 1:
      active_qmatrix_4x4[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][7][4]);
      break;
    case 2:
      active_qmatrix_4x4[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][6][4]);
      break;
    default:
      printf("Quantizer::loadmatrix(), wrong sip type\n");
      break;
    }
    active_qmatrix_8x8[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][8][4]);
    active_qmatrix_16x16[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][14][4]);
    break;
  case 1: //!< Intra_Chroma
    if (CbCrFlag==0)  //!< Cb
    {
      active_qmatrix_4x4[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][1][4]);
      active_qmatrix_8x8[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][10][4]);
    }
    else //!< Cr
    {
      active_qmatrix_4x4[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][2][4]);
      active_qmatrix_8x8[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][11][4]);
    }
    break;
  case 2:  //!< Inter_Luma
    active_qmatrix_4x4[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][3][4]);
    active_qmatrix_8x8[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][9][4]);
    active_qmatrix_16x16[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][15][4]);
    break;
  case 3:  //!< Inter_Chroma
    if (CbCrFlag==0)  //!< Cb
    {
      active_qmatrix_4x4[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][4][4]);
      active_qmatrix_8x8[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][12][4]);
    }
    else //!< Cr
    {
      active_qmatrix_4x4[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][5][4]);
      active_qmatrix_8x8[matrix_type] = &(param.qmatrix[matrix_type][param.quantization_set_id][13][4]);
    }
    break;
  default:
    printf("Quantizer::loadmatrix(), wrong emodel\n");
    break;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Decide which category of the rounding offsets to use
*
* \param[in] intra       Whether this is an intra block
* \param[in] intra_frame Whether this is an inter frame
* \param[in] v           The coefficient before rounding
* \param[in] low_zeromv  Whether the MV percentage is low
* \return Category
*/
//-----------------------------------------------------------------------------
int Quantizer::get_offset_index(int intra, int intra_frame, int v, int low_zeromv)
{
  int offset_index;

  switch (param.q_adaptoffset)
  {
  case 2:
    if (intra)
      offset_index = (intra_frame) ? 1 : 3;
    else
      offset_index = imin(v >> 16, 1);
    break;

  case 4:
    if (intra)
      offset_index = (intra_frame) ? 1 : 3;
    else
      offset_index = (low_zeromv) ? 1 : imin(v >> 16, 1);
    break;

  default:
    offset_index = 0;
    break;
  }

  return offset_index;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get upper bound for offset
*
* \return Upper bound
*/
//-----------------------------------------------------------------------------
int Quantizer::get_offset_limit()
{
  int offset_limit;

  switch (param.q_adaptoffset)
  {
  case 1:
  case 2:
  case 3:
    offset_limit = 65535;
    break;

  default:
    offset_limit = 32768;
    break;
  }

  return offset_limit;
}

//-----------------------------------------------------------------------------
/*!
* \brief Determine whether to update offset
*
* \param[in] v           Scaled coefficient (16.16 fixed-point)
* \param[in] vq          Quantized coefficient
* \param[in] low_zeromv  Whether the MV percentage is low
* \return Flag indicating update
*/
//-----------------------------------------------------------------------------
bool Quantizer::get_update_offset(int v, int vq, int low_zeromv)
{
  bool update_offset;

  switch (param.q_adaptoffset)
  {
  case 1:
    update_offset = (v >= 0x8000);
    break;

  case 2:
    update_offset = (v >= 0x18000);
    break;

  case 3:
    update_offset = (vq >= 1);
    break;

  case 4:
    update_offset = ((low_zeromv && vq >= 1) || (!low_zeromv && v >= 0x18000));
    break;

  default:
    update_offset = false;
    break;
  }

  return update_offset;
}

//-----------------------------------------------------------------------------
/*!
* \brief Quantize 8x8 block
*
* Quantizes all coefficients in a block and adapts the offsets used in quantization.
*
* \param[in,out] blk Block to quantize
* \param[in] firstcoeff Postion of first coefficient to quantize
*/
//-----------------------------------------------------------------------------
void Quantizer::quant8x8(Block *blk8, int firstcoeff, int parity) 
{
  int k;
  int offset_limit = get_offset_limit();

  int qblk_sum = 0;
  int vqadd = 0;
  int v_max = 0;
  int vq_max = 0, sgn_max = 0, index_max = -1;
  int err_max = -1;
  bool mod_off_max = false;

  int offsetValue; // cq_start, cq_end

  for (k = firstcoeff; k < 64; k++) 
  {
    int v = blk8->val[k];    
    int sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * quant_scale8x8[k]*16/active_qmatrix_8x8[0][k]; // cq_start, cq_end

    int offset_index = get_offset_index(intra, intra_frame, v, low_zeromv);

    int vq = (v + offset8x8[offset_index][k]) >> 16;

    offsetValue = offset8x8[offset_index][k]; // cq_start, cq_end

    bool update_offset_flag = get_update_offset(v, vq, low_zeromv);
    if (update_offset_flag)
    {
        offset8x8[offset_index][k] = update_offset(offset8x8[offset_index][k], v, vq, offset_limit);
    }
    if(param.qmatrix_enable_flag[1]) // cq_start
    {
      vq = max(0, (v + (offsetValue*(128-active_qmatrix_8x8[1][k])/128) + ((1<<16)*active_qmatrix_8x8[1][k]/128))) >> 16;
      if(vq>0)
        vq = max((v + offsetValue + ((1<<16)*active_qmatrix_8x8[1][k]/128)), 0) >> 16;
    } // cq_end

    if (param.transform_selection)
    {
      int vqshift = vq << 16;
      int v_err_d = abs(v - vqshift);

      if (v_err_d >= err_max )
      {
        index_max = k;
        err_max  = v_err_d;
        vq_max  = vq;
        v_max = v;
        sgn_max = sgn;
        mod_off_max = update_offset_flag;
        vqadd = (v >= vqshift) ? 1 : -1;
      }
      qblk_sum += vq;
    }

    blk8->val[k] = vq * sgn;
  }

  if (param.transform_selection && parity != -1 && (qblk_sum & 1) != parity && qblk_sum > 0)
  {
    int v = v_max;
    int vq = vq_max + vqadd;
    k = index_max;
    blk8->val[k] = vq * sgn_max;

    int offset_index = get_offset_index(intra, intra_frame, v, low_zeromv);

    if (mod_off_max)
    {
      offset8x8[offset_index][k] += shift_right_round(-v + (vq_max << 16), 10);
    }

    bool update_offset_flag = get_update_offset(v, vq, low_zeromv);
    if (update_offset_flag)
    {
      offset8x8[offset_index][k] = update_offset(offset8x8[offset_index][k], v, vq, offset_limit);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Quantize 8x8 block with rdoq
*
* Quantizes all coefficients in a block and adapts the offsets used in quantization.
*
* \param[in,out] blk Block to quantize
* \param[in] eenc EntropyEncoder
* \param[in] type Entropy model index (EntropyModel::I_luma, etc.)
* \param[in] lambda_mode 
* \param[in] context 
* \param[in] ztree Zerotree used to decode positions of nonzero coefficients
*/
//-----------------------------------------------------------------------------
void Quantizer::quant_rdoq8x8(Block *blk, EntropyEncoder *eenc, int type, double lambda_mode, int context, ZeroTree *ztree, int parity) 
{
  int k, ztk;
  int i;
  int sgn; // sign
  int v;   // value of coefficient
  int vq;  // quantized value
  int zti;

  double cij;
  double fq;
  double dfq;
  int Lfloor, Lceil;
  double cost_floor, cost_ceil, cost_zero;
  double bit_0, bit_1;
  Block check_one(8, 8);
  Block check_two(8, 8);
  Block keep_blk(8, 8);
  int count0=0;
  int count1=0;
  int count2=0;
  double stored_cost_cij[64];

  int min_one=ZERO;
  int max_one=ZERO;
  double jsum=0.0;
  int second_loop_end=ZTsize64;
  double min_cost=FLT_INF;

  int total_val=0;
  int total_ar_val=0;
  int offset_index = 0;

  double ar_cost=0;
  double ar_distortion=0;
  double rdoq_distortion=0;
  double rdoq_cost=0;
  int ar_count=0;
  int rdoq_count=0;
  double c_ar, c_rd;
  Block ar_blk(8, 8);
  const int *ptr2;

  // Hidden Flag
  int qblk_parity, qblk_sum = 0;  
  int vq_max = 0, sgn_max = 0, index_max = -1;
  int vqadd = 0;
  double vqshift;
  double err_max = -FLT_INF, v_err_d;
  int second_coeff = 0, second_coeff_max = 0, second_sgn_max = 0, second_index_max = -1;
  int second_qblk_parity;
  unsigned int second_qblk_sum = 0;  
  double second_err_max = -1, second_v_err_d;

#ifdef TS_RDOQ_8X8_ENHANCE
  int mod_coeff = -1;
  int ori_coeff, new_coeff = 0;
  double rdoq_cost_2 = 0;
  double rdoq_distortion_2 = 0;
  int rdoq_count_2 = 0;
  int total_val_2 = 0;
  double c_rd_2;
  int save_blk_val;
  int save_ztk = 0;
  int mod_off, mod_off_max = 0, v_max = 0;
#endif //TS_RDOQ_8X8_ENHANCE

  ptr2 = Block::ZigZag_Table_8x8;

  for (k = ZERO; k < ZTsize64; k++)
    keep_blk.val[k]=blk->val[k];

  for (k=ZERO; k<ZTsize64; k++) 
  {
    v = blk->val[ptr2[k]];    
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * quant_scale8x8[ptr2[k]]*16/active_qmatrix_8x8[0][ptr2[k]]; // cq_start, cq_end

#ifdef TS_RDOQ_8X8_ENHANCE
    mod_off = 0;
#endif // TS_RDOQ_8X8_ENHANCE

    if(low_zeromv)
      offset_index=ONE; 
    else
      offset_index = imin(v >> 16, ONE);

    vq = (v + offset8x8[offset_index][ptr2[k]]) >> 16;

    if ((low_zeromv && vq >= 1) || (!low_zeromv && v >= 65536+32768)) 
    {
#ifdef TS_RDOQ_8X8_ENHANCE
      mod_off=1;
#endif // TS_RDOQ_8X8_ENHANCE

      offset8x8[1][ptr2[k]] = update_offset(offset8x8[1][ptr2[k]], v, vq, 32768);
    }

    if (param.transform_selection && parity != -1)
    {
      vqshift = (double) (vq<<16);
      v_err_d = fabs((double) v - vqshift);

      if(v_err_d >= err_max )
      {
        index_max = k;
        err_max  = v_err_d;
        vq_max  = vq;
        sgn_max = sgn;
        if( v>=vqshift )
          vqadd = 1;
        else
          vqadd = -1;
#ifdef TS_RDOQ_8X8_ENHANCE
        v_max = v;
        mod_off_max = mod_off;
#endif // TS_RDOQ_8X8_ENHANCE
      }
      qblk_sum += vq;
    }

    ar_blk.val[ptr2[k]] = vq * sgn;
  }

  if (param.transform_selection && parity != -1 )
  {
    // get the parity of the sum of coefficients
    qblk_parity = (qblk_sum & 01);

    if (qblk_parity != parity && qblk_sum>0)
    {
      vq = vq_max + vqadd;
      k = index_max;
      ar_blk.val[ptr2[k]] = vq * sgn_max;

#ifdef TS_RDOQ_8X8_ENHANCE
      v = v_max;

      if(mod_off_max)
        offset[ONE][ptr2[k]] += shift_right_round( -v + (vq_max<<16), 10); // add 1/1024 of difference

      if ((low_zeromv && vq >= 1) || (!low_zeromv && v >= 65536+32768)) 
      {
        offset8x8[1][ptr2[k]] = update_offset(offset8x8[1][ptr2[k]], v, vq, 32768);
      }
#endif // TS_RDOQ_8X8_ENHANCE
    }
  }

  ar_cost = eenc->cost_ztree_rdo8x8(&ar_blk, type, context, ztree);
  for (i= ZTsize64-1, k = i; k >=0 ; k--) 
  {
    ztk = ptr2[k];
    zti = ptr2[i];

    v = keep_blk.val[ztk];  
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * quant_scale8x8[ztk]*16/active_qmatrix_8x8[0][ztk]; // cq_start, cq_end

    cij = (double)v ;

    vq = ar_blk.val[ztk];   
    sgn = (vq < 0) ? -1 : 1;
    vq = vq * sgn;

    dfq = (double)(vq<<16) ;

    ar_distortion += (cij-dfq)*(cij-dfq)/coeff_norm_rdoq8x8[ztk];
    ar_cost += eenc->cost_one_coeff8x8(vq, zti);

    total_ar_val+=vq;    

    if(vq) ar_count++;

    if (abs(ar_blk.val[ztk]) > 0) i-= abs(ar_blk.val[ztk]);
    if (i<0) i=0;
  }

  // grouping

  for (k = ZERO; k < ZTsize64; k++) 
  { 
    ztk = ztree->leaf_index[ptr2[k]];

    v = blk->val[ztk];    
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * quant_scale8x8[ztk]*16/active_qmatrix_8x8[0][ztk]; // cq_start, cq_end

    fq = (double)v/ 65536; // vq = (v + offset[ZERO][k]) >> 16;

    if(fq < 0.5)
    {
      check_one.val[ztk]=0; count0++;
      blk->val[ztk] = 0;
    }
    else if(fq < 1.0)
    {
      check_one.val[ztk]=1; count1++;
      max_one = k;
    }
    else
    {
      check_one.val[ztk]=2; count2++;
      if(fq < 1.5)
      {
        min_one = k;
        max_one = k;
      }
      else
      {
        min_one = k+1;
        max_one = k+1;
      }
    }    
  }

  if(count0==ZTsize64) return;  

  // first loop
  if(min_one<ZTsize64)
  {    
    for(k=ZERO; k<ZTsize64; k++) check_two.val[k] = ZERO;
    for(k=ZERO; k<min_one; k++)
    {
      ztk = ztree->leaf_index[ptr2[k]];
      check_two.val[ztk]=check_one.val[ztk];
    }

    if(max_one>=ZTsize64) max_one=ZTsize64-1;

    for (k = min_one; k <= max_one; k++)
    {      
      ztk = ztree->leaf_index[ptr2[k]];

      if(check_one.val[ztk])
      {        
        v = blk->val[ztk];    
        sgn = (v < 0) ? -1 : 1;
        v = v * sgn;
        v = v * quant_scale8x8[ztk]*16/active_qmatrix_8x8[0][ztk]; // cq_start, cq_end

        cij = (double)v ;

        stored_cost_cij[ztk] = cij*cij/coeff_norm_rdoq8x8[ztk];
        jsum += stored_cost_cij[ztk];        
      }
    }

    second_loop_end=min_one;

    bit_0 = lambda_mode*eenc->cost_ztree_rdo8x8(&check_two, type, context, ztree);
    min_cost = jsum+bit_0;

    for (k = min_one; k <= max_one; k++)
    {      
      ztk = ztree->leaf_index[ptr2[k]];

      if(check_one.val[ztk])
      {        
        v = blk->val[ztk];    
        sgn = (v < 0) ? -1 : 1;
        v = v * sgn;
        v = v * quant_scale8x8[ztk]*16/active_qmatrix_8x8[0][ztk]; // cq_start, cq_end

        cij = (double)v ;

        dfq = (double)(1<<16) ;
        cost_ceil = (cij-dfq)*(cij-dfq)/coeff_norm_rdoq8x8[ztk]+lambda_mode*(eenc->cost_one_coeff8x8(1, ZTsize64-1));

        check_two.val[ztk]=check_one.val[ztk];
        bit_1 = lambda_mode*eenc->cost_ztree_rdo8x8(&check_two, type, context, ztree);

        if(jsum-stored_cost_cij[ztk]+cost_ceil+bit_1 < min_cost)
        {          
          jsum=jsum-stored_cost_cij[ztk]+cost_ceil;

          second_loop_end=k+1;

          min_cost = jsum+bit_1;          
        }
        else if(low_zeromv)
        {
          jsum=jsum-stored_cost_cij[ztk]+cost_ceil;
        }
        else
        {
          check_two.val[ztk]=0;
          jsum=jsum-stored_cost_cij[ztk]+cost_ceil;          
        }        
      }      
    }

    // second_loop
#ifdef TS_RDOQ_8X8_ENHANCE
    if(frame_type!=ft_I_frame && second_loop_end<ZTsize64 && (context || count2 || count1>2 || low_zeromv==0 || second_loop_end))
#else
    if(second_loop_end<ZTsize64 && (context || count2 || count1>2 || low_zeromv==0 || second_loop_end))
#endif //TS_RDOQ_8X8_ENHANCE    
    {
      for (k = second_loop_end; k<ZTsize64; k++) 
      {
        ztk = ztree->leaf_index[ptr2[k]];
        blk->val[ztk] = ZERO;
        check_one.val[ztk]=ZERO;
      }
    }

  }

  for (i= ZTsize64 - 1, k = i; k >=0 ; k--) 
  {
    ztk = ptr2[k];
    zti = ptr2[i];

    v = keep_blk.val[ztk];    
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * quant_scale8x8[ztk]*16/active_qmatrix_8x8[0][ztk]; // cq_start, cq_end

    cij = (double)v ;

    v = blk->val[ztk];    
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * quant_scale8x8[ztk]*16/active_qmatrix_8x8[0][ztk]; // cq_start, cq_end

    if(check_one.val[ztk]==0)
    {     
      vq = 0;      
      second_v_err_d = cij;
      second_coeff = 1;
    }
    else if(check_one.val[ztk]==1)
    {      
      cost_floor = cij*cij/coeff_norm_rdoq8x8[ztk];

      dfq = (double)(1<<16) ;

      cost_ceil = (cij-dfq)*(cij-dfq)/coeff_norm_rdoq8x8[ztk] + lambda_mode*(eenc->cost_one_coeff8x8(1, zti));

      check_one.val[ztk]=0;
      bit_0 = lambda_mode*eenc->cost_ztree_rdo8x8(&check_one, type, context, ztree);
      check_one.val[ztk]=1;
      bit_1 = lambda_mode*eenc->cost_ztree_rdo8x8(&check_one, type, context, ztree);

      if(cost_ceil+bit_1 < cost_floor+bit_0)
      {
        check_one.val[ztk]=1;
        vq = 1;
        second_v_err_d = dfq - cij;
        second_coeff = 0;
      }
      else
      {
        check_one.val[ztk]=0;
        vq = 0;
#ifdef TS_RDOQ_8X8_ENHANCE
        mod_coeff = k;
        new_coeff = sgn;
#endif //TS_RDOQ_8X8_ENHANCE
        second_v_err_d = cij;
        second_coeff = 1;
      }     
    }
    else
    {      
      fq = (double)v/ 65536; // vq = (v + offset[ZERO][k]) >> 16;

      Lfloor = (int)floor(fq);   // use the double version instead of the single precision floorf
      Lceil = Lfloor+1;

      if(fq-(double)Lfloor < (double)Lceil-fq)
      {        
        vq = Lfloor;        
        second_v_err_d = cij - (double)(Lfloor<<16);
        second_coeff = Lceil;
      }
      else
      {        
        dfq = (double)(Lfloor<<16) ;
        cost_floor = (cij-dfq)*(cij-dfq)/coeff_norm_rdoq8x8[ztk] + lambda_mode*(eenc->cost_one_coeff8x8(Lfloor, zti));

        dfq = (double)(Lceil<<16) ;
        cost_ceil = (cij-dfq)*(cij-dfq)/coeff_norm_rdoq8x8[ztk] + lambda_mode*(eenc->cost_one_coeff8x8(Lceil, zti));

        cost_zero = cij*cij/coeff_norm_rdoq8x8[ztk];

        if(cost_floor<cost_zero)
        {
          if(cost_ceil < cost_floor)
          {
            vq = Lceil;
            second_v_err_d = (double)(Lceil<<16) - cij;
            second_coeff = Lfloor;
          }
          else
          {
            vq = Lfloor;
            second_v_err_d = cij - (double)(Lfloor<<16);
            second_coeff = Lceil;
          }
        }
        else
        {
          if(cost_ceil < cost_zero)
          {
            vq = Lceil;            
            second_v_err_d = (double)(Lceil<<16) - cij;
            second_coeff = Lfloor;
          }
          else
          {
            vq=0;
            check_one.val[ztk]=0;
            second_v_err_d = cij;
            second_coeff = 1;
          }
        }        
      }      
    }     

    if(second_v_err_d > 0 )
      second_v_err_d = second_v_err_d*second_v_err_d/coeff_norm_rdoq8x8[ztk];
    if(second_v_err_d >= second_err_max)
    {
      second_index_max = ztk;
      second_err_max  = second_v_err_d;
      second_sgn_max = sgn;
      second_coeff_max = second_coeff;
    }
    second_qblk_sum += vq;

    dfq = (double)(vq<<16) ;
    rdoq_distortion += (cij-dfq)*(cij-dfq)/coeff_norm_rdoq8x8[ztk];
    rdoq_cost += eenc->cost_one_coeff8x8(vq, zti);

    total_val+=vq;

    if(vq) rdoq_count++;

    if (abs(vq) > 0) i-= abs(vq);
    if (i<0) i=0;

    blk->val[ztk] = vq * sgn;    
  }

  if (param.transform_selection && parity != -1 )
  {
    // get the parity of the sum of coefficients
    second_qblk_parity = (second_qblk_sum & 01);

    if (second_qblk_parity != parity && second_qblk_sum>0 )
    {
#ifdef TS_RDOQ_8X8_ENHANCE
      if(mod_coeff>=0)
      {
        ori_coeff = 0;        
        save_ztk = ptr2[mod_coeff];
        check_one.val[save_ztk]=1;        
        blk->val[save_ztk] = new_coeff;
        rdoq_cost_2 = eenc->cost_ztree_rdo8x8(&check_one, type, context, ztree);

        for (i= ZTsize64-1, k = i; k >=0 ; k--) 
        {
          ztk = ptr2[k];
          zti = ptr2[i];

          v = keep_blk.val[ztk];  
          sgn = (v < 0) ? -1 : 1;
          v = v * sgn;
          v = v * quant_scale8x8[ztk]*16/active_qmatrix_8x8[0][ztk]; // cq_start, cq_end

          cij = (double)v ;

          vq = blk->val[ztk];          
          sgn = (vq < 0) ? -1 : 1;
          vq = vq * sgn;          
          dfq = (double)(vq<<16) ;

          rdoq_distortion_2 += (cij-dfq)*(cij-dfq)/coeff_norm_rdoq8x8[ztk];
          rdoq_cost_2 += eenc->cost_one_coeff8x8(vq, zti);
          total_val_2+=vq;    

          if(vq) rdoq_count_2++;

          if (abs(blk->val[ztk]) > 0) i-= abs(blk->val[ztk]);
          if (i<0) i=0;
        }

        check_one.val[save_ztk] = abs(ori_coeff);
        blk->val[save_ztk] = ori_coeff;
      }
      save_blk_val = blk->val[second_index_max];
#endif //TS_RDOQ_8X8_ENHANCE

      blk->val[second_index_max] = second_coeff_max * second_sgn_max;
      if(second_coeff_max==0)
        check_one.val[second_index_max]=0;
      else
        check_one.val[second_index_max]=1;

      rdoq_cost = eenc->cost_ztree_rdo8x8(&check_one, type, context, ztree);
      rdoq_distortion = 0;
      rdoq_count = 0;
      total_val = 0;

      for (i= ZTsize64-1, k = i; k >=0 ; k--) 
      {
        ztk = ptr2[k];
        zti = ptr2[i];

        v = keep_blk.val[ztk];  
        sgn = (v < 0) ? -1 : 1;
        v = v * sgn;
        v = v * quant_scale8x8[ztk]*16/active_qmatrix_8x8[0][ztk]; // cq_start, cq_end

        cij = (double)v ;

        vq = blk->val[ztk];   
        sgn = (vq < 0) ? -1 : 1;
        vq = vq * sgn;
        dfq = (double)(vq<<16) ;

        rdoq_distortion += (cij-dfq)*(cij-dfq)/coeff_norm_rdoq8x8[ztk];
        rdoq_cost += eenc->cost_one_coeff8x8(vq, zti);
        total_val+=vq;    

        if(vq) rdoq_count++;

        if (abs(blk->val[ztk]) > 0) i-= abs(blk->val[ztk]);
        if (i<0) i=0;
      }
#ifdef TS_RDOQ_8X8_ENHANCE
      if ( mod_coeff>=0 ) 
      {
        c_rd = rdoq_cost*lambda_mode+rdoq_distortion;
        c_rd_2 = rdoq_cost_2*lambda_mode+rdoq_distortion_2;
        if ( c_rd > c_rd_2 )    
        {
          blk->val[second_index_max] = save_blk_val;
          blk->val[save_ztk] = new_coeff;
          rdoq_distortion = rdoq_distortion_2;
          rdoq_cost = rdoq_cost_2;
          total_val = total_val_2;    
          rdoq_count = rdoq_count_2;
        }
      }
#endif // TS_RDOQ_8X8_ENHANCE
    }
    else
      rdoq_cost += eenc->cost_ztree_rdo8x8(&check_one, type, context, ztree);
  }
  else
    rdoq_cost += eenc->cost_ztree_rdo8x8(&check_one, type, context, ztree);

  c_ar = ar_cost*lambda_mode+ar_distortion;
  c_rd = rdoq_cost*lambda_mode+rdoq_distortion;

#ifdef TS_RDOQ_8X8_ENHANCE
  if( c_ar < 1.05*c_rd )
  {    
    for (k = ZERO; k < ZTsize64; k++)
      blk->val[k]=ar_blk.val[k];
  }
#else // TS_RDOQ_8X8_ENHANCE
  if(c_ar < c_rd && (low_zeromv==0 || (ar_distortion<rdoq_distortion)))
  {    
    for (k = ZERO; k < ZTsize64; k++)
      blk->val[k]=ar_blk.val[k];
  }
  else if((low_zeromv) && (reference_idc==REFERENCE_PIC) && (total_val==0))
  {
    c_ar = ar_cost*lambda_mode*ar_distortion/rdoq_distortion+ar_distortion;
    if(c_ar < c_rd)
    {
      for (k = ZERO; k < ZTsize64; k++)
        blk->val[k]=ar_blk.val[k];
    }    
  }
  else if(reference_idc==REFERENCE_PIC && frame_type==ft_P_frame && low_zeromv && (ar_distortion<rdoq_distortion) && (ar_blk.val[0]==blk->val[0]) && (rdoq_count < ar_count))
  {
    if(ar_blk.val[0]==0)
    {      
      c_ar = ar_cost*lambda_mode*ar_distortion/rdoq_distortion+ar_distortion;
      if(c_ar < c_rd)
      {
        for (k = ZERO; k <ZTsize64; k++)
          blk->val[k]=ar_blk.val[k];
      }
    }    
  }
#endif // TS_RDOQ_8X8_ENHANCE
}

//-----------------------------------------------------------------------------
/*!
* \brief Inverse quantize 8x8 block
*
* \param[in,out] blk Block to inverse quantize
*/
//-----------------------------------------------------------------------------
void Quantizer::iquant8x8(Block *blk8) 
{
  int k;
#ifdef Q_OFFSET // cq_start
  int v, sgn, vq;    
#endif // cq_end
  for (k = ZERO; k < 64; k++) // cq_start
  {
#ifdef Q_OFFSET
    v = blk8->val[k] ;

    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;

    if (v>0)
    {
      blk8->val[k] = (v*128 - active_qmatrix_8x8[1][k]);
      if (!param.bugfix93)
      {
        blk8->val[k] = (short)blk8->val[k];
      }
    }

    vq = blk8->val[k] * iquant_scale8x8[k] * active_qmatrix_8x8[0][k];

    blk8->val[k] = sgn*((vq +1024)>>11);
#else
    blk8->val[k] = (blk8->val[k] * iquant_scale8x8[k] * active_qmatrix_8x8[0][k]+8)>>4;
#endif  
  }  
}

//-----------------------------------------------------------------------------
/*!
* \brief Load the value of the offset for intra
*
* \param[in] bank Storage for the offset to load from
*/
//-----------------------------------------------------------------------------
void Quantizer::I_load(int bank)
{
  int i,j;
  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      offset[j][i]=I_offset[bank][j][i];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Save the value of the offset for intra
*
* \param[out] bank Storage for the offset to save to
*/
//-----------------------------------------------------------------------------
void Quantizer::I_save(int bank) 
{
  int i,j;
  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      I_offset[bank][j][i]=offset[j][i];
    }
  }

}

//-----------------------------------------------------------------------------
/*!
* \brief Load the value of the offset for AIF
*/
//-----------------------------------------------------------------------------
void Quantizer::load()
{
  int i,j;

  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      offset[j][i]=AIF_offset[j][i];
    }
  }
  if (param.trans8x8_enable) 
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<param.copy_offset;i++)
      {
        offset8x8[j][i]=AIF_offset_8x8[j][i];
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Save the value of the offset for AIF
*/
//-----------------------------------------------------------------------------
void Quantizer::save() 
{
  int i,j;

  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      AIF_offset[j][i]=offset[j][i];
    }
  }
  if (param.trans8x8_enable) 
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<param.copy_offset;i++)
      {
        AIF_offset_8x8[j][i]=offset8x8[j][i];
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Save the value of the offset for AIF
*/
//-----------------------------------------------------------------------------
void Quantizer::P_save_1() 
{
  int i,j;

  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      P_AIF_offset_1[j][i]=offset[j][i];
    }
  }
  if (param.trans8x8_enable) 
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<param.copy_offset;i++)
      {
        P_AIF_offset_1_8x8[j][i]=offset8x8[j][i];
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Save the value of the offset for SAIF
*/
//-----------------------------------------------------------------------------
void Quantizer::P_save_2() 
{
  int i,j;

  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      P_AIF_offset_5[j][i]=offset[j][i];
    }
  }
  if (param.trans8x8_enable) 
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<param.copy_offset;i++)
      {
        P_AIF_offset_5_8x8[j][i]=offset8x8[j][i];
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Save the value of the offset for AIF
*/
//-----------------------------------------------------------------------------
void Quantizer::P_save_all(int k) 
{
  int i,j;

  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      if(k==1)
      {
        offset[j][i]=P_AIF_offset_1[j][i];
      }
      else if(k==2)
      {
        offset[j][i]=P_AIF_offset_5[j][i];
      }  
    }
  }
  if (param.trans8x8_enable) 
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<param.copy_offset;i++)
      {
        if(k==1)
        {
          offset8x8[j][i]=P_AIF_offset_1_8x8[j][i];
        }
        else if(k==2)
        {
          offset8x8[j][i]=P_AIF_offset_5_8x8[j][i];
        }  
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Save the value of the offset for AIF
*/
//-----------------------------------------------------------------------------
void Quantizer::B_save_1() 
{
  int i,j;

  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      B_AIF_offset_1[j][i]=offset[j][i];
    }
  }
  if (param.trans8x8_enable) 
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<param.copy_offset;i++)
      {
        B_AIF_offset_1_8x8[j][i]=offset8x8[j][i];
      }
    }
  }
}
//-----------------------------------------------------------------------------
/*!
* \brief Save the value of the offset for SAIF
*/
//-----------------------------------------------------------------------------
void Quantizer::B_save_2() 
{
  int i,j;

  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      B_AIF_offset_5[j][i]=offset[j][i];
    }
  }
  if (param.trans8x8_enable) 
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<param.copy_offset;i++)
      {
        B_AIF_offset_5_8x8[j][i]=offset8x8[j][i];
      }
    }
  }
}
//-----------------------------------------------------------------------------
/*!
* \brief Save the value of the offset for AIF
*/
//-----------------------------------------------------------------------------
void Quantizer::B_save_all(int k) 
{
  int i,j;

  for(j=0;j<4;j++)
  {
    for(i=0;i<16;i++)
    {
      if(k==1)
      {
        offset[j][i]=B_AIF_offset_1[j][i];
      }
      else if(k==2)
      {
        offset[j][i]=B_AIF_offset_5[j][i];
      }
    }
  }
  if (param.trans8x8_enable) 
  {
    for(j=0;j<4;j++)
    {
      for(i=0;i<param.copy_offset;i++)
      {
        if(k==1)
        {
          offset8x8[j][i]=B_AIF_offset_1_8x8[j][i];
        }
        else if(k==2)
        {
          offset8x8[j][i]=B_AIF_offset_5_8x8[j][i];
        }      
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief read quantize scaling table for 4x4
*
* \param[in] f Pointer to file to read from
* \param[in] sip_type SIP type of the block
*/
//-----------------------------------------------------------------------------
void Quantizer::read_qtables(FILE *f, int sip_type)
{
  int i, j, q;
  double dq;

  int *iquant_scale;
  int *quant_scale;
  double *coeff_norm_rdoq;

  for (i = ZERO; i < NQPS; i++) 
  {

    iquant_scale = inv_scale[sip_type] + i * TX_NCOEFF;

    for (j = ZERO; j < TX_NCOEFF; j++) 
    {
      fscanf(f, "%d ", &q);
      iquant_scale[j] = q;
    }
    fscanf(f,"\n");
  }

  for (i = ZERO; i < NQPS; i++) 
  {
    quant_scale  = fwd_scale[sip_type] + i * TX_NCOEFF;
    for (j = ZERO; j < TX_NCOEFF; j++) 
    {
      fscanf(f, "%d ", &q);
      quant_scale[j] = q;
    }
    fscanf(f,"\n");
  }

  for (i = ZERO; i < NQPS; i++) 
  {
    coeff_norm_rdoq = coeff_norm_rdoq_qp[sip_type] + i * TX_NCOEFF;
    for (j = ZERO; j < TX_NCOEFF; j++) 
    {
      fscanf(f, "%lf ", &dq);
      coeff_norm_rdoq[j] = dq;
    }
    fscanf(f,"\n");
  }
  rescan_d(coeff_norm_rdoq, sip_type, 0);

  return;
}

#ifndef READ_QTABLES_FROM_FILE
void Quantizer::copy_qtables(int *f, int sip_type)
{
  int i, j;
  
  int *iquant_scale;
  int *quant_scale;
  double *coeff_norm_rdoq;
  
  for (i = ZERO; i < NQPS; i++) 
  {
    
    iquant_scale = inv_scale[sip_type] + i * TX_NCOEFF;
    
    for (j = ZERO; j < TX_NCOEFF; j++) 
    {
      iquant_scale[j] = *f++;
    }
  }
  
  for (i = ZERO; i < NQPS; i++) 
  {
    quant_scale  = fwd_scale[sip_type] + i * TX_NCOEFF;
    for (j = ZERO; j < TX_NCOEFF; j++) 
    {
      quant_scale[j] = 1;
    }
  }
  
  for (i = ZERO; i < NQPS; i++) 
  {
    coeff_norm_rdoq = coeff_norm_rdoq_qp[sip_type] + i * TX_NCOEFF;
    for (j = ZERO; j < TX_NCOEFF; j++) 
    {
      coeff_norm_rdoq[j] = 1.0;
    }
  }
  rescan_d(coeff_norm_rdoq, sip_type, 0);
  
  return;
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief read quantize scaling table for 8x8
*
* \param[in] f Pointer to file to read from
*/
//-----------------------------------------------------------------------------
void Quantizer::read_qtables_8x8(FILE *f)
{
  int i, j, q;

  int *iquant_scale8x8;
  int *quant_scale8x8;
  double dq;
  double *coeff_norm_rdoq8x8;

  for (i = ZERO; i < NQPS; i++) 
  {
    iquant_scale8x8 = inv_scale8x8 + i*64;

    for (j = ZERO; j < 64; j++) 
    {
      fscanf(f, "%d ", &q);
      iquant_scale8x8[j] = q;
    }
    fscanf(f,"\n");
  }

  for (i = ZERO; i < NQPS; i++) 
  {
    quant_scale8x8  = fwd_scale8x8 + i * 64;
    for (j = ZERO; j < 64; j++) 
    {
      fscanf(f, "%d ", &q);
      quant_scale8x8[j] = q;
    }
    fscanf(f,"\n");
  }

  for (i = ZERO; i < NQPS; i++) 
  {
    coeff_norm_rdoq8x8 = coeff_norm_rdoq_qp8x8 + i * ZTsize64;
    for (j = ZERO; j < ZTsize64; j++) 
    {
      fscanf(f, "%lf ", &dq);
      coeff_norm_rdoq8x8[j] = dq;
    }
    fscanf(f,"\n");
  }

  return;
}

#ifndef READ_QTABLES_FROM_FILE
void Quantizer::copy_qtables_8x8(int *f)
{
  int i, j;
  
  int *iquant_scale8x8;
  int *quant_scale8x8;
  double *coeff_norm_rdoq8x8;
  
  for (i = ZERO; i < NQPS; i++) 
  {
    iquant_scale8x8 = inv_scale8x8 + i * 64;
    
    for (j = ZERO; j < 64; j++) 
    {
      iquant_scale8x8[j] = *f++;
    }
  }
  
  for (i = ZERO; i < NQPS; i++) 
  {
    quant_scale8x8  = fwd_scale8x8 + i * 64;
    for (j = ZERO; j < 64; j++) 
    {
      quant_scale8x8[j] = 1;
    }
  }
  
  for (i = ZERO; i < NQPS; i++) 
  {
    coeff_norm_rdoq8x8 = coeff_norm_rdoq_qp8x8 + i * ZTsize64;
    for (j = ZERO; j < ZTsize64; j++) 
    {
      coeff_norm_rdoq8x8[j] = 1.0;
    }
  }
  
  return;
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Update the quantization rounding offset
*
* \param[in out] offset The quantization rounding offset
* \param[in]     v      The coefficient before rounding
* \param[in]     vq     The coefficient after rounding
* \param[in]     bound  The quantization rounding offset upper bound
*/
//-----------------------------------------------------------------------------
int Quantizer::update_offset(int offset, int v, int vq, int bound)
{  
  offset += shift_right_round(v - (vq<<16), 10); // add 1/1024 of difference
  offset = clip(offset, 0, bound);

  return offset;
}

//-----------------------------------------------------------------------------
/*!
* \brief Set quantizer for 16x16
*
* \param[in] qp Array of quantization parameter values (one for each coefficient)
*/
//-----------------------------------------------------------------------------
void Quantizer::set_qp16x16(int *qp) 
{
  int k;
  
  for (k = ZERO; k < TX_NCOEFF16; k ++) 
    qp16x16[k] = clip(qp[k], ZERO, NQPS - ONE);    
}

#define QUANT16_SCALING  8
#define IQUANT16_SCALING 6
const int qstep[TX_SIZE16] =  {256, 245, 235, 225, 215, 206, 197, 189, 181, 173, 166, 159, 152, 146, 140, 134};
const int iqstep[TX_SIZE16] = {64, 67, 70, 73, 76, 79, 83, 87, 91, 95, 99, 103, 108, 112, 117, 123};

//-----------------------------------------------------------------------------
/*!
* \brief Quantize 16x16 block
*
* Quantizes all coefficients in a block.
*
* \param[in,out] blk Block to quantize
*
* \param[in]     type: 0: mode decision, 1: encode
*/
//-----------------------------------------------------------------------------
void Quantizer::quant16x16(Block *blk, int type) 
{
  int k;
  int sgn; 
  int v;   
  int vq;      // quantized value

  int qp_per;
  int qp_rem;
  int offset_limit = get_offset_limit();

  for (k = 0; k < TX_NCOEFF16; k ++) 
  {
    qp_per = qp16x16[k] / 16; 
    qp_rem = qp16x16[k] % 16; 

    v = blk->val[k] >> 6;    
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    v = v * qstep[qp_rem]*16/active_qmatrix_16x16[0][k]; // cq_start, cq_end
    v >>= qp_per;

    int offset_index = get_offset_index(intra, intra_frame, v, low_zeromv);
        
    vq = (v + offset16x16[offset_index][k]) >> 16;

    bool update_offset_flag = (type == 1) && get_update_offset(v, vq, low_zeromv);
    if (update_offset_flag)
    {
      offset16x16[offset_index][k] = update_offset(offset16x16[offset_index][k], v, vq, offset_limit);
    }
    if (param.qmatrix_enable_flag[1]) // cq_start
    {
      vq = max((v + (offset16x16[ZERO][k]*(128-active_qmatrix_16x16[1][k])/128) + ((1<<16)*active_qmatrix_16x16[1][k]/128)), 0) >> 16;
      if (vq>0)
        vq = max((v + offset16x16[ZERO][k] + ((1<<16)*active_qmatrix_16x16[1][k]/128)), 0) >> 16;
    } // cq_end
    blk->val[k] = vq * sgn;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Inverse quantize 16x16 block
*
* \param[in,out] blk Block to inverse quantize
*/
//-----------------------------------------------------------------------------
void Quantizer::iquant16x16(Block *blk) 
{
  int k;
  int qp_per;
  int qp_rem;
  int vq;
  int v, sgn; // cq_start, cq_end
  for (k = 0; k < TX_NCOEFF16; k ++)
  {
    qp_per = qp16x16[k] / 16; 
    qp_rem = qp16x16[k] % 16; 
    v = blk->val[k]; // cq_start
    sgn = (v < 0) ? -1 : 1;
    v = v * sgn;
    if (v>0)
    {
      blk->val[k] = sgn*(v*128 - active_qmatrix_16x16[1][k]);
      if (!param.bugfix93)
      {
        blk->val[k] = (short)blk->val[k]; 
      }
    }
    vq = blk->val[k] * iqstep[qp_rem]*active_qmatrix_16x16[0][k];
    // for high precision, need to check
    blk->val[k] = (((vq << qp_per) +1024)>>11); // cq_end
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
