/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file RefFrame.cpp
*
* \brief Implementation of RefFrame class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "RefFrame.H"
#include "Global.H"
#include "Param.H"

#ifdef MVCOMP_ZERORES
#include "MotionField.H"
#endif

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in luma sample units
* \param[in] w width of frame in luma sample units
* \param[in] mem_on indication of AIF
* \param[in] decoder_mode 0: Encoder mode, 1: Decoder mode
*/
//-----------------------------------------------------------------------------
RefFrame::RefFrame(int h, int w, int mem_on, int decoder_mode)
#else
#ifdef NO_AIF_CHROMA
//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in luma sample units
* \param[in] w width of frame in luma sample units
* \param[in] chroma_on indication of chroma AIF
* \param[in] decoder_mode 0: Encoder mode, 1: Decoder mode
*/
//-----------------------------------------------------------------------------
RefFrame::RefFrame(int h, int w, int chroma_on, int decoder_mode)
#else
//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in luma sample units
* \param[in] w width of frame in luma sample units
* \param[in] decoder_mode 0: Encoder mode, 1: Decoder mode
*/
//-----------------------------------------------------------------------------
RefFrame::RefFrame(int h, int w, int decoder_mode)
#endif
#endif
{
  int border;
  
  border = TWO * MOTION_BLK_SIZE;

  this->decoder_mode = decoder_mode;
  
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  malloc_enable = mem_on;
  if(malloc_enable)
  {
    luma = LumaFrames::create(h, w, border, decoder_mode);
    if (!decoder_mode)
      luma_hor = new int [(h+(border<<1))*((w+(border<<1))<<2)];
    chroma = ChromaFrames::create(h, w, border, param.chroma_on_fly);
    frame = new Frame(h, w);
  }
#else

  luma = LumaFrames::create(h, w, border, decoder_mode);
  if (!decoder_mode)
    luma_hor = new int [(h+(border<<1))*((w+(border<<1))<<2)];

#ifdef NO_AIF_CHROMA
  chroma_enable = chroma_on;
  if(chroma_enable)
  {
    chroma = ChromaFrames::create(h, w, border, param.chroma_on_fly);
    frame = new Frame(h, w);
  }
#else
  chroma = ChromaFrames::create(h, w, border);
  frame = new Frame(h, w);
#endif
#endif
  
  init ();

#ifdef MVCOMP_ZERORES
  // We want to store mvs -- we store the motion field, in fact.
  saved_mf = new MotionField(h, w);
#endif
  
#ifdef FAST_INTERPOL
  interpol_frame_no = -1;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
RefFrame::~RefFrame()
{
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  if(malloc_enable)
  {
    delete luma;
    if (!decoder_mode)
      delete [] luma_hor;
    delete chroma;
    delete frame;
  }
#else
  delete luma;
  if (!decoder_mode)
    delete [] luma_hor;
#ifdef NO_AIF_CHROMA
  if(chroma_enable)
  {
    delete chroma;
    delete frame;
  }
#else
  delete chroma;
  delete frame;
#endif
#endif
#ifdef MVCOMP_ZERORES
  delete saved_mf;
#endif
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Generate interpolated arrays
*
* Generates interpolated arrays RefFrame::luma and RefFrame::chroma based on
* the content of RefFrame::frame.
*/
//-----------------------------------------------------------------------------
void RefFrame::interpolate()
{
#ifdef FAST_INTERPOL
  if (interpol_frame_no != frame->frame_no)
  {
    luma->makemap(frame);
  }
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  interpolate_chroma();
#else
#ifdef NO_AIF_CHROMA
  if(chroma_enable)
    interpolate_chroma();
#else
  interpolate_chroma();
#endif
#endif
#else // FAST_INTERPOL
  luma->makemap(frame);
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  chroma->makemap(frame);
#else
#ifdef NO_AIF_CHROMA
  if(chroma_enable)
    chroma->makemap(frame);
#else
  chroma->makemap(frame);
#endif
#endif
#endif // FAST_INTERPOL
}

#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
//-----------------------------------------------------------------------------
/*!
* \brief Generate interpolated arrays
*
* Generates interpolated arrays RefFrame::chroma based on
* the content of RefFrame::frame.
*/
//-----------------------------------------------------------------------------
void RefFrame::interpolate_chroma()
{
#ifdef FAST_INTERPOL
  if (interpol_frame_no != frame->frame_no)
  {
    chroma->makemap(frame);
    interpol_frame_no = frame->frame_no;
  }
#else
  chroma->makemap(frame);
#endif
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Generate AIF interpolated arrays
*
* Generates interpolated arrays RefFrame::luma and RefFrame::chroma based on
* the content of RefFrame::frame.
*
* \param[in] cur Pointer to current frame
* \param[in] i List index (LIST_0 or LIST_1)
* \param[in] j Reference frame index
*/
//-----------------------------------------------------------------------------
void RefFrame::AIF_interpolate(Frame *cur, int i, int j, AIF_adaptive_filter *saif_half)
{
  luma->AIF_makemap(frame,cur, i, j, saif_half);
#ifndef MEM_REDUCTION_FOR_LUMA_INTERPORATION
#ifdef NO_AIF_CHROMA
  if(chroma_enable)
    chroma->AIF_makemap(frame);
#else
  chroma->AIF_makemap(frame);
#endif
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Initialize reference frame.
*
* Initialize the reference frame to not-used state.
*
*/
//-----------------------------------------------------------------------------
void RefFrame::init ()
{
  empty = TRUE;
  frame_no = ZERO;
}


//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
