/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file RefFrameManager.cpp
*
* \brief Implementation of RefFrameManager class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "RefFrameManager.H"
#include "Global.H"
#include "Param.H"
#include "AIF_adaptive_filter.H"

#ifdef MVCOMP_ZERORES
#include "MotionField.H"
#endif

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Allocates memory for reference frames and resets fields.
*
* \param[in] h Height of frame in sample units
* \param[in] w Width of frame in sample units
* \param[in] decoder_mode 0: Encoder mode, 1: Decoder mode
*/
//-----------------------------------------------------------------------------
RefFrameManager::RefFrameManager(int h, int w, int decoder_mode)
{
  int i, j;
  
  // NOTE: when creating RefFrame objects, always use the same height and width
  // as getStride() relies on all reference frames to have the same size

  this->decoder_mode = decoder_mode;
  if (decoder_mode)
  {
    param.chroma_on_fly = TRUE;
  }

  ref_process = new RefProcess();
  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      set_inter_pel(i,j,power2(EIGHTH_PEL));
    }
  }
  
  for (i = ZERO; i < MAX_REF_FRAMES; i++)
  {
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
    ref[i] = new RefFrame(h, w, 1, decoder_mode);
#else
#ifdef NO_AIF_CHROMA
    ref[i] = new RefFrame(h, w, 1, decoder_mode);
#else
    ref[i] = new RefFrame(h, w, decoder_mode);    
#endif
#endif 
    if(param.aif)
    {
#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
      AIF_ref[i] = new RefFrame(h, w, 0, decoder_mode);
      AIF_ref[i]->luma = ref[i]->luma;
      if (!decoder_mode)
        AIF_ref[i]->luma_hor = ref[i]->luma_hor;
      AIF_ref[i]->chroma = ref[i]->chroma;
      AIF_ref[i]->frame = ref[i]->frame;
      AIF_ref[i]->empty = ref[i]->empty;
      AIF_ref[i]->frame_no = ref[i]->frame_no;
#else
#ifdef NO_AIF_CHROMA
      AIF_ref[i] = new RefFrame(h, w, 0);
      AIF_ref[i]->chroma = ref[i]->chroma;
      AIF_ref[i]->frame = ref[i]->frame;
#else
      AIF_ref[i] = new RefFrame(h, w);
#endif
#endif
    }
    list[LIST_0][i] = list[LIST_1][i] = i;
    lRelation[LIST_0][i] = lRelation[LIST_1][i] = i;
    for (j = ZERO; j < MAX_REF_FRAMES; j++) 
    {
      mv_scale[LIST_0][i][j] = mv_scale[LIST_1][i][j] = power2(RESCALE_BITS);
      
    }
  }
  
  n_ref_frames = ZERO;
  single_ref_frame = ZERO;
  if (!decoder_mode)
    allocate_intermediate_memory(h, w);

  deferred_refresh_flag = 0;

#ifdef MVCOMP_ZERORES
  cur_frame_no = 0;
  cur_frame_type = 0;
#endif

  init_rpps_id();

  prev_rpps_id = ZERO;
  target_rpps_id = ZERO;
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
* Frees memory for reference frames.
*/
//-----------------------------------------------------------------------------
RefFrameManager::~RefFrameManager()
{
  int i;

  if (!decoder_mode)
    free_intermediate_memory();

  delete ref_process;

  for (i = ZERO; i < MAX_REF_FRAMES; i++)
  {
    delete ref[i];
    if (param.aif)
    {
      delete AIF_ref[i];
    }
  }
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
/*!
* \brief Allocate intermediate memory
*
* Allocate the intermediate memory for frame interpolation
* Allocated memory should be released when interpolated image has been created
*
* \param[in] h Height of frame in luma sample units
* \param[in] w Width of frame in luma sample units
*/
//-----------------------------------------------------------------------------
void RefFrameManager::allocate_intermediate_memory(int h, int w)
{
  int i, j, step;
  int width, height;
  int border;
  int k;
  
  border = TWO * MOTION_BLK_SIZE;
  
  width = w + TWO * border;
  height = h + TWO * border;

  step = (param.B_subpel > 1 || param.P_subpel> 1) ? 1 : 2;
  
  for (i=ZERO; i<FRACTIONAL_PEL; i+=step)
  {
    for (j=ZERO; j<FRACTIONAL_PEL; j+=step)
    {
      if (i&1)
        intermediate[i][j] = NULL;
      else
        intermediate[i][j] = new int[width*height];
    }
  }

  buf = new int[imax(width, height)*(FRACTIONAL_PEL>>1)];
  
  for (k = ZERO; k < MAX_REF_FRAMES; k++)
  {
    for (i=ZERO; i<FRACTIONAL_PEL; i+=step)
      for (j=ZERO; j<FRACTIONAL_PEL; j+=step)
        ref[k]->luma->intermediate[i][j] = intermediate[i][j]; 
    ref[k]->luma->buf = buf;

    if(param.aif)
    {
      for (i=ZERO; i<FRACTIONAL_PEL; i+=step)
        for (j=ZERO; j<FRACTIONAL_PEL; j+=step)
          AIF_ref[k]->luma->intermediate[i][j] = intermediate[i][j];
      AIF_ref[k]->luma->buf = buf;
      
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Free intermediate memory
*
* Free the intermediate memory for frame interpolation
* Allocated memory should be released when interpolated image has been created
*
* \param[in] h Height of frame in luma sample units
* \param[in] w Width of frame in luma sample units
*/
//-----------------------------------------------------------------------------
void RefFrameManager::free_intermediate_memory()
{
  int i, j, step;
  
  step = (param.B_subpel > 1 || param.P_subpel > 1) ? 1 : 2;
  
  for (i=ZERO; i<FRACTIONAL_PEL; i+=step)
  {
    for (j=ZERO; j<FRACTIONAL_PEL; j+=step)
    {
      if (intermediate[i][j])
        delete[] intermediate[i][j];
    }
  }
  delete[] buf;  
}

//-----------------------------------------------------------------------------
/*!
* \brief Set subpel size
*
* This function will be called when subpel size is change
*
* \param[in] interpel subpel size
* \param[in] curr_list List to be considered
* \param[in] cur_ref Reference frame index
*/
//-----------------------------------------------------------------------------
void RefFrameManager::set_inter_pel(int curr_list, int curr_ref, int interpel)
{
  ref_process->inter_pel[curr_list][curr_ref]=interpel;
  if(ref_process->inter_pel[curr_list][curr_ref] == PEL_ONE_EIGHTH) ref_process->log_inter_pel[curr_list][curr_ref] = LOG_PEL_ONE_EIGHTH;
  else if (ref_process->inter_pel[curr_list][curr_ref] == PEL_ONE_QUAD) ref_process->log_inter_pel[curr_list][curr_ref] = LOG_PEL_ONE_QUAD;
  else if (ref_process->inter_pel[curr_list][curr_ref] == PEL_ONE_HALF) ref_process->log_inter_pel[curr_list][curr_ref] = LOG_PEL_ONE_HALF;
  else if (ref_process->inter_pel[curr_list][curr_ref] == PEL_ONE_FULL) ref_process->log_inter_pel[curr_list][curr_ref] = LOG_PEL_ONE_FULL;
  else 
  {
    printf ("Error unknown inter_pel %d", interpel);
  }
}

#ifdef MEM_REDUCTION_FOR_LUMA_INTERPORATION
//-----------------------------------------------------------------------------
/*!
* \brief Apply standard filter to frames in the reference frame buffer
*
* The interpolated arrays are generated for luma and chroma if filter type is standard. 
*
*/
//-----------------------------------------------------------------------------

void RefFrameManager::standard_interpolation()
{
  int i;

  for (i = ZERO; i < n_ref_frames; i++) 
  {

     if(ref_process->filter_type[LIST_0][i]==S_FIF)
     {
       ref[list[LIST_0][i]]->interpolate();
     }else
     {
       ref[list[LIST_0][i]]->interpolate_chroma();
     }
  }
}
#endif

//-----------------------------------------------------------------------------
/*!
* \brief Add a frame to the reference frame buffer
*
* Adds the frame f to the reference frame buffer. The content of the frame
* is copied into one of the slots of the reference frame buffer and interpolated
* arrays are generated for luma and chroma. If the buffer is full prior to insertion,
* the frame with the lowest frame number is removed from the buffer.
*
* \param[in] f %Frame to be added to reference frame buffer
*/
//-----------------------------------------------------------------------------
#ifndef MVCOMP_ZERORES
void RefFrameManager::add_frame(Frame *f)
#else
void RefFrameManager::add_frame(Frame *f, MotionField *mf_frame, int *inter_pel, int *log_inter_pel, int *saved_ref_list, int *saved_frame_nos)
#endif
{
  int tmp, i, list_min;
  
  list_min = INT_INF;
  tmp = ZERO;
  
  for (i = ZERO; i < MAX_REF_FRAMES; i ++) 
  {
    if (ref[i]->empty == TRUE)
    {
      tmp = i;
      break;
    }
    else if (ref[i]->frame_no < list_min)
    {
      tmp = i;
      list_min = ref[i]->frame_no;
    }
  }
#ifdef MVCOMP_ZERORES
  // Copy the motion field -- we copy everything for now.
  ref[tmp]->saved_mf->copy_field_from(mf_frame);

  // Save the reference lists in effect.
  memcpy(&ref[tmp]->saved_ref_list[0][0], saved_ref_list != NULL ? saved_ref_list : &list[LIST_0][0], sizeof(ref[tmp]->saved_ref_list));

  // Copy the frame-no's used for references for this frame's mf.
  // Use -1 for non-existing.
  if (saved_frame_nos != NULL)
  {
      for (i = ZERO; i < MAX_REF_FRAMES; i++)
          ref[tmp]->saved_ref_framenos[i] = saved_frame_nos[i];
  }
  else
  {
      for (i = ZERO; i < MAX_REF_FRAMES; i++)
          ref[tmp]->saved_ref_framenos[i] = ref[i]->empty ? -1 : ref[i]->frame_no;
  }

  // Copy the interpolation in effect for later mv scaling use.
  memcpy(ref[tmp]->saved_inter_pel, inter_pel, sizeof(ref[tmp]->saved_inter_pel));
  memcpy(ref[tmp]->saved_log_inter_pel, log_inter_pel, sizeof(ref[tmp]->saved_log_inter_pel));
#endif


  ref[tmp]->frame->copy(f);
#ifndef MEM_REDUCTION_FOR_LUMA_INTERPORATION
  if(param.aif)
  {
    AIF_ref[tmp]->frame->copy(f);
  }
  
#ifdef FAST_INTERPOL
  ref[tmp]->interpol_frame_no = -1;
#endif
  ref[tmp]->interpolate();
#endif
  ref[tmp]->empty = FALSE;
  ref[tmp]->frame_no = f->frame_no;
  n_ref_frames = imin(n_ref_frames + ONE, MAX_REF_FRAMES);
}

//-----------------------------------------------------------------------------
/*!
* \brief Replace the frame in the reference frame buffer with AIF reference
*
* The frame added to the reference frame buffer is replaced with AIF interpolated frame
*
* \param[in] cur %Frame to be interpolated
* \param[in] saif_half pointer to aif instance
*/
//-----------------------------------------------------------------------------
void RefFrameManager::AIF_replace_frame(Frame *cur, AIF_adaptive_filter *saif_half)
{
  int i, j;
  char flag_replaced[MAX_REF_FRAMES];
  
  for(j=0;j<MAX_REF_FRAMES;j++)
  {
    flag_replaced[j] = 0;
  }

  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      if (ref[list[i][j]]->empty != TRUE)
      {
        if((ref_process->filter_type[i][j]==AIF_SAIF) && (flag_replaced[list[i][j]]==0))
        {
          flag_replaced[list[i][j]] = 1;
#ifdef FAST_INTERPOL
          ref[list[i][j]]->interpol_frame_no = -1;
#endif
          AIF_ref[list[i][j]]->AIF_interpolate(cur,i,j,saif_half);
          AIF_ref[list[i][j]]->empty = FALSE;
          AIF_ref[list[i][j]]->frame_no = ref[list[i][j]]->frame_no;
        }
      }
    }
  }  
}

//-----------------------------------------------------------------------------
/*!
* \brief Generate reference list
*
* Generates the reference list list_no. For a B-frame, this needs to be called twice; once
* for each list. Also generates scaling factors that may be used for rescaling
* motion vectors.
*
* \param[in] frame_type Type of frame for which to generate list (ft_P_frame or ft_B_frame)
* \param[in] list_no Index of list to be generated (LIST_0 or LIST_1)
* \param[in] frame_no Number of frame for which to generate list
*/
//-----------------------------------------------------------------------------
void RefFrameManager::prepare_list(int frame_type, int list_no, int frame_no)
{
  int i, j, fno_i, fno_i1, tmp;
  int sorting_rule;
  int time_scale;
  
  time_scale = ONE - TWO * list_no; // invert time :) when dealing with list 1
  
  frame_no = time_scale * frame_no; 
  
  for (j = ZERO; j < MAX_REF_FRAMES - ONE; j++) 
  {
    if (ref[list[list_no][j + ONE]]->empty == TRUE)
      break; // stop if no more entries
    
    i = j;
    while (i >= ZERO)
    {
      fno_i  = ref[list[list_no][i      ]]->frame_no * time_scale;
      fno_i1 = ref[list[list_no][i + ONE]]->frame_no * time_scale;
      
      switch (frame_type)
      {
          //if a P frame then rule is based on distance first, then direction.
        case ft_P_frame:
          sorting_rule = (iabs(fno_i1 - frame_no) < iabs(fno_i - frame_no) || (iabs(fno_i1 - frame_no) == iabs(fno_i - frame_no) && fno_i1 < fno_i));
          break;
          //if a B frame then rule is based on direction first, then distance.
        case ft_B_frame:
        default:
          sorting_rule = ((fno_i < frame_no) && (fno_i1 < frame_no) && (fno_i < fno_i1)) ||((fno_i > frame_no) && (fno_i > fno_i1));
          break;
      }
      
      if (sorting_rule == TRUE)
      {
        tmp = list[list_no][i];
        list[list_no][i] = list[list_no][i + 1];
        list[list_no][i + 1] = tmp;
        i--;
      }
      else
        break;
    }
  }
  
  // if both lists are identical and contain 2 or more elements, swap first two elements in list 1
  if (list_no == LIST_1 &&
    ref[list[list_no][ONE]]->empty == FALSE &&
    ref[list[list_no][ZERO]]->frame_no == ref[list[LIST_0][ZERO]]->frame_no)
  {
    tmp = list[list_no][ZERO];
    list[list_no][ZERO] = list[list_no][ONE];
    list[list_no][ONE] = tmp;
  }
  
  // compute list size
  j = MAX_REF_FRAMES;
  while (j > ZERO && ref[list[list_no][j - ONE]]->empty == TRUE)
    j--;
  
  if (single_ref_frame == TRUE)
    j = imin(j, ONE);
  
  if (param.adapt_list_size == TRUE && frame_type == ft_B_frame && j > 0) 
  {
    while (j > 1 && (time_scale * ref[list[list_no][j - ONE]]->frame_no > frame_no))
      j--;
  }
  
  lsize[list_no] = imin(j, param.nref_frames);
  frame_no = frame_no * time_scale;
  
  for (i=ZERO;i < lsize[list_no]; i++) 
  {
    for (j=ZERO;j < lsize[list_no]; j++) 
    {
      mv_scale[list_no][i][j] = power2(RESCALE_BITS) * (ref[list[list_no][i]]->frame_no - frame_no) / (ref[list[list_no][j]]->frame_no - frame_no) ;
    }
  }

#ifdef MV_CODING_ENHANCE
  for (i = 0; i < lsize[list_no]; i++)
  {
    dist[list_no][i] = frame_no - ref[list[list_no][i]]->frame_no;
    inv_dist[list_no][i] = 4096 / dist[list_no][i];
  }
#endif

#ifdef MVCOMP_ZERORES
  cur_frame_no = frame_no;
  cur_frame_type = frame_type;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to reference luma arrays
*
* \param[in] index Index into list
* \param[in] cur_list Index of list to use (LIST_0 or LIST_1)
* \return Pointer to interpolated luma arrays
*/
//-----------------------------------------------------------------------------
LumaFrames *RefFrameManager::get_luma(int index, int cur_list)
{
  return ref[list[cur_list][index]]->luma;
}


//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to AIF reference luma arrays
*
* \param[in] index Index into list
* \param[in] cur_list Index of list to use (LIST_0 or LIST_1)
* \return Pointer to interpolated luma arrays
*/
//-----------------------------------------------------------------------------
LumaFrames *RefFrameManager::AIF_get_luma(int index, int cur_list)
{
  return AIF_ref[list[cur_list][index]]->luma;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to reference chroma arrays
*
* \param[in] index Index into list
* \param[in] cur_list Index of list to use (LIST_0 or LIST_1)
* \return Pointer to interpolated chroma arrays
*/
//-----------------------------------------------------------------------------
ChromaFrames *RefFrameManager::get_chroma(int index,int cur_list)
{
  return ref[list[cur_list][index]]->chroma;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to AIF reference chroma arrays
*
* \param[in] index Index into list
* \param[in] cur_list Index of list to use (LIST_0 or LIST_1)
* \return Pointer to interpolated chroma arrays
*/
//-----------------------------------------------------------------------------
ChromaFrames *RefFrameManager::AIF_get_chroma(int index,int cur_list)
{
  return AIF_ref[list[cur_list][index]]->chroma;
}

//-----------------------------------------------------------------------------
/*!
* \brief Identifies list indices that point to the same reference frame
*
* Sets the lRelation array which maps an index from LIST_0 into and index
* from LIST_1 if both refer to the same reference frame. This may be used
* in motion estimation to switch between LIST_0 and LIST_1 modes.
*/
//-----------------------------------------------------------------------------
void RefFrameManager::get_lRelation()
{
  int i, j;
  
  // Lets reset relationship
  for (j = ZERO; j < MAX_REF_FRAMES; j++) 
    lRelation[LIST_0][j] = lRelation[LIST_1][j] = -1;
  
  for (j = ZERO; j < lsize[LIST_0]; j++) 
  {
    if (ref[list[LIST_0][j]]->empty == TRUE)
      break; // stop if no more entries
    
    for (i = ZERO; i < lsize[LIST_1]; i++) 
    {
      if (ref[list[LIST_1][i]]->empty == TRUE)
        break; // stop if no more entries
      
      if (list[LIST_0][j] == list[LIST_1][i]) 
      {
        lRelation[LIST_0][j] = i;
        lRelation[LIST_1][i] = j;
        break;
      }
      
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Set parameters for reference frame processing and create AIF refefence pictures 
*
* \param[in] cur Frame to be interpolated
*/
//-----------------------------------------------------------------------------
void RefFrameManager::process_list(Frame * /*cur*/, AIF_adaptive_filter *saif_half)
{
  int i, j;
  int counter=0;
  
  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      set_inter_pel(i,j,power2(ref_process->inter_type[i][j])); //S.Kato Set inter_pel[2][4] and log_inter_pel[2][4]
      if(param.aif)
      {
        saif_half->aif_filterflag[i][j]=ref_process->filter_type[i][j]+1;
        if(ref_process->filter_type[i][j]==S_FIF||ref_process->filter_type[i][j]>=N_FIF)
        {
          saif_half->AIF_FilterFlag=1;
          aif_loop[i][j]=0;
        }
        else
        {
          if(counter==0)
          {
            saif_half->AIF_FilterFlag = ref_process->filter_type[i][j]+1;
            counter++;
          }
          aif_loop[i][j]=2;
        }
      }      
    }
  }
}


//-----------------------------------------------------------------------------
/*!
* \brief Initilization and definition of default RPPS
*/
//-----------------------------------------------------------------------------
void RefFrameManager::init_rpps_id()
{
  int i,j;  
  for(i=0;i<256;i++)
  {
    ref_pic_parameter_set_id_used[i]=ZERO;
    ref_pic_parameter_set_id_shared[i]=ZERO;
  }
  
  //ID=ZERO: all 1/4, S_FIF, IC_ON, ROUND_LASTSTEP
  ref_pic_parameter_set_id_used[ZERO]=ONE;
  ref_pic_parameter_set_id_shared[ZERO]=ONE;
  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      inter_type_set[ZERO][i][j]=QUAD_PEL;
      filter_type_set[ZERO][i][j]=S_FIF;
      ic_disable_set[ZERO][i][j]=IC_ON;
      round_type_set[ZERO][i][j]=ROUND_LASTSTEP;
    }
  }

  //ID=ONE: all 1/4, SAIF, IC_ON, ROUND_LASTSTEP
  ref_pic_parameter_set_id_used[ONE]=ONE;
  ref_pic_parameter_set_id_shared[ONE]=ONE;
  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      inter_type_set[ONE][i][j]=QUAD_PEL;
      filter_type_set[ONE][i][j]=AIF_SAIF;
      ic_disable_set[ONE][i][j]=IC_ON;
      round_type_set[ONE][i][j]=ROUND_LASTSTEP;
    }
  }

  //ID=TWO: all 1/8, S_FIF, IC_ON, ROUND_LASTSTEP
  ref_pic_parameter_set_id_used[TWO]=ONE;
  ref_pic_parameter_set_id_shared[TWO]=ONE;
  for(i=0;i<LISTS;i++)
  {
    for(j=0;j<MAX_REF_FRAMES;j++)
    {
      inter_type_set[TWO][i][j]=EIGHTH_PEL;
      filter_type_set[TWO][i][j]=S_FIF;
      ic_disable_set[TWO][i][j]=IC_ON;
      round_type_set[TWO][i][j]=ROUND_LASTSTEP;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Find the unused smallest rpps_id 
* 
* \return rpps_id
* If rpps_id is bigger than MAX_RPPS_ID, return -1
*/
//-----------------------------------------------------------------------------
int RefFrameManager::find_rpps_id()
{
  int id;
  int i, j;
  
  int count;
  
  
  for(id=0;id<MAX_RPPS_ID;id++)
  {
    if(ref_pic_parameter_set_id_used[id]==ONE)
    {
      
      count=0;
      
      for(i=0;i<LISTS;i++)
      {
        for(j=0;j<MAX_REF_FRAMES;j++)
        {
          
          if(ref_process->inter_type[i][j]!=inter_type_set[id][i][j])
          {
            count=1;
            break;
          }
          if(ref_process->filter_type[i][j]!=filter_type_set[id][i][j])
          {
            count=1;
            break;
          }
          if(ref_process->ic_disable[i][j]!=ic_disable_set[id][i][j])
          {
            count=1;
            break;
          }
          if(ref_process->round_type[i][j]!=round_type_set[id][i][j])
          {
            count=1;
            break;
          }
        }
      }
      if(count==0)
      {
        return id;
      }
      
    }
  }
  
  for(id=0;id<MAX_RPPS_ID;id++)
  {
    if(ref_pic_parameter_set_id_used[id]==ZERO)
    {
      ref_pic_parameter_set_id_used[id]=ONE;
      for(i=0;i<LISTS;i++)
      {
        for(j=0;j<MAX_REF_FRAMES;j++)
        {
          inter_type_set[id][i][j]=ref_process->inter_type[i][j];
          filter_type_set[id][i][j]=ref_process->filter_type[i][j];
          ic_disable_set[id][i][j]=ref_process->ic_disable[i][j];
          round_type_set[id][i][j]=ref_process->round_type[i][j];
        }
      }
      return id;
    }
  }
  
  printf("Error Exceed max_rpps_id");
  return -1;
  
}

//-----------------------------------------------------------------------------
// Methods for debugging
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
/*!
* \brief Print reference lists
*
*/
//-----------------------------------------------------------------------------
void RefFrameManager::print_lists()
{
  int list_no, j;
  
  for (list_no = ZERO; list_no < LISTS; list_no++)
  {
    printf("List %d : [ ", list_no);
    for (j = ZERO; j < lsize[list_no]; j++) 
    {
      if (ref[list[list_no][j]]->empty == TRUE)
        printf("x ");
      else
        printf("%d ",ref[list[list_no][j]]->frame_no);
    }
    printf(" ]\n");
  }
  
  for (list_no = ZERO; list_no < LISTS; list_no++)
  {
    printf("ListRelation %d : [ ", list_no);
    for (j = ZERO; j < lsize[list_no]; j++) 
    {
      if (ref[list[list_no][j]]->empty == TRUE)
        printf("x ");
      else
        printf("%d ",lRelation[list_no][j]);
    }
    printf(" ]\n");
  }
  
}

//-----------------------------------------------------------------------------
/*!
* \brief Refresh reference frame buffer. Set all frames to EMPTY.
* 
*/
//-----------------------------------------------------------------------------
void RefFrameManager::init (void)
{
  int list_no;
  int n;

  for (n=0; n<MAX_REF_FRAMES; n++)
  {
    ref[n]->init ();
  }
  for (list_no = ZERO; list_no < LISTS; list_no++)
  {
    for (n=0; n<MAX_REF_FRAMES; n++)
    {
      list[list_no][n] = n;
  }
    lsize[list_no] = 0;
  }
  n_ref_frames = 0;
  deferred_refresh_flag = 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief Performs deferred refresh.
* 
*  Performs deferred refresh. All frames except 
*  frame_no==deferred_refresh_frame_number are purges.
*  The remaining frame is put to the first buffer space.
*  This operation is to be carried out after deferred_refresh_flag is set to 1
*  along with a valid deferred_refresh_frame_number value.
*  Subsequently frame numbers are checked until one with a value 
*  > deferred_refresh_frame_number. Then this function is called.
*  After that the reference frame buffer only has one reference frame left,
*  whose frame number is deferred_refresh_frame_number.
*  For those frames with frame number <= deferred_refresh_frame_number, they 
*  will just be added to the reference frame buffer as per normal.
*/
//-----------------------------------------------------------------------------
void RefFrameManager::perform_deferred_refresh ()
{
  int n;
  int list_no;
  int index = -1;
  RefFrame *ref_temp;

  //===============================================================
  //  Check for which one is the frame to retain. Free the rest.
  //===============================================================
  for (n=0; n<MAX_REF_FRAMES; n++)
  {
    if (!ref[n]->empty && ref[n]->frame_no==deferred_refresh_frame_number)
    {
      index = n;
    }
    else
    {
      ref[n]->init ();
    }
  }
  if (index>=0)
  {
    n_ref_frames = 1;
    n_ref_frames = 1;
    ref_temp = ref[0];    // these three lines swap the reference frame 
    ref[0] = ref[index];  // to the location 0
    ref[index] = ref_temp;
    for (list_no = ZERO; list_no < LISTS; list_no++)
    {
      for (n=0; n<MAX_REF_FRAMES; n++)
    {
        list[list_no][n] = n;
    }
      lsize[list_no] = 1;
  }
  }
  else
  {
    //===============================================================
    //  The list is cleared totally.
    //===============================================================
    n_ref_frames = 0;
    for (list_no = ZERO; list_no < LISTS; list_no++)
    {
      for (n=0; n<MAX_REF_FRAMES; n++)
    {
        list[list_no][n] = n;
    }
      lsize[list_no] = 0;
  }
  }
  deferred_refresh_flag = 0;
}

//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to luma reference data 
*
* \param[in] list LIST_0 or LIST_1
* \param[in] ref_index reference index
* \param[in] mv motion vector
* \param[in] i vertical position
* \param[in] j horizontal position
* \param[in] h height of block to be fetched
* \param[in] w width of block to be fetched
* \return luma pointer
*/
//-----------------------------------------------------------------------------
unsigned char *RefFrameManager::get_luma_ptr(int list, int ref_index, int *mv, int i, int j, int h, int w, unsigned char *data)
{
  LumaFrames *lf = get_luma(ref_index, list);
  int temp_aif_loop = 0;
  if (param.aif)
  {
    temp_aif_loop = aif_loop[list][ref_index];

    if (temp_aif_loop != 0)
    {
      assert(temp_aif_loop == 2);
      lf = AIF_get_luma(ref_index, list);
    }
  }
  
  int temp_log_inter_pel = ref_process->log_inter_pel[list][ref_index];
  
  int u = clip(i + (mv[vertical  ] >> temp_log_inter_pel) + lf->border, 0, lf->height - h);
  int v = clip(j + (mv[horizontal] >> temp_log_inter_pel) + lf->border, 0, lf->width  - w);
  
  int k = (mv[vertical  ] << (LOG_FRACTIONAL_PEL - temp_log_inter_pel)) & (FRACTIONAL_PEL - 1);
  int l = (mv[horizontal] << (LOG_FRACTIONAL_PEL - temp_log_inter_pel)) & (FRACTIONAL_PEL - 1);
  
  if (decoder_mode)
  {
    u = clip(u, 2, lf->height - h - 3);
    v = clip(v, 2, lf->width - w - 3);
    assert(h == MOTION_BLK_SIZE);
    assert(w == MOTION_BLK_SIZE);

    if (temp_aif_loop)
    {
      lf->interpolate_on_the_fly_AIF(k, l, u, v, data);
    }
    else
    {
      lf->interpolate_on_the_fly(k, l, u, v, data);
    }
    return data;
  }
  else
    return lf->data[k][l] + lf->width * u + v;
}

  
//-----------------------------------------------------------------------------
/*!
* \brief Get pointer to chroma reference data 
*
* NOTE: all sizes/positions are in luma sample units
*
* \param[in] list LIST_0 or LIST_1
* \param[in] ref_index reference index
* \param[in] mv motion vector
* \param[in] i vertical position
* \param[in] j horizontal position
* \param[in] h height of block to be fetched
* \param[in] w width of block to be fetched
* \return chroma pointer
*/
//-----------------------------------------------------------------------------
unsigned char *RefFrameManager::get_chroma_ptr(int list, int ref_index, int *mv, int i, int j, int h, int w, unsigned char *data)
{
  ChromaFrames *cf;

  if (param.aif)
  {
    int temp_aif_loop = aif_loop[list][ref_index];
    
    if (temp_aif_loop == 0)
    {
      cf = get_chroma(ref_index, list);
    }
    else if (temp_aif_loop == 2)
    {
      cf = AIF_get_chroma(ref_index, list);
    }
    else
    {
      printf("Error: bad value of aif_loop: %d\n", temp_aif_loop);
      exit(1);
    }
  }
  else
  {
    cf = get_chroma(ref_index, list);
  }
  
  int temp_log_inter_pel = ref_process->log_inter_pel[list][ref_index];
  
  int u = clip(((i + (mv[vertical  ] >> temp_log_inter_pel)) >> 1) + cf->border, 0, cf->height - (h >> 1));
  int v = clip(((j + (mv[horizontal] >> temp_log_inter_pel)) >> 1) + cf->border, 0, cf->width  - (w >> 1));
  
  int k = (mv[vertical  ] << (LOG_FRACTIONAL_PEL - temp_log_inter_pel)) & (2 * FRACTIONAL_PEL - 1);
  int l = (mv[horizontal] << (LOG_FRACTIONAL_PEL - temp_log_inter_pel)) & (2 * FRACTIONAL_PEL - 1);
  
  if (data != NULL)
  {
    cf->interpolate_chroma_on_the_fly(k, l, u, v, data);
    return data;
  }
  else
  {
    assert(cf->decoder_mode == 0);
    return cf->data[k][l] + 2 * (cf->width * u + v);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Get width of one line of reference data (luma/chroma) 
*
* NOTE: this function relies on the fact that all reference data has the same
*       width
*
* \return width in samples
*/
//-----------------------------------------------------------------------------
int RefFrameManager::get_stride()
{
  if (decoder_mode)
    return MOTION_BLK_SIZE;     
  else
    return ref[0]->luma->width;
}
//-----------------------------------------------------------------------------
/*!
* \brief Set AIF filter coefficients 
*
* \param[in] saif_half: pointer to AIF data
*/
//-----------------------------------------------------------------------------
void RefFrameManager::getAIFcoef(AIF_adaptive_filter *saif_half) 
{
  for(int j=0;j<15;j++)
    for(int i=0;i<7;i++)
      iFilterCoef_SAIF[j][i]=saif_half->iFilterCoef_SAIF[j][i]; 

  AIF_mode_flag1 = saif_half->AIF_mode_flag1;
  AIF_mode_flag2 = saif_half->AIF_mode_flag2;
}

//-----------------------------------------------------------------------------
/*!
* \brief Set AIF filter coefficients 
*
* \param[in] saif_half: pointer to AIF data
*/
//-----------------------------------------------------------------------------
void RefFrameManager::setAIFcoef(LumaFrames *lf) 
{
  for(int j=0;j<15;j++)
    for(int i=0;i<7;i++)
      lf->iFilterCoef_SAIF[j][i]=iFilterCoef_SAIF[j][i];   

  lf->AIF_mode_flag1 = AIF_mode_flag1;
  lf->AIF_mode_flag2 = AIF_mode_flag2;
}
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
