/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file SimpleProbEst.cpp
* \brief Implementation of SimpleProbEst class
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "SimpleProbEst.H"
#include "Param.H"
#include "Global.H"

#define NEW_KSHIFT (30-SYNCHRO_MAX_CONFIDENCE) // keep log table with 30 bits of precision.
#define SHIFTED_SYNCHRO_MAX_CONFIDENCE (SYNCHRO_MAX_CONFIDENCE<<NEW_KSHIFT)

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Reset
*
* Resets estimate to 50/50
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::reset() 
{
  p = power2(PROB_BITS - 1);
}

//-----------------------------------------------------------------------------
/*!
* \brief Update probability estimate
*
* \param[in] val Coded symbol
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::update(int val) 
{
  if (type!='S')
  {
    printf("Unexpected SPE type!\n");
    abort();
  }
  if (val == ZERO)
    p -= shift_right_round(p - power2(PROB_BITS) + param.ac_exp_limit, param.loss_rate);
  else
    p -= shift_right_round(p - param.ac_exp_limit, param.loss_rate);
  
#ifdef CABAC_SYNC_POINTS
    nb++;
#endif
#ifdef TRACE_STATS  
  symcount[val]++;
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Set probability estimate
*
* Sets estimate to p.
*
* \param[in] p Probability estimate
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::set_prob(int p) 
{
  this->p = p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy proba settings from src.
*
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::copyInitialProba(ProbEst *src)
{
  // The copies should be between identical contexts.
  // Can compare name pointers as they're original literal strings.
  if (idCtx != src->idCtx || idName != src->idName)
  {
    printf("copyInitialProba error: src ctxt %d (%s) to ctxt %d (%s)\n",
           src->idCtx, src->idName, idCtx, idName);
    abort();
  }
  p = src->p;
  i_p = src->i_p;
}

//-----------------------------------------------------------------------------
/*!
* \brief Restore current probability from stored value
*
* Restore current Probabiliy status from default value
*
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::restoreDefaultState()
{
  p=i_p;
}

#ifdef CABAC_SYNC_POINTS
int SimpleProbEst::KFunc(int mnb0, int mnb1)
{
  if (mnb0 < 0 || mnb0 >= MAX_NBLOG)
    mnb0 = MAX_NBLOG-1;
  if (mnb1 < 0 || mnb1 >= MAX_NBLOG)
    mnb1 = MAX_NBLOG-1;

  int lp0 = tLog[mnb0];
  int lp1 = tLog[mnb1];

  int t = SHIFTED_SYNCHRO_MAX_CONFIDENCE;
  //int t = 14;
  if (lp0 > t) lp0 = t;
  if (lp1 > t) lp1 = t;
  int diff = lp0-lp1;

  if (diff > 0)
  {
    int k= (diff + 3*(1<<(NEW_KSHIFT-1)) -1) >>NEW_KSHIFT;
    return k;
  }
  else if (diff == 0)
    return 1;
  else
  {
    int k= (-diff + 3*(1<<(NEW_KSHIFT-1)) -1)>>NEW_KSHIFT;
    return -k;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Update symbol information for the nominated superblock.
*
* The probability and mean number of symbols related to this superblock are updated.
*
* \param[in] i The index of the superblock in the line.
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::updateSynchro(int i)
{
  if (type!='S')
  {
    printf("Unexpected SPE type!\n");
    abort();
  }
  //-- store proba found
  p2d[i] = p; 
  //-- store reliability info
  mnb[i] += shift_right_round((nb<<SYNCHRO_SHIFT_CONFIDENCE)-mnb[i], LOSS_RATE_SYNCHRO);

}

//-----------------------------------------------------------------------------
/*!
* \brief Initialize superblock info from an average of two available neighbours.
*
* Information from neighbouring superblocks (left, up) is examined in order to
* fabricate starting values for the current superblock.
*
* \param[in] i Superblock index in the current line being processed.
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::averaging(int i)
{
  int s = mnb[i-1]+mnb[i];
  int nm=0;
  int p2, nm2;
  if ( (s!=0) && (mnb[i]!=0) )
  {
    { //-- average variation (no division)
      if (mnb[i-1]<mnb[i])
      {
        int k=KFunc(mnb[i], mnb[i-1]);
        p2=p2d[i];
        if (k < 32)
        {
          if (p2d[i-1] > p2d[i])
            p2 += shift_right_round(p2d[i-1]-p2d[i], k);
          else
            p2 -= shift_right_round(p2d[i]-p2d[i-1], k);
        }

        nm2 = mnb[i];
        nm2 += shift_right_round(mnb[i-1]-mnb[i], k);
        //printf("simple0: %d %d %d %d k=%d -> nm=%d p=%d\n", mnb[i-1], mnb[i], p2d[i-1], p2d[i], k, p2, nm2);
      }
      else
      {
        int k=KFunc(mnb[i-1], mnb[i]);
        p2 = p2d[i-1];
        if (k < 0)
          printf("-ve!");
        if (k < 32)
          p2 += shift_right_round(p2d[i]-p2d[i-1], k);
        nm2 = mnb[i-1];
        if (k < 32)
          nm2 += shift_right_round(mnb[i]-mnb[i-1], k);
        //printf("simple1: %d %d %d %d k=%d -> nm=%d p=%d\n", mnb[i-1], mnb[i], p2d[i-1], p2d[i], k, p2, nm2);
      }
      p = p2;
      nm = nm2;
    }
  }
  else
  { // no symbols yet coded up to now
    // keep with proba previously proposed
    p = p2d[i-1];
    nm = mnb[i-1];
    //printf("simplenull: p=%d nm=%d\n", p, nm);
  }
  mnb[i] = nm;
}


//-----------------------------------------------------------------------------
/*!
* \brief Reset all synchronisation information to an initial state.
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::resetSynchro()
{
  if (type!='S')
  {
    printf("Unexpected SPE type!\n");
    abort();
  }
  p2d[0] = p; 
  mnb[0] = 0;
  for(int i=1; i < MAX_AC_SYNCHRO; i++)
  {
    p2d[i] = 0;
    mnb[i] = 0;
    nb = 0;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Initialize synchronisation information for a particular superblock.
*
* Information from neighbouring superblocks (left, up) is examined in order to
* fabricate starting values for the current superblock.
*
* \param[in] i Superblock index in the current line being processed.
*/
//-----------------------------------------------------------------------------
void SimpleProbEst::initSynchro(int i)
{
  if (type!='S')
  { printf("type not supported!\n"); abort();}

  nb=0; // count how many symb in current synchro bloc
  if (i == 0)
  {
    // take proba from prev line
    p = p2d[0];
    mnb[i] = mnb[0];
  } 
  else
  {
    if (p2d[i])
    { // compute proba averaging (init p for coding, and associated reliability)
      averaging(i);
    }
    else
    { // no symbols observed yet here, use previous proba
      p= p2d[i-1];
      mnb[i] = mnb[i-1];
    }
  }
  //if (idCtx==0) printf("init Synchro %2d mnb= %6d\n", i, mnb[i]);
  return;
}

#endif // CABAC_SYNC_POINTS

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
