/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file Trace.cpp
*
* \brief Implementation of Trace class
*
*
*/


#include "Global.H"
#ifdef TRACE_DEBUG

#include "Trace.H"

#ifdef SUBSTREAM_MT
// Our substream indicator is kept in thread local storage.
#include <pthread.h>
extern pthread_key_t substream_number_key;
#define edec_ss (reinterpret_cast<long>(pthread_getspecific(substream_number_key)))
#elif defined(FRAME_SUBSTREAMS)
// Our substream indicator is kept in our edec_ss field, updated per
//superblock.
int Trace::edec_ss;
#endif

#ifdef FRAME_SUBSTREAMS
FILE* Trace::m_pTraceFiles[100];
int Trace::m_nSubStreams;
#else
FILE* Trace::m_pTraceFile;
#endif
FILE* Trace::m_ptmpTraceFile;
FILE* Trace::m_pbestTraceFile;
bool  Trace::m_bWriteTrace;
bool  Trace::m_bDecoderTrace;
double Trace::m_dBitsCount;

#ifdef FRAME_SUBSTREAMS
#define m_pTraceFile m_pTraceFiles[edec_ss]
#endif

#ifdef FRAME_SUBSTREAMS
char *Trace::replace_percent_x(char const *src, char const *percent, int val, char *res, size_t res_size)
{
  char valstr[50];
  if (res_size < strlen(src)+100)
    abort(); // We should have more than this.
  char const *dpos = strstr(src, percent);
  snprintf(valstr, sizeof(valstr), "%03d", val);
  if (dpos == NULL)
  {
    // Suffix the value.
    strcpy(res, src);
    strcat(res, valstr);
  }
  else
  {
    // Replace %d with the value.
    memcpy(res, src, dpos-src);
    res[dpos-src] = 0;
    strcat(res, valstr);
    strcat(res, dpos+strlen(percent)); // Skip the %d.
  }
  return res;
}
#endif

#ifdef FRAME_SUBSTREAMS
void Trace::openTrace(char* pcString, int iMode, int nSubStreams)
#else
void Trace::openTrace(char* pcString, int iMode)
#endif
{
#ifdef FRAME_SUBSTREAMS
  // More than one substream is only expected in the decoder.
  char res[200]; // Temporary name.
  m_nSubStreams = nSubStreams;
  if (m_nSubStreams == 0)
    m_nSubStreams = 1;
  if (m_nSubStreams > sizeof(m_pTraceFiles)/sizeof(m_pTraceFiles[0]))
  {
    printf("too many substreams (%d) to trace\n", m_nSubStreams);
    exit(EXIT_FAILURE);
  }
#endif
  switch(iMode)
  {
  case 0:
#ifdef FRAME_SUBSTREAMS
    // open a trace per substream.
    {
      for (int i = 0; i < m_nSubStreams; i++)
        if (m_pTraceFiles[i] == NULL)
          m_pTraceFiles[i] = fopen(replace_percent_x(pcString, "%d", i, res, sizeof(res)), "wt");
      break;
    }
#else
    {
      if(m_pTraceFile == NULL)
        m_pTraceFile = fopen(pcString, "wt");
      break;
    }
#endif
  case 1:
    {
      if(m_ptmpTraceFile == NULL)
        m_ptmpTraceFile = fopen(pcString, "wt");
      break;
    }
  case 2:
    {
      if(m_pbestTraceFile == NULL)
        m_pbestTraceFile = fopen(pcString, "wt");
      break;
    }
  }
}

void Trace::closeTrace(int iMode)
{
  switch(iMode)
  {
  case 0:
#ifdef FRAME_SUBSTREAMS
    // close all traces.
    {
      for (int i = 0; i < m_nSubStreams; i++)
        if(m_pTraceFiles[i] != NULL)
        {
          fclose(m_pTraceFiles[i]);
          m_pTraceFiles[i] = NULL;
        }
      break;
    }
#else
    {
      if(m_pTraceFile != NULL)
      {
        fclose(m_pTraceFile);
        m_pTraceFile = NULL;
      }
      break;
    }
#endif
  case 1:
    {
      if(m_ptmpTraceFile != NULL)
      {
        fclose(m_ptmpTraceFile);
        m_ptmpTraceFile = NULL;
      }
      break;
    }
  case 2:
    {
      if(m_pbestTraceFile != NULL)
      {
        fclose(m_pbestTraceFile);
        m_pbestTraceFile = NULL;
      }
      break;
    }
  }
}

void Trace::printTrace(char* pcString, int val, int iMode)
{
  if(m_bWriteTrace || m_bDecoderTrace)
  {
  switch(iMode)
  {
  case 0:
    {
      if(m_pTraceFile != NULL)
        fprintf(m_pTraceFile, "%s %d ", pcString, val);
      break;
    }
  case 1:
    {
      if(m_ptmpTraceFile != NULL)
        fprintf(m_ptmpTraceFile, "%s %d ", pcString, val);
      break;
    }
  case 2:
    {
      if(m_pbestTraceFile != NULL)
        fprintf(m_pbestTraceFile, "%s %d ", pcString, val);
      break;
    }
  }
  }
}

void Trace::printTrace(char* pcString, int val, int ctx, int iMode)
{
  if(m_bWriteTrace || m_bDecoderTrace || iMode == 3)
  {
  switch(iMode)
  {
  case 0:
    {
      if(m_pTraceFile != NULL)
        fprintf(m_pTraceFile, "%s %d ctx %d \n", pcString, val, ctx);
      break;
    }
  case 1:
    {
      if(m_ptmpTraceFile != NULL)
        fprintf(m_ptmpTraceFile, "%s %d ctx %d \n", pcString, val, ctx);
      break;
    }
  case 2:
    {
      if(m_pbestTraceFile != NULL)
        fprintf(m_pbestTraceFile, "%s %d ctx %d \n", pcString, val, ctx);
      break;
    }
  }
  }
}

void Trace::printTrace(char* pcString, int iMode)
{
   if(m_bWriteTrace || m_bDecoderTrace || iMode == 3)
  {
  switch(iMode)
  {
  case 0:
    {
      if(m_pTraceFile != NULL)
        fprintf(m_pTraceFile, "%s ", pcString);
      break;
    }
  case 1:
    {
      if(m_ptmpTraceFile != NULL)
        fprintf(m_ptmpTraceFile, "%s ", pcString);
      break;
    }
  case 2:
    {
      if(m_pbestTraceFile != NULL)
        fprintf(m_pbestTraceFile, "%s ", pcString);
      break;
    }
  }
   }
}

void Trace::printTrace(int iVal, int iMode)
{
   if(m_bWriteTrace || m_bDecoderTrace || iMode == 3)
  {
    switch(iMode)
    {
    case 0:
      {
        if(m_pTraceFile != NULL)
          fprintf(m_pTraceFile, "%d ", iVal);
        break;
      }
    case 1:
      {
        if(m_ptmpTraceFile != NULL)
          fprintf(m_ptmpTraceFile, "%d ", iVal);
        break;
      }
    case 2:
      {
        if(m_pbestTraceFile != NULL)
          fprintf(m_pbestTraceFile, "%d ", iVal);
        break;
      }
    }
   }
}

void Trace::printTrace(int iMode)
{
   if(m_bWriteTrace || m_bDecoderTrace )
  {
    switch(iMode)
    {
    case 0:
      {
        if(m_pTraceFile != NULL)
          fprintf(m_pTraceFile, "\n ");
        break;
      }
    case 1:
      {
        if(m_ptmpTraceFile != NULL)
          fprintf(m_ptmpTraceFile, "\n ");
        break;
      }
    case 2:
      {
        if(m_pbestTraceFile != NULL)
          fprintf(m_pbestTraceFile, "\n ");
        break;
      }
    }
   }
}

void Trace::copyFile(int iModeSrc, int iModeDst)
{
  FILE* pSrc, *pDst;
  if(iModeSrc == 1)
    pSrc = fopen("TraceEncoderTmp.txt", "rt");
  if(iModeSrc == 2)
    pSrc = fopen("TraceEncoderBest.txt", "rt");

  if(iModeDst == 2)
    pDst = fopen("TraceEncoderBest.txt", "wt");
  if(iModeDst == 0)
    pDst = (m_pTraceFile != NULL ? m_pTraceFile : fopen("TraceEncoder.txt", "wt"));

  char line[500];
  fgets(line, 500, pSrc);
  while(!feof(pSrc))
  {
    fprintf(pDst, "%s", line);
    fgets(line, 500, pSrc);
    if(feof(pSrc))
    {
      fprintf(pDst, "%s", line);
    }
  }

  fclose(pSrc);
  if(iModeDst == 2 || m_pTraceFile == NULL)
    fclose(pDst);
}

void Trace::initFrameTrace(int iVal, int iMode, int iDec)
{
  if(iDec == 0)
    m_bDecoderTrace = true;

  m_dBitsCount = 0;

  switch(iMode)
  {
  case 0:
#ifdef FRAME_SUBSTREAMS
    // into all traces.
    {
      for (int i = 0; i < m_nSubStreams; i++)
        if(m_pTraceFiles[i] != NULL)
          fprintf(m_pTraceFiles[i], "============ Frame %d ========================\n ", iVal);
      break;
    }
#else
    {
      if(m_pTraceFile != NULL)
        fprintf(m_pTraceFile, "============ Frame %d ========================\n ", iVal);
      break;
    }
#endif
  case 1:
    {
      if(m_ptmpTraceFile != NULL)
        fprintf(m_ptmpTraceFile, "============ Frame %d ========================\n ", iVal);
      break;
    }
  case 2:
    {
      if(m_pbestTraceFile != NULL)
        fprintf(m_pbestTraceFile, "============ Frame %d ========================\n ", iVal);
      break;
    }
  }
}

void Trace::initBlockTrace(int iValj, int iVali, int iMode)
{
  switch(iMode)
  {
  case 0:
    {
      if(m_pTraceFile != NULL)
        fprintf(m_pTraceFile, "------- Block %d %d ---------\n ", iValj, iVali);
      break;
    }
  case 1:
    {
      if(m_ptmpTraceFile != NULL)
        fprintf(m_ptmpTraceFile, "------- Block %d %d ---------\n ", iValj, iVali);
      break;
    }
  case 2:
    {
      if(m_pbestTraceFile != NULL)
        fprintf(m_pbestTraceFile, "------- Block %d %d ---------\n ", iValj, iVali);
      break;
    }
  }
}

void Trace::countBits(double d)
{
  m_dBitsCount+= d;
}

void Trace::PrintBits()
{
  printf("bits: %4.f\n", m_dBitsCount);
}

#endif
