/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
* \file TransformAlternate.cpp
*
* \brief Implementation of TransformAlternate class
*
*
*/

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "TransformAlternate.H"

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

TransformAlternate::TransformAlternate(int frametype, int inter8x8type)
{
  m_frametype = frametype;
  m_inter8x8type = inter8x8type;
}

const int IntraTransform4x4[16] = {26, 81, 156, 184, 105, 181, 32, -122, 175, -3, -153, 106, 141, -144, 123, -60};
const int InterTransform4x4[16] = {62, 134, 169, 123, 157, 124, -68, -121, 149, -88, -101, 160, 105, -144, 140, -89};

void TransformAlternate::forward(Block *blk)
{
  if (blk->get_height() == 2)
    forward_2x8(blk);
  else if (blk->get_width() == 2)
    forward_8x2(blk);
  else if (blk->get_width() == 8)
    forward_8x8(blk);
  else
    forward_4x4(blk);
}

void TransformAlternate::forward_4x4(Block *blk)
{
  int i, j, k, p;
  int tmp[TX_NCOEFF];
  int ival;
  int a[4];  
  
  const int *T = (m_frametype == ft_I_frame) ? IntraTransform4x4 : InterTransform4x4;

  for ( i=0; i<TX_SIZE; i++ ) 
  {
    for( j = 0; j<TX_SIZE; j++ )
    {
      a[j] = blk->val[TX_SIZE*i+j];
    }

    for( j = 0, p = 0; j<TX_SIZE; j++, p+=TX_SIZE )
    {
      tmp[i+TX_SIZE*j] = 0;
      for( k = 0; k<TX_SIZE; k++)
      {
        tmp[i+TX_SIZE*j] += T[p+k]*a[k];
      }
      tmp[i+TX_SIZE*j] = ( tmp[i+TX_SIZE*j] + (1<<6) ) >> 7;
    }
  }
  
  for ( i=0; i<TX_SIZE; i++ ) 
  {
    for( j = 0; j<TX_SIZE; j++ )
    {
      a[j] = tmp[TX_SIZE*i+j];
    }

    for( j = 0, p = 0; j<TX_SIZE; j++, p+=TX_SIZE )
    {
      ival = 0;
      for( k = 0; k<TX_SIZE; k++)
      {
        ival += T[p+k]*a[k];
      }
      blk->val[i+TX_SIZE*j] = ( ival + (1<<10) ) >> 11;
    }
  }
}

void TransformAlternate::inverse(Block *blk)
{
  int blk_height = blk->get_height();
  int blk_width = blk->get_width();
  
  if (blk_height == 2)
    inverse_2x8(blk);
  else if (blk_width == 2)
    inverse_8x2(blk);
  else if (blk_width == 4)
    inverse_4x4(blk);
  else
    inverse_8x8(blk);
}

void TransformAlternate::inverse_4x4(Block *blk)
{
  int i, j, k, p;
  int tmp[TX_NCOEFF];
  int ival;
  int a[4];
  
  const int *T = (m_frametype == ft_I_frame) ? IntraTransform4x4 : InterTransform4x4;

  for ( i=0; i<TX_SIZE; i++ ) 
  {
    for( j = 0; j<TX_SIZE; j++ )
    {
      a[j] = blk->val[TX_SIZE*i+j];
    }

    for( j = 0; j<TX_SIZE; j++ )
    {
      tmp[i+TX_SIZE*j] = 0;
      for( k = 0, p = 0; k<TX_SIZE; k++, p+=TX_SIZE )
      {
        tmp[i+TX_SIZE*j] += T[p+j]*a[k];
      }
      tmp[i+TX_SIZE*j] = ( tmp[i+TX_SIZE*j] + (1<<5) ) >> 6;
    }
  }

  for ( i=0; i<TX_SIZE; i++ ) 
  {
    for( j = 0; j<TX_SIZE; j++ )
    {
      a[j] = tmp[TX_SIZE*i+j];
    }

    for( j = 0; j<TX_SIZE; j++)
    {
      ival = 0;
      for( k = 0, p = 0; k<TX_SIZE; k++, p+=TX_SIZE )
      {
        ival += T[p+j]*a[k];
      }
      blk->val[i+TX_SIZE*j] = ( ival + (1<<7) ) >> 8;
    }
  }
}

const int IntraTransform8x8_MxN[64] = {10,    31,    77,   142,   207,   248,   265,   247,
                                      164,   266,   303,   238,    87,   -50,  -134,  -149,
                                      132,   162,    61,  -123,  -199,  -103,    80,   211,
                                      244,   180,  -159,  -245,    70,   260,    46,  -211,
                                      234,   -19,  -250,    82,   223,  -166,  -166,   181,
                                      266,  -199,   -80,   249,  -212,   -26,   226,  -144,
                                      136,  -200,   152,   -59,   -58,   171,  -194,    91,
                                      108,  -180,   216,  -232,   237,  -230,   188,   -86};
const int IntraTransform2x2[4]  = {106, 233, 233, -106};

void TransformAlternate::forward_8x2(Block *blk)
{
  int i, j, k, p;
  int tmp[TX_NCOEFF];
  int ival;
  int a[8]; 

  assert(m_frametype == ft_I_frame); // 8x2 not supported for inter blocks
  const int *Tv = IntraTransform8x8_MxN;
  const int *Th = IntraTransform2x2;

  for ( i=0; i<2; i++ ) 
  {
    for( j = 0; j<8; j++ )
    {
      a[j] = blk->val[i+2*j];
    }

    for( j = 0, p = 0; j<8; j++, p+=8 )
    {
      tmp[8*i+j] = 0;
      for( k = 0; k < 8; k++ )
      {
        tmp[8*i+j] += Tv[p+k]*a[k];
      }
      tmp[8*i+j] = ( tmp[8*i+j] + (1<<6) ) >> 7;
    }
  }
  
  for ( i=0; i<8; i++ ) 
  {
    for( j = 0; j<2; j++ )
    {
      a[j] = tmp[i+8*j];
    }

    for( j = 0, p = 0; j<2; j++, p+=2 )
    {
      ival = 0;
      for( k = 0; k<2; k++)
      {
        ival += Th[p+k]*a[k];
      }
      blk->val[2*i+j] = ( ival + (1<<11) ) >> 12;
    }
  }
}

void TransformAlternate::inverse_8x2(Block *blk)
{
  int i, j, k, p;
  int tmp[TX_NCOEFF];
  int ival;
  int a[8];

  assert(m_frametype == ft_I_frame);
  
  const int *Tv = IntraTransform8x8_MxN;
  const int *Th = IntraTransform2x2;

  for ( i=0; i<2; i++ ) 
  {
    for( j = 0; j<8; j++ )
    {
      a[j] = blk->val[i+2*j];
    }

    for( j = 0; j<8; j++ )
    {
      tmp[8*i+j] = 0;
      for( k = 0, p = 0; k < 8; k++, p+=8  )
      {
        tmp[8*i+j] += Tv[p+j]*a[k];
      }
      tmp[8*i+j] = ( tmp[8*i+j] + (1<<5) ) >> 6;
    }
  }
  
  for ( i=0; i<8; i++ ) 
  {
    for( j = 0; j<2; j++ )
    {
      a[j] = tmp[i+8*j];
    }

    for( j = 0; j<2; j++ )
    {
      ival = 0;
      for( k = 0, p = 0; k<2; k++, p+=2 )
      {
        ival += Th[p+j]*a[k];
      }
      blk->val[2*i+j] = ( ival + (1<<8) ) >> 9;
    }
  }
}

void TransformAlternate::forward_2x8(Block *blk)
{
  int i, j, k, p;
  int tmp[TX_NCOEFF];
  int ival;
  int a[8];

  assert(m_frametype == ft_I_frame);
  
  const int *Tv = IntraTransform2x2;  
  const int *Th = IntraTransform8x8_MxN;
  
  for ( i=0; i<8; i++ ) 
  {
    for( j = 0; j<2; j++ )
    {
      a[j] = blk->val[i+8*j];
    }

    for( j = 0, p = 0; j<2; j++, p+=2 )
    {
      tmp[2*i+j] = 0;
      for( k = 0; k < 2; k++ )  
      {    
        tmp[2*i+j] += Tv[p+k]*a[k];  
      }
      tmp[2*i+j] = ( tmp[2*i+j] + (1<<5) ) >> 6;
    }
  }
  
  for ( i=0; i<2; i++ ) 
  {
    for( j = 0; j<8; j++ )
    {
      a[j] = tmp[i+2*j];
    }

    for( j = 0, p = 0; j<8; j++, p+=8 )
    {
      ival = 0;
      for( k = 0; k<8; k++ )
      {
        ival += Th[p+k]*a[k];
      }
      blk->val[8*i+j] = ( ival + (1<<12) ) >> 13;
    }
  }
}

void TransformAlternate::inverse_2x8(Block *blk)
{
  int i, j, k, p;
  int tmp[TX_NCOEFF];
  int ival;
  int a[8];

  assert(m_frametype == ft_I_frame);
  
  const int *Tv = IntraTransform2x2;
  const int *Th = IntraTransform8x8_MxN;

  for ( i=0; i<8; i++ ) 
  {
    for( j = 0; j<2; j++ )
    {
      a[j] = blk->val[i+8*j];
    }

    for( j = 0; j<2; j++)
    {
      tmp[2*i+j] = 0;
      for( k = 0, p = 0; k < 2; k++, p+=2  )
      {
        tmp[2*i+j] += Tv[p+j]*a[k];
      }
      tmp[2*i+j] = ( tmp[2*i+j] + (1<<4) ) >> 5;
    }
  }
  
  for ( i=0; i<2; i++ ) 
  {
    for( j = 0; j<8; j++ )
    {
      a[j] = tmp[i+2*j];
    }

    for( j = 0; j<8; j++ )
    {
      ival = 0;
      for( k = 0, p = 0; k<8; k++, p+=8 )
      {
        ival += Th[p+j]*a[k];
      }
      blk->val[8*i+j] = ( ival + (1<<9) ) >> 10;
    }
  }
}


const int IntraTransform8x8[64] = {
    50,    95,   155,   224,   289,   332,   354,   346,
   226,   349,   389,   309,   135,   -70,  -236,  -306,
   243,   256,    82,  -171,  -269,  -152,    82,   255,
   394,   147,  -300,  -288,   122,   355,    86,  -306,
   344,  -135,  -319,   205,   260,  -246,  -230,   251,
   297,  -349,    31,   331,  -335,   -37,   339,  -206,
   145,  -252,   228,  -102,   -92,   267,  -290,   130,
   127,  -251,   320,  -359,   360,  -311,   219,   -87};

const int InterTransform8x8[3][64] =
{
  {
    126,   217,   293,   356,   356,   280,   186,    91,
    234,   340,   296,   110,  -130,  -328,  -367,  -251,
    238,   238,    30,  -222,  -229,    17,   233,   239,
    351,   123,  -331,  -243,   223,   305,  -107,  -361,
    314,   -84,  -326,   206,   219,  -320,   -96,   331,
    316,  -318,   -19,   337,  -334,    20,   288,  -286,
    182,  -264,   205,   -78,   -67,   226,  -293,   188,
    147,  -265,   316,  -340,   337,  -306,   245,  -132
  },
  { // inter_KLT8x8_smooth
    243,   247,   263,   270,   270,   263,   247,   243,
    246,   352,   308,   123,  -117,  -316,  -355,  -240,
    187,   185,   -36,  -300,  -305,   -43,   187,   193,
    355,   127,  -326,  -235,   231,   310,  -104,  -358,
    296,  -106,  -359,   165,   178,  -352,  -114,   317,
    316,  -318,   -20,   336,  -335,    19,   288,  -286,
    175,  -275,   189,   -97,   -86,   210,  -302,   183,
    147,  -265,   316,  -340,   336,  -306,   245,  -132
  },
  { // inter_KLT8x8_DC
    256,   256,   256,   256,   256,   256,   256,   256,
    246,   353,   308,   122,  -118,  -317,  -355,  -239,
    178,   177,   -44,  -309,  -314,   -50,   179,   184,
    355,   127,  -326,  -235,   231,   310,  -104,  -358,
    294,  -108,  -363,   160,   173,  -356,  -116,   315,
    316,  -319,   -20,   336,  -335,    19,   288,  -286,
    175,  -274,   190,   -96,   -85,   211,  -302,   183,
    147,  -265,   316,  -340,   336,  -306,   245,  -132
  }
};

void TransformAlternate::forward_8x8(Block *blk)
{
  int i, j, k, p;
  int tmp[TX_NCOEFF8];
  int ival;
  int a[TX_SIZE8];
  
  const int *T = (m_frametype == ft_I_frame) ? IntraTransform8x8 : InterTransform8x8[m_inter8x8type];
  
  for (i = 0; i < TX_SIZE8; i ++) 
  {
    for (j = 0; j < TX_SIZE8; j ++ )
      a[j] = blk->val[TX_SIZE8 * i + j];

    for (j = 0, p = 0; j < TX_SIZE8; j ++, p += TX_SIZE8)
    {
      tmp[i + TX_SIZE8 * j] = 0;
      for (k = 0; k < TX_SIZE8; k ++)
        tmp[i + TX_SIZE8 * j] += T[p + k] * a[k];

      tmp[i + TX_SIZE8 * j] = (tmp[i + TX_SIZE8 * j] + (1 << 10)) >> 11;
    }
  }
  
  for (i = 0; i < TX_SIZE8; i ++) 
  {
    for (j = 0; j < TX_SIZE8; j ++ )
      a[j] = tmp[TX_SIZE8 * i + j];

    for (j = 0, p = 0; j < TX_SIZE8; j ++, p += TX_SIZE8)
    {
      ival = 0;
      for (k = 0; k < TX_SIZE8; k ++)
        ival += T[p + k] * a[k];

      blk->val[i + TX_SIZE8 * j] = (ival + (1 << 10) ) >> 11;
    }
  }
}

void TransformAlternate::inverse_8x8(Block *blk)
{
  int i, j, k, p;
  int tmp[TX_NCOEFF8];
  int ival;
  int a[TX_SIZE8];
  
  const int *T = (m_frametype == ft_I_frame) ? IntraTransform8x8 : InterTransform8x8[m_inter8x8type];
  
  for (i = 0; i < TX_SIZE8; i ++) 
  {
    for (j = 0; j < TX_SIZE8; j ++ )
      a[j] = blk->val[TX_SIZE8 * i + j];

    for (j = 0; j < TX_SIZE8; j ++)
    {
      tmp[i + TX_SIZE8 * j] = 0;
      for (k = 0, p = 0; k < TX_SIZE8; k ++, p += TX_SIZE8)
        tmp[i + TX_SIZE8 * j] += T[p+j] * a[k];

      tmp[i + TX_SIZE8 * j] = (tmp[i + TX_SIZE8 * j] + (1 << 7)) >> 8;
    }
  }

  for (i = 0; i < TX_SIZE8; i ++) 
  {
    for (j = 0; j < TX_SIZE8; j ++ )
      a[j] = tmp[TX_SIZE8 * i + j];

    for(j = 0; j < TX_SIZE8; j ++)
    {
      ival = 0;
      for (k = 0, p = 0; k < TX_SIZE8; k ++, p += TX_SIZE8)
        ival += T[p + j] * a[k];

      blk->val[i + TX_SIZE8 * j] = (ival + (1 << 7)) >> 8;
    }
  }
}

int KLT16x16[] =
{
  6,    9,   12,   15,   19,   22,   27,   30,   33,   36,   39,   41,   43,   44,   45,   44,
  24,   31,   37,   40,   42,   42,   39,   33,   24,   13,    1,  -12,  -23,  -32,  -39,  -41,
  37,   43,   41,   36,   24,    7,  -13,  -30,  -40,  -43,  -36,  -22,   -2,   18,   35,   43,
  40,   41,   29,    8,  -19,  -37,  -43,  -33,   -9,   20,   39,   43,   31,    6,  -25,  -45,
  54,   42,    0,  -37,  -49,  -32,    3,   34,   40,   22,   -6,  -31,  -35,  -17,   12,   34,
  35,   14,  -22,  -40,  -20,   19,   41,   27,  -12,  -42,  -34,    3,   42,   48,    0,  -51,
  45,    6,  -45,  -36,   16,   44,   20,  -27,  -40,   -8,   38,   35,  -13,  -42,  -17,   35,
  41,  -15,  -48,    2,   41,   16,  -31,  -30,   15,   41,    8,  -44,  -24,   41,   31,  -36,
  43,  -31,  -39,   35,   29,  -29,  -28,   22,   31,  -14,  -39,   12,   45,  -17,  -42,   30,
  31,  -39,  -10,   49,  -13,  -42,   24,   34,  -25,  -35,   36,   25,  -45,   -1,   38,  -22,
  25,  -41,   14,   32,  -47,    0,   52,  -31,  -32,   49,   -1,  -40,   25,   13,  -31,   17,
  24,  -38,   26,    2,  -34,   40,   -4,  -41,   44,    1,  -43,   41,   -2,  -37,   45,  -19,
  19,  -39,   37,  -12,  -20,   41,  -39,    9,   26,  -44,   38,  -10,  -25,   47,  -45,   20,
  20,  -39,   43,  -38,   22,    6,  -33,   49,  -43,   17,   11,  -29,   38,  -39,   29,  -11,
  12,  -27,   39,  -46,   48,  -40,   24,   -9,  -10,   27,  -38,   42,  -41,   33,  -21,    8,
  5,  -12,   18,  -25,   33,  -39,   43,  -47,   47,  -44,   41,  -37,   30,  -20,   11,   -4
};

//-----------------------------------------------------------------------------
/*!
 * \brief Forward transform
 *
 * \param[in,out] blk Block of data to transform
 */
//-----------------------------------------------------------------------------
void TransformAlternate::forward_transform(Block *blk)
{
  int i, j, k, p, ival;
  int a[256];
  int tmp[256];
  int tsize;
  int scale;
  int shift2;
  
  tsize = 16;
  assert(blk->get_height() == tsize && blk->get_width() == tsize);

  scale = 3;  
  shift2 = (1 << (scale - 1));
  
  for (i = 0; i < tsize; i ++) 
  {
    for (j = 0; j < tsize; j ++ )
      a[j] = blk->val[tsize * i + j];
    
    for (j = 0, p = 0; j < tsize; j ++, p += tsize)
    {
      tmp[i + tsize * j] = 0;
      for (k = 0; k < tsize; k ++)
        tmp[i + tsize * j] += KLT16x16[p + k] * a[k];
    }
  }
  
  for (i = 0; i < tsize; i ++) 
  {
    for (j = 0; j < tsize; j ++ )
      a[j] = (tmp[tsize * i + j] + shift2) >> scale;
    
    for (j = 0, p = 0; j < tsize; j ++, p += tsize)
    {
      ival = 0;
      for (k = 0; k < tsize; k ++)
        ival += KLT16x16[p + k] * a[k];
      
      // val is "short", use val to high precision, need to check
      blk->val[i + tsize * j] = (ival + shift2) >> scale; 
    }
  }
}
//-----------------------------------------------------------------------------
/*!
 * \brief Inverse transform
 *
 * \param[in,out] blk Block of data to transform
 * \param[in] TransType transform type
 */
//-----------------------------------------------------------------------------
void TransformAlternate::inverse_transform(Block *blk)
{
  int i, j, k, p;
  int tmp[256];
  int ival;
  int a[256];
  
  int tsize;
  int scale, shift2;
  
  tsize = 16;
  assert(blk->get_height() == tsize && blk->get_width() == tsize);

  scale = 7;  
  shift2 = (1 << (scale - 1));
    
  for (i = 0; i < tsize; i ++) 
  {
    for (j = 0; j < tsize; j ++ )
      a[j] = blk->val[tsize * i + j];
    
    for (j = 0; j < tsize; j ++)
    {
      tmp[i + tsize * j] = 0;
      for (k = 0, p = 0; k < tsize; k ++, p += tsize)
        tmp[i + tsize * j] += KLT16x16[p+j] * a[k];
    }
  }
  
  for (i = 0; i < tsize; i ++) 
  {
    for (j = 0; j < tsize; j ++ )
      a[j] = (tmp[tsize * i + j] + shift2) >> scale;
    
    for(j = 0; j < tsize; j ++)
    {
      ival = 0;
      for (k = 0, p = 0; k < tsize; k ++, p += tsize)
        ival += KLT16x16[p + j] * a[k];
      
      blk->val[i + tsize * j] = (ival + shift2) >> scale;
    }
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
