/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
 * \file UnaryProbEst.cpp
 * \brief Implementation of UnaryProbEst class
 */
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "UnaryProbEst.H"
#include "Param.H"
#include "Global.H"

#ifdef CABAC_SYNC_POINTS
static bool bTabUpDnSuff = false;
static int UpSuff[2<<PROB_BITS];
static int DnSuff[2<<PROB_BITS];
#endif

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
 * \brief Reset estimator
 *
 * Sets suffix and suffix size to zero and probability estimate for prefix to 1/2.
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::reset()
{
  p = power2(PROB_BITS - 1);
  suffix = ZERO;
  suffix_size = ZERO;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Set probability estimate
 *
 * Sets probability and adjusts size of suffix.
 *
 * \param[in] p Probability estimate
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::set_prob(int p)
{
  suffix = ZERO;          // Fix for the release / debug mode mismatch.  suffix was not initialised before.
  suffix_size = ZERO;
  while (p < 0x2000) 
  {
    p = TWO * p - shift_right_round(p * p, PROB_BITS);
    suffix_size++;
  }
  this->p = p;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Updates suffix size
 *
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::adjust_suffix()
{
  if (p < 0x2000) 
  {
    p = (p << 1) - 0x400;  // bug fix   changed 0x800 to 0x400
    suffix_size++;
  }
  else if (p > 0x3c00 && suffix_size != ZERO) 
  { // bug fix   changed 0x3800 to 0x3c00
    p = (p + 0x400) >> 1; // bug fix   changed 0x800 to 0x400
    suffix_size--;
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Copy proba settings from src.
 *
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::copyInitialProba(ProbEst *src)
{
  UnaryProbEst *usrc = (reinterpret_cast<UnaryProbEst *>(src));
  // The copies should be between identical contexts.
  // Can compare name pointers as they're original literal strings.
  if (idCtx != usrc->idCtx || idName != usrc->idName)
  {
    printf("copyInitialProba error: src ctxt %d (%s) to ctxt %d (%s)\n",
           usrc->idCtx, usrc->idName, idCtx, idName);
    abort();
  }
  i_p = usrc->i_p;
  p = usrc->p;
  suffix_size = usrc->suffix_size;
}

void UnaryProbEst::set_default_prob(unsigned int p)
{
  i_p = p;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Restore current probability from stored value
 *
 * Restore current Probabiliy status from default value
 *
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::restoreDefaultState()
{
  set_prob(i_p);
#ifdef CABAC_SYNC_POINTS
  p2d[0] = p; 
  s2d[0] = suffix_size;
  mnb[0] = 0;
  // mnb[0] = 4;
  for(int i = 1; i < MAX_AC_SYNCHRO; i++)
  {
    p2d[i] = 0;
    s2d[i] = 0;
    mnb[i] = 0;
    nb = 0;
  }
#endif // CABAC_SYNC_POINTS
}

#ifdef CABAC_SYNC_POINTS
//-----------------------------------------------------------------------------
/*!
 * \brief Initialize the global up/down suffix table information.
 *
 * This routine only initializes the global table once.
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::BuildTabUpDnSuff()
{
  if (!bTabUpDnSuff)
  {
    int lni = 0;
    int mni = (1<<PROB_BITS)-1;
    DnSuff[0] = 0;
    for(int i = 0; i <= mni; i++)
    {
      int ni = 2*i - shift_right_round_us(i*i, PROB_BITS);
      if (ni > mni) ni = mni;
      UpSuff[i] = ni;
      for(int j = lni+1; j <= ni; j++)
        DnSuff[j] = i;
      lni = ni;
    }
    bTabUpDnSuff = true;
  }
}
#endif // CABAC_SYNC_POINTS

//-----------------------------------------------------------------------------
/*!
 * \brief Update probability information, including # symbols coded in superblock.
 *
 * \param[in] val Coded symbol.
 *
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::update(int val) 
{
  if (type!='U')
  {
    printf("Unexpected UPE type!\n");
    abort();
  }
  
  if (val == ZERO)
  {
    p -= shift_right_round(p - power2(PROB_BITS) + param.ac_exp_limit, param.loss_rate);
  }
  else
  {
    p -= shift_right_round(p - param.ac_exp_limit, param.loss_rate);
  }
  
#ifdef CABAC_SYNC_POINTS
  {
    nb += (2<<suffix_size); // update nb symb in synchro bloc
  }
#endif
  
#ifdef TRACE_STATS  
  symcount[val]++;
#endif
}

#ifdef CABAC_SYNC_POINTS
//-----------------------------------------------------------------------------
/*!
 * \brief Update symbol information for the nominated superblock.
 *
 * The probability and mean number of symbols related to this superblock are updated.
 *
 * \param[in] i The index of the superblock in the line.
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::updateSynchro(int i)
{
  if (type!='U')
  {
    printf("Unexpected UPE type!\n");
    abort();
  }
  //-- store proba found
  p2d[i] = p; 
  s2d[i] = suffix_size;
  //-- store reliability info
  //mnb[i] = nb;
  mnb[i] += shift_right_round((nb<<10)-mnb[i], LOSS_RATE_SYNCHRO);
  //if (idCtx==0) printf("update Synchro %2d mnb= %6d\n", i, mnb[i]);
  return; 
}


//-----------------------------------------------------------------------------
/*!
 * \brief Initialize superblock info from an average of two available neighbours.
 *
 * Information from neighbouring superblocks (left, up) is examined in order to
 * fabricate starting values for the current superblock.
 *
 * \param[in] i Superblock index in the current line being processed.
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::saveraging(int i)
{
  int s = mnb[i-1]+mnb[i];
  int nm = 0;
  int p2, nm2,s2;
  if ( (s!=0) && (mnb[i]!=0) )
  {
    { //-- average variation (no division)
      int ss = s2d[i];
      int pp = p2d[i];
      // rescale if necessary p2d[i-1]
      while (ss > (int)s2d[i-1])
      { ss--; pp = DnSuff[pp]; }
      while (ss < (int)s2d[i-1])
      { ss++; pp = UpSuff[pp]; }
      
      if (mnb[i-1]<mnb[i])
      {
        int k=KFunc(mnb[i], mnb[i-1]);
        p2=pp;
        if (k < 32)
          p2 += shift_right_round(p2d[i-1]-pp, k);
        s2 = s2d[i-1];
        nm2 = mnb[i];
        if (k < 32)
          nm2 += shift_right_round(mnb[i-1]-mnb[i], k);
        //printf("unary0: %d %d %d(%d) %d(%d) k=%d -> nm=%d p=%d s=%d\n", mnb[i-1], mnb[i], p2d[i-1], s2d[i-1], p2d[i], s2d[i], k, p2, nm2, s2);
      }
      else
      {
        int k=KFunc(mnb[i-1], mnb[i]);
        p2 = p2d[i-1];
        if (k < 32)
          p2 += shift_right_round(pp-p2d[i-1], k);
        s2 = s2d[i-1];
        nm2 = mnb[i-1];
        if (k < 32)
          nm2 += shift_right_round(mnb[i]-mnb[i-1], k);
        //printf("unary1: %d %d %d(%d) %d(%d) k=%d -> nm=%d p=%d s=%d\n", mnb[i-1], mnb[i], p2d[i-1], s2d[i-1], p2d[i], s2d[i], k, p2, nm2, s2);
      }
      
      // rescale if necessary p2
      while (p2 < 0x2000)
      { s2++; p2 = UpSuff[p2];}
      while ((p2 > 0x4000) && (s2 > 0))
      { s2--; p2 = DnSuff[p2];}
      
      
      p = p2;
      suffix_size = s2;
      nm = nm2;
      //printf("unary: nm=%d p=%d\n", nm2, p2);
    }
  }
  else
  { // no symbols yet coded up to now
    p = p2d[i-1];
    suffix_size = s2d[i-1];
    nm = mnb[i-1];
    //printf("unary0: nm=%d p=%d\n", nm, p);
  }
  mnb[i] = nm;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Initialize synchronisation information for a particular superblock.
 *
 * Information from neighbouring superblocks (left, up) is examined in order to
 * fabricate starting values for the current superblock.
 *
 * \param[in] i Superblock index in the current line being processed.
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::initSynchro(int i)
{
  if (type != 'U')
  {
    printf("Unexpected UPE type!\n");
    abort();
  }
  
  nb = 0; // count how many symb in current synchro bloc
  if (i == 0)
  {
    // take proba from prev line
    p = p2d[0];
    suffix_size = s2d[0];
    mnb[i] = mnb[0];
  } 
  else
  {
    if (p2d[i])
    { // compute proba averaging (init p for coding, and associated reliability)
      saveraging(i);
    }
    else
    { // no symbols observed yet here, use previous proba
      p = p2d[i-1];
      suffix_size = s2d[i-1];
      mnb[i] = mnb[i-1];
    }
  }
  return;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Reset all synchronisation information to an initial state.
 */
//-----------------------------------------------------------------------------
void UnaryProbEst::resetSynchro()
{
  if (type != 'U')
  {
    printf("Unexpected UPE type!\n");
    abort();
  }
  p2d[0] = p;
  mnb[0] = 0;
  s2d[0] = suffix_size;
  // mnb[0] = 4;
  for(int i = 1; i < MAX_AC_SYNCHRO; i++)
  {
    p2d[i] = 0;
    mnb[i] = 0;
    s2d[i] = 0;
    nb = 0;
  }
}
#endif // CABAC_SYNC_POINTS

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
