/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

/*!
 * \file WFrame.cpp
 *
 * \brief Implementation of WFrame class
 *
 *
 */

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "WFrame.H"
#include "Global.H"
#include "Param.H" 

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
 * \brief Constructor
 *
 * \param[in] h height of frame in luma sample units
 * \param[in] w width of frame in luma sample units
 */
//-----------------------------------------------------------------------------
WFrame::WFrame(int h, int w)
{
  height = h;
  width = w;
  
  height2 = height >> 1;
  width2  = width  >> 1;
  height4 = height >> 2;
  width4  = width  >> 2;
  height8 = height >> 3;
  width8  = width  >> 3;
  
  m_size = width * height + 2 * width2 * height2;
  
  m_data = new short[m_size];
  m_luma = m_data;
  m_crcb[0] = m_luma + width * height;
  m_crcb[1] = m_crcb[0] + width2 * height2;
  
  clear();
}

//-----------------------------------------------------------------------------
/*!
 * \brief Destructor
 *
 */
//-----------------------------------------------------------------------------
WFrame::~WFrame()
{
  delete[] m_data;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
 * \brief Clear frame
 *
 * Sets all luma and chroma samples to 0.
 */
//-----------------------------------------------------------------------------
void WFrame::clear()
{
  memset(m_data, 0, m_size * sizeof(short));
}

//-----------------------------------------------------------------------------
/*!
 * \brief Copy frame
 *
 * Copies data of frame f into current frame.
 *
 * \note Both frames should have the same size for this to work properly.
 *
 * \param[in] f %Frame from which data is copied
 */
//-----------------------------------------------------------------------------
void WFrame::copy(WFrame *f)
{
  memcpy(m_data, f->m_data, m_size * sizeof(short));
}

//-----------------------------------------------------------------------------
/*!
 * \brief Add a frame
 *
 * \param[in] f %Frame to add
 */
//-----------------------------------------------------------------------------
void WFrame::add(WFrame *f)
{
  int i;
  
  for (i = 0; i < m_size; i++)
  {
    m_data[i] += f->m_data[i];
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Subtract a frame
 *
 * \param[in] f %Frame to subtract
 */
//-----------------------------------------------------------------------------
void WFrame::sub(WFrame *f)
{
  int i;
  
  for (i = 0; i < m_size; i++)
  {
    m_data[i] -= f->m_data[i];
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Compute SAD in a frame
 *
 * \param[in] f %Frame to compute SAD
 *
 * \return sad
 */
//-----------------------------------------------------------------------------
double WFrame::sad(WFrame *f)
{
  int i;
  double diff = 0.0;
  
  for (i = 0; i < m_size; i++)
  {
    diff += (double) abs(m_data[i] - f->m_data[i]);
  }
  
  return diff;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Compute SSE (Sum of square error) in a frame
 *
 * \param[in] f %Frame to compute SAD
 *
 * \return sse
 */
//-----------------------------------------------------------------------------
double WFrame::sse(WFrame *f)
{
  int i;
  int diff;
  double sum = 0.0;
  
  for (i=0; i<m_size; i++) 
  {
    diff = m_data[i] - f->m_data[i];
    sum += (double)(diff * diff);
  }
  
  return sum;
}

//-----------------------------------------------------------------------------
/*!
 * \brief Convert from Frame class
 *
 * \param[in] f %Frame to convert from
 */
//-----------------------------------------------------------------------------
void WFrame::expand(Frame *f)
{
  int i;
  
  for (i=0; i<m_size; i++)
  {
    m_data[i] = (f->data[i] - MID_SAMPLE_VAL) << SAMPLE_ADD_PREC;
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Convert to Frame class
 *
 * Converts 16-bit samples to 8-bit samples and clips them to the
 * [ZERO,MAX_SAMPLE_VAL] range.
 *
 * \param[in] f %Frame to convert to
 */
//-----------------------------------------------------------------------------
void WFrame::collapse(Frame *f)
{
  int i;
  short *src = m_data;
  unsigned char *dst = f->data;
  
  if (SAMPLE_ADD_PREC == ZERO) 
  {
    for (i=0; i<m_size; i++)
    {
      dst[i] = clip(MID_SAMPLE_VAL + src[i], ZERO, MAX_SAMPLE_VAL);
    }
  }
  else 
  {
    for (i=0; i<m_size; i++)
    {
      dst[i] = clip(MID_SAMPLE_VAL + shift_right_round(src[i], SAMPLE_ADD_PREC), ZERO, MAX_SAMPLE_VAL);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Convert sign to Frame class
 *
 * Converts 16-bit samples to 8-bit samples and clips them to the
 * [ZERO,MAX_SAMPLE_VAL] range.
 *
 * \param[in] f %Frame to convert to
 */
//-----------------------------------------------------------------------------
void WFrame::collapse_sign(Frame *f)
{
  int i;
  short *src = m_data;
  unsigned char *dst = f->data;
  
  for (i = 0; i < m_size; i++)
  {
    dst[i] = 2 + ((src[i] >> 31) & -2);
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Convert modulus to Frame class
 *
 * Converts 16-bit samples to 8-bit samples and clips them to the
 * [ZERO,MAX_SAMPLE_VAL] range.
 *
 * \param[in] f %Frame to convert to
 */
//-----------------------------------------------------------------------------
void WFrame::collapse_modulus(Frame *f)
{
  int i;
  short *src = m_data;
  unsigned char *dst = f->data;
  
  if (SAMPLE_ADD_PREC == ZERO) 
  {
    for (i = 0; i < m_size; i++)
    {
      dst[i] = imin(abs(src[i]), MAX_SAMPLE_VAL);
    }
  }
  else 
  {
    for (i = 0; i < m_size; i++)
    {
      dst[i] = imin(abs(shift_right_round(src[i], SAMPLE_ADD_PREC)), MAX_SAMPLE_VAL);
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Retrieve luma block (block-aligned)
 *
 * \param[out] blk block
 * \param[in] i Vertical position (in block-height units)
 * \param[in] j Horizontal position (in block-width units)
 */
//-----------------------------------------------------------------------------
void WFrame::get_blk_luma(Block *blk, int i, int j)
{
  int k, l;
  int gridx = blk->get_width();
  int gridy = blk->get_height();
  int blk_stride = blk->get_stride();
  
  i *= gridy;
  j *= gridx;
  
  short *src = get_luma_ptr(i, j);
  
  for (k = 0; k < imin(gridy, height - i); k++)
  {
    for (l = 0; l < imin(gridx, width - j); l++) 
    {
      blk->val[blk_stride * k + l] = src[width * k + l];
    }
  }
}
//-----------------------------------------------------------------------------
/*!
 * \brief Store MxN luma block (block-aligned)
 *
 * \param[in] sub-block (dimensions are specified in width+height)
 * \param[in] i Vertical position (in block units x height)
 * \param[in] j Horizontal position (in block units x width)
 */
//-----------------------------------------------------------------------------
void WFrame::put_blk_luma(Block *blk, int i, int j)
{
  int k,l;
  short *p;
  int gridx = blk->get_width();
  int gridy = blk->get_height();
  
  p = get_luma_ptr(gridy * i, gridx * j);
  
  for (k = ZERO; k < min(gridy, height - gridy * i); k++)
  {
    for (l = ZERO; l < min(gridx, width - gridx * j); l++)
    {
      p[width * k + l] = blk->val[blk->get_stride() * k + l];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Retrieve chroma block (block-aligned)
 *
 * \param[out] blk sub-block (dimensions are specified in width+height)
 * \param[in] i Vertical position (in block units x height)
 * \param[in] j Horizontal position (in block units x width)
 */
//-----------------------------------------------------------------------------
void WFrame::get_blk_chroma(Block *blk, int i, int j, int c)
{
  int k,l;
  int gridx = blk->get_width();
  int gridy = blk->get_height();
  
  short *src = get_chroma_ptr(gridy*i, gridx*j, c);
  
  for (k = 0; k < min(gridy, height2 - gridy * i); k ++)
  {
    for (l = 0; l < min(gridx, width2 - gridx * j); l ++) 
    {
      blk->val[blk->get_stride() * k + l] = src[width2 * k + l];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Store chroma block (block-aligned)
 *
 * \param[out] blk sub-block (dimensions are specified in width+height)
 * \param[in] i Vertical position (in block units x height)
 * \param[in] j Horizontal position (in block units x width)
 */
//-----------------------------------------------------------------------------
void WFrame::put_blk_chroma(Block *blk, int i, int j, int c)
{
  int k,l;
  int gridx = blk->get_width();
  int gridy = blk->get_height();
  short *dst = get_chroma_ptr(gridy*i, gridx*j, c);
  
  for (k = 0; k < min(gridy, height2 - gridy * i); k ++)
  {
    for (l = 0; l < min(gridx, width2 - gridx * j); l ++) 
    {
      dst[width2 * k + l] = blk->val[blk->get_stride() * k + l];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Store motion compensation block
 *
 * \param[in] i Vertical position (8x8 luma block units)
 * \param[in] j Horizontal position (8x8 luma block units)
 * \param[in] lu Array of luma samples
 * \param[in] ch Array of chroma samples
 */
//-----------------------------------------------------------------------------
void WFrame::put_mc_block(int i, int j, short lu[64], short ch[2][16])
{
  int k, l;
  short *p;
  int uv;
  
  p = get_luma_ptr(MOTION_BLK_SIZE*i, MOTION_BLK_SIZE*j);
  
  for (k=0; k<8; k++)
  {
    for (l=0; l<8; l++)
    {
      p[width*k+l] = lu[8*k+l];
    }
  }
  
  for (uv=0; uv<2; uv++) 
  {
    p = get_chroma_ptr((MOTION_BLK_SIZE/2)*i, (MOTION_BLK_SIZE/2)*j, uv);
    
    for (k=0; k<4; k++)
    {
      for (l=0; l<4; l++)
      {
        p[width2*k+l] = ch[uv][4*k+l];
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
 * \brief Store motion compensation block with motion sharing
 *
 * \param[in] i Vertical position (8x8 luma block units)
 * \param[in] j Horizontal position (8x8 luma block units)
 * \param[in] lu Array of luma samples
 * \param[in] ch Array of chroma samples
 * \param[in] shift_y Vertical component of partition size for motion sharing
 * \param[in] shift_y Horizontal component of partition size for motion sharing 
 */
//-----------------------------------------------------------------------------
void WFrame::put_mc_block_mvshare(int i, int j, short lu[64], short ch[2][16], int shift_y, int shift_x)
{
  
  int k, l;
  short *p;
  int uv;
  
  if(shift_y>0)
  {
    p = get_luma_ptr(MOTION_BLK_SIZE*i, MOTION_BLK_SIZE*j);
    
    for (k=0; k<shift_y; k++)
    {
      for (l=0; l<8; l++)
      {
        p[width*k+l] = lu[8*k+l];// 4x8
      }
    }
    
    for (uv=0; uv<2; uv++) 
    {
      p = get_chroma_ptr((MOTION_BLK_SIZE/2)*i, (MOTION_BLK_SIZE/2)*j, uv);
      
      for (k=0; k<shift_y/2; k++)
      {
        for (l=0; l<4; l++)
        {
          p[width2*k+l] = ch[uv][4*k+l];  // 2x4
        }
      }
    }
  }
  else if(shift_x>0)
  {
    p = get_luma_ptr(MOTION_BLK_SIZE*i, MOTION_BLK_SIZE*j);
    
    for (k=0; k<8; k++)
    {
      for (l=0; l<shift_x; l++)
      {
        p[width*k+l] = lu[8*k+l];  // 1x4
      }
    }
    
    for (uv=0; uv<2; uv++) 
    {
      p = get_chroma_ptr((MOTION_BLK_SIZE/2)*i, (MOTION_BLK_SIZE/2)*j, uv);
      
      for (k=0; k<4; k++)
      {
        for (l=0; l<shift_x/2; l++)
        {
          p[width2*k+l] = ch[uv][4*k+l]; // 1x2
        }
      }
    }
  }  
}

//-----------------------------------------------------------------------------
/*!
 * \brief Store motion compensation chroma block
 *
 * \param[in] i Vertical position (8x8 luma block units)
 * \param[in] j Horizontal position (8x8 luma block units)
 * \param[in] ch Array of chroma samples
 */
//-----------------------------------------------------------------------------
void WFrame::put_mc_block_chroma(int i, int j, short ch[2][16])
{
  int k, l;
  short *p;
  int uv;
  
  for (uv=0; uv<2; uv++) 
  {
    p = get_chroma_ptr((MOTION_BLK_SIZE/2)*i, (MOTION_BLK_SIZE/2)*j, uv);
    
    for (k=0; k<4; k++)
    {
      for (l=0; l<4; l++)
      {
        p[width2*k+l] = ch[uv][4*k+l];
      }
    }
  }
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
