/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file WienerFilter.cpp
*
* \brief Implementation of WienerFilter class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Global.H"
#include "EntropyEncoder.H"
#include "Buffer.H"
#include "WienerFilter.H"

#ifdef WIENER_MT
#include <pthread.h>
#endif

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* \param[in] h height of frame in 4x4 luma block units
* \param[in] w width of frame in 4x4 luma block units
*/
//-----------------------------------------------------------------------------
WienerFilter::WienerFilter(int hh, int ww, Param *param)
{
  int i;

  int qtable [FILTER_PRECISION_TABLE_NUMBER]={8,16,32,64,128,256,512,1024,2048,4096,8192,16384};
  int qtable2[FILTER_PRECISION_TABLE_NUMBER]={3, 4, 5, 6,  7,  8,  9,  10,  11,  12,  13,   14};
  int qtable3[FILTER_PRECISION_TABLE_NUMBER]={4, 8,16,32, 64,128,256, 512,1024,2048,4096, 8192};
  
  for (i=0;i<FILTER_PRECISION_TABLE_NUMBER;i++)
  {
    filter_precision_table      [i]=qtable [i];
    filter_precision_table_shift[i]=qtable2[i];
    filter_precision_table_half [i]=qtable3[i];
  }
  
  height  = hh;
  width   = ww;
  height2 = hh>>1;
  width2  = ww>>1;
  
  fs_max_rec   = (*param).fs_max_rec;
  fs_max_qpe   = (*param).fs_max_qpe;
  fs_max_pred  = (*param).fs_max_pred;
  
  for (i = 0; i < 3; i++)
  {
    wlf_flag[i] = 0;
    filtertype[i] = 0; // Coefficients of a 2D FIR filter
    filter_length_RD_rec_x [i]=1;
    filter_length_RD_rec_y [i]=1;
    filter_length_RD_pred_x[i]=1;
    filter_length_RD_pred_y[i]=1;
    filter_length_RD_qpe_x [i]=1;
    filter_length_RD_qpe_y [i]=1;
  }

  rd_wlf_eenc = new EntropyEncoder;
  rd_wlf_buff = new Buffer;
  rd_wlf_eenc->reset(rd_wlf_buff);
  get_mem2Dint   (&coeffs      , 3, 1+3*MAX_WIENER_FILTER_LENGTH*MAX_WIENER_FILTER_LENGTH);
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
WienerFilter::~WienerFilter() 
{

  delete  rd_wlf_buff;
  delete  rd_wlf_eenc;  

  free_mem2Dint   (coeffs);
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Sets the Lagrangian factor
*
* \param[in] set_lambda %Lagrangian factor
*/
//-----------------------------------------------------------------------------
void WienerFilter::set_lambda(double set_lambda)
{
  lambda=set_lambda;
}


//-----------------------------------------------------------------------------
/*!
* \brief Calculation of autocorrelation function
*
* \param[in] on                   % horizontal offset in the autocorrelation matrix
* \param[in] om                   % vertical offset in the autocorrelation matrix
* \param[in] width_f              % width of image data
* \param[in] height_f             % height of image data
* \param[in] tap1                 % filter size 
* \param[in] tap2                 % filter size 
* \param[in] tap_minus_one_half1  % one half of filter size minus one
* \param[in] tap_minus_one_half2  % one half of filter size minus one
* \param[in] *im1                 % pointer to frist image data
* \param[in] *im2                 % pointer to second image data
*
*/
//-----------------------------------------------------------------------------
template <typename T, typename U>
void WienerFilter::calc_auto_correlation(int on, int om, int tap1, int tap2, int tap_minus_one_half1, int tap_minus_one_half2, const Plane<T> &im1, const Plane<U> &im2)
{
  int m,n,i,j,o,p;
  int taptap1 = tap1 * tap1;
  int taptap2 = tap2 * tap2;
  double *accum = new double[taptap1 * taptap2];
  double *vec2 = new double[taptap2];
  int height_f = im1.get_height();
  int width_f = im1.get_width();

  double val2 = (double)(width_f*height_f);

  for (i = 0; i < taptap1 * taptap2; i++)
  {
    accum[i] = 0.0;
  }
  for (j = 0; j < height_f; j++) 
  {    
    for (i = 0; i < width_f; i++)
    {
      int mx, my;
      for (my = 0; my < tap2; my++)
      {
        const U *im2_row = im2[j - tap_minus_one_half2 + my] + (i - tap_minus_one_half2);
        for (mx = 0; mx < tap2; mx++)
        {
          vec2[tap2 * my + mx] = (double)im2_row[mx];
        }
      }

      int nx, ny;
      double *accum_p = accum;
      for (ny = 0; ny < tap1; ny++)
      {
        const T *im1_row = im1[j - tap_minus_one_half1 + ny] + (i - tap_minus_one_half1);
        for (nx = 0; nx < tap1; nx++)
        {
          double val = (double)im1_row[nx];
          for (m = 0; m < taptap2; m++)
          {
            *accum_p++ += val * vec2[m];
          }
        }
      }
    }
  }

  for (n = 0; n < taptap1; n++)
  {
    o = n + on;
    for (m = 0; m < taptap2; m++)
    {
      double val=accum[n * taptap2 + m];
      p = m + om;
      
      a[o][p] += val;
      a[p][o] += val;

      an[o][p] += val2;
      an[p][o] += val2;
    }
  }

  delete[] vec2;
  delete[] accum;
}

//-----------------------------------------------------------------------------
/*!
* \brief Calculation of crosscorrelation function
*
* \param[in] on                   % offset in the crosscorrelation vactor
* \param[in] width_f              % width of image data
* \param[in] height_f             % height of image data
* \param[in] tap                  % filter size 
* \param[in] tap_minus_one_half   % one half of filter size minus one
* \param[in] *xY                  % pointer to frist image
* \param[in] *dY                  % pointer to second image
*
*/
//-----------------------------------------------------------------------------
template <typename T>
void WienerFilter::calc_cross_correlation(int on, int tap, int tap_minus_one_half, const Plane<T> &xY, unsigned char *dY)
{
  int m,n,i,j;
  int width_f = xY.get_width();
  int height_f = xY.get_height();
  unsigned char *p_dY = dY;

  for (j = 0; j < height_f; j++) 
  {
    for (i = 0; i < width_f; i++)
    {
      double org = (double)(*p_dY++);
      const T *ptr = xY[j - tap_minus_one_half] + (i - tap_minus_one_half);
      int pitch = xY.get_pitch();
      double *rho_p = &rho[on];

      for (n = 0; n < tap; n++) 
      {
        for (m = 0; m < tap; m++) 
        {
          *rho_p++ += org * ptr[n * pitch + m];
        }
      }
    }
  }      
  
  for (n = 0; n < tap * tap; n++)
  {
    rho_n[on + n] += (double)(width_f * height_f);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Calculation of mean
*
* \param[in] on                   % offset in the crosscorrelation vactor
* \param[in] width_f              % width of image data
* \param[in] height_f             % height of image data
* \param[in] *dY                  % pointer to second image
*
*/
//-----------------------------------------------------------------------------
void WienerFilter::calc_mean_cross (int on, int width_f, int height_f, unsigned char *dY)
{
  int j;
  int tmp=height_f*width_f;
  unsigned char *p_uc=dY;
  double *p_d=&(rho[on]);
  
  for (j = 0; j < tmp; j++) 
  {
    *p_d+=(double)(*(p_uc++));
  }  
    
  rho_n[on]+=(double)(tmp);
}

//-----------------------------------------------------------------------------
/*!
* \brief Calculation of mean
*
* \param[in] on                   % horizontal offset in the autocorrelation matrix
* \param[in] om                   % vertical offset in the autocorrelation matrix
* \param[in] width_f              % width of image data
* \param[in] height_f             % height of image data
* \param[in] tap                  % filter size 
* \param[in] tap_minus_one_half   % one half of filter size minus one
* \param[in] *im                  % pointer to image data
*
*/
//-----------------------------------------------------------------------------
template <typename T>
void WienerFilter::calc_mean_auto(int on, int om, int tap, int tap_minus_one_half, const Plane<T> &im)
{
  int i,j;
  int x,y;
  int width_f = im.get_width();
  int height_f = im.get_height();
  double val2 = (double)(width_f*height_f);
  int o = on;

  for (j = 0; j < tap; j++)
  {
    for (i = 0; i < tap; i++)
    {
      double mean = 0.0;

      for (y = 0; y < height_f; y++)
      {
        const T *ptr = im[y + j - tap_minus_one_half] + (i - tap_minus_one_half);
        int sum = 0;

        for (x = 0; x < width_f; x++)
        {
          sum += ptr[x];
        }

        mean += (double)sum;
      }

      a  [   o][  om] += mean;
      a  [  om][   o] += mean;
      an [   o][  om] += val2;
      an [  om][   o] += val2;
      o++;
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Setting certain lines and columns of the autocorrelation matrix to zero
*
* \param[in] tap                  % filter size max
* \param[in] tap_minus_one_half   % one half of filter size max minus one
* \param[in] tap_check            % filter size to be realized
* \param[in] on                   % horizontal and vertical offset in the autocorrelation matrix
*
*/
//-----------------------------------------------------------------------------
void WienerFilter::set_lines_and_columns_of_auto_to_zero(int tap, int tap_minus_one_half, int tap_check, int on)
{
  int k,l;
  int taptap=tap*tap;
  int new_x,new_y;  
  int length=fs_max_rec*fs_max_rec+fs_max_pred*fs_max_pred+fs_max_qpe*fs_max_qpe+1;
  int tmp=(tap_check-1)>>1;
  int tmp1;
  
  for (k = 0; k < taptap; k++)
  {
    new_x =k%tap-tap_minus_one_half;
    new_y =k/tap-tap_minus_one_half;
    tmp1=k+on;
    if (new_x<-tmp || new_x>tmp || new_y<-tmp || new_y>tmp || tap_check==0)
    {
      for (l = 0; l < length; l++)
      {
        a  [tmp1][l]=0.0;
        a  [l][tmp1]=0.0;
      }
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Estimation of a non separable 2D Wiener Filter of equal filter size in horizontal and vertical direction
*
* \param[in] *dY                  % pointer to desired image signal
* \param[in] *xY                  % pointer to reconstructed image signal
* \param[in] *pY                  % pointer to prediction image signal
* \param[in] *qpe_m               % pointer to modulus of quantized prediction error signal
* \param[in] *qpe_s               % pointer to sign of quantized prediction error signal
* \param[in] *eenc                % pointer to entropy encoder
* \param[in] component            % color component
* \param[in] height_f             % height of image data
* \param[in] width_f              % width of image data
*
* \return RD-Costs
*/
//-----------------------------------------------------------------------------
double WienerFilter::estimate_filter_2D_equal_size(unsigned char *dY, unsigned char *xY, unsigned char *pY, unsigned char *qpe_m, unsigned char *qpe_s, EntropyEncoder *eenc,int component)
{
  double rd_costs=9999999.0;    
  double rd_cost_precision=0.0;
  double rd_cost_min_precision=0.0;
  int    ii;
  int    prec_best=0;
  int    precision_best=0;
  int    o;
  int    i, j, m, n, l, k, k1, s, count;
  int    new_x, new_y;
  int    tap, tap_rec, tap_pred, tap_qpe;
  int    rate=0;
  int    rate_wlf_off=0;
  double rounding=0.0;
  double mse=0.0;
  double mse_d=0.0;
  double mse_wlf_off=0.0;
  double rd_cost=0.0;
  double rd_cost_min=0.0;
  double rd_cost_wlf_off=0.0;
  
  double *p_d=NULL;
  int    *p_i=NULL;  
  int filtersize_rec_best =1;
  int filtersize_pred_best=1;
  int filtersize_qpe_best =1;
  int TAP_rec_1_2 =(fs_max_rec -1)>>1;
  int TAP_pred_1_2=(fs_max_pred-1)>>1;
  int TAP_qpe_1_2 =(fs_max_qpe -1)>>1;
  int o_rec =0;
  int o_pred=fs_max_rec*fs_max_rec;
  int o_qpe =o_pred+fs_max_pred*fs_max_pred;
  int length=o_qpe+fs_max_qpe*fs_max_qpe+1;

  int height_f, width_f;
  int tmp;
  int *p_int2;
  
  if (component==0)
  {
    height_f=height;
    width_f=width;
  }
  else
  {
    height_f=height2;
    width_f=width2;
  }
  
// Get memory
  image = new unsigned char[width_f*height_f];

  get_mem1Ddouble(&rho         , length);
  get_mem1Ddouble(&rho_n       , length);
  get_mem1Ddouble(&rho_RD      , length);
  get_mem2Ddouble(&a           , length,length);
  get_mem2Ddouble(&a_1         , length, length);
  get_mem2Ddouble(&an          , length,length);
  get_mem2Ddouble(&a_RD        , length,length);
  get_mem1Ddouble(&h           , length);
  get_mem1Ddouble(&h1          , length);
  get_mem1Ddouble(&r1          , length);
  get_mem1Dint   (&null        , length);
  get_mem1Dint   (&coeffs_save           , 1+3*MAX_WIENER_FILTER_LENGTH*MAX_WIENER_FILTER_LENGTH);
  get_mem1Dint   (&coeffs_best           , 1+3*MAX_WIENER_FILTER_LENGTH*MAX_WIENER_FILTER_LENGTH);
  get_mem1Dint   (&coeffs_best_precision , 1+3*MAX_WIENER_FILTER_LENGTH*MAX_WIENER_FILTER_LENGTH);
  
  Plane<unsigned char> p_xY(width_f, height_f, (MAX_WIENER_FILTER_LENGTH - 1) / 2);
  Plane<unsigned char> p_pY(width_f, height_f, (MAX_WIENER_FILTER_LENGTH - 1) / 2);
  Plane<short> p_qpe(width_f, height_f, (MAX_WIENER_FILTER_LENGTH - 1) / 2);

  for (i = 0; i < height_f; i++)
  {
    for (j = 0; j < width_f; j++)
    {
      p_xY[i][j] = xY[width_f * i + j];
      p_pY[i][j] = pY[width_f * i + j];
      p_qpe[i][j] = qpe_m[width_f * i + j] * (qpe_s[width_f * i + j] - 1);
    }
  }

  p_xY.mirror();
  p_pY.mirror();
  p_qpe.mirror();

  for (i = 0; i < length; i++)
  {
    rho   [i] = rho_n [i] = 0.0;
    for (j = 0; j < length; j++)
    {
      a  [i][j] = an [i][j] = 0.0;
    }
  }
  
//Calculate Cross Correlation  
  calc_cross_correlation(o_rec   , fs_max_rec , TAP_rec_1_2 , p_xY , dY);
  calc_cross_correlation(o_pred  , fs_max_pred, TAP_pred_1_2, p_pY , dY);
  calc_cross_correlation(o_qpe   , fs_max_qpe , TAP_qpe_1_2 , p_qpe, dY);
  calc_mean_cross        (length-1, width_f, height_f, dY);
        
//Calculate Auto Correlation
  calc_auto_correlation(o_rec , o_rec  , fs_max_rec , fs_max_rec , TAP_rec_1_2 , TAP_rec_1_2 , p_xY, p_xY);
  calc_auto_correlation(o_rec , o_pred , fs_max_rec , fs_max_pred, TAP_rec_1_2 , TAP_pred_1_2, p_xY, p_pY);
  calc_auto_correlation(o_rec , o_qpe  , fs_max_rec , fs_max_qpe , TAP_rec_1_2 , TAP_qpe_1_2 , p_xY, p_qpe);
  calc_auto_correlation(o_pred, o_pred , fs_max_pred, fs_max_pred, TAP_pred_1_2, TAP_pred_1_2, p_pY, p_pY);
  calc_auto_correlation(o_pred, o_qpe  , fs_max_pred, fs_max_qpe , TAP_pred_1_2, TAP_qpe_1_2 , p_pY, p_qpe);
  calc_auto_correlation(o_qpe , o_qpe  , fs_max_qpe , fs_max_qpe , TAP_qpe_1_2 , TAP_qpe_1_2 , p_qpe, p_qpe);

//Calculate Means
  calc_mean_auto(o_rec , length-1, fs_max_rec , TAP_rec_1_2 , p_xY);
  calc_mean_auto(o_pred, length-1, fs_max_pred, TAP_pred_1_2, p_pY);
  calc_mean_auto(o_qpe , length-1, fs_max_qpe , TAP_qpe_1_2 , p_qpe);

  a [length-1][length-1]=1.0;
  an[length-1][length-1]=1.0;
  
    
  for (i = 0; i < length; i++) 
  {
    rho [i]/=rho_n[i];
  }

  for (n=0;n<length;n++) 
  {
    for (m=0;m<length;m++) 
    {
      a [n][m]/=an [n][m];
    }
  }

  for (n=0;n<length;n++) 
  {
    memcpy(a_RD[n],a[n],length*sizeof(double));
  }
  memcpy(rho_RD,rho,length*sizeof(double));
  
  //get_rate for Wiener Filter turned off
  rd_wlf_eenc->copy_coding_state(eenc);
  rate_wlf_off =rd_wlf_eenc->get_nbits();

  //get_mse for Wiener Filter turned off
  mse_wlf_off = get_mse(dY,xY,height_f*width_f);
  
  //calc RD costs for Wiener Filter turned off
  rd_cost_wlf_off= mse_wlf_off + lambda * (double)(rate_wlf_off)/(double)(height_f*width_f);
  rd_cost_min=rd_cost_wlf_off;
  filtersize_rec_best  = 0;
  filtersize_pred_best = 0;
  filtersize_qpe_best  = 0;

  wlf_flag[component]=0; //turn off
  
  for (tap=0 ; tap<(TAP_rec_1_2+1)*(TAP_pred_1_2+1)*(TAP_qpe_1_2+1) ; tap++)
  {
    tap_rec   = tap%(TAP_rec_1_2+1);
    tap_pred  = (tap/(TAP_rec_1_2+1))%(TAP_pred_1_2+1);
    tap_qpe   = (tap/((TAP_rec_1_2+1)*(TAP_pred_1_2+1)))%(TAP_qpe_1_2+1);

    tap_rec   = tap_rec  == 0 ? 1 : ((tap_rec -1)<<1)+3;
    tap_pred  = tap_pred == 0 ? 1 : ((tap_pred-1)<<1)+3;
    tap_qpe   = tap_qpe  == 0 ? 1 : ((tap_qpe -1)<<1)+3;
    
    for (n=0;n<length;n++) 
    {
      memcpy(a[n],a_RD[n],length*sizeof(double));
    }
    memcpy(rho,rho_RD,length*sizeof(double));
    
    set_lines_and_columns_of_auto_to_zero(fs_max_rec , TAP_rec_1_2 , tap_rec , o_rec );
    set_lines_and_columns_of_auto_to_zero(fs_max_pred, TAP_pred_1_2, tap_pred, o_pred);
    set_lines_and_columns_of_auto_to_zero(fs_max_qpe , TAP_qpe_1_2 , tap_qpe , o_qpe );
    
    memset(null,1000,length*sizeof(int));

    //Take out ZERO-Lines
    count=0;
    for (k = 0; k < length; k++)
    {
      l=0;
      s=0;
      while(l< length)
      {
        if(a[k][l]>0.000001 || a[k][l]<-0.000001)
        {
          s=1;
          break;
          l= length;
        }
        else
        {
          l++;
        }
      }

      if (s==0)
      {
        null[count]=k;
        count++;
      }
      else 
      {
        memcpy(a_1[k-count],a[k],length*sizeof(double));
        r1[k-count]=rho[k];
      }
    }

    //Take out ZERO-Columns
    k1=0;
    for(l = 0; l < length; l++)
    {
      if(l==null[k1]) 
      {
        k1++;
      }
      else
      {
        for(k = 0; k < length; k++)
        {
          a_1[k][l-k1]=a_1[k][l];
        }
      }
    }
    
    
    if (length-count>0)
    {
      get_mem2Ddouble(&a_final  , length-count, length-count);
      get_mem1Ddouble(&r_final  , length-count);
      get_mem1Ddouble(&h_final  , length-count);

      for(k = 0; k < length-count; k++)
      {
        memcpy(a_final[k],a_1[k],(length-count)*sizeof(double));
      }
      memcpy(r_final,r1,(length-count)*sizeof(double));

      
      sgaus(a_final, h_final, r_final, length-count);

      free_mem1Ddouble(r_final);
      free_mem2Ddouble(a_final);
    }

    k1=0;
    p_d=&(h1[0]);
    for(k = 0; k < length; k++)
    {
      if(k==null[k1]) //This needs to be a ZERO-Column
      {
        *(p_d++)=0.0;
        k1++;
      }
      else 
      {
        *(p_d++)=h_final[k-k1];
      }
    }

    if (length-count>0)
    {
      free_mem1Ddouble(h_final);
    }

    rd_cost_min_precision=9999999.0;
    for (ii=0;ii<FILTER_PRECISION_TABLE_NUMBER;ii++)
    {      
      //Quantize filter coefficients
      p_d=&(h1[0]);
      p_i=&(coeffs[component][0]);
      for (m = 0; m < length; m++) 
      {
        if (*p_d>0.0)
        {
          *(p_i++) =  (int) ( *(p_d++) * (double)(filter_precision_table[ii]) + 0.5);
        }
        else
        {
          *(p_i++) = -(int) (-*(p_d++) * (double)(filter_precision_table[ii]) + 0.5);
        }
      }

      memcpy(coeffs_save, coeffs[component],  length * sizeof(int));

      k1=0;
      p_int2=&(coeffs[component][0]);    
      
      tmp=(tap_rec-1)>>1;  
      for (k = 0; k < fs_max_rec*fs_max_rec; k++)
      {
        new_x =k%fs_max_rec-TAP_rec_1_2;
        new_y =k/fs_max_rec-TAP_rec_1_2;
        if (!(new_x<-tmp || new_x>tmp || new_y<-tmp || new_y>tmp || tap_rec==0))
        {
          *(p_int2++)=coeffs_save[k];
        }
      }
      
      tmp=(tap_pred-1)>>1;
      for (k = 0; k < fs_max_pred*fs_max_pred; k++)
      {
        new_x =k%fs_max_pred-TAP_pred_1_2;
        new_y =k/fs_max_pred-TAP_pred_1_2;
        if (!(new_x<-tmp || new_x>tmp || new_y<-tmp || new_y>tmp || tap_pred==0))
        {
          *(p_int2++)=coeffs_save[k+fs_max_rec*fs_max_rec];
        }
      }
      
      tmp=(tap_qpe-1)>>1;
      for (k = 0; k < fs_max_qpe*fs_max_qpe; k++)
      {
        new_x =k%fs_max_qpe-TAP_qpe_1_2;
        new_y =k/fs_max_qpe-TAP_qpe_1_2;
        if (!(new_x<-tmp || new_x>tmp || new_y<-tmp || new_y>tmp || tap_qpe==0))
        {
          *(p_int2++)=coeffs_save[k+fs_max_rec*fs_max_rec+fs_max_pred*fs_max_pred];
        }
      }
      
      *(p_int2++)=coeffs_save[fs_max_rec*fs_max_rec+fs_max_pred*fs_max_pred+fs_max_qpe*fs_max_qpe];
      
      memset(p_int2,0,(length-k1)*sizeof(double));
  
      filter_length_RD_rec_x [component]  = tap_rec;
      filter_length_RD_rec_y [component]  = tap_rec;
      filter_length_RD_pred_x[component]  = tap_pred;
      filter_length_RD_pred_y[component]  = tap_pred;
      filter_length_RD_qpe_x [component]  = tap_qpe;
      filter_length_RD_qpe_y [component]  = tap_qpe;
      
      filter_precision[component]=ii;
      
      filter_2D_equal_size(p_xY, image, p_pY, p_qpe, component);

      rd_wlf_eenc->copy_coding_state(eenc);
      rd_wlf_eenc->encode_wlf_coeffs(coeffs,filter_length_RD_rec_x,filter_length_RD_rec_y, filter_length_RD_pred_x, filter_length_RD_pred_y,filter_length_RD_qpe_x,filter_length_RD_qpe_y,component,filter_precision, filter_precision_table);

      rate =rd_wlf_eenc->get_nbits();

      mse =get_mse(dY,image,height_f*width_f);
      
      rd_cost_precision= mse + lambda * (double)(rate)/(double)(height_f*width_f);        
      
      if (rd_cost_precision < rd_cost_min_precision)
      {
        precision_best=ii;
        rd_cost_min_precision =rd_cost_precision;
        memcpy(coeffs_best_precision, coeffs[component], length * sizeof(int));
      }
    }
    
    filter_precision[component]=precision_best;
    
    memcpy(coeffs[component], coeffs_best_precision,  length * sizeof(int));
    
    k1=tap_rec*tap_rec+tap_pred*tap_pred+tap_qpe*tap_qpe+1;
    
    for (o=0;o<10;o++)
    {
      mse_d = filter_2D_equal_size_d(&rounding, dY, p_xY, image, p_pY, p_qpe, component);
      int sign = (rounding > 0.0) ? 1 : -1;
      int delta = (int)(sign * rounding * 0.25 * filter_precision_table[precision_best] + 0.5);
      if (delta == 0)
        break; // Not modifying coeff, so there is no point in re-testing the same coeffs over and over again
      coeffs[component][k1 - 1] -= sign * delta;
    }

    filter_2D_equal_size(p_xY, image, p_pY, p_qpe, component);

    //get_rate
    rd_wlf_eenc->copy_coding_state(eenc);
    rd_wlf_eenc->encode_wlf_coeffs(coeffs,filter_length_RD_rec_x,filter_length_RD_rec_y, filter_length_RD_pred_x, filter_length_RD_pred_y,filter_length_RD_qpe_x,filter_length_RD_qpe_y,component, filter_precision, filter_precision_table);
    rate =rd_wlf_eenc->get_nbits();
    //get_mse
    mse =get_mse(dY,image,height_f*width_f);
    
    //calc RD costs
    rd_cost= mse + lambda * (double)(rate)/(double)(height_f*width_f);
    
    if (rd_cost < rd_cost_min)
    {
      filtersize_rec_best  = tap_rec;
      filtersize_pred_best = tap_pred;
      filtersize_qpe_best  = tap_qpe;
      prec_best            = filter_precision[component];
      rd_cost_min          = rd_cost;
      wlf_flag[component]  = 1; //turn on
      memcpy(coeffs_best, coeffs[component], length * sizeof(int));
    }
    
    memcpy(coeffs[component], coeffs_best,  length * sizeof(int));
  }

  memcpy(coeffs[component], coeffs_best,  length * sizeof(int));

  filter_length_RD_rec_x [component]  = filtersize_rec_best;
  filter_length_RD_rec_y [component]  = filtersize_rec_best;
  filter_length_RD_pred_x[component]  = filtersize_pred_best;
  filter_length_RD_pred_y[component]  = filtersize_pred_best;
  filter_length_RD_qpe_x [component]  = filtersize_qpe_best;
  filter_length_RD_qpe_y [component]  = filtersize_qpe_best;
  filter_precision[component]=prec_best;
  
  delete[] image;
  free_mem1Ddouble(h);
  free_mem1Ddouble(rho);
  free_mem1Ddouble(rho_n);
  free_mem1Ddouble(rho_RD);
  free_mem2Ddouble(a);
  free_mem2Ddouble(an);
  free_mem2Ddouble(a_RD);
  
  free_mem1Ddouble(r1);
  free_mem1Ddouble(h1);
  free_mem1Dint   (null);
  free_mem2Ddouble(a_1);
  free_mem1Dint   (coeffs_save);
  free_mem1Dint   (coeffs_best);
  free_mem1Dint   (coeffs_best_precision);

  return rd_costs;
}

//-----------------------------------------------------------------------------
/*!
* \brief Non separable 2D Wiener Filter of equal filter size in horizontal and vertical direction
*
* \param[in] *xY                  % pointer to reconstructed image signal
* \param[in] *oY                  % pointer to output image signal
* \param[in] *pY                  % pointer to prediction image signal
* \param[in] *qpe_m               % pointer to modulus of quantized prediction error signal
* \param[in] *qpe_s               % pointer to sign of quantized prediction error signal
* \param[in] component            % color component
* \param[in] height_f             % height of image data
* \param[in] width_f              % width of image data
*
* \param[out] Rounding eror 
*
*/
//-----------------------------------------------------------------------------
double WienerFilter::filter_2D_equal_size_d(double *rounding, unsigned char *org, const Plane<unsigned char> &xY, unsigned char *oY, const Plane<unsigned char> &pY, const Plane<short> &qpe, int component)
{
  int i, j, m, n, sum;
  int tmp_rec1  = filter_length_RD_rec_x [component];
  int tmp_pred1 = filter_length_RD_pred_x[component];
  int tmp_qpe1  = filter_length_RD_qpe_x [component];
  int tmp_rec  = (tmp_rec1  - 1) >> 1;
  int tmp_pred = (tmp_pred1 - 1) >> 1;
  int tmp_qpe  = (tmp_qpe1  - 1) >> 1;

  unsigned char *p_oY = oY;
  unsigned char *p_org = org;
  const unsigned char *uc_ptr;
  const short *s_ptr;
  int pitch;
  double mse = 0.0;
  double precision = (double)(filter_precision_table[filter_precision[component]]);
  *rounding = 0.0;

  for (j = 0; j < xY.get_height(); j++) 
  {
    for (i = 0; i < xY.get_width(); i++) 
    {
      int *co = coeffs[component];
      
      sum = 0;
      
      uc_ptr = xY[j - tmp_rec] + (i - tmp_rec);
      pitch = xY.get_pitch() - tmp_rec1;

      for (n = 0; n < tmp_rec1; n++)
      {        
        for (m = 0; m < tmp_rec; m++)
        {
          sum += (*uc_ptr++) * (*co++);
          sum += (*uc_ptr++) * (*co++);
        }
        sum += (*uc_ptr++) * (*co++);

        uc_ptr += pitch;
      }
            
      uc_ptr = pY[j - tmp_pred] + (i - tmp_pred);
      pitch = pY.get_pitch() - tmp_pred1;

      for (n = 0; n < tmp_pred1; n++)
      {
        for (m = 0; m < tmp_pred; m++)
        {
          sum += (*uc_ptr++) * (*co++);
          sum += (*uc_ptr++) * (*co++);
        }
        sum += (*uc_ptr++) * (*co++);

        uc_ptr += pitch;
      }

      s_ptr = qpe[j - tmp_qpe] + (i - tmp_qpe);
      pitch = qpe.get_pitch() - tmp_qpe1;

      for (n = 0; n < tmp_qpe1; n++) 
      {//y
        for (m = 0; m < tmp_qpe; m++)
        {//x
          sum += (*s_ptr++) * (*co++);
          sum += (*s_ptr++) * (*co++);
        }
        sum += (*s_ptr++) * (*co++);

        s_ptr += pitch;
      }
     
      sum +=(*(co++));
      
      double org_val = (double)(*p_org++);
      double avg = (double)sum / precision;

      mse += (avg - org_val) * (avg - org_val);

      *rounding += avg;

      sum  +=  filter_precision_table_half [filter_precision[component]];
      sum >>=  filter_precision_table_shift[filter_precision[component]];

      *rounding -= (double)sum;

      *p_oY++ = clip(sum, 0, 255);
    }
  }

  *rounding /= xY.get_height() * xY.get_width();
  return mse / (xY.get_height() * xY.get_width());
}

//-----------------------------------------------------------------------------
/*!
 * \brief Non separable 2D Wiener Filter of equal filter size in horizontal and vertical direction
 *
 * \param[in] xY                  % reference to reconstructed image signal
 * \param[in] oY                  % pointer to output image signal
 * \param[in] pY                  % reference to prediction image signal
 * \param[in] qpe                 % reference to quantized prediction error signal
 * \param[in] component           % color component
 */
//-----------------------------------------------------------------------------
#ifdef WIENER_MT
struct WienerThreadArgs
{
  pthread_t tid;
  WienerFilter *wf;
  const Plane<unsigned char> *xY;
  unsigned char *oY;
  const Plane<unsigned char> *pY;
  const Plane<short> *qpe;
  int component;
  int first_row;
  int nrows;
};
static void *launch_wiener(void *targ)
{
  struct WienerThreadArgs *targs = (struct WienerThreadArgs *)targ;
  targs->wf->filter_2D_equal_size_rows(*targs->xY,
                                       targs->oY,
                                       *targs->pY,
                                       *targs->qpe,
                                       targs->component,
                                       targs->first_row,
                                       targs->nrows);
  return NULL;
}
void WienerFilter::filter_2D_equal_size(const Plane<unsigned char> &xY, unsigned char *oY, const Plane<unsigned char> &pY, const Plane<short> &qpe, int component)
{
  struct WienerThreadArgs *targs = new WienerThreadArgs[param.decode_nwfthreads];
  int rows_per_thread = (xY.get_height()+param.decode_nwfthreads-1)/param.decode_nwfthreads;

  for (int t = 0; t < param.decode_nwfthreads; t++)
  {
    targs[t].wf = this;
    targs[t].xY = &xY;
    targs[t].oY = oY;
    targs[t].pY = &pY;
    targs[t].qpe = &qpe;
    targs[t].component = component;
    targs[t].first_row = t*rows_per_thread;
    if (t < param.decode_nwfthreads-1 || param.decode_nwfthreads == 1)
      targs[t].nrows = rows_per_thread;
    else
      targs[t].nrows = xY.get_height()-(targs[t-1].first_row+targs[t-1].nrows);
  }
  // launch.
  if (param.decode_nwfthreads == 1)
  {
    // direct call.
    launch_wiener(&targs[0]);
  }
  else
  {
    for (int t = 0; t < param.decode_nwfthreads; t++)
    {
      pthread_create(&targs[t].tid, NULL, launch_wiener, &targs[t]);
    }
    for (int t = 0; t < param.decode_nwfthreads; t++)
    {
      pthread_join(targs[t].tid, NULL);
    }
  }
  delete [] targs;
}

void WienerFilter::filter_2D_equal_size_rows(const Plane<unsigned char> &xY, unsigned char *oY, const Plane<unsigned char> &pY, const Plane<short> &qpe, int component, int first_row, int nrows)
#else
void WienerFilter::filter_2D_equal_size(const Plane<unsigned char> &xY, unsigned char *oY, const Plane<unsigned char> &pY, const Plane<short> &qpe, int component)
#endif
{
  int i, j, m, n, sum;
  int tmp_rec1  = filter_length_RD_rec_x [component];
  int tmp_pred1 = filter_length_RD_pred_x[component];
  int tmp_qpe1  = filter_length_RD_qpe_x [component];
  int tmp_rec  = (tmp_rec1  - 1) >> 1;
  int tmp_pred = (tmp_pred1 - 1) >> 1;
  int tmp_qpe  = (tmp_qpe1  - 1) >> 1;

#ifdef WIENER_MT
  unsigned char *p_oY = oY+first_row*xY.get_width();;
#else
  unsigned char *p_oY = oY;
#endif
  const unsigned char *uc_ptr;
  const short *s_ptr;
  int pitch;

#ifdef WIENER_MT
  for (j = first_row; j < first_row+nrows; j++) 
#else
  for (j = 0; j < xY.get_height(); j++) 
#endif
  {
    for (i = 0; i < xY.get_width(); i++) 
    {
      int *co = coeffs[component];
      
      sum = 0;
      
      uc_ptr = xY[j - tmp_rec] + (i - tmp_rec);
      pitch = xY.get_pitch() - tmp_rec1;

      for (n = 0; n < tmp_rec1; n++)
      {        
        for (m = 0; m < tmp_rec; m++)
        {
          sum += (*uc_ptr++) * (*co++);
          sum += (*uc_ptr++) * (*co++);
        }
        sum += (*uc_ptr++) * (*co++);

        uc_ptr += pitch;
      }
            
      uc_ptr = pY[j - tmp_pred] + (i - tmp_pred);
      pitch = pY.get_pitch() - tmp_pred1;

      for (n = 0; n < tmp_pred1; n++)
      {
        for (m = 0; m < tmp_pred; m++)
        {
          sum += (*uc_ptr++) * (*co++);
          sum += (*uc_ptr++) * (*co++);
        }
        sum += (*uc_ptr++) * (*co++);

        uc_ptr += pitch;
      }

      s_ptr = qpe[j - tmp_qpe] + (i - tmp_qpe);
      pitch = qpe.get_pitch() - tmp_qpe1;

      for (n = 0; n < tmp_qpe1; n++) 
      {//y
        for (m = 0; m < tmp_qpe; m++)
        {//x
          sum += (*s_ptr++) * (*co++);
          sum += (*s_ptr++) * (*co++);
        }
        sum += (*s_ptr++) * (*co++);

        s_ptr += pitch;
      }
     
      sum +=(*co++);
      
      sum  +=  filter_precision_table_half [filter_precision[component]];
      sum >>=  filter_precision_table_shift[filter_precision[component]];

      *p_oY++ = clip(sum, 0, 255);
    }
  }  
}

//-----------------------------------------------------------------------------
/*!
 * \brief Non separable 2D Wiener Filter of equal filter size in horizontal and vertical direction
 *
 * \param[in] xY                  % reference to reconstructed image signal
 * \param[in] oY                  % pointer to output image signal
 * \param[in] pY                  % reference to prediction image signal
 * \param[in] qpe_m               % reference to quantized prediction error signal (modulus)
 * \param[in] qpe_s               % reference to quantized prediction error signal (sign)
 * \param[in] component           % color component
 */
//-----------------------------------------------------------------------------
void WienerFilter::filter_2D_equal_size(unsigned char *xY, unsigned char *oY, unsigned char *pY, unsigned char *qpe_m, unsigned char *qpe_s, int component)
{
  int width_f, height_f;

  if (component == 0)
  {
    width_f = width;
    height_f = height;
  }
  else
  {
    width_f = width2;
    height_f = height2;
  }

  Plane<unsigned char> p_xY(width_f, height_f, (MAX_WIENER_FILTER_LENGTH - 1) / 2);
  Plane<unsigned char> p_pY(width_f, height_f, (MAX_WIENER_FILTER_LENGTH - 1) / 2);
  Plane<short> p_qpe(width_f, height_f, (MAX_WIENER_FILTER_LENGTH - 1) / 2);
  int i, j;

  for (i = 0; i < height_f; i++)
  {
    for (j = 0; j < width_f; j++)
    {
      p_xY[i][j] = xY[width_f * i + j];
      p_pY[i][j] = pY[width_f * i + j];
      p_qpe[i][j] = qpe_m[width_f * i + j] * (qpe_s[width_f * i + j] - 1);
    }
  }

  p_xY.mirror();
  p_pY.mirror();
  p_qpe.mirror();

  filter_2D_equal_size(p_xY, oY, p_pY, p_qpe, component);
}

//-----------------------------------------------------------------------------
/*!
* \brief Solve Wiener Hopf equations
*
* \param[in] **a                  % pointer to 2D autocorrelation matrix
* \param[in]  *x                  % pointer to 1D coefficient vector
* \param[in]  *b                  % pointer to 1D crosscorrelation vector
* \param[in]   n                  % dimension of matrix and vectors
*/
//-----------------------------------------------------------------------------
void WienerFilter::sgaus (double **a, double *x, double *b, int n)
{
  int i, j, k;
  double q, s;

  for (k = 0; k < n - 1; k++)
  {
    if (a[k][k] != 0.0)
    {
      for (i = k + 1; i < n; i++)
      {
        q = a[i][k] / a[k][k];
        for (j = k + 1; j <= i; j++)
          a[i][j] -= q * a[j][k];
        b[i] -= q * b[k];
      }
    }
  }

  if (a[n - 1][n - 1] == 0.0)   
    x[n - 1] = 0.0;
  else
    x[n - 1] = b[n - 1] / a[n - 1][n - 1];
  for (k = n - 2; k >= 0; k--)
  {
    if (a[k][k] == 0.0)
    {                          
      x[k] = 0.0;
    }
    else
    {
      s = b[k];
      for (j = k + 1; j < n; j++)
        s -= a[j][k] * x[j];
      x[k] = s / a[k][k];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Calculation of mean squared error
*
* \param[in] *image1              % pointer to first image data
* \param[in] *image2              % pointer to second image data
* \param[in] size                 % Size of image data
*
* \return mse
*/
//-----------------------------------------------------------------------------
double WienerFilter::get_mse(unsigned char *image1,unsigned char *image2, int size)
{
  int    i;
  double mse=0.0;
  int    tmp;
  unsigned char *i1=image1;
  unsigned char *i2=image2;
  for (i=0;i<size;i++)
  {
    tmp=(*i1++)-(*i2++);
    mse+=(double)(tmp*tmp);
  }
  return mse/(double)size;
}

void WienerFilter::encode_params(EntropyEncoder *eenc)
{
  int c;
  
  for (c = 0; c < 3; c++)
  {
    eenc->encode_wlf_flag(&wlf_flag[c]);
    if (wlf_flag[c])
    {
      eenc->encode_wlf_filtertype(&filtertype[c]);
      eenc->encode_wlf_coeffs(coeffs, filter_length_RD_rec_x, filter_length_RD_rec_y, filter_length_RD_pred_x, filter_length_RD_pred_y, filter_length_RD_qpe_x, filter_length_RD_qpe_y, c, filter_precision, filter_precision_table);
    }
  }  
}

void WienerFilter::decode_params(EntropyDecoder *edec)
{
  int c;
  
  for (c = 0; c < 3; c++)
  {
    edec->decode_wlf_flag(&wlf_flag[c]);
    if (wlf_flag[c])
    {
      edec->decode_wlf_filtertype(&filtertype[c]);  
      if (0 == filtertype[c] || 1 == filtertype[c])//2D
      {
        edec->decode_wlf_hint_2D(coeffs, filter_length_RD_rec_x, filter_length_RD_rec_y, filter_length_RD_pred_x, filter_length_RD_pred_y, filter_length_RD_qpe_x, filter_length_RD_qpe_y, c, filter_precision, filter_precision_table);
      }
      else
      {
        edec->decode_wlf_hint_1D(coeffs, filter_length_RD_rec_x, filter_length_RD_rec_y, filter_length_RD_pred_x, filter_length_RD_pred_y, filter_length_RD_qpe_x, filter_length_RD_qpe_y, c, filter_precision, filter_precision_table);
      }
    }
  }  
}

void WienerFilter::estimate(Frame *in, Frame *orig, Frame *pred, Frame *qpe_m, Frame *qpe_s, double lambdaY, double lambdaC, EntropyEncoder *eenc, Param *param)
{
  int c;
  
  for (c = 0; c < 3; c++)
  {
    unsigned char *p_in = (c == 0) ? in->luma : in->crcb[c-1];
    unsigned char *p_orig = (c == 0) ? orig->luma : orig->crcb[c-1];
    
    wlf_flag[c] = param->wlf_enable[c];
    
    if (1 == wlf_flag[c])
    {
      assert(0 == filtertype[c] && (filter_length_RD_rec_x[c] == filter_length_RD_rec_y[c]) && (filter_length_RD_pred_x[c] == filter_length_RD_pred_y[c]) && (filter_length_RD_qpe_x[c] == filter_length_RD_qpe_y[c]));
      unsigned char *p_pred = (c == 0) ? pred->luma : pred->crcb[c-1];
      unsigned char *p_qpe_m = (c == 0) ? qpe_m->luma : qpe_m->crcb[c-1];
      unsigned char *p_qpe_s = (c == 0) ? qpe_s->luma : qpe_s->crcb[c-1];
     
      set_lambda((c == 0) ? lambdaY : lambdaC);
      estimate_filter_2D_equal_size(p_orig, p_in, p_pred, p_qpe_m, p_qpe_s, eenc, c);
    }    
  }
}

void WienerFilter::filter(Frame *in, Frame *out, Frame *pred, Frame *qpe_m, Frame *qpe_s)
{
  int c;
  
  for (c = 0; c < 3; c++)
  {
    unsigned char *p_in = (c == 0) ? in->luma : in->crcb[c-1];
    unsigned char *p_out = (c == 0) ? out->luma : out->crcb[c-1];

    if (1 == wlf_flag[c])
    {
      assert(0 == filtertype[c] && (filter_length_RD_rec_x[c] == filter_length_RD_rec_y[c]) && (filter_length_RD_pred_x[c] == filter_length_RD_pred_y[c]) && (filter_length_RD_qpe_x[c] == filter_length_RD_qpe_y[c]));

      unsigned char *p_pred = (c == 0) ? pred->luma : pred->crcb[c-1];
      unsigned char *p_qpe_m = (c == 0) ? qpe_m->luma : qpe_m->crcb[c-1];
      unsigned char *p_qpe_s = (c == 0) ? qpe_s->luma : qpe_s->crcb[c-1];
      
      filter_2D_equal_size(p_in, p_out, p_pred, p_qpe_m, p_qpe_s, c);
    }
    else  
    {
      int height = (c == 0) ? in->height : in->height2;
      int width = (c == 0) ? in->width : in->width2;
      
      memcpy(p_out, p_in, height * width * sizeof(unsigned char));
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Allocate 2D memory array -> int array2D[rows][columns]
*
* \param[in,out] array2D Pointer to allocate memory
* \param[in] rows number of vertical pel
* \param[in] columns number of horizontal pel
*
* \return memory size in bytes
*/
//-----------------------------------------------------------------------------
int WienerFilter::get_mem2Dint(int ***array2D, int rows, int columns)
{
  int i;

  if((*array2D      = (int**)calloc(rows,        sizeof(int*))) == NULL)
    no_mem_exit("get_mem2Dint: array2D");
  if(((*array2D)[0] = (int* )calloc(rows*columns,sizeof(int ))) == NULL)
    no_mem_exit("get_mem2Dint: array2D");

  for(i=1 ; i<rows ; i++)
    (*array2D)[i] =  (*array2D)[i-1] + columns  ;

  return rows*columns*sizeof(int);
}

//-----------------------------------------------------------------------------
/*!
* \brief free 2D memory array
*
* which was allocated with get_mem2Dint()
*
* \param[in] array2D Pointer to allocate memory
*/
//-----------------------------------------------------------------------------
void WienerFilter::free_mem2Dint(int **array2D)
{
  if (array2D)
  {
    if (array2D[0]) 
      free (array2D[0]);
    else exit(0);

    free (array2D);

  } 
  else
  {
    exit(0);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief  Exit program if memory allocation failed (using error())
*
* \param[in] where string indicating which memory allocation failed
*/
//-----------------------------------------------------------------------------
void WienerFilter::no_mem_exit(const char *where)
{
  printf ("Memory error in %s\n", where);
  exit(0);
}

//-----------------------------------------------------------------------------
/*!
* \brief Allocate 1D memory array -> double array1D[columns]
*
* \param[in,out] array1D Pointer to allocate memory
* \param[in] rows number of vertical pel
*
* \return memory size in bytes
*/
//-----------------------------------------------------------------------------
int WienerFilter::get_mem1Ddouble(double **array1D, int raws)
{
  int j;

  if((*array1D      = (double*)calloc(raws, sizeof(double))) == NULL)
    no_mem_exit("get_mem1Ddouble: array3D");

  for(j = 0; j < raws; j++)
    (*array1D)[j] = 0.0;
  return sizeof(double)*raws;
}

//-----------------------------------------------------------------------------
/*!
* \brief Allocate 2D short memory array -> double array2D[rows][columns]
*
* \param[in,out] array2D Pointer to allocate memory
* \param[in] rows number of vertical pel
* \param[in] columns number of horizontal pel
*
* \return memory size in bytes
*/
//-----------------------------------------------------------------------------
int WienerFilter::get_mem2Ddouble(double ***array2D, int rows, int columns)
{
  int i,j;

  if((*array2D      = (double**)calloc(rows, sizeof(double*))) == NULL)
    no_mem_exit("get_mem2Ddouble: array2D");

  if(((*array2D)[0] = (double* )calloc(columns*rows,sizeof(double ))) == NULL)
    no_mem_exit("get_mem3Ddouble: array2D");

  for(i=1;i<rows;i++)
    (*array2D)[i] = (*array2D)[i-1] + columns ;

  for(i = 0; i < rows; i++)
    for(j = 0; j < columns; j++)
      (*array2D)[i][j] = 0.0;
  return sizeof(double)*rows*columns;
}

//-----------------------------------------------------------------------------
/*!
* \brief free 1D short memory array
*
* which was allocated with get_mem1Ddouble
*
* \param[in] array1D Pointer to allocate memory
*/
//-----------------------------------------------------------------------------
void WienerFilter::free_mem1Ddouble(double *array1D)
{
  if (array1D)
    free (array1D);
  else
    no_mem_exit("free_mem1Ddouble: array1D");
}

//-----------------------------------------------------------------------------
/*!
* \brief free 2D short memory array
*
* which was allocated with get_mem2Ddouble()
*
* \param[in] array2D Pointer to allocate memory
*/
//-----------------------------------------------------------------------------
void WienerFilter::free_mem2Ddouble(double **array2D)
{

  if (array2D)
  {
    if (array2D[0])
      free (array2D[0]);
    else
      no_mem_exit("free_mem2Ddouble: array2D");

    free (array2D);
  }
  else
  {
    printf("free_mem2DDouble: trying to free unused memory");
    exit(-1);
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Allocate 1D memory array -> int array1D[columns]
*
* \param[in,out] array1D Pointer to allocate memory
* \param[in] rows number of vertical pel
*
* \return memory size in bytes
*/
//-----------------------------------------------------------------------------
int WienerFilter::get_mem1Dint(int **array1D, int raws)
{
  int j;

  if((*array1D      = (int*)calloc(raws, sizeof(int))) == NULL)
    no_mem_exit("get_mem1Dint: array3D");

  for(j = 0; j < raws; j++)
    (*array1D)[j] = 0;
  return sizeof(int)*raws;
}

//-----------------------------------------------------------------------------
/*!
* \brief free 1D int memory array
*
* which was allocated with get_mem1Dint
*
* \param[in] array1D Pointer to allocate memory
*/
//-----------------------------------------------------------------------------
void WienerFilter::free_mem1Dint(int *array1D)
{
  if (array1D)
    free (array1D);
  else
    no_mem_exit("free_mem1Dint: array1D");
}
//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------
