/* ====================================================================================================================

  The copyright in this software is being made available under the license included below.

  Copyright (c) 2010, France Telecom S.A., NTT Corp, NTT DOCOMO, Inc., DOCOMO Communications Laboratories USA, Inc.,
  Panasonic Corp., Technicolor S.A. All rights reserved.

  By receiving and/or using this software, you agree to be bound by the following license. Subject to the terms of
  this agreement, this license is a non-exclusive license to distribute and use the software, in source and binary
  forms, with or without modification, permitted only for the purpose of developing standards within the Joint
  Collaborative Team on Video Coding and for testing and promoting such standards. The license is subject to the
  following conditions:

    * Distributions of source code must retain the above copyright notice, this list of conditions and the following
      disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions, and the
      following disclaimer in the documentation and/or other materials provided with such distribution.
    * Neither the name of France Telecom S.A., NTT Corp, NTT DOCOMO, Inc.,
      DOCOMO Communications Laboratories USA, Inc., Panasonic Corp., Technicolor S.A nor the names of their
      contributors may be used to endorse or promote products derived from this software without specific prior
      written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, EQUIPMENT OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THIS SOFTWARE MAY BE SUBJECT TO OTHER THIRD PARTY AND CONTRIBUTOR RIGHTS, INCLUDING PATENT RIGHTS, AND NO SUCH
  RIGHTS ARE GRANTED UNDER THIS LICENSE. ALL USERS AGREE TO INDEMNIFY AND HOLD HARMLESS THE ABOVE MENTIONED PARTIES
  TO THE COPYRIGHT AGAINST ANY AND ALL CLAIMS OR DEMANDS, INCLUDING WITHOUT LIMITATION INTELLECTUAL PROPERTY RIGHTS
  OF THIRD PARTIES, WHICH MIGHT ARISE UNDER THE USE OR DISTRIBUTION OF THIS SOFTWARE.

 * ====================================================================================================================
*/

//-----------------------------------------------------------------------------
/*!
* \file ZeroTree.cpp
*
* \brief Implementation of ZeroTree class
*
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Include headers
//-----------------------------------------------------------------------------

#include "Param.H"
#include "ZeroTree.H"

//-----------------------------------------------------------------------------
// Constructor/destructor
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Constructor
*
* Constructs a zerotree based on the pattern read from a zerotree configuration
* file.
*
* \param[in] pattern Data pattern read from zerotree configuration file
*/
//-----------------------------------------------------------------------------
#ifdef FRAME_SUBSTREAMS
ZeroTree::ZeroTree(int *pattern, int do_reset, int ctx0)
#else
ZeroTree::ZeroTree(int *pattern, int do_reset)
#endif
{
  int i, j;

  nleaves = pattern[0];

  size = TWO * nleaves - ONE;

  nodes = new ZTNode[size];
  leaf_index = new int[nleaves];
  m_structure = new int[size];

  for (i = ZERO; i < size; i++) 
  {
    nodes[i].id = i - nleaves;
  m_structure[i] = pattern[i + 1];
  }

  for (i = ZERO; i < nleaves; i++)
    leaf_index[i] = pattern[i + size + ONE];

  tree_idx  = ONE;      // tree indices start at position 1 of pattern
  coeff_idx = size+ONE; // coeff indices start at position size+1 of pattern
  node_idx  = nleaves;  // int nodes come after lead nodes

  root = grow_children(pattern);

  for (i = ZERO; i < TWO; i++)
    for (j = ZERO; j < TWO; j++) 
    {
      bit_costi[i][j] = new int[nleaves - ONE];
    }

    i_prob = new int[2*nleaves-2];

    for (i=0; i<2*(nleaves-1); i++)
      i_prob[i] = pattern[3*nleaves+i];

#ifdef FRAME_SUBSTREAMS
    local_pe = new SimpleProbEst[2*(nleaves-1)];
#else
    pe = new SimpleProbEst[2*(nleaves-1)];
#endif

    // Do nothing here -- we will read the probas in later,
    // and then do a reset_bit_costs().
    if (do_reset)
      reset_bit_costs();

#ifdef TRACE_STATS
    stats = new int[power2(nleaves)];
    for (i=0; i<power2(nleaves); i++)
      stats[i] = ZERO;
#endif
#ifdef FRAME_SUBSTREAMS
    first_ctx_id = ctx0;
    n_ctxs = 2*(nleaves-1);
    if (first_ctx_id+n_ctxs > MAX_ZTREE_CTXS)
        abort();
#endif
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy initial probabilities to local areas.
*
* Initial probabilities were originally stored in the pattern[] and copied
* from there to i_prob[].  If we've initialized initial probas using an EntropyModel
* loop, the initial probas are stored directly with the probability.  We copy
* the initial proba from there to both i_prob and pattern to prevent code
* change elsewhere.
*
* \param[cs_src] The EntropyModel into which initial probabilities have been read.
* \param[pattern] Original ztree pattern to initialize.
*/
//-----------------------------------------------------------------------------
#ifdef FRAME_SUBSTREAMS
void ZeroTree::copy_initial_probs_back(EntropyModel *cs_src, int *pattern)
#else
void ZeroTree::copy_initial_probs_back(EntropyModel * /*cs_src*/, int *pattern)
#endif
{
  int i;
#ifdef FRAME_SUBSTREAMS
  // Probas are stored directly in EntropyModel if first_ctx_id >= 0 (decoder)
  // otherwise in local_pe (encoder).
  SimpleProbEst *src = get_pe(cs_src);
  for (i = 0; i < 2*(nleaves-1); i++)
  {
    i_prob[i] = src[i].i_p;
    if (src != local_pe)
      local_pe[i].i_p = src[i].i_p;
    pattern[3*pattern[0]+i] = i_prob[i];
  }
#else
  for (i = 0; i < 2*(nleaves-1); i++)
  {
    i_prob[i] = pe[i].i_p;
    pattern[3*pattern[0]+i] = i_prob[i];
  }
#endif

  reset_bit_costs();
}

#ifdef FRAME_SUBSTREAMS
void ZeroTree::copy_initial_probs_from(ZeroTree *zt_src, EntropyModel *cs_src)
#else
void ZeroTree::copy_initial_probs_from(ZeroTree *zt_src, EntropyModel * /*cs_src*/)
#endif
{
  int i;
  
#ifdef FRAME_SUBSTREAMS
  // Probas are stored directly in EntropyModel if first_ctx_id >= 0 (decoder)
  // otherwise in local_pe (encoder).
  SimpleProbEst *src = zt_src->get_pe(cs_src);
  SimpleProbEst *dst = get_pe(cs_src);

  for (i = 0; i < 2*(nleaves-1); i++)
  {
    i_prob[i] = src[i].i_p;
    dst[i].i_p = src[i].i_p;
    if (dst != local_pe)
      local_pe[i].i_p = src[i].i_p;
  }
#else
  for (i = 0; i < 2*(nleaves-1); i++)
  {
    pe[i].i_p = zt_src->pe[i].i_p;
    i_prob[i] = zt_src->pe[i].i_p;
  }
#endif

  reset_bit_costs();
}

//-----------------------------------------------------------------------------
/*!
* \brief Destructor
*
*/
//-----------------------------------------------------------------------------
ZeroTree::~ZeroTree()
{
  int i,j;

  delete[] nodes;
  delete[] m_structure;

  for (i = ZERO; i < TWO; i++)
    for (j = ZERO; j < TWO; j++) 
    {
      delete[] bit_costi[i][j];
    }

    delete[] i_prob;
#ifdef FRAME_SUBSTREAMS
    delete[] local_pe;
#else
    delete[] pe;
#endif
#ifdef TRACE_STATS
    delete[] stats;
#endif
    delete[] leaf_index;
}

//-----------------------------------------------------------------------------
// Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
/*!
* \brief Add a node to zerotree and further grow tree if needed
*
* Is called during the zerotree construction phase. Adds a node and further
* grows children nodes if not a leaf node.
* 
* \param[in] pattern Data pattern read from zerotree configuration file
* \return Root node
*/
//-----------------------------------------------------------------------------
ZTNode *ZeroTree::grow_children(int *pattern) 
{
  ZTNode *node;

  if (pattern[tree_idx++] == 0)
    node = &nodes[pattern[coeff_idx++]];
  else 
  {
    node = &nodes[node_idx++];
    node->left = grow_children(pattern);
    node->right = grow_children(pattern);
  }

  return node;
}

#ifdef FRAME_SUBSTREAMS
int ZeroTree::cost_nodes(Block *blk, ArithEncoder *aenc, EntropyModel *cs)
#else
int ZeroTree::cost_nodes(Block *blk, ArithEncoder *aenc)
#endif
{
  int i;
  int num = blk->get_num_elements();

  for (i = 0; i < num; i++)
  {
    nodes[i].value = (blk->val[i] != 0);
  }
  
  int cost = 0;
  for (i = num - 2; i >= 0; i--)
  {
    int lval = nodes[num + i].left->value;
    int rval = nodes[num + i].right->value;
    int val = lval | rval;
    nodes[num + i].value = val;
    
    if (val)
    {
      int both = lval & rval;

#ifdef FRAME_SUBSTREAMS
      cost += aenc->cost(both, &get_pe(cs)[2 * i]);
#else
      cost += aenc->cost(both, &pe[2 * i]);
#endif
      
      if (!both)
#ifdef FRAME_SUBSTREAMS
        cost += aenc->cost(lval, &get_pe(cs)[2 * i + 1]);
#else
        cost += aenc->cost(lval, &pe[2 * i + 1]);
#endif
    }
  }
  
  return cost;
}

//-----------------------------------------------------------------------------
/*!
* \brief Update values of zerotree leaves
*
* Initially called with the root node as argument. Recursively updates values
* of leaf nodes.
* 
* \param[in] node Current node
* \param[in] val  Values of a sub-tree
*/
//-----------------------------------------------------------------------------
void ZeroTree::update_values(Block *val)
{
  int i;
  for (i = 0; i < nleaves; i++)
  {
    nodes[i].value = (val->val[i] != 0);
  }

  for (i = 2*nleaves - 2; i >= nleaves; i--)
  {
    nodes[i].value = nodes[i].left->value | nodes[i].right->value;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Resets rate costs
*
* Computes rate costs for coding elements at each node.
*/
//-----------------------------------------------------------------------------
void ZeroTree::reset_bit_costs() 
{
  int j,k;

  for (j = ZERO; j < nleaves - ONE; j++) 
  {
    for (k = ZERO; k < TWO; k++) 
    {
      bit_costi[k][0][j] = param.bitsM[0][i_prob[2*j+k]];
      bit_costi[k][1][j] = param.bitsM[1][i_prob[2*j+k]];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Updates rate costs
*
* Computes rate costs for coding elements at each node.
*/
//-----------------------------------------------------------------------------
#ifdef FRAME_SUBSTREAMS
void ZeroTree::update_bit_costs(SimpleProbEst *pe)
#else
void ZeroTree::update_bit_costs() 
#endif
{
  int j,k;

  for (j = ZERO; j < nleaves - ONE; j++) 
  {
    for (k = ZERO; k < TWO; k++) 
    {
      bit_costi[k][0][j] = param.bitsM[0][pe[2*j+k].p];
      bit_costi[k][1][j] = param.bitsM[1][pe[2*j+k].p];
    }
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Computes cost for coding a node and its children
*
* When node is the root node, returns the cost for coding entire tree.
*
* \param[in] node Node for which to compute the cost
* \return Cost for coding node
*/
//-----------------------------------------------------------------------------
int ZeroTree::get_ztcost(ZTNode *node) 
{
  int mode;
  int r;

  if (node->id < 0 || node->value == 0) 
  {
    return 0;
  }
  else 
  {
    r =  get_ztcost(node->left);
    r += get_ztcost(node->right);

    mode = 2 * node->left->value + node->right->value;

    switch (mode) 
    {
    case 0:
      printf("!!!!\n");
      break;
    case 1:
      r += bit_costi[0][0][node->id];
      r += bit_costi[1][0][node->id];
      break;
    case 2:
      r += bit_costi[0][0][node->id];
      r += bit_costi[1][1][node->id];
      break;
    case 3:
      r += bit_costi[0][1][node->id];
      break;
    }

    return r;
  }
}

//-----------------------------------------------------------------------------
/*!
* \brief Copy the coding state
*
* Copies the coding state from the specified source object to the current object
*
* \param[in,out] ztree Pointer to specified source object
*/
//-----------------------------------------------------------------------------
void ZeroTree::copy_coding_state(ZeroTree  *ztree)
{

  this->tree_idx  = ztree->tree_idx;
  this->coeff_idx = ztree->coeff_idx;
  this->node_idx  = ztree->node_idx;
  this->nleaves = ztree->nleaves;
  memcpy(this->i_prob, ztree->i_prob, 2 * (ztree->nleaves - 1) * sizeof (int));
#ifdef FRAME_SUBSTREAMS
  // In decoding environment, contexts are stored in the global entropymodel and have already
  // been copied.
  // Otherwise (in the encoder), we copy them here.
  if (first_ctx_id < 0)
    memcpy(this->local_pe,     ztree->local_pe,     2 * (ztree->nleaves - 1) * sizeof (SimpleProbEst));
#else
  memcpy(this->pe,     ztree->pe,     2 * (ztree->nleaves - 1) * sizeof (SimpleProbEst));
#endif
}

//-----------------------------------------------------------------------------
// End of file
//-----------------------------------------------------------------------------

